/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 07 1.10.1.3 source/albert/graph2d/cag.cpp, 2d, ioc.v400, 000302 
/*================
||
||  File:   CAG.C
||
||  What:   Implementation for CAG nodes
||
||  Change History:
||
*/

#include <cag.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
#define _CAG_CPP_
#include <ipagetun.h>
#endif

// #include <istring.hpp>

//===================//
// ICAGNode routines //
//===================//


/*================
||
||  ICAGNode constructor
||
*/

ICAGNode::ICAGNode ()
{
}

ICAGNode::ICAGNode (const ICAGNode&)
{
}


/*================
||
||  ICAGNode destructor
||
*/

ICAGNode::~ICAGNode ()
{
}


/*================
||
||  ICAGNode assignment operator
||
*/

ICAGNode& ICAGNode::operator= (const ICAGNode&)
{
    return *this;
}


/*================
||
||  ICAGNode transition type request
||
*/

ICAGNode::TransitionType ICAGNode::transitionType () const
{
    return kFill;   // default type is kFill
}


/*================
||
||  ICAGNode type request
||
*/

ICAGNode::CAGType ICAGNode::type () const
{
    return kUndefined;
}


/*================
||
||  ICAGNode static Sign function
||
*/

ICAGNode::WrapCount ICAGNode::Sign (ICAGNode::WrapCount x)
{
    if (x > 0)
        return 1;
    else if (x == 0)
        return 0;
    else
        return -1;
}


//====================//
// ICAGChild routines //
//====================//


/*================
||
||  ICAGChild constructor
||
*/

ICAGChild::ICAGChild (ICAGNode* parent)
    : fParent (parent)
{
}


/*================
||
||  ICAGChild destructor
||
*/

ICAGChild::~ICAGChild ()
{
}


//====================//
// ICAGUnary routines //
//====================//


/*================
||
||  ICAGUnary constructor
||
*/

ICAGUnary::ICAGUnary (ICAGNode* parent)
    : ICAGChild (parent),
      fLeftFrameLink (0 ),
      fRightFrameLink (0 )
{
}


/*================
||
||  ICAGUnary destructor
||
*/

ICAGUnary::~ICAGUnary ()
{
    delete fLeftFrameLink;
    delete fRightFrameLink;
}


/*================
||
||  ICAGUnary frame link functions
||
*/

ICAGLeftFrameLink* ICAGUnary::leftFrameLink ()
{
    if (fLeftFrameLink == 0 )
        fLeftFrameLink = new ICAGLeftFrameLink (this);
    return fLeftFrameLink;
}

ICAGRightFrameLink* ICAGUnary::rightFrameLink ()
{
    if (fRightFrameLink == 0 )
        fRightFrameLink = new ICAGRightFrameLink (this);
    return fRightFrameLink;
}


//======================//
// ICAGEvenOdd routines //
//======================//


/*================
||
||  ICAGEvenOdd constructor
||
*/

ICAGEvenOdd::ICAGEvenOdd (ICAGNode* parent)
    : ICAGUnary (parent)
{
    fFillCount = 0;
    fLeftFrameCount = 0;
    fLeftFrameCountNorm = 0;
    fRightFrameCount = 0;
    fRightFrameCountNorm = 0;

}


/*================
||
||  ICAGEvenOdd destructor
||
*/

ICAGEvenOdd::~ICAGEvenOdd ()
{
}


/*================
||
||  ICAGEvenOdd transition functions
||
*/

void ICAGEvenOdd::transition (ICAGNode::WrapCount fill)
{
    fFillCount += fill;
    parent()->transition (fFillCount & 1);
}

void ICAGEvenOdd::transitionLeftFrame (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCount += leftFrame;
    evaluate();
}

void ICAGEvenOdd::transitionRightFrame (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCount += rightFrame;
    evaluate();
}

void ICAGEvenOdd::evaluate ()
{
    WrapCount oldLeftFrameCount = fLeftFrameCountNorm;
    WrapCount oldRightFrameCount = fRightFrameCountNorm;

    if (fLeftFrameCount != fRightFrameCount)    // in the frame?
    {
        fLeftFrameCountNorm = 1;
        fRightFrameCountNorm = 0;
    }
    else if (fLeftFrameCount)   // inside?
    {
        fLeftFrameCountNorm = 1;
        fRightFrameCountNorm = 1;
    }
    else    // outside
    {
        fLeftFrameCountNorm = 0;
        fRightFrameCountNorm = 0;
    }

    if (oldLeftFrameCount != fLeftFrameCountNorm)
        parent()->transitionLeftFrame (fLeftFrameCountNorm);
    if (oldRightFrameCount != fRightFrameCountNorm)
        parent()->transitionRightFrame (fRightFrameCountNorm);
}


/*================
||
||  ICAGNode type request
||
*/

ICAGNode::CAGType ICAGEvenOdd::type () const
{
    return kEvenOdd;
}


//======================//
// ICAGWindNum routines //
//======================//


/*================
||
||  ICAGWindNum constructor
||
*/

ICAGWindNum::ICAGWindNum (ICAGNode* parent)
    : ICAGUnary (parent)
{
    fFillCount = 0;
    fLeftFrameCount = 0;
    fLeftFrameCountNorm = 0;
    fRightFrameCount = 0;
    fRightFrameCountNorm = 0;
}


/*================
||
||  ICAGWindNum destructor
||
*/

ICAGWindNum::~ICAGWindNum ()
{
}


/*================
||
||  ICAGWindNum transition functions
||
*/

void ICAGWindNum::transition (ICAGNode::WrapCount fill)
{
    ICAGNode::WrapCount oldFillCount = fFillCount;
    fFillCount += fill;

    if (!(oldFillCount * fFillCount))       // count has transitioned to/from zero?
        parent()->transition (fFillCount ? 1 : 0);
}

void ICAGWindNum::transitionLeftFrame (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCount += leftFrame;
    evaluate();
}

void ICAGWindNum::transitionRightFrame (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCount += rightFrame;
    evaluate();
}

void ICAGWindNum::evaluate ()
{
    WrapCount oldLeftFrameCount = fLeftFrameCountNorm;
    WrapCount oldRightFrameCount = fRightFrameCountNorm;

    if (Sign (fLeftFrameCount) != Sign (fRightFrameCount))  // in the frame?
    {
        fLeftFrameCountNorm = 1;
        fRightFrameCountNorm = 0;
    }
    else if (fLeftFrameCount)   // inside?
    {
        fLeftFrameCountNorm = 1;
        fRightFrameCountNorm = 1;
    }
    else    // outside
    {
        fLeftFrameCountNorm = 0;
        fRightFrameCountNorm = 0;
    }

    if (oldLeftFrameCount != fLeftFrameCountNorm)
        parent()->transitionLeftFrame (fLeftFrameCountNorm);
    if (oldRightFrameCount != fRightFrameCountNorm)
        parent()->transitionRightFrame (fRightFrameCountNorm);
}


/*================
||
||  ICAGNode type request
||
*/

ICAGNode::CAGType ICAGWindNum::type () const
{
    return kWindNum;
}


//============================//
// ICAGLeftFrameLink routines //
//============================//


/*================
||
||  ICAGLeftFrameLink constructor
||
*/

ICAGLeftFrameLink::ICAGLeftFrameLink (ICAGNode* parent)
    : ICAGUnary (parent)
{
}


/*================
||
||  ICAGLeftFrameLink destructor
||
*/

ICAGLeftFrameLink::~ICAGLeftFrameLink ()
{
}


/*================
||
||  ICAGLeftFrameLink transition function
||
*/

void ICAGLeftFrameLink::transition (ICAGNode::WrapCount leftFrame)
{
    parent()->transitionLeftFrame (leftFrame);
}


/*================
||
||  ICAGLeftFrameLink transition type request
||
*/

ICAGNode::TransitionType ICAGLeftFrameLink::transitionType () const
{
    return kLeftFrame;
}


/*================
||
||  ICAGNode type request
||
*/

ICAGNode::CAGType ICAGLeftFrameLink::type () const
{
    return ((ICAGLeftFrameLink *)this)->parent()->type();
}


//=============================//
// ICAGRightFrameLink routines //
//=============================//


/*================
||
||  ICAGRightFrameLink constructor
||
*/

ICAGRightFrameLink::ICAGRightFrameLink (ICAGNode* parent)
    : ICAGUnary (parent)
{
}


/*================
||
||  ICAGRightFrameLink destructor
||
*/

ICAGRightFrameLink::~ICAGRightFrameLink ()
{
}


/*================
||
||  ICAGRightFrameLink transition function
||
*/

void ICAGRightFrameLink::transition (ICAGNode::WrapCount rightFrame)
{
    parent()->transitionRightFrame (rightFrame);
}


/*================
||
||  ICAGRightFrameLink transition type request
||
*/

ICAGNode::TransitionType ICAGRightFrameLink::transitionType () const
{
    return kRightFrame;
}


/*================
||
||  ICAGNode type request
||
*/

ICAGNode::CAGType ICAGRightFrameLink::type () const
{
    return ((ICAGRightFrameLink *)this)->parent()->type();
}


//========================//
// ICAGRightLink routines //
//========================//


/*================
||
||  ICAGRightLink constructor
||
*/

ICAGRightLink::ICAGRightLink (ICAGBinary* parent)
    : fParent (parent)
{
}


/*================
||
||  ICAGRightLink destructor
||
*/

ICAGRightLink::~ICAGRightLink ()
{
}


/*================
||
||  ICAGRightLink transition functions
||
*/

void ICAGRightLink::transition (ICAGNode::WrapCount fill)
{
    fParent->transitionRightSide (fill);
}

void ICAGRightLink::transitionLeftFrame (ICAGNode::WrapCount leftFrame)
{
    fParent->transitionLeftFrameRightSide (leftFrame);
}

void ICAGRightLink::transitionRightFrame (ICAGNode::WrapCount rightFrame)
{
    fParent->transitionRightFrameRightSide (rightFrame);
}


//=====================//
// ICAGBinary routines //
//=====================//


/*================
||
||  ICAGBinary constructor
||
*/

ICAGBinary::ICAGBinary (ICAGNode* parent)
    : ICAGChild (parent), fRightLink (this)
{
}


/*================
||
||  ICAGBinary destructor
||
*/

ICAGBinary::~ICAGBinary ()
{
}


/*================
||
||  ICAGBinary right link function
||
*/

ICAGRightLink* ICAGBinary::rightLink ()
{
    return &fRightLink;
}


//==================//
// ICAGAdd routines //
//==================//


/*================
||
||  ICAGAdd constructor
||
*/

ICAGAdd::ICAGAdd (ICAGNode* parent)
    : ICAGBinary (parent)
{
    fFillCount = 0;
    fLeftFrameCount = 0;
    fRightFrameCount = 0;

    fFillCountLeftSide = 0;
    fLeftFrameCountLeftSide = 0;
    fRightFrameCountLeftSide = 0;

    fFillCountRightSide = 0;
    fLeftFrameCountRightSide = 0;
    fRightFrameCountRightSide = 0;
}


/*================
||
||  ICAGAdd destructor
||
*/

ICAGAdd::~ICAGAdd ()
{
}


/*================
||
||  ICAGAdd transition functions
||
*/

void ICAGAdd::transition (ICAGNode::WrapCount fill)
{
    ICAGNode::WrapCount oldFillCount = fFillCount;
    fFillCountLeftSide = fill;
    fFillCount = (fFillCountLeftSide || fFillCountRightSide);

    if (oldFillCount != fFillCount)     // count has changed?
        parent()->transition (fFillCount);
}

void ICAGAdd::transitionRightSide (ICAGNode::WrapCount fill)
{
    ICAGNode::WrapCount oldFillCount = fFillCount;
    fFillCountRightSide = fill;
    fFillCount = (fFillCountLeftSide || fFillCountRightSide);

    if (oldFillCount != fFillCount)     // state has changed?
        parent()->transition (fFillCount);
}

void ICAGAdd::transitionLeftFrame (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCountLeftSide = leftFrame;

    ICAGNode::WrapCount oldLeftFrameCount = fLeftFrameCount;
    fLeftFrameCount = Sign (fLeftFrameCountLeftSide + fLeftFrameCountRightSide);

    if (oldLeftFrameCount != fLeftFrameCount)
        parent()->transitionLeftFrame (fLeftFrameCount);
}

void ICAGAdd::transitionRightFrame (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCountLeftSide = rightFrame;

    ICAGNode::WrapCount oldRightFrameCount = fRightFrameCount;
    fRightFrameCount = Sign (fRightFrameCountLeftSide + fRightFrameCountRightSide);

    if (oldRightFrameCount != fRightFrameCount)
        parent()->transitionRightFrame (fRightFrameCount);
}

void ICAGAdd::transitionLeftFrameRightSide (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCountRightSide = leftFrame;

    ICAGNode::WrapCount oldLeftFrameCount = fLeftFrameCount;
    fLeftFrameCount = Sign (fLeftFrameCountLeftSide + fLeftFrameCountRightSide);

    if (oldLeftFrameCount != fLeftFrameCount)
        parent()->transitionLeftFrame (fLeftFrameCount);
}

void ICAGAdd::transitionRightFrameRightSide (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCountRightSide = rightFrame;

    ICAGNode::WrapCount oldRightFrameCount = fRightFrameCount;
    fRightFrameCount = Sign (fRightFrameCountLeftSide + fRightFrameCountRightSide);

    if (oldRightFrameCount != fRightFrameCount)
        parent()->transitionRightFrame (fRightFrameCount);
}


//=======================//
// ICAGSubtract routines //
//=======================//


/*================
||
||  ICAGSubtract constructor
||
*/

ICAGSubtract::ICAGSubtract (ICAGNode* parent)
    : ICAGBinary (parent)
{
    fFillCount = 0;
    fLeftFrameCount = 0;
    fRightFrameCount = 0;

    fFillCountLeftSide = 0;
    fLeftFrameCountLeftSide = 0;
    fRightFrameCountLeftSide = 0;

    fFillCountRightSide = 0;
    fLeftFrameCountRightSide = 0;
    fRightFrameCountRightSide = 0;
}


/*================
||
||  ICAGSubtract destructor
||
*/

ICAGSubtract::~ICAGSubtract ()
{
}


/*================
||
||  ICAGSubtract transition functions
||
*/

void ICAGSubtract::transition (ICAGNode::WrapCount value)
{
    ICAGNode::WrapCount oldFillCount = fFillCount;
    fFillCountLeftSide = value;
    fFillCount = fFillCountLeftSide && !fFillCountRightSide;

    if (oldFillCount != fFillCount)     // state has changed?
        parent()->transition (fFillCount);
}

void ICAGSubtract::transitionRightSide (ICAGNode::WrapCount value)
{
    ICAGNode::WrapCount oldFillCount = fFillCount;
    fFillCountRightSide = value;
    fFillCount = fFillCountLeftSide && !fFillCountRightSide;

    if (oldFillCount != fFillCount)     // state has changed?
        parent()->transition (fFillCount);
}

void ICAGSubtract::transitionLeftFrame (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCountLeftSide = leftFrame;
    evaluate();
}

void ICAGSubtract::transitionRightFrame (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCountLeftSide = rightFrame;
    evaluate();
}

void ICAGSubtract::transitionLeftFrameRightSide (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCountRightSide = leftFrame;
    evaluate();
}

void ICAGSubtract::transitionRightFrameRightSide (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCountRightSide = rightFrame;
    evaluate();
}

void ICAGSubtract::evaluate ()
{
    ICAGNode::WrapCount oldLeftFrameCount = fLeftFrameCount;
    ICAGNode::WrapCount oldRightFrameCount = fRightFrameCount;

    // If outside the left or inside the right, evaluate to outside; else treat as even-odd
    if (!(fLeftFrameCountLeftSide || fRightFrameCountLeftSide) ||
        (fLeftFrameCountRightSide && fRightFrameCountRightSide))
    {
        fLeftFrameCount = fRightFrameCount = 0;
    }
    else    // treat it as even-odd
    {
        fLeftFrameCount = fLeftFrameCountLeftSide + fLeftFrameCountRightSide;
        fRightFrameCount = fRightFrameCountLeftSide + fRightFrameCountRightSide;
    }

    if (oldLeftFrameCount != fLeftFrameCount)
        parent()->transitionLeftFrame (fLeftFrameCount);
    if (oldRightFrameCount != fRightFrameCount)
        parent()->transitionRightFrame (fRightFrameCount);
}


//========================//
// ICAGIntersect routines //
//========================//


/*================
||
||  ICAGIntersect constructor
||
*/

ICAGIntersect::ICAGIntersect (ICAGNode* parent)
    : ICAGBinary (parent)
{
    fFillCount = 0;
    fLeftFrameCount = 0;
    fRightFrameCount = 0;

    fFillCountLeftSide = 0;
    fLeftFrameCountLeftSide = 0;
    fRightFrameCountLeftSide = 0;

    fFillCountRightSide = 0;
    fLeftFrameCountRightSide = 0;
    fRightFrameCountRightSide = 0;
}


/*================
||
||  ICAGIntersect destructor
||
*/

ICAGIntersect::~ICAGIntersect ()
{
}


/*================
||
||  ICAGIntersect transition functions
||
*/

void ICAGIntersect::transition (ICAGNode::WrapCount value)
{
    ICAGNode::WrapCount oldFillCount = fFillCount;
    fFillCountLeftSide = value;
    fFillCount = fFillCountLeftSide && fFillCountRightSide;

    if (oldFillCount != fFillCount)     // state has changed?
        parent()->transition (fFillCount);
}

void ICAGIntersect::transitionRightSide (ICAGNode::WrapCount value)
{
    ICAGNode::WrapCount oldFillCount = fFillCount;
    fFillCountRightSide = value;
    fFillCount = fFillCountLeftSide && fFillCountRightSide;

    if (oldFillCount != fFillCount)     // state has changed?
        parent()->transition (fFillCount);
}

void ICAGIntersect::transitionLeftFrame (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCountLeftSide = leftFrame;

    ICAGNode::WrapCount oldLeftFrameCount = fLeftFrameCount;
    fLeftFrameCount = (fLeftFrameCountLeftSide && fLeftFrameCountRightSide);

    if (oldLeftFrameCount != fLeftFrameCount)
        parent()->transitionLeftFrame (fLeftFrameCount);
}

void ICAGIntersect::transitionRightFrame (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCountLeftSide = rightFrame;

    ICAGNode::WrapCount oldRightFrameCount = fRightFrameCount;
    fRightFrameCount = (fRightFrameCountLeftSide && fRightFrameCountRightSide);

    if (oldRightFrameCount != fRightFrameCount)
        parent()->transitionRightFrame (fRightFrameCount);
}

void ICAGIntersect::transitionLeftFrameRightSide (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCountRightSide = leftFrame;

    ICAGNode::WrapCount oldLeftFrameCount = fLeftFrameCount;
    fLeftFrameCount = (fLeftFrameCountLeftSide && fLeftFrameCountRightSide);

    if (oldLeftFrameCount != fLeftFrameCount)
        parent()->transitionLeftFrame (fLeftFrameCount);
}

void ICAGIntersect::transitionRightFrameRightSide (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCountRightSide = rightFrame;

    ICAGNode::WrapCount oldRightFrameCount = fRightFrameCount;
    fRightFrameCount = (fRightFrameCountLeftSide && fRightFrameCountRightSide);

    if (oldRightFrameCount != fRightFrameCount)
        parent()->transitionRightFrame (fRightFrameCount);
}


//==========================//
// ICAGExclusiveOr routines //
//==========================//


/*================
||
||  ICAGExclusiveOr constructor
||
*/

ICAGExclusiveOr::ICAGExclusiveOr (ICAGNode* parent)
    : ICAGBinary (parent)
{
    fFillCount = 0;
    fLeftFrameCount = 0;
    fRightFrameCount = 0;

    fFillCountLeftSide = 0;
    fLeftFrameCountLeftSide = 0;
    fRightFrameCountLeftSide = 0;

    fFillCountRightSide = 0;
    fLeftFrameCountRightSide = 0;
    fRightFrameCountRightSide = 0;
}


/*================
||
||  ICAGExclusiveOr destructor
||
*/

ICAGExclusiveOr::~ICAGExclusiveOr ()
{
}


/*================
||
||  ICAGExclusiveOr transition functions
||
*/

void ICAGExclusiveOr::transition (ICAGNode::WrapCount value)
{
    ICAGNode::WrapCount oldFillCount = fFillCount;
    fFillCountLeftSide = value;
    fFillCount = (fFillCountLeftSide && !fFillCountRightSide) ||
                    (!fFillCountLeftSide && fFillCountRightSide);

    if (oldFillCount != fFillCount)     // state has changed?
        parent()->transition (fFillCount);
}

void ICAGExclusiveOr::transitionRightSide (ICAGNode::WrapCount value)
{
    ICAGNode::WrapCount oldFillCount = fFillCount;
    fFillCountRightSide = value;
    fFillCount = (fFillCountLeftSide && !fFillCountRightSide) ||
                    (!fFillCountLeftSide && fFillCountRightSide);

    if (oldFillCount != fFillCount)     // state has changed?
        parent()->transition (fFillCount);
}

void ICAGExclusiveOr::transitionLeftFrame (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCountLeftSide = leftFrame;

    ICAGNode::WrapCount oldLeftFrameCount = fLeftFrameCount;
    fLeftFrameCount = fLeftFrameCountLeftSide + fLeftFrameCountRightSide;

    if (oldLeftFrameCount != fLeftFrameCount)
        parent()->transitionLeftFrame (fLeftFrameCount);
}

void ICAGExclusiveOr::transitionRightFrame (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCountLeftSide = rightFrame;

    ICAGNode::WrapCount oldRightFrameCount = fRightFrameCount;
    fRightFrameCount = fRightFrameCountLeftSide + fRightFrameCountRightSide;

    if (oldRightFrameCount != fRightFrameCount)
        parent()->transitionRightFrame (fRightFrameCount);
}

void ICAGExclusiveOr::transitionLeftFrameRightSide (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCountRightSide = leftFrame;

    ICAGNode::WrapCount oldLeftFrameCount = fLeftFrameCount;
    fLeftFrameCount = fLeftFrameCountLeftSide + fLeftFrameCountRightSide;

    if (oldLeftFrameCount != fLeftFrameCount)
        parent()->transitionLeftFrame (fLeftFrameCount);
}

void ICAGExclusiveOr::transitionRightFrameRightSide (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCountRightSide = rightFrame;

    ICAGNode::WrapCount oldRightFrameCount = fRightFrameCount;
    fRightFrameCount = fRightFrameCountLeftSide + fRightFrameCountRightSide;

    if (oldRightFrameCount != fRightFrameCount)
        parent()->transitionRightFrame (fRightFrameCount);
}


//===================//
// ICAGRoot routines //
//===================//


/*================
||
||  ICAGRoot constructor
||
*/

ICAGRoot::ICAGRoot ()
{
    fFillCount = 0;
    fLeftFrameCount = 0;
    fRightFrameCount = 0;
}


/*================
||
||  ICAGRoot destructor
||
*/

ICAGRoot::~ICAGRoot ()
{
}


/*================
||
||  ICAGRoot transition functions
||
*/

void ICAGRoot::transition (ICAGNode::WrapCount fill)
{
    fFillCount = fill;
}

void ICAGRoot::transitionLeftFrame (ICAGNode::WrapCount leftFrame)
{
    fLeftFrameCount = leftFrame;
}

void ICAGRoot::transitionRightFrame (ICAGNode::WrapCount rightFrame)
{
    fRightFrameCount = rightFrame;
}


/*================
||
||  ICAGRoot get functions
||
*/

ICAGNode::WrapCount ICAGRoot::fillCount () const
{
    return fFillCount;
}

ICAGNode::WrapCount ICAGRoot::leftFrameCount () const
{
    return fLeftFrameCount;
}

ICAGNode::WrapCount ICAGRoot::rightFrameCount () const
{
    return fRightFrameCount;
}



