/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 44  1.19.1.7  source/albert/graph2d/grafcoll.cpp, 2d, ioc.v400, 000302  
/*================
||
||  File:   GrafCollectibles.C
||
||  What:   Implementation for polymorphic geometries
||
||  Classes defined in this file:
||  * IAreaGeometryHandle
||  *IMAreaGeometry
||  *ITransformedGeometry
||  *IRectGeometry
||  *IRectilinearGeometry
||  *IEllipseGeometry
||  *IPolygonGeometry
||  *ILoopGeometry
||  *IAddGeometry
||  *ISubtractGeometry
||  *IIntersectGeometry
||  *IExclusiveOrGeometry
||
||  Change History:
||
*/

#include <ilanglvl.hpp>
#include <istring.hpp>

#include <igrtypes.hpp>
#include <igbase2d.hpp>
#include <igarea2d.hpp>
#include <igloop2d.hpp>
#include <grafcoll.hpp>
#include <grafextr.hpp>
#include <cag.hpp>
#include <imatrix.hpp>
#include <inumeric.hpp>
#include <grassert.hpp>
#include <grtype.hpp>
#include <idatstrm.hpp>
#include <grstrmod.hpp>
#include <iprimlck.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
#define _GRAFCOLL_CPP_
#include <ipagetun.h>
#endif

//=====================//
// IAreaGeometryHandle //
//=====================//

/*================
||
||  IAreaGeometryHandle constructors
||
*/

StreamableDefinitionsMacro(IAreaGeometryHandle, gGraphicsStreamModule);

IAreaGeometryHandle::IAreaGeometryHandle()
{
    fGeometry = 0 /*NIL*/;
}

IAreaGeometryHandle::IAreaGeometryHandle(IMAreaGeometry* geometry)
{
    fGeometry = geometry;
    if (geometry != 0 /*NIL*/)
        geometry->addReference();
}

IAreaGeometryHandle::IAreaGeometryHandle(const IAreaGeometryHandle& source)
{
    IMAreaGeometry* geometry = source.fGeometry;
    fGeometry = geometry;
    if (geometry != 0 /*NIL*/)
    {

// #ifndef SS
#ifdef NO_OC_APPLICABLE
        // If this object is in a different heap, copy the geometry instead of referencing it;
        // or if this object is on the stack, copy the geometry if it's not in the default heap
        // because the heap may go away(yuck)
        TMemoryHeap* heap = TMemoryHeap::HeapOf(this);
        TMemoryHeap* otherHeap = TMemoryHeap::HeapOf(geometry);
        if (((heap == 0 /*NIL*/) && (otherHeap != TMemoryHeap::GetDefaultHeap())) ||
            ((heap != 0 /*NIL*/) && (heap != otherHeap)))
            fGeometry = ::GraphicCopy(*geometry, *heap);
#endif // NO_OC_APPLICABLE
// #endif // not SS

        fGeometry->addReference();
    }
}

/*================
||
||  IAreaGeometryHandle destructor
||
*/

IAreaGeometryHandle::~IAreaGeometryHandle()
{
    if (fGeometry != 0 /*NIL*/)
        fGeometry->removeReference();
}

/*================
||
||  IAreaGeometryHandle::operator=
||
*/

IAreaGeometryHandle& IAreaGeometryHandle::operator=(const IAreaGeometryHandle& source)
{
    if (fGeometry != source.fGeometry)
    {
        // Always add the new reference before removing the old reference!!!!!!!!!!!!!!
        // Otherwise you might lose the new object!

        IMAreaGeometry* geometry = source.fGeometry;
        IMAreaGeometry* oldGeometry = fGeometry;

        fGeometry = geometry;
        if (geometry != 0 /*NIL*/)
        {

// #ifndef SS
#ifdef NO_OC_APPLICABLE
            // If this object is in a different heap, copy the geometry instead of referencing it;
            // or if this object is on the stack, copy the geometry if it's not in the default heap
            // because the heap may go away(yuck)
            TMemoryHeap* heap = TMemoryHeap::HeapOf(this);
            TMemoryHeap* otherHeap = TMemoryHeap::HeapOf(geometry);
            if (((heap == 0 /*NIL*/) && (otherHeap != TMemoryHeap::GetDefaultHeap())) ||
                ((heap != 0 /*NIL*/) && (heap != otherHeap)))
                fGeometry = ::GraphicCopy(*geometry, *heap);
#endif // NO_OC_APPLICABLE
// #endif // not SS


            fGeometry->addReference();
        }

        if (oldGeometry != 0 /*NIL*/)
            oldGeometry->removeReference();
    }
    return *this;
}

bool IAreaGeometryHandle::operator==(const IAreaGeometryHandle& handle) const
{
    if (fGeometry == handle.fGeometry)
        return true;
    else if ((fGeometry != 0 /*NIL*/) && (handle.fGeometry != 0 /*NIL*/)) {
        if (fGeometry->isEmpty() && handle.fGeometry->isEmpty()) //added for defect 25849
            return true;
        else
            return (*fGeometry == *handle.fGeometry);
    }
    else
        return false;
}

/*================
||
||  IAreaGeometryHandle::operator!=
||
*/

bool IAreaGeometryHandle::operator!=(const IAreaGeometryHandle& handle) const
{
    return !operator==(handle);
}

/*================
||
||  IAreaGeometryHandle::operator<
||
*/

bool IAreaGeometryHandle::operator<(const IAreaGeometryHandle& handle) const
{
    if (fGeometry == handle.fGeometry)
        return false;
    else if (fGeometry == 0 /*NIL*/)
        return true;
    else if (handle.fGeometry == 0 /*NIL*/)
        return false;
    else
        return (*fGeometry < *handle.fGeometry);
}

/*================
||
||  IAreaGeometryHandle::isEmpty
||
*/

bool IAreaGeometryHandle::isEmpty() const
{
    return (fGeometry == 0 /*NIL*/) || (fGeometry->isEmpty());
}

/*================
||
||  IAreaGeometryHandle::setToEmpty
||
*/

void IAreaGeometryHandle::setToEmpty()
{
    if ( (fGeometry != 0 /*NIL*/) && (!fGeometry->isEmpty()) )//added for defect 25849
    {
        fGeometry->removeReference();
        fGeometry = 0 /*NIL*/;
    }
}

/*================
||
||  IAreaGeometryHandle::isRectilinear
||
*/

bool IAreaGeometryHandle::isRectilinear() const
{
    if (fGeometry != 0 /*NIL*/)
        return fGeometry->isRectilinear();
    else
        return true;
}

/*================
||
||  IAreaGeometryHandle::isRectangular
||
*/

bool IAreaGeometryHandle::isRectangular() const
{
    if (fGeometry != 0 /*NIL*/)
        return fGeometry->isRectangular();
    else
        return true;
}

/*================
||
||  IAreaGeometryHandle::isSingular
||
*/

bool IAreaGeometryHandle::isSingular() const
{
    if (fGeometry != 0 /*NIL*/)
        return fGeometry->isSingular();
    else
        return true;
}

/*================
||
||  IAreaGeometryHandle::intersects
||
*/

bool IAreaGeometryHandle::intersects(const IGRect2D& src) const
{
    if (fGeometry != 0 /*NIL*/)
        return fGeometry->intersects(src);
    else
        return false;
}

/*================
||
||  IAreaGeometryHandle::contains
||
*/

bool IAreaGeometryHandle::contains(const IGPoint2D& src) const
{
    if (fGeometry != 0 /*NIL*/)
        return fGeometry->contains(src);
    else
        return false;
}

bool IAreaGeometryHandle::contains(const IGRect2D& src) const
{
    if (fGeometry != 0 /*NIL*/)
        return fGeometry->contains(src);
    else
        return false;
}

/*================
||
||  IAreaGeometryHandle::bounds
||
*/

IGRect2D IAreaGeometryHandle::bounds() const
{
    if ( (fGeometry != 0 ) && (!fGeometry->isEmpty() ) )
        return fGeometry->bounds();
    else
        return IGRect2D::zeroRect();
}

/*================
||
||  IAreaGeometryHandle cAG Operations
||
*/

void IAreaGeometryHandle::add(const IAreaGeometryHandle& handle)
{
    IMAreaGeometry* targetGeometry = fGeometry;
    IMAreaGeometry* srcGeometry = handle.fGeometry;
    if ( (targetGeometry == 0 )  || (targetGeometry->isEmpty()) ) // 0 + X == X
    {
        if ( (srcGeometry != 0 ) && (!srcGeometry->isEmpty()))
            *this = handle;
    }
    else if ( (srcGeometry != 0 ) && (!srcGeometry->isEmpty()) ) // X + 0 == X
    {
        // Do some trivial rejections
        IGRect2D bounds(targetGeometry->bounds());
        IGRect2D handleBounds(srcGeometry->bounds());

        if (targetGeometry->isRectangular() && bounds.contains(handleBounds))
        {
            // If I contain handle, do nothing
        }
        else if (srcGeometry->isRectangular() && handleBounds.contains(bounds))
        {
            *this = handle;
        }
        else if (*targetGeometry == *srcGeometry)
        {
            // If equal, do nothing
        }
        else
            *this = targetGeometry->cloneAndAdd(handle);
    }
}

void IAreaGeometryHandle::subtract(const IAreaGeometryHandle& handle)
{
    IMAreaGeometry* targetGeometry = fGeometry;
    if ( (targetGeometry != 0 ) && (!targetGeometry->isEmpty()) )// 0 - X == 0
    {
        IMAreaGeometry* srcGeometry = handle.fGeometry;
        if ((srcGeometry != 0 ) && (!srcGeometry->isEmpty()) ) // 0 - 0 == 0
        {
            // Do some trivial rejections
            IGRect2D bounds(targetGeometry->bounds());
            IGRect2D handleBounds(srcGeometry->bounds());

            if (bounds.intersects(handleBounds))
            {
                if (srcGeometry->isRectangular() && handleBounds.contains(bounds))
                    setToEmpty();
                else if (*targetGeometry == *srcGeometry)
                    setToEmpty();
                else
                    *this = targetGeometry->cloneAndSubtract(handle);
            }
        }
    }
}

void IAreaGeometryHandle::intersect(const IAreaGeometryHandle& handle)
{
    IMAreaGeometry* targetGeometry = fGeometry;
    if ( (targetGeometry != 0 /*NIL*/)  && (!targetGeometry->isEmpty()/*NIL*/) ) // 0 * X == 0
    {
        IMAreaGeometry* srcGeometry = handle.fGeometry;
        if ( (srcGeometry == 0 /*NIL*/) || (srcGeometry->isEmpty()) )    // X * 0 == 0
            *this = handle;
        else
        {
            // Do some trivial rejections
            IGRect2D bounds(targetGeometry->bounds());
            IGRect2D handleBounds(srcGeometry->bounds());

            if (!bounds.intersects(handleBounds))
                setToEmpty();
            else if (targetGeometry->isRectangular() && bounds.contains(handleBounds))
                *this = handle;
            else if (srcGeometry->isRectangular() && handleBounds.contains(bounds))
            {
                // If handle contains me, do nothing
            }
            else if (*targetGeometry == *srcGeometry)
            {
                // If equal, do nothing
            }
            else
                *this = targetGeometry->cloneAndIntersect(handle);
        }
    }
}

void IAreaGeometryHandle::exclusiveOr(const IAreaGeometryHandle& handle)
{
    if ((fGeometry == 0 ) || (fGeometry->isEmpty()) )  // 0 ^ X == X
    {
        if ( (handle.fGeometry != 0 ) && (!handle.fGeometry->isEmpty()) )
            *this = handle;
    }
    else if ( (handle.fGeometry != 0) && (!handle.fGeometry->isEmpty()))   // X ^ 0  == X
    {
        if (*fGeometry == *handle.fGeometry)    // if they are equal, set to empty
            setToEmpty();
        else
            *this = fGeometry->cloneAndExclusiveOr(handle);
    }
}

/*================
||
||  IAreaGeometryHandle::extract
||
*/

void IAreaGeometryHandle::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
    if (fGeometry != 0 /*NIL*/)
        fGeometry->extract(extractor, parent, matrix);
}

/*================
||
||  IAreaGeometryHandle::transformBy
||
*/

void IAreaGeometryHandle::transformBy(const IGrafMatrix& m)
{
    if ((fGeometry != 0 /*NIL*/) && !m.isIdentity())
        *this = fGeometry->cloneAndTransform(m);
}

void IAreaGeometryHandle::writeToStream(IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
    ::writeObject(fGeometry, toWhere);
}

void IAreaGeometryHandle::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    if (fGeometry != 0 /*NIL*/)
        fGeometry->removeReference();

    ::readObject(fGeometry, fromWhere );
    if (fGeometry != 0 /*NIL*/)
        fGeometry->addReference();

}


/*================
||
||  IAreaGeometryHandle time stamp function
||
*/

unsigned long IAreaGeometryHandle::timeStamp() const
{
    if (fGeometry == 0 /*NIL*/)
        return (unsigned long) 0;
    else
        return fGeometry->timeStamp();
}


//========================//
// IMAreaGeometry routines //
//========================//

/*================
||
||  IMAreaGeometry constructors
||
*/

IMAreaGeometry::IMAreaGeometry()
{
}

IMAreaGeometry::IMAreaGeometry(const IMAreaGeometry& source)
    : fTimeStamp(source.fTimeStamp)
{
}


/*================
||
||  IMAreaGeometry destructor
||
*/

IMAreaGeometry::~IMAreaGeometry()
{
}


/*================
||
||  IMAreaGeometry::isRectilinear
||
*/

bool IMAreaGeometry::isRectilinear() const
{
    return false;
}


/*================
||
||  IMAreaGeometry::isRectangular
||
*/

bool IMAreaGeometry::isRectangular() const
{
    return false;
}


/*================
||
||  IMAreaGeometry::isSingular
||
*/

bool IMAreaGeometry::isSingular() const
{
    return true;
}


/*================
||
||  IMAreaGeometry cAG functions
||
*/

IAreaGeometryHandle IMAreaGeometry::cloneAndAdd(const IAreaGeometryHandle& handle)
{
    return IAreaGeometryHandle(new (::kSameHeap, this) IAddGeometry(IAreaGeometryHandle(this), handle));
}

IAreaGeometryHandle IMAreaGeometry::cloneAndSubtract(const IAreaGeometryHandle& handle)
{
    return IAreaGeometryHandle(new (::kSameHeap, this) ISubtractGeometry(IAreaGeometryHandle(this), handle));
}

IAreaGeometryHandle IMAreaGeometry::cloneAndIntersect(const IAreaGeometryHandle& handle)
{
    return IAreaGeometryHandle(new (::kSameHeap, this) IIntersectGeometry(IAreaGeometryHandle(this), handle));
}

IAreaGeometryHandle IMAreaGeometry::cloneAndExclusiveOr(const IAreaGeometryHandle& handle)
{
    return IAreaGeometryHandle(new (::kSameHeap, this) IExclusiveOrGeometry(IAreaGeometryHandle(this), handle));
}


/*================
||
||  IMAreaGeometry::cloneAndTransform
||
*/

IAreaGeometryHandle IMAreaGeometry::cloneAndTransform(const IGrafMatrix& m)
{
    return IAreaGeometryHandle(new (::kSameHeap, this) ITransformedGeometry(IAreaGeometryHandle(this), m));
}

void IMAreaGeometry::writeToStream(IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
}

void IMAreaGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    updateTimeStamp();
}

/*================
||
||  IMAreaGeometry operators
||
*/

IMAreaGeometry& IMAreaGeometry::operator=(const IMAreaGeometry& source)
{
    if (this != &source)
    {
        fTimeStamp = source.fTimeStamp;
    }
    return *this;
}


/*================
||
||  IMAreaGeometry::unreachable
||
*/

void IMAreaGeometry::unreachable()
{
    delete this;
}


#if !IC_RTTI_ENABLED
/*================
||
||  IMAreaGeometry substitute for RTTI
||
||  *** This will go away when we get RTTI ***
||
*/

bool IMAreaGeometry::isSameClass(const IMAreaGeometry& geometry) const
{
    return (GraphicTypeid(*this) == GraphicTypeid(geometry));
}
#endif // IC_RTTI_ENABLED



/*================
||
||  IMAreaGeometry time stamp functions
||
*/

unsigned long IMAreaGeometry::fgTimeStampBase;

unsigned long IMAreaGeometry::timeStamp() const
{
    return fTimeStamp;
}

void IMAreaGeometry::updateTimeStamp()
{
    fTimeStamp = ++fgTimeStampBase;
}

bool IMAreaGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool IMAreaGeometry::operator==(const IMAreaGeometry& obj) const
{
	return false;
}

bool IMAreaGeometry::operator<(const IMAreaGeometry& obj) const
{
	return fTimeStamp < obj.fTimeStamp;
}

//===============================//
// ITransformedGeometry routines //
//===============================//

StreamableDefinitionsMacro(ITransformedGeometry, gGraphicsStreamModule);

/*================
||
||  ITransformedGeometry constructors
||
*/

ITransformedGeometry::ITransformedGeometry()
{
    fBoundsValid = false;
}

ITransformedGeometry::ITransformedGeometry(
    const IAreaGeometryHandle& handle,
    const IGrafMatrix& modelMatrix)
    : fGeometry(handle), fMatrix (modelMatrix)
{
    fBoundsValid = false;
    updateTimeStamp();
}

ITransformedGeometry::ITransformedGeometry(const ITransformedGeometry& source)
    : IMAreaGeometry(source), fGeometry (source.fGeometry),
      fMatrix (source.fMatrix)
{
    fBoundsValid = source.fBoundsValid;
    if (fBoundsValid)
        fBounds = source.fBounds;
}


/*================
||
||  ITransformedGeometry destructor
||
*/

ITransformedGeometry::~ITransformedGeometry()
{
}


/*================
||
||  ITransformedGeometry::extract
||
*/

void ITransformedGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
	IGrafMatrix save = ((ITransformedGeometry *)this)->fMatrix;
    ((ITransformedGeometry *)this)->fMatrix.concatWith(matrix);
    fGeometry.extract(extractor, parent, fMatrix);
    ((ITransformedGeometry *)this)->fMatrix = save;
}


/*================
||
||  ITransformedGeometry::cloneAndTransform
||
*/

IAreaGeometryHandle ITransformedGeometry::cloneAndTransform(const IGrafMatrix& m)
{
    // If I'm shared, create a new object; else modify and return myself
    if (count() > 1)
    {
        IGrafMatrix concatMatrix (fMatrix);
        concatMatrix.concatWith(m);
        if (concatMatrix.isIdentity())
            return fGeometry;
        else
            return IAreaGeometryHandle(new (::kSameHeap, this) ITransformedGeometry(fGeometry, concatMatrix));
    }
    else
    {
        IGrafMatrix concatMatrix (fMatrix);
        concatMatrix.concatWith(m);

        if (concatMatrix.isIdentity())
            return fGeometry;
        else
        {
            fMatrix = concatMatrix;
            updateTimeStamp();
            fBoundsValid = false;
            return IAreaGeometryHandle(this);
        }
    }
}


/*================
||
||  ITransformedGeometry::isEmpty
||
*/

bool ITransformedGeometry::isEmpty() const
{
    return fGeometry.isEmpty();
}


/*================
||
||  ITransformedGeometry::isRectilinear
||
*/

bool ITransformedGeometry::isRectilinear() const
{
    return (fMatrix.isRectilinear() && fGeometry.isRectilinear());
}


/*================
||
||  ITransformedGeometry::isRectangular
||
*/

bool ITransformedGeometry::isRectangular() const
{
    return (fMatrix.isRectilinear() && fGeometry.isRectangular());
}


/*================
||
||  ITransformedGeometry::isSingular
||
*/

bool ITransformedGeometry::isSingular() const
{
    return fGeometry.isSingular();
}


/*================
||
||  ITransformedGeometry Hit Testing
||
*/

IGRect2D ITransformedGeometry::bounds() const
{
    if (!fBoundsValid)
    {
        ((ITransformedGeometry*) this)->fBounds = fMatrix.transformBounds(fGeometry.bounds());
        ((ITransformedGeometry*) this)->fBoundsValid = true;    // *** need ~const!
    }
    return fBounds;
}

bool ITransformedGeometry::intersects(const IGRect2D& rect) const
{
    // *** This isn't very accurate!  It may return true when it should return false!
    return fGeometry.intersects(fMatrix.untransformBounds (rect));
}

bool ITransformedGeometry::contains(const IGPoint2D& point) const
{
    return fGeometry.contains(fMatrix.untransformPoint(point));
}

bool ITransformedGeometry::contains(const IGRect2D& rect) const
{
    // *** This isn't very accurate!  It may return false when it should return true!
    return fGeometry.contains(fMatrix.untransformBounds (rect));
}


void ITransformedGeometry::writeToStream(IDataStream& toWhere ) const
{
    IStreamOutFrame streamFrame(toWhere);
    IMAreaGeometry::writeToStream(toWhere);

    fGeometry >>= toWhere;
    fMatrix >>= toWhere;
}

void ITransformedGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    IMAreaGeometry::readFromStream(fromWhere);

    fGeometry <<= fromWhere;
    fMatrix <<= fromWhere;
    fBoundsValid = false;
}

/*================
||
||  ITransformedGeometry operators
||
*/

ITransformedGeometry& ITransformedGeometry::operator=(const ITransformedGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);
        fGeometry = source.fGeometry;
        fMatrix = source.fMatrix;
        fBounds = source.fBounds;
        fBoundsValid = source.fBoundsValid;
    }
    return *this;
}

bool ITransformedGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool ITransformedGeometry::operator==(const IMAreaGeometry& collectible) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(collectible))
    {
        const ITransformedGeometry& geometry = dynamic_cast<const ITransformedGeometry&>(collectible);
#else
    if (isSameClass(collectible))
    {
        ITransformedGeometry& geometry = *((ITransformedGeometry*)&collectible);
#endif // IC_RTTI_ENABLED
        return ((timeStamp() == geometry.timeStamp()) ||
                ((fMatrix == geometry.fMatrix) && (fGeometry == geometry.fGeometry)));
    }
    else
        return false;
}



//========================//
// IRectGeometry routines //
//========================//

StreamableDefinitionsMacro(IRectGeometry, gGraphicsStreamModule);

/*================
||
||  IRectGeometry constructors
||
*/

IRectGeometry::IRectGeometry()
    : IGRect2D(IGRect2D::zeroRect())
{
}

IRectGeometry::IRectGeometry(const IGRect2D& rect)
    : IGRect2D(rect)
{
    updateTimeStamp();
}

IRectGeometry::IRectGeometry(const IRectGeometry& source)
    : IMAreaGeometry(source), IGRect2D(source)
{
}


/*================
||
||  IRectGeometry destructor
||
*/

IRectGeometry::~IRectGeometry()
{
}


/*================
||
||  IRectGeometry cAG functions
||
*/

IAreaGeometryHandle IRectGeometry::cloneAndAdd(const IAreaGeometryHandle& handle)
{
    if (handle.contains(*this))
        return handle;

// #ifndef SS
#ifdef NO_RECTILINEARGEOMETRY
    else if (handle.isRectilinear())
    {
        IAreaGeometryHandle newHandle(new (::kSameHeap, this) IRectilinearGeometry(*this));
        newHandle.add(handle);
        return newHandle;
    }
#endif // NO_RECTILINEARGEOMETRY
// #endif // not SS

    else
        return IMAreaGeometry::cloneAndAdd(handle);
}

IAreaGeometryHandle IRectGeometry::cloneAndSubtract(const IAreaGeometryHandle& handle)
{
    if (handle.contains(*this))
        return IAreaGeometryHandle();

// #ifndef SS
#ifdef NO_RECTILINEARGEOMETRY
    else if (handle.isRectilinear())
    {
        IAreaGeometryHandle newHandle(new (::kSameHeap, this) IRectilinearGeometry(*this));
        newHandle.subtract(handle);
        return newHandle;
    }
#endif // NO_RECTILINEARGEOMETRY
// #endif // not SS

    else
        return IMAreaGeometry::cloneAndSubtract(handle);
}

IAreaGeometryHandle IRectGeometry::cloneAndIntersect(const IAreaGeometryHandle& handle)
{
    if (handle.contains(*this))
        return IAreaGeometryHandle(this);
    else if (handle.isRectangular())
    {
        IRectGeometry* newRect = (count() > 1) ? new (::kSameHeap, this) IRectGeometry(*this) : this;
        newRect->intersectWith(handle.bounds());
        newRect->updateTimeStamp();
        return IAreaGeometryHandle(newRect);
    }

// #ifndef SS
#ifdef NO_RECTILINEARGEOMETRY
    else if (handle.isRectilinear())
    {
        IAreaGeometryHandle newHandle(new (::kSameHeap, this) IRectilinearGeometry(*this));
        newHandle.intersect(handle);
        return newHandle;
    }
#endif // NO_RECTILINEARGEOMETRY
// #endif // not SS

    else
        return IMAreaGeometry::cloneAndIntersect(handle);
}

IAreaGeometryHandle IRectGeometry::cloneAndExclusiveOr(const IAreaGeometryHandle& handle)
{

// #ifndef SS
#ifdef NO_RECTILINEARGEOMETRY
    if (handle.isRectilinear())
    {
        IAreaGeometryHandle newHandle(new (::kSameHeap, this) IRectilinearGeometry(*this));
        newHandle.exclusiveOr(handle);
        return newHandle;
    }
    else
#endif // NO_RECTILINEARGEOMETRY
// #endif // not SS

        return IMAreaGeometry::cloneAndExclusiveOr(handle);
}


/*================
||
||  IRectGeometry::extract
||
*/

void IRectGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
    ICAGEvenOdd* node = new (::kSameHeap, this) ICAGEvenOdd(parent);
    extractor.adoptCAGNode(node);
    extractor.extractRect(*this, node, matrix);
}


/*================
||
||  IRectGeometry::cloneAndTransform
||
*/

IAreaGeometryHandle IRectGeometry::cloneAndTransform(const IGrafMatrix& m)
{

// #ifndef SS
#ifdef NO_RECTILINEARGEOMETRY
    if (m.isRectilinear())
    {
        IGPoint2D topLeft = m.transformPoint(topLeft());
        IGPoint2D bottomRight = m.transformPoint(bottomRight());
        IGRect2D newRect(topLeft, bottomRight);
        newRect.orderPoints();

        // If I'm shared, create a new object; else modify and return myself
        if (count() > 1)
            return IAreaGeometryHandle(new (::kSameHeap, this) IRectGeometry(newRect));
        else
        {
            *this = newRect;
            updateTimeStamp();
            return IAreaGeometryHandle(this);
        }
    }
    else
#endif // NO_RECTILINEARGEOMETRY
// #endif // not SS

        return IMAreaGeometry::cloneAndTransform(m);
}


/*================
||
||  IRectGeometry::isEmpty
||
*/

bool IRectGeometry::isEmpty() const
{
    return IGRect2D::isEmpty();
}


/*================
||
||  IRectGeometry::isRectilinear
||
*/

bool IRectGeometry::isRectilinear() const
{
    return true;
}


/*================
||
||  IRectGeometry::isRectangular
||
*/

bool IRectGeometry::isRectangular() const
{
    return true;
}


/*================
||
||  IRectGeometry Hit Testing
||
*/

IGRect2D IRectGeometry::bounds() const
{
    return *this;
}

bool IRectGeometry::intersects(const IGRect2D& rect) const
{
    return IGRect2D::intersects(rect);
}

bool IRectGeometry::contains(const IGPoint2D& p) const
{
    return IGRect2D::contains(p);
}

bool IRectGeometry::contains(const IGRect2D& rect) const
{
    return IGRect2D::contains(rect);
}


void IRectGeometry::writeToStream(IDataStream& toWhere ) const
{
    IStreamOutFrame streamFrame(toWhere);
    IMAreaGeometry::writeToStream(toWhere);
    IGRect2D::writeToStream(toWhere);
}

void IRectGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    IMAreaGeometry::readFromStream(fromWhere);
    IGRect2D::readFromStream(fromWhere);
}

/*================
||
||  IRectGeometry operators
||
*/

IRectGeometry& IRectGeometry::operator=(const IRectGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);
        IGRect2D::operator=(source);
    }
    return *this;
}

bool IRectGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool IRectGeometry::operator==(const IMAreaGeometry& collectible) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(collectible))
        return IGRect2D::operator==(dynamic_cast<const IRectGeometry&>(collectible));
#else
    if (isSameClass(collectible))
        return IGRect2D::operator==(*((IRectGeometry*)&collectible));
#endif // IC_RTTI_ENABLED
    else
        return false;
}



// #ifndef SS
#ifdef NO_RECTILINEARGEOMETRY

//======================//
// IRectilinearGeometry //
//======================//

// To Do:
//  *   cloneAndTransform of rectilinear rotation matrix.
//  *   Handle rectilinear polygons.  Even-odd is easy; winding number is harder.

const GCoordinate *  IRectilinearGeometry::kStartList = 0;
const GCoordinate *  IRectilinearGeometry::kEndList   = 0;

void
IRectilinearGeometry::initStartEnd()
{
    if (!kStartList) {
        IPrimalLock lock;
        if (!kStartList) {
            kStartList = new GCoordinate(-kInfinity);
            kEndList = new GCoordinate(kInfinity);
        }
    }
}

const long IRectilinearGeometry::kDataIncrement = 64;
const long IRectilinearGeometry::kLocalBufferSize = 200;

#ifdef IRECTILINEARGEOMETRY_DEBUG

/*================
||
||  IRectilinearGeometry self-check function; checks for internal consistency
||
*/

void IRectilinearGeometry::check() const
{
    initStartEnd();

    long sizeData = 0;
    for(long i = 0; i < fSizeLength; i++)
        sizeData += fLength[i];
    InternalAssert(sizeData + 1 == fSizeData, "Sizes don't match");

    if (fSizeLength > 1)
    {
        IGRect2D bounds(kInfinity, kInfinity, -kInfinity, -kInfinity);
        GCoordinate* dataPtr = fData;
        GCoordinate y = *kStartList;

        InternalAssert(*dataPtr == *kStartList, "Start value is wrong");

        for(i = 0; i < fSizeLength; i++)
        {
            GCoordinate prevY = y;
            y = *dataPtr++;
            InternalAssert((dataPtr == fData + 1) || (y > prevY), "Y values are out of order");
            bounds.fBottom = y;

            short length = fLength[i] - 2;
            if (length > 0)
            {
                if (y < bounds.fTop)
                    bounds.fTop = y;

                if (dataPtr[0] < bounds.fLeft)
                    bounds.fLeft = dataPtr[0];

                for(long j = 0; j < length; j++, dataPtr++)
                {
                    InternalAssert(dataPtr[0] < dataPtr[1], "X values are out of order");

                    if (dataPtr[1] > bounds.fRight)
                        bounds.fRight = dataPtr[1];
                }
                dataPtr++;
            }
        }

        InternalAssert(*dataPtr == *kEndList, "end value is wrong");
        InternalAssert(bounds == fBounds, "Bounds is wrong");
    }
}
#endif  // IRECTILINEARGEOMETRY_DEBUG


/*================
||
||  IRectilinearGeometry constructors & destructor
||
*/

IRectilinearGeometry::IRectilinearGeometry()
{
    initStartEnd();

    fBounds = IGRect2D::zeroRect();

    fSizeData = 2;
    fData = new (::kSameHeap, this) GCoordinate[fSizeData];
    GCoordinate* data = fData;
    *data++ = *kStartList;
    *data++ = *kEndList;

    fSizeLength = 1;
    fLength = new (::kSameHeap, this) short[fSizeLength];
    *fLength = 1;
}

IRectilinearGeometry::IRectilinearGeometry(const IGRect2D& g)
{
    initStartEnd();

    fBounds = g;

    fSizeData = 6;
    fData = new (::kSameHeap, this) GCoordinate[fSizeData];
    GCoordinate* data = fData;
    *data++ = *kStartList;
    *data++ = g.fTop;
    *data++ = g.fLeft;
    *data++ = g.fRight;
    *data++ = g.fBottom;
    *data++ = *kEndList;

    fSizeLength = 3;
    fLength = new (::kSameHeap, this) short[fSizeLength];
    short* length = fLength;
    *length++ = 1;
    *length++ = 3;
    *length++ = 1;

    updateTimeStamp();
}

IRectilinearGeometry::IRectilinearGeometry(const IRectilinearGeometry& source)
    : IMAreaGeometry(source)
{
    long i;
    GCoordinate *sd,*dd;
    short *ss, *ds;

    fBounds = source.fBounds;
    fSizeData = source.fSizeData;
    fSizeLength = source.fSizeLength;
    fData = new (::kSameHeap, this) GCoordinate[fSizeData];
    fLength = new (::kSameHeap, this) short[fSizeLength];
    for(i=0, sd = source.fData, dd=fData; i<fSizeData; i++)
        *dd++ = *sd++;
    for(i=0, ss = source.fLength, ds=fLength; i<fSizeLength; i++)
        *ds++ = *ss++;
}


IRectilinearGeometry::IRectilinearGeometry(
    GCoordinate* adoptedData,
    long    dataSize,
    short*  adoptedLength,
    long    lengthSize,
    IGRect2D  bounds )
{
    fData = adoptedData;
    fSizeData = dataSize;
    fLength = adoptedLength;
    fSizeLength = lengthSize;
    fBounds = bounds;

    updateTimeStamp();
}


IRectilinearGeometry::~IRectilinearGeometry()
{
    delete fLength;
    delete fData;
}


/*================
||
||  IRectilinearGeometry::cloneAndTransform
||
*/

IAreaGeometryHandle IRectilinearGeometry::cloneAndTransform(const IGrafMatrix& m)
{
    initStartEnd();

    if ((m.type() == IGrafMatrix::kTranslate) || (m.type() == IGrafMatrix::kScale))
    {
        IRectilinearGeometry* newGeometry = (count() > 1) ? new (::kSameHeap, this) IRectilinearGeometry(*this) : this;

        IGRect2D bounds(*kEndList, *kEndList, *kStartList, *kStartList);

        GCoordinate* dataPtr;
        short* lengthPtr;
        bool topFound = false;
        GCoordinate y = *kStartList;
        for(dataPtr = newGeometry->fData, lengthPtr = newGeometry->fLength; *dataPtr != *kEndList; lengthPtr++)
        {
            short length = *lengthPtr - 1;

            y = *dataPtr;
            if (y != *kStartList)    // don't transform negative infinity
            {
                IGPoint2D point = m.transformPoint(IGPoint2d(0, y));
                y = point.fY;

                // Expand the top and bottom of the bounds
                if (!topFound)
                {
                    bounds.fTop = y;
                    topFound = true;
                }
            }

            *dataPtr++ = y;

            // Transform the x pairs and expand the bounds if there are any x pairs
            if (length > 0)
            {
                for(short i = 0; i < length; i += 2, dataPtr += 2)
                {
                    dataPtr[0] = m.transformPoint(IGPoint2d(dataPtr[0], 0)).fX;
                    dataPtr[1] = m.transformPoint(IGPoint2d(dataPtr[1], 0)).fX;
                }

                GCoordinate x = dataPtr[-length];
                if (x < bounds.fLeft)
                    bounds.fLeft = x;

                x = dataPtr[-1];
                if (x > bounds.fRight)
                    bounds.fRight = x;
            }
        }

        bounds.fBottom = y;
        newGeometry->fBounds = bounds;
        newGeometry->updateTimeStamp();
        return IAreaGeometryHandle(newGeometry);
    }
    else
        return IMAreaGeometry::cloneAndTransform(m);
}


/*================
||
||  IRectilinearGeometry extract function
||
*/

void IRectilinearGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
    initStartEnd();

    if (fSizeLength > 1)
    {
        ICAGEvenOdd* node = new (::kSameHeap, this) ICAGEvenOdd(parent);
        extractor.adoptCAGNode(node);

        GCoordinate* dataPtr;
        short* lengthPtr;

        for(dataPtr = fData, lengthPtr = fLength; *dataPtr != *kEndList; lengthPtr++)
        {
            short length = *lengthPtr - 1;
            GCoordinate y = *dataPtr++;
            GCoordinate nextY = dataPtr[length];

            for(; length > 0; length -= 2, dataPtr += 2)
                extractor.extractRect(IGRect2D(dataPtr[0], y, dataPtr[1], nextY), node, matrix);
        }
    }
}


/*================
||
||  IRectilinearGeometry assignment operator
||
*/

IRectilinearGeometry& IRectilinearGeometry::operator=(const IRectilinearGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);

        fBounds = source.fBounds;
        fSizeData = source.fSizeData;
        fSizeLength = source.fSizeLength;

        delete fData;
        delete fLength;
        fData = new (::kSameHeap, this) GCoordinate[fSizeData];
        fLength = new (::kSameHeap, this) short[fSizeLength];

        long i;
        GCoordinate *sd,*dd;
        short *ss, *ds;

        for(i=0, sd = source.fData, dd=fData; i<fSizeData; i++)
            *dd++ = *sd++;
        for(i=0, ss = source.fLength, ds=fLength; i<fSizeLength; i++)
            *ds++ = *ss++;
    }
    return *this;
}


/*================
||
||  IRectilinearGeometry Query functions
||
*/

bool IRectilinearGeometry::isEmpty() const
{
    return (fSizeLength == 1);
}

bool IRectilinearGeometry::isRectilinear() const
{
    return true;
}

bool IRectilinearGeometry::isRectangular() const
{
    return (fSizeData <= 6);
}

bool IRectilinearGeometry::isSingular() const
{
    return isRectangular();
}

IGRect2D IRectilinearGeometry::bounds() const
{
    return fBounds;
}


/*================
||
||  IRectilinearGeometry::add
||
*/

IAreaGeometryHandle IRectilinearGeometry::cloneAndAdd(const IAreaGeometryHandle& handle)
{
    if (contains(handle.bounds()))
        return IAreaGeometryHandle(this);

#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(*handle.geometry()))
    {
        IRectilinearGeometry*   r = dynamic_cast<const IRectilinearGeometry*>(handle.geometry());
#else
    if (isSameClass(*handle.geometry()))
    {
        IRectilinearGeometry*   r = (IRectilinearGeometry*) handle.geometry();
#endif // IC_RTTI_ENABLED
        GCoordinate*    newData = 0 /*NIL*/;
        long    newDataSize = 0;
        short*  newLength = 0 /*NIL*/;
        long    newLengthSize = 0;
        IGRect2D  newBounds;

        addData(  fData, fSizeData, fLength, fSizeLength,
                    r->fData, r->fSizeData, r->fLength, r->fSizeLength,
                    newData, newDataSize, newLength, newLengthSize, newBounds );

        return IAreaGeometryHandle(new (::kSameHeap, this) IRectilinearGeometry(newData, newDataSize, newLength, newLengthSize, newBounds));
    }
    else if (handle.isRectangular())
    {
        IRectilinearGeometry* newGeometry = (count() > 1) ? new (::kSameHeap, this) IRectilinearGeometry(*this) : this;
        newGeometry->add(IRectilinearGeometry( handle.bounds() ) );
        newGeometry->updateTimeStamp();
        return IAreaGeometryHandle(newGeometry);
    }

    else
        return IMAreaGeometry::cloneAndAdd(handle);
}

void IRectilinearGeometry::add(const IRectilinearGeometry& geometry)
{
    addData(  fData, fSizeData, fLength, fSizeLength,
                geometry.fData, geometry.fSizeData, geometry.fLength, geometry.fSizeLength,
                fData, fSizeData, fLength, fSizeLength, fBounds );
}


void IRectilinearGeometry::addData(
    GCoordinate* myData,
    long    myDataSize,
    short*  myLength,
    long    myLengthSize,
    GCoordinate* otherData,
    long    otherDataSize,
    short*  otherLength,
    long    otherLengthSize,
    GCoordinate*& newData,
    long&   newDataSize,
    short*& newLength,
    long&   newLengthSize,
    IGRect2D& newBounds )
{
    initStartEnd();

    // set or allocate the local length array.
    short   localLengthBuffer [ kLocalBufferSize ];
    short*  localLength;
    long allocatedSizeLength = myLengthSize + otherLengthSize - 1;  // the max it can be
    bool    usingLocalBufferForLength;
    if (allocatedSizeLength > kLocalBufferSize )
    {
        localLength = new (::kSameHeap, myLength) short [ allocatedSizeLength ];
        usingLocalBufferForLength = false;
    }
    else
    {
        localLength = localLengthBuffer;
        usingLocalBufferForLength = true;
    }

    // set the initial local data array.
    GCoordinate localDataBuffer [ kLocalBufferSize ];
    GCoordinate*    localData = localDataBuffer;
    long allocatedSizeData = kLocalBufferSize;
    bool    usingLocalBufferForData = true;

    long    freeData = allocatedSizeData;       // number of free data slots

    GCoordinate* data1 = myData;                // "this" rectilinear data
    GCoordinate* data2 = otherData;             // the parameter's rectilinear data
    GCoordinate* data3 = localData;             // the new rectilinear data
    GCoordinate *prevData3 = data3;

    short* len1 = myLength;                     // "this" length data
    short* len2 = otherLength;                  // the parameter's length data
    short* len3 = localLength;                  // the new length data

    newBounds.fLeft = kInfinity;
    newBounds.fRight = -kInfinity;

    while ((*data1 != *kEndList) || (*data2 != *kEndList))
    {
        GCoordinate* endD1 = &data1[*len1];
        GCoordinate* endD2 = &data2[*len2];
        GCoordinate* d3 = data3;
        {
            GCoordinate* d1 = data1;
            GCoordinate* d2 = data2;

            int state1 = false;
            int state2 = false;
            int state3 = false;

            *d3++ = *((*d1 < *d2) ? d2 : d1);   // store the new y value
            freeData--;
            d1++;
            d2++;

            while ((d1 != endD1) || (d2 != endD2))
            {
                GCoordinate d1val = *d1;
                GCoordinate d2val = *d2;
                if (d2 == endD2)
                {
                    state1 = ~state1;
                    d1++;
                }
                else if (( d1 == endD1 ) || (d1val > d2val) )
                {
                    state2 = ~state2;
                    d1val = d2val;
                    d2++;
                }
                else if (d1val < d2val )
                {
                    state1 = ~state1;
                    d1++;
                }
                else    // if (*d1 == *d2)
                {
                    state1 = ~state1;
                    state2 = ~state2;
                    d1++;
                    d2++;
                }

                int newState3 = (state1 | state2);
                if (newState3 != state3)
                {
                    if (--freeData <= 1 )  // out of space?
                    {
                        allocatedSizeData += kDataIncrement;
                        freeData += kDataIncrement;
                        GCoordinate* newLocalData = new (::kSameHeap, myData) GCoordinate [ allocatedSizeData ];
                        memmove(newLocalData, localData, ( d3 - localData ) * sizeof( GCoordinate ) );
                        d3 = newLocalData + (d3 - localData );
                        data3 = newLocalData + (data3 - localData );
                        prevData3 = newLocalData + (prevData3 - localData );
                        if (usingLocalBufferForData )
                            usingLocalBufferForData = false;
                        else
                            delete localData;
                        localData = newLocalData;
                    }
                    *d3++ = d1val;
                    state3 = newState3;
                }
            }
        }

        *len3 = d3 - data3;     // length of new y-event

        if (*endD1 <= *endD2)
        {
            data1 = endD1;
            len1++;
        }
        if (*endD2 <= *endD1)
        {
            data2 = endD2;
            len2++;
        }

        // Eliminate redundant y-events
        if (prevData3 != data3)     // just did first line?
        {
            bool redundant = false;
            if (len3[0] == len3[-1])
            {
                GCoordinate* data3x = data3 + 1;
                GCoordinate* prevData3x = prevData3 + 1;
                for(; (prevData3x != data3) && (*data3x == *prevData3x); data3x++, prevData3x++);

                redundant = (prevData3x == data3);
            }

            if (!redundant)
            {
                prevData3 = data3;
                data3 = d3;

                if (len3[0] > 1)
                {
                    if (prevData3[1] < newBounds.fLeft )
                        newBounds.fLeft = prevData3[1];
                    if (data3[-1] > newBounds.fRight )
                        newBounds.fRight = data3[-1];
                }
                len3++;
            }
        }
        else
        {
            data3 = d3;
            len3++;
        }
    }
    *data3++ = *kEndList;

    long localDataSize = data3 - localData;
    long localLengthSize = len3 - localLength;

    // Update the bounds.
    if (localDataSize == 2 )
        newBounds = IGRect2D::zeroRect();
    else
    {
        newBounds.fTop = (localLength[ 0 ] == 1 ) ? localData [ 1 ] : localData [ 0 ];
        newBounds.fBottom = data3 [ - len3 [ -1 ] - 1 ];
    }

    // Update the data array.
    if (localDataSize > newDataSize )
    {
        delete newData;
        newData = new (::kSameHeap, myData) GCoordinate [ localDataSize ];
    }
    newDataSize = localDataSize;
    memmove(newData, localData, newDataSize * sizeof( GCoordinate ) );
    if (!usingLocalBufferForData )
        delete localData;

    // Update the length array.
    if (localLengthSize > newLengthSize )
    {
        delete newLength;
        newLength = new (::kSameHeap, myLength) short [ localLengthSize ];
    }
    newLengthSize = localLengthSize;
    memmove(newLength, localLength, newLengthSize * sizeof( short ) );
    if (!usingLocalBufferForLength )
        delete localLength;
}


/*================
||
||  IRectilinearGeometry::subtract
||
*/

IAreaGeometryHandle IRectilinearGeometry::cloneAndSubtract(const IAreaGeometryHandle& handle)
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(*handle.geometry()))
    {
        IRectilinearGeometry*   r = dynamic_cast<const IRectilinearGeometry*>(handle.geometry());
#else
    if (isSameClass(*handle.geometry()))
    {
        IRectilinearGeometry*   r = (IRectilinearGeometry*) handle.geometry();
#endif // IC_RTTI_ENABLED
        GCoordinate*    newData = 0 /*NIL*/;
        long    newDataSize = 0;
        short*  newLength = 0 /*NIL*/;
        long    newLengthSize = 0;
        IGRect2D  newBounds;

        subtractData( fData, fSizeData, fLength, fSizeLength,
                        r->fData, r->fSizeData, r->fLength, r->fSizeLength,
                        newData, newDataSize, newLength, newLengthSize, newBounds );

        return IAreaGeometryHandle(new (::kSameHeap, this) IRectilinearGeometry(newData, newDataSize, newLength, newLengthSize, newBounds));
    }
    else if (handle.isRectangular())
    {
        IRectilinearGeometry* newGeometry = (count() > 1) ? new (::kSameHeap, this) IRectilinearGeometry(*this) : this;
        newGeometry->subtract(IRectilinearGeometry( handle.bounds() ) );
        newGeometry->updateTimeStamp();
        return IAreaGeometryHandle(newGeometry);
    }
    else
        return IMAreaGeometry::cloneAndSubtract(handle);
}

void IRectilinearGeometry::subtract(const IRectilinearGeometry& geometry )
{
    subtractData( fData, fSizeData, fLength, fSizeLength,
                    geometry.fData, geometry.fSizeData, geometry.fLength, geometry.fSizeLength,
                    fData, fSizeData, fLength, fSizeLength, fBounds );
}


void IRectilinearGeometry::subtractData(
    GCoordinate* myData,
    long    myDataSize,
    short*  myLength,
    long    myLengthSize,
    GCoordinate* otherData,
    long    otherDataSize,
    short*  otherLength,
    long    otherLengthSize,
    GCoordinate*& newData,
    long&   newDataSize,
    short*& newLength,
    long&   newLengthSize,
    IGRect2D& newBounds )
{
    initStartEnd();

    // set or allocate the local length array.
    short   localLengthBuffer [ kLocalBufferSize ];
    short*  localLength;
    long allocatedSizeLength = myLengthSize + otherLengthSize - 1;  // the max it can be
    bool    usingLocalBufferForLength;
    if (allocatedSizeLength > kLocalBufferSize )
    {
        localLength = new (::kSameHeap, myLength) short [ allocatedSizeLength ];
        usingLocalBufferForLength = false;
    }
    else
    {
        localLength = localLengthBuffer;
        usingLocalBufferForLength = true;
    }

    // set the initial local data array.
    GCoordinate localDataBuffer [ kLocalBufferSize ];
    GCoordinate*    localData = localDataBuffer;
    long allocatedSizeData = kLocalBufferSize;
    bool    usingLocalBufferForData = true;

    long    freeData = allocatedSizeData;       // number of free data slots

    GCoordinate* data1 = myData;                // "this" rectilinear data
    GCoordinate* data2 = otherData;             // the parameter's rectilinear data
    GCoordinate* data3 = localData;             // the new rectilinear data
    GCoordinate *prevData3 = data3;

    short* len1 = myLength;                     // "this" length data
    short* len2 = otherLength;                  // the parameter's length data
    short* len3 = localLength;                  // the new length data

    newBounds.fLeft = kInfinity;
    newBounds.fRight = -kInfinity;

    while (*data1 != *kEndList)
    {
        GCoordinate* endD1 = &data1[*len1];
        GCoordinate* endD2 = &data2[*len2];
        GCoordinate* d3 = data3;
        {
            GCoordinate* d1 = data1;
            GCoordinate* d2 = data2;

            int state1 = false;
            int state2 = false;
            int state3 = false;

            *d3++ = *((*d1 < *d2) ? d2 : d1);   // store the new y value
            freeData--;
            d1++;
            d2++;

            while (d1 != endD1)
            {
                GCoordinate d1val = *d1;
                GCoordinate d2val = *d2;
                if ((d2 == endD2) || (d1val < d2val))
                {
                    state1 = ~state1;
                    d1++;
                }
                else if (d1val > d2val)
                {
                    state2 = ~state2;
                    d1val = d2val;
                    d2++;
                }
                else    // if (*d1 == *d2)
                {
                    state1 = ~state1;
                    state2 = ~state2;
                    d1++;
                    d2++;
                }

                int newState3 = (state1 & ~state2);
                if (newState3 != state3)
                {
                    if (--freeData <= 1 )  // out of space?
                    {
                        allocatedSizeData += kDataIncrement;
                        freeData += kDataIncrement;
                        GCoordinate* newLocalData = new (::kSameHeap, myData) GCoordinate [ allocatedSizeData ];
                        memmove(newLocalData, localData, ( d3 - localData ) * sizeof( GCoordinate ) );
                        d3 = newLocalData + (d3 - localData );
                        data3 = newLocalData + (data3 - localData );
                        prevData3 = newLocalData + (prevData3 - localData );
                        if (usingLocalBufferForData )
                            usingLocalBufferForData = false;
                        else
                            delete localData;
                        localData = newLocalData;
                    }
                    *d3++ = d1val;
                    state3 = newState3;
                }
            }
        }

        *len3 = d3 - data3;     // length of new y-event

        if (*endD1 <= *endD2)
        {
            data1 = endD1;
            len1++;
        }
        if (*endD2 <= *endD1)
        {
            data2 = endD2;
            len2++;
        }

        // Eliminate redundant y-events
        if (prevData3 != data3)     // just did first line?
        {
            bool redundant = false;
            if (len3[0] == len3[-1])
            {
                GCoordinate* data3x = data3 + 1;
                GCoordinate* prevData3x = prevData3 + 1;
                for(; (prevData3x != data3) && (*data3x == *prevData3x); data3x++, prevData3x++);

                redundant = (prevData3x == data3);
            }

            if (!redundant)
            {
                prevData3 = data3;
                data3 = d3;

                if (len3[0] > 1)
                {
                    if (prevData3[1] < newBounds.fLeft )
                        newBounds.fLeft = prevData3[1];
                    if (data3[-1] > newBounds.fRight )
                        newBounds.fRight = data3[-1];
                }
                len3++;
            }
        }
        else
        {
            data3 = d3;
            len3++;
        }
    }
    *data3++ = *kEndList;

    long localDataSize = data3 - localData;
    long localLengthSize = len3 - localLength;

    // Update the bounds.
    if (localDataSize == 2 )
        newBounds = IGRect2D::zeroRect();
    else
    {
        newBounds.fTop = (localLength[ 0 ] == 1 ) ? localData [ 1 ] : localData [ 0 ];
        newBounds.fBottom = data3 [ - len3 [ -1 ] - 1 ];
    }

    // Update the data array.
    if (localDataSize > newDataSize )
    {
        delete newData;
        newData = new (::kSameHeap, myData) GCoordinate [ localDataSize ];
    }
    newDataSize = localDataSize;
    memmove(newData, localData, newDataSize * sizeof( GCoordinate ) );
    if (!usingLocalBufferForData )
        delete localData;

    // Update the length array.
    if (localLengthSize > newLengthSize )
    {
        delete newLength;
        newLength = new (::kSameHeap, myLength) short [ localLengthSize ];
    }
    newLengthSize = localLengthSize;
    memmove(newLength, localLength, newLengthSize * sizeof( short ) );
    if (!usingLocalBufferForLength )
        delete localLength;
}


/*================
||
||  IRectilinearGeometry::intersect
||
*/

IAreaGeometryHandle IRectilinearGeometry::cloneAndIntersect(const IAreaGeometryHandle& handle)
{
    IGRect2D  otherBounds(handle.bounds());

    if (contains(otherBounds))
        return handle;

#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(*handle.geometry()))
    {
        IRectilinearGeometry*   r = dynamic_cast<const IRectilinearGeometry*>(handle.geometry());
#else
    if (isSameClass(*handle.geometry()))
    {
        IRectilinearGeometry*   r = (IRectilinearGeometry*) handle.geometry();
#endif // IC_RTTI_ENABLED
        GCoordinate*    newData = 0 /*NIL*/;
        long    newDataSize = 0;
        short*  newLength = 0 /*NIL*/;
        long    newLengthSize = 0;
        IGRect2D  newBounds;

        intersectData(fData, fSizeData, fLength, fSizeLength,
                        r->fData, r->fSizeData, r->fLength, r->fSizeLength,
                        newData, newDataSize, newLength, newLengthSize, newBounds );

        return IAreaGeometryHandle(new (::kSameHeap, this) IRectilinearGeometry(newData, newDataSize, newLength, newLengthSize, newBounds));
    }
    else if (handle.isRectangular())
    {
        IRectilinearGeometry* newGeometry = (count() > 1) ? new (::kSameHeap, this) IRectilinearGeometry(*this) : this;
        newGeometry->intersect(IRectilinearGeometry( handle.bounds() ) );
        newGeometry->updateTimeStamp();
        return IAreaGeometryHandle(newGeometry);
    }
    else
        return IMAreaGeometry::cloneAndIntersect(handle);
}

void IRectilinearGeometry::intersect(const IRectilinearGeometry& geometry)
{
    intersectData(fData, fSizeData, fLength, fSizeLength,
                    geometry.fData, geometry.fSizeData, geometry.fLength, geometry.fSizeLength,
                    fData, fSizeData, fLength, fSizeLength, fBounds );
}


void IRectilinearGeometry::intersectData(
    GCoordinate* myData,
    long    myDataSize,
    short*  myLength,
    long    myLengthSize,
    GCoordinate* otherData,
    long    otherDataSize,
    short*  otherLength,
    long    otherLengthSize,
    GCoordinate*& newData,
    long&   newDataSize,
    short*& newLength,
    long&   newLengthSize,
    IGRect2D& newBounds )
{
    initStartEnd();

    // set or allocate the local length array.
    short   localLengthBuffer [ kLocalBufferSize ];
    short*  localLength;
    long allocatedSizeLength = myLengthSize + otherLengthSize - 1;  // the max it can be
    bool    usingLocalBufferForLength;
    if (allocatedSizeLength > kLocalBufferSize )
    {
        localLength = new (::kSameHeap, myLength) short [ allocatedSizeLength ];
        usingLocalBufferForLength = false;
    }
    else
    {
        localLength = localLengthBuffer;
        usingLocalBufferForLength = true;
    }

    // set the initial local data array.
    GCoordinate localDataBuffer [ kLocalBufferSize ];
    GCoordinate*    localData = localDataBuffer;
    long allocatedSizeData = kLocalBufferSize;
    bool    usingLocalBufferForData = true;

    long    freeData = allocatedSizeData;       // number of free data slots

    GCoordinate* data1 = myData;                // "this" rectilinear data
    GCoordinate* data2 = otherData;             // the parameter's rectilinear data
    GCoordinate* data3 = localData;             // the new rectilinear data
    GCoordinate *prevData3 = data3;

    short* len1 = myLength;                     // "this" length data
    short* len2 = otherLength;                  // the parameter's length data
    short* len3 = localLength;                  // the new length data

    newBounds.fLeft = kInfinity;
    newBounds.fRight = -kInfinity;

    while ((*data1 != *kEndList) && (*data2 != *kEndList))
    {
        GCoordinate* endD1 = &data1[*len1];
        GCoordinate* endD2 = &data2[*len2];
        GCoordinate* d3 = data3;
        {
            GCoordinate* d1 = data1;
            GCoordinate* d2 = data2;

            int state1 = false;
            int state2 = false;
            int state3 = false;

            *d3++ = *((*d1 < *d2) ? d2 : d1);   // store the new y value
            freeData--;
            d1++;
            d2++;

            while ((d1 != endD1) && (d2 != endD2))
            {
                GCoordinate d1val = *d1;
                GCoordinate d2val = *d2;
                if (d1val > d2val)
                {
                    state2 = ~state2;
                    d1val = d2val;
                    d2++;
                }
                else if (d1val < d2val )
                {
                    state1 = ~state1;
                    d1++;
                }
                else    // if (*d1 == *d2)
                {
                    state1 = ~state1;
                    state2 = ~state2;
                    d1++;
                    d2++;
                }

                int newState3 = (state1 & state2);
                if (newState3 != state3)
                {
                    if (--freeData <= 1 )  // out of space?
                    {
                        allocatedSizeData += kDataIncrement;
                        freeData += kDataIncrement;
                        GCoordinate* newLocalData = new (::kSameHeap, myData) GCoordinate [ allocatedSizeData ];
                        memmove(newLocalData, localData, ( d3 - localData ) * sizeof( GCoordinate ) );
                        d3 = newLocalData + (d3 - localData );
                        data3 = newLocalData + (data3 - localData );
                        prevData3 = newLocalData + (prevData3 - localData );
                        if (usingLocalBufferForData )
                            usingLocalBufferForData = false;
                        else
                            delete localData;
                        localData = newLocalData;
                    }
                    *d3++ = d1val;
                    state3 = newState3;
                }
            }
        }

        *len3 = d3 - data3;     // length of new y-event

        if (*endD1 <= *endD2)
        {
            data1 = endD1;
            len1++;
        }
        if (*endD2 <= *endD1)
        {
            data2 = endD2;
            len2++;
        }

        // Eliminate redundant y-events
        if (prevData3 != data3)     // just did first line?
        {
            bool redundant = false;
            if (len3[0] == len3[-1])
            {
                GCoordinate* data3x = data3 + 1;
                GCoordinate* prevData3x = prevData3 + 1;
                for(; (prevData3x != data3) && (*data3x == *prevData3x); data3x++, prevData3x++);

                redundant = (prevData3x == data3);
            }

            if (!redundant)
            {
                prevData3 = data3;
                data3 = d3;

                if (len3[0] > 1)
                {
                    if (prevData3[1] < newBounds.fLeft )
                        newBounds.fLeft = prevData3[1];
                    if (data3[-1] > newBounds.fRight )
                        newBounds.fRight = data3[-1];
                }
                len3++;
            }
        }
        else
        {
            data3 = d3;
            len3++;
        }
    }
    *data3++ = *kEndList;

    long localDataSize = data3 - localData;
    long localLengthSize = len3 - localLength;

    // Update the bounds.
    if (localDataSize == 2 )
        newBounds = IGRect2D::zeroRect();
    else
    {
        newBounds.fTop = (localLength[ 0 ] == 1 ) ? localData [ 1 ] : localData [ 0 ];
        newBounds.fBottom = data3 [ - len3 [ -1 ] - 1 ];
    }

    // Update the data array.
    if (localDataSize > newDataSize )
    {
        delete newData;
        newData = new (::kSameHeap, myData) GCoordinate [ localDataSize ];
    }
    newDataSize = localDataSize;
    memmove(newData, localData, newDataSize * sizeof( GCoordinate ) );
    if (!usingLocalBufferForData )
        delete localData;

    // Update the length array.
    if (localLengthSize > newLengthSize )
    {
        delete newLength;
        newLength = new (::kSameHeap, myLength) short [ localLengthSize ];
    }
    newLengthSize = localLengthSize;
    memmove(newLength, localLength, newLengthSize * sizeof( short ) );
    if (!usingLocalBufferForLength )
        delete localLength;
}


/*================
||
||  IRectilinearGeometry::exclusiveOr
||
*/

IAreaGeometryHandle IRectilinearGeometry::cloneAndExclusiveOr(const IAreaGeometryHandle& handle)
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(*handle.geometry()))
    {
        IRectilinearGeometry*   r = dynamic_cast<const IRectilinearGeometry*>(handle.geometry());
#else
    if (isSameClass(*handle.geometry()))
    {
        IRectilinearGeometry*   r = (IRectilinearGeometry*) handle.geometry();
#endif // IC_RTTI_ENABLED
        GCoordinate*    newData = 0 /*NIL*/;
        long    newDataSize = 0;
        short*  newLength = 0 /*NIL*/;
        long    newLengthSize = 0;
        IGRect2D  newBounds;

        exclusiveOrData(  fData, fSizeData, fLength, fSizeLength,
                            r->fData, r->fSizeData, r->fLength, r->fSizeLength,
                            newData, newDataSize, newLength, newLengthSize, newBounds );

        return IAreaGeometryHandle(new (::kSameHeap, this) IRectilinearGeometry(newData, newDataSize, newLength, newLengthSize, newBounds));
    }
    else if (handle.isRectangular())
    {
        IRectilinearGeometry* newGeometry = (count() > 1) ? new (::kSameHeap, this) IRectilinearGeometry(*this) : this;
        newGeometry->exclusiveOr(IRectilinearGeometry( handle.bounds() ) );
        newGeometry->updateTimeStamp();
        return IAreaGeometryHandle(newGeometry);
    }
    else
        return IMAreaGeometry::cloneAndExclusiveOr(handle);
}

void IRectilinearGeometry::exclusiveOr(const IRectilinearGeometry& geometry)
{
    exclusiveOrData(  fData, fSizeData, fLength, fSizeLength,
                        geometry.fData, geometry.fSizeData, geometry.fLength, geometry.fSizeLength,
                        fData, fSizeData, fLength, fSizeLength, fBounds );
}


void IRectilinearGeometry::exclusiveOrData(
    GCoordinate* myData,
    long    myDataSize,
    short*  myLength,
    long    myLengthSize,
    GCoordinate* otherData,
    long    otherDataSize,
    short*  otherLength,
    long    otherLengthSize,
    GCoordinate*& newData,
    long&   newDataSize,
    short*& newLength,
    long&   newLengthSize,
    IGRect2D& newBounds )
{
    initStartEnd();

    // set or allocate the local length array.
    short   localLengthBuffer [ kLocalBufferSize ];
    short*  localLength;
    long allocatedSizeLength = myLengthSize + otherLengthSize - 1;  // the max it can be
    bool    usingLocalBufferForLength;
    if (allocatedSizeLength > kLocalBufferSize )
    {
        localLength = new (::kSameHeap, myLength) short [ allocatedSizeLength ];
        usingLocalBufferForLength = false;
    }
    else
    {
        localLength = localLengthBuffer;
        usingLocalBufferForLength = true;
    }

    // set the initial local data array.
    GCoordinate localDataBuffer [ kLocalBufferSize ];
    GCoordinate*    localData = localDataBuffer;
    long allocatedSizeData = kLocalBufferSize;
    bool    usingLocalBufferForData = true;

    long    freeData = allocatedSizeData;       // number of free data slots

    GCoordinate* data1 = myData;                // "this" rectilinear data
    GCoordinate* data2 = otherData;             // the parameter's rectilinear data
    GCoordinate* data3 = localData;             // the new rectilinear data
    GCoordinate *prevData3 = data3;

    short* len1 = myLength;                     // "this" length data
    short* len2 = otherLength;                  // the parameter's length data
    short* len3 = localLength;                  // the new length data

    newBounds.fLeft = kInfinity;
    newBounds.fRight = -kInfinity;

    while ((*data1 != *kEndList) || (*data2 != *kEndList))
    {
        GCoordinate* endD1 = &data1[*len1];
        GCoordinate* endD2 = &data2[*len2];
        GCoordinate* d3 = data3;
        {
            GCoordinate* d1 = data1;
            GCoordinate* d2 = data2;

            int state1 = false;
            int state2 = false;
            int state3 = false;

            *d3++ = *((*d1 < *d2) ? d2 : d1);   // store the new y value
            freeData--;
            d1++;
            d2++;

            while ((d1 != endD1) || (d2 != endD2))
            {
                GCoordinate d1val = *d1;
                GCoordinate d2val = *d2;
                if (d2 == endD2)
                {
                    state1 = ~state1;
                    d1++;
                }
                else if (( d1 == endD1 ) || (d1val > d2val) )
                {
                    state2 = ~state2;
                    d1val = d2val;
                    d2++;
                }
                else if (d1val < d2val )
                {
                    state1 = ~state1;
                    d1++;
                }
                else    // if (*d1 == *d2)
                {
                    state1 = ~state1;
                    state2 = ~state2;
                    d1++;
                    d2++;
                }

                int newState3 = (state1 != state2);
                if (newState3 != state3)
                {
                    if (--freeData <= 1 )  // out of space?
                    {
                        allocatedSizeData += kDataIncrement;
                        freeData += kDataIncrement;
                        GCoordinate* newLocalData = new (::kSameHeap, myData) GCoordinate [ allocatedSizeData ];
                        memmove(newLocalData, localData, ( d3 - localData ) * sizeof( GCoordinate ) );
                        d3 = newLocalData + (d3 - localData );
                        data3 = newLocalData + (data3 - localData );
                        prevData3 = newLocalData + (prevData3 - localData );
                        if (usingLocalBufferForData )
                            usingLocalBufferForData = false;
                        else
                            delete localData;
                        localData = newLocalData;
                    }
                    *d3++ = d1val;
                    state3 = newState3;
                }
            }
        }

        *len3 = d3 - data3;     // length of new y-event

        if (*endD1 <= *endD2)
        {
            data1 = endD1;
            len1++;
        }
        if (*endD2 <= *endD1)
        {
            data2 = endD2;
            len2++;
        }

        // Eliminate redundant y-events
        if (prevData3 != data3)     // just did first line?
        {
            bool redundant = false;
            if (len3[0] == len3[-1])
            {
                GCoordinate* data3x = data3 + 1;
                GCoordinate* prevData3x = prevData3 + 1;
                for(; (prevData3x != data3) && (*data3x == *prevData3x); data3x++, prevData3x++);

                redundant = (prevData3x == data3);
            }

            if (!redundant)
            {
                prevData3 = data3;
                data3 = d3;

                if (len3[0] > 1)
                {
                    if (prevData3[1] < newBounds.fLeft )
                        newBounds.fLeft = prevData3[1];
                    if (data3[-1] > newBounds.fRight )
                        newBounds.fRight = data3[-1];
                }
                len3++;
            }
        }
        else
        {
            data3 = d3;
            len3++;
        }
    }
    *data3++ = *kEndList;

    long localDataSize = data3 - localData;
    long localLengthSize = len3 - localLength;

    // Update the bounds.
    if (localDataSize == 2 )
        newBounds = IGRect2D::zeroRect();
    else
    {
        newBounds.fTop = (localLength[ 0 ] == 1 ) ? localData [ 1 ] : localData [ 0 ];
        newBounds.fBottom = data3 [ - len3 [ -1 ] - 1 ];
    }

    // Update the data array.
    if (localDataSize > newDataSize )
    {
        delete newData;
        newData = new (::kSameHeap, myData) GCoordinate [ localDataSize ];
    }
    newDataSize = localDataSize;
    memmove(newData, localData, newDataSize * sizeof( GCoordinate ) );
    if (!usingLocalBufferForData )
        delete localData;

    // Update the length array.
    if (localLengthSize > newLengthSize )
    {
        delete newLength;
        newLength = new (::kSameHeap, myLength) short [ localLengthSize ];
    }
    newLengthSize = localLengthSize;
    memmove(newLength, localLength, newLengthSize * sizeof( short ) );
    if (!usingLocalBufferForLength )
        delete localLength;
}


/*================
||
||  IRectilinearGeometry::contains IGPoint2D
||
*/
bool IRectilinearGeometry::contains(const IGPoint2D& p) const
{
    if (fBounds.contains(p))
    {
        short *length = fLength + 1;
        GCoordinate *data = fData + 1;

        while (*data <= p.fY)
            data += *length++;
        data -= *--length - 1;
        for(short l = *length; l > 1; l -= 2)
        {
            if ((p.fX >= *data) && (p.fX < *(data+1)))  // data points into x-values here
                return true;
            data += 2;
        }
    }
    return false;
}


/*================
||
||  IRectilinearGeometry::contains IGRect2D
||
*/
bool IRectilinearGeometry::contains(const IGRect2D& rect) const
{
    initStartEnd();

    if (fBounds.intersects(rect))
    {
        short *length = fLength + 1;
        GCoordinate *data = fData + 1;

        while (*data <= rect.fTop)
            data += *length++;
        data -= *--length;
        while ((rect.fBottom > *data) && (*data != *kEndList))
        {
            GCoordinate* x = data + 1;
            short l;
            for(l = *length; l > 1; l -= 2)
            {
                if ((rect.fLeft >= *x++) && (rect.fRight <= *x++))  // not every span, ANY span is OK
                {
                    l = -1;
                    break;
                }
            }
            if (l != -1)
                return false;       // no x-span completely contained the rect width
            data += *length++;
        }
        return (*data != *kEndList);
    }
    else
        return false;
}





/*================
||
||  IRectilinearGeometry::intersects
||
*/
bool IRectilinearGeometry::intersects(const IGRect2D& rect) const
{
    initStartEnd();

    if (fBounds.intersects(rect))
    {
        short *length = fLength + 1;
        GCoordinate *data = fData + 1;

        while (*data <= rect.fTop)
            data += *length++;
        data -= *--length;
        while ((rect.fBottom > *data) && (*data != *kEndList))
        {
            GCoordinate* x = data + 1;
            for(short l = *length; l > 1; l -= 2)
            {
                if ((rect.fRight > *x++) && (rect.fLeft < *x++))        // a single span overlaps, exit early
                    return true;
            }
            data += *length++;
        }
    }
    return false;
}



bool IRectilinearGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool IRectilinearGeometry::operator==(const IMAreaGeometry other) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(other))
#else
    if (isSameClass(other))
#endif // IC_RTTI_ENABLED
    {
        if (timeStamp() == other.timeStamp())
            return true;
        else if ((fBounds == other.fBounds) && (fSizeData == other.fSizeData) && (fSizeLength == other.fSizeLength))
        {
            GCoordinate* myData;
            GCoordinate* areaData;
            long i;
            for(i = 0, myData = fData, areaData = other.fData; i < fSizeData; i++, myData++, areaData++)
                if (*myData != *areaData)
                    return false;
            return true;
        }
        else
            return false;
    }
    else
        return false;
}

/*================
||
||  IRectilinearGeometry::operator>>=
||
*/

IBaseStream& IRectilinearGeometry::operator>>= (IBaseStream& toWhere) const
{
//    WriteVersion (toWhere);
    IMAreaGeometry::operator>>= (toWhere);

    fBounds >>= toWhere;
    fSizeData >>= toWhere;
    fSizeLength >>= toWhere;

    long i;
    GCoordinate *sd;
    short *ss;

    for(i = 0, sd = fData; i < fSizeData; i++, sd++)
        *sd >>= toWhere;
    for(i = 0, ss = fLength; i < fSizeLength; i++, ss++)
        *ss >>= toWhere;

    return toWhere;
}

/*================
||
||  IRectilinearGeometry::operator<<=
||
*/

IBaseStream& IRectilinearGeometry::operator<<= (IBaseStream& fromWhere)
{
//    VersionInfo theVersion = ReadVersion (fromWhere, 0, 0);
    IMAreaGeometry::operator<<= (fromWhere);

    fBounds <<= fromWhere;
    fSizeData <<= fromWhere;
    fSizeLength <<= fromWhere;

    delete fData;
    fData = new (::kSameHeap, this) GCoordinate[fSizeData];
    delete fLength;
    fLength = new (::kSameHeap, this) short[fSizeLength];

    long i;
    GCoordinate *sd;
    short *ss;

    for(i = 0, sd = fData; i < fSizeData; i++, sd++)
        *sd <<= fromWhere;
    for(i = 0, ss = fLength; i < fSizeLength; i++, ss++)
        *ss <<= fromWhere;

    return fromWhere;
}


#endif // NO_RECTILINEARGEOMETRY
// #endif // not SS



//===========================//
// IEllipseGeometry routines //
//===========================//

StreamableDefinitionsMacro(IEllipseGeometry, gGraphicsStreamModule);


/*================
||
||  IEllipseGeometry constructors
||
*/

IEllipseGeometry::IEllipseGeometry()
    : IGEllipse2D(IGRect2D::zeroRect())
{
}

IEllipseGeometry::IEllipseGeometry(const IGEllipse2D& ellipse)
    : IGEllipse2D(ellipse)
{
    updateTimeStamp();
}

IEllipseGeometry::IEllipseGeometry(const IEllipseGeometry& source)
    : IMAreaGeometry(source), IGEllipse2D(source)
{
}


/*================
||
||  IEllipseGeometry destructor
||
*/

IEllipseGeometry::~IEllipseGeometry()
{
}


/*================
||
||  IEllipseGeometry::extract
||
*/

void IEllipseGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
    ICAGEvenOdd* node = new (::kSameHeap, this) ICAGEvenOdd(parent);
    extractor.adoptCAGNode(node);
    extractor.extractEllipse(*this, node, matrix);
}


/*================
||
||  IEllipseGeometry::cloneAndTransform
||
*/

IAreaGeometryHandle IEllipseGeometry::cloneAndTransform(const IGrafMatrix& m)
{

// #ifndef SS
#ifdef NO_RECTILINEARGEOMETRY
    if (m.isRectilinear())
    {
        IGRect2D bounds = IGEllipse2D::bounds();
        IGPoint2D topLeft = m.transformPoint(bounds.topLeft());
        IGPoint2D bottomRight = m.transformPoint(bounds.bottomRight());
        IGRect2D newRect(topLeft, bottomRight);
        newRect.orderPoints();

        // If I'm shared, create a new object; else modify and return myself
        if (count() > 1)
            return IAreaGeometryHandle(new (::kSameHeap, this) IEllipseGeometry(IGEllipse2D(newRect)));
        else
        {
            *this = IGEllipse2D(newRect);
            updateTimeStamp();
            return IAreaGeometryHandle(this);
        }
    }
    else

#endif // NO_RECTILINEARGEOMETRY
// #endif // not SS

        return IMAreaGeometry::cloneAndTransform(m);
}


/*================
||
||  IEllipseGeometry::isEmpty
||
*/

bool IEllipseGeometry::isEmpty() const
{
    return bounds().isEmpty();
}


/*================
||
||  IEllipseGeometry Hit Testing
||
*/

IGRect2D IEllipseGeometry::bounds() const
{
    return IGEllipse2D::bounds();
}

bool IEllipseGeometry::intersects(const IGRect2D& rect) const
{
    return IGEllipse2D::intersects(rect);
}

bool IEllipseGeometry::contains(const IGPoint2D& point) const
{
    return IGEllipse2D::contains(point);
}

bool IEllipseGeometry::contains(const IGRect2D& rect) const
{
    return (IGEllipse2D::contains(rect.topLeft()) &&
            IGEllipse2D::contains(rect.topRight()) &&
            IGEllipse2D::contains(rect.bottomLeft()) &&
            IGEllipse2D::contains(rect.bottomRight()));
}


void IEllipseGeometry::writeToStream(IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
    IMAreaGeometry::writeToStream(toWhere);
    IGEllipse2D::writeToStream(toWhere);
}

void IEllipseGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    IMAreaGeometry::readFromStream(fromWhere);
    IGEllipse2D::readFromStream(fromWhere);
}

/*================
||
||  IEllipseGeometry operators
||
*/

IEllipseGeometry& IEllipseGeometry::operator=(const IEllipseGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);
        IGEllipse2D::operator=(source);
    }
    return *this;
}

bool IEllipseGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool IEllipseGeometry::operator==(const IMAreaGeometry& collectible) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(collectible))
    {
        const IEllipseGeometry& ellipseGeometry = dynamic_cast<const IEllipseGeometry&>(collectible);
#else
    if (isSameClass(collectible))
    {
        IEllipseGeometry& ellipseGeometry = *((IEllipseGeometry*)&collectible);
#endif // IC_RTTI_ENABLED
        return (IGEllipse2D::operator==(ellipseGeometry));
    }
    else
        return false;
}


//===========================//
// IPolygonGeometry routines //
//===========================//

StreamableDefinitionsMacro(IPolygonGeometry, gGraphicsStreamModule);


/*================
||
||  IPolygonGeometry constructors
||
*/

IPolygonGeometry::IPolygonGeometry()
{
    fBoundsValid = false;
}

IPolygonGeometry::IPolygonGeometry(const IGPolygon2D& polygon)
    : IGPolygon2D(polygon)
{
    fBoundsValid = false;
    updateTimeStamp();
}

IPolygonGeometry::IPolygonGeometry(const IPolygonGeometry& source)
    : IMAreaGeometry(source), IGPolygon2D(source),
      fBounds(source.fBounds), fBoundsValid(source.fBoundsValid)
{
}


/*================
||
||  IPolygonGeometry destructor
||
*/

IPolygonGeometry::~IPolygonGeometry() { }


/*================
||
||  IPolygonGeometry::extract
||
*/

void IPolygonGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{


    bool eoFill = eOFill();
    ICAGUnary* node;
    if (eoFill)
        node = new (::kSameHeap, this) ICAGEvenOdd(parent);
    else
        node = new (::kSameHeap, this) ICAGWindNum(parent);

    extractor.adoptCAGNode(node);
    extractor.extractPolygon (*this, node, matrix);
}


/*================
||
||  IPolygonGeometry::cloneAndTransform
||
*/

const unsigned long IPolygonGeometry::kLargePolygon = 16;

IAreaGeometryHandle IPolygonGeometry::cloneAndTransform(const IGrafMatrix& m)
{
    if (numberOfPoints() > kLargePolygon)
        return IMAreaGeometry::cloneAndTransform(m);
    else
    {
        // If I'm shared, return a new object; else return myself
        IPolygonGeometry* polygon;
        if (count() > 1)
            polygon = new (::kSameHeap, this) IPolygonGeometry(*this);
        else
            polygon = this;

        polygon->transformBy(m);
        polygon->fBoundsValid = false;
        polygon->updateTimeStamp();
        return IAreaGeometryHandle(polygon);
    }
}


/*================
||
||  IPolygonGeometry::isEmpty
||
*/

bool IPolygonGeometry::isEmpty() const
{
    return (numberOfPoints() < 3);   // should probably be more precise than this
}


/*================
||
||  IPolygonGeometry::isRectilinear
||
*/

bool IPolygonGeometry::isRectilinear() const
{
    return IGPolygon2D::isRectilinear();
}


/*================
||
||  IPolygonGeometry::isRectangular
||
*/

bool IPolygonGeometry::isRectangular() const
{
    // 2 points: true
    // 3 points: Triangle
    // 4 points: If rectilinear, MAY be a rectangle, but it may be degenerate!
    // 5 or more points: Forget it!

    return (numberOfPoints() <= 2);
}


/*================
||
||  IPolygonGeometry Hit Testing
||
*/

IGRect2D IPolygonGeometry::bounds() const
{
    if (!fBoundsValid)
    {
        ((IPolygonGeometry*) this)->fBounds = IGPolygon2D::bounds();
        ((IPolygonGeometry*) this)->fBoundsValid = true;
    }

    return fBounds;
}

bool IPolygonGeometry::intersects(const IGRect2D& rect) const
{
    return IGPolygon2D::intersects(rect);
}

bool IPolygonGeometry::contains(const IGPoint2D& p) const
{
    return IGPolygon2D::contains(p);
}

bool IPolygonGeometry::contains(const IGRect2D&) const
{
    return false;   // *** Not implemented ***
}


void IPolygonGeometry::writeToStream(IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
    IMAreaGeometry::writeToStream(toWhere);
    IGPolygon2D::writeToStream(toWhere);
}

void IPolygonGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    IMAreaGeometry::readFromStream(fromWhere);
    IGPolygon2D::readFromStream(fromWhere);
    fBoundsValid = false;
}

/*================
||
||  IPolygonGeometry operators
||
*/

IPolygonGeometry& IPolygonGeometry::operator=(const IPolygonGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);
        IGPolygon2D::operator=(source);
        fBounds = source.fBounds;
        fBoundsValid = source.fBoundsValid;
    }
    return *this;
}

bool IPolygonGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool IPolygonGeometry::operator==(const IMAreaGeometry& collectible) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(collectible))
    {
        const IPolygonGeometry& poly = dynamic_cast<const IPolygonGeometry&>(collectible);
#else
    if (isSameClass(collectible))
    {
        IPolygonGeometry& poly = *((IPolygonGeometry*)&collectible);
#endif // IC_RTTI_ENABLED
        return ((timeStamp() == poly.timeStamp()) || IGPolygon2D::operator==(poly));
    }
    else
        return false;
}


//========================//
// ILoopGeometry routines //
//========================//

StreamableDefinitionsMacro(ILoopGeometry, gGraphicsStreamModule);


/*================
||
||  ILoopGeometry constructors
||
*/

ILoopGeometry::ILoopGeometry()
{
    fBoundsValid = false;
}

ILoopGeometry::ILoopGeometry(const IGLoop2D& loop)
    : IGLoop2D(loop)
{
    fBoundsValid = false;
    updateTimeStamp();
}

ILoopGeometry::ILoopGeometry(const ILoopGeometry& source)
    : IMAreaGeometry(source), IGLoop2D(source),
      fBounds(source.fBounds), fBoundsValid(source.fBoundsValid)
{
}


/*================
||
||  ILoopGeometry destructor
||
*/

ILoopGeometry::~ILoopGeometry() { }


/*================
||
||  ILoopGeometry::extract
||
*/

void ILoopGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
    bool eoFill = eOFill();
    ICAGUnary* node;
    if (eoFill)
        node = new (::kSameHeap, this) ICAGEvenOdd(parent);
    else
        node = new (::kSameHeap, this) ICAGWindNum(parent);

    extractor.adoptCAGNode(node);
    extractor.extractLoop(*this, node, matrix);
}


/*================
||
||  ILoopGeometry::cloneAndTransform
||
*/

const unsigned long ILoopGeometry::kLargeLoop = 16;

IAreaGeometryHandle ILoopGeometry::cloneAndTransform(const IGrafMatrix& m)
{
    if (numberOfPoints() > kLargeLoop)
        return IMAreaGeometry::cloneAndTransform(m);
    else
    {
        // If I'm shared, return a new object; else return myself
        ILoopGeometry* loop;
        if (count() > 1)
            loop = new (::kSameHeap, this) ILoopGeometry(*this);
        else
            loop = this;

        loop->transformBy(m);
        loop->fBoundsValid = false;
        loop->updateTimeStamp();
        return IAreaGeometryHandle(loop);
    }
}


/*================
||
||  ILoopGeometry::isEmpty
||
*/

bool ILoopGeometry::isEmpty() const
{
    return ((order() == 0) || (numberOfPoints() < 3));    // should probably be more precise
}


/*================
||
||  ILoopGeometry Hit Testing
||
*/

IGRect2D ILoopGeometry::bounds() const
{
    if (!fBoundsValid)
    {
        ((ILoopGeometry*) this)->fBounds = IGLoop2D::bounds();
        ((ILoopGeometry*) this)->fBoundsValid = true;
    }

    return fBounds;
}

bool ILoopGeometry::intersects(const IGRect2D& rect) const
{
    return IGLoop2D::intersects(rect);
}

bool ILoopGeometry::contains(const IGPoint2D& p) const
{
    return IGLoop2D::contains(p);
}

bool ILoopGeometry::contains(const IGRect2D&) const
{
    return false;   // *** Not implemented ***
}


void ILoopGeometry::writeToStream(IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
    IMAreaGeometry::writeToStream(toWhere);
    IGLoop2D::writeToStream(toWhere);
}

void ILoopGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    IMAreaGeometry::readFromStream(fromWhere);
    IGLoop2D::readFromStream(fromWhere);
    fBoundsValid = false;
}

/*================
||
||  ILoopGeometry operators
||
*/

ILoopGeometry& ILoopGeometry::operator=(const ILoopGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);
        IGLoop2D::operator=(source);
        fBounds = source.fBounds;
        fBoundsValid = source.fBoundsValid;
    }
    return *this;
}

bool ILoopGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}
bool ILoopGeometry::operator==(const IMAreaGeometry& collectible) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(collectible))
    {
        const ILoopGeometry& geometry = dynamic_cast<const ILoopGeometry&>(collectible);
#else
    if (isSameClass(collectible))
    {
        ILoopGeometry& geometry = *((ILoopGeometry*) &collectible);
#endif // IC_RTTI_ENABLED
        return ((timeStamp() == geometry.timeStamp()) || IGLoop2D::operator==(geometry));
    }
    else
        return false;
}



#ifdef NOT_AVAIL

//===========================//
// IGlyphRunGeometry routines //
//===========================//

StreamableDefinitionsMacro(IGlyphRunGeometry, gGraphicsStreamModule);

/*================
||
||  IGlyphRunGeometry constructors
||
*/

IGlyphRunGeometry::IGlyphRunGeometry()
{
}

IGlyphRunGeometry::IGlyphRunGeometry(const IGGlyphRun& glyphrun)
    : IGGlyphRun (glyphrun)
{

// reject image fonts, as they can't be rendered by Gdi/Gpi path
    ParameterException (glyphrun.font()->isBitmap(), kImageFontUsedInArea );

    updateTimeStamp();
}

IGlyphRunGeometry::IGlyphRunGeometry(const IGlyphRunGeometry& source)
    : IMAreaGeometry(source), IGGlyphRun (source)
{
}

/*================
||
||  IGlyphRunGeometry destructor
||
*/

IGlyphRunGeometry::~IGlyphRunGeometry()
{
}


/*================
||
||  IGlyphRunGeometry::extract
||
*/

void IGlyphRunGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
    ICAGEvenOdd* node = new (::kSameHeap, this) ICAGEvenOdd(parent);
    extractor.adoptCAGNode(node);
    extractor.extractGlyphRun (*this, node, matrix);
}


/*================
||
||  IGlyphRunGeometry::cloneAndTransform
||
*/

IAreaGeometryHandle IGlyphRunGeometry::cloneAndTransform(const IGrafMatrix& m)
{
        return IMAreaGeometry::cloneAndTransform(m);
}


/*================
||
||  IGlyphRunGeometry::isEmpty
||
*/

bool IGlyphRunGeometry::isEmpty() const
{
    return bounds().isEmpty();
}


/*================
||
||  IGlyphRunGeometry Hit Testing
||
*/

IGRect2D IGlyphRunGeometry::bounds() const
{
    return IGGlyphRun::bounds();
}

bool IGlyphRunGeometry::intersects(const IGRect2D& rect) const
{
    return IGGlyphRun::bounds().intersects(rect);
}

bool IGlyphRunGeometry::contains(const IGPoint2D& point) const
{
    return IGGlyphRun::bounds().contains(point);
}

bool IGlyphRunGeometry::contains(const IGRect2D& rect) const
{
    return IGGlyphRun::bounds().contains(rect);

}


/*================
||
||  IGlyphRunGeometry streaming operators
||
*/
void IGlyphRunGeometry::writeToStream(IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
    IMAreaGeometry::writeToStream(toWhere);
    IGGlyphRun::writeToStream(toWhere);
}

void IGlyphRunGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    IMAreaGeometry::readFromStream(fromWhere);
    IGGlyphRun::readFromStream(fromWhere);
}

/*================
||
||  IGlyphRunGeometry operators
||
*/

IGlyphRunGeometry& IGlyphRunGeometry::operator=(const IGlyphRunGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);
        IGGlyphRun::operator=(source);
    }
    return *this;
}

bool IGlyphRunGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool IGlyphRunGeometry::operator==(const IMAreaGeometry& collectible) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(collectible))
    {
        const IGlyphRunGeometry& geometry = dynamic_cast<const IGlyphRunGeometry&>(collectible);
#else
    if (isSameClass(collectible))
    {
	IGlyphRunGeometry& geometry = *((IGlyphRunGeometry*) &obj);
#endif // IC_RTTI_ENABLED
        return (IGGlyphRun::operator==(geometry));
    }
    else
        return false;
}



#endif // NOT_AVAIL




//=======================//
// IAddGeometry routines //
//=======================//

StreamableDefinitionsMacro(IAddGeometry, gGraphicsStreamModule);


/*================
||
||  IAddGeometry constructors
||
*/

IAddGeometry::IAddGeometry()
{
    fBoundsValid = false;
}

IAddGeometry::IAddGeometry(const IAreaGeometryHandle& left, const IAreaGeometryHandle& right)
    : fLeft(left), fRight (right)
{
    fBoundsValid = false;
    updateTimeStamp();
}

IAddGeometry::IAddGeometry(const IAddGeometry& source)
    : IMAreaGeometry(source), fLeft(source.fLeft), fRight (source.fRight),
      fBounds(source.fBounds), fBoundsValid(source.fBoundsValid)
{
}


/*================
||
||  IAddGeometry destructor
||
*/

IAddGeometry::~IAddGeometry()
{
}


/*================
||
||  IAddGeometry cAG functions
||
*/

IAreaGeometryHandle IAddGeometry::cloneAndAdd(const IAreaGeometryHandle& handle)
{
    IAddGeometry* newGeometry;
    if (count() > 1)
        newGeometry = new (::kSameHeap, this) IAddGeometry(*this);
    else
        newGeometry = this;

    newGeometry->fLeft.add(handle);
    newGeometry->fBoundsValid = false;
    newGeometry->updateTimeStamp();
    return IAreaGeometryHandle(newGeometry);
}


/*================
||
||  IAddGeometry extract function
||
*/

void IAddGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
    ICAGAdd* node = new (::kSameHeap, this) ICAGAdd(parent);
    extractor.adoptCAGNode(node);
    fLeft.extract(extractor, node, matrix);
    fRight.extract(extractor, node->rightLink(), matrix);
}


/*================
||
||  IAddGeometry query functions
||
*/

bool IAddGeometry::isEmpty() const
{
    return fLeft.isEmpty() && fRight.isEmpty();
}

bool IAddGeometry::isRectilinear() const
{
    return fLeft.isRectilinear() && fRight.isRectilinear();
}

bool IAddGeometry::isSingular() const
{
    return false;
}


/*================
||
||  IAddGeometry hit functions
||
*/

IGRect2D IAddGeometry::bounds() const
{
    if (!fBoundsValid)
    {
        ((IAddGeometry*) this)->fBounds = fLeft.bounds();
        ((IAddGeometry*) this)->fBounds.extendTo (fRight.bounds());
        ((IAddGeometry*) this)->fBoundsValid = true;    // *** need ~const!
    }
    return fBounds;
}

bool IAddGeometry::intersects(const IGRect2D& rect) const
{
    return fLeft.intersects(rect) || fRight.intersects (rect);
}

bool IAddGeometry::contains(const IGPoint2D& point) const
{
    return bounds().contains(point) &&
            (fLeft.contains(point) || fRight.contains (point));
}

bool IAddGeometry::contains(const IGRect2D& rect) const
{
    return fLeft.contains(rect) || fRight.contains (rect);
}


/*================
||
||  IAddGeometry operator=
||
*/

IAddGeometry& IAddGeometry::operator=(const IAddGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);
        fLeft = source.fLeft;
        fRight = source.fRight;
        fBounds = source.fBounds;
        fBoundsValid = source.fBoundsValid;
    }
    return *this;
}


bool IAddGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool IAddGeometry::operator==(const IMAreaGeometry& collectible) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(collectible))
    {
        const IAddGeometry& geometry = dynamic_cast<const IAddGeometry&>(collectible);
#else
    if (isSameClass(collectible))
    {
        IAddGeometry& geometry = *((IAddGeometry*) &collectible);
#endif // IC_RTTI_ENABLED
        return ((timeStamp() == geometry.timeStamp()) ||
                ((fLeft == geometry.fLeft) && (fRight == geometry.fRight)));
    }
    else
        return false;
}


void IAddGeometry::writeToStream(IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
    IMAreaGeometry::writeToStream(toWhere);

    fLeft >>= toWhere;
    fRight >>= toWhere;
}

void IAddGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    IMAreaGeometry::readFromStream(fromWhere);

    fLeft <<= fromWhere;
    fRight <<= fromWhere;

    fBoundsValid = false;
}


//============================//
// ISubtractGeometry routines //
//============================//

StreamableDefinitionsMacro(ISubtractGeometry, gGraphicsStreamModule);


/*================
||
||  ISubtractGeometry constructors
||
*/

ISubtractGeometry::ISubtractGeometry()
{
    fBoundsValid = false;
}

ISubtractGeometry::ISubtractGeometry(const IAreaGeometryHandle& left, const IAreaGeometryHandle& right)
    : fLeft(left), fRight (right)
{
    fBoundsValid = false;
    updateTimeStamp();
}

ISubtractGeometry::ISubtractGeometry(const ISubtractGeometry& source)
    : IMAreaGeometry(source), fLeft(source.fLeft), fRight (source.fRight),
      fBounds(source.fBounds), fBoundsValid(source.fBoundsValid)
{
}


/*================
||
||  ISubtractGeometry destructor
||
*/

ISubtractGeometry::~ISubtractGeometry()
{
}


/*================
||
||  ISubtractGeometry cAG functions
||
*/

IAreaGeometryHandle ISubtractGeometry::cloneAndSubtract(const IAreaGeometryHandle& handle)
{
    ISubtractGeometry* newGeometry;
    if (count() > 1)
        newGeometry = new (::kSameHeap, this) ISubtractGeometry(*this);
    else
        newGeometry = this;

    newGeometry->fRight.add(handle);
    newGeometry->fBoundsValid = false;
    newGeometry->updateTimeStamp();
    return IAreaGeometryHandle(newGeometry);
}

IAreaGeometryHandle ISubtractGeometry::cloneAndIntersect(const IAreaGeometryHandle& handle)
{
    ISubtractGeometry* newGeometry;
    if (count() > 1)
        newGeometry = new (::kSameHeap, this) ISubtractGeometry(*this);
    else
        newGeometry = this;

    newGeometry->fLeft.intersect(handle);
    newGeometry->fBoundsValid = false;
    newGeometry->updateTimeStamp();
    return IAreaGeometryHandle(newGeometry);
}


/*================
||
||  ISubtractGeometry extract function
||
*/

void ISubtractGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
    ICAGSubtract* node = new (::kSameHeap, this) ICAGSubtract(parent);
    extractor.adoptCAGNode(node);
    fLeft.extract(extractor, node, matrix);
    fRight.extract(extractor, node->rightLink(), matrix);
}


/*================
||
||  ISubtractGeometry query functions
||
*/

bool ISubtractGeometry::isEmpty() const
{
    return fLeft.isEmpty() && fRight.isEmpty();
}

bool ISubtractGeometry::isRectilinear() const
{
    return fLeft.isRectilinear() && fRight.isRectilinear();
}

bool ISubtractGeometry::isSingular() const
{
    return false;
}


/*================
||
||  ISubtractGeometry hit functions
||
*/

IGRect2D ISubtractGeometry::bounds() const
{
    if (!fBoundsValid)
    {
        // Can't subtract the bounding rectangles because they are loose bounds!
        // But can we do better than this?

        ((ISubtractGeometry*) this)->fBounds = fLeft.bounds();
        ((ISubtractGeometry*) this)->fBoundsValid = true;   // *** need ~const!
    }
    return fBounds;
}

bool ISubtractGeometry::intersects(const IGRect2D& rect) const
{
    return fLeft.intersects(rect); // *** Not fully implemented
}

bool ISubtractGeometry::contains(const IGPoint2D& point) const
{
    return bounds().contains(point) && (fLeft.contains(point) && !fRight.contains(point));
}

bool ISubtractGeometry::contains(const IGRect2D& rect) const
{
    return fLeft.contains(rect) && !fRight.intersects(rect);
}


/*================
||
||  ISubtractGeometry operator=
||
*/

ISubtractGeometry& ISubtractGeometry::operator=(const ISubtractGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);
        fLeft = source.fLeft;
        fRight = source.fRight;
        fBounds = source.fBounds;
        fBoundsValid = source.fBoundsValid;
    }
    return *this;
}


bool ISubtractGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool ISubtractGeometry::operator==(const IMAreaGeometry& collectible) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(collectible))
    {
        const ISubtractGeometry& geometry = dynamic_cast<const ISubtractGeometry&>(collectible);
#else
    if (isSameClass(collectible))
    {
        ISubtractGeometry& geometry = *((ISubtractGeometry*)&collectible);
#endif // IC_RTTI_ENABLED
        return ((timeStamp() == geometry.timeStamp()) ||
                ((fLeft == geometry.fLeft) && (fRight == geometry.fRight)));
    }
    else
        return false;
}

void ISubtractGeometry::writeToStream(IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
    IMAreaGeometry::writeToStream(toWhere);

    fLeft >>= toWhere;
    fRight >>= toWhere;
}

void ISubtractGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    IMAreaGeometry::readFromStream(fromWhere);

    fLeft <<= fromWhere;
    fRight <<= fromWhere;

    fBoundsValid = false;
}


//=============================//
// IIntersectGeometry routines //
//=============================//

StreamableDefinitionsMacro(IIntersectGeometry, gGraphicsStreamModule);


/*================
||
||  IIntersectGeometry constructors
||
*/

IIntersectGeometry::IIntersectGeometry()
{
    fBoundsValid = false;
}

IIntersectGeometry::IIntersectGeometry(const IAreaGeometryHandle& left, const IAreaGeometryHandle& right)
    : fLeft(left), fRight (right)
{
    fBoundsValid = false;
    updateTimeStamp();
}

IIntersectGeometry::IIntersectGeometry(const IIntersectGeometry& source)
    : IMAreaGeometry(source), fLeft(source.fLeft), fRight (source.fRight),
      fBounds(source.fBounds), fBoundsValid(source.fBoundsValid)
{
}


/*================
||
||  IIntersectGeometry destructor
||
*/

IIntersectGeometry::~IIntersectGeometry()
{
}


/*================
||
||  IIntersectGeometry cAG functions
||
*/

IAreaGeometryHandle IIntersectGeometry::cloneAndIntersect(const IAreaGeometryHandle& handle)
{
    IIntersectGeometry* newGeometry;
    if (count() > 1)
        newGeometry = new (::kSameHeap, this) IIntersectGeometry(*this);
    else
        newGeometry = this;

    newGeometry->fLeft.intersect(handle);
    newGeometry->fBoundsValid = false;
    newGeometry->updateTimeStamp();
    return IAreaGeometryHandle(newGeometry);
}


/*================
||
||  IIntersectGeometry extract function
||
*/

void IIntersectGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
    ICAGIntersect* node = new (::kSameHeap, this) ICAGIntersect(parent);
    extractor.adoptCAGNode(node);
    fLeft.extract(extractor, node, matrix);
    fRight.extract(extractor, node->rightLink(), matrix);
}


/*================
||
||  IIntersectGeometry query functions
||
*/

bool IIntersectGeometry::isEmpty() const
{
    return (bounds().isEmpty()) || (fLeft.isEmpty() && fRight.isEmpty());
}

bool IIntersectGeometry::isRectilinear() const
{
    return fLeft.isRectilinear() && fRight.isRectilinear();
}

bool IIntersectGeometry::isSingular() const
{
    return false;
}


/*================
||
||  IIntersectGeometry hit functions
||
*/

IGRect2D IIntersectGeometry::bounds() const
{
    if (!fBoundsValid)
    {
        ((IIntersectGeometry*) this)->fBounds = fLeft.bounds();
        ((IIntersectGeometry*) this)->fBounds.intersectWith(fRight.bounds());
        ((IIntersectGeometry*) this)->fBoundsValid = true;  // *** need ~const!
    }
    return fBounds;
}

bool IIntersectGeometry::intersects(const IGRect2D& rect) const
{
    return fLeft.intersects(rect) && fRight.intersects (rect); // *** Not fully implemented
}

bool IIntersectGeometry::contains(const IGPoint2D& point) const
{
    return bounds().contains(point) &&
            (fLeft.contains(point) && fRight.contains (point));
}

bool IIntersectGeometry::contains(const IGRect2D& rect) const
{
    return fLeft.contains(rect) && fRight.contains (rect);
}


/*================
||
||  IIntersectGeometry operator=
||
*/

IIntersectGeometry& IIntersectGeometry::operator=(const IIntersectGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);
        fLeft = source.fLeft;
        fRight = source.fRight;
        fBounds = source.fBounds;
        fBoundsValid = source.fBoundsValid;
    }
    return *this;
}


bool IIntersectGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool IIntersectGeometry::operator==(const IMAreaGeometry& collectible) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(collectible))
    {
        const IIntersectGeometry& geometry = dynamic_cast<const IIntersectGeometry&>(collectible);
#else
    if (isSameClass(collectible))
    {
        IIntersectGeometry& geometry = *((IIntersectGeometry*) &collectible);
#endif // IC_RTTI_ENABLED
        return ((timeStamp() == geometry.timeStamp()) ||
                ((fLeft == geometry.fLeft) && (fRight == geometry.fRight)));
    }
    else
        return false;
}


void IIntersectGeometry::writeToStream(IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
    IMAreaGeometry::writeToStream(toWhere);

    fLeft >>= toWhere;
    fRight >>= toWhere;
}

void IIntersectGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    IMAreaGeometry::readFromStream(fromWhere);

    fLeft <<= fromWhere;
    fRight <<= fromWhere;

    fBoundsValid = false;
}


//===============================//
// IExclusiveOrGeometry routines //
//===============================//

StreamableDefinitionsMacro(IExclusiveOrGeometry, gGraphicsStreamModule);


/*================
||
||  IExclusiveOrGeometry constructors
||
*/

IExclusiveOrGeometry::IExclusiveOrGeometry()
{
    fBoundsValid = false;
}

IExclusiveOrGeometry::IExclusiveOrGeometry(const IAreaGeometryHandle& left, const IAreaGeometryHandle& right)
    : fLeft(left), fRight (right)
{
    fBoundsValid = false;
    updateTimeStamp();
}

IExclusiveOrGeometry::IExclusiveOrGeometry(const IExclusiveOrGeometry& source)
    : IMAreaGeometry(source), fLeft(source.fLeft), fRight (source.fRight),
      fBounds(source.fBounds), fBoundsValid(source.fBoundsValid)
{
}


/*================
||
||  IExclusiveOrGeometry destructor
||
*/

IExclusiveOrGeometry::~IExclusiveOrGeometry()
{
}


/*================
||
||  IExclusiveOrGeometry cAG functions
||
*/

IAreaGeometryHandle IExclusiveOrGeometry::cloneAndExclusiveOr(const IAreaGeometryHandle& handle)
{
    IExclusiveOrGeometry* newGeometry;
    if (count() > 1)
        newGeometry = new (::kSameHeap, this) IExclusiveOrGeometry(*this);
    else
        newGeometry = this;

    newGeometry->fLeft.exclusiveOr(handle);
    newGeometry->fBoundsValid = false;
    newGeometry->updateTimeStamp();
    return IAreaGeometryHandle(newGeometry);
}


/*================
||
||  IExclusiveOrGeometry extract function
||
*/

void IExclusiveOrGeometry::extract(
    IGrafExtractor& extractor,
    ICAGNode* parent,
    const IGrafMatrix& matrix) const
{
    ICAGExclusiveOr* node = new (::kSameHeap, this) ICAGExclusiveOr(parent);
    extractor.adoptCAGNode(node);
    fLeft.extract(extractor, node, matrix);
    fRight.extract(extractor, node->rightLink(), matrix);
}


/*================
||
||  IExclusiveOrGeometry query functions
||
*/

bool IExclusiveOrGeometry::isEmpty() const
{
    return fLeft.isEmpty() && fRight.isEmpty();
}

bool IExclusiveOrGeometry::isRectilinear() const
{
    return fLeft.isRectilinear() && fRight.isRectilinear();
}

bool IExclusiveOrGeometry::isSingular() const
{
    return false;
}


/*================
||
||  IExclusiveOrGeometry hit functions
||
*/

IGRect2D IExclusiveOrGeometry::bounds() const
{
    if (!fBoundsValid)
    {
        ((IExclusiveOrGeometry*) this)->fBounds = fLeft.bounds();
        ((IExclusiveOrGeometry*) this)->fBounds.extendTo (fRight.bounds());
        ((IExclusiveOrGeometry*) this)->fBoundsValid = true;    // *** need ~const!
    }
    return fBounds;
}

bool IExclusiveOrGeometry::intersects(const IGRect2D& rect) const
{
    return fLeft.intersects(rect) || fRight.intersects (rect); // *** Not fully implemented
}

bool IExclusiveOrGeometry::contains(const IGPoint2D& point) const
{
    if (bounds().contains(point))
    {
        bool leftContains = fLeft.contains(point);
        bool rightContains = fRight.contains(point);
        return ((leftContains && !rightContains) || (!leftContains && rightContains));
    }
    else
        return false;
}

bool IExclusiveOrGeometry::contains(const IGRect2D& rect) const
{
    return (fLeft.contains(rect) && !fRight.intersects (rect)) ||
            (fRight.contains(rect) && !fLeft.intersects (rect));
}


/*================
||
||  IExclusiveOrGeometry operator=
||
*/

IExclusiveOrGeometry& IExclusiveOrGeometry::operator=(const IExclusiveOrGeometry& source)
{
    if (this != &source)
    {
        IMAreaGeometry::operator=(source);
        fLeft = source.fLeft;
        fRight = source.fRight;
        fBounds = source.fBounds;
        fBoundsValid = source.fBoundsValid;
    }
    return *this;
}


bool IExclusiveOrGeometry::operator!=(const IMAreaGeometry& obj) const
{
	return !operator==(obj);
}

bool IExclusiveOrGeometry::operator==(const IMAreaGeometry& collectible) const
{
#if IC_RTTI_ENABLED
    if (typeid(*this) == typeid(collectible))
    {
        const IExclusiveOrGeometry& geometry = dynamic_cast<const IExclusiveOrGeometry&>(collectible);
#else
    if (isSameClass(collectible))
    {
        IExclusiveOrGeometry& geometry = *((IExclusiveOrGeometry*)&collectible);
#endif // IC_RTTI_ENABLED
        return (timeStamp() == geometry.timeStamp() ||
                (fLeft == geometry.fLeft && fRight == geometry.fRight));
    }
    else
        return false;
}


void IExclusiveOrGeometry::writeToStream(IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
    IMAreaGeometry::writeToStream(toWhere);

    fLeft >>= toWhere;
    fRight >>= toWhere;
}

void IExclusiveOrGeometry::readFromStream(IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
    IMAreaGeometry::readFromStream(fromWhere);

    fLeft <<= fromWhere;
    fRight <<= fromWhere;

    fBoundsValid = false;
}


