/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 49 1.25.1.3 source/albert/graph2d/grafline.cpp, 2d, ioc.v400, 000302 
/*================
||
||  File:   GraphicLines.C
||
||  What:   Albert 2D Line Primitives code
||
*/

#include <igraph2d.hpp>
#include <igrport.hpp>
#include <igrafatt.hpp>
#include <grtype.hpp>
#include <idatstrm.hpp>
#include <grstrmod.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
#define _GRAFLINE_CPP_
#include <ipagetun.h>
#endif

/*=============================================
||
||  Line Oriented Graphic Primitives
||
*/

/*================================
||  ILine::
||
*/

StreamableDefinitionsMacro(ILine, gGraphicsStreamModule);

// ----- Constructors and Destructors -----

ILine::ILine() : IMGraphic(), fLine(0 /*NIL*/)
{}

ILine::ILine(const ILine& line) : IMGraphic(line), fLine(0 /*NIL*/)
{
	if (line.fLine != 0 /*NIL*/) {
		fLine = new IGLine2D(*line.fLine);
	}
}

ILine::ILine(IGLine2D* line, IGrafBundle* bundle) : IMGraphic(bundle), fLine(line)
{}

ILine::~ILine()
{
	delete fLine;
}

void ILine::adoptGeometry(IGLine2D* line)
{
	delete fLine;
	fLine = line;
}

IGLine2D* ILine::orphanGeometry()
{
	IGLine2D* line = fLine;
	fLine = 0 /*NIL*/;
	return line;
}

const IGLine2D* ILine::geometry()
{
	return fLine;
}

void ILine::draw(IGrafPort& port) const
{
	if (fLine)
	{
		ILinkedGrafPort bundlePort(&port, attributeState());
		bundlePort.draw(*fLine);
	}
}

void ILine::transformBy(const IGrafMatrix& m)
{
	if (fLine == 0 /*NIL*/) return;

	fLine->transformBy(m);
	updateTimeStamp();
}

IGRect2D ILine::geometricBounds() const
{
	if (fLine == 0 /*NIL*/)
		return IGRect2D();

	return fLine->bounds();
}

bool ILine::intersects (const IGRect2D &  rect) const
{
  bool  retval = false;
  if (isHitEnabled () == true)
  {
    if (fLine != 0 /* NIL */)
    {
      retval = fLine->intersects (rect);
    }

    ((ILine *) this)->setHit (retval);  // for performance reasons,
                                        // hit state is cached.
  }

  return (retval);

} // ILine::intersects


IGRect2D ILine::looseFitBounds(const IGrafPort* port) const
{
	return IMGraphic::looseFitBounds(port);
}


void ILine::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
    IMGraphic::writeToStream(toWhere);
    bool nilPointer = (fLine == 0 /*NIL*/) ? true : false;
    long(nilPointer) >>= toWhere;
    if (! nilPointer)
    	*fLine >>=(toWhere);
}

void ILine::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
    IMGraphic::readFromStream(fromWhere);

    bool nilPointer;
    long temp;
    temp <<= fromWhere;
    nilPointer = bool(temp);
    if (! nilPointer)
    {
		if (fLine == 0 /*NIL*/)
            fLine = new IGLine2D;
        *fLine <<=(fromWhere);
    }
    else
    {
       delete fLine;
       fLine = 0 /*NIL*/;
    }

}

ILine& ILine::operator=(const ILine& other)
{
    if (&other != this)
    {
        IMGraphic::operator=(other);
	    if (other.fLine != 0 /*NIL*/)
	    {
	    	if (fLine == 0 /*NIL*/)
	    		fLine = new IGLine2D(*other.fLine);
	    	else
	    		*fLine = (*other.fLine);
	    }else
	    {
	    	delete fLine;
	    	fLine = 0 /*NIL*/;
	    }
        fBoundsTimeStamp = other.fBoundsTimeStamp;
    }
    return(*this);
}

bool ILine::operator==(const IMDrawable& src) const
{
	if (!IMGraphic::operator==(src))
		return false;

	ILine* srcGraphic = (ILine*)&src;
	return (fLine == srcGraphic->fLine || *fLine == *srcGraphic->fLine);
}

/*================================
||  IPolyline::
||
*/

// ----- Constructors and Destructors -----

StreamableDefinitionsMacro(IPolyline, gGraphicsStreamModule);

IPolyline::IPolyline() : IMGraphic(), fBounds(0 /*NIL*/), fPolyline(0 /*NIL*/)
{}

IPolyline::IPolyline(const IPolyline& polyline) :
	IMGraphic(polyline),
	fBounds(0 /*NIL*/),
	fPolyline(0 /*NIL*/)
{
	if (polyline.fPolyline != 0 /*NIL*/) {
		fPolyline = new IGPolyline2D(*polyline.fPolyline);
	}
}

IPolyline::IPolyline(IGPolyline2D* polyline, IGrafBundle* bundle) :
	IMGraphic(bundle),
	fPolyline(polyline),
	fBounds(0 /*NIL*/)
{}

IPolyline::~IPolyline()
{
	delete fBounds;
	delete fPolyline;
}

void IPolyline::adoptGeometry(IGPolyline2D* polyline)
{
	delete fPolyline;
	fPolyline = polyline;
}

IGPolyline2D* IPolyline::orphanGeometry()
{
	IGPolyline2D* polyline = fPolyline;
	fPolyline = 0 /*NIL*/;
	return polyline;
}

const IGPolyline2D* IPolyline::geometry()
{
	return fPolyline;
}

void IPolyline::draw(IGrafPort& port) const
{
	if (fPolyline)
	{
		ILinkedGrafPort bundlePort(&port, attributeState());
		bundlePort.draw(*fPolyline);
	}
}

void IPolyline::transformBy(const IGrafMatrix& m)
{
	if (fPolyline == 0 /*NIL*/) return;

	fPolyline->transformBy(m);
	updateTimeStamp();
}

IGRect2D IPolyline::geometricBounds() const
{
	if (fPolyline == 0 /*NIL*/) return IGRect2D();
	return fPolyline->bounds();
}


bool IPolyline::intersects (const IGRect2D &  rect) const
{
  bool  retval = false;
  if (isHitEnabled () == true)
  {
    if (fPolyline != 0 /* NIL */)
    {
      retval = fPolyline->intersects (rect);
    }

    ((IPolyline *) this)->setHit (retval);  // for performance reasons,
                                            // hit state is cached.
  }

  return (retval);

} // IPolyline::intersects


IGRect2D IPolyline::looseFitBounds(const IGrafPort* port) const
{
	return IMGraphic::looseFitBounds(port);
}


void IPolyline::writeToStream (IDataStream& toWhere) const
{
    IStreamOutFrame  streamFrame (toWhere);
    IMGraphic::writeToStream (toWhere);

    bool  polylineNotNil (fPolyline != 0 /* NIL */);
    long (polylineNotNil) >>= toWhere;
    if (polylineNotNil)
    {
      *fPolyline >>= toWhere;
    }

    bool  boundsNotNil (fBounds != 0 /*NIL*/);
    long (boundsNotNil) >>= toWhere;
    if (boundsNotNil)
    {
      *fBounds >>= toWhere;
    }

    fBoundsTimeStamp >>= toWhere;

} // IPolyline::writeFromStream



void IPolyline::readFromStream (IDataStream& fromWhere)
{
    IStreamInFrame  streamFrame (fromWhere);
    IMGraphic::readFromStream (fromWhere);

    long  temp;
    temp <<= fromWhere;
    bool  notNil = bool (temp);
    if (notNil)
    {
      if (fPolyline == 0 /* NIL */)
      {
	fPolyline = new IGPolyline2D;
      }

      *fPolyline <<= fromWhere;
    }
    else
    {
      delete fPolyline;
      fPolyline = 0 /* NIL */;
    }


    temp <<= fromWhere;
    notNil = bool (temp);
    if (notNil)
    {
      if (fBounds == 0 /* NIL */)
      {
    	fBounds = new IGRect2D;
      }

      *fBounds <<= fromWhere;
    }
    else
    {
      delete fBounds;
      fBounds = 0 /*NIL*/;
    }

    fBoundsTimeStamp <<= fromWhere;

} // IPolyline::readFromStream



IPolyline& IPolyline::operator= (const IPolyline& other)
{
  if (&other != this)
  {
    IMGraphic::operator= (other);

    if (other.fPolyline != 0 /* NIL */)
    {
      if (fPolyline == 0 /* NIL */)
	fPolyline = new IGPolyline2D (*(other.fPolyline));
      else
	*fPolyline = *(other.fPolyline);
    }
    else
    {
      delete fPolyline;
      fPolyline = 0 /* NIL */;
    }


    if (other.fBounds != 0 /* NIL */)
    {
      if (fBounds == 0 /* NIL */)
	fBounds = new IGRect2D (*(other.fBounds));
      else
	*fBounds = *(other.fBounds);
    }
    else
    {
      delete fBounds;
      fBounds = 0 /* NIL */;
    }

    fBoundsTimeStamp = other.fBoundsTimeStamp;

  } // if not self assignment

  return (*this);

} // IPolyline::operator=


bool IPolyline::operator==(const IMDrawable& src) const
{
	if (!IMGraphic::operator==(src))
		return false;

	IPolyline* srcGraphic = (IPolyline*)&src;
	return (fPolyline == srcGraphic->fPolyline || *fPolyline == *srcGraphic->fPolyline);
}

/*================================
||  ICurve::
||
*/

// ----- Constructors and Destructors -----

ICurve::ICurve() : IMGraphic(), fBounds(0 /*NIL*/), fCurve(0 /*NIL*/)
{}

ICurve::ICurve(const ICurve& curve) : IMGraphic(curve), fBounds(0 /*NIL*/), fCurve(0 /*NIL*/)
{
	if (curve.fCurve != 0 /*NIL*/) {
		fCurve = new IGCurve2D(*curve.fCurve);
		fMatrix = curve.fMatrix;
	}
}

ICurve::ICurve(IGCurve2D* curve, IGrafBundle* bundle) :
	IMGraphic(bundle),
	fCurve(curve),
	fBounds(0 /*NIL*/)
{}

ICurve::~ICurve()
{
	delete fBounds;
	delete fCurve;
}

void ICurve::adoptGeometry(IGCurve2D* curve)
{
	delete fCurve;
	fCurve = curve;
}

IGCurve2D* ICurve::orphanGeometry()
{
	IGCurve2D* curve = fCurve;
	fCurve = 0 /*NIL*/;
	return curve;
}

const IGCurve2D* ICurve::geometry()
{
	return fCurve;
}


void ICurve::draw(IGrafPort& port) const
{
	if (fCurve)
	{
    	ILinkedGrafPort bundlePort(&port, attributeState());
    	ILinkedGrafPort port(&bundlePort, kModelMatrix, &fMatrix);
    	port.draw(*fCurve);
	}
}

void ICurve::transformBy(const IGrafMatrix& m)
{
	if (fCurve == 0 /*NIL*/) return;

	fMatrix.concatWith(m);
	updateTimeStamp();
}

IGRect2D ICurve::geometricBounds() const
{
	if (fCurve == 0 /*NIL*/) return IGRect2D();
	if (fMatrix.isIdentity())
		return fCurve->bounds();
	else
		return fMatrix.transformBounds(fCurve->bounds());
}


bool ICurve::intersects (const IGRect2D &  rect) const
{
  bool  retval = false;
  if (isHitEnabled () == true)
  {
    if (fCurve != 0 /* NIL */)
    {
      retval = fCurve->intersects (rect);
    }

    ((ICurve *) this)->setHit (retval);  // for performance reasons,
                                         // hit state is cached.
  }

  return (retval);

} // ICurve::intersects


IGRect2D ICurve::looseFitBounds(const IGrafPort* port) const
{
	return IMGraphic::looseFitBounds(port);
}


StreamableDefinitionsMacro(ICurve, gGraphicsStreamModule);

void ICurve::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
    IMGraphic::writeToStream(toWhere);
    bool nilPointer = (fCurve == 0 /*NIL*/) ? true : false;
    long(nilPointer) >>= toWhere;
    if (! nilPointer)
        *fCurve >>=(toWhere);
}

void ICurve::readFromStream(IDataStream& fromWhere)
{
    IStreamInFrame streamFrame(fromWhere);
    IMGraphic::readFromStream(fromWhere);

    bool nilPointer;
    long temp;
    temp <<= fromWhere;
    nilPointer = bool(temp);
    if (! nilPointer)
    {
	if (fCurve == 0 /*NIL*/)
            fCurve = new IGCurve2D;
        *fCurve <<=(fromWhere);
    }
    else
    {
       delete fCurve;
       fCurve = 0 /*NIL*/;
    }

    delete fBounds;
    fBounds = 0 /*NIL*/;
}

ICurve& ICurve::operator=(const ICurve& other)
{
    if (&other != this)
    {
        IMGraphic::operator=(other);
	    if (other.fCurve != 0 /*NIL*/)
	    {
	    	if (fCurve == 0 /*NIL*/) 	
	    		fCurve = new IGCurve2D(*other.fCurve);
			else
	    		*fCurve = (*other.fCurve);
	    }else
	    {
	    	delete fCurve;
	    	fCurve = 0 /*NIL*/;
	    }


	    if (other.fBounds != 0 /*NIL*/)
	    {
	    	if (fBounds == 0 /*NIL*/) 	
		  fBounds = new IGRect2D (*(other.fBounds));
		else
		  *fBounds = *(other.fBounds);
	    }
	    else
	    {
	    	delete fBounds;
	    	fBounds = 0 /*NIL*/;
	    }

        fBoundsTimeStamp = other.fBoundsTimeStamp;
    }
    return(*this);
}

bool ICurve::operator==(const IMDrawable& src) const
{
	if (!IMGraphic::operator==(src))
		return false;

	ICurve* srcGraphic = (ICurve*)&src;
	return (fCurve == srcGraphic->fCurve || *fCurve == *srcGraphic->fCurve);
}
