// IBM Open Class Library
//
// Licensed Materials - Property of IBM
//
// "Restricted Materials of IBM"
//
// (C) Copyright IBM Corporation, 1992, 1997 All Rights Reserved
//
// Revision: 34 1.11.1.5 source/albert/graph2d/graftext/istdprnd.cpp, 2d, ioc.v400, 000302 
#if !(__IBMCPP__ >= 400)
template <class C>
inline C min(const C& x, const C& y) { return (x < y) ? x : y; };

template <class C>
inline C max(const C& x, const C& y) { return (x > y) ? x : y; };
#endif

//................................................................................

#include <istdprnd.hpp>
#include <istdcrnd.hpp>
#include <icrend.hpp>
#include <iunicode.hpp>
#include <itxtbdry.hpp>
#include <istyles.hpp>
//#include "iunicode.hpp"
#include <iugnrl.hpp>
#include <ibundles.hpp> // for IFillBundle
#include <isynchr.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
#define _ISTDPRND_CPP_
#include <ipagetun.h>
#endif

//................................................................................

#pragma pack(push,4)

class IStyleRunIterator {
public:
							IStyleRunIterator(const ITextIterator& start, const ITextIterator& limit);
							~IStyleRunIterator();

	IStyleRunIterator&		operator=(const IStyleRunIterator& source);

	bool 					isValid();
	ITextIterator			runStart();
	ITextIterator			runLimit();
	const ITextStyleSet*	style();
	void					next();

private:
	ITextIterator			fRunStart;
	ITextIterator			fRunLimit;
	ITextIterator			fLimit;
	const ITextStyleSet*	fStyle;
};

#pragma pack(pop)

IStyleRunIterator::IStyleRunIterator(const ITextIterator& iter, const ITextIterator& limit)
	: fRunStart(iter), fRunLimit(iter), fStyle(NULL), fLimit(limit)
{
	next();
}

IStyleRunIterator::~IStyleRunIterator()
{
}

IStyleRunIterator& IStyleRunIterator::operator=(const IStyleRunIterator& source)
{
	if (this != &source) {
		fRunStart = source.fRunStart;
		fRunLimit = source.fRunLimit;
		fStyle = source.fStyle;
		fLimit = source.fLimit;
	}
	return *this;
}

ITextIterator IStyleRunIterator::runStart()
{
	return fRunStart;
}

ITextIterator IStyleRunIterator::runLimit()
{
	return fRunLimit;
}

bool IStyleRunIterator::isValid()
{
	return fRunStart < fLimit;
}

const ITextStyleSet* IStyleRunIterator::style()
{
	return fStyle;
}

void IStyleRunIterator::next()
{
	fRunStart = fRunLimit;

	if (fRunLimit < fLimit) {
		IText::offset_type runStart;
		IText::length_type runLength;

		fStyle = fRunStart.text()->stylesAt(fRunLimit.offset(), runStart, runLength);
		fRunLimit = fRunStart.text()->begin() + (runStart + runLength);
		if (fRunLimit > fLimit) {
			fRunLimit = fLimit;
		}
	}
}

//................................................................................

#pragma pack(push,4)

class IParagraphIterator {
public:
	IParagraphIterator(const ITextIterator& iter);
	~IParagraphIterator();

	ITextIterator			runLimit();
	void					next();

private:
	IText*					fText;
	ITextIterator			fOffset;
};

#pragma pack(pop)

IParagraphIterator::IParagraphIterator(const ITextIterator& iter)
	: fText(iter.text()), fOffset(iter)
{
	next();
}

IParagraphIterator::~IParagraphIterator()
{
}

ITextIterator IParagraphIterator::runLimit()
{
	return fOffset;
}

void
IParagraphIterator::next()
{
	if (fOffset < fText->end()) {
		fOffset++;
		size_t paraOffset = fText->end().offset();
		ICharacterPropertyIterator prop(IUnicode::kParagraphTerminator, IUnicode::kParagraphTerminator);
		while (prop) {
			UniChar delim = *prop;
			size_t temp = fText->find(delim, fOffset.offset());
			if (temp < paraOffset) {
				paraOffset = temp;
			}
			prop++;
		}

		fOffset = fText->begin() + paraOffset;
	}
}

//................................................................................

#pragma pack(push,4)

class ILineCursor {

private:

	IGrafPort& fGrafPort;

	// current position is segment start on each call to nextSegment
	// limit is first offset after paragraph separator, or end of text
	ITextIterator fStart, fCurrentPosition, fLimit;
	
	// vertical line metrics
	GCoordinate fAscent, fDescent, fLeading;
	
	// line metrics
	GCoordinate fTotalAdvance, fVisibleAdvance;
	IText::length_type fLength;
	
	// segment metrics
	// reset at start of each segment
	// segment decimal advance is the advance from the start of the segment to the first decimal
	GCoordinate fSegmentTotalAdvance, fSegmentVisibleAdvance, fSegmentDecimalAdvance;
	IText::length_type fSegmentLength;
	
	// segment flag
	bool fAtTabStop;
	
	// line flag;  don't call nextSegment once this is true
	// true iff fCurrentPosition == fLimit;
	bool fAtLineEnd;
	
	// first segment flag - used in fitFirstWord
	bool fFirstSegmentDone;
	
	// char renderer can measure a run of text in a single style
	ICharacterRendererCache fCharRenderer;
	
public:

	ILineCursor(const ITextIterator& start, const ITextIterator& limit, IGrafPort& g);
	
	void 					setInfo(ILayoutInfo& info) const;
	
	bool 					atTabStop() const;
	bool 					atLineEnd() const;
	
	GCoordinate 			segmentTotalAdvance() const;
	GCoordinate 			segmentVisibleAdvance() const;
	GCoordinate 			segmentDecimalAdvance() const;
	IText::length_type 		segmentLength() const;

	void 					nextSegment(
									GCoordinate segFormatWidth, 
									GCoordinate decimanFormatWidth,
									GCoordinate averageCharWidth);

	void 					adjustTotalAdvanceTo(GCoordinate newTotalAdvance);
	
private:
	ITextIterator			advanceToWordLimit(const ITextIterator& pos);
	ITextIterator			advanceToNextWordLimit(const ITextIterator& pos);
	ITextIterator			retreatToWordLimit(const ITextIterator& pos);
	ITextIterator			retreatToPreviousWordLimit(const ITextIterator& pos);

	ITextIterator			fitFirstWord(const ITextIterator& start, GCoordinate formatWidth, GCoordinate& advance);

	ITextIterator			findSegmentLimit(ITextIterator pos);
	
	void					updateVerticalMetrics(const ITextIterator& start, const ITextIterator& limit);
	
	ITextIterator			findDecimalBreak(
									const ITextIterator& start,
									const ITextIterator& limit,
									GCoordinate& visibleAdvance,
									GCoordinate decimalFormatWidth,
									GCoordinate initialAverageCharWidth);
	
	ITextIterator			findLineBreak(
									const ITextIterator& start,
									const ITextIterator& limit,
									GCoordinate& visibleAdvance,
									GCoordinate segmentFormatWidth,
									GCoordinate initialAverageCharWidth);

	GCoordinate				handleAdvance(
								IGrafPort& g,
								ITextIterator textStart,
								const ITextIterator& textLimit,
								GCoordinate tabAdvance);

	ITextPlacement			handlePixelToIndex(
								IGrafPort& g,
								const IGPoint2D& pixel,
								ITextIterator textStart,
								const ITextIterator& textLimit);

private:
	// Don't call;  these are just to silence the compiler
	ILineCursor(const ILineCursor& rhs) : fGrafPort(rhs.fGrafPort) {}
	ILineCursor& operator=(const ILineCursor& rhs) { return *this; }
};

#pragma pack(pop)

ILineCursor::ILineCursor(const ITextIterator& start, const ITextIterator& limit, IGrafPort& g) :
	fGrafPort(g),
	fStart(start),
	fCurrentPosition(start),
	fLimit(limit),
	fAscent(0.0),
	fDescent(0.0),
	fLeading(0.0),
	fTotalAdvance(0.0),
	fVisibleAdvance(0.0),
	fLength(0),
	fSegmentTotalAdvance(0.0),
	fSegmentVisibleAdvance(0.0),
	fSegmentDecimalAdvance(0.0),
	fSegmentLength(0),
	fAtTabStop(false),
	fAtLineEnd(false),
	fFirstSegmentDone(false),
	fCharRenderer()
{
	IText::offset_type d1;

	const ITextStyleSet* curStyle = start.text()->stylesAt(start.offset(), d1, d1);
	fCharRenderer.setStyle(curStyle);
}

GCoordinate
ILineCursor::handleAdvance(
								IGrafPort& g,
								ITextIterator textStart,
								const ITextIterator& textLimit,
								GCoordinate tabAdvance)
{
	IStyleRunIterator styleIter(textStart, textLimit);
	GCoordinate advance = 0.0;

	while (styleIter.isValid()) {

		fCharRenderer.setStyle(styleIter.style());

		advance += fCharRenderer.handleAdvance(g, styleIter.runStart(), styleIter.runLimit(), tabAdvance);

		styleIter.next();
	}

	// check code:
#if 0
	{
		ITextIterator checkStart = textStart;
		GCoordinate checkAdvance = 0;
		IText::offset_type d1;

		while (checkStart < textLimit) {
			fCharRenderer.setStyle(checkStart.text()->stylesAt(checkStart.offset(), d1, d1));
			checkAdvance += fCharRenderer.handleCharWidth(g, checkStart);
			++checkStart;
		}

		if (checkAdvance != advance)
			throw IException("Check advance doesn't agree with advance!");
	}
#endif
	return advance;
}

ITextPlacement
ILineCursor::handlePixelToIndex(
								IGrafPort& g,
								const IGPoint2D& pixel,
								ITextIterator textStart,
								const ITextIterator& textLimit)
{
	IStyleRunIterator styleIter(textStart, textLimit);
	GCoordinate advance = 0.0;

	while (styleIter.isValid()) {

		fCharRenderer.setStyle(styleIter.style());

		GCoordinate newAdvance = fCharRenderer.handleAdvance(g, styleIter.runStart(), styleIter.runLimit(), 0);
		
		if (pixel.fX <= advance + newAdvance)
			return fCharRenderer.handlePixelToIndex(g, 
							IGPoint2D(pixel.fX-advance, pixel.fY), styleIter.runStart(), styleIter.runLimit());

		advance += newAdvance;
		styleIter.next();
	}

	return ITextPlacement(textStart);
}

void ILineCursor::setInfo(ILayoutInfo& info) const
{
	info.setCharStart(fStart);
	info.setCharLimit(fStart + fLength);
	info.fAscent = fAscent;
	info.fDescent = fDescent;
	info.fLeading = fLeading;
	info.fAdvance = fVisibleAdvance;
	info.fTotalAdvance = fTotalAdvance;
	info.fGraphicStart = 0.0;	//formattter uses this.
	info.fLeftMargin = 0.0;		//renderer will set
}

bool ILineCursor::atTabStop() const
{
	return fAtTabStop;
}

bool ILineCursor::atLineEnd() const
{
	return fAtLineEnd;
}

GCoordinate ILineCursor::segmentTotalAdvance() const
{
	return fSegmentTotalAdvance;
}

GCoordinate ILineCursor::segmentDecimalAdvance() const
{
	return fSegmentDecimalAdvance;
}

GCoordinate ILineCursor::segmentVisibleAdvance() const
{
	return fSegmentVisibleAdvance;
}

IText::length_type ILineCursor::segmentLength() const
{
	return fSegmentLength;
}

/**
* If pos is not at a word end, move to next word end.
*/
ITextIterator ILineCursor::advanceToWordLimit(const ITextIterator& pos)
{
	if (pos.offset() != 0 && pos != fLimit) {
	
		if (!fCharRenderer.handleIsWhiteSpace(pos) || fCharRenderer.handleIsWhiteSpace(pos-1))
			return advanceToNextWordLimit(pos);
		else
			return pos;
	}
	else
		return pos;
}

/**
* Move pos to next word end.
*/
ITextIterator ILineCursor::advanceToNextWordLimit(const ITextIterator& pos)
{
	ITextIterator wordLimit = pos;
	bool prevIsWhiteSpace, nextIsWhiteSpace = fCharRenderer.handleIsWhiteSpace(wordLimit);
		
	while (wordLimit < fLimit) {
	
		prevIsWhiteSpace = nextIsWhiteSpace;
		++wordLimit;
		nextIsWhiteSpace = fCharRenderer.handleIsWhiteSpace(wordLimit);
		
		if (!prevIsWhiteSpace && nextIsWhiteSpace)
			return wordLimit;
	}
	
	return wordLimit;
}

/**
* If pos is not at a word end, move to previous word end.
*/
ITextIterator ILineCursor::retreatToWordLimit(const ITextIterator& pos)
{
	if (pos.offset() != 0) {
		
		bool atWhiteSpace = (pos == fLimit) || fCharRenderer.handleIsWhiteSpace(pos);

		if (!atWhiteSpace || fCharRenderer.handleIsWhiteSpace(pos-1))
			return retreatToPreviousWordLimit(pos);
		else
			return pos;
	}
	else
		return pos;
}

/**
* Move to previous word end.
*/
ITextIterator ILineCursor::retreatToPreviousWordLimit(const ITextIterator& pos)
{
	ITextIterator wordLimit = pos;

	if (wordLimit.offset() != 0)
		--wordLimit;

	bool prevIsWhiteSpace, nextIsWhiteSpace 
				= (wordLimit == fLimit) || fCharRenderer.handleIsWhiteSpace(wordLimit);
		
	while (wordLimit.offset() != 0) {
	
		prevIsWhiteSpace = nextIsWhiteSpace;
		--wordLimit;
		nextIsWhiteSpace = fCharRenderer.handleIsWhiteSpace(wordLimit);
		
		if (prevIsWhiteSpace && !nextIsWhiteSpace)
			return wordLimit+1;
	}
	
	return wordLimit;
}

/**
* Find the offset from fCurrentPosition which fits in formatWidth.
*/
ITextIterator ILineCursor::fitFirstWord(const ITextIterator& start, GCoordinate formatWidth, GCoordinate& advance)
{
	IGPoint2D hitPoint(formatWidth, 0);
	
	ITextPlacement place = handlePixelToIndex(fGrafPort, hitPoint, start, fLimit);
	ITextIterator firstWordLimit = place.fPos;
	
	advance = (firstWordLimit == start)? 0 : handleAdvance(fGrafPort, start, firstWordLimit, 0);
		
	if ((advance==formatWidth || place.fPlacement==ITextPlacement::BEFORE_OFFSET) && firstWordLimit > start)
		firstWordLimit--;

	// ensure at least one character on the line
	if (!fFirstSegmentDone && firstWordLimit == start)
		++firstWordLimit;

	if (firstWordLimit != place.fPos)
		advance = handleAdvance(fGrafPort, start, firstWordLimit, 0);
		
	return firstWordLimit;
}

/**
* Find the position after the tab closest to pos, or the segment limit.
*/
ITextIterator ILineCursor::findSegmentLimit(ITextIterator pos)
{
	while (pos < fLimit) {
		if (*pos == UWhiteSpace::kHorizontalTabulation) {
			++pos;
//			if (pos == fLimit-1)
//				++pos;
			break;
		}
		else
			++pos;
	}
	
	return pos;
}

/**
* Update ascent, descent, leading on range.
*/
void ILineCursor::updateVerticalMetrics(const ITextIterator& start, const ITextIterator& limit)
{
	GCoordinate ascent = 0, descent = 0, leading = 0;
	IStyleRunIterator styleIter(start, limit);

	while (styleIter.isValid()) {

		fCharRenderer.setStyle(styleIter.style());

		ascent = fCharRenderer.getAscent(styleIter.runStart(), styleIter.runLimit(), fGrafPort);
		descent = fCharRenderer.getDescent(styleIter.runStart(), styleIter.runLimit(), fGrafPort);
		leading = fCharRenderer.getLeading(styleIter.runStart(), styleIter.runLimit(), fGrafPort);

		fAscent = max(fAscent, ascent);
		fDescent = max(fDescent, descent);
		fLeading = max(fLeading, leading);
		
		styleIter.next();
	}
}


/**
* segmentFormatWidth - the maximum allowed advance of this segment
* decimalFormatWidth - if not 0, the maximum advance after the FIRST decimal in the segment
*/
void 
ILineCursor::nextSegment(
						GCoordinate segmentFormatWidth, 
						GCoordinate decimalFormatWidth, 
						GCoordinate initialAverageCharWidth)
{
	fSegmentTotalAdvance = 0.0;
	fSegmentVisibleAdvance = 0.0;
	fSegmentDecimalAdvance = 0.0;
	fSegmentLength = 0;
	fAtTabStop = false;
	
	if (fCurrentPosition == fLimit) {
		fAtLineEnd = true;
		updateVerticalMetrics(fCurrentPosition, fCurrentPosition);
		return;
	}
	
	ITextIterator segmentLimit = findSegmentLimit(fCurrentPosition);
	
	ITextIterator visibleLimit = 
				findLineBreak(fCurrentPosition, segmentLimit, 
							fSegmentVisibleAdvance, segmentFormatWidth, initialAverageCharWidth);

	// check decimal constraint
	
	if (decimalFormatWidth != 0.0)
		visibleLimit = findDecimalBreak(fCurrentPosition, visibleLimit, 
							fSegmentVisibleAdvance, decimalFormatWidth, initialAverageCharWidth);
	
	// now, if we're within format width, advance to next non-whitespace character, or segmentLimit
	ITextIterator segmentBreak = visibleLimit;
	
	if (fSegmentVisibleAdvance < segmentFormatWidth)
		while (segmentBreak < segmentLimit && fCharRenderer.handleIsWhiteSpace(segmentBreak))
			++segmentBreak;
		
	// if segmentBreak didn't move, then visible advance is total advance
	if (segmentBreak == visibleLimit) {
	
		fSegmentTotalAdvance = fSegmentVisibleAdvance;
		
		// but, if visibleLimit is not at a word limit, but is instead at the
		// segment limit, then visible advance needs to be computed

		// find last non-whitespace character in segment:
		while (visibleLimit > fCurrentPosition) {
			--visibleLimit;
			if (!fCharRenderer.handleIsWhiteSpace(visibleLimit)) {
				++visibleLimit;
				break;
			}
		}
		
		if (visibleLimit < segmentBreak)
			fSegmentVisibleAdvance = handleAdvance(fGrafPort, fCurrentPosition, visibleLimit, 0);
	}
	else
		fSegmentTotalAdvance = handleAdvance(fGrafPort, fCurrentPosition, segmentBreak, 0);
	
	fSegmentLength = segmentBreak.offset() - fCurrentPosition.offset();
	
	if (fSegmentLength != 0) {
		if (segmentBreak[-1] == UWhiteSpace::kHorizontalTabulation)
			fAtTabStop = true;
	}
	
	updateVerticalMetrics(fCurrentPosition, segmentBreak);	
	
	fLength += fSegmentLength;
	fCurrentPosition = segmentBreak;

	if (fSegmentVisibleAdvance != 0)
		fVisibleAdvance = fTotalAdvance + fSegmentVisibleAdvance;
	else
		fSegmentVisibleAdvance = fVisibleAdvance - fTotalAdvance;

	fTotalAdvance += fSegmentTotalAdvance;
	
	if (segmentBreak == fLimit)
		fAtLineEnd = true;
		
	fFirstSegmentDone = true;
}

/**
* Returns the break after the first decimal which fits in decimalFormatWidth.
* Also computes fSegmentDecimalAdvance.
* Does nothing if there are no decimals in the given range.
*/
ITextIterator 
ILineCursor::findDecimalBreak(
						const ITextIterator& start, 
						const ITextIterator& limit,
						GCoordinate& visibleAdvance,
						GCoordinate decimalFormatWidth,
						GCoordinate initialAverageCharWidth)
{
	ITextIterator firstDecimal = start;
	
	while (firstDecimal < limit) {
		if (fCharRenderer.handleIsDecimal(firstDecimal++))
			break;
	}
	
	if (firstDecimal == limit)
		return limit;
	
	fSegmentDecimalAdvance = handleAdvance(fGrafPort, start, firstDecimal, 0);
	
	// fFirstSegmentDone gets set to true b/c we already have some text on the line,
	// and we don't want to add the first character after the decimal if that character
	// won't fit.  See findLineBreak and fitFirstWord.
	fFirstSegmentDone = true;
	
	ITextIterator decimalBreak = 
			findLineBreak(firstDecimal, limit, 
						visibleAdvance, decimalFormatWidth, initialAverageCharWidth);
						
	visibleAdvance += fSegmentDecimalAdvance;
	
	return decimalBreak;
}

/**
* finds the line break between start and limit, and returns the 
* visible advance in visibleAdvance
*/

ITextIterator 
ILineCursor::findLineBreak(
						const ITextIterator& start, 
						const ITextIterator& limit,
						GCoordinate& visibleAdvance,
						GCoordinate segmentFormatWidth,
						GCoordinate initialAverageCharWidth)
{
	GCoordinate averageCharWidth = initialAverageCharWidth;
	GCoordinate currentAdvance;
	ITextIterator minGuess = start, maxGuess = limit, guess;
	
	visibleAdvance = 0;
	/*
		minGuess is the greatest offset which definitely fits on the line
		maxGuess is the least offset which possibly fits
	*/
	
	while (minGuess < maxGuess) { 

#pragma info(none)
		if (averageCharWidth < 1)
			averageCharWidth = 1;

		guess = start + (segmentFormatWidth / averageCharWidth);
#pragma info(restore)
		
		if (guess > maxGuess)
			guess = maxGuess;
		else {

			guess = advanceToWordLimit(guess);
		
			if (guess <= minGuess)
				guess = advanceToNextWordLimit(minGuess);
			if (guess > maxGuess)
				guess = maxGuess;
		}
		// now guess is in (minGuess, maxGuess]

		currentAdvance = handleAdvance(fGrafPort, start, guess, 0);
		
		if (currentAdvance >= segmentFormatWidth) {

			fAtLineEnd = true; // OK to do this here?

			maxGuess = retreatToPreviousWordLimit(guess);
			if (maxGuess <= start)  {
				// if you're in here, the first word won't fit.

				if (fCharRenderer.handleIsWhiteSpace(start)) {
					minGuess = start+1;
					while (minGuess < limit && fCharRenderer.handleIsWhiteSpace(minGuess))
						minGuess++;
					visibleAdvance = handleAdvance(fGrafPort, start, minGuess, 0);
				}
				else
					minGuess = fitFirstWord(start, segmentFormatWidth, visibleAdvance);
				// and you'll exit now, since minGuess >= maxGuess
			}
		}
		else {
			minGuess = guess;
			visibleAdvance = currentAdvance;
		}
			
		IText::length_type currentLength = guess.offset() - start.offset();
#pragma info(none)
		averageCharWidth = (currentLength == 0)? 0 : fSegmentVisibleAdvance / currentLength;
#pragma info(restore)
	}
	
	return minGuess;
}

void ILineCursor::adjustTotalAdvanceTo(GCoordinate newTotalAdvance)
{
	GCoordinate delta = newTotalAdvance - fTotalAdvance;
	fTotalAdvance = newTotalAdvance;

	if (fSegmentVisibleAdvance != 0.0)
		fVisibleAdvance += delta;
}

//................................................................................
IParagraphRendererCache::IParagraphRendererCache()
	: fIsOwned(false), fRenderer(NULL)
{
}
IParagraphRendererCache::IParagraphRendererCache(const IParagraphRendererCache& source)
	: fIsOwned(source.fIsOwned), fRenderer(NULL)
{
	if (source.fIsOwned && source.fRenderer) {
		fRenderer = source.fRenderer->clone();
	}
}

IParagraphRendererCache::~IParagraphRendererCache()
{
	if (fIsOwned) {
		delete fRenderer;
		fRenderer = NULL;
	}
}

IParagraphRenderer* IParagraphRendererCache::clone() const
{
	return new IParagraphRendererCache(*this);
}

void IParagraphRendererCache::writeToStream(IDataStream& toWhere) const
{
	short temp = fIsOwned;
	temp >>= toWhere;
	//hey {jf} -
	//flatten(fRenderer);
}

void IParagraphRendererCache::readFromStream(IDataStream& fromWhere)
{
	short temp;
	temp<<= fromWhere;
	fIsOwned = temp;
	//hey {jf} -
	//resurrect();
}

void IParagraphRendererCache::layout(IGrafPort& g, ILayoutInfo& infoOut,
	const GCoordinate& lineDimension, ITextIterator lineStart,
	const ITextIterator& paragraphStart, const ITextIterator& paragraphLimit,
	GCoordinate averageCharWidth) const
{
	fRenderer->layout(g, infoOut, lineDimension, lineStart, paragraphStart, paragraphLimit, averageCharWidth);
}


void IParagraphRendererCache::renderWithHighlight(IGrafPort& g,
	const ILayoutInfo& info, const IGPoint2D& baseLine, const ITextIterator& selStart,
	const ITextIterator& selEnd, const IColor& hilight) const
{
	fRenderer->renderWithHighlight(g, info, baseLine, selStart, selEnd, hilight);
}

void IParagraphRendererCache::render(IGrafPort& g,
	const ILayoutInfo& info, const IGPoint2D& baseline) const
{
	fRenderer->render(g, info, baseline);
}

ITextPlacement IParagraphRendererCache::pixelToIndex(
	IGrafPort& g, const ILayoutInfo& info, const IGPoint2D& where) const
{
	return fRenderer->pixelToIndex(g, info, where);
}

IGPoint2D IParagraphRendererCache::indexToPixel(
	IGrafPort& g, const ILayoutInfo& info, const ITextPlacement& index) const
{
	return fRenderer->indexToPixel(g, info, index);
}

void IParagraphRendererCache::renderCaret(
	const ILayoutInfo& info, const ITextPlacement& index) const
{
	fRenderer->renderCaret(info, index);
}

void IParagraphRendererCache::setStyle(const ITextStyleSet* style)
{
	IParagraphRenderer* temp = IParagraphRenderer::extractFrom(style);
	if (temp != NULL) {
		if (fIsOwned) {
			delete fRenderer;
			fRenderer = NULL;
		}
		fRenderer = temp;
		fIsOwned = false;
	}
	else {
		if (!fIsOwned) {
			fRenderer = new IStandardParagraphRenderer();
		}
		fRenderer->setStyle(style);
		fIsOwned = true;
	}
}

//................................................................................

void IStandardParagraphRenderer::TextBox(	const IText& text,
							IGrafPort& g,
							const IGRect2D& drawRect)
{
	TextBox(text, g, drawRect, true, true);
}

void IStandardParagraphRenderer::TextBox(	const IText& text,
							IGrafPort& g,
							const IGRect2D& drawRect,
							bool fillInc,
							bool hLine)
{
	const ITextIterator textStart = text.begin();
	const ITextIterator textEnd = text.end();
	ITextIterator curLineStart = text.begin();
	ITextIterator curParagraphStart = curLineStart;
	ITextIterator curParagraphLimit = curLineStart;
	IGPoint2D pen = drawRect.topLeft();

	if (!fillInc) {
	    if (hLine) {
	        pen.fY += drawRect.height();
	    }
	    else {
	        pen.fX += drawRect.width();
	    }
	}

	ILayoutInfo theLine;

	IParagraphRendererCache renderer;

	while (curLineStart < textEnd) {
		if (curLineStart >= curParagraphLimit) {
			IText::offset_type offset = text.find(UWhiteSpace::kParagraphSeparator, curLineStart.offset());
			if (offset == IText::npos) {
				curParagraphLimit = textEnd;
			}
			else {
				curParagraphLimit = textStart + offset + 1;	// skip the paragraph mark
			}

			renderer.setStyleFrom(curParagraphStart);
		}

		renderer.layout(g, theLine, drawRect.width(), curLineStart, curParagraphStart, curParagraphLimit, 10);

		if (!fillInc) {
			if (hLine) {
				pen.fY -= theLine.height();
			}
			else {
				pen.fX -= theLine.height();
			}
		}

		renderer.render(g, theLine, pen); // can only render horizontal...

		if (fillInc) {
			if (hLine) {
				pen.fY += theLine.height();
			}
			else {
				pen.fX += theLine.height();
			}
		}

		curLineStart = theLine.charLimit();
	}
}

IStandardParagraphRenderer::IStandardParagraphRenderer()
	: fLeadingMargin(0.0),
	fTrailingMargin(0.0),
	fFirstLineIndent(0.0),
	fMinFirstLineSpacing(0.0),
	fExtraFirstLineSpacing(0.0),
	fExtraLineSpacing(0.0), // added {jbr}
	fMinLineSpacing(0.0),
	fFlush(IParagraphJustificationStyle::kStart),
	fTabRuler(NULL),
	fCacheStyle(NULL),
	fRenderer()
{
}

IStandardParagraphRenderer::IStandardParagraphRenderer(const ITextStyleSet* pStyle)
	: fLeadingMargin(0.0),
	fTrailingMargin(0.0),
	fFirstLineIndent(0.0),
	fMinFirstLineSpacing(0.0),
	fExtraFirstLineSpacing(0.0),
	fExtraLineSpacing(0.0), // added {jbr}
	fMinLineSpacing(0.0),
	fFlush(IParagraphJustificationStyle::kStart),
	fTabRuler(NULL),
	fCacheStyle(pStyle),
	fRenderer()
{
	initRenderer(pStyle);
}

IStandardParagraphRenderer::IStandardParagraphRenderer(const IStandardParagraphRenderer& source)
	: fLeadingMargin(0.0),
	fTrailingMargin(0.0),
	fFirstLineIndent(0.0),
	fMinFirstLineSpacing(0.0),
	fExtraFirstLineSpacing(0.0),
	fExtraLineSpacing(0.0), // added {jbr}
	fMinLineSpacing(0.0),
	fFlush(IParagraphJustificationStyle::kStart),
	fTabRuler(NULL),
	fCacheStyle(NULL),
	fRenderer()
{
	//initRenderer(source.fCacheStyle);
}

IStandardParagraphRenderer::~IStandardParagraphRenderer()
{
}

void IStandardParagraphRenderer::setStyle(const ITextStyleSet* style)
{
	initRenderer(style);
}

IParagraphRenderer* IStandardParagraphRenderer::clone() const
{
	return new IStandardParagraphRenderer(*this);
}

void IStandardParagraphRenderer::writeToStream(IDataStream& toWhere) const
{
}

void IStandardParagraphRenderer::readFromStream(IDataStream& fromWhere)
{
}

void IStandardParagraphRenderer::layout(IGrafPort& g, ILayoutInfo& info, const GCoordinate& lineBound,
	ITextIterator lineStart, const ITextIterator& paragraphStart, const ITextIterator& limit, GCoordinate averageCharWidth) const
{
// {jbr} Temporary workaround for 0-length lines:
	if (lineStart == limit) {

		info.setCharStart(lineStart);
		info.setCharLimit(limit);
		info.clearSegments();
		//info.fText = lineStart.text();

		info.fAdvance = info.fTotalAdvance = 0;
		info.fLeftMargin = fLeadingMargin;
		if (lineStart==paragraphStart)
			info.fLeftMargin += fFirstLineIndent;

		// client is responsible for ascent, descent, leading, and graphic start

		return;
	}

	info.clearSegments();

	GCoordinate curFirstLineIndent = (lineStart == paragraphStart) ? fFirstLineIndent : 0.0;

	GCoordinate effectiveLeftMargin = fLeadingMargin + curFirstLineIndent;
	GCoordinate effectiveRightMargin = lineBound - fTrailingMargin;

	ITabStop tab(effectiveLeftMargin, ITabStop::kLeft); // assume first segment controlled by this tab

	bool hasSegments = false;

	ILineCursor iter(lineStart, limit, g);

	GCoordinate prevTotalAdvance = 0; // limit adjustments of segment

	do {
		GCoordinate rfw = effectiveRightMargin - tab.position(); // available width past tab stop
		GCoordinate tfw = effectiveRightMargin - effectiveLeftMargin - prevTotalAdvance; // total available width

		GCoordinate segmentFormatWidth = rfw;
		GCoordinate decimalFormatWidth = 0.0; // default no decimal tabs
		switch (tab.type()) {
			case ITabStop::kAuto:
			case ITabStop::kLeft:
				// default
				break;
			case ITabStop::kCenter:
				segmentFormatWidth = min(tfw, rfw * 2.0);
				break;
			case ITabStop::kRight:
				segmentFormatWidth = tfw;
				break;
			case ITabStop::kDecimal:
				segmentFormatWidth = tfw;
				decimalFormatWidth = rfw;
				break;
		}


		iter.nextSegment(segmentFormatWidth, decimalFormatWidth, averageCharWidth);

		if (!hasSegments && iter.atTabStop()) {
			hasSegments = true;
		}

		if (hasSegments) {
			GCoordinate offset = tab.position();

			switch (tab.type()) {
				case ITabStop::kCenter:
					offset = tab.position() - iter.segmentVisibleAdvance() / 2.0;
					break;

				case ITabStop::kRight:
					offset = tab.position() - iter.segmentVisibleAdvance();
					break;

				case ITabStop::kDecimal:
					offset = tab.position() - iter.segmentDecimalAdvance();
					break;

				default:
					break;
			}
			offset -= effectiveLeftMargin;

			if (offset < prevTotalAdvance) {
				offset = prevTotalAdvance;
			}

			prevTotalAdvance = offset + iter.segmentTotalAdvance();
			iter.adjustTotalAdvanceTo(prevTotalAdvance);

			info.addSegment(ISegmentInfo(offset, iter.segmentLength()));

			setNextTab(tab, effectiveLeftMargin + prevTotalAdvance); // renderer supplies next tab after this position
		}
	} while (!iter.atLineEnd() && tab.position() <= effectiveRightMargin); // <= to allow right tab at right margin

	// if the last segment ended in a tab, adjust the line advance to the next tab.	 This will
	// be used when calculating the width of the last tab.

	if (iter.atTabStop()) {
		iter.adjustTotalAdvanceTo(min(tab.position(), effectiveRightMargin) - effectiveLeftMargin);
	}

	// copy line info data into cache

	iter.setInfo(info);

	// compute final left margin

	switch(fFlush) {
		case IParagraphJustificationStyle::kEnd:
			info.fLeftMargin = effectiveRightMargin - info.fAdvance;
			break;

		case IParagraphJustificationStyle::kCenter:
			info.fLeftMargin = (effectiveLeftMargin + effectiveRightMargin - info.fAdvance) / 2;
			break;

		default:
			info.fLeftMargin = effectiveLeftMargin; // leftortop, or something else weird
			break;
	}

	// Adjust first and interline ascents.	This pushes down the baseline, which is what we want.
	// Adjusting the leading instead puts the space after the baseline.	 'Spacing' is a poor word,
	// really we're adjusting the ascents.	This should instead adjust the baseline-to-baseline
	// distance, but we don't have enough information to do that because we don't have the info
	// for the previous line.  That's what we really want, but then we're dependent on the styles
	// of the previous line, and we'd have to tweak the formatting code to make sure this line's
	// info was always recomputed when the line before it changed.	That would be yucksters, so
	// we hack the ascents for now. !!!

	if (lineStart == paragraphStart) {
		info.fAscent = max(fMinFirstLineSpacing, info.fAscent + fExtraFirstLineSpacing);
	}
	else {
		info.fAscent = max(fMinLineSpacing, info.fAscent + fExtraLineSpacing);
	}
}

static ISynchronized<IStandardTabRuler*> gDefaultRuler;

void IStandardParagraphRenderer::setNextTab(ITabStop& tab, GCoordinate position) const
{
    // for now, default is a left tab every 36 pixels
    // we need a tab bar interface that the renderer can query, and a style to carry the information.

	if (fTabRuler != NULL) {
		tab = fTabRuler->nextTab(position);
		/*
	    ITabStop nextTab = fTabRuler->nextTab(position);
	    if (nextTab.type() != ITabStop::kAuto) {
	    	tab = nextTab;
	    }
	    else {
	    	//hey {jf} - this should have a real huge value
	    	tab = ITabStop(999999999, ITabStop::kAuto);
	    }
	    */
	}
	else {
        IStandardTabRuler *ruler = gDefaultRuler.get();
        if (!ruler) {
            ruler = synchronizedInit(gDefaultRuler, new IStandardTabRuler);
        }
	    tab = ruler->nextTab(position);
	}
}


void IStandardParagraphRenderer::renderWithHighlight(IGrafPort& g,
	const ILayoutInfo& info, const IGPoint2D& baseLine, const ITextIterator& selStart,
	const ITextIterator& selStop, const IColor& hilight) const
{
		/*hey {jf} - render highlight here
	if (	//selStart != null && selStop != null &&
			selStart != selStop &&
			selStart < info.charLimit() &&
			selStop > info.charStart()) {

		ILayoutInfo info;
		GCoordinate hStart = indexToPixel(info, selStart).fX;
		GCoordinate hEnd = indexToPixel(info, selStop).fX;

		g.setColor(highlight);
		g.fillRect(x + hStart, y, hEnd - hStart, info.height());
	}
		*/

	if (selStart != selStop && selStart < info.charLimit() && selStop > info.charStart()) {

		IGPoint2D hStart = indexToPixel(g, info, selStart),
					hStop = indexToPixel(g, info, selStop);

		IGRect2D bgRect(hStart.fX + baseLine.fX, baseLine.fY/* - info.fAscent*/, 
				hStop.fX + baseLine.fX, baseLine.fY + info.fAscent + info.fDescent + info.fLeading);

		g.draw(bgRect, IFillBundle(hilight));
	}

	render(g, info, baseLine);
}

/**
 Render segments of a line.	 If a line has tabs, it will have more than one segment.  Segments have no internal tabs,
 vertical tabs, or newlines, but may end with one.	It's passed through to the character renderer so it can
 determine whether to show invisibles or not (right now it draws the missing char for debugging, there's no
 information about invisibles passed in yet).
[ NOTE:  I've taken this behavior out.  Right now, we don't pass the invisible at the end
 of the segment to the character renderer.  If we want a "show invisibles" finction, we'll
 have to re-think where the logic to handle these characters really ought to go.
							--rtg 5/7/97		]
 */
GCoordinate IStandardParagraphRenderer::renderSegment(
	IGrafPort& g,
	ITextIterator start,
	const ITextIterator& limit,
	const IGPoint2D& p) const
{
	// look at the last character of the segment to render and if it's a segment or
	// line terminator, fudge the segment length down by 1 so we don't try to draw it
	// (there's gotta be a better way to do this!)		--rtg 5/7/97
	ITextIterator	workingLimit = limit - 1;
    int switch_val = *workingLimit;
    switch (switch_val) {
		case UASCII::kHorizontalTabulation:
		case UASCII::kLineFeed:
		case UASCII::kCarriageReturn:
		case UGeneralPunctuation::kParagraphSeparator:
		case UGeneralPunctuation::kLineSeparator:
			break;

		default:
			++workingLimit;
			break;
	}
	
	IGPoint2D baseLine = p;
	IStyleRunIterator styleIter(start, workingLimit);
	while(styleIter.isValid()) {
		const ITextStyleSet* style = styleIter.style();
		ITextIterator runLimit = styleIter.runLimit();
		((IStandardParagraphRenderer*)this)->fRenderer.setStyle(style);
		GCoordinate advance = fRenderer.handleDrawChars(g, baseLine, styleIter.runStart(), styleIter.runLimit());
		baseLine.fX += advance;
		styleIter.next();
	}
	return baseLine.fX - p.fX;
}

void IStandardParagraphRenderer::render(IGrafPort& g, const ILayoutInfo& info,
	const IGPoint2D& baseLine) const
{
	ITextIterator start = info.charStart();

	GCoordinate leftMarginX = baseLine.fX + info.fLeftMargin;

	IGPoint2D p(leftMarginX, baseLine.fY + info.fAscent);

	if (info.segmentCount() != 0) {
		for (size_t i = 0; i < info.segmentCount(); ++i) {
			const ISegmentInfo& s = info[i];

			p.fX = leftMarginX + s.fAdvance;

			p.fX += renderSegment(g, start, start + s.fLength, p);

			start += s.fLength;
		}
	} else {
		renderSegment(g, start, info.charLimit(), p);
	}
}

ITextPlacement IStandardParagraphRenderer::pixelToIndex(IGrafPort& g,
														const ILayoutInfo& info,
														const IGPoint2D& whereIn) const
{
	IGPoint2D where(whereIn);

	where.fX -= info.fLeftMargin;

	if (where.fX <= 0 || info.charStart() >= info.charLimit()) {
		return ITextPlacement(info.charStart(), ITextPlacement::AFTER_OFFSET);
	}
	else if (where.fX >= info.fTotalAdvance) {
		return ITextPlacement(info.charLimit(), ITextPlacement::BEFORE_OFFSET);
	}
	else {
		ITextPlacement result(info.charLimit());

		ITextIterator start = info.charStart();
		ITextIterator limit = info.charLimit();
		GCoordinate limitX = info.fTotalAdvance;
		//GCoordinate limitY = where.fY;

		/* Determine which segment the offset lies in.	For now, all segments are left to right, in order.
		   Iterate until a segment starts past the offset, then back up.  That segment extends to the start
		   of the next, or to the end of the total advance of the line.	 Adjust the offset and limit to be
		   relative to the start of the segment. The limit is needed to resolve the tab width. !!! LimitY is
		   bogus and should be removed. */

		if (info.segmentCount() != 0) {
			size_t i = 0;
			while (i < info.segmentCount()) {
				const ISegmentInfo& s = info[i];
				if (where.fX < s.fAdvance)
					break;

				start += s.fLength;
				++i;
			}
			if (i < info.segmentCount()) {
				limitX = info[i].fAdvance;
			}
			--i;
			const ISegmentInfo& s = info[i];

			limit = start;
			start -= s.fLength;
			where.fX -= s.fAdvance;
			limitX -= s.fAdvance;
		}

		IStyleRunIterator styleIter(start, limit);

		GCoordinate advance = 0.0;

		while (styleIter.isValid()) {
			const ITextStyleSet* style = styleIter.style();
			((IStandardParagraphRenderer*)this)->fRenderer.setStyle(style);

			ITextIterator runLimit(styleIter.runLimit());
			// could optimize this out in some places {jbr}
			GCoordinate newAdvance = fRenderer.handleAdvance(g, styleIter.runStart(), styleIter.runLimit(), limitX);
			if (where.fX <= advance + newAdvance) {
				result = fRenderer.handlePixelToIndex(
										g,
										IGPoint2D(where.fX - advance, where.fY),
									//hey {jf} - why was this extra point here?
									//	IGPoint2D(limitX, limitY),
										styleIter.runStart(),
										styleIter.runLimit());

				return result;
			}
			advance += newAdvance;
			styleIter.next();
		}

//hey {jf} - fix this
//		if (!styleIter.isValid()) {
//			throw new IllegalArgumentException("Debug: pixel to offset - reached end of iterator before end of line");
//		}
		return result;
	}
}

IGPoint2D IStandardParagraphRenderer::indexToPixel(IGrafPort& g,
													const ILayoutInfo& info,
													const ITextPlacement& offset) const
{
	IGPoint2D result(0,0);

	GCoordinate res = 0.0;
	ITextIterator pos = offset.fPos; // this would matter with bidi text, but we don't support it yet.

	if (pos < info.charStart()) {
		res = info.fLeftMargin;
	}
	else if (pos >= info.charLimit()) {
		res = info.fLeftMargin + info.fTotalAdvance;
	}
	else {
		ITextIterator start = info.charStart();
		ITextIterator limit = info.charLimit();
		res = info.fLeftMargin;

		/* Determine which segment the offset is in, and adjust the 'left margin' to the start of
		   that segment.  Then add in the offset within the segment. */

		if (info.segmentCount() != 0) {
			for (int i = 0; i < info.segmentCount(); ++i) {
				const ISegmentInfo& s = info[i];
				if (pos < start + s.fLength) {
					limit = start + s.fLength;
					res += s.fAdvance;
					break;
				}
				start += s.fLength;
			}
		}

		IStyleRunIterator styleIter(start, limit);

		while (styleIter.isValid()) {

			const ITextStyleSet* style = styleIter.style();
			((IStandardParagraphRenderer*)this)->fRenderer.setStyle(style);

			if (pos > styleIter.runLimit()) {
				res += fRenderer.handleAdvance(g, styleIter.runStart(), styleIter.runLimit(), 0);
			}
			else {
				IGPoint2D temp = fRenderer.handleIndexToPixel(g, styleIter.runStart(), styleIter.runLimit(), offset);
				res += temp.fX;
				break;
			}

			styleIter.next();
		}
	}

	result.fX = res;
	result.fY = info.fAscent;

	return result;
}

void IStandardParagraphRenderer::renderCaret(const ILayoutInfo& info,
	const ITextPlacement& index) const
{
//hey {jf} - write this routine
}

void IStandardParagraphRenderer::initRenderer(const ITextStyleSet* set)
{
	if (fCacheStyle != set) {

		fCacheStyle = set;

//		fLeadingMargin = LeadingMarginStyle.extractDefaultFrom(set);
		IParagraphStartIndentStyle leadingStyle;
		set->extract(leadingStyle);
		fLeadingMargin = leadingStyle.startIndent();

//		fTrailingMargin = TrailingMarginStyle.extractDefaultFrom(set);
		IParagraphEndIndentStyle trailingStyle;
		set->extract(trailingStyle);
		fTrailingMargin = trailingStyle.endIndent();

//		fFirstLineIndent =  FirstLineIndentStyle.extractDefaultFrom(set);
		IParagraphFirstIndentStyle indentStyle;
		set->extract(indentStyle);
		fFirstLineIndent = indentStyle.firstIndent();

//		fFlush = FlushStyle.extractDefaultFrom(set);
		IParagraphJustificationStyle justStyle;
		set->extract(justStyle);
		fFlush = justStyle.justification();
	}
}

//................................................................................
