/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 66 1.24.1.5 source/albert/graph2d/ibundles.cpp, 2d, ioc.v400, 000302 
/*================
||
||  File:   Bundles.C
||
||  What:   Defines IAttributeState class members and members of
||			IAttributeState subclasses such as IGrafBundle
||
||  Class defined in this file:
||      IAttributeState
||      IFillBundle
||      IFrameBundle
||      IFillAndFrameBundle
||      IGrafBundle
||
||  Change History:
||
*/

#include <grtype.hpp>
#include <ibundles.hpp>
#include <stdio.h>
#include <igrafatt.hpp>
#include <ipaints.hpp>
#include <imatrix.hpp>
#include <inumeric.hpp>
#include <idatstrm.hpp>
#include <grstrmod.hpp>
#include <iprimlck.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
#define _IBUNDLES_CPP_
#include <ipagetun.h>
#endif

//=============================
// IAttributeTimeStamp routines
//=============================

struct IAttributeTimeStamp {
	IAttributeTimeStamp();
	unsigned long fAttributeTimeStamp;
};

IAttributeTimeStamp::IAttributeTimeStamp()
  : fAttributeTimeStamp(0)
{
}

static IAttributeTimeStamp* gStaticTimeStamp = 0;
static IAttributeTimeStamp *staticTimeStamp()
{
	if (!gStaticTimeStamp) {
		IPrimalLock lock;
		if (!gStaticTimeStamp) {
			gStaticTimeStamp = new IAttributeTimeStamp;
			++(gStaticTimeStamp->fAttributeTimeStamp);
		}
	}
	else  {
		IPrimalLock lock;
		++(gStaticTimeStamp->fAttributeTimeStamp);
	}
	return gStaticTimeStamp;
}

const unsigned long&  IAttributeState::staticAttTimeStamp()
{
	return staticTimeStamp()->fAttributeTimeStamp;
}

//================
// IAttributeState routines
//================

// StreamableDefinitionsMacro(IAttributeState, gGraphicsStreamModule);


IAttributeState::IAttributeState() :
	fPrinting(kPrintingUndefined),
	fTimeStamp( IAttributeState::staticAttTimeStamp())
{}

IAttributeState::IAttributeState(const IAttributeState& source) :
	fTimeStamp(source.fTimeStamp),
	fPrinting(source.fPrinting)
{}

IAttributeState::~IAttributeState()
{}

IAttributeState& IAttributeState::operator=(const IAttributeState& source)
{
	if (&source != this)
	{
		fTimeStamp = source.fTimeStamp;
		fPrinting = source.fPrinting;
	}
	return *this;
}

bool IAttributeState::isPrinting() const
{
	return printingFlag() == kIsPrinting;
}

IAttributeState::EPrintingState IAttributeState::printingFlag() const
{
	return fPrinting;
}

void IAttributeState::setPrintingFlag(IAttributeState::EPrintingState printFlag)
{
	fPrinting = printFlag;
	updateTimeStamp ();
}

const IPaint* IAttributeState::fillPaint() const
{
	return 0 /*NIL*/;
}

const IColorTransferMode* IAttributeState::fillTransferMode() const
{
	return 0 /*NIL*/;
}

const IPaint* IAttributeState::framePaint () const
{
	return 0 /*NIL*/;
}

const IColorTransferMode* IAttributeState::frameTransferMode() const
{
	return 0 /*NIL*/;
}

const IPen* IAttributeState::framePen () const
{
	return 0 /*NIL*/;
}

const ICap* IAttributeState::frameEndCap () const
{
	return 0 /*NIL*/;
}

const IJoint* IAttributeState::frameJoint () const
{
	return 0 /*NIL*/;
}

const IImageTransferMode* IAttributeState::imageTransferMode() const
{
	return 0 /*NIL*/;
}

const IImageSamplingControl* IAttributeState::imageSampling() const
{
	return 0 /*NIL*/;
}

// pure virtual
IAttributeState::EDrawOperation IAttributeState::drawingOperation () const
{
	return IAttributeState::kUnset;
}

void IAttributeState::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
	(unsigned long)fPrinting >>= toWhere;
}

void IAttributeState::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
	unsigned long  temp;
	temp <<= fromWhere;
	fPrinting = (EPrintingState) temp;
	updateTimeStamp ();
}

unsigned long IAttributeState::timeStamp() const
{
	return fTimeStamp;
}

void IAttributeState::setTimeStamp(unsigned long seed)
{
	fTimeStamp = seed;
}

void IAttributeState::updateTimeStamp()
{
	fTimeStamp = IAttributeState::staticAttTimeStamp();
}

void IAttributeState::resetTimeStamp()
{
	fTimeStamp = (unsigned long)0;
}

bool IAttributeState::operator==(const IAttributeState& obj) const
{
	bool retval = false;
	if (GraphicTypeid(*this) == GraphicTypeid(obj))
	{
		retval = (fPrinting == obj.fPrinting);
	}
	return retval;
}

bool IAttributeState::operator!=(const IAttributeState& obj) const
{
	return !operator==(obj);
}



#if defined(_AIX) && IC_RTTI_ENABLED && __IBMCPP__ >= 360 && __IBMCPP__ < 400
/*================
||
||  class IBuggyBundle is a workaround for an AIX RTTI 96a compiler code generation bug.
||
*/


// class static
IAttributeState::EDrawOperation IBuggyBundle::fgDrawOperation = IAttributeState::kFill;

StreamableDefinitionsMacro(IBuggyBundle, gGraphicsStreamModule);

IBuggyBundle::IBuggyBundle() :
	fBaseMode(IColorTransferMode::kSourceCopy),
	fPaint(IBaseColor(IBaseColor::kBlack))
{
	fMode = &fBaseMode;
}

IBuggyBundle::IBuggyBundle(const IBuggyBundle& copy) :
	IAttributeState(copy),
	fPaint(copy.fPaint)
{
	if (GraphicTypeid(*copy.fMode) == GraphicTypeid(fBaseMode))
	{
		fBaseMode = *copy.fMode;
		fMode = &fBaseMode;
	}
	else
	{
		fMode = GraphicCopy(*copy.fMode);
	}
}

IBuggyBundle::IBuggyBundle(const IBaseColor& fillColor) :
	fBaseMode(IColorTransferMode::kSourceCopy),
	fPaint(IBaseColor(fillColor))
{
	fMode = &fBaseMode;
}

IBuggyBundle::IBuggyBundle(
	const IBaseColor& fillColor,
	const IColorTransferMode& transferMode) :

	fPaint(fillColor)
{
	if (GraphicTypeid(transferMode) == GraphicTypeid(fBaseMode))
	{
		fBaseMode = transferMode;
		fMode = &fBaseMode;
	}
	else
	{
		fMode = GraphicCopy(transferMode);
	}
}

IBuggyBundle::~IBuggyBundle()
{
	if (fMode != &fBaseMode) delete fMode;
}

IBuggyBundle& IBuggyBundle::operator=(const IBuggyBundle& obj)
{
	if (&obj != this)
	{
		IAttributeState::operator=(obj);

		fPaint = obj.fPaint;
		setFillTransferMode(*obj.fMode);
	}

	return *this;
}

void IBuggyBundle::setFillColor(const IBaseColor& color)
{
	fPaint.setColor(color);
}

const IPaint* IBuggyBundle::fillPaint() const
{
	return &fPaint;
}

void IBuggyBundle::setFillTransferMode(const IColorTransferMode& transferMode)
{
	if (GraphicTypeid(transferMode) != GraphicTypeid(*fMode))
	{
		if (fMode != &fBaseMode)
			delete fMode;

		if (GraphicTypeid(transferMode) == GraphicTypeid(fBaseMode))
		{	// copy it into the fBaseMode
			fBaseMode = transferMode;
			fMode = &fBaseMode;
		}
		else
		{	// clone the obj
			fMode = GraphicCopy(transferMode);
		}
	}
	else
	{	// reuse the storage...
		*fMode = transferMode;
	}
}

const IColorTransferMode* IBuggyBundle::fillTransferMode() const
{
	return fMode;
}

IAttributeState::EDrawOperation IBuggyBundle::drawingOperation() const
{
	// return the class invariant even if the base class changed the
	// state.
	return fgDrawOperation;
}

void IBuggyBundle::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
	IAttributeState::writeToStream(toWhere);
	fPaint >>= toWhere;

	bool streamPtr = true;
	if (fMode == &fBaseMode)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fMode, toWhere);
	}
	else
	{
		fBaseMode >>= toWhere;
	}
}

void IBuggyBundle::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
	IAttributeState::readFromStream(fromWhere);

	fPaint <<= fromWhere;

	bool streamPtr;
	long temp;
	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fMode != &fBaseMode)
		delete fMode;
	if (streamPtr)
	{
		::readObject(fMode, fromWhere);
	}
	else
	{
		fBaseMode <<= fromWhere;
		fMode = &fBaseMode;
	}
}

bool IBuggyBundle::operator==(const IAttributeState& obj)  const
{
	if (GraphicTypeid(*this) == GraphicTypeid(obj))
	{
#if IC_RTTI_ENABLED
		const IBuggyBundle& bundle = dynamic_cast<const IBuggyBundle&>(obj);
#else
		const IBuggyBundle& bundle = *((const IBuggyBundle*)&obj);
#endif // IC_RTTI_ENABLED
		return ((fPaint == bundle.fPaint) && (*fMode == *bundle.fMode));
	}
	else
		return false;
}
#endif // defined(_AIX) && IC_RTTI_ENABLED && __IBMCPP__ >= 360 && __IBMCPP__ < 400



/*================
||
||  class IFillBundle
||
*/

// class static
IAttributeState::EDrawOperation IFillBundle::fgDrawOperation = IAttributeState::kFill;

StreamableDefinitionsMacro(IFillBundle, gGraphicsStreamModule);

IFillBundle::IFillBundle() :
	fBaseMode(IColorTransferMode::kSourceCopy),
	fPaint(IBaseColor(IBaseColor::kBlack))
{
	fMode = &fBaseMode;
}

IFillBundle::IFillBundle(const IFillBundle& copy) :
	IAttributeState(copy),
	fPaint(copy.fPaint)
{
	if (GraphicTypeid(*copy.fMode) == GraphicTypeid(fBaseMode))
	{
		fBaseMode = *copy.fMode;
		fMode = &fBaseMode;
	}
	else
	{
		fMode = GraphicCopy(*copy.fMode);
	}
}

IFillBundle::IFillBundle(const IBaseColor& fillColor) :
	fBaseMode(IColorTransferMode::kSourceCopy),
	fPaint(IBaseColor(fillColor))
{
	fMode = &fBaseMode;
}

IFillBundle::IFillBundle(
	const IBaseColor& fillColor,
	const IColorTransferMode& transferMode) :

	fPaint(fillColor)
{
	if (GraphicTypeid(transferMode) == GraphicTypeid(fBaseMode))
	{
		fBaseMode = transferMode;
		fMode = &fBaseMode;
	}
	else
	{
		fMode = GraphicCopy(transferMode);
	}
}

IFillBundle::~IFillBundle()
{
	if (fMode != &fBaseMode) delete fMode;
}

IFillBundle& IFillBundle::operator=(const IFillBundle& obj)
{
	if (&obj != this)
	{
		IAttributeState::operator=(obj);

		fPaint = obj.fPaint;
		setFillTransferMode(*obj.fMode);
	}

	return *this;
}

void IFillBundle::setFillColor(const IBaseColor& color)
{
	fPaint.setColor(color);
}

const IPaint* IFillBundle::fillPaint() const
{
	return &fPaint;
}

void IFillBundle::setFillTransferMode(const IColorTransferMode& transferMode)
{
	if (GraphicTypeid(transferMode) != GraphicTypeid(*fMode))
	{
		if (fMode != &fBaseMode)
			delete fMode;

		if (GraphicTypeid(transferMode) == GraphicTypeid(fBaseMode))
		{	// copy it into the fBaseMode
			fBaseMode = transferMode;
			fMode = &fBaseMode;
		}
		else
		{	// clone the obj
			fMode = GraphicCopy(transferMode);
		}
	}
	else
	{	// reuse the storage...
		*fMode = transferMode;
	}
}

const IColorTransferMode* IFillBundle::fillTransferMode() const
{
	return fMode;
}

IAttributeState::EDrawOperation IFillBundle::drawingOperation() const
{
	// return the class invariant even if the base class changed the
	// state.
	return fgDrawOperation;
}

void IFillBundle::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
	IAttributeState::writeToStream(toWhere);
	fPaint >>= toWhere;

	bool streamPtr = true;
	if (fMode == &fBaseMode)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fMode, toWhere);
	}
	else
	{
		fBaseMode >>= toWhere;
	}
}

void IFillBundle::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
	IAttributeState::readFromStream(fromWhere);

	fPaint <<= fromWhere;

	bool streamPtr;
	long temp;
	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fMode != &fBaseMode)
		delete fMode;
	if (streamPtr)
	{
		::readObject(fMode, fromWhere);
	}
	else
	{
		fBaseMode <<= fromWhere;
		fMode = &fBaseMode;
	}
}

bool IFillBundle::operator==(const IAttributeState& obj)  const
{
	if (GraphicTypeid(*this) == GraphicTypeid(obj))
	{
#if IC_RTTI_ENABLED
		const IFillBundle& bundle = dynamic_cast<const IFillBundle&>(obj);
#else
		const IFillBundle& bundle = *((const IFillBundle*)&obj);
#endif // IC_RTTI_ENABLED
		return ((fPaint == bundle.fPaint) && (*fMode == *bundle.fMode));
	}
	else
		return false;
}


/*================
||
||  class IFrameBundle
||
*/

IAttributeState::EDrawOperation
IFrameBundle::fgDrawOperation = IAttributeState::kFrame;

StreamableDefinitionsMacro(IFrameBundle, gGraphicsStreamModule);

IFrameBundle::IFrameBundle() :
	fBaseMode(IColorTransferMode::kSourceCopy),
	fPaint(IBaseColor(IBaseColor::kBlack)),
	fPen(IPen::kSolid, 1)
{
	fMode = &fBaseMode;
}

IFrameBundle::IFrameBundle(const IFrameBundle& copy) :
	IAttributeState(copy),
	fPaint(copy.fPaint),
	fPen(copy.fPen)
{
	if (GraphicTypeid(*copy.fMode) == GraphicTypeid(fBaseMode))
	{
		fBaseMode = *copy.fMode;
		fMode = &fBaseMode;
	}
	else
	{
		fMode = GraphicCopy(*copy.fMode);
	}
}

IFrameBundle::IFrameBundle(
	const IBaseColor& frameColor,
	GCoordinate penWidth,
	IPen::EPenBalance balance) :

	fBaseMode(IColorTransferMode::kSourceCopy),
	fPaint(frameColor),
	fPen(IPen::kSolid, penWidth, balance)
{
	fMode = &fBaseMode;
}

IFrameBundle::IFrameBundle(
	const IBaseColor& frameColor,
	const IColorTransferMode& transferMode,
	GCoordinate penWidth,
	IPen::EPenBalance balance) :

	fPaint(frameColor),
	fPen(IPen::kSolid, penWidth, balance)
{
	
	if (GraphicTypeid(transferMode) == GraphicTypeid(fBaseMode))
	{
		fBaseMode = transferMode;
		fMode = &fBaseMode;
	}
	else
	{
		fMode = GraphicCopy(transferMode);
	}
}

IFrameBundle::~IFrameBundle()
{
	if (fMode != &fBaseMode) delete fMode;
}

IFrameBundle& IFrameBundle::operator=(const IFrameBundle& obj)
{
	if (&obj != this)
	{
		IAttributeState::operator=(obj);
			
		fPaint = obj.fPaint;
		setFrameTransferMode(*obj.fMode);
		fPen = obj.fPen;
	}

	return *this;
}

void IFrameBundle::setFrameColor(const IBaseColor& color)
{
	fPaint.setColor(color);
}

const IPaint* IFrameBundle::framePaint()  const
{
	return &fPaint;
}

const IColorTransferMode* IFrameBundle::frameTransferMode() const
{
	return fMode;
}

void IFrameBundle::setFrameTransferMode(const IColorTransferMode& transferMode)
{
	if (GraphicTypeid(transferMode) != GraphicTypeid(*fMode))
	{
		if (fMode != &fBaseMode)
			delete fMode;

		if (GraphicTypeid(transferMode) == GraphicTypeid(fBaseMode))
		{	// copy it into the fBaseMode
			fBaseMode = transferMode;
			fMode = &fBaseMode;
		}
		else
		{	// clone the obj
			fMode = GraphicCopy(transferMode);
		}
	}
	else
	{	// reuse the storage...
		*fMode = transferMode;
	}
}

void IFrameBundle::setPenWidth(GCoordinate penWidth)
{
	fPen.setPenWidth(penWidth);
}

void IFrameBundle::setPenBalance(IPen::EPenBalance balance)
{
	fPen.setPenBalance(balance);
}

const IPen* IFrameBundle::framePen() const
{
	return &fPen;
}

IAttributeState::EDrawOperation IFrameBundle::drawingOperation() const
{
	return fgDrawOperation;
}

void IFrameBundle::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
	IAttributeState::writeToStream(toWhere);

	fPaint >>= toWhere;
	fPen >>= toWhere;

	bool streamPtr = true;
	if (fMode == &fBaseMode)
		streamPtr = false;
	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fMode, toWhere);
	}
	else
	{
		fBaseMode >>= toWhere;
	}

}

void IFrameBundle::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
	IAttributeState::readFromStream(fromWhere);

	fPaint <<= fromWhere;
	fPen <<= fromWhere;

	bool streamPtr;
	long temp;
	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fMode != &fBaseMode)
		delete fMode;
	if (streamPtr)
	{
		::readObject(fMode, fromWhere);
	}
	else
	{
		fBaseMode <<= fromWhere;
		fMode = &fBaseMode;
	}

}

bool IFrameBundle::operator==(const IAttributeState& obj) const
{
	if (GraphicTypeid(*this) == GraphicTypeid(obj))
	{
#if IC_RTTI_ENABLED
		const IFrameBundle& other = dynamic_cast<const IFrameBundle&>(obj);
#else
		const IFrameBundle& other = *((const IFrameBundle*) &obj);
#endif // IC_RTTI_ENABLED
		return
			(fPaint == other.fPaint) &&
			(*fMode == *other.fMode) &&
			(fPen == other.fPen);
	}
	else
		return false;
}

/*================
||
||  class IFillAndFrameBundle
||
*/

IAttributeState::EDrawOperation
IFillAndFrameBundle::fgDrawOperation = IAttributeState::kFillAndFrame;

StreamableDefinitionsMacro(IFillAndFrameBundle, gGraphicsStreamModule);

IFillAndFrameBundle::IFillAndFrameBundle() :
	fBaseFillMode(IColorTransferMode::kSourceCopy),
	fFillPaint(IBaseColor(IBaseColor::kBlack)),
	fBaseFrameMode(IColorTransferMode::kSourceCopy),
	fFramePaint(IBaseColor(IBaseColor::kBlack)),
	fPen(IPen::kSolid, 1)
{
	fFillMode = &fBaseFillMode;
	fFrameMode = &fBaseFrameMode;
}

IFillAndFrameBundle::IFillAndFrameBundle(const IFillAndFrameBundle& copy) :
	IAttributeState(copy),
	fFillPaint(copy.fFillPaint),
	fFramePaint(copy.fFramePaint),
	fPen(copy.fPen)
{
	if (GraphicTypeid(*copy.fFillMode) == GraphicTypeid(fBaseFillMode))
	{
		fBaseFillMode = *copy.fFillMode;
		fFillMode = &fBaseFillMode;
	}
	else
	{
		fFillMode = GraphicCopy(*copy.fFillMode);
	}

	if (GraphicTypeid(*copy.fFrameMode) == GraphicTypeid(fBaseFrameMode))
	{
		fBaseFrameMode = *copy.fFrameMode;
		fFrameMode = &fBaseFrameMode;
	}
	else
	{
		fFrameMode = GraphicCopy(*copy.fFrameMode);
	}
}

IFillAndFrameBundle::IFillAndFrameBundle(
	const IBaseColor& fillColor,
	const IBaseColor& frameColor,
	GCoordinate penWidth,
	IPen::EPenBalance balance) :

	fBaseFillMode(IColorTransferMode::kSourceCopy),
	fBaseFrameMode(IColorTransferMode::kSourceCopy),
	fFillPaint(fillColor),
	fFramePaint(frameColor),
	fPen(IPen::kSolid, penWidth, balance)
{
	fFillMode = &fBaseFillMode;
	fFrameMode = &fBaseFrameMode;
}

IFillAndFrameBundle::IFillAndFrameBundle(
	const IBaseColor& fillColor,
	const IColorTransferMode& fillTransferMode,
	const IBaseColor& frameColor,
	const IColorTransferMode& frameTransferMode,
	GCoordinate penWidth,
	IPen::EPenBalance balance) :

	fFillPaint(fillColor),
	fFramePaint(frameColor),
	fPen(IPen::kSolid, penWidth, balance)
{
	if (GraphicTypeid(fillTransferMode) == GraphicTypeid(fBaseFillMode))
	{
		fBaseFillMode = fillTransferMode;
		fFillMode = &fBaseFillMode;
	}
	else
	{
		fFillMode = GraphicCopy(fillTransferMode);
	}

	if (GraphicTypeid(frameTransferMode) == GraphicTypeid(fBaseFrameMode))
	{
		fBaseFrameMode = frameTransferMode;
		fFrameMode = &fBaseFrameMode;
	}
	else
	{
		fFrameMode = GraphicCopy(frameTransferMode);
	}
}

IFillAndFrameBundle::~IFillAndFrameBundle()
{
	if (fFillMode != &fBaseFillMode) delete fFillMode;
	if (fFrameMode != &fBaseFrameMode) delete fFrameMode;
}

IFillAndFrameBundle& IFillAndFrameBundle::operator=(const IFillAndFrameBundle& obj)
{
	if (&obj != this)
	{
		IAttributeState::operator=(obj);
			
		fFillPaint = obj.fFillPaint;
		fFramePaint = obj.fFramePaint;
		setFillTransferMode(*obj.fFillMode);
		setFrameTransferMode(*obj.fFrameMode);
		fPen = obj.fPen;
	}
	
	return *this;
}

void IFillAndFrameBundle::setFillColor(const IBaseColor& color)
{
	fFillPaint.setColor(color);
}

const IPaint*
IFillAndFrameBundle::fillPaint() const
{
	return &fFillPaint;
}

void IFillAndFrameBundle::setFillTransferMode(
	const IColorTransferMode& transferMode)
{
	if (GraphicTypeid(transferMode) != GraphicTypeid(*fFillMode))
	{
		if (fFillMode != &fBaseFillMode)
			delete fFillMode;

		if (GraphicTypeid(transferMode) == GraphicTypeid(fBaseFillMode))
		{
			// copy it into the fBaseFillMode
			fBaseFillMode = transferMode;
			fFillMode = &fBaseFillMode;
		}
		else
		{	
			// clone the obj
			fFillMode = GraphicCopy(transferMode);
		}
	}
	else
	{	
		// reuse the storage...
		*fFillMode = transferMode;
	}
}

const IColorTransferMode* IFillAndFrameBundle::fillTransferMode()
	const
{
	return fFillMode;
}

void IFillAndFrameBundle::setFrameColor(const IBaseColor& color)
{
	fFramePaint.setColor(color);
}

const IPaint* IFillAndFrameBundle::framePaint() const
{
	return &fFramePaint;
}

const IColorTransferMode* IFillAndFrameBundle::frameTransferMode() const
{
	return 0 /*NIL*/;
}

void IFillAndFrameBundle::setFrameTransferMode(
	const IColorTransferMode&  transferMode)
{
	if (GraphicTypeid(transferMode) != GraphicTypeid(*fFrameMode))
	{
		if (fFrameMode != &fBaseFrameMode)
			delete fFrameMode;

		if (GraphicTypeid(transferMode) == GraphicTypeid(fBaseFrameMode))
		{	
			// copy it into the fBaseFrameMode
			fBaseFrameMode = transferMode;
			fFrameMode = &fBaseFrameMode;
		}
		else
		{	
			fFrameMode = GraphicCopy(transferMode);
		}
	}
	else
	{	// reuse the storage...
		*fFrameMode = transferMode;
	}

}

void IFillAndFrameBundle::setPenWidth(GCoordinate penWidth)
{
	fPen.setPenWidth(penWidth);
}

void IFillAndFrameBundle::setPenBalance(IPen::EPenBalance balance)
{
	fPen.setPenBalance(balance);
}

const IPen* IFillAndFrameBundle::framePen() const
{
	return &fPen;
}

IAttributeState::EDrawOperation IFillAndFrameBundle::drawingOperation() const
{
	return fgDrawOperation;
}

void IFillAndFrameBundle::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
	IAttributeState::writeToStream(toWhere);

	fFillPaint >>= toWhere;
	fFramePaint >>= toWhere;
	fPen >>= toWhere;

	bool streamPtr = true;
	if (fFillMode == &fBaseFillMode)
		streamPtr = false;
	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fFillMode, toWhere);
	}
	else
	{
		fBaseFillMode >>= toWhere;
	}

	streamPtr = true;
	if (fFrameMode == &fBaseFrameMode)
		streamPtr = false;
	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fFrameMode, toWhere);
	}
	else
	{
		fBaseFrameMode >>= toWhere;
	}
}

void IFillAndFrameBundle::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
	IAttributeState::readFromStream(fromWhere);

	fFillPaint <<= fromWhere;
	fFramePaint <<= fromWhere;
	fPen <<= fromWhere;

	bool streamPtr;
	long temp;
	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fFillMode != &fBaseFillMode)
		delete fFillMode;
	if (streamPtr)
	{
		::readObject(fFillMode, fromWhere);
	}
	else
	{
		fBaseFillMode <<= fromWhere;
		fFillMode = &fBaseFillMode;
	}
	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fFrameMode != &fBaseFrameMode)
		delete fFrameMode;
	if (streamPtr)
	{
		::readObject(fFrameMode, fromWhere);
	}
	else
	{
		fBaseFrameMode <<= fromWhere;
		fFrameMode = &fBaseFrameMode;
	}
}

bool IFillAndFrameBundle::operator==(const IAttributeState& obj)  const
{
	if (GraphicTypeid(*this) == GraphicTypeid(obj))
	{
#if IC_RTTI_ENABLED
		const IFillAndFrameBundle& other = dynamic_cast<const IFillAndFrameBundle&>(obj);
#else
		const IFillAndFrameBundle& other = *((const IFillAndFrameBundle*) &obj);
#endif // IC_RTTI_ENABLED
		return
			(fFillPaint == other.fFillPaint)   &&
			(fFramePaint == other.fFramePaint) &&
			(*fFillMode == *other.fFillMode)   &&
			(*fFrameMode == *other.fFrameMode) &&
			(fPen == other.fPen);
	}
	else
		return false;
}

/*================
||
||  class IGrafBundle
||
*/

StreamableDefinitionsMacro(IGrafBundle, gGraphicsStreamModule);

IGrafBundle::IGrafBundle() :
	fFillPaint(0 /*NIL*/),
	fFillMode(0 /*NIL*/),
	fFramePaint(0 /*NIL*/),
	fFrameMode(0 /*NIL*/),
	fPen(0 /*NIL*/),
	fJoint(0 /*NIL*/),
	fEndCap(0 /*NIL*/),
	fImageMode(0 /*NIL*/),
	fSamplingControl(0 /*NIL*/),
	fDrawOperation(IAttributeState::kUnset)
{}

IGrafBundle::IGrafBundle(
	const IPaint& fPaint,
	IAttributeState::EDrawOperation attribute) :

	fFillPaint(0 /*NIL*/),
	fFillMode(0 /*NIL*/),
	fFramePaint(0 /*NIL*/),
	fFrameMode(0 /*NIL*/),
	fPen(0 /*NIL*/),
	fJoint(0 /*NIL*/),
	fEndCap(0 /*NIL*/),
	fImageMode(0 /*NIL*/),
	fSamplingControl(0 /*NIL*/),
	fDrawOperation(attribute)
{
	if (attribute != kUnset)
	{
		if (GraphicTypeid(fPaint) == GraphicTypeid(fBaseFramePaint))
		{
			if (attribute == kFrame)
			{
				fBaseFramePaint = fPaint;
				fFramePaint = &fBaseFramePaint;
			}
			else if (attribute == kFill)
			{
				fBaseFillPaint = fPaint;
				fFillPaint = &fBaseFillPaint;
			}
			else if (attribute == kFillAndFrame)
			{
				fBaseFillPaint = fPaint;
				fFillPaint = &fBaseFillPaint;
				fBaseFramePaint = fPaint;
				fFramePaint = &fBaseFramePaint;
			}
		}
		else
		{
			if (attribute == kFrame)
			{
				fFramePaint = GraphicCopy(fPaint);
			}
			else if (attribute == kFill)
			{
				fFillPaint = GraphicCopy(fPaint);
			}
			else if (attribute == kFillAndFrame)
			{
				fFramePaint = GraphicCopy(fPaint);
				fFillPaint = GraphicCopy(fPaint);
			}
		}
	}
}

IGrafBundle::IGrafBundle(
	const IPaint& fillPaint,
	const IPaint& framePaint,
	IAttributeState::EDrawOperation attribute) :

	fFillPaint(0 /*NIL*/),
	fFillMode(0 /*NIL*/),
	fFramePaint(0 /*NIL*/),
	fFrameMode(0 /*NIL*/),
	fPen(0 /*NIL*/),
	fJoint(0 /*NIL*/),
	fEndCap(0 /*NIL*/),
	fImageMode(0 /*NIL*/),
	fSamplingControl(0 /*NIL*/),
	fDrawOperation(attribute)
{
	if (GraphicTypeid(fillPaint) == GraphicTypeid(fBaseFillPaint))
	{
		fBaseFillPaint = fillPaint;
		fFillPaint = &fBaseFillPaint;
	}
	else
	{
		fFillPaint = GraphicCopy(fillPaint);
	}

	if (GraphicTypeid(framePaint) == GraphicTypeid(fBaseFramePaint))
	{
		fBaseFramePaint = framePaint;
		fFramePaint = &fBaseFramePaint;
	}
	else
	{
		fFramePaint = GraphicCopy(framePaint);
	}
}

IGrafBundle::IGrafBundle(
	const IBaseColor& color,
	IAttributeState::EDrawOperation attribute) :

	fFillPaint(0 /*NIL*/),
	fFillMode(0 /*NIL*/),
	fFramePaint(0 /*NIL*/),
	fFrameMode(0 /*NIL*/),
	fPen(0 /*NIL*/),
	fJoint(0 /*NIL*/),
	fEndCap(0 /*NIL*/),
	fImageMode(0 /*NIL*/),
	fSamplingControl(0 /*NIL*/),
	fDrawOperation(attribute)
{
	if (attribute == kFill)
	{
		fBaseFillPaint.setColor(color);
		fFillPaint = &fBaseFillPaint;
	}
	else if (attribute == kFrame)
	{
		fBaseFramePaint.setColor(color);
		fFramePaint = &fBaseFramePaint;
	}
	else if (attribute == kFillAndFrame)
	{
		fBaseFillPaint.setColor(color);
		fFillPaint = &fBaseFillPaint;
		fBaseFramePaint.setColor(color);
		fFramePaint = &fBaseFramePaint;
	}
}

IGrafBundle::IGrafBundle(
	const IBaseColor& fillColor,
	const IBaseColor& frameColor,
	IAttributeState::EDrawOperation attribute) :

	fFillPaint(0 /*NIL*/),
	fFillMode(0 /*NIL*/),
	fFramePaint(0 /*NIL*/),
	fFrameMode(0 /*NIL*/),
	fPen(0 /*NIL*/),
	fJoint(0 /*NIL*/),
	fEndCap(0 /*NIL*/),
	fImageMode(0 /*NIL*/),
	fSamplingControl(0 /*NIL*/),
	fDrawOperation(attribute)
{
	fBaseFillPaint.setColor(fillColor);
	fFillPaint = &fBaseFillPaint;

	fBaseFramePaint.setColor(frameColor);
	fFramePaint = &fBaseFramePaint;
}

IGrafBundle::IGrafBundle(const IGrafBundle& obj) :
	fFillPaint(0 /*NIL*/),
	fFillMode(0 /*NIL*/),
	fFramePaint(0 /*NIL*/),
	fFrameMode(0 /*NIL*/),
	fPen(0 /*NIL*/),
	fJoint(0 /*NIL*/),
	fEndCap(0 /*NIL*/),
	fImageMode(0 /*NIL*/),
	fSamplingControl(0 /*NIL*/),
	fDrawOperation(obj.fDrawOperation)
{

	if (obj.fFillPaint)
	{
		if (GraphicTypeid(*obj.fFillPaint) == GraphicTypeid(fBaseFillPaint))
		{
			fBaseFillPaint = *obj.fFillPaint;
			fFillPaint = &fBaseFillPaint;
		}
		else
		{
			fFillPaint = GraphicCopy(*obj.fFillPaint);
		}
	}

	if (obj.fFramePaint)
	{
		if (GraphicTypeid(*obj.fFramePaint) == GraphicTypeid(fBaseFramePaint))
		{
			fBaseFramePaint = *obj.fFramePaint;
			fFramePaint = &fBaseFramePaint;
		}
		else
		{
			fFramePaint = GraphicCopy(*obj.fFramePaint);
		}
	}


	if (obj.fFillMode)
	{
		if (GraphicTypeid(*obj.fFillMode) == GraphicTypeid(fBaseFillMode))
		{
			fBaseFillMode = *obj.fFillMode;
			fFillMode = &fBaseFillMode;
		}
		else
		{
			fFillMode = GraphicCopy(*obj.fFillMode);
		}
	}


	if (obj.fFrameMode)
	{
		if (GraphicTypeid(*obj.fFrameMode) == GraphicTypeid(fBaseFrameMode))
		{
			fBaseFrameMode = *obj.fFrameMode;
			fFrameMode = &fBaseFrameMode;
		}
		else
		{
			fFrameMode = GraphicCopy(*obj.fFrameMode);
		}
	}


	if (obj.fPen)
	{
		if (GraphicTypeid(*obj.fPen) == GraphicTypeid(fBasePen))
		{
			fBasePen = *obj.fPen;
			fPen = &fBasePen;
		}
		else
		{
			fPen = GraphicCopy(*obj.fPen);
		}
	}


	if (obj.fJoint)
	{
		if (GraphicTypeid(*obj.fJoint) == GraphicTypeid(fBaseJoint))
		{
			fBaseJoint = *obj.fJoint;
			fJoint = &fBaseJoint;
		}
		else
		{
			fJoint = GraphicCopy(*obj.fJoint);
		}
	}


	if (obj.fEndCap)
	{
		if (GraphicTypeid(*obj.fEndCap) == GraphicTypeid(fBaseEndCap))
		{
			fBaseEndCap = *obj.fEndCap;
			fEndCap = &fBaseEndCap;
		}
		else
		{
			fEndCap = GraphicCopy(*obj.fEndCap);
		}
	}


	if (obj.fImageMode)
	{
		if (GraphicTypeid(*obj.fImageMode) == GraphicTypeid(fBaseImageMode))
		{
			fBaseImageMode = *obj.fImageMode;
			fImageMode = &fBaseImageMode;
		}
		else
		{
			fImageMode = GraphicCopy(*obj.fImageMode);
		}
	}
	
	
	if (obj.fSamplingControl)
	{
		if (GraphicTypeid(*obj.fSamplingControl) == GraphicTypeid(fBaseSamplingControl))
		{
			fBaseSamplingControl = *obj.fSamplingControl;
			fSamplingControl = &fBaseSamplingControl;
		}
		else
		{
			fSamplingControl = GraphicCopy(*obj.fSamplingControl);
		}
	}
}

IGrafBundle::IGrafBundle(const IAttributeState& obj) :
	fFillPaint(0 /*NIL*/),
	fFillMode(0 /*NIL*/),
	fFramePaint(0 /*NIL*/),
	fFrameMode(0 /*NIL*/),
	fPen(0 /*NIL*/),
	fJoint(0 /*NIL*/),
	fEndCap(0 /*NIL*/),
	fImageMode(0 /*NIL*/),
	fSamplingControl(0 /*NIL*/),
	fDrawOperation(obj.drawingOperation())
{
	
	if (obj.fillPaint())
	{
		if (GraphicTypeid(*obj.fillPaint()) == GraphicTypeid(fBaseFillPaint))
		{
			fBaseFillPaint = *obj.fillPaint();
			fFillPaint = &fBaseFillPaint;
		}
		else
		{
			fFillPaint = GraphicCopy(*obj.fillPaint());
		}
	}
	
	if (obj.framePaint())
	{
		if (GraphicTypeid(*obj.framePaint()) == GraphicTypeid(fBaseFramePaint))
		{
			fBaseFramePaint = *obj.framePaint();
			fFramePaint = &fBaseFramePaint;
		}
		else
		{
			fFramePaint = GraphicCopy(*obj.framePaint());
		}
	}
	
	
	if (obj.fillTransferMode())
	{
		if (GraphicTypeid(*obj.fillTransferMode()) == GraphicTypeid(fBaseFillMode))
		{
			fBaseFillMode = *obj.fillTransferMode();
			fFillMode = &fBaseFillMode;
		}
		else
		{
			fFillMode = GraphicCopy(*obj.fillTransferMode());
		}
	}
	
	
	if (obj.frameTransferMode())
	{
		if (GraphicTypeid(*obj.frameTransferMode()) == GraphicTypeid(fBaseFrameMode))
		{
			fBaseFrameMode = *obj.frameTransferMode();
			fFrameMode = &fBaseFrameMode;
		}
		else
		{
			fFrameMode = GraphicCopy(*obj.frameTransferMode());
		}
	}
	
	
	if (obj.framePen())
	{
		if (GraphicTypeid(*obj.framePen()) == GraphicTypeid(fBasePen))
		{
			fBasePen = *obj.framePen();
			fPen = &fBasePen;
		}
		else
		{
			fPen = GraphicCopy(*obj.framePen());
		}
	}
	
	
	if (obj.frameJoint())
	{
		if (GraphicTypeid(*obj.frameJoint()) == GraphicTypeid(fBaseJoint))
		{
			fBaseJoint = *obj.frameJoint();
			fJoint = &fBaseJoint;
		}
		else
		{
			fJoint = GraphicCopy(*obj.frameJoint());
		}
	}
	
	
	if (obj.frameEndCap())
	{
		if (GraphicTypeid(*obj.frameEndCap()) == GraphicTypeid(fBaseEndCap))
		{
			fBaseEndCap = *obj.frameEndCap();
			fEndCap = &fBaseEndCap;
		}
		else
		{
			fEndCap = GraphicCopy(*obj.frameEndCap());
		}
	}
	
	
	if (obj.imageTransferMode())
	{
		if (GraphicTypeid(*obj.imageTransferMode()) ==
	GraphicTypeid(fBaseImageMode))
		{
			fBaseImageMode = *obj.imageTransferMode();
			fImageMode = &fBaseImageMode;
		}
		else
		{
			fImageMode = GraphicCopy(*obj.imageTransferMode());
		}
	}
	
	
	if (obj.imageSampling())
	{
		if (GraphicTypeid(*obj.imageSampling()) == GraphicTypeid(fBaseSamplingControl))
		{
			fBaseSamplingControl = *obj.imageSampling();
			fSamplingControl = &fBaseSamplingControl;
		}
		else
		{
			fSamplingControl = GraphicCopy(*obj.imageSampling());
		}
	}
}

IGrafBundle::~IGrafBundle()
{
	if (fFillPaint != &fBaseFillPaint) delete fFillPaint;
	
	if (fFramePaint != &fBaseFramePaint) delete fFramePaint;
	
	if (fFillMode != &fBaseFillMode) delete fFillMode;
	
	if (fFrameMode != &fBaseFrameMode) delete fFrameMode;
	
	if (fPen != &fBasePen) delete fPen;
	
	if (fJoint != &fBaseJoint) delete fJoint;
	
	if (fEndCap != &fBaseEndCap) delete fEndCap;
	
	if (fImageMode != &fBaseImageMode) delete fImageMode;
	
	if (fSamplingControl != &fBaseSamplingControl) delete fSamplingControl;
}

const IPaint* IGrafBundle::fillPaint() const
{
	return fFillPaint;
}

void IGrafBundle::setFillPaint(const IPaint& obj)
{
	if (!fFillPaint || GraphicTypeid(obj) != GraphicTypeid(*fFillPaint))
	{
		if (fFillPaint && fFillPaint != &fBaseFillPaint)
			delete fFillPaint;
	
		if (GraphicTypeid(obj) == GraphicTypeid(fBaseFillPaint))
		{
			fBaseFillPaint = obj;
			fFillPaint = &fBaseFillPaint;
		}
		else
		{
			fFillPaint = GraphicCopy(obj);
		}
	}
	else
	{	// reuse the storage...
		*fFillPaint = obj;
	}
}

void IGrafBundle::removeFillPaint()
{
	if (fFillPaint != &fBaseFillPaint)
		delete fFillPaint;
	fFillPaint = NULL;
}

const IBaseColor* IGrafBundle::fillColor() const
{
	return fFillPaint->color();
}

void IGrafBundle::setFillColor(const IBaseColor& color)
{
	if (fFillPaint)
		fFillPaint->setColor(color);
	else
	{
		fBaseFillPaint.setColor(color);
		fFillPaint = &fBaseFillPaint;
	}
}

const IColorTransferMode* IGrafBundle::fillTransferMode() const
{
	return fFillMode;
}

void IGrafBundle::setFillTransferMode(const IColorTransferMode& obj)
{
	if (!fFillMode || GraphicTypeid(obj) != GraphicTypeid(*fFillMode))
	{
		if (fFillMode != &fBaseFillMode)
			delete fFillMode;
		
		if (GraphicTypeid(obj) == GraphicTypeid(fBaseFillMode))
		{
			fBaseFillMode = obj;
			fFillMode = &fBaseFillMode;
		}
		else
		{
			fFillMode = GraphicCopy(obj);
		}
	}
	else
	{	
		*fFillMode = obj;
	}
}

void IGrafBundle::removeFillTransferMode()
{
	if (fFillMode != &fBaseFillMode)
		delete fFillMode;
	fFillMode = NULL;
}

const IPaint* IGrafBundle::framePaint() const
{
	return fFramePaint;
}

void IGrafBundle::setFramePaint(const IPaint& obj)
{
	if (!fFramePaint || GraphicTypeid(obj) != GraphicTypeid(*fFramePaint))
	{
		if (fFramePaint != &fBaseFramePaint)
			delete fFramePaint;
		
		if (GraphicTypeid(obj) == GraphicTypeid(fBaseFramePaint))
		{
			fBaseFramePaint = obj;
			fFramePaint = &fBaseFramePaint;
		}
		else
		{
			fFramePaint = GraphicCopy(obj);
		}
	}
	else
	{	
		*fFramePaint = obj;
	}
}

void IGrafBundle::removeFramePaint()
{
	if (fFramePaint != &fBaseFramePaint)
		delete fFramePaint;
	fFramePaint = NULL;
}

const IBaseColor* IGrafBundle::frameColor() const
{
	return fFramePaint->color();
}

void IGrafBundle::setFrameColor(const IBaseColor& color)
{
	if (fFramePaint)
		fFramePaint->setColor(color);
	else
	{
		fBaseFramePaint.setColor(color);
		fFramePaint = &fBaseFramePaint;
	}
}

const IColorTransferMode* IGrafBundle::frameTransferMode() const
{
	return fFrameMode;
}

void IGrafBundle::setFrameTransferMode(const IColorTransferMode& obj)
{
	if (!fFrameMode || GraphicTypeid(obj) != GraphicTypeid(*fFrameMode))
	{
		if (fFrameMode != &fBaseFrameMode)
			delete fFrameMode;
		
		if (GraphicTypeid(obj) == GraphicTypeid(fBaseFrameMode))
		{
			fBaseFrameMode = obj;
			fFrameMode = &fBaseFrameMode;
		}
		else
		{
			fFrameMode = GraphicCopy(obj);
		}
	}
	else
	{	
		*fFrameMode = obj;
	}
}

void IGrafBundle::removeFrameTransferMode()
{
	if (fFrameMode != &fBaseFrameMode) delete fFrameMode;
	fFrameMode = NULL;
}

const IPen* IGrafBundle::framePen() const
{
	return fPen;
}

void IGrafBundle::setFramePen(const IPen& obj)
{
	if (!fPen || GraphicTypeid(obj) != GraphicTypeid(*fPen))
	{
		if (fPen != &fBasePen)
			delete fPen;
		
		if (GraphicTypeid(obj) == GraphicTypeid(fBasePen))
		{
			fBasePen = obj;
			fPen = &fBasePen;
		}
		else
		{
			fPen = GraphicCopy(obj);
		}
	}
	else
	{	
		*fPen = obj;
	}
}

void IGrafBundle::removeFramePen()
{
	if (fPen != &fBasePen) delete fPen;
	fPen = NULL;
}

const ICap*	IGrafBundle::frameEndCap() const
{
	return fEndCap;
}

void IGrafBundle::setFrameEndCap(const ICap& obj)
{
	if (!fEndCap || GraphicTypeid(obj) != GraphicTypeid(*fEndCap))
	{
		if (fEndCap != &fBaseEndCap)
			delete fEndCap;
		
		if (GraphicTypeid(obj) == GraphicTypeid(fBaseEndCap))
		{
			fBaseEndCap = obj;
			fEndCap = &fBaseEndCap;
		}
		else
		{
			fEndCap = GraphicCopy(obj);
		}
	}
	else
	{	
		*fEndCap = obj;
	}
}

void IGrafBundle::removeFrameEndCap()
{
	if (fEndCap != &fBaseEndCap) delete fEndCap;
	fEndCap = NULL;
}

const IJoint* IGrafBundle::frameJoint() const
{
	return fJoint;
}

void IGrafBundle::setFrameJoint(const IJoint& obj)
{
	if (!fJoint || GraphicTypeid(obj) != GraphicTypeid(*fJoint))
	{
		if (fJoint != &fBaseJoint)
			delete fJoint;
		
		if (GraphicTypeid(obj) == GraphicTypeid(fBaseJoint))
		{
			fBaseJoint = obj;
			fJoint = &fBaseJoint;
		}
		else
		{
			fJoint = GraphicCopy(obj);
		}
	}
	else
	{	
		*fJoint = obj;
	}
}

void IGrafBundle::removeFrameJoint()
{
	if (fJoint != &fBaseJoint) delete fJoint;
	fJoint = NULL;
}

const IImageTransferMode* IGrafBundle::imageTransferMode() const
{
	return fImageMode;
}

void IGrafBundle::setImageTransferMode(const IImageTransferMode& obj)
{
	if (!fImageMode || GraphicTypeid(obj) != GraphicTypeid(*fImageMode))
	{
		if (fImageMode != &fBaseImageMode)
			delete fImageMode;
		
		if (GraphicTypeid(obj) == GraphicTypeid(fBaseImageMode))
		{
			fBaseImageMode = obj;
			fImageMode = &fBaseImageMode;
		}
		else
		{
			fImageMode = GraphicCopy(obj);
		}
	}
	else
	{	
		*fImageMode = obj;
	}
}

void IGrafBundle::removeImageTransferMode()
{
	if (fImageMode != &fBaseImageMode) delete fImageMode;
	fImageMode = NULL;
}

const IImageSamplingControl* IGrafBundle::imageSampling() const
{
	return fSamplingControl;
}

void IGrafBundle::setImageSampling(const IImageSamplingControl& obj)
{
	if (!fSamplingControl || GraphicTypeid(obj) !=
			GraphicTypeid(*fSamplingControl))
	{
		if (fSamplingControl != &fBaseSamplingControl)
			delete fSamplingControl;
		
		if (GraphicTypeid(obj) == GraphicTypeid(fBaseSamplingControl))
		{
			fBaseSamplingControl = obj;
			fSamplingControl = &fBaseSamplingControl;
		}
		else
		{
			fSamplingControl = GraphicCopy(obj);
		}
	}
	else
	{	
		*fSamplingControl = obj;
	}
}

void IGrafBundle::removeImageSampling()
{
	if (fSamplingControl != &fBaseSamplingControl) delete fSamplingControl;
	fSamplingControl = NULL;
}

IAttributeState::EDrawOperation IGrafBundle::drawingOperation() const
{
	return fDrawOperation;
}

void IGrafBundle::setDrawingOperation(IAttributeState::EDrawOperation drawOper)
{
	fDrawOperation = drawOper;
}

IGrafBundle& IGrafBundle::operator=(const IGrafBundle& obj)
{
	if (obj.fFillPaint)
	{
		setFillPaint(*obj.fFillPaint);
	}
	else
	{
		if (fFillPaint != &fBaseFillPaint)
			delete fFillPaint;
		fFillPaint = NULL;
	}
	
	if (obj.fFramePaint)
	{
		setFramePaint(*obj.fFramePaint);
	}
	else
	{
		if (fFramePaint != &fBaseFramePaint)
			delete fFramePaint;
		fFramePaint = NULL;
	}
	
	if (obj.fFillMode)
	{
		setFillTransferMode(*obj.fFillMode);
	}
	else
	{
		if (fFillMode != &fBaseFillMode)
			delete fFillMode;
		fFillMode = NULL;
	}
	
	if (obj.fFrameMode)
	{
		setFrameTransferMode(*obj.fFrameMode);
	}
	else
	{
		if (fFrameMode != &fBaseFrameMode)
			delete fFrameMode;
		fFrameMode = NULL;
	}
	
	if (obj.fPen)
	{
		setFramePen(*obj.fPen);
	}
	else
	{
		if (fPen != &fBasePen)
			delete fPen;
		fPen = NULL;
	}
	
	if (obj.fJoint)
	{
		setFrameJoint(*obj.fJoint);
	}
	else
	{
		if (fJoint != &fBaseJoint)
			delete fJoint;
		fJoint = NULL;
	}
	
	if (obj.fEndCap)
	{
		setFrameEndCap(*obj.fEndCap);
	}
	else
	{
		if (fEndCap != &fBaseEndCap)
			delete fEndCap;
		fEndCap = NULL;
	}
	
	if (obj.fImageMode)
	{
		setImageTransferMode(*obj.fImageMode);
	}
	else
	{
		if (fImageMode != &fBaseImageMode)
			delete fImageMode;
		fImageMode = NULL;
	}
	
	if (obj.fSamplingControl)
	{
		setImageSampling(*obj.fSamplingControl);
	}
	else
	{
		if (fSamplingControl != &fBaseSamplingControl)
			delete fSamplingControl;
		fSamplingControl = NULL;
	}
	
	setDrawingOperation(obj.fDrawOperation);

	return (*this);
}

void IGrafBundle::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
	bool streamPtr = true;
	if (fFillPaint == &fBaseFillPaint)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fFillPaint, toWhere);
	}
	else
	{
		fBaseFillPaint >>= toWhere;
	}

	streamPtr = true;
	if (fFillMode == &fBaseFillMode)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fFillMode, toWhere);
	}
	else
	{
		fBaseFillMode >>= toWhere;
	}

	streamPtr = true;
	if (fFramePaint == &fBaseFramePaint)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fFramePaint, toWhere);
	}
	else
	{
		fBaseFramePaint >>= toWhere;
	}

	streamPtr = true;
	if (fFrameMode == &fBaseFrameMode)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fFrameMode, toWhere);
	}
	else
	{
		fBaseFrameMode >>= toWhere;
	}

	streamPtr = true;
	if (fPen == &fBasePen)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fPen, toWhere);
	}
	else
	{
		fBasePen >>= toWhere;
	}

	streamPtr = true;
	if (fJoint == &fBaseJoint)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fJoint, toWhere);
	}
	else
	{
		fBaseJoint >>= toWhere;
	}

	streamPtr = true;
	if (fEndCap == &fBaseEndCap)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fEndCap, toWhere);
	}
	else
	{
		fBaseEndCap >>= toWhere;
	}

	streamPtr = true;
	if (fImageMode == &fBaseImageMode)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fImageMode, toWhere);
	}
	else
	{
		fBaseImageMode >>= toWhere;
	}

	streamPtr = true;
	if (fSamplingControl == &fBaseSamplingControl)
		streamPtr = false;

	long(streamPtr) >>= toWhere;
	if (streamPtr)
	{
		::writeObject(fSamplingControl, toWhere);
	}
	else
	{
		fBaseSamplingControl >>= toWhere;
	}
}

void IGrafBundle::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
	bool streamPtr;
	long temp;

	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fFillPaint != &fBaseFillPaint)
		delete fFillPaint;
	if (streamPtr)
	{
		::readObject(fFillPaint, fromWhere);
	}else
	{
		fBaseFillPaint <<= fromWhere;
		fFillPaint = &fBaseFillPaint;
	}

	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fFillMode != &fBaseFillMode)
		delete fFillMode;
	if (streamPtr)
	{
		::readObject(fFillMode, fromWhere);
	}
	else
	{
		fBaseFillMode <<= fromWhere;
		fFillMode = &fBaseFillMode;
	}

	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fFramePaint != &fBaseFramePaint)
		delete fFramePaint;
	if (streamPtr)
	{
		::readObject(fFramePaint, fromWhere);
	}
	else
	{
		fBaseFramePaint <<= fromWhere;
		fFramePaint = &fBaseFramePaint;
	}

	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fFrameMode != &fBaseFrameMode)
		delete fFrameMode;
	if (streamPtr)
	{
		::readObject(fFrameMode, fromWhere);
	}
	else
	{
		fBaseFrameMode <<= fromWhere;
		fFrameMode = &fBaseFrameMode;
	}

	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fPen != &fBasePen)
		delete fPen;
	if (streamPtr)
	{
		::readObject(fPen, fromWhere);
	}
	else
	{
		fBasePen <<= fromWhere;
		fPen = &fBasePen;
	}

	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fJoint != &fBaseJoint)
		delete fJoint;
	if (streamPtr)
	{
		::readObject(fJoint, fromWhere);
	}else
	{
		fBaseJoint <<= fromWhere;
		fJoint = &fBaseJoint;
	}

	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fEndCap != &fBaseEndCap)
		delete fEndCap;
	if (streamPtr)
	{
		::readObject(fEndCap, fromWhere);
	}
	else
	{
		fBaseEndCap <<= fromWhere;
		fEndCap = &fBaseEndCap;
	}

	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fImageMode != &fBaseImageMode)
		delete fImageMode;
	if (streamPtr)
	{
		::readObject(fImageMode, fromWhere);
	}
	else
	{
		fBaseImageMode <<= fromWhere;
		fImageMode = &fBaseImageMode;
	}

	temp <<= fromWhere;
	streamPtr = bool(temp);
	if (fSamplingControl != &fBaseSamplingControl)
		delete fSamplingControl;
	if (streamPtr)
	{
		::readObject(fSamplingControl, fromWhere);
	}
	else
	{
		fBaseSamplingControl <<= fromWhere;
		fSamplingControl = &fBaseSamplingControl;
	}
}

bool IGrafBundle::operator==(const IAttributeState& obj) const
{
	
	if ((GraphicTypeid(*this) != GraphicTypeid(obj)) ||
	
		((fFillPaint && !obj.fillPaint()) ||
		 (!fFillPaint && obj.fillPaint()) ||
		 (fFillPaint && (*fFillPaint != *obj.fillPaint()))) ||

		((fFillMode && !obj.fillTransferMode()) ||
		 (!fFillMode && obj.fillTransferMode()) ||
		 (fFillMode && (*fFillMode != *obj.fillTransferMode()))) ||

		((fFramePaint && !obj.framePaint()) ||
		 (!fFramePaint && obj.framePaint()) ||
		 (fFramePaint && (*fFramePaint != *obj.framePaint()))) ||

		((fFrameMode && !obj.frameTransferMode()) ||
		 (!fFrameMode && obj.frameTransferMode()) ||
		 (fFrameMode && (*fFrameMode != *obj.frameTransferMode()))) ||

		((fPen && !obj.framePen()) ||
		 (!fPen && obj.framePen()) ||
		 (fPen && (*fPen != *obj.framePen()))) ||

		((fJoint && !obj.frameJoint()) ||
		 (!fJoint && obj.frameJoint()) ||
		 (fJoint && (*fJoint != *obj.frameJoint()))) ||

		((fEndCap && !obj.frameEndCap()) ||
		 (!fEndCap && obj.frameEndCap()) ||
		 (fEndCap && (*fEndCap != *obj.frameEndCap()))) ||
	
		((fImageMode && !obj.imageTransferMode()) ||
		 (!fImageMode && obj.imageTransferMode()) ||
		 (fImageMode && (*fImageMode != *obj.imageTransferMode()))) ||
	
		((fSamplingControl && !obj.imageSampling()) ||
		 (!fSamplingControl && obj.imageSampling()) ||
		 (fSamplingControl && (*fSamplingControl != *obj.imageSampling()))))

		return (false);
	else
		return true;
}
