// Revision: 01 1.11.1.3 source/albert/graph2d/icoordsy.cpp, 2d, ioc.v400, 000302 
/*******************************************************************************
* FILE NAME: icoordsy.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icoordsy.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#include <ibase.hpp>
#ifdef IC_WIN
#define INCL_WINMESSAGEMGR          // for WM_CALCVALIDRECTS
extern "C" {
  #include <windows.h>
}
#endif // IC_WIN

#include <icoordsy.hpp>
#include <iexcept.hpp>
#include <ipoint.hpp>
#include <irect.hpp>
#include <itrace.hpp>
#include <iprimlck.hpp>

#pragma priority(-2147481424)

// Segment definitions
#ifdef IC_PAGETUNE
#define _ICOORDSY_CPP_
#include <ipagetun.h>
#endif

class SetupOrientation
{
/*
	This class examines the environment variable ICL_COORD_SYSTEM when
	an object is created and sets ICoordinateSystem:fsapplication
	based on the value.  If the value of ICL_COORD_SYSTEM is UPPERLEFT, then
	ICoordinateSystem::originUpperLeft is set as the application orientation.
	Otherwise, the application orieintation is unmodified.
*/
public:
	static bool isDisabled();

private:
	static bool fDisabled;
    static bool fInited;
};



// defaulted always to lefthand (upperleft)
ICoordinateSystem::EOrientation
	ICoordinateSystem::fsapplication = ICoordinateSystem::kOriginUpperLeft ;

// static object to set the orientation for debugging
bool SetupOrientation::fDisabled = false;
bool SetupOrientation::fInited = false;


/*------------------------------------------------------------------------------
| SetupOrientation::SetupOrientation                                           |
------------------------------------------------------------------------------*/
bool SetupOrientation::isDisabled()
{
	IMODTRACE_ALL("SetupOrientation::SetupOrientation");
    
    if (!fInited) {
        IPrimalLock lock;
        if (!fInited) {
#ifdef IC_DEVELOP
        	// Not supported in prior versions
        	IString strTrace(getenv("ICLUI_COORD_SYSTEM"));
        	if (strTrace.length())
        	{
        		strTrace.strip().upperCase();
        		if (strTrace == "UPPERLEFT")
        		{
        			ITRACE_DEVELOP("Set coordinate system to upper left");
        			ICoordinateSystem::fsapplication = ICoordinateSystem::kOriginUpperLeft;
        			fDisabled = true;
        		}
        		else if (strTrace == "LOWERLEFT")
        		{
        			ITRACE_DEVELOP("Set coordinate system to lower left");
        			ICoordinateSystem::fsapplication = ICoordinateSystem::kOriginLowerLeft;
        			fDisabled = true;
        		}
        	}
#endif
            fInited = true;
        }
    }
    return fDisabled;
}


/*------------------------------------------------------------------------------
| ICoordinateSystem::ICoordinateSystem   (hidden ... all static class)         |
------------------------------------------------------------------------------*/
ICoordinateSystem::ICoordinateSystem()
{}

/*------------------------------------------------------------------------------
| ICoordinateSystem::convertToApplication                                      |
------------------------------------------------------------------------------*/
IRectangle ICoordinateSystem::convertToApplication(
	const IRectangle& nativeRectangle,
	const ISize& referenceSize)
{
	IMODTRACE_ALL("ICoordinateSystem::convertToApplication");
	// With the 2 coordinate systems supported, the
	// conversion algorithm is the same in either direction.
	return convertRectangle(nativeRectangle, referenceSize);
}


/*------------------------------------------------------------------------------
| ICoordinateSystem::convertToNative                                           |
------------------------------------------------------------------------------*/
IRectangle ICoordinateSystem::convertToNative(
	const IRectangle& applicationRectangle,
	const ISize& referenceSize)
{
	IMODTRACE_ALL("ICoordinateSystem::convertToNative");
	// With the 2 coordinate systems supported, the
	// conversion algorithm is the same in either direction.
	return convertRectangle(applicationRectangle, referenceSize);
}

/*------------------------------------------------------------------------------
| ICoordinateSystem::convertRectangle                                          |
------------------------------------------------------------------------------*/
IRectangle  ICoordinateSystem::convertRectangle(
	const IRectangle& rectangle,
	const ISize& referenceSize)
{
	ITRACE_ALL(IString("rectangle=") + rectangle.asString());
	IRectangle result = rectangle;
	if (isConversionNeeded())
	{
		// since there are only 2 coordinate systems supported we assume
		// that application is the opposite of native.  The desired y coordinate
		// for the origin is the distance between the far corner of the
		// reference rectangle and the corner of the rectangle to be converted.
		// IRectangle::maxY gives the "far corner" y coordinate, which we
		// subtract from the reference size height to give our desired new
		// coordinate.
		result.moveTo(
			IPoint(rectangle.minX(), referenceSize.height() - rectangle.maxY()));
	}
	ITRACE_ALL(IString("result=") + result.asString());
	return result;
}


/*------------------------------------------------------------------------------
| ICoordinateSystem::convertToApplication                                      |
------------------------------------------------------------------------------*/
IPoint ICoordinateSystem::convertToApplication(
	const IPoint& nativePoint,
	const ISize& referenceSize)
{
	IPoint result = nativePoint;
	if (isConversionNeeded())
		result.setY(referenceSize.height() - result.y());
	return result;
}

/*------------------------------------------------------------------------------
| ICoordinateSystem::convertToNative                                           |
------------------------------------------------------------------------------*/
IPoint ICoordinateSystem::convertToNative(
	const IPoint& applicationPoint,
	const ISize& referenceSize)
{
	IPoint result = applicationPoint;
	if (isConversionNeeded())
		result.setY(referenceSize.height() - result.y());
	return result;
}

/*------------------------------------------------------------------------------
| ICoordinateSystem::setApplicationOrientation                                 |
------------------------------------------------------------------------------*/
ICoordinateSystem::EOrientation ICoordinateSystem::setApplicationOrientation(EOrientation coordsystem)
{
#ifdef IC_DEVELOP
	// In debug build, check for orientation set by environment variable
	if (!SetupOrientation::isDisabled())
		fsapplication = coordsystem;
#else
	fsapplication = coordsystem;
#endif
	return fsapplication;
}

// Note: These functions are inlined on all platforms except Windows due to
//       variables they reference not being exported.
#ifdef IC_WIN
/*------------------------------------------------------------------------------
| ICoordinateSystem::applicationOrientation                                    |
------------------------------------------------------------------------------*/
ICoordinateSystem::EOrientation ICoordinateSystem::applicationOrientation()
{
	return fsapplication;
}

/*------------------------------------------------------------------------------
| ICoordinateSystem::isConversionNeeded                                        |
------------------------------------------------------------------------------*/
bool ICoordinateSystem::isConversionNeeded()
{
	return (fsapplication != nativeOrientation()) ;
}
#endif //IC_WIN


