/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 39 1.15.1.5 source/albert/graph2d/igrafatt.cpp, 2d, ioc.v400, 000302 
/*================
||
||  File:   GraphicsAttributes.C *
||
||  What:   Routines for the Albert Attribute State classes
||
||  Class defined in this file:
||		IPen
||		ICap
||		IJoint
||		IImageTransferMode
||		IColorTransferMode
||		IImageSamplingControl
||
*/

#include <igrafatt.hpp>
#include <inumeric.hpp>
#include <grassert.hpp>
#include <idatstrm.hpp>
#include <grstrmod.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
#define _IGRAFATT_CPP_
#include <ipagetun.h>
#endif

/*================
||
||  class IPen
||
||
*/

StreamableDefinitionsMacro(IPen, gGraphicsStreamModule);

IPen::IPen()
{
	setPenParams(kSolid, 1.0, kCenterFrame);
}

IPen::~IPen()
{}

IPen::IPen(IPen::EPenType type, GCoordinate penWidth, IPen::EPenBalance balance)
{
	// Currently setPenData is non-virtual. If it does become virtual
	// then this needs to be changed.
	setPenParams(type, penWidth, balance);
}

void IPen::setPenParams(IPen::EPenType type, GCoordinate penWidth, IPen::EPenBalance balance)
{
	fPenType = type;
#ifdef IC_MOTIF
	if (fPenType == IPen::kHairline ){
		fPenWidth = 0.0; //1 pixel width line 
		fPenBalance = IPen::kUnknownFrame;
    	}
	else if (fPenType == IPen::kUnknownPen){
		fPenWidth = 1.0;
		fPenBalance = IPen::kUnknownFrame;
    	}
#else
	if (fPenType == IPen::kHairline || fPenType == IPen::kUnknownPen){
		fPenWidth = 1.0;
		fPenBalance = IPen::kUnknownFrame;
    	}
#endif
	else {
		fPenWidth = penWidth;
		fPenBalance = balance;
    }
	if (fPenWidth < 0.0) {
		fPenWidth = 1.0; // this is a safe parameter
		// THROW an INVAILID PARAMETER EXCEPTION
	}
}
void IPen::setPenWidth(GCoordinate penWidth)
{
	setPenParams(fPenType, penWidth, fPenBalance);
}

void IPen::setPenType(IPen::EPenType penType)
{
	setPenParams(penType, fPenWidth, fPenBalance);
}

void IPen::setPenBalance(IPen::EPenBalance penBalance)
{
	setPenParams(fPenType, fPenWidth, penBalance);
}

IPen::EPenType IPen::penParams(GCoordinate& penWidth, IPen::EPenBalance& balance)	const
{
    penWidth = fPenWidth;
	balance = fPenBalance;
	return fPenType;
}


bool IPen::operator!=(const IPen& obj) const
{
	return !operator==(obj);
}

bool IPen::operator==(const IPen& data) const
{
	if (data.fPenType != fPenType)
		return false;
	if (data.fPenWidth != fPenWidth)
		return false;
	if (data.fPenBalance != fPenBalance)
		return false;
    return  true;
}

void IPen::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
    long(fPenType) >>= toWhere;
    long(fPenBalance) >>= toWhere;
    fPenWidth >>= toWhere; 		
}

void IPen::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
    long temp;
    temp <<= fromWhere;
    fPenType = EPenType(temp);

    temp <<= fromWhere;
    fPenBalance = EPenBalance(temp);

    fPenWidth <<= fromWhere;
}

IPen& IPen::operator=(const IPen& source)
{
    if (&source != this)
    {
		fPenType = source.fPenType;
		fPenBalance = source.fPenBalance;
		fPenWidth = source.fPenWidth;
    }
    return *this;
}


IPen::IPen(const IPen& source)
{	
	fPenType = source.fPenType;
	fPenBalance = source.fPenBalance;
	fPenWidth = source.fPenWidth;

}


IPen::EPenType IPen::penType() const
{
    return fPenType;
}


IPen::EPenBalance IPen::penBalance() const
{
    return fPenBalance;
}

GCoordinate IPen::penWidth() const
{
    return fPenWidth;
}

IGRect2D IPen::adjustBounds(const IGRect2D& bounds) const
{
	IGRect2D bbox = bounds;
	if (fPenType == kHairline)
	{
		bbox.inset(IGPoint2D(-1.,-1.));
	}
	else
	{
		GCoordinate width = penWidth();
		if (fPenBalance == kCenterFrame)
		{
			bbox.inset(IGPoint2D(-1.,-1.) * width);
		}
		// else if (fPenBalance == kInsetFrame)
		//     bounding box should remain unchanged...
	}

	return bbox;
}

/*================
||
||  class ICap
||
||
*/

StreamableDefinitionsMacro(ICap, gGraphicsStreamModule);

ICap::ICap(ICap::ECapType type)
{
	fCapType = type;
}

ICap::ICap(const ICap& src)
{
	fCapType = src.fCapType;
}

ICap::~ICap()
{}

ICap& ICap::operator=(const ICap& src)
{
   	if (&src != this)
		fCapType = src.fCapType;
    return *this;
}

ICap::ECapType ICap::capType() const
{
    return fCapType;
}


bool ICap::operator!=(const ICap& obj) const
{
	return !operator==(obj);
}

bool ICap::operator==(const ICap& src) const
{
    if (fCapType == src.fCapType)
    	return true;
	else
		return false;
}

void ICap::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
    long(fCapType) >>= toWhere;
}

void ICap::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
    long temp;
    temp <<= fromWhere;
    fCapType = ECapType(temp);
}

IGRect2D ICap::adjustBounds(const IGRect2D& bounds, const IPen* pen) const
{
	IGRect2D bbox = bounds;
	GCoordinate width;
	if (pen) {
		if (pen->penBalance() == IPen::kCenterFrame)
			width = pen->penWidth() * 0.5;
		else	// kInsetFrame
			width = 0.;		// no adjustment;
	}
	else	// hairline assumed
		width = 1.0;

	switch (fCapType)
	{
	case kRound:
	case kFlush:
		bbox.inset(IGPoint2D(-1.,-1.) * width);
		break;
	case kSquare:
		bbox.inset(IGPoint2D(-1.,-1.) * (width * SquareRoot(2.)));
		break;
	}
	return bbox;
}

/*================
||
||  class IJoint
||
||
*/

/*
 * There is no information of whether or not the trunc angle is used
 * or not. If indeed it were the case that the trunc angle is useful
 * then  we must add the one extra call to take care of it.
 */

StreamableDefinitionsMacro(IJoint, gGraphicsStreamModule);

IJoint::IJoint(IJoint::EJointType type, GCoordinate miterLimit)
{
	fJointType = type;
	fmiterLimit = (type == IJoint::kMiterLimit) ? miterLimit : 1.0 ; 	
}

IJoint::~IJoint()
{}

bool IJoint::operator!=(const IJoint& obj) const
{
	return !operator==(obj);
}
bool IJoint::operator==(const IJoint& src) const
{
   	if (src.fJointType != fJointType)
		return false;
	if (fJointType == IJoint::kMiterLimit)
		return (fmiterLimit == src.fmiterLimit);
    return true;
}

void IJoint::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
    long(fJointType) >>= toWhere;
    fmiterLimit >>= toWhere;
}

void IJoint::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
    long temp;
    temp <<= fromWhere;
    fJointType = EJointType(temp);
    fmiterLimit <<= fromWhere;
}

IJoint& IJoint::operator=(const IJoint& src)
{
    fJointType = src.fJointType;
	fmiterLimit = src.fmiterLimit;
    return *this;
}

IJoint::IJoint(const IJoint& src)
{
	fJointType = src.fJointType;
	fmiterLimit = src.fmiterLimit;
}


IJoint::EJointType IJoint::jointType() const
{
    return fJointType;
}

GCoordinate IJoint::miterLimit() const
{	
	return fmiterLimit;
}

void IJoint::setJointType(IJoint::EJointType type, GCoordinate miterLimit)
{	
	fJointType = type;
	fmiterLimit = miterLimit;
}

IGRect2D IJoint::adjustBounds(const IGRect2D& bounds, const IPen* pen) const
{
	IGRect2D bbox = bounds;
	if (fJointType != kBevel)
	{
		GCoordinate width;
		if (pen) {
			if (pen->penBalance() == IPen::kCenterFrame)
				width = pen->penWidth() * 0.5;
			else	// kInsetFrame
				width = 0.;		// no adjustment;
		}
		else	// hairline assumed
			width = 1.0;

		bbox.inset(IGPoint2D(-1.,-1.) * width);
	}

	return bbox;
}

/*================
||
||  class IImageTransferMode
||
||
*/
StreamableDefinitionsMacro(IImageTransferMode, gGraphicsStreamModule);

IImageTransferMode::IImageTransferMode() : fMode(IImageTransferMode::kSrcCopy)
{}

IImageTransferMode::IImageTransferMode(const IImageTransferMode& obj) : fMode(obj.fMode)
{}

IImageTransferMode::IImageTransferMode(EImageMode mode) : fMode(mode)
{}

IImageTransferMode::~IImageTransferMode()
{}

IImageTransferMode::EImageMode IImageTransferMode::mode() const
{
	return fMode;
}

IImageTransferMode& IImageTransferMode::operator=(const IImageTransferMode& obj)
{
	fMode = obj.fMode;
	return *this;
}

void IImageTransferMode::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
	long(fMode) >>= toWhere;
}

void IImageTransferMode::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
	long temp;
	temp <<= fromWhere;
	fMode = EImageMode(temp);
}

bool IImageTransferMode::operator==(const IImageTransferMode& obj) const
{
	return (fMode == obj.fMode);
}

bool IImageTransferMode::operator!=(const IImageTransferMode& other) const
{
	return !operator==(other);
}

/*================
||
||  class IColorTransferMode
||
||
*/
StreamableDefinitionsMacro(IColorTransferMode, gGraphicsStreamModule);

IColorTransferMode::IColorTransferMode() : fMode(IColorTransferMode::kSourceCopy)
{}

IColorTransferMode::IColorTransferMode(const IColorTransferMode& obj) : fMode(obj.fMode)
{}

IColorTransferMode::IColorTransferMode(EColorMode mode) : fMode(mode)
{}

IColorTransferMode::~IColorTransferMode()
{}

IColorTransferMode::EColorMode IColorTransferMode::mode() const
{
	return fMode;
}

IColorTransferMode& IColorTransferMode::operator=(const IColorTransferMode& obj)
{
	fMode = obj.fMode;
	return *this;
}

void IColorTransferMode::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
    long(fMode) >>= toWhere;
}

void IColorTransferMode::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
	long temp;
	temp <<= fromWhere;
	fMode = EColorMode(temp);
}

bool IColorTransferMode::operator==(const IColorTransferMode& obj) const
{
	return (fMode == obj.fMode);
}

bool IColorTransferMode::operator!=(const IColorTransferMode& other) const
{
	return !operator==(other);
}


/*================
||
||  class IImageSamplingControl
||
||
*/
StreamableDefinitionsMacro(IImageSamplingControl, gGraphicsStreamModule);

IImageSamplingControl::IImageSamplingControl() : fMode(IImageSamplingControl::kColorOnColor)
{}

IImageSamplingControl::IImageSamplingControl(const IImageSamplingControl& obj) : fMode(obj.fMode)
{}

IImageSamplingControl::IImageSamplingControl(ESamplingMode mode) : fMode(mode)
{}

IImageSamplingControl::~IImageSamplingControl()
{}

IImageSamplingControl& IImageSamplingControl::operator=(const IImageSamplingControl& obj)
{
	fMode = obj.fMode;
	return *this;
}

IImageSamplingControl::ESamplingMode IImageSamplingControl::mode() const
{
	return fMode;
}

void IImageSamplingControl::writeToStream(IDataStream& toWhere) const
{
	IStreamOutFrame streamFrame(toWhere);
    long(fMode) >>= toWhere;
}

void IImageSamplingControl::readFromStream(IDataStream& fromWhere)
{
	IStreamInFrame streamFrame(fromWhere);
	long temp;
	temp <<= fromWhere;
	fMode = ESamplingMode(temp);
}

bool IImageSamplingControl::operator==(const IImageSamplingControl& obj) const
{
	return (fMode == obj.fMode);
}

bool IImageSamplingControl::operator!=(const IImageSamplingControl& other) const
{
	return !operator==(other);
}
