/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 87 1.30.1.6 source/albert/graph2d/igraph2d.cpp, 2d, ioc.v400, 000302 
/*================
||
||  File:   Graphics.C
||
||  What:   Albert 2D IMGraphic Primitives code
||
||
*/

#include <igraph2d.hpp>
#include <igrport.hpp>
#include <grtype.hpp>
#include <grstate.hpp>
#include <igrafatt.hpp>
#include <idatstrm.hpp>
#include <grstrmod.hpp>
#include <iprimlck.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
#define _IGRAPH2D_CPP_
#include <ipagetun.h>
#endif

/*
 *   IMDrawable methods. This is an abstract base class for both 2D and 3D
 *   IMGraphic objects.
 */

 /*
	* IMDrawable Constructors: All constructors are protected.
	*/

IMDrawable::IMDrawable()
{
}

IMDrawable::IMDrawable (const IMDrawable& other)
{
}


IMDrawable& IMDrawable::operator= (const IMDrawable& other )
{

	return (*this);
}

// a virtual Destructor
IMDrawable::~IMDrawable()
{
}

bool IMDrawable::operator== (const IMDrawable& source) const
{
	if (GraphicTypeid(*this) != GraphicTypeid(source))
		return false;
	return true;
}
bool IMDrawable::operator!= (const IMDrawable& source) const
{
	return !operator==(source);
}


void IMDrawable::writeToStream( IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
}

void IMDrawable::readFromStream( IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
}

//==========//
// IMGraphic //
//==========//

/*===================
||
||  IMGraphic::Destructor
||
*/

IMGraphic::~IMGraphic()
{
	delete fBundle;
}


IGRect2D IMGraphic::looseFitBounds(const IGrafPort* port) const
{
	IGRect2D bounds = geometricBounds();

	const IAttributeState* bundle;
	if (port)
		bundle = port->attributes();
	else
		bundle = attributeState();

	const IPen* pen = 0 /*NIL*/;
	const IJoint* joint = 0 /*NIL*/;
	const ICap* cap = 0 /*NIL*/;

	if (bundle) {	// get the pen and joint from the bundle if any
		pen = bundle->framePen();
		cap = bundle->frameEndCap();
		joint = bundle->frameJoint();
	}

	if (!pen || (!cap && !joint)) {	// if anything missing, get them from the IRootGrafState
		IRootGrafState defaultState;
		bundle = defaultState.attributeState();
		if (!pen)
			pen = bundle->framePen();
		if (!cap)
			cap = bundle->frameEndCap();
	}

	GCoordinate penWid = 0.0;
	GCoordinate penBal = IPen::kCenterFrame;
	if (pen->penType() != IPen::kHairline) {
		penWid = pen->penWidth();
		penBal = pen->penBalance();
	} else {
		return pen->adjustBounds(bounds);
	}
	// if the pen is inset or the width == 0.0, then the joint is not affecting the bounds
	if (penBal == IPen::kInsetFrame || penWid == 0.0)
		return bounds;

	if (cap)
		bounds = cap->adjustBounds(bounds, pen);
	else
		bounds = joint->adjustBounds(bounds, pen);

	return bounds;
}


bool  IMGraphic::intersects (const IGRect2D &  rect) const
{
	bool  retval = false;

	if (isHitEnabled () == true)
	{
		IGRect2D  bounds = this->geometricBounds ();
		retval = bounds.intersects (rect);
		((IMGraphic *) this)->setHit (retval);  // for performance reasons,
												// hit state is cached.
	}

	return (retval);

} // IMGraphic::intersects


/*===================
||
||  IMGraphic::Transform Methods
||  Default implementation calls TransformObjectBy
||
*/

void IMGraphic::scaleBy(const IGPoint2D& scale, const IGPoint2D& centerOfScale)
{
	IGrafMatrix matrix;
	matrix.scaleBy (scale, centerOfScale);
	transformBy (matrix);
}

void IMGraphic::rotateBy(GDegrees degrees, const IGPoint2D& center)
{
	IGrafMatrix matrix(degrees, center);

	transformBy(matrix);
}

void IMGraphic::translateBy(const IGPoint2D& delta)
{
	IGrafMatrix matrix;
	matrix.setToTranslate (delta);
	transformBy (matrix);
}

void IMGraphic::adoptAttributeState(IAttributeState* adoptedBundle)
{
	delete fBundle;
	fBundle = adoptedBundle;

	updateTimeStamp();
}

IAttributeState* IMGraphic::orphanAttributeState()
{
	IAttributeState* bundle = fBundle;
	fBundle = 0 /*NIL*/;
	updateTimeStamp();
	return bundle;
}

const IAttributeState* IMGraphic::attributeState () const
{
	return (fBundle);
}

void IMGraphic::adoptBundle( IGrafBundle* adoptedBundle )
{
	delete fBundle;
	fBundle = adoptedBundle;

	updateTimeStamp();
}

IGrafBundle* IMGraphic::orphanBundle()
{
	updateTimeStamp();
	IAttributeState* attState  = orphanAttributeState();
	if (!attState)
		return (IGrafBundle* )attState;
	IGrafBundle bundle;
	if (GraphicTypeid(bundle) == GraphicTypeid( *attState) ) {
		return (IGrafBundle*) attState;
	} else {
		IGrafBundle* nbundle = new IGrafBundle( *attState );
		return nbundle;
	}
}

const IGrafBundle* IMGraphic::bundle () const
{
	const IAttributeState* attState = attributeState();
	IMGraphic* nonConstThis = (IMGraphic *) this;
	if (!attState)
		return (const IGrafBundle*)attState;

	IGrafBundle bundle;
	if (GraphicTypeid(bundle) == GraphicTypeid( *attState) ) {
		return (const IGrafBundle*) attState;
	} else {
		IGrafBundle* nbundle = new IGrafBundle( *attState );
		nonConstThis->adoptBundle( nbundle );
		return nbundle;
	}
}

void IMGraphic::writeToStream( IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
	IMDrawable::writeToStream(toWhere);
	::writeObject(fBundle, toWhere);
}

void IMGraphic::readFromStream( IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
	IMDrawable::readFromStream(fromWhere);
	delete fBundle;
	::readObject(fBundle, fromWhere);
}


bool IMGraphic::operator==(const IMDrawable& src) const
{
	bool  retval = true;

	// subclass will rely on type check here, 'cuz they call the
	// base class comparison operator
	if (GraphicTypeid(*this) != GraphicTypeid(src))
	{
		retval = false;
	}
	else
	{
		IMGraphic *  srcGraphic = (IMGraphic *) &src;
		if (fBundle && srcGraphic->fBundle)
		{
			retval = (*(srcGraphic->fBundle) == *fBundle) &&
				 (srcGraphic->fHitFlag == fHitFlag) &&
				 (srcGraphic->fHitEnableFlag == fHitEnableFlag);
		}
		else if (fBundle || srcGraphic->fBundle)
		{
			retval = false;
		}

	}

	return retval;

}


bool IMGraphic::isHitEnabled () const
{
	return fHitEnableFlag;
}


void IMGraphic::setHitEnabled (bool  flag)
{
	fHitEnableFlag = flag;
}


void IMGraphic::setHit (bool  flag)
{
	fHitFlag = flag;
}


/*===================
||
||  IMGraphic::Constructors
||  Protected for abstract class
||
*/


IMGraphic::IMGraphic() :
	fHitFlag (false),
	fHitEnableFlag (true)
{
	fBundle = 0 /*NIL*/;
	updateTimeStamp();
}

IMGraphic::IMGraphic(IGrafBundle* adoptBundle) :
	fHitFlag (false),
	fHitEnableFlag (true)
{
	fBundle = adoptBundle;
	updateTimeStamp();
}

IMGraphic::IMGraphic (const IMGraphic& other) :
	IMDrawable (other)
{
	if (other.fBundle)
		fBundle = GraphicCopy (*other.fBundle);
	else
		fBundle = 0 /*NIL*/;
	fHitFlag       = other.fHitFlag;
	fHitEnableFlag = other.fHitEnableFlag;
	fSeed = other.fSeed;
}

/*===================
||
||  IMGraphic::operator=
||  Protected for abstract class
||
*/

IMGraphic& IMGraphic::operator= (const IMGraphic& other )
{
	if (&other != this ) {
		if (fBundle != other.fBundle) {
			delete fBundle;
			if (other.fBundle)
				fBundle = GraphicCopy (*other.fBundle);
			else
				fBundle = 0 /*NIL*/;
		}
		fHitFlag       = other.fHitFlag;
		fHitEnableFlag = other.fHitEnableFlag;
		fSeed = other.fSeed;
	}
	return (*this);
}

struct ISeed {
	ISeed();
	unsigned long fNewGraphicSeed;
};
ISeed::ISeed()
  : fNewGraphicSeed(0)
{
}
static ISeed* gStaticSeed = 0;
static ISeed *staticSeed()
{
	if (!gStaticSeed) {
		IPrimalLock lock;
		if (!gStaticSeed) {
			gStaticSeed = new ISeed;
			++(gStaticSeed->fNewGraphicSeed);
		}
	}
	else {
		IPrimalLock lock;
		++(gStaticSeed->fNewGraphicSeed);
	}
	return gStaticSeed;
}

const unsigned long& IMGraphic::seed() { return staticSeed()->fNewGraphicSeed; }

/*===================
||
||  IMGraphic::TimeStamp Stuff
||
*/

unsigned long IMGraphic::timeStamp() const
{
	return fSeed;
}
void IMGraphic::setTimeStamp(const unsigned long& newSeed)
{
	fSeed = newSeed;
}

void IMGraphic::updateTimeStamp()
{
	fSeed = IMGraphic::seed();	
}

void IMGraphic::resetTimeStamp()
{
	fSeed = 0;
}

/*================================
||  IGraphicHierarchy::
||
*/

// ----- Constructors and Destructors -----

IGraphicHierarchy::IGraphicHierarchy() : IMGraphic()
{}

IGraphicHierarchy::IGraphicHierarchy(IGrafBundle* bundle) : IMGraphic(bundle)
{}

IGraphicHierarchy::IGraphicHierarchy(const IGraphicHierarchy& graphicHierarchy) :
	IMGraphic(graphicHierarchy)
{}

IGraphicHierarchy::~IGraphicHierarchy ()
{}


void IGraphicHierarchy::writeToStream( IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
	IMGraphic::writeToStream(toWhere);
}

void IGraphicHierarchy::readFromStream( IDataStream& fromWhere )
{
	IStreamInFrame streamFrame(fromWhere);
	IMGraphic::readFromStream(fromWhere);
}

IGraphicHierarchy& IGraphicHierarchy::operator= (const IGraphicHierarchy& other )
{
	if ( &other != this )
	{
		IMGraphic::operator= (other);
	}
	return (*this);
}

bool IGraphicHierarchy::operator==(const IMDrawable& src) const
{
	return IMGraphic::operator==(src);
}

/*
 * Support for GraphicGroups which are implemented to provide one method
 * for creating hierarchies.
 */
IGraphicIterator::IGraphicIterator()
{}

IGraphicIterator::~IGraphicIterator()
{}

IGraphicGroupIterator::IGraphicGroupIterator(const IGraphicGroup* group) :
	fIterator(group->fChildren)
{
	fIterator.setToFirst();
}

IGraphicGroupIterator::~IGraphicGroupIterator ()
{}

void IGraphicGroupIterator::reset ()
{
	fIterator.setToFirst ();
}

IMGraphic* IGraphicGroupIterator::first()
{
	fIterator.setToFirst();
	return ( fIterator.isValid() ? fIterator.element():0 /*NIL*/ );
}

IMGraphic* IGraphicGroupIterator::last ()
{
	fIterator.setToLast();
	return ( fIterator.isValid() ? fIterator.element():0 /*NIL*/ );
}

IMGraphic*   IGraphicGroupIterator::next ()
{
	fIterator.setToNext();
	return ( fIterator.isValid() ? fIterator.element() : 0 /*NIL*/ );
}

IMGraphic*   IGraphicGroupIterator::previous ()
{
	fIterator.setToPrevious();
	return ( fIterator.isValid () ? fIterator.element() : 0 /*NIL*/ );
}

StreamableDefinitionsMacro(IGraphicGroup, gGraphicsStreamModule);

IGraphicGroup::IGraphicGroup()
{
	fBounds = IGRect2D::zeroRect();
}

IGraphicGroup::IGraphicGroup(const IGraphicGroup& other) : IGraphicHierarchy(other)
{
	fMatrix = other.fMatrix;
	fBounds = other.fBounds;

	IGraphicIterator *iter = other.createIterator();

	IMGraphic* member = iter->first();
	while ( member != 0 /*NIL*/ ) {
		IMGraphic* memberCopy = GraphicCopy (*member);
		adoptLast( memberCopy );
		member = iter->next();
	}
	updateTimeStamp();
	delete iter;
}

IGraphicGroup::IGraphicGroup( IGrafBundle* bundle) :
	IGraphicHierarchy( bundle )
{
}

IGraphicGroup::~IGraphicGroup()
{
	deleteAll();
}

void IGraphicGroup::deleteAll()
{
	GraphicDeque::Cursor iter( fChildren );
	for( iter.setToFirst();  iter.isValid(); iter.setToNext() )
		delete iter.element();
	fChildren.removeAll();
}
		
// === Parental Functions ===
void IGraphicGroup::adoptLast(IMGraphic* obj)
{
	fChildren.addAsLast( obj );
	updateTimeStamp();
}

void IGraphicGroup::adoptFirst(IMGraphic* obj)
{
	fChildren.addAsFirst( obj );
	updateTimeStamp();
}

IMGraphic* IGraphicGroup::orphan(const IMGraphic& obj)
{
	updateTimeStamp();

	IMGraphic* objToBeRemoved = 0 /*NIL*/;

	IMGraphic* lastElement = fChildren.lastElement();
	if (lastElement != 0 /*NIL*/)
		{
			IMGraphic* firstElement;
			do {
				firstElement = fChildren.firstElement ();
				fChildren.removeFirst();
				if( *firstElement == obj ) {
					objToBeRemoved = firstElement;
				}
				else {
					fChildren.addAsLast( firstElement );
				}
			} while ( firstElement != lastElement );
		}

	return objToBeRemoved;
}

IMGraphic* IGraphicGroup::orphanLast()
{
	updateTimeStamp();
	IMGraphic* lastElement;

	if ( ( lastElement = fChildren.lastElement() ) != 0 /*NIL*/ )
		fChildren.removeLast();

	return lastElement;
}

IMGraphic* IGraphicGroup::orphanFirst()
{
	updateTimeStamp();

	IMGraphic* firstElement;
		
	if( ( firstElement = fChildren.firstElement() ) != 0 /*NIL*/ )
		fChildren.removeFirst();

	return firstElement;
}

long    IGraphicGroup::count() const
{
	return fChildren.numberOfElements();
}


IGraphicIterator* IGraphicGroup::createIterator() const
{
	IGraphicGroupIterator* returnVal = new IGraphicGroupIterator (this);
	return (returnVal);
}

// === IMGraphic ===
void IGraphicGroup::draw (IGrafPort& port) const
{
	IGraphicGroupIterator iterator (this);
	const IAttributeState* state = attributeState();
	if (!state) {
		if (fMatrix.isIdentity()) {
			for (IMGraphic* child = iterator.first(); child != 0 /*NIL*/; child = iterator.next())
				child->draw(port);
		}
		else {
			ILinkedGrafPort newport(&port, kModelMatrix, &fMatrix);
			for (IMGraphic* child = iterator.first(); child != 0 /*NIL*/; child = iterator.next())
				child->draw(newport);
		}
	}
	else {
		ILinkedGrafPort bundlePort(&port, state);
		if (fMatrix.isIdentity()) {
			for (IMGraphic* child = iterator.first(); child != 0 /*NIL*/; child = iterator.next())
				child->draw(bundlePort);
		}
		else {
			ILinkedGrafPort newport(&bundlePort, kModelMatrix, &fMatrix);
			for (IMGraphic* child = iterator.first(); child != 0 /*NIL*/; child = iterator.next())
				child->draw(newport);
		}
	}
}

const IGrafMatrix* IGraphicGroup::matrix() const
{
		return &fMatrix;
}

IGRect2D IGraphicGroup::geometricBounds() const
{
	IGraphicGroupIterator iterator (this);
	bool isFirst = true;
	IGRect2D tmpRect;

	for ( IMGraphic* child = iterator.first ();
	child != 0 /*NIL*/;
	child = iterator.next ())
		{
			if (isFirst) {
	
	tmpRect = child->geometricBounds();
	isFirst = false;
			}
			else
	{
		tmpRect.extendTo (child->geometricBounds());
	}

		}
	if (isFirst)
		return (IGRect2D::zeroRect());
	else {
		const IGrafMatrix* grafmatrix = matrix();
		if (grafmatrix->isIdentity())
			return tmpRect;
		tmpRect = grafmatrix->transformBounds( tmpRect );
		return tmpRect;
	}
}

IGRect2D IGraphicGroup::looseFitBounds(const IGrafPort* port) const
{
	IGraphicGroupIterator iterator (this);
	bool isFirst = true;
	IGRect2D tmpRect;

	for ( IMGraphic* child = iterator.first ();
	child != 0 /*NIL*/;
	child = iterator.next ())
		{
			if (isFirst) {
	
	tmpRect = child->looseFitBounds(port);
	isFirst = false;
			}
			else
	{
		tmpRect.extendTo (child->looseFitBounds(port));
	}

		}
	if (isFirst)
		return (IGRect2D::zeroRect());
	else {
		const IGrafMatrix* grafmatrix = matrix();
		if (grafmatrix->isIdentity())
			return tmpRect;
		tmpRect = grafmatrix->transformBounds( tmpRect );
		return tmpRect;
	}
}


void IGraphicGroup::transformBy (const IGrafMatrix& matrix)
{
		fBounds  = matrix.transformBounds( fBounds );
		fMatrix.concatWith( matrix );
		updateTimeStamp();
}


bool  IGraphicGroup::intersects (const IGRect2D &  rect) const
{
	bool  retval = false;

	if (this->isHitEnabled () == false)
	{
		return retval;
	}

	IGraphicGroupIterator  iterator (this);

	for (IMGraphic* child = iterator.first ();
			 child != 0 /*NIL*/;
			 child = iterator.next ())
	{
		//
		// If any of the children are hit, then the group is hit.
		//

		if (child->isHitEnabled () == true)
		{
			retval = retval || child->intersects (rect);
		}
	
	} // for

	((IGraphicGroup *) this)->setHit (retval);  // for performance reasons,
					                                    // hit state is cached.

	return retval;

} // IGraphicGroup::intersects


// === Base Functionality ===

IGraphicGroup&  IGraphicGroup::operator= ( const IGraphicGroup& src )
{
	if (&src != this ) {
		IGraphicHierarchy::operator=(src);

		fMatrix = src.fMatrix;
		fBounds = src.fBounds;
		fBoundsTimeStamp = src.fBoundsTimeStamp;

		deleteAll();

		IGraphicIterator *iter = src.createIterator();

		IMGraphic* member = iter->first();
		while ( member != 0 /*NIL*/ ) {
			IMGraphic* memberCopy = GraphicCopy (*member);
			adoptLast( memberCopy );
			member = iter->next();
		}
		delete iter;
	}
	return (*this);
}

bool IGraphicGroup::operator==(const IMDrawable& srcGraphic) const
{
	if (!IMGraphic::operator==(srcGraphic))
		return false;
	IGraphicGroup* src = (IGraphicGroup*) &srcGraphic;
	if (fMatrix != src->fMatrix)
		return false;
	IGraphicIterator* srcIter = src->createIterator();
	IGraphicIterator* myIter =  createIterator();

	IMGraphic* srcMember = srcIter->first();
	IMGraphic* myMember  = myIter->first();
	while (srcMember != 0 /*NIL*/ && myMember != 0 /*NIL*/) {
		if (*srcMember != *myMember){
			delete srcIter;
			delete myIter;
			return false;
		}
		srcMember = srcIter->next();
		myMember = myIter->next();
	}
	delete myIter;
	delete srcIter;
	//
	// Making sure that both the members are indeed 0 /*NIL*/
	//
	if (srcMember == myMember)
		return true;
	return false;
}


void IGraphicGroup::writeToStream( IDataStream& toWhere ) const
{
	IStreamOutFrame streamFrame(toWhere);
	IMGraphic::writeToStream(toWhere);
	fMatrix >>= toWhere;

	fChildren.numberOfElements () >>= toWhere;
	IGraphicIterator*  myIter =  createIterator ();
	IMGraphic*         myMember  = myIter->first ();
	while (myMember != 0 /*NIL*/)
	{
		::writeObject (myMember, toWhere);
		myMember = myIter->next ();
	}
	delete myIter;

	fBounds >>= toWhere;
	fBoundsTimeStamp >>= toWhere;

} // IGraphicGroup::writeToStream


void IGraphicGroup::readFromStream( IDataStream& fromWhere )
{
	IStreamInFrame streamFrame (fromWhere);
	IMGraphic::readFromStream (fromWhere);
	fMatrix <<= fromWhere;

	deleteAll ();

	long noOfElements;
	noOfElements <<= fromWhere;
	for (long i = 0;  i < noOfElements; i++)
	{
		IMGraphic* member;
		::readObject (member, fromWhere);
		adoptLast (member);
	}

	fBounds <<= fromWhere;
	fBoundsTimeStamp <<= fromWhere;

} // IGraphicGroup::readFromStream
