/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 61 1.34.1.7 source/albert/graph2d/igtxtrun.cpp, 2d, ioc.v400, 000302 
/*******************************************************************************
* FILE NAME: igtxtrun.cpp                                                      *
*******************************************************************************/

#include <ilanglvl.hpp>

#if defined(IC_WIN)
#include <gdidev.hpp>
#elif defined(IC_PM)
#include <gpidev.hpp>
#include <winemul.h>
#elif defined(IC_MOTIF)
#include <xdevice.hpp>
#endif

#include <grassert.hpp>
#include <igtxtrun.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
#define _IGTXTRUN_CPP_
#include <ipagetun.h>
#endif

///////////////////////////////////////////////////////
//
// IGTextRun
//  Graphically displays a single run of styled text.
//  Where all of the characters in the range of text
//  to be drawn contain the same styles.
//
///////////////////////////////////////////////////////

/*------- Constructors -------*/

#ifndef IC_MOTIF
const char* IGTextRun::gDefaultFontName = 0 /*NIL*/;
#else
const char* IGTextRun::gDefaultFontName = "courier" ;
#endif


IGTextRun::IGTextRun(
	const IText*  textToAlias,
	offset_type runBegin,
	offset_type runEnd,
    const IGPoint2D& anchor,
    EAttachment attachment) :

    fText(0 /*NIL*/),
		fTextStorage(0), //hash pointer
    fStyles(0 /*NIL*/),
    fBegin(runBegin),
    fEnd(runEnd),
	fAnchor(anchor),
	fAttachment(kDefault),
	fEscapement(0),
	fHadMetrics(false),
	fExternalStyleSet(false),	// used to see if anyone had called setStyles()...
	fStylesCached(false),
	fSubTextCached(false),
	fTypeFace(gDefaultFontName),
	fPointSize(0),
	fWeight(0),
	fItalic(0),
	fUnderline(0),
	fStrikeOut(0),
	fTextColor(0),
	fOutline(0),
	fBitmap(0),
	fFixedPitch(0)
{
	setText(textToAlias);
	setAttachment(attachment);
	long end = resetStyles() + fBegin;
	if (fEnd > end)	// force the end to be the length of the style run...
		fEnd = end;
}

// for convenience, this will take the entire IText as a single run...
IGTextRun::IGTextRun(
	const IText*  textToAlias,
    const IGPoint2D& anchor,
    EAttachment attachment) :

    fText(0 /*NIL*/),
		fTextStorage(0),
    fStyles(0 /*NIL*/),
    fBegin(0),
    fEnd(IText::npos),
	fAnchor(anchor),
	fAttachment(kDefault),
	fEscapement(0),
	fHadMetrics(false),
	fExternalStyleSet(false),	// used to see if anyone had called setStyles()...
	fStylesCached(false),
	fSubTextCached(false),
	fTypeFace(gDefaultFontName),
	fPointSize(0),
	fWeight(0),
	fItalic(0),
	fUnderline(0),
	fStrikeOut(0),
	fTextColor(0),
	fOutline(0),
	fBitmap(0),
	fFixedPitch(0)
{
	setText(textToAlias);
	setAttachment(attachment);
  fEnd = resetStyles() + fBegin; // force the end to be the length of the style run...
}

IGTextRun::IGTextRun(const IGTextRun& that) :
    fText(that.fText),
		fTextStorage(that.fTextStorage),
    fStyles(that.fStyles),
    fBegin(that.fBegin),
    fEnd(that.fEnd),
	fAnchor(that.fAnchor),
	fAttachment(that.fAttachment),
	fEscapement(that.fEscapement),
	fHadMetrics(that.fHadMetrics),
	fExternalStyleSet(that.fExternalStyleSet),	// used to see if anyone had called setStyles()...
	fStylesCached(that.fStylesCached),
	fSubTextCached(that.fSubTextCached),
	fTypeFace(gDefaultFontName),
	fPointSize(0),
	fWeight(0),
	fItalic(0),
	fUnderline(0),
	fStrikeOut(0),
	fTextColor(0),
	fOutline(0),
	fBitmap(0),
	fFixedPitch(0)
{
	copyCachedStuff(that);
}


IGTextRun& IGTextRun::operator=(const IGTextRun& that)
{
	if (this != &that) {
		fText = that.fText;
		fTextStorage = that.fTextStorage;
		fStyles = that.fStyles;
		fBegin = that.fBegin;
		fEnd = that.fEnd;
		fAnchor = that.fAnchor;
		fAttachment = that.fAttachment;
		fEscapement = that.fEscapement;
		fHadMetrics = that.fHadMetrics;
		fExternalStyleSet = that.fExternalStyleSet;	// used to see if anyone had called setStyles()...
		fStylesCached = that.fStylesCached;
		fSubTextCached = that.fSubTextCached;

		copyCachedStuff(that);
	}
	
	return *this;
}

IGTextRun::~IGTextRun()
{}

IGTextRun::IGTextRun() :
    fText(0 /*NIL*/),
		fTextStorage(0),
    fStyles(0 /*NIL*/),
    fBegin(0),
    fEnd(0),
	fAnchor(IGPoint2D::origin()),
	fAttachment(kDefault),
	fEscapement(0),
	fHadMetrics(false),
	fExternalStyleSet(false),	// used to see if anyone had called setStyles()...
	fStylesCached(false),
	fSubTextCached(false),
	fTypeFace(gDefaultFontName),
	fPointSize(0),
	fWeight(0),
	fItalic(0),
	fUnderline(0),
	fStrikeOut(0),
	fTextColor(0),
	fOutline(0),
	fBitmap(0),
	fFixedPitch(0)
{}

/*---------- internal stuff ------------*/

void IGTextRun::copyCachedStuff(const IGTextRun& that)
{
	if (fStylesCached) { // copy the cached stuff...
		fTypeFace = that.fTypeFace;
		fEscapement = that.fEscapement;
		fWeight = that.fWeight;
		fItalic = that.fItalic;
		fUnderline = that.fUnderline;
		fStrikeOut = that.fStrikeOut;
		fPointSize =that.fPointSize;
		fOutline = that.fOutline;
		fBitmap = that.fBitmap;
		fFixedPitch = that.fFixedPitch;

	}

	if (fSubTextCached) {
		fUnicodeString = that.fUnicodeString;
		fUnicodeStringLength = that.fUnicodeStringLength;
		fSubText = that.fSubText;
	}

	if (that.fTextColor) {
		if (!fTextColor)
			fTextColor = &fColor;
		fColor = that.fColor;
	}

	if (fHadMetrics) {
		fBounds = that.fBounds;
		fNextAnchor = that.fNextAnchor;
		fActualDevice = that.fActualDevice;
		fStamp1 = that.fStamp1;
		fStamp2 = that.fStamp2;
	}
}

/*---------- Setters/Getters ------------*/

void IGTextRun::setText(const IText* textToAlias)
{
	IText::offset_type	 chunkPos;
	IText::length_type	 chunkLength;
	const UniChar* store = textToAlias ? textToAlias->storage_chunk(0, chunkPos, chunkLength) : 0 /* NIL */;

	if (fText == textToAlias && fTextStorage == store)
		return;
	fHadMetrics = 0;
	fText = textToAlias;
	fTextStorage = store;
	if (!fExternalStyleSet) {
	fStyles = 0 /*NIL*/;
	fStylesCached = false;
	}
	fSubTextCached = false;
}

const ITextStyleSet* IGTextRun::styles() const {
	return fStyles;
}

// Character renderer uses this, instead of setting styles in IText like other clients.
// The character renderer knows when its underlying styles have changed;  thus, it can
// safely set styles in an IGTextRun once, then supress style recaching (which is expensive)
// if the text (but not the styles) change.
// Since this function is private, IStandardCharacterRenderer is a friend of IGTextRun.
// IStandardCharacterRenderer does not access any other protected / private members of IGTextRun.

void IGTextRun::setStyles(const ITextStyleSet& styleSetToAlias) {
	fStyles = &styleSetToAlias;
	fExternalStyleSet = true;
	cacheRenderStyles();
}

// returns the length from fBegin to the end of this run...
size_t IGTextRun::resetStyles() {
	length_type count = 0;
	offset_type starting = 0;
	if (fText && !fText->empty())
		fStyles = fText->stylesAt(fBegin, starting, count);
	return count + starting - fBegin;
}

void IGTextRun::setBegin(offset_type newBegin) {
	setBeginAndEnd(newBegin, fEnd);
}

void IGTextRun::setEnd(offset_type newEnd) {
	setBeginAndEnd(fBegin, newEnd);
}

void IGTextRun::setBeginAndEnd(offset_type newBegin, offset_type newEnd) {
	if (fBegin == newBegin && fEnd == newEnd)
		return;

	fHadMetrics = 0;
	fBegin = newBegin;
	fEnd = newEnd;
	if (!fExternalStyleSet) {
	fStyles = 0 /*NIL*/;
	fStylesCached = false;
	}
	fSubTextCached = false;
}

// find the run at newBegin, and set the end automatically
void IGTextRun::setRunBeginsAt(offset_type newBegin) {
	if (fExternalStyleSet)	// the only way to turn off external styles set... yak!
		fExternalStyleSet = false;

	bool beginWasSame = (fBegin == newBegin);
	bool stylesCached = fStylesCached;
	fBegin = newBegin;
	length_type length = resetStyles();
	fStylesCached = false;
	offset_type end = fBegin + length;
	if (fEnd == end && beginWasSame) {	// if the begin/end are not changed
		fStylesCached = stylesCached;	// restore the styles cached
		return;
	}

	fHadMetrics = 0;
	fEnd = end;
	fSubTextCached = false;
}

void IGTextRun::setAnchor(const IGPoint2D& location) {
	if (fAnchor == location)
		return;

	fHadMetrics = 0;
	fAnchor = location;
}

void IGTextRun::setAttachment(EAttachment attachment)
{
	if (fAttachment == attachment)
		return;

	fHadMetrics = 0;

	// one extra style from the textRun, this should be one of the styles style though ...
	fAttachment = attachment;
	switch (fAttachment)
	{
	default:
	case IGTextRun::kLeftToRight:
		fEscapement = 0;
		break;
	case IGTextRun::kTopToBottom:
		fEscapement = 2700;
		break;
	case IGTextRun::kRightToLeft:
		fEscapement = 1800;
		break;
	case IGTextRun::kBottomToTop:
		fEscapement = 900;
		break;
	}
}

/*-------- Comparison --------*/

bool IGTextRun::operator==(const IGTextRun& that) const
{
	if (fText && that.fText)
	{
		if (fText != that.fText &&
		    fTextStorage != that.fTextStorage  &&
			  fText->compare(fBegin, fEnd - fBegin, *(that.fText), that.fBegin, that.fEnd - that.fBegin))
			return false;
		else
			return (
				// fAnchor == that.fAnchor &&
				fAttachment == that.fAttachment);
	}
	else
		return !(fText || that.fText);
}

bool IGTextRun::operator!=(const IGTextRun& that) const
{
	return !(*this == that);
}

/*-------- Others --------*/

void IGTextRun::cacheRenderStyles()
{
	if (!fStyles)	// if the styleSet is not set for the run...
		resetStyles();

	// default values:
	fTypeFace = gDefaultFontName;
	fPointSize = 0;
	fWeight = 0;
	fItalic = 0;
	fUnderline = 0;
	fStrikeOut = 0;
	fTextColor = 0 /*NIL*/;
	fOutline = 0;
	fBitmap = 0;
	fFixedPitch = 0;

	if (fStyles && !fStyles->empty())
	{

		ITextTypefaceStyle typefaceStyle;
		if (fStyles->contains(typefaceStyle))
		{
			fStyles->extract(typefaceStyle);
			fTypeFace = typefaceStyle.typeface();
		}
		
		ITextPointSizeStyle pointSizeStyle;
		if (fStyles->contains(pointSizeStyle))
		{
			fStyles->extract(pointSizeStyle);		
			fPointSize = pointSizeStyle.pointSize();
		}
		
		ITextBoldfaceStyle boldfaceStyle;
		if (fStyles->contains(boldfaceStyle))
		{
			fStyles->extract(boldfaceStyle);
			if (boldfaceStyle.isBoldface())
#ifndef IC_MOTIF
				fWeight = FW_BOLD;
#else
				fWeight = 1;
#endif
		}
		
		ITextItalicStyle italicStyle;
		if (fStyles->contains(italicStyle))
		{
			fStyles->extract(italicStyle);
			if (italicStyle.isItalic())
				fItalic = 1;
		}
		
		ITextUnderlineStyle underlineStyle;
		if (fStyles->contains(underlineStyle))
		{
			fStyles->extract(underlineStyle);
			if (underlineStyle.isUnderlined())
				fUnderline = 1;
		}
		
		ITextStrikethroughStyle strikethroughStyle;
		if (fStyles->contains(strikethroughStyle))
		{
			fStyles->extract(strikethroughStyle);
			if (strikethroughStyle.isStrikethrough())
				fStrikeOut = 1;
		}
		
		ITextColorStyle colorStyle;
		if (fStyles->contains(colorStyle))
		{
			fStyles->extract(colorStyle);
			fTextColor = &fColor;
			fColor = IBaseColor(
					CharIntensity(colorStyle.red()),
					CharIntensity(colorStyle.green()),
					CharIntensity(colorStyle.blue()));
		}
		ITextFixedPitchStyle fixedPitchStyle;
		if (fStyles->contains(fixedPitchStyle))
		{
			fStyles->extract(fixedPitchStyle);
			if (fixedPitchStyle.isFixedPitch())
				fFixedPitch = 1;
		}
	}
		
	  ITextOutlineStyle outlineStyle;
		if (fStyles->contains(outlineStyle))
		{
			fStyles->extract(outlineStyle);
			if (outlineStyle.isOutlined())
					fOutline = 1;
		}
		
	  ITextBitmapStyle bitampStyle;
		if (fStyles->contains(bitampStyle))
		{
			fStyles->extract(bitampStyle);
			if (bitampStyle.isBitmap())
					fBitmap = 1;
		}
			
		fStylesCached = true;

}

void IGTextRun::cacheRenderSubtext(bool isUnicodeSupported)
{
// {jbr} Changed this 6/16/97
	if (isUnicodeSupported) {
		IText::offset_type chunkPos, chunkLength;
		fUnicodeString = fText->storage_chunk(fBegin, chunkPos, chunkLength);

		if (chunkPos + chunkLength >= fEnd) {
			fUnicodeString += (fBegin - chunkPos);
			fUnicodeStringLength = fEnd-fBegin;
		}
		else {
			IText::length_type subtextLength = fEnd-fBegin;
			fSubText = fText->substr(fBegin, subtextLength);
			fUnicodeString = fSubText.data();
			fUnicodeStringLength = subtextLength;
		}
	}
	else {
		fSubText = fText->substr(fBegin, fEnd-fBegin);
	}
	fSubTextCached = true;
}

bool IGTextRun::hadMetrics(
	IPresSpaceHandle perActualDevice,
	unsigned long stamp1,
	unsigned long stamp2) const
{
	if (perActualDevice != 0 /*NIL*/)	// if it has a device...
		return
			bool(
				fHadMetrics &&
				(fActualDevice == perActualDevice) &&
				(fStamp1 == stamp1) &&
				(fStamp2 == stamp2));
	else	// otherwise any metrics is as good...
		return fHadMetrics;
}

void IGTextRun::updateMetrics(
	IGRect2D& box,
	IGPoint2D& posn,
	IPresSpaceHandle perActualDevice,
	unsigned long stamp1,
	unsigned long stamp2)
{
	if (perActualDevice != 0 /*NIL*/)
	{
		if (fHadMetrics &&
			(fActualDevice == perActualDevice) &&
			(fStamp1 == stamp1) &&
			(fStamp2 == stamp2))
			return;		// already had the good bounding box...
	}
	else
	{
		if (fHadMetrics)
			return;		// already had a almost good bounding box...
	}
	fBounds = box;
	fNextAnchor = posn;
#ifndef IC_MOTIF
	fActualDevice = perActualDevice;
#endif
	fHadMetrics = true;
	fStamp1 = stamp1;
	fStamp2 = stamp2;
}

IGRect2D IGTextRun::bounds(const IGrafPort* port) const
{
	const IGrafState* state = 0 /*NIL*/;
  IGrafDevice* dev = 0 /*NIL*/;
	
	if (!port) {
#ifndef IC_MOTIF
		if (fHadMetrics)
#else
   		IXDevice::determineTextMetrics(*((IGTextRun*)this),
									0, state);
#endif
			return fBounds;
	}
	else {
		IGrafPort* nonConstPort = (IGrafPort*)port;

		state = nonConstPort->state();
		dev = nonConstPort->device();
		if (dev)
		{
#ifdef IC_MOTIF
   		IXDevice::determineTextMetrics(*((IGTextRun*)this),
									(IXDevice*)dev, state);
#endif
		}
	}

#if defined(IC_WIN)
  // the following only works on Windows (may be os/2) only
  IGdiDevice::determineTextMetrics(*((IGTextRun*)this),(IGdiDevice*)dev, state );
#elif defined(IC_PM)
  IGpiDevice::determineTextMetrics(*((IGTextRun*)this), (IGpiDevice*)dev, state );
#endif

	return fBounds;
}

IGPoint2D IGTextRun::nextAnchor(const IGrafPort* port) const
{
	const IGrafState* state = 0 /*NIL*/;
  IGrafDevice* dev;
	if (!port) {
#ifndef IC_MOTIF
		if (fHadMetrics)
#else
   		IXDevice::determineTextMetrics(*((IGTextRun*)this),
                                    0, state);
#endif
			return fNextAnchor;
	}
	else {
		IGrafPort* nonConstPort = (IGrafPort*)port;

		state = nonConstPort->state();
		dev = nonConstPort->device();
		if (dev)
		{
#ifdef IC_MOTIF
   		IXDevice::determineTextMetrics(*((IGTextRun*)this),
                                    (IXDevice*)dev, state);
#endif
		}

	}


#if defined(IC_WIN)
  // the following only works on Windows (may be os/2) only
  IGdiDevice::determineTextMetrics(*((IGTextRun*)this), (IGdiDevice*)dev, state);
#elif defined(IC_PM)
  IGpiDevice::determineTextMetrics(*((IGTextRun*)this), (IGpiDevice*)dev, state);
#endif

	return fNextAnchor;
}
