// NOSHIP ...
/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 62 1.21.1.5 source/albert/graph2d/igtxtrun.hpp, 2d, ioc.v400, 000302 
#ifndef _IGTXTRUN_
#define _IGTXTRUN_

/*******************************************************************************
* INTERNAL FILE: igtxtrun.hpp                                                  *
*******************************************************************************/

#include <igbase2d.hpp>
#include <ibcolor.hpp>

#include <itext.hpp>
#include <istylset.hpp>
#include <i2dghand.hpp>

#if __IBMCPP__ >= 400
#pragma namemangling(compat)
#endif

class IGrafPort;

#pragma enum(4)
#pragma pack(push,4)

///////////////////////////////////////////////////////
//
// IGTextRun
//  Graphically displays a single run of styled text.
//  All of the characters in the range of text
//  to be drawn must contain the same styles.
//
///////////////////////////////////////////////////////
class IGTextRun
{
  public:
	enum EAttachment {
		kLeftToRight,
		kRightToLeft,
		kTopToBottom,
		kBottomToTop,
		kDefault = kLeftToRight
	};

	typedef size_t offset_type;
	typedef size_t length_type;
	
	IGTextRun(
		const IText* textToAlias,
		offset_type runBegin,
		offset_type runEnd,
		const IGPoint2D& anchor = IGPoint2D::origin(),
		EAttachment attachment = kDefault);

	// for convenience, make the entire IText as a single run, using styleset of the 1st run
	IGTextRun(
		const IText* textToAlias,
		const IGPoint2D& anchor = IGPoint2D::origin(),
		EAttachment attachment = kDefault);

	IGTextRun(const IGTextRun&);
	IGTextRun& operator=(const IGTextRun&);

	~IGTextRun();		// not virtual! don't inherit from this class.
	IGTextRun();

	const IText* text() const;
	void setText(const IText* textToAlias);

	offset_type begin() const;
	void setBegin(offset_type);

	offset_type end() const;
	void setEnd(offset_type);

	void setBeginAndEnd(offset_type newBegin, offset_type newEnd);

	void setRunBeginsAt(offset_type newBegin);	// find the run at newBegin, and set the end automatically

	IGPoint2D anchor() const;
	void setAnchor(const IGPoint2D& location);

	EAttachment attachment() const;
	void setAttachment(EAttachment attachment);

	IGRect2D bounds(const IGrafPort* = 0 /*NIL*/) const;
	IGPoint2D nextAnchor(const IGrafPort* = 0 /*NIL*/) const;
	
	bool operator==(const IGTextRun&) const;
	bool operator!=(const IGTextRun&) const;

  private:
	// this function should be obsolete, caller should get styles from the IText object
	const ITextStyleSet* styles() const;

friend class ICustomButtonDrawHandler; // ICustomButtonDrawHandler::drawForeground
friend class IStandardCharacterRenderer; // calls setStyles;  see igtxtrun.cpp for more info...
	void setStyles(const ITextStyleSet& styleSetToAlias);

	static const char* gDefaultFontName;

  public:
	// not for public consumption...  for font contexts and graf devices...
	void cacheRenderStyles();
	void cacheRenderSubtext(bool isUnicodeSupported);

	bool hadMetrics(
		IPresSpaceHandle perActualDevice = 0 /*NIL*/,
		unsigned long timeStamp1 = 0,
		unsigned long timeStamp2 = 0) const;

	void updateMetrics(
		IGRect2D& box,
		IGPoint2D& posn,
		IPresSpaceHandle perActualDevice = 0 /*NIL*/,
		unsigned long timeStamp1 = 0,
		unsigned long timeStamp2 = 0);

  private:
	size_t resetStyles();
	void copyCachedStuff(const IGTextRun& that);

	// private date members:
	const IText* fText;				// aliased text, not owned
	const UniChar* fTextStorage;	// text storage point for caching purposes	
	const ITextStyleSet* fStyles;	// aliased style set, not owned
	offset_type fBegin;
	offset_type fEnd;
	IGPoint2D fAnchor;
	IGPoint2D fNextAnchor;
	EAttachment fAttachment;
	IGRect2D fBounds;
	IPresSpaceHandle fActualDevice;
	unsigned long fStamp1, fStamp2;	// pseudo time stamps of matrices and perhaps bundles...
	bool fHadMetrics;

	IBaseColor fColor;				// storage for color

	bool fExternalStyleSet;			// used to see if anyone had called setStyles()...

  public:
	// methods used by graf devices...

	// begin of text cache for renderer ...
	bool fSubTextCached;
	const UniChar* fUnicodeString;	// used for NT rendering
	length_type fUnicodeStringLength;	// used for NT rendering
	IText fSubText;	// used for Win95 and others doesn't do Unicode...
	// end of text cache for renderer ...

	// begin of styles cache for renderer...
	int fEscapement;
	bool fStylesCached;
	IBaseColor* fTextColor;
	const char* fTypeFace;
	GCoordinate fPointSize;
	int fWeight;
	unsigned char fItalic;
	unsigned char fUnderline;
	unsigned char fStrikeOut;
	unsigned char fOutline;
	unsigned char fBitmap;
	unsigned char fFixedPitch;
	
	// end of styles cache for renderer...
};

#pragma pack(pop)
#pragma enum(pop)

#if __IBMCPP__ >= 400
#pragma namemangling()
#endif

inline const IText* IGTextRun::text() const {
	return fText;
}

inline IGTextRun::offset_type IGTextRun::begin() const {
	return fBegin;
}

inline IGTextRun::offset_type IGTextRun::end() const {
	return fEnd;
}

inline IGPoint2D IGTextRun::anchor() const {
	return fAnchor;
}

inline IGTextRun::EAttachment IGTextRun::attachment() const
{
	return fAttachment;
}

#endif // _IGTXTRUN_
