/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 97 1.15.1.3 source/albert/graph2d/iknots.cpp, 2d, ioc.v400, 000302 
/*================
||
||  File:   Knots.c
||
||  What:   Knotvector related methods for implementing curves and surfaces
||          These are strictly private routines
||
||  Change history:
||
*/

#include <prspline.hpp>

#ifdef GRAPH2D_DEBUG
#include <stdio.h>
#endif

#include <grassert.hpp>
#include <inumeric.hpp>
#include <iknots.hpp>
#include <float.h>

// Segment definitions
#ifdef IC_PAGETUNE
#define _IKNOTS_CPP_
#include <ipagetun.h>
#endif

/******************************************************************************
    Evaluation code
 *****************************************************************************/

//
// Compute the derivatives.  Assumes the order-1'th set of spline
// values is passed in dvals, and returns a modified set of dvals with
// derivatives.  From Bartel's net code, but fixed to be insensitive
// to knot vector spacing.
//

static void  ComputeDeriv(
    unsigned long order,
    long brkPoint,
    long deriv,
    const GParametric * const knots,
    GParametric * dvals )
{
    register GParametric omega;
    long di, dr, s;
    register GParametric knotScale;

    if ((order == 2) && (deriv == 1))   // Special cases "bvals" loop will miss
    {
        dvals[0] = 1.0; dvals[1] = 0.0;
    }

    if ((order == 3) && (deriv == 2))
    {
        dvals[0] = 1.0; dvals[1] = 0.0; dvals[2] = 0.0;
    }

    knotScale = knots[brkPoint + 1] - knots[brkPoint];

    for (dr = order - deriv + 1; dr <= order; dr++)
    {
        di = brkPoint - dr + 1;
        for (s = dr - 2; s >= 0L; s--)
        {
            di++;
            omega = knotScale * ((GParametric)(dr-1)) / (knots[di+dr-1] - knots[di]);
            dvals[s + 1] += -omega * dvals[s];
            dvals[s] *= omega;
        }
    }
}

//
// Compute the Spline functions B[i,k](u).  For each call, this computes the
// the k values B[i-k][k](u)...B[i][k](u) and returns them in bvals.  If dvals1
// or dvals are not NULL, the corresponding derivative values are computed as
// well.  This is a critical inner loop for anything that evalutes curves or
// surfaces.  This code is loosely based on some routines posted to the net by
// Richard Bartels.
//
// Note:  The order and knot vectors are passed in as arguments so this routine
// can be used by TGCurve3D, TGSurface and the rendering code.
//
// Another note:  This code does NO ERROR CHECKING.  The brkPoint, knots, etc.
// are assumed to be for a valid interval of a valid spline.
//
// A final note: This code has been tuned extensivly with the MPW C
// compiler.  If you think you can speed it up, try MEASURING it
// first.  Many attempts to do so (common sub expression removal, more
// pointers) have only slowed it down.
//

void ComputeSplineValues(
    GParametric u,
    long brkPoint,
    unsigned long order,
    const GParametric * const knots,
    register GParametric * bvals,
    GParametric * dvals1,
    GParametric * dvals2 )
{
    register GParametric omega;
    register const GParametric * knotsptr = &(knots[brkPoint]);

    bvals[0] = 1.0;

    //
    // Common cases (linear, cubic and quadratic) are unrolled.
    //
    if (order == 2)
    {
        omega = (u - *knotsptr) / (knotsptr[1] - *knotsptr);
        bvals[0] = omega;
        bvals[1] = 1-omega;
        if (dvals1)
        {
            dvals1[0] = 1.0;    dvals1[1] = 0.0;
        }
    }
    else
    if ((order == 3) && (!dvals1) && (!dvals2))
    {
        bvals[1] = 0.0;
        omega = (u - *knotsptr) / (knotsptr[1] - *knotsptr);
        bvals[1] += (1-omega) * bvals[0];
        bvals[0] *= omega;
        bvals[2] = 0.0;
        omega = (u - knotsptr[-1]) / (knotsptr[1] - knotsptr[-1]);
        bvals[2] += (1-omega) * bvals[1];
        bvals[1] *= omega;
        omega = (u - *knotsptr) / (knotsptr[2] - *knotsptr);
        bvals[1] += (1-omega) * bvals[0];
        bvals[0] *= omega;
    }
    else
    if ((order == 4) && (!dvals1) && (!dvals2))
    {
        bvals[1] = 0.0;
        omega = (u - *knotsptr) / (knotsptr[1] - *knotsptr);
        bvals[1] += (1-omega) * bvals[0];
        bvals[0] *= omega;
        bvals[2] = 0.0;
        omega = (u - knotsptr[-1]) / (knotsptr[1] - knotsptr[-1]);
        bvals[2] += (1-omega) * bvals[1];
        bvals[1] *= omega;
        omega = (u - *knotsptr) / (knotsptr[2] - *knotsptr);
        bvals[1] += (1-omega) * bvals[0];
        bvals[0] *= omega;
        bvals[3] = 0.0;
        omega = (u - knotsptr[-2]) / (knotsptr[1] - knotsptr[-2]);
        bvals[3] += (1-omega) * bvals[2];
        bvals[2] *= omega;
        omega = (u - knotsptr[-1]) / (knotsptr[2] - knotsptr[-1]);
        bvals[2] += (1-omega) * bvals[1];
        bvals[1] *= omega;
        omega = (u - *knotsptr) / (knotsptr[3] - *knotsptr);
        bvals[1] += (1-omega) * bvals[0];
        bvals[0] *= omega;
    }
    else
    {
        register long r, s;
        long i;

        for (r = 2; r <= order; r++)
        {
            i = brkPoint - r + 1;
            bvals[r - 1] = 0.0;
            knotsptr = &(knots[i]);
            for (s = r-2; s >= 0; s--)
            {
                i++;
                knotsptr++;
                /* Stashing the "u - *knotsptr" only slowed it down... */
                omega = (u - *knotsptr) / (knots[i + r - 1] - *knotsptr);

                /* Unfolding this with pointers actually slowed down with MPW C++... */
                bvals[s + 1] += (1.0 - omega) * bvals[s];
                bvals[s] *= omega;
            }

            if ((dvals1) && (r == order-1))
            {
                for (s = 0L; s < order; s++)
                    dvals1[s] = (s < r) ? bvals[s] : 0.0;
            }
            if ((dvals2) && (r == order-2))
            {
                for (s = 0L; s < order; s++)
                    dvals2[s] = (s < r) ? bvals[s] : 0.0;
            }
        }
    }

    /*
    // Compute derivatives as well.  This was formed by mashing together
    // the two routines published by Bartels on Usenet.  They could
    // perhaps be rolled together even tighter; the kv denom is an
    // obvious example.  Perhaps the loops could be folded?
    */

    if (dvals1)
        ComputeDeriv( order, brkPoint, 1, knots, dvals1 );
    if (dvals2)
        ComputeDeriv( order, brkPoint, 2, knots, dvals2 );
}

//
// Find the index of the start of the interval for a given value of u.
// This is the Killer B's
//

long
FindInterval(
    const GParametric& u,
    const GParametric * const knots,
    unsigned long numPts,
    unsigned long order )
{
    long i;
    const GParametric * uPtr;

    if (u == knots[numPts])     // Allow maximum parameter value
        i = numPts - 1;
    else
    {
        i = numPts - 1 + order;
        uPtr = &(knots[i]);
        while ( (u < *uPtr) && (i > 0) )
            uPtr--, i--;
    }
    return( i );
}


//
// Given a breakpoint, find the start of the next interval
//

long
FindNextInterval( long curBrk, const GParametric * const knots,
                  unsigned long numPts )
{
    long i;

    curBrk++;
    i = curBrk + 1;

    while ( (knots[i] == knots[curBrk]) && (i <= numPts) )
        i++;

    return( i - 1 );
}

/******************************************************************************
    Discontinuities
 *****************************************************************************/

unsigned long GetNextDiscon(
    unsigned long startIndex,
    unsigned long discon,
    unsigned long order,
    unsigned long numPoints,
    const GParametric * const knots )
{
    InitializedAssert( order, "" );

    GParametric curU = knots[startIndex];
    long same, numKnots = order + numPoints;
    long i = startIndex;

    // Sanity check on discon

    InternalAssert( (discon < order), "" );


    // Skip past the current index to the next interval
    while ((i < numKnots) && (curU == knots[i]))
        i++;

    if (i >= numKnots)
        return numKnots - 1;    // Already hit the end

    // count how many same knots in a row, quitting if it's enough
    do {
        curU = knots[i];
        same = 1;
        i++;
        while ((i < numKnots) && (same < order - discon) && (knots[i] == curU))
        {
            i++; same++;
        }
    }
    while ((i < numKnots) && (same < order - discon));

    if (same >= order - discon)
        return i - same;
    else
        return numKnots - 1;
}


// Some algorithms (e.g., degree reduction, deformation) can only deal with
// piecewise Bezier curves.  These must keep track of the c0 discontinuties
// in the curve's knot vector, so that the control point indexing is correct
// and the knot vector in the new curve has c0 discon's in the same place.
// These routines deal with an array of discons that record which Bezier
// segments have a c0 discontinuity.

//
// Create an array indicating which segments in the curve start
// with a c0 discontinuity.  Assumes curve is piecewise Bezier.
//

long
MakeBezierDiscons(
                  const GParametric * knots,
                  unsigned long order,
                  unsigned long numPoints,
                  long *& discons )
{
    long d, seg, numDiscons;

    // Allocate based on rough-guess worst case
    discons = new long[(numPoints - 1) / (order - 1)];

    d = order - 1;                  // Use order for Beziers...
    discons[0] = -1;                // end flag
    seg = 1;
    numDiscons = 0;

    while (seg <= (numPoints - numDiscons - 1) / d)
    {
        // check for multiplicity of order
        if (knots[seg * d + numDiscons + 1] == knots[seg * d + numDiscons])
        {
            discons[numDiscons++] = seg - 1;
            discons[numDiscons] = -1;
        }
        seg++;
    }

    return numDiscons;
}

//
// Given an array discons, create the appropriate knot vector
//
void
CreateKnotsFromDiscons(
                       GParametric * knots,
                       unsigned long order,
                       unsigned long numPoints,
                       long * discons )
{
    long i, seg, d, numDiscons;

    knots[0] = 0.0;             // Hmmm...maybe keep track of the orig. range?
    numDiscons = 0;
    d = order - 1;              // Easier to work in degrees for Beziers...
    seg = 0;

    while (seg * d + numDiscons < numPoints)
    {
        for (i = 1; i <= d; i++)
            knots[d * seg + i + numDiscons] = (GParametric) seg;

        if (seg == discons[numDiscons])
        {
            numDiscons++;
            knots[d * seg + d + numDiscons] = (GParametric) seg;
        }

        seg++;
    }
    knots[d + numPoints] = (GParametric) (seg - 1);
}


// check whether knots are sorted in ascending order and multiplicity<=order

bool IsKnotvectorConsistent( const GParametric *Knots,
                                unsigned long n,
                                unsigned long order )
{
    long i, mul=1;
    GParametric last = Knots[0];
    bool ok = true;

    for( i=1; i<n; i++ ) {
        if( last>Knots[i] ) {
#ifdef GRAPH2D_DEBUG
            printf( "Knotvector check: Knot[%ld]=%lf < previous knot=%lf\n", i, Knots[i], last );
#endif
            ok = false;
            mul = 1;
        } else if( last==Knots[i] ) {
            mul++;
            if( mul>order ) {
#ifdef GRAPH2D_DEBUG
                printf( "Knotvector check: Knot[%ld]=%lf has multiplicity %ld (order=%ld)\n", i, Knots[i], mul, order );
#endif
                ok = false;
            }
        } else
            mul = 1;
        last = Knots[i];
    }
    return ok;
}

/******************************************************************************
    Refinement
 *****************************************************************************/

//
// Compute an alpha refinement matrix, given an original knot vector oldKnots,
// a refinement knot vector newKnots, the number numNew of new control points
// (over the existing numPts).  This uses the Lyche-Morken algorithm.
// (See "Making the Oslo Algorithm More Efficient", SIAM J. of Numerical
//  analysis, June 1986, p. 663, Algorithm 1.).
//

void ComputeAlphaMatrix(
                   const GParametric * oldKnots,
                   const GParametric * newKnots,
                   unsigned long u_numPts,
                   unsigned long u_numNew,
                   unsigned long u_order,
                   GParametric **& alpha )
{
    long numPts = u_numPts;     // Due to bug in Max/Min
    long numNew = u_numNew;     // ditto
    long order = u_order;       // ditto

    long        odelta, delta, ih, il, iu, nu, p;
    GParametric b1, b, d1, d2, tj;
    IStash<GCoordinate> z( order + 1 ), ah( order + 1 );
    long        i, j;

    alpha = new GParametric*[order];
    for (i = 0; i<order; i++)
        alpha[i] = new GParametric[numNew - order];

    /* first calculate the alpha matrix */

    for (j = 0; j < numNew-order; j++)
    {
        i = j + 1;
        odelta = FindInterval( newKnots[j], oldKnots, numPts, order );
        delta = odelta;

        while ((newKnots[i] == oldKnots[delta]) && (i < j + order))
        {
            i++;
            delta--;
        }

        ih = delta + 1;
        nu = 0;

        for (p = 1; p < order; p++)
            if (newKnots[j + p] == oldKnots[ih])
                ih++;
            else
            {
                nu++;
                z[nu] = newKnots[j + p];
            }

        ah[order] = 1.0;
        for (p = 1; p <= nu; p++)
        {
            b1 = 0.0;
            tj = z[p];
            if (p > delta)
                b1 = (tj - oldKnots[0]) * ah[1 + order - delta]
                    / (oldKnots[p + order - nu] - oldKnots[0]);
            il = Max( 1, delta - p + 1 );
            iu = Min( delta, numPts-1 + nu - p );

            for (i = il; i <= iu; i++)
            {
                d1 = tj - oldKnots[i];
                d2 = oldKnots[i + p + order - nu - 1] - tj;
                if ((d1 + d2) == 0.0)
                    b = 1.0e10;       // Magic number found by Anupam.
                                      // Defect #2956.
                else
                    b = ah[i + order - delta] / (d1 + d2);
                ah[i + order - delta - 1] = d2 * b + b1;
                b1 = d1 * b;
            }
            ah[iu+order-delta] = b1;
            if (iu < delta)
                ah[iu+order-delta] +=
                        + (oldKnots[numPts-1+order]-tj) * ah[iu+order-delta +1]
                        / (oldKnots[numPts-1+order] - oldKnots[iu+1]);
        }
        il = Max( delta - nu, 0 );
        iu = Min( delta, numPts-1 );
        ih = Max( odelta - (order - 1), 0 );

        /* NOTE: The AlphaMul should be taught to know about il and iu for
         * each value of j.  This would save a number of multiplies by zero.
         * [I tried this 11-Mar-93...it ran slower! (heap overhead)   -jp]
         */
        for (i = 0;  i< order; i++) alpha[i][j] = 0.0;
        for (i = il; i<= iu; i++)    alpha[i-ih][j] = ah[i+order-delta];
    }
}



void RefineKVUniform( unsigned long fOrder,
                      unsigned long fNumPoints,
                      const GParametric* fKnots,
                      IRawArray<GParametric>& newKnots )
{
    long i, o, numberKnots;
    numberKnots = 2*fNumPoints + 1;
    IRawArray<GParametric> tmpKnots( numberKnots );
    newKnots.swap( tmpKnots );

    o = 0;                      // index into new Knots
    for (i = 0; i< fOrder; i++ ) // knots at beginning of knotvector stay same
        newKnots[o++] = fKnots[i];

    // new knots halfway between existing ones
    for (i = fOrder; i<=fNumPoints; i++)
    {
        if( fKnots[i]!=fKnots[i-1] )
            newKnots[o++] = (fKnots[i-1]+fKnots[i] ) / 2.0;
        newKnots[o++] = fKnots[i];
    }

    // knots at end of knotvector stay same
    for ( i=fNumPoints+1; i<fNumPoints+fOrder; i++)
        newKnots[o++] = fKnots[i];

    newKnots.resize( o );
}

// Returns false if the knot vector was already in Bezier form
void RefineKVToBezier(
                      unsigned long fOrder,
                      unsigned long fNumPoints,
                      const GParametric* fKnots,
                      IRawArray<GParametric>& newKnots )
{
    long i, j, n, same, tmp;

    n = fNumPoints + fOrder;
    i = n + (n - 2 * fOrder) * (fOrder - 1); // Worst case new knots

    IRawArray<GParametric> tmpKnots( n + i );
    newKnots.swap( tmpKnots );

    newKnots[0] = fKnots[0];    // Always copy first

    j = 1;
    same = 1;
    for (i = 1; i < n - 1; i++)
    {
        newKnots[j] = fKnots[i]; // Look for sequences of k-1 fold knots...
        if (fKnots[i] == fKnots[i + 1])
            same++;
        else
        {
            if (same < fOrder - 1)
            {
                for (tmp = 1; tmp < fOrder - same; tmp++)
                {
                    j++;
                    newKnots[j] = fKnots[i];
                }
            }
            same = 1;
        }
        j++;
    }
    newKnots[j++] = fKnots[n - 1];
    newKnots.resize( j );
}

void RefineKVToPinned( unsigned long fOrder,
                       unsigned long fNumPoints,
                       const IRawArray<GParametric>& fKnots,
                       IRawArray<GParametric>& newKnots )
{
    long i;
    for (i = 0; i < fOrder; i++)
        newKnots[i] = fKnots[fOrder - 1];
    for (i = fOrder; i < fNumPoints; i++)
        newKnots[i] = fKnots[i];
    for (i = fNumPoints; i < fNumPoints + fOrder; i++)
        newKnots[i] = fKnots[fNumPoints];
}

/******************************************************************************
    Sets / Tests
 *****************************************************************************/


//
// A couple popular knot vector schemes...
//
void SetBezierKnots(
    unsigned long fOrder,
    unsigned long fNumPoints,
    IRawArray<GParametric>& fKnots )
{
    unsigned long i, j, ind, segments;

    /* only if number controlpoints is multiple of order */

    segments = (fNumPoints-1) / (fOrder-1);
    if ( segments * (fOrder-1) == fNumPoints-1 )
    {
        ind = 0; /* index into Knots */
        for (i = 0; i < fOrder; i++)
            fKnots[ind++] = 0.0;
        for (i = 0; i < segments-1; i++)
            for (j = 0; j < fOrder-1; j++)
                fKnots[ind++] = i + 1.0;
        for (i = 0; i < fOrder; i++)
            fKnots[ind++] = segments;
    }
}

void SetPinnedKnots(
    unsigned long fOrder,
    unsigned long fNumPoints,
    IRawArray<GParametric>& fKnots )
{
    unsigned long i;

    for(i = 0; i < fOrder; i++)
        fKnots[i] = 0.0;
    for (i = fOrder; i < fNumPoints; i++)
        fKnots[i] = i - fOrder + 1;
    for (i = fNumPoints; i < fNumPoints + fOrder; i++)
        fKnots[i] = fNumPoints - fOrder + 1;

}

bool IsBezierKnots( unsigned long fOrder,
                       unsigned long fNumPoints,
                       const IRawArray<GParametric>& fKnots )
{
    register unsigned long i, same;
    bool isbez = true;
    GParametric curU;

    if (! IsPinnedKnots( fOrder, fNumPoints, fKnots ))
        return false;   // Bezier implies pinned

    i = fOrder;
    while ((i < fNumPoints) && (isbez))
    {
        curU = fKnots[i];
        same = 1;
        i++;
        while ((i < fNumPoints) && (curU == fKnots[i]))
        {
            i++;
            same++;
        }
        isbez = (same >= fOrder-1); // Must have at least k-1 sequential knots
    }
    return isbez;
}

bool IsPinnedKnots( unsigned long fOrder,
                       unsigned long fNumPoints,
                       const IRawArray<GParametric>& fKnots )
{
    bool pinned = true;
    long    i;
    GCoordinate min, max;

    min = fKnots[0];
    max = fKnots[fNumPoints+fOrder-1];
    for( i=1; (i<fOrder) && pinned; i++ )
        pinned = pinned && fKnots[i]==min && fKnots[fNumPoints+i]==max;
    return pinned;
}

void MoveAndSortKnot( unsigned long index, GParametric u,
                      unsigned long fOrder,
                      unsigned long fNumPoints,
                      IRawArray<GParametric>& fKnots )
{
    InitializedAssert( fOrder, "" );

    long i, brkPt, max = fNumPoints + fOrder - 1;
    ParameterAssert( (index <= max), "" );

    if (fKnots[index] == u)     // No brainer
        return;

    // Look through the knot vector to see where the knot belongs

    brkPt = FindInterval( u, &fKnots[0], fNumPoints, fOrder );
    if (u > fKnots[max])        // Watch end case
        brkPt = max;

    if (index < brkPt)
        for (i = index; i < brkPt; i++)
            fKnots[i] = fKnots[i + 1];

    if (index > brkPt)
    {
        for (i = index; i > brkPt; i--)
            fKnots[i] = fKnots[i - 1];
        if (brkPt > 0)
            brkPt++;
    }

    fKnots[brkPt] = u;

    // check to see if they hosed the kv by exceeding valid multiplicity

    while ((brkPt < fOrder + fNumPoints - 1) && (fKnots[brkPt + 1] == u))
        brkPt++;
    for (i = brkPt; (i>=0) && (i >= brkPt - fOrder) && (fKnots[i] == u); i--);
    ParameterAssert( (brkPt - i <= fOrder), "Multiplicty > order" );
}


/******************************************************************************
    privateMergeKnots
 *****************************************************************************/


/*
    Knotvector merging algorithm
    ----------------------------
    Used by makeCompatible routines (in IGCurve2D and TGCurve3D) and sweep
    routines, when scale and twist knotvectors are merged into trajectory
    knotvector.

    Description:
    This is a merge algorithm (an OR operation) with a few special cases.
    Src and Dst knots are merged into a new knotvector, Result. This is a
    new heap block passed back by the routine.

    Assumptions:
    The two incoming knotvectors Src and Dst have the same range, i.e. their
    first and last knots respectively have the same numeric value.
    the knotvectors are legal, ie there are no breakpoints with
    multiplicity>order
    no other assumptions regarding internal knotvector structure.
    Src and Dst discontinuities have to be preserved. Only c0 and C1
    are preserved,
    as DstOrder and DstOrder-1 multiplicity breakpoints, respectively.

    NOTE: This has been rewritten from scratch to avoid problems with
    infinite loops and glitches due to round-off errors.  This does not
    strictly follow the order constraint above, but as long as the kv's
    have the same range, this shouldn't be a problem.
*/

// Copy knot vectors to ResultKnots until next breakpoint is reached.
static void
Advance( const GParametric * kv1,
         const GParametric * kv2,
         long& ind1, long& ind2,
         long num1, long num2,
         IRawArray<GParametric>& ResultKnots, long& ResultNumberKnots,
         GParametric tol )
{
    long mult1, mult2;
    mult1 = 1;
    while ((ind1 < num1) && Compare( kv1[ind1],
                                               kv1[ind1 + 1], tol))
    {
        mult1++;
        ResultKnots[ResultNumberKnots++] = kv1[ind1++];
    }
    ResultKnots[ResultNumberKnots++] = kv1[ind1++];

    mult2 = 0;
    while ((ind2 < num2) && Compare( kv2[ind2],
                                               kv1[ind1 - 1], tol ))
    {
        mult2++;
        ind2++;
    }

    while ((mult2 - mult1) > 0)
    {
        ResultKnots[ResultNumberKnots++] = kv2[ind2 - 1];
        mult2--;
    }
}


void privateMergeKnots(
    const GParametric *SrcKnots,    // incoming knots to be merged ()
    unsigned long  SrcNumberKnots,  // number of source knots
    unsigned long  SrcOrder,        // source order (to determine src discons)
    const GParametric *DstKnots,    // destination knotvector
    unsigned long  DstNumberKnots,  // number of destination knots
    unsigned long  DstOrder,        // destination order
    IRawArray<GParametric> &ResultKnots     // output knotvector
    )
{
    // Remove two least significant bits from comparisons
    GParametric tol = (Max( SrcKnots[SrcNumberKnots - 1],
                           DstKnots[DstNumberKnots - 1])
                       - Min( SrcKnots[0], DstKnots[0] )) * DBL_EPSILON * 4.0;

    // allocate worst case array (no knots match except boundary knots
    // (same range!))
    ResultKnots = IRawArray<GParametric>( (unsigned long)((Abs((long)SrcOrder-(long)DstOrder) + 1)
                                          * (DstNumberKnots + SrcNumberKnots)) );
    long ResultNumberKnots = 0;

    long curSrc = 0, curDst = 0;

    while ((curSrc < SrcNumberKnots) && (curDst < DstNumberKnots))
    {
        if (SrcKnots[curSrc] < DstKnots[curDst])
            Advance( SrcKnots, DstKnots, curSrc, curDst,
                     SrcNumberKnots, DstNumberKnots,
                     ResultKnots, ResultNumberKnots, tol );
        else
            Advance( DstKnots, SrcKnots, curDst, curSrc,
                     DstNumberKnots, SrcNumberKnots,
                     ResultKnots, ResultNumberKnots, tol );
    }

    // Finish the remainder
    if (curSrc < SrcNumberKnots)
        while (curSrc < SrcNumberKnots)
            ResultKnots[ResultNumberKnots++] = SrcKnots[curSrc++];

    if (curDst < DstNumberKnots)
        while (curDst < DstNumberKnots)
            ResultKnots[ResultNumberKnots++] = DstKnots[curDst++];


    ResultKnots.resize( ResultNumberKnots );
}
