/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 41 1.8.2.4 source/albert/graph2d/numerics.hpp, 2d, ioc.v400, 000302 
/*================
||
||  File:   NumericalStuff.h
||
||  What:   Various numerical routines from "Numerical recipes"
||
||  Classes defined in this file:
||
||  * IGaussianIntegrator
||  * INewtonRootFinder
||  * IChebyshev
||
||  Change History:
||
*/


#ifndef _NUMERICS_
#define _NUMERICS_

#ifndef IC_GRAFTYPES
#include <igrtypes.hpp>
#endif

#ifndef IC_PRSPLINEROUTINES
#include "prspline.hpp"
#endif

#if __IBMCPP__ >= 400
#pragma namemangling(compat)
#endif

class IGaussianIntegrator
{
/*
    Family: Geometry;
    audience: Geometry implementation;
    Info:

    This performs numerical integration of a function.  This must be
    subclassed, and the subclass must provide an "integrand".  See
    Numerical recipes (Press, et. al) for more information.;
    */

  public:

    IGaussianIntegrator( long order );
	virtual ~IGaussianIntegrator();

    double gaussianQuadrature( double from, double to );

    // Override with the function to be integrated
    virtual double integrand( double x ) = 0;
};

class INewtonRootFinder
{
/*
    Family: Geometry;
    audience: Geometry implementation;
    Info:

    This finds the root of a function, using Newton iteration  This must be
    subclassed, and the subclass must provide a "function" to find the root of.
    See Numerical recipes (Press, et. al) for more information.;
    */

  public:

    INewtonRootFinder();        // Could initialize accuracy, max iterations
	virtual ~INewtonRootFinder();

    GParametric findRoot( GParametric lowBracket,
                          GParametric highBracket,
                          GParametric tolerance );

    // Override with the function to find the root of

    virtual void function( GParametric u, double& y, double& yDerivative ) = 0;
};

#ifdef NO_REPARAM_YET
class IChebyshev
{
/*
    Family: Geometry;
    audience: Geometry implementation;
    Info:

    This finds the Chebyshev approximation a function.  This must be
    subclassed, and the subclass must provide a "function" to be approximated.
    See Numerical recipes (Press, et. al) for more information.;
    */

  public:
    IChebyshev();

    void chebyshevFit( GParametric from,
                       GParametric to,
                       unsigned long numCoefs,
                       IStash<GCoordinate>& coefficients );

    // Override with the function to approximate
    virtual double function( GParametric u ) = 0;
};
#endif // NO_REPARAM_YET

#if __IBMCPP__ >= 400
#pragma namemangling()
#endif

#endif // _NUMERICS_
