/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 73 1.5.1.3 source/albert/graph2d/nurbs.cpp, 2d, ioc.v400, 000302 

// nurbs.cpp

#include <nurbs.hpp>
#include <iknots.hpp>
#include <prspline.hpp>
#include <inumeric.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
#define _NURBS_CPP_
#include <ipagetun.h>
#endif

INurbDiscretiser::INurbDiscretiser(IGRect2D clipBounds, double resolution)
{
	fClipBounds = clipBounds;
	fResolution = resolution;
}

INurbDiscretiser::INurbDiscretiser(IGRect2D clipBounds)
{
	fClipBounds = clipBounds;
	fResolution = 0.1;
}


void INurbDiscretiser::DiscretizeOrderTwo(const GParametric* knotVector, IGRPoint2DArray pointArray)
{
	int numPts = pointArray.numberOfPoints();

	begin();
	for (int i = 0; i < numPts; i++) {
		addPoint(pointArray[i].divW());
		// check for c0 discontinuity...
		if (knotVector[i + 1] == knotVector[i + 2]) {
			endPath();
		}
	}
	endPath();
}

void INurbDiscretiser::DiscretizeNurb(
	unsigned long uorder,
	const GParametric* knots,
	IGRPoint2DArray controlPoints,
	bool closed)
{
	long numPts = controlPoints.numberOfPoints();
	long i,j, steps, lastBrkPoint;
	long order = uorder;
	IGRPoint2D *tPtr, *deltaPtr, *deltaPtr1;
	IGRect2D segmentBounds;
	bool c0disCon = false;
	bool skippedSegment = false;
	IGRPoint2DArray delta(order +1);
	IGRPoint2DArray t(order +1);

	bool RatFlag = false;
	//IGRPoint2D controlPtr[] = controlPoints;
	const IGRPoint2D* controlPtr = &controlPoints[0];
	for (i = 0; i < numPts; i++) {
		if (controlPtr->fW != 1.0) {
			RatFlag = true;
			break;
		}
		controlPtr++;
	}

	GCoordinate u, u1, first_u, last_u, tmp_u;

	// Find the interval for the first span
	first_u = knots[order-1];
	u = first_u;
	last_u = knots[numPts];
	long brkPoint = FindInterval(u, knots, numPts, order);
	u1 = knots[brkPoint + 1];

	begin();

	// Step through the spans of the curve, forward differencing each span.
	while (u < last_u) {
		// Compute the bounds of the current segment
		segmentBounds = GetSegmentBounds(brkPoint, order, controlPoints, RatFlag);

		// Is this segment visible?
		skippedSegment = !NonEmptyIntersects(fClipBounds, segmentBounds);
		if (!skippedSegment) {
			steps = GetCurveSteps(brkPoint, order, &controlPoints[0], fResolution);

			// So we have at least order initial samples to prime the forward diff
			steps = (steps < order) ? order : steps;

			if (steps > 5000) // If the curve is too huge,
			{         // just draw the control mesh
				for (j = brkPoint - order + 1; j < brkPoint; j++)
					addPoint(controlPoints[j].divW());
				skippedSegment = true;  // So no attempt is made to use FD info
			}
		}           // If the segment was skipped, and we're drawing
		else          // a loop, add the ctl points so the loop
		{           // topology remains correct
			if (closed)
				for (j = brkPoint - order + 1; j < brkPoint; j++)
					addPoint(controlPoints[j].divW());
		}

		if (!skippedSegment)  // Segment is visible and of reasonable size
		{
			tmp_u = (u1 - u) / (double)steps;
			tPtr = &t[0];
			for (j = 0; j < order; j++) {
				NurbEval(
					&controlPoints[0],
					(long)brkPoint,
					u + j * tmp_u,
					knots,
					(unsigned long)order,
					RatFlag,
					tPtr++);
			}

			// Note first point on the curve or a c0 discontinuity
			delta[0] = t[0];

			// store the first point if not a closed loop
			if ((u == first_u) || c0disCon) {
				// Pin it to the bounds in case of round-off error
				IGPoint2D vertex = RatFlag ? delta[0].divW() : delta[0].dropW();
				segmentBounds.pin(vertex);
				addPoint(vertex);
			}

			// Build the forward difference table from the initial samples on the curve.
			deltaPtr = &(delta[1]);

			if (RatFlag) {
				for (i = 1; i < order; i++) {
					tPtr = &t[0];
					register IGRPoint2D* tPtr1 = &(t[1]);
					for (j = 0; j < order - i; j++) {
						tPtr->fX = tPtr1->fX - tPtr->fX;
						tPtr->fY = tPtr1->fY - tPtr->fY;
						tPtr->fW = tPtr1->fW - tPtr->fW;
						tPtr++;
						tPtr1++;
					}
					*deltaPtr++ = t[0];
				}
			}
			else {
				for (i = 1; i < order; i++) {
					tPtr = &t[0];
					register IGRPoint2D* tPtr1 = &(t[1]);
					for (j = 0; j < order - i; j++) {
						tPtr->fX = tPtr1->fX - tPtr->fX;
						tPtr->fY = tPtr1->fY - tPtr->fY;
						tPtr++;
						tPtr1++;
					}
					*deltaPtr++ = t[0];
				}
			}

			// Step through the curve span
			i = 0;
			while (i<steps)
			{
				// Update the forward differences
				deltaPtr = &(delta[0]);
				deltaPtr1 = &(delta[1]);
				if (RatFlag) {
					for (j = 0; j < order-1; j++) {
						deltaPtr->fX += deltaPtr1->fX;
						deltaPtr->fY += deltaPtr1->fY;
						deltaPtr->fW += deltaPtr1->fW;
						deltaPtr++;
						deltaPtr1++;
					}
				}
				else {
					for (j = 0; j < order-1; j++) {
						deltaPtr->fX += deltaPtr1->fX;
						deltaPtr->fY += deltaPtr1->fY;
						deltaPtr++;
						deltaPtr1++;
					}
				}

				if (i < steps - 1) {
					// Pin it to the bounds in case of round-off error
					IGPoint2D vertex =
					RatFlag ? delta[0].divW() : delta[0].dropW();
					segmentBounds.pin(vertex);
					addPoint(vertex);
				}
				i++;
			}
		}

		// Move to the next interval on the curve.
		u = u1;
		lastBrkPoint = brkPoint;
		brkPoint = FindNextInterval(brkPoint, knots, numPts);
		u1 = knots[brkPoint + 1];

		if (closed || (!skippedSegment)) {
			c0disCon = true;
			bool sameKnot, c1disCon = true;
			if (u == last_u) {
				for (j = brkPoint; j < brkPoint + order; j++) {
					sameKnot = (u == knots[j]);
					c0disCon = c0disCon && (u == knots[j]);
					if (j < (brkPoint + order - 1))
						c1disCon = c1disCon && sameKnot;
				}
			}
			else {
				for (j = brkPoint; j >= brkPoint - (order - 1); j--) {
					sameKnot = (u == knots[j]);
					c0disCon = c0disCon && (u == knots[j]);
					if (j >= brkPoint - order + 2)
						c1disCon = c1disCon && sameKnot;
				}
			}

			if (c0disCon || c1disCon) {
				addPoint(controlPoints[lastBrkPoint].divW());
				if (c0disCon)
					endPath();
			}
			else {
				if (skippedSegment)
					addPoint(controlPtr[lastBrkPoint].divW());
				else {
					// Pin it to the bounds in case of round-off error
					IGPoint2D vertex = RatFlag ? delta[0].divW() : delta[0].dropW();
					segmentBounds.pin(vertex);
					addPoint(vertex);
				}
			}
		}
		else
			endPath();
	}
	endPath();
}

int INurbDiscretiser::GetCurveSteps(int brkPoint, int order, const IGRPoint2D* V, double tolerance)
{
	long i, numsteps = 0;
	IGRPoint2D tmp;
	GCoordinate xstep, ystep;

	// Should never see a linear curve (polyline) here, but if we do there's
	// an obvious answer
	if (order == 2)
		return 1;


	// Find the second derivative for a the hypothetical quadratic segment
	// formed by each subsequent set of three control points in the segment.
	// Use the maximum of this as the stepsize for the over-all segment.
	for (i = brkPoint - order + 1; i < brkPoint - 1; i++)
	{
		tmp = V[i].divW() - 2 * V[i+1].divW() + V[i+2].divW();
		xstep = Abs(2 * tmp.fX) / tolerance;
		ystep = Abs(2 * tmp.fY) / tolerance;
		numsteps = Max(numsteps, (long)SquareRoot(Max((double)xstep, (double)ystep)));
	}

	return numsteps;
}

bool INurbDiscretiser::NonEmptyIntersects(IGRect2D a, IGRect2D b)
{
	return ((a.fLeft < b.fRight) && (a.fRight > b.fLeft) && (a.fTop < b.fBottom) && (a.fBottom > b.fTop));
}
