/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 31 1.9.2.6 source/albert/graph2d/prspline.hpp, 2d, ioc.v400, 000302 
/*================
||
||  File:   prSplineRoutines.h
||
||  What:   Useful routines for IGCurve2D and other objects
||          These are strictly private routines
||
||  Change history:
||
*/

#ifndef _PRSPLINE_
#define _PRSPLINE_

#ifndef IC_GRAFTYPES
#include <igrtypes.hpp>
#endif

#ifndef IC_RAWARRAY
#include <irawarr.hpp>
#endif


#ifndef IC_PRRATIONALOPERATORS
#include "prratopr.hpp"
#endif

#ifndef IC_NUMERICS
#include <inumeric.hpp>
#endif

#ifndef IC_KNOTS
#include <iknots.hpp>
#endif

#ifndef IC_BASECURVE
#include <ibcurve.hpp>
#endif

#ifndef HIDE_TEMPLATE_DEFINITIONS
#define MAKE_TStash_DEFINITIONS_VISIBLE
#else
#define inline
#endif


#if __IBMCPP__ >= 400
#pragma namemangling(compat)
#endif

class IGPoint2D;
class IGRPoint2D;
class TGPoint3D;
class TGRPoint3D;
class IGRPoint2DArray;

#define kLocalMaxOrder 6

#pragma pack(push,4)

template <class AType>
class IStash {
/*
    Family: Geometry;
    audience: Curve internals programmers;
    Info:

Spline computations frequently require a local stash of fOrder
number of elements.  Rather than always take them from the heap,
it's wise to allocate them off the stack if you know there's less
than a typical small number of them, since numbers greater than
4 (for cubics) are rarely used.  This class takes care of
/the bookeeping.  Amazingly enough, using an inline [] operator
generates exactly the same code as using normal array indexing! ;

*/
  public:

    IStash( long unsigned num );
    ~IStash();

    AType& operator[]( long which );

  private:
    AType localCoords[kLocalMaxOrder];
    AType * fArray;
};

#pragma pack(pop)

#ifdef MAKE_TStash_DEFINITIONS_VISIBLE

template <class AType>
inline IStash<AType>::IStash( long unsigned num )
{
    if (num > kLocalMaxOrder)
        fArray = new AType[ num ];
    else
        fArray = localCoords;
}

template <class AType>
inline IStash<AType>::~IStash()
{
    if (fArray != localCoords)
        delete [] fArray;
}

template <class AType>
inline
AType& IStash<AType>::operator[]( long which )
{
    return fArray[which];
}

#endif

// Multiply a set of basis functions to a set of control points

template<class PointType>
void BasisMul(
    long brkPoint,
    unsigned long order,
    const GParametric * const bvals,
    const PointType * const points,
    PointType& p)
{
    register long i;
    register const PointType * cp;
    register const GParametric * bp;
    register GCoordinate tmp;

    SetToZero( p );

    bp = &(bvals[order - 1]);
    cp = &(points[brkPoint - order + 1]);

    for (i = 0; i < order; i++)
    {
        tmp = *bp--;
        p += *cp * tmp;
        cp++;
    }
}

// Return the geometric bounds of the curve segment starting at knot
// index brkPoint
extern
IGRect2D GetSegmentBounds( long brkPoint, long order,
                         const IGRPoint2DArray& pts, bool ratFlag = true );

//
// evaluate a point on a NURB curve, using a very efficient method.
// The code (and the notation) is from the Killer B's, p. 387.  This
// is used instead of IGCurve2D[3D]::evaluate when you already know what
// the breakpoint is, and you have raw pointer access to the curve data.
//

template<class PointType>
void NurbEval( const PointType * V,             // Local control points
              long delta,                       // Breakpoint
              GCoordinate u,                    // Parameter to eval. at.
              const GParametric * const knots,  // Local knot vector
              unsigned long order,              // Order of the curve
              bool /*RatFlag*/,             // true if rational w != 1.0
              PointType * p )                   // Output
{
    long i;
    IStash<GParametric> coef( order + 1 );
    GCoordinate * cf_ptr;
    const PointType * Vptr;

    ComputeSplineValues( u, delta, order, knots, &coef[0], 0 /*NIL*/, 0 /*NIL*/ );

    BasisMul( delta, order, &coef[0], V, *p );
}


                                // For "matrix" cubic spline computation
extern void CubicBasisMatrixMul( const BasisMatrix a,
                                 const BasisMatrix b,
                                 BasisMatrix destination );

//
// Find distance from point to line -segment- measuring distances
// to endpoints if needed
//

template<class PointType>
GParametric PointSegDist (
                        const PointType& p0,
                        const PointType& p1,
                        const PointType& test,
                        GCoordinate& dist )
{
    PointType v0, vm;
    GCoordinate t;
    GCoordinate denom;
    PointType p;

    // Find two vectors, p0p1 & p0test
    v0 = p1 - p0;
    vm = test - p0;

    // Ratio of dot prodcuts gives the projection of vm onto v0
    denom = v0.dotProduct( v0 );
    if (denom == 0.0)
        t = 0.0;
    else
        t = v0.dotProduct( vm ) / denom;

    // Distance to line is point on v0 corresponding to t minus testpoint
    dist = ((p0 + v0 * t) - test).vectorLength();

    // if t < 0 or t > 1, then measure distance to the nearest endpoint
    if (t < 0)
    {
        dist = (p0 - test).vectorLength();
        t = 0.0;
    }
    if (t > 1)
    {
        dist = (p1 - test).vectorLength();
        t = 1.0;
    }
    return t;
}

//
// This is a floating point compare that truncates the least
// significat bits.  LSB are likely to be corrupted when knot
// vectors are merged, due to rescaling operations.  By lopping
// the bits off, we greatly reduce the chances of introducing
// unintentional discontinuities.  Tolerance should be
// set to (maxKnot - minKnot) * DBL_EPSILON * 4
//
#if !(__IBMCPP__ >= 400)
static
#endif
inline bool Compare(double a, double b, double tolerance)
{
    return Abs( a - b ) < tolerance;
}

//
// These basis conversion matrices are shared by the degree reduction code
// and the Arc length reparameterization code.  They convert from Chebyshev
// polynomials to Bezier.
//
extern GCoordinate *** ChebToBez;
extern void MakeReductionMatrices( long order );

extern long comb(long d, long i);

// Save casting clutter
#if !(__IBMCPP__ >= 400)
static
#endif
inline double fchoose(long d, long i)
{   return (double)comb(d, i); }


#if __IBMCPP__ >= 400
#pragma namemangling()
#endif

#ifdef HIDE_TEMPLATE_DEFINITIONS
#undef inline
#endif

#endif // _PRSPLINE_
