/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 36 1.12.1.3 source/albert/graph2d/redcurve.inl, 2d, ioc.v400, 000302 
/*================
||
||  File:   redcurve.inl
||
||  What:   Reduce the order of a curve.
||
||
||  Change History:
||
*/

#ifndef _REDCURVE_INL_
#define _REDCURVE_INL_

#ifndef _IGLOOP2D_
#include <igloop2d.hpp>
#endif

#ifndef _PRSPLINE_
#include "prspline.hpp"
#endif

#ifndef _GRASSERT_
#include <grassert.hpp>
#endif

#ifndef _INUMERIC_
#include <inumeric.hpp>          // For Log2 and Abs
#endif

#ifndef _IKNOTS_
#include <iknots.hpp>
#endif

#ifndef _PRRATOPR_
#include "prratopr.hpp"
#endif

#ifndef MAKE_TBaseCurve_DEFINITIONS_VISIBLE
#error "This file is compiled only when instantiating templates"
#endif


//
// Degree reduction is taken from the paper "Degree reduction of Bezier
// Curves" by Watkins and Worsey, CAD Sept 88.  The technique is pretty
// straightforward: first convert the Bezier polynomials to Chebycheff
// polynomials, drop the least significant Chebycheff term, and then convert
// back to Bezier form.  The magnitude of the Chebycheff term dropped gives
// a measure of the error introduced by the process, if it's too large then
// the curve is subdivided and the process repeated until the error falls
// to a tolerable level.
//

// Basis conversion matrices - Carl Sagan pointers!

static GCoordinate *** BezToCheb = 0 /*NIL*/;
GCoordinate *** ChebToBez = 0 /*NIL*/;
static const long kLocalMaxReduce = 35;

// Storage management for double-indexed matrices
// Because the matrix invert routine expects a contiguous array of numbers,
// we allocate one block of numbers and assign the row pointers to point
// into this block.

GCoordinate **
AllocMatrix( long size )
{
    long i;
    GCoordinate ** mat;

    // Allocate the matrix as one big block with all the rows
    // pointing into it.

#ifdef NO_OC_APPLICABLE
// #ifndef SS
    TRY
// #endif // not SS
#endif // NO_OC_APPLICABLE

                mat = new GCoordinate*[size]; mat[0] = 0 /*NIL*/;
                mat[0] = new GCoordinate[ size * size ];

#ifdef NO_OC_APPLICABLE
// #ifndef SS
    EXCEPT
        delete [] mat[0];
        delete [] mat;
        RETHROW 0 /*NIL*/;
    ENDTRY
// #endif // not SS
#else // NO_OC_APPLICABLE
        if (!mat || !mat[0])
        {
        delete [] mat[0];
        delete [] mat;
                InternalAssert(false, "Matrix allocation failed");
        }
#endif // NO_OC_APPLICABLE

    for (i = 1; i < size; i++)
        mat[i] = &(mat[0][i * size]);
    return mat;
}

void
FreeMatrix( GCoordinate ** mat )
{
    delete [] mat[0];
    delete [] mat;
}

// Multiply two matrices defined by pointer-to-array style arrays
// This belongs in a math library someplace...

void ShiftMatrixMul( GCoordinate ** a,
                     GCoordinate ** b,
                     GCoordinate ** dest, long size )
{
    long i, j, k;
    for (i = 0; i < size; i++)
        for (j = 0; j < size; j++)
        {
            dest[i][j] = 0.0;
            for (k = 0; k < size; k++)
                dest[i][j] += a[i][k] * b[k][j];
        }
}

static long
ipow( long x, long p )
{
    register long r, i;

    r = 1;
    for (i = 1; i <= p; i++)
        r *= x;
    return( r );
}


//
// Create the change of basis matrices to go to/from Chebycheff & Bezier.
// Beware of errors in the formulas given by Watkins & Co - See J. Peterson,
// "Letter to the Editor", CAD, August 91 for the corrections.
//

void
MakeReductionMatrices( long order )
{
    long degree, i, j;
    GCoordinate ** BezToPol, ** PolToCheb, ** PolToBez, ** ChebToPol;
    static long currentMaxOrder = kLocalMaxReduce;

    // Alloc the storage if it's not there already

    if (! BezToCheb)
    {
        currentMaxOrder = Max( kLocalMaxReduce, order + 1 );
        BezToCheb = new GCoordinate**[ currentMaxOrder ];
        ChebToBez = new GCoordinate**[ currentMaxOrder ];
        for (i = 0; i < currentMaxOrder; i++)
        {
            BezToCheb[i] = 0 /*NIL*/;
            ChebToBez[i] = 0 /*NIL*/;
        }
    }
    else
        if (currentMaxOrder <= order)
        {
#ifdef NO_OC_APPLICABLE
// #ifndef SS
            BezToCheb = (GCoordinate ***)
            TMemoryHeap::ReallocateAny( BezToCheb,
                                        (order+1) * sizeof( GCoordinate ** ));
            ChebToBez = (GCoordinate ***)
            TMemoryHeap::ReallocateAny( ChebToBez,
                                        (order+1) * sizeof( GCoordinate ** ));
// #endif // not SS
#else // NO_OC_APPLICABLE
            BezToCheb = (GCoordinate ***) realloc( BezToCheb, size_t((order+1) * sizeof( GCoordinate ** )));
            ChebToBez = (GCoordinate ***) realloc( ChebToBez, size_t((order+1) * sizeof( GCoordinate ** )));
#endif // NO_OC_APPLICABLE

            for (i = currentMaxOrder; i <= order; i++)
            {
                BezToCheb[i] = 0 /*NIL*/;
                ChebToBez[i] = 0 /*NIL*/;
            }
            currentMaxOrder = order + 1;
        }

    if (! BezToCheb[order])
    {
        BezToCheb[order] = AllocMatrix( order );
        ChebToBez[order] = AllocMatrix( order );
    }
    else
        return;

    BezToPol = AllocMatrix( order );
    PolToCheb = AllocMatrix( order );
    PolToBez = AllocMatrix( order );
    ChebToPol = AllocMatrix( order );

    degree = order - 1;         // Degree works best for beziers

    //
    // Compute the Chebycheff/Bezier matrices by first finding the
    // Chebycheff/polynomial and Bezier/polynomial matrices, and
    // multiplying them together.
    //

    ChebToPol[0][0] = 1.0;
    ChebToPol[0][1] = 0.0;
    ChebToPol[1][0] = -1.0;
    ChebToPol[1][1] = 2.0;
    for (i = 0; i < order; i++)
        for (j = 0; j < order; j++)
        {
            // BezToPol

            if (j >= i)
                BezToPol[i][j] = ipow(-1, i + j + 2)
                                * fchoose(degree, i) * fchoose(degree-i, degree-j);
            else
                BezToPol[i][j] = 0.0;

            // PolToBez

            if (j >= i)
                PolToBez[i][j] = fchoose( j, j - i ) / fchoose( degree, i );
            else
                PolToBez[i][j] = 0.0;

            // Note:  The formulas for these two in Watkin's paper are WRONG!

            // PolToCheb

            PolToCheb[i][j] = 0.0;
            if ((i <= 1) && (j <= i))
                PolToCheb[i][j] = 1.0 / ((double)i + 1.0);

            if ((i >= 2) && (j == 0))
                PolToCheb[i][j] = fchoose(2 * i - 1, i - 1) / Power( 2, 2 * i -1 );

            if ((i >= 2) && (j >= 1) && (j == i))
                PolToCheb[i][j] = fchoose(2 * i - 1, i - j) / Power(2, 2 * i - 1);

            if ((i >= 2) && (j >= 1) && (j != i))
                PolToCheb[i][j] = (fchoose(2*i-1, i-j) + fchoose(2*i-1, i-(j+1)))
                                 / Power(2, 2 * i - 1);

            // ChebToPol

            if ((i > 1) || (j > 1))
                ChebToPol[i][j] = 0.0;

            if ((j <= i) && (i >= 2))
            {
                if (j >= 1)
                    ChebToPol[i][j] = 4 * ChebToPol[i-1][j-1]
                                     - 2 * ChebToPol[i-1][j]-ChebToPol[i-2][j];
                else
                    ChebToPol[i][j] = -2*ChebToPol[i-1][j] - ChebToPol[i-2][j];
            }
        }

    ShiftMatrixMul( BezToPol, PolToCheb, BezToCheb[order], order );
    ShiftMatrixMul( ChebToPol, PolToBez, ChebToBez[order], order );

    FreeMatrix( BezToPol );
    FreeMatrix( PolToCheb );
    FreeMatrix( ChebToPol );
    FreeMatrix( PolToBez );
}

//
// Reduce a curve by one degree.  Assumes the input is a complete
// peicewise Bezier curve of the specified order (does no error checking!)
//

template <class RPointType, class PointType>
void IBaseCurve<RPointType, PointType>::reduceOneDegree(
                                         IRawArray<RPointType>& pts,
                                         unsigned long order,
                                         unsigned long numPts,
                                         GCoordinate tolerance,
                                         long * discons,
                                         unsigned long numDiscons,
                                         unsigned long * newPts )
{
    long i, j, curDiscon, seg, numSegs, degree, origPts, curTmpChunk;
    GCoordinate err, subDivs, tmp;
    long dimension = sizeof(PointType) / sizeof(GCoordinate);

    RPointType * left, * right; // Subdivision storage
    RPointType endPoints[2];
    RPointType * tmpPts, * tmpPtr, * ptsPtr;

    degree = order - 1;         // For beziers, degrees are more natural

    numSegs = (numPts - numDiscons - 1) / degree;

    // Make storage for tmp arrays...

    left = new RPointType[order];
    right = new RPointType[order];
    tmpPts = new RPointType[order];

    curTmpChunk = 2;
    origPts = numPts;
    pts.resize( origPts * curTmpChunk );

    MakeReductionMatrices( order );
    MakeReductionMatrices( order-1 );

    endPoints[0] = pts[0];      // HEY - do this for discons too!
    seg = 0;
    curDiscon = 0;

    while (seg < numSegs)
    {
        // If we've hit a discon, slide the indexing forward.

        if (seg == discons[curDiscon])
            curDiscon++;

        endPoints[1] = pts[seg * degree + degree + curDiscon];

        // Convert segment from Bezier to Chebycheff polynomial basis

        tmpPtr = tmpPts;
        for (i = 0; i < order; i++)
        {
            ptsPtr = &pts[seg * degree + curDiscon];

            SetToZero( *tmpPtr );

            for (j = 0; j < order; j++)
            {
                tmp = BezToCheb[order][j][i];
                *tmpPtr += *ptsPtr * tmp;
                ptsPtr++;
            }
            tmpPtr++;
        }

        // check the highest (unused) Chebycheff term to find the error
        err = 0.0;
        for (i = 0; i < dimension; i++)
            err = Max( err, Abs( tmpPts[degree][i] ));

        InternalAssert( (err <= 100),
                        "Degree reduction went bezerk (err > 100?, tolerence too small?)..." );

        if (err == 0.0)
            subDivs = 0;
        else
            subDivs = 1.0 / degree * Log2( err / tolerance );

        if (subDivs > 1.0)
        {
            register long r;
            //
            // If the error was too large, then the segment needs to
            // be split so we can try again on the pieces.
            //

            // May need more control point storage (ugh)
            if ((numSegs + 1) * degree + numDiscons + 1 >= origPts * curTmpChunk)
            {
                curTmpChunk++;
                pts.resize( curTmpChunk * origPts );
            }

            // Make room for the new segment

            for (i = (numPts-1)+degree; i >= (seg + 2) * degree; i--)
            {
                pts[i + curDiscon] = pts[i - degree + curDiscon];
            }

            // Now split the segment (ooh fer lexical scoping...)

            for (i = 0; i < order; i++)
            {
                SetToZero( left[i] );
                SetToZero( right[i] );

                ptsPtr = &(pts[degree * seg + curDiscon]);
                for (r = 0; r <= i; r++)
                {
                    tmp = fchoose(i, r) / Power( 2, i );
                    left[i] += *ptsPtr * tmp;
                    ptsPtr++;
                }

                ptsPtr = &(pts[i + degree * seg + curDiscon]);
                for (r = i; r <= degree; r++)
                {
                    tmp = fchoose(degree-i, degree-r) / Power( 2, degree - i );
                    right[i] += *ptsPtr * tmp;
                    ptsPtr++;
                }
            }

            // Copy subdivided parts back to original array

            for (i = 0; i <= degree; i++)
                pts[i + degree * seg + curDiscon] = left[i];
            for (i = 0; i <= degree; i++)
                pts[degree + i + degree * seg + curDiscon] = right[i];

            numSegs++;
            numPts = degree * numSegs + numDiscons + 1;

            // Update the discons array to account for the new segment
            i = curDiscon;
            while (discons[i] > 0)
            {
                if (discons[i] > seg) // Only downstream
                    discons[i]++;
                i++;
            }

        }
        else
        {
            // Now convert segment back to Bezier, dropping a term on the way

            for (i = 0; i < degree; i++)
            {
                ptsPtr = &pts[seg * (degree-1) + i + curDiscon];
                SetToZero( *ptsPtr );
                tmpPtr = tmpPts;

                for (j = 0; j < degree; j++)
                {
                    tmp = ChebToBez[degree][j][i];
                    *ptsPtr += *tmpPtr * tmp;
                    tmpPtr++;
                }
                ptsPtr++;
            }

            // Force endpoints back to their original locations

            pts[seg * (degree - 1) + (degree - 1) + curDiscon] = endPoints[1];
            seg++;
        }
    }

    pts[0] = endPoints[0];

    *newPts = (degree-1) * numSegs + numDiscons + 1;
    delete [] left;
    delete [] right;
    delete [] tmpPts;
}

//
// Reduce the order by repeatedly dropping the order by one.
//

template <class RPointType, class PointType>
void
IBaseCurve<RPointType, PointType>::approximateLowerOrder( unsigned long newOrder,
                                                          GCoordinate tolerance )
{
    long * discons, numDiscons;

    if (newOrder == fOrder)
        return;

    ParameterException( (newOrder > fOrder), IGraphicException::kAboveRange );
    ParameterException( (newOrder <= 2), IGraphicException::kBelowRange );


    refineToBeziers();              // This makes the seed update calls

    numDiscons = MakeBezierDiscons( &fKnots[0], fOrder, fNumPoints, discons );

    // Find and flag c0 discontinuities in the piecewise Bezier


    tolerance /= fOrder - newOrder;

    while (fOrder > newOrder)
    {
        reduceOneDegree( fPoints,
                           fOrder,
                           fNumPoints,
                           tolerance,
                           discons,
                           numDiscons,
                           &fNumPoints );
        fOrder--;
    }

    // reset the storage to match the actual size used.
    fPoints.resize( fNumPoints );

    // Rebuild the knot vector, taking into account the updated discon's.

    fKnots.resize( fNumPoints + fOrder );

    CreateKnotsFromDiscons( &fKnots[0], fOrder, fNumPoints, discons );

    delete [] discons;
}


#endif // _REDCURVE_INL_
