/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 43 1.17.1.4 source/albert/graph2d/regnextr.hpp, 2d, ioc.v400, 000302 
/*================
||
||  File:   RegionGrafExtractor.h
||
||  What:   Routines which support the area geometry extraction on Windows GDI platform
||
||			IRegionGrafExtractor
||
||			This class constructs a region by extracting basic geometries from TGAarea,
||			drawing them into a path, converting path into a region and combining it with
||			regions created from other geometries.  The region is defined in logical units.
||
||
||  Change History:
||		?           Isai Scheinberg -   Original
||		20 May 96   tt              -   First rewrite
||
*/

#ifndef _REGNEXTR_
#define _REGNEXTR_

#include <ibase.hpp>
#include <i2dghand.hpp>
#include <igarea2d.hpp>
#include <igrtypes.hpp>
#include <grafdevs.hpp>
#include <grafextr.hpp>
#include <cag.hpp>

#if defined(IC_WIN)
#include <gdidev.hpp>
#elif defined(IC_PM)
#include <gpidev.hpp>
#elif defined(IC_MOTIF)
#include <xdevice.hpp>
#endif

#if __IBMCPP__ >= 400
#pragma namemangling(compat)
#endif

class IGLine2D;
class IGRect2D;
class IGPoint2D;
class IGEllipse2D;
class IGPolygon2D;
class IGPolyline2D;
class IGLoop2D;
// class IGGlyphRun;

class ICAGNode;
class ICAGUnary;
class ICAGRoot;

/*================
||
||      IRegionGrafExtractor
||
||      IRegionGrafExtractor is the means by which IGdiDevice class finds what
||      basic geometries were used to construct IGArea
||
*/

#ifndef IC_MOTIF //for WinNT / OS/2

#define HRGN IRegionHandle

class IStreamableRegionHandle : public IRegionHandle {
// This is hack to get IRawArray<I?RegionHandle> to compile because
// the IRawArray likes template classes have streaming operators. This
// class and the way it is used bebow are not supposely streamed and used
// at runtime at all!

public:
	IStreamableRegionHandle(HRGN region) : IRegionHandle(region) {}
    IDataStream& operator>>=(IDataStream& to) const {
    	return to;
    }
    IDataStream& operator<<=(IDataStream& from) {
		return from;
    }
};


class IRegionGrafExtractor: public IGrafExtractor {
public:

#if defined(IC_WIN)
	IRegionGrafExtractor(IGdiDevice* device);
#elif defined(IC_PM)
	IRegionGrafExtractor(IGpiDevice* device);
#endif

	virtual ~IRegionGrafExtractor ();

	// === Area-defining geometries ===

	void extractRect(
		const IGRect2D&,
		ICAGUnary*,
		const IGrafMatrix& cs = IGrafMatrix::identity());
	void extractEllipse(
		const IGEllipse2D&,
		ICAGUnary*,
		const IGrafMatrix& cs = IGrafMatrix::identity());
	void extractPolygon(
		const IGPolygon2D&,
		ICAGUnary*,
		const IGrafMatrix& cs = IGrafMatrix::identity());
	void extractLoop(
		const IGLoop2D&,
		ICAGUnary*,
		const IGrafMatrix& cs = IGrafMatrix::identity());
/*
	void extractGlyphRun(
		const IGGlyphRun&,
		ICAGUnary*,
		const IGrafMatrix& cs = IGrafMatrix::identity());
*/

	int region(IRegionHandle& hRGN) const;

protected:

private:
	IRegionGrafExtractor();
	IRegionGrafExtractor(const IRegionGrafExtractor&);
	IRegionGrafExtractor& operator=(const IRegionGrafExtractor& Src );

	void UpdateRegionArray(ICAGNode* node);

#if defined(IC_WIN)
   IGdiDevice* fGdiDevice;    // referenced IGdiDevice, from which the
#elif defined(IC_PM)
   IGpiDevice* fGdiDevice;    // referenced IGpiDevice, from which the
#endif

	IPresSpaceHandle fHDC;		// deviceContext is referenced

	HRGN fRegion;				// region created from IGArea
	int fRegionType;			// NULLREGION or other type

	IRawArray<IStreamableRegionHandle> fRegionArray;
};

#else //for X only

class IStreamableRegionHandle : public IRegionHandle {
// This is hack to get IRawArray<I?RegionHandle> to compile because
// the IRawArray likes template classes have streaming operators. This
// class and the way it is used bebow are not supposely streamed and used
// at runtime at all!

public:
  IStreamableRegionHandle(Region region) : IRegionHandle(region) {}
    IDataStream& operator>>=(IDataStream& to) const {
      return to;
    }
    IDataStream& operator<<=(IDataStream& from) {
    return from;
    }
};

class IRegionGrafExtractor: public IGrafExtractor {
public:

  IRegionGrafExtractor(IXDevice* device);

  virtual ~IRegionGrafExtractor ();

  // === Area-defining geometries ===

  void extractRect(
    const IGRect2D&,
    ICAGUnary*,
    const IGrafMatrix& cs = IGrafMatrix::identity());
  void extractEllipse(
    const IGEllipse2D&,
    ICAGUnary*,
    const IGrafMatrix& cs = IGrafMatrix::identity());
  void extractPolygon(
    const IGPolygon2D&,
    ICAGUnary*,
    const IGrafMatrix& cs = IGrafMatrix::identity());
  void extractLoop(
    const IGLoop2D&,
    ICAGUnary*,
    const IGrafMatrix& cs = IGrafMatrix::identity());

	int region(IRegionHandle& hRGN) const;

protected:

private:
        void render( ICAGRoot&);

private:
  IRegionGrafExtractor();
  IRegionGrafExtractor(const IRegionGrafExtractor&);
  IRegionGrafExtractor& operator=(const IRegionGrafExtractor& Src );

  void UpdateRegionArray(ICAGNode* node);

  IXDevice* fGdiDevice;   // referenced IXDevice, from which the

  IPresSpaceHandle fHDC;    // deviceContext is referenced

  Region fRegion;       // ptr to region created from IGArea
  int fRegionType;      // NULL REGION or valid type

  IRawArray<IStreamableRegionHandle> fRegionArray;
};

#endif //IC_MOTIF

#if __IBMCPP__ >= 400
#pragma namemangling()
#endif

#endif // _REGNEXTR_
