/*----------------------------------------------------------------------------*/
/* FILE NAME: ibuffer2.cpp                                                    */
/*                                                                            */
/* DESCRIPTION:                                                               */
/*   This file contains the implementation of the IBuffer comparison          */
/*   function "compare".                                                      */
/*                                                                            */
/*                                                                            */
/* COPYRIGHT:                                                                 */
/*   IBM Open Class Library                                                   */
/*   Licensed Materials - Property of IBM                                     */
/*                                                                            */
/*   5645-001                                                                 */
/*   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.           */
/*                                                                            */
/*   US Government Users Restricted Rights - Use, duplication, or             */
/*   disclosure restricted by GSA ADP Schedule Contract with IBM Corp.        */
/*                                                                            */
/*----------------------------------------------------------------------------*/
#ifdef __MVS__
    #pragma csect(code,"IBUFFER2")
    #pragma csect(static,"SBUFFER2")
    #pragma comment (copyright, \
    "Licensed Materials - Property of IBM  5645-001 (C) \
    Copyright IBM Corp. 1992, 1997  All Rights Reserved. \
    US Government Users Restricted Rights - Use, duplication, or \
    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.")
#endif

extern "C"
  {
  #include <string.h>
  #include <stdlib.h>
  }

#include <istring.hpp>     // For I18N support routine

#include <ibuffer.hpp>


#include <iexcbase.hpp>

#include <icconsta.h>

#include <iexcept.hpp>

#include <errno.h>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IBUFFER2_CPP_
  #include <ipagetun.h>
#endif

#ifdef IC_NLS
/*-------------------------- Comparsion ---------------------------*/
static int      /*  LOCALE sensitive compare                       */
  locale_ncmp          ( const void* pBuf1,
                         const void* pArg,
                         size_t      cmpLen);

#endif
/*------------------------------------------------------------------------------
| IBuffer::compare                                                             |
|                                                                              |
| Initialize result to "less than".                                            |
|                                                                              |
| If neither string is null, we compare as many bytes as in the                |
| shorter of the two, using memcmp.                                            |
|                                                                              |
| If these sections compare equal, then, if the lengths are the same,          |
| set the result to "equal."  Otherwise, set the result to either              |
| "less than" or "greater than," depending on which string is longer           |
| (the longer one is "greater").                                               |
|                                                                              |
| If these sections aren't equal, then simply set the result to                |
| "less than" or "greater than" according to the result of the call            |
| to memcmp.                                                                   |
|                                                                              |
| If the receiver isn't null, but the argument is, then return                 |
| "greater than."  If both the receiver and argument are null,                 |
| return "equal."                                                              |
|                                                                              |
| Notes:                                                                       |
|   1. A quick check is made to see if the receiver's buffer and the           |
|      argument coincide (in which case "equal" is returned).                  |
------------------------------------------------------------------------------*/
IBuffer::Comparison IBuffer :: compare ( const void*  pArg,
                                         unsigned     argLen ) const
  {
  Comparison result = lessThan;

  if (!pArg)  result = greaterThan;
  else
    if ( this->contents() == pArg
         &&
         this->length() == argLen )
      result = equal;
    else
      if ( this->length() > 0 )
        if ( argLen > 0 )
         {
          int rc;
#ifdef IC_NLS
          if (IString::isInternationalized())
             rc = locale_ncmp( this->contents(), pArg,
                               ( this->length() < argLen ) ?
                                            this->length() : argLen );
          else  /*  use old semematic   */
#endif
            rc = memcmp( this->contents(), pArg,
                         ( this->length() < argLen ) ?
                                      this->length() : argLen );

          if ( rc == 0 )
            result = ( this->length() == argLen ) ?
                       equal
                     :
                       ( this->length() < argLen ) ?
                                          lessThan : greaterThan;
          else
            result = ( rc < 0 ) ? lessThan : greaterThan;

         }
        else
          result = greaterThan;
      else
        if ( argLen == 0 )
          result = equal;
  return result;
  }

#ifdef IC_NLS
/*------------------------------------------------------------------------------
| int locale_ncmp                                                              |
|                                                                              |
| Initialize rc  to  -1 (lessThan).                                            |
|                                                                              |
| Duplicate buffer1 and buffer2 with null terminator.                          |
| Use strcoll() to compare the 2 duplicated buffers, taking into               |
| consideration of the fact that there might be null characters inside         |
| the buffers itself.                                                          |
|                                                                              |
| Return code from strcoll() is returned to the caller.                        |
|                                                                              |
| Notes:                                                                       |
|   strcoll() could fail due to internal malloc problem.                       |
------------------------------------------------------------------------------*/
static int  locale_ncmp ( const void*  pBuf1,
                          const void*  pBuf2,
                          size_t       cmpLen )
{
  int rc = -1;                     /*  Set rc to lessThan           */
  unsigned sLen = cmpLen + 1;      /*  Length to include null char  */

  char * pW1 = (char *) pBuf1;     /*  Set up initial work buffers pointers  */
  char * pW2 = (char *) pBuf2;
  char * pWBuf1 = NULL;
  char * pWBuf2 = NULL;

  /*  if buffer1 is NOT null ended              */
  if (pW1  &&  (*(pW1 + cmpLen - 1) != '\0'))
   {
     pWBuf1 = new char[sLen];             /*  Get work buffer1 via new  */
     if (pWBuf1)                      /*  if work buffer1 exist     */
      {
        memcpy(pWBuf1, pBuf1, cmpLen);    /*  Duplicate (work) buffer1  */
        *(pWBuf1 + cmpLen) = '\0';        /*  Null terminate it         */
        pW1 = pWBuf1;
      }
     else
       pW1 = NULL;                        /*  Memory storage problem    */
   }

                  /*  if work buffer1 exist and buffer2 is NOT null ended  */
  if (pW1  &&  pW2  &&  (*(pW2 + cmpLen - 1) != '\0'))
   {
     pWBuf2 = new char[sLen];                /*  Get work buffer2 via new  */
     if (pWBuf2)                      /*  if work buffer1 exist     */
      {
        memcpy(pWBuf2, pBuf2, cmpLen);       /*  Duplicate (work) buffer2  */
        *(pWBuf2 + cmpLen) = '\0';           /*  Null terminate it         */
        pW2 = pWBuf2;
      }
     else
      {
        pW2 = NULL;                          /*  Memory storage problem    */
                          /*  if work buffer1 obtained via new, delete it  */
        delete [] pWBuf1;
      }
   }


  if (pW1  &&  pW2)                  /*  if both work buffers pointer exist  */
   {
     unsigned iBuf1 = 0,
              iBuf2 = 0,
              iX1, iX2;

     sLen = 0;                       /*  Clear the length  */

     for (;;)
      {
        iX1 = 0;                     /*  Reset indexes for NULLs found  */
        iX2 = 0;

        while ('\0' == *pW1)     /*  Skip over the possible NULL characters  */
         {
           ++pW1;
           ++iX1;
           ++iBuf1;
           if (iBuf1 == cmpLen)  break;     /*  Done, end of buffer          */
         }

        while ('\0' == *pW2)     /*  Skip over the possible NULL characters  */
         {
           ++pW2;
           ++iX2;
           ++iBuf2;
           if (iBuf2 == cmpLen)  break;     /*  Done, end of buffer          */
         }

        if (iX1 != iX2)       /*  Not equal number of NULL characters found  */
         {
           if (iX1 < iX2)  rc = 1;        /*  Indicate pBuf1 > pBuf2         */
           break;                         /*  Done, no need to continue      */
         }

        if (iBuf1 == cmpLen)     /*  if end of first buffer  */
         {
           if (iX1 == cmpLen)       /*  if both buffers contain only NULL  */
              rc = 0;               /*  charcaters                         */
           break;                /*  Done with compare       */
         }

        errno = 0;                     /*  Clear errno                  */
        rc = strcoll(pW1, pW2);        /*  Collate the 2 strings        */
        if (errno || rc)  break;       /*  if strcoll failed            */

        sLen = strlen(pW1) + 1;        /*  How far did the compare go?  */
                                       /*  including the NULL character */
        iBuf1 += sLen;
        if (iBuf1 >= cmpLen)  break;   /*  Done, no need to continue    */
        iBuf2 += sLen;
        pW1 += sLen;                   /*  Advance the pointer within   */
        pW2 += sLen;                   /*  buffers                      */
      }                       /*  loop through possible strings in the  */
                              /*  buffers                               */

                          /*  Delete the work buffers created via new  */
     delete [] pWBuf1;
     delete [] pWBuf2;

     if (errno)      /*  Bad rc from strcoll() call  */
      {
        char cMsg[20];
        int save_errno = errno;
        sprintf(cMsg, "strcoll errno = %d\n", errno);
#ifdef IC_MOTIF
  IInvalidRequest exception(cMsg, save_errno, IException::recoverable);
  ITHROW( exception );
#else
        ITHROWCLIBERROR( cMsg, IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
#endif
      }

   }

 return rc;
}


#endif
