/*----------------------------------------------------------------------------*/
/* FILE NAME: idbcsb7.cpp                                                     */
/*                                                                            */
/* DESCRIPTION:                                                               */
/*   This file contains the implementation of the editing functions of        */
/*   the class IDBCSBuffer.                                                   */
/*                                                                            */
/*                                                                            */
/* COPYRIGHT:                                                                 */
/*   IBM Open Class Library                                                   */
/*   Licensed Materials - Property of IBM                                     */
/*                                                                            */
#ifdef __MVS__
/*   5645-001                                                                 */
/*   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.           */
#endif
#ifdef __OS400__
/*   ????-???                                                                 */
/*   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.           */
#endif
#ifdef __OS2__
/*   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.           */
#endif
#ifdef __WINDOWS__
/*   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.           */
#endif
#ifdef _AIX
/*   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.           */
#endif
#ifdef _sparc
/*   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.           */
#endif
#ifdef __sparc
/*   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.           */
#endif
/*                                                                            */
/*   US Government Users Restricted Rights - Use, duplication, or             */
/*   disclosure restricted by GSA ADP Schedule Contract with IBM Corp.        */
/*                                                                            */
/*----------------------------------------------------------------------------*/
#ifdef __MVS__
    #pragma csect(code,"IDBCSB7")
    #pragma csect(static,"SDBCSB7")
    #pragma comment (copyright, \
    "Licensed Materials - Property of IBM  5645-001 (C) \
    Copyright IBM Corp. 1992, 1997  All Rights Reserved. \
    US Government Users Restricted Rights - Use, duplication, or \
    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.")
#endif

/*==============================================================================
 Notes:
   1. Each of these functions returns a pointer to an IDBCSBuffer object.
      The resulting buffer contains the edited result.  It may or not
      be a different buffer than the receiver, depending on the nature
      of the change and the use count of the receiver.
   2. It is presumed that the caller has properly adjusted the use
      count for the receiver prior to calling any of these functions
      (that is, the use count is accurate).
   3. It is presumed that the caller will attach the resulting buffer
      to an IString so that the built-in use count in the returned
      buffer is accurate.
   4. Typical usage is as follows:
        ...
        // Preserve old contents.
        IString old( aString.pBuffer );
        ...
        // Reset aString to point to (possibly) new edited buffer:
        aString.pBuffer = aString.pBuffer->xxxxxxx();
        ...
==============================================================================*/

#include <istring.hpp>

#include <idbcsbuf.hpp>

#include <istrtest.hpp>

#include <ctype.h>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IDBCSB7_CPP_
  #include <ipagetun.h>
#endif

#include "inlsutil.hpp"


/*------------------------------------------------------------------------------
| IDBCSBuffer::center                                                          |
|                                                                              |
| If the requested length is equal to the length of the receiver,              |
| then the receiver is returned unmodified.  Otherwise, the                    |
| receiver's length is reset to the argument length and space is               |
| allocated for the centered buffer.  The new buffer is built                  |
| one of two ways, depending on whether the new length is larger               |
| or smaller than the original length of the receiver's buffer.                |
|                                                                              |
| In the former case, the amount of padding required before and                |
| after the centered string is calculated.  Then, the new string               |
| is filled with pad characters, the old contents are copied, and              |
| then the remainder is filled with more pad characters.                       |
|                                                                              |
| In the latter case, a substring of the receiver (the center portion          |
| of the requested length) is copied to the result buffer.                     |
------------------------------------------------------------------------------*/
IDBCSBuffer * IDBCSBuffer :: center ( unsigned newLength,
                                  char     padCharacter )
{
  IDBCSBuffer *result = this;
#ifdef IC_NLS
 if (IString::isInternationalized())
 {
   if ( newLength != this->length() )
   {
     //-----------------------------------------------------------------------
     // Initialize parameters to likely values:
     //-----------------------------------------------------------------------
     unsigned prefix       = 0,
              suffix       = 0,
              fromReceiver = this->length(),
              startPos     = 1;

     //-----------------------------------------------------------------------
     // Adjust initialization parameters:
     //-----------------------------------------------------------------------
     if ( newLength > this->length() )
     {
       prefix = ( newLength - this->length() ) / 2;
       suffix = newLength - this->length() - prefix;

     //-----------------------------------------------------------------------
     // Allocate space and copy receiver to middle:
     //-----------------------------------------------------------------------
     result = (IDBCSBuffer *)this->newBuffer( 0, prefix,
                               this->contents() + startPos - 1, fromReceiver,
                               0, suffix,
                               padCharacter );

     }
     else
     {
       fromReceiver = newLength;
       startPos     = ( this->length() - fromReceiver ) / 2 + 1;

       // -- Convert MB string to wchar string buffer
       // -- Find correspondig start position index in the IWBuffer
       // -- Convert IWBuffer to MB string starting from the wc index


       int      wcStartPos;

       IWBuffer    srcWBuf(this->contents(), this->length() );

       wcStartPos = findCharNum(this->contents(),this->length(), startPos );

       char  *pMBCS = new char[newLength+1];
       int       mbDestLen;

       mbDestLen  =  cvtWCS2MBS(  (wchar_t *) srcWBuf + wcStartPos - 1,
                                  srcWBuf.charNum() - wcStartPos,
                                  pMBCS,
                                  newLength,
                                  1, padCharacter );



     //-----------------------------------------------------------------------
     // Allocate space and copy receiver to middle:
     //-----------------------------------------------------------------------
     result = (IDBCSBuffer *)this->newBuffer( 0, prefix,
                               pMBCS, newLength,
                               0, suffix,
                               padCharacter );
     delete[] pMBCS;

     }

   }
   else
     this->addRef();

 } else {   // invoke old version of code
#endif
   if ( newLength != this->length() )
   {
     //-----------------------------------------------------------------------
     // Initialize parameters to likely values:
     //-----------------------------------------------------------------------
     unsigned prefix       = 0,
              suffix       = 0,
              fromReceiver = this->length(),
              startPos     = 1;

     //-----------------------------------------------------------------------
     // Adjust initialization parameters:
     //-----------------------------------------------------------------------
     if ( newLength > this->length() )
     {
       prefix = ( newLength - this->length() ) / 2;
       suffix = newLength - this->length() - prefix;
     }
     else
     {
       fromReceiver = newLength;
       startPos     = ( this->length() - fromReceiver ) / 2 + 1;
       //---------------------------------------------------------------------
       // If the byte at startPos is the second byte of a DBCS character, then
       // the startPos is decremented so that the string will get cut from the
       // previous byte.
       //---------------------------------------------------------------------
       while ( startPos > 1 && this->charType( startPos ) > IStringEnum::mbcs1 )
         startPos--;
     }

     //-----------------------------------------------------------------------
     // Allocate space and copy receiver to middle:
     //-----------------------------------------------------------------------
     result = (IDBCSBuffer *)this->newBuffer( 0, prefix,
                               this->contents() + startPos - 1, fromReceiver,
                               0, suffix,
                               padCharacter );

     //-----------------------------------------------------------------------
     // If the last byte of the new string is the first byte of a DBCS
     // character, then that byte gets set to the padCharacter.
     //-----------------------------------------------------------------------
     if ( startPos + fromReceiver <= this->length () )
     {
       IStringEnum::CharType typeOfFirstAfterSubstring =
         this->charType ( startPos + fromReceiver );
       unsigned len = result->length ();
       if ( typeOfFirstAfterSubstring > IStringEnum::mbcs1 )
         for ( unsigned i = 1; i < typeOfFirstAfterSubstring && i <= len; i++)
           result->contents()[len - i] = padCharacter;
     }
   }
   else
     this->addRef();
#ifdef IC_NLS
 }
#endif

  return result;
}

// Class to maintain a linked list of occurrences of patterns.
struct Occur {
  Occur(unsigned u)
    : pNext(0), pos(u) {}
  ~Occur() { delete pNext; };

  // Pointer to next occurrence.
  Occur *pNext;
  // Index of this occurrence.
  unsigned pos;
};

#ifdef IC_NLS

/*------------------------------------------------------------------------------
| IDBCSBuffer::change                                                          |
|                                                                              |
| This function is implemented by:                                             |
|   1. If it is not EBCDIC platform or it is not internationalized, delegate   |
|      the work to IDBCSBuffer::change(); otherwise do the following,          |
|   2. Locate the character position to start searching from 'startPos'.       |
|   3. Convert receiver and pattern strings to wide strings.                   |
|   4. Make a pass through of the receiver, noting the number of occurrences   |
|      of the pattern and the character position of each.                      |
|   5. Convert replacement string to wide string.                              |
|   6. Calculate the new length of the destination wide string, based on the   |
|      relative number of characters of the pattern and replacement wide       |
|      strings.                                                                |
|   7. Fill the new wide buffer, alternating pieces of the old buffer with the |
|      new replacement string.                                                 |
|   8. Convert the resulting wide string back to MBCS string.                  |
|                                                                              |
| The information collected during the first pass through of the receiver is   |
| maintained in a linked list of Occur objects. Each has a data member to      |
| record the character position in the receiver of the pattern occurrence and  |
| a pointer to the Occur object for the next occurrence of the pattern.        |
------------------------------------------------------------------------------*/
IDBCSBuffer * IDBCSBuffer :: change ( const char *pPattern,
                                  unsigned    patternLen,
                                  const char *pReplacement,
                                  unsigned    replacementLen,
                                  unsigned    startPos,
                                  unsigned    numChanges )
{
#ifdef IC_MVS400
  if (IString::isInternationalized()) {

    IDBCSBuffer *result = this;

    if (startPos == 0) startPos = 1;

    // Locate the character position to start searching
    unsigned numPos = findNextCharNum(contents(), length(), startPos);

    if (numPos) {
      unsigned count = 0;           // Record the number of occurrences

      // Initialize occurrence chain with a dummy entry at head.
      // 'pOccur' is used to step through the chain; it points
      // to the last occurrence during search, current occurrence
      // during pattern replacement. Note that when 'head' is
      // destructed, it will delete all the other Occur objects.
      Occur head(0);
      Occur *pOccur = &head;
      IWBuffer receiverWBuf(this->contents(), this->length());
      IWBuffer patternWBuf(pPattern, patternLen);

      // Search for occurrences till all we need are found...
      while (count < numChanges) {
        if ((numPos = receiverWBuf.positionOf(patternWBuf, numPos)) != 0) {
          // Occurrence found ...
          ++count;
          pOccur = pOccur->pNext = new Occur(numPos);
          numPos += patternWBuf.charNum();
        } else {
          // No more occurrences ...
          break;
        }
      }

      // Do we find some match?
      if (count != 0) { // Yes...
        // Add dummy Occur at the end to ensure tail end of receiver
        // gets copy to destination.
        pOccur->pNext = new Occur(receiverWBuf.charNum() + 1);
        pOccur = head.pNext;
        IWBuffer replacementWBuf(pReplacement, replacementLen);
        unsigned newWLen = receiverWBuf.charNum();

        // Determine the size of the destination wide string.
        if (patternWBuf.charNum() > replacementWBuf.charNum()) {
          newWLen -= checkMultiplication(
                       patternWBuf.charNum() - replacementWBuf.charNum(),
                       count);

        } else if (patternWBuf.charNum() < replacementWBuf.charNum()) {
          newWLen = checkAddition(newWLen,
                      checkMultiplication(
                        replacementWBuf.charNum() - patternWBuf.charNum(),
                        count));
        }

        // Allocate memory for the destination wide string and then copy
        // the portion from receiver wide string up to the first occurrence
        // of pattern.
        wchar_t *pWSource  = receiverWBuf;
        wchar_t *pDestWBuf = new wchar_t[newWLen];
        wchar_t *pWDest    = pDestWBuf;
        unsigned copyLen   = pOccur->pos - 1;
        unsigned previousPos;

        wcsncpy(pWDest, pWSource, copyLen);

        pWSource += copyLen;
        pWDest   += copyLen;

        // For each occurrence, replace pattern and copy up to the next
        // occurrence.
        while (count--) {
          previousPos = pOccur->pos;
          pOccur = pOccur->pNext;
          wcsncpy(pWDest, replacementWBuf, replacementWBuf.charNum());
          pWSource += patternWBuf.charNum();
          pWDest   += replacementWBuf.charNum();
          copyLen = (pOccur->pos - previousPos) - patternWBuf.charNum();
          wcsncpy(pWDest, pWSource, copyLen);
          pWSource += copyLen;
          pWDest   += copyLen;
        }

        char *pDestMBBuf = new char[newWLen * MB_CUR_MAX];
        unsigned newMBLen = cvtWCS2MBS(pDestWBuf, newWLen, pDestMBBuf);

        result = (IDBCSBuffer *)this->newBuffer(pDestMBBuf, newMBLen);


        delete[] pDestMBBuf;
        delete[] pDestWBuf;

      } else { // Cannot find a match, no change.
        this->addRef();
      }

    } else { // No character to start search, no change.
      this->addRef();
    }

    return result;

  } else { // !internationalized
#endif

    return (IDBCSBuffer *)IBuffer::change(pPattern, patternLen,
                                          pReplacement, replacementLen,
                                          startPos, numChanges);
#ifdef IC_MVS400
  }
#endif

}
#endif

/*------------------------------------------------------------------------------
| IDBCSBuffer::insert                                                          |
|                                                                              |
| This function accepts as arguments:                                          |
|   - pointer to string to be inserted                                         |
|   - length  of string to be inserted                                         |
|   - position at which to insert the string                                   |
|   - pad character.                                                           |
|                                                                              |
| It returns the receiver unmodified if both the string to be inserted is      |
| null and the index at which it is to be inserted is not off the end of       |
| the receiver buffer.                                                         |
|                                                                              |
| Otherwise, space is allocated for the combined buffers.  The insertion       |
| is accomplished in four steps:                                               |
|                                                                              |
|   1. The portion of the receiver prior to the index at which                 |
|      the new string is to be inserted is copied (may be null).               |
|   2. If the receiver isn't this long, then it is padded out                  |
|      to the insert index using the pad character.                            |
|   3. The to-be-inserted string is copied (may be null).                      |
|   4. The portion of the receiver following the index at which                |
|      the new string is to be inserted is copied (may be null).               |
|                                                                              |
| All this is accomplished with a single call to newBuffer().                  |
------------------------------------------------------------------------------*/
IDBCSBuffer * IDBCSBuffer :: insert ( const char *pInsert,
                                  unsigned    len,
                                  unsigned    index,
                                  char        padCharacter )
{
  IDBCSBuffer *result = this;

#ifdef IC_NLS
 if (IString::isInternationalized())
 {
   if ( len != 0 || index > this->length() )
   {

     unsigned    len1 = ( this->length() < index ) ?
                          this->length()
                          :
                          index,
                 len2,
                 len3;
     const char *p1   = this->contents(),
                *p2,
                *p3;

     //-----------------------------------------------------------------------
     // See if we need to pad the receiver:
     //-----------------------------------------------------------------------
     if ( index > this->length() )
     {
       //---------------------------------------------------------------------
       // Second initializer is pad characters:
       //---------------------------------------------------------------------
       len2 = index - this->length(),
       p2   = 0,
       //---------------------------------------------------------------------
       // Third initializer is the string to be inserted:
       //---------------------------------------------------------------------
       len3 = len,
       p3   = pInsert;

       //----------------------------------------------------------------------
       // Return new buffer:
       //----------------------------------------------------------------------
       result = (IDBCSBuffer *)this->newBuffer( p1, len1,
                                 p2, len2,
                                 p3, len3,
                                 padCharacter );
     }
     else
     {

       IWBuffer    srcWBuf(p1, this->length() );

       wchar_t   *pwcIns = (wchar_t *)malloc (sizeof (wchar_t) * len  +1 );
       int       wcInsLen;

       IWBuffer    insWBuf(pInsert, len );

       int  wcIndex = findCharNum(p1, this->length(), index);


       IWBuffer    destWBuf(NULL, len + this->length() );

       int retwcDestLen =  insWCS2WCS( (wchar_t *) srcWBuf,
                                       srcWBuf.charNum(),
                                       (wchar_t *) insWBuf,
                                       insWBuf.charNum(),
                                       wcIndex,
                                       (wchar_t *) destWBuf);


        int  mbDestLen = len + this->length();
        char *pmbDest =  new char[ mbDestLen + 1];
        int  retmbDestLen;
        int  wcStartPos = 0;


       retmbDestLen  =  cvtWCS2MBS(  (wchar_t *) destWBuf,
                                     retwcDestLen,
                                     pmbDest,
                                     mbDestLen,
                                     1, padCharacter);

       //----------------------------------------------------------------------
       // Return new buffer:
       //----------------------------------------------------------------------
       result = (IDBCSBuffer *)this->newBuffer(pmbDest, retmbDestLen,
                                NULL, 0, NULL, 0,
                               padCharacter );
       delete[] pmbDest;


     }

   }
   else
     this->addRef();

 } else {   // invoke old version of code
#endif

   if ( len != 0 || index > this->length() )
   {
     //-----------------------------------------------------------------------
     // If the index is somewhere in the existing string, then
     // check to see if the insertion point is between bytes of a DBCS
     // character.  If so, then move the insertion point back one.
     //-----------------------------------------------------------------------
     if ( index < this->length() )
       while ( index > 0 && this->charType( index + 1 ) > IStringEnum::mbcs1 )
         index--;

     unsigned    len1 = ( this->length() < index ) ?
                          this->length()
                          :
                          index,
                 len2,
                 len3;
     const char *p1   = this->contents(),
                *p2,
                *p3;

     //-----------------------------------------------------------------------
     // See if we need to pad the receiver:
     //-----------------------------------------------------------------------
     if ( index > this->length() )
     {
       //---------------------------------------------------------------------
       // Second initializer is pad characters:
       //---------------------------------------------------------------------
       len2 = index - this->length(),
       p2   = 0,
       //---------------------------------------------------------------------
       // Third initializer is the string to be inserted:
       //---------------------------------------------------------------------
       len3 = len,
       p3   = pInsert;
     }
     else
     {
       //---------------------------------------------------------------------
       // Second initializer is the string to be inserted:
       //---------------------------------------------------------------------
       len2 = len,
       p2   = pInsert,
       //---------------------------------------------------------------------
       // Third initializer is the rest of the receiver:
       //---------------------------------------------------------------------
       len3 = this->length() - index,
       p3   = this->contents() + index;
     }

     //-----------------------------------------------------------------------
     // Return new buffer:
     //-----------------------------------------------------------------------
     result = (IDBCSBuffer *)this->newBuffer( p1, len1,
                               p2, len2,
                               p3, len3,
                               padCharacter );
   }
   else
     this->addRef();
#ifdef IC_NLS
 }
#endif

  return result;
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::leftJustify                                                     |
|                                                                              |
| If the new length is the same as the receiver's current length,              |
| then the receiver is returned unmodified.                                    |
|                                                                              |
| Otherwise, the new buffer is generated via an invocation of the              |
| newBuffer() function.  The new contents are built from the                   |
| receiver (either its full length or truncated at the new length              |
| if that is shorter) and a string of pad characters (of sufficient            |
| length to fill the result out to the specified length).                      |
------------------------------------------------------------------------------*/
IDBCSBuffer * IDBCSBuffer :: leftJustify ( unsigned newLength,
                                       char     padCharacter )
{
  IDBCSBuffer *result = this;
  unsigned len2    = 0;

#ifdef IC_NLS
 if (IString::isInternationalized())
 {
   if ( newLength > this->length() ) {

     result = (IDBCSBuffer *)this->newBuffer( this->contents(), this->length(),
                               0, newLength - this->length(),
                               0, 0,
                               padCharacter );

   } else if ( newLength < this->length() ) {

     // -- Convert MB string to wchar string buffer
     // -- Find correspondig start position index in the IWBuffer
     // -- Convert IWBuffer to MB string starting from the wc index

     IWBuffer    srcWBuf(this->contents(), this->length() );

     char  *pMBCS = new char[newLength+1];
     int       retmbDestLen;
     int       wcStartPos = 0;

     retmbDestLen  =  cvtWCS2MBS( (wchar_t *) srcWBuf,
                                  srcWBuf.charNum(),
                                  pMBCS,
                                  newLength,
                                  1, padCharacter);

     //--------------------------------------------------------------------
     // Allocate space and copy receiver to middle:
     //--------------------------------------------------------------------

     result = (IDBCSBuffer *)this->newBuffer( pMBCS, newLength,
                               0, 0 ,
                               0, 0,
                               padCharacter );
     delete[] pMBCS;

   } else { // newLength == this->length()

     this->addRef();
   }

 } else {   // invoke old version of code
#endif
    if ( newLength != this->length() )
    {
      if ( newLength > this->length() )
        len2 = newLength - this->length();

      result = (IDBCSBuffer *)this->newBuffer( this->contents(),
                                ( this->length() < newLength ) ?
                                  this->length()
                                  :
                                  newLength,
                                0, len2,
                                0, 0,
                                padCharacter );
      //-----------------------------------------------------------------------
      // If the last byte of the resulting string is DBCS1, then replace it
      // with the padCharacter.
      //-----------------------------------------------------------------------
        if ( newLength < this->length () )
        {
          IStringEnum::CharType typeOfFirstAfterSubstring =
            this->charType ( newLength + 1 );
          if ( typeOfFirstAfterSubstring > IStringEnum::mbcs1 )
            for ( unsigned i = 1;
                  i < typeOfFirstAfterSubstring && i <= newLength;
                  i++)
              result->contents()[newLength - i] = padCharacter;
        }
    }
    else
      this->addRef();
#ifdef IC_NLS
 }
#endif

  return result;
}



/*------------------------------------------------------------------------------
| IDBCSBuffer::lowerCase                                                       |
|                                                                              |
| Starting at the beginning of the buffer, scan for the next capital           |
| letter and if found, convert it to lower case.  Quit as soon as              |
| the search for a capital (via indexOfAnyOf()) fails.                         |
------------------------------------------------------------------------------*/
IDBCSBuffer * IDBCSBuffer :: lowerCase ( )
{
  IDBCSBuffer *result = this;

#ifdef IC_NLS
  if (IString::isInternationalized())
  {
    if (this->length() > 0)
    {
       IWBuffer wBufferThis(this->contents(), this->length());
       int iNumWChar = wBufferThis.charNum();
       for (int iStringIdx = 1; iStringIdx <= iNumWChar; iStringIdx++)
          wBufferThis[iStringIdx] = towlower(wBufferThis[iStringIdx]);
       char *pcChar = new char[iNumWChar * MB_CUR_MAX];
       int iNumChar = cvtWCS2MBS(wBufferThis, iNumWChar, pcChar);
       result = (IDBCSBuffer *)this->newBuffer(pcChar, iNumChar);
       ::delete [] pcChar;
    }
    else
      this->addRef();
  } else {
#endif
    if ( this->useCount() > 1 )
      result = (IDBCSBuffer *)this->newBuffer( this->contents(),
                                               this->length() );
    else
      this->addRef();
    char *p = result->contents();
	char *pEnd = p + result->length();
    while ( p < pEnd ) {
	  *p = tolower(*p);
	  ++p;
	}
#ifdef IC_NLS
  }
#endif

  return result;
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::overlayWith                                                     |
|                                                                              |
| This member function overlayWith() returns the receiver                      |
| unmodified if both the length of the overlay string is zero and              |
| the overlay occurs within the current receiver (not off the end).            |
|                                                                              |
| Otherwise, space is allocated for the combined buffers.  The                 |
| overlay is accomplished in four steps:                                       |
|                                                                              |
|   1. The portion of the receiver prior to the index at which                 |
|      the new string is to be overlayed is copied (may be null).              |
|   2. If the receiver isn't long enough, then it is padded out                |
|      to this index using the pad character.                                  |
|   3. The overlay string is copied (may be null).                             |
|   4. The portion of the receiver following the portion of it                 |
|       which was overlaid is copied (may be null).                            |
------------------------------------------------------------------------------*/
IDBCSBuffer * IDBCSBuffer :: overlayWith ( const char *pOverlay,
                                       unsigned    len,
                                       unsigned    index,
                                       char        padCharacter )
{
  IDBCSBuffer *result = this;

  if ( len != 0 || index > this->length() )
  {
    //------------------------------------------------------------------------
    // Treat index 0 like index 1:
    //------------------------------------------------------------------------
    if ( index == 0 ) index = 1;
#ifdef IC_NLS
    if (IString::isInternationalized())
    {
      char *pFinal;
      unsigned iFinalLength;
      if (index <= this->length())
      {
         char *pFirst = new char[index - 1];
         getSubString(this->contents(), this->length(), 1,
                      padCharacter, pFirst, index - 1);
         if (index + len <= this->length())
         {
            char *pSecond = new char[this->length()];
            unsigned iRemain = getRemainString(this->contents(),
                                               this->length(),
                                               index + len,
                                               padCharacter,
                                               pSecond);
            iFinalLength = index + len + iRemain - 1;
            pFinal = new char[iFinalLength];
            memcpy(pFinal, pFirst, index - 1);
            memcpy(pFinal + index - 1, pOverlay, len);
            memcpy(pFinal + index + len - 1, pSecond, iRemain);
            delete [] pSecond;
         }
         else
         {
            iFinalLength = index + len - 1;
            pFinal = new char[iFinalLength];
            memcpy(pFinal, pFirst, index - 1);
            memcpy(pFinal + index - 1, pOverlay, len);
         }
         delete [] pFirst;
      }
      else
      {
         iFinalLength = index + len - 1;
         pFinal = new char[iFinalLength];
         memset(pFinal, padCharacter, iFinalLength);
         memcpy(pFinal, this->contents(), this->length());
         memcpy(pFinal + index - 1, pOverlay, len);
      }
      result = (IDBCSBuffer *)this->newBuffer(pFinal, iFinalLength);
      delete [] pFinal;
    } else {
#endif
      unsigned  len1 = ( this->length() < index - 1 ) ?
                         this->length()
                         :
                         index - 1,
                len2,
                len3;
      const char *p1 = this->contents(),
                 *p2,
                 *p3;

    //------------------------------------------------------------------------
    // Rest is pad+overlay or overlay+rest:
    //------------------------------------------------------------------------
      if ( len1 < index - 1 )
      {
      //----------------------------------------------------------------------
      // Second initializer is pad characters:
      //----------------------------------------------------------------------
        len2 = ( index - 1 ) - len1;
        p2   = 0;
      //----------------------------------------------------------------------
      // Third initializer is the string to be overlayed:
      //----------------------------------------------------------------------
        len3 = len;
        p3   = pOverlay;
      }
      else
      {
      //----------------------------------------------------------------------
      // Second initializer is the string to be overlayed:
      //----------------------------------------------------------------------
        len2 = len;
        p2   = pOverlay;
      //----------------------------------------------------------------------
      // Third initializer is the rest of the receiver:
      //----------------------------------------------------------------------
        if ( this->length() >= index + len )
          len3 = this->length() - index - len + 1,
          p3   = this->contents() + index + len - 1;
        else
          len3 = 0,
          p3   = 0;
      }

    //------------------------------------------------------------------------
    // Obtain new buffer:
    //------------------------------------------------------------------------
      result = (IDBCSBuffer *)this->newBuffer( p1, len1,
                                p2, len2,
                                p3, len3,
                                padCharacter );

    //------------------------------------------------------------------------
    // If the byte at (index - 1) is the first byte of a DBCS character,
    // then that byte should be replaced with padCharacter.
    // If the byte at (index + len) is the second byte of a DBCS
    // character, then that byte should be replaced with padCharacter.
    //------------------------------------------------------------------------
      if ( index <= this->length () )
      {
        IStringEnum::CharType typeOfFirstOverlayed = this->charType ( index );
        if ( typeOfFirstOverlayed > IStringEnum::mbcs1 )
          for ( unsigned i = 1; i < typeOfFirstOverlayed && i < index; i++)
            result->contents()[index - 1 - i] = ' ';
      }

      for ( unsigned i = 0; index + len + i <= this->length (); i++ )
        if ( this->charType( index + len + i ) > IStringEnum::mbcs1 )
          result->contents()[index + len - 1 + i] = padCharacter;
        else
          break;
#ifdef IC_NLS
    }
#endif

  }
  else
    this->addRef();

  return result;
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::reverse                                                         |
|                                                                              |
| If the string is not null, reverse its bytes.  Since the IDBCSBuffer is      |
| not guaranteed to not have null characters, we can't use strrev()            |
| and instead roll our own.                                                    |
------------------------------------------------------------------------------*/
IDBCSBuffer * IDBCSBuffer :: reverse ( )
{
  IDBCSBuffer *result = this;
  if ( this->length() > 1 )
  {
#ifdef IC_NLS
    if (IString::isInternationalized())
    {
      IWBuffer wBufferThis(this->contents(), this->length());
      wchar_t *pwStart = wBufferThis;
      wchar_t *pwEnd = wBufferThis + wBufferThis.charNum() - 1;
      wchar_t wSwap;
      while (pwStart < pwEnd)
      {
        wSwap = *pwEnd;
        *pwEnd-- = *pwStart;
        *pwStart++ = wSwap;
      }
      // the original length should be large enough
      char *pcChar = new char[this->length()];
      int iNumChar = cvtWCS2MBS(wBufferThis, wBufferThis.charNum(), pcChar);
      result = (IDBCSBuffer *)this->newBuffer( pcChar, iNumChar);
      delete [] pcChar;
    } else {
#endif
      result = (IDBCSBuffer *)this->newBuffer( this->contents(),
                                               this->length() );
      char *p0 = this->contents();
      char *p2 = result->contents() + result->length() - 1;
      char *pend = this->contents() + this->length();
      size_t len1, len2;
      while (p0 < pend)
      {
         len1 = charLength(p0);
         for (len2 = len1; len2 > 0; len2--)
         {
            *p2 = *(p0 + len2 - 1);
            p2--;
         }
         p0 += len1;
      }
#ifdef IC_NLS
    }
#endif
  }
  else
    this->addRef();

  return result;
}

/*------------------------------------------------------------------------------
| IDBCSBuffer::strip                                                           |
|                                                                              |
| Accepts a string of characters to be stripped. A "mode" indicator specifies  |
| whether leading, trailing, or, both leading and trailing characters meeting  |
| the criteria are to be stripped.                                             |
|                                                                              |
| This method is implemented with the basic algorithm:                         |
|    1. Find the first character that fails the criteria                       |
|    2. Find the last character that fails the criteria                        |
|    3. Allocate space for and copy the intervening characters                 |
|    4. Return the resulting buffer.                                           |
------------------------------------------------------------------------------*/
IDBCSBuffer * IDBCSBuffer :: strip ( const char             *pChars,
                                 unsigned                len,
                                 IStringEnum::StripMode  mode )
  {
  IDBCSBuffer *result = this;

  if ( this->length() != 0 )
  {
#ifdef IC_NLS
    if (IString::isInternationalized())
    {
      IWBuffer wBufferThis(this->contents(), this->length());
      IWBuffer wRemoveChar(pChars, len);
      int iStripLen = wRemoveChar.charNum();
      int iInLen = wBufferThis.charNum();
      int iStart, iSearch, iEnd;
      iStart = 1;
      iEnd = iInLen;
      if (mode != IStringEnum::trailing)
      {
         for (iStart = 1; iStart <= iInLen; iStart++)
         {
            for (iSearch = 1; iSearch <= iStripLen; iSearch++)
            {
               if (wBufferThis[iStart] == wRemoveChar[iSearch])
                  break;
            }
          if (iSearch > iStripLen)
             break;
         }
      }
      if (mode != IStringEnum::leading)
      {
        for (iEnd = iInLen; iEnd > 0; iEnd--)
        {
          for (iSearch = 1; iSearch <= iStripLen; iSearch++)
          {
             if (wBufferThis[iEnd] == wRemoveChar[iSearch])
                break;
          }
          if (iSearch > iStripLen)
             break;
        }
      }
      if (iEnd >= iStart)
      {
        char *pcChar = new char[(iEnd - iStart + 1) * MB_CUR_MAX];
        int iNumChar = cvtWCS2MBS(wBufferThis + iStart - 1, iEnd - iStart + 1,
                                  pcChar);
        result = (IDBCSBuffer *)this->newBuffer( pcChar, iNumChar);
        delete [] pcChar;
      }
      else
      {
        result = (IDBCSBuffer *)this->null();
        result->addRef();
      }
    } else {
#endif
       unsigned start = 1,
                stop  = this->length();
       if ( mode != IStringEnum::trailing )
         start = this->indexOfAnyBut( pChars, len, 1 );
       if ( mode != IStringEnum::leading
            &&
            start != 0 )
         {
         stop = this->lastIndexOfAnyBut( pChars, len, this->length() );
      // If last char to keep is MBCS, keep all bytes.
         if (stop > 0)
            stop += charLength ( stop ) - 1;
         }
       if ( start != 1
            ||
            stop != this->length() )
         {
         if ( start != 0 )
           result = (IDBCSBuffer *)this->newBuffer(this->contents() + start-1,
                                                   stop - start + 1 );
         else
           {
           result = (IDBCSBuffer *)this->null();
           result->addRef();
           }
         }
       else
         this->addRef();
#ifdef IC_NLS
    }
#endif
  }
  else
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::strip                                                           |
|                                                                              |
| This method uses an IStringTest object to determine which characters are to  |
| be stripped. A "mode" indicator specifies whether leading, trailing, or,     |
| both leading and trailing characters meeting the criteria are to be          |
| stripped.                                                                    |
|                                                                              |
| 1. Find the first character that fails the criteria                          |
| 2. Find the last character that fails the criteria                           |
| 3. Allocate space for and copy the intervening characters                    |
| 4. Return the resulting buffer.                                              |
------------------------------------------------------------------------------*/
IDBCSBuffer * IDBCSBuffer :: strip ( const IStringTest      &aTest,
                                 IStringEnum::StripMode  mode )
  {
  IDBCSBuffer *result = this;

  if ( this->length() != 0 )
    {
#ifdef IC_NLS
    if (IString::isInternationalized())
    {
      IWBuffer wBufferThis(this->contents(), this->length());
      int iInLen = wBufferThis.charNum();
      int iStart, iEnd;
      iStart = 1;
      iEnd = iInLen;
      if (mode != IStringEnum::trailing)
      {
         for (iStart = 1; iStart <= iInLen; iStart++)
         {
            if (!aTest.test(wBufferThis[iStart]))
               break;
         }
      }
      if (mode != IStringEnum::leading)
      {
        for (iEnd = iInLen; iEnd > 0; iEnd--)
        {
          if (!aTest.test(wBufferThis[iEnd]))
             break;
        }
      }
      if (iEnd >= iStart)
      {
        char *pcChar = new char[(iEnd - iStart + 1) * MB_CUR_MAX];
        int iNumChar = cvtWCS2MBS(wBufferThis + iStart - 1, iEnd - iStart + 1,
                                  pcChar);
        result = (IDBCSBuffer *)this->newBuffer( pcChar, iNumChar);
        delete [] pcChar;
      }
      else
      {
        result = (IDBCSBuffer *)this->null();
        result->addRef();
      }
    } else {
#endif
    unsigned start = 1,
             stop  = this->length();
    if ( mode != IStringEnum::trailing )
      start = this->indexOfAnyBut( aTest, 1 );
    if ( mode != IStringEnum::leading
         &&
         start != 0 )
      {
      stop = this->lastIndexOfAnyBut( aTest, this->length() );
      // If last char to keep is DBCS, keep two bytes.
      if (stop > 0)
         stop += charLength ( stop ) - 1;
      }
    if ( start != 1
         ||
         stop != this->length() )
      {
      if ( start != 0 )
        result = (IDBCSBuffer *)this->newBuffer( this->contents() + start - 1,
                                                 stop - start + 1 );
      else
        {
        result = (IDBCSBuffer *)this->null();
        result->addRef();
        }
      }
    else
      this->addRef();
#ifdef IC_NLS
    }
#endif
    }
  else
    this->addRef();

  return result;
  }

/*------------------------------------------------------------------------------
| IDBCSBuffer::translate                                                       |
|                                                                              |
| The translation is accomplished by:                                          |
|                                                                              |
|   1. Determine the number of logical characters in the input string          |
|      (pInputChars) that are to be translated.                                |
|   2. Set up a translation table that consists of an entry for each           |
|      logical character to be translated, and contains information            |
|      such as the input character, input character length, output character,  |
|      output character length, and difference between outlen and inlen.       |
|   3. Set up a translation occurrence linked list which consists of           |
|      one node per translation that is to occur.  Each node contains 1) the   |
|      position in self.data that contains a character to be translated,       |
|      and 2) an index into the transTable that contains the translation       |
|      info.  This is accomplished by finding all occurrences of each          |
|      inputCharacter (which is a logical character) in the string (via        |
|      indexOfAnyOf()).                                                        |
|   4. A new buffer is allocated (with a new length if required) and filled    |
|      in by processing each node in the translation occurrence linked list.   |
------------------------------------------------------------------------------*/
class transTable
{
  public:
    char in[MB_LEN_MAX];                 // Input character to be translated
    char out[MB_LEN_MAX];                // Output character
    char inLen;                          // Length of input character (1 for
                                         //  single byte, 2 for double byte)
    char outLen;                         // Length of output character (1 for
                                         //  single byte, 2 for double byte)
    int diff;                            // outlen - inlen
};

class transOccurrence : public Occur
{
  public:
    unsigned int n;                      // corresponding index in transTable
                                         // constructor
    transOccurrence(unsigned int i, unsigned int j);
    ~transOccurrence ( );
};

transOccurrence :: transOccurrence ( unsigned int i, unsigned int j )
  : Occur( i ),
    n( j )
  {
  }


transOccurrence :: ~transOccurrence ( )
  {
  }

IDBCSBuffer * IDBCSBuffer :: translate ( const char *pInputChars,
                                     unsigned    inputLen,
                                     const char *pOutputChars,
                                     unsigned    outputLen,
                                     char        padCharacter )
{
  IDBCSBuffer *result = this;
#ifdef IC_NLS
  if (IString::isInternationalized())
  {
    IWBuffer wBufferIn(pInputChars, inputLen);
    IWBuffer wBufferPad(&padCharacter, 1);
    IWBuffer wBufferOut(pOutputChars, outputLen);
    int iNumIn = wBufferIn.charNum();
    int iNumOut = wBufferOut.charNum();
    IWBuffer wBufferThis(this->contents(), this->length());
    int iNumWChar = wBufferThis.charNum();
    for (int iStringIdx = 1; iStringIdx <= iNumWChar; iStringIdx++)
       for (int iTranIdx = 1; iTranIdx <= iNumIn; iTranIdx++)
          if (wBufferThis[iStringIdx] == wBufferIn[iTranIdx])
          {
             if (iTranIdx > iNumOut)
                 wBufferThis[iStringIdx] = wBufferPad[1];
             else
                 wBufferThis[iStringIdx] = wBufferOut[iTranIdx];
             break;
          }
    char *pcChar = new char[iNumWChar * MB_CUR_MAX];
    int iNumChar = cvtWCS2MBS(wBufferThis, iNumWChar, pcChar);
    result = (IDBCSBuffer *)this->newBuffer(pcChar, iNumChar);
    delete [] pcChar;

  } else {
#endif

  //--------------------------------------------------------------------------
  // Determine the number of characters (NOT bytes) to be translated.
  //--------------------------------------------------------------------------
    unsigned int numChars = 0,
                 n        = 0;
    while ( n < inputLen )
    {
      n += charLength ( &pInputChars [n] );
      numChars++;
    }

  //--------------------------------------------------------------------------
  // Set up the translation table.  There is a separate translation table
  // entry for each input-output pair.  Each entry contains inputChar,
  // inputChar length, outputChar, outputChar length, and the amount of
  // difference in the two lengths.
  //--------------------------------------------------------------------------
    transTable* transTbl = new transTable[ numChars ];
    const char *pi = pInputChars;
    const char *po = pOutputChars;
    for ( n = 0; n < numChars; n++)
    {
       unsigned i;
       transTbl[n].inLen = (char) charLength ( pi );
       for (i = 0; i < transTbl[n].inLen; i++)
         transTbl[n].in[i] = *pi++;

       if ( po >= pOutputChars + outputLen )
       {
         transTbl[n].outLen = 1;
         transTbl[n].out[0] = padCharacter;
       }
       else
       {
         transTbl[n].outLen = (char) charLength ( po );
         for (i = 0; i < transTbl[n].outLen; i++)
           transTbl[n].out[i] = *po++;
       }
       transTbl[n].diff = transTbl[n].outLen - transTbl[n].inLen;
    }


  //--------------------------------------------------------------------------
  //                        Set up transOccurrence.
  //--------------------------------------------------------------------------
  // transOccurrence is a linked list consisting of one node per translation
  // that is to occur.  Each node contains 1) the position in self.data that
  // contains a character to be translated, and 2) an index into the transTable
  // that contains the translation info (inputChars, outputChars, lengths).
  //--------------------------------------------------------------------------
    pi = pInputChars;
    unsigned int pos = 0;
    transOccurrence head = transOccurrence( 0, 0 );
    transOccurrence* pTransOcc = &head;
    unsigned int count = 0;
    int lengthChange = 0;

  //--------------------------------------------------------------------------
  // Set up the transOccurrence linked list by finding all occurrences of
  // each inputCharacter (which is a logical character) in the string.  Also,
  // keep track of the amount of string length change that will occur so that
  // the new resulting string can be allocated at the correct length.
  //--------------------------------------------------------------------------
    while ( ( pos = this->indexOfAnyOf( pi,
                                        inputLen,
                                        pos + 1 ) ) != 0 )
    {
       size_t len = charLength ( pos );
       for (n = 0; n < numChars; n++)
       {
         if (! memcmp( &this->contents()[pos-1], &transTbl[n].in, len ) )
           break;
       }
       pTransOcc = (transOccurrence*)
                   ( pTransOcc -> pNext = new transOccurrence( pos-1, n ) );
       count++;
       lengthChange += transTbl[n].diff;
    }

  //--------------------------------------------------------------------------
  // Allocate the new, translated string; this string will reflect any length
  // change calculated above.
  //--------------------------------------------------------------------------
    char* pOld = this->contents();
    result = (IDBCSBuffer *)this->newBuffer( this->contents(), this->length(),
                                             0, lengthChange,
                                             0, 0, 0 );
    pTransOcc = (transOccurrence*)head.pNext;
    char* pSource = pOld;
    char* pDest = result->contents();
    unsigned int previousPos = 0;

  //--------------------------------------------------------------------------
  // Copy the contents of the old string into the new string (result) until
  // we get to a position that requires a translation.  At that point, copy
  // the translation character into the new string, based on the
  // information in the transTable.  Continue this until all nodes in
  // transOccurrence have been processed (number of nodes == count).
  //--------------------------------------------------------------------------
    while (count--)
    {
       unsigned int len = pTransOcc->pos - previousPos;
       memcpy( pDest, pSource, len );
       pDest += len;
       pSource += len;
       memcpy( pDest, transTbl[pTransOcc->n].out,
               len = transTbl[pTransOcc->n].outLen );
       pDest += len;
       pSource += transTbl[pTransOcc->n].inLen;
       previousPos = pSource - pOld;
       pTransOcc = (transOccurrence*)pTransOcc->pNext;
    }

  //--------------------------------------------------------------------------
  // Finish filling in the new string (result) with the remaining string
  // contents.
  //--------------------------------------------------------------------------
    while ((*pDest++ = *pSource++) != 0)
    { ; /* empty */ }

    delete [] transTbl;
#ifdef IC_NLS
  }
#endif
  return result;
}


/*------------------------------------------------------------------------------
| IDBCSBuffer::upperCase                                                       |
|                                                                              |
| Starting at the beginning of the string, scan for the next lower             |
| case letter and if found, convert it to upper case.  Quit as soon            |
| as the search for a lowercase (via indexOfAnyOf()) fails.                    |
------------------------------------------------------------------------------*/
IDBCSBuffer * IDBCSBuffer :: upperCase ( )
{
  IDBCSBuffer *result = this;
#ifdef IC_NLS
  if (IString::isInternationalized())
  {
    if (this->length() > 0)
    {
       IWBuffer wBufferThis(this->contents(), this->length());
       int iNumWChar = wBufferThis.charNum();
       for (int iStringIdx = 1; iStringIdx <= iNumWChar; iStringIdx++)
          wBufferThis[iStringIdx] = towupper(wBufferThis[iStringIdx]);
       char *pcChar = new char[iNumWChar * MB_CUR_MAX];
       int iNumChar = cvtWCS2MBS(wBufferThis, iNumWChar, pcChar);
       result = (IDBCSBuffer *)this->newBuffer(pcChar, iNumChar);
       delete [] pcChar;
    }
    else
       this->addRef();
  } else {
#endif
    if ( this->useCount() > 1 )
      result = (IDBCSBuffer *)this->newBuffer( this->contents(),
                                               this->length() );
    else
      this->addRef();
    char *p = result->contents();
    char *pEnd = p + result->length();
    while ( p < pEnd ) {
	  *p = toupper(*p);
      ++p;
    }
#ifdef IC_NLS
  }
#endif
  return result;
}
