/*******************************************************************************
* FILE NAME: idllimp.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the functions in idllimp.hpp      *
*                                                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Revision: 87 1.7.1.18 source/core/base/idllimp.cpp, appsupport, ioc.v400, 000302 

#include <ibase.hpp>
#include <idllstat.hpp>
#include <idllimp.hpp>
#include <iexcept.hpp>
#include <itrace.hpp>
#include <ireslock.hpp>
#include <istring.hpp>
#include <istparse.hpp>
#include <stdlib.h>
#include <stdio.h>

#ifdef IC_WIN
#include <windows.h>
#endif

#ifdef IC_PM
  #define INCL_DOSMODULEMGR
  #define INCL_DOSRESOURCES    /* Resource types */
  #define INCL_DOSINFOSEG
  #define INCL_DOSPROCESS
  #define INCL_WINERRORS
  #include <os2.h>
#endif


#ifdef IC_MOTIF
  #include <sys/ldr.h>       // unload
  #include <load.h>          // load()
  #include <sys/types.h>
  #include <locale.h>
  #include <filehdr.h>
  #include <aouthdr.h>
  #include <scnhdr.h>
  #include <loader.h>
  #include <stdio.h>
  #include <xcoff.h>
  #include <errno.h>
  #include <unistd.h>
#endif



#ifdef IC_PAGETUNE
    #define _IDLLIMP_CPP_
    #include <ipagetun.h>
#endif

/***************************************************************/
/*                                                             */
/*  Reference Counting Strategy.                               */
/*                                                             */
/*  We need to reference count the dlls because the OS (well   */
/*  at least OS/2) allows you to open the same dll multiple    */
/*  times but kills the dll on the first attempt to close it.  */
/*                                                             */
/*  It is the responsibility of IDLLModule to locate an        */
/*  existing instance of IRefCountedDLLImp (really a derived   */
/*  class) in the static collection of implementation objects  */
/*  and update its reference count by calling addRef.          */
/*                                                             */
/*  addRef calls openDLL as necessary (a reference count of    */
/*  zero) to load the dll.  removeRef calls closeDLL as        */
/*  necessary (the reference count is zero after decrementing  */
/*  it) to unload the dll.                                     */
/***************************************************************/

///////////////////////////////////////////////////////////////////////////////
///////   IRefCountedDLLImp Functions
///////////////////////////////////////////////////////////////////////////////

/*------------------------------------------------------------------------------
| IRefCountedDLLImp::IRefCountedDLLImp                                         |
------------------------------------------------------------------------------*/
IRefCountedDLLImp::IRefCountedDLLImp( const char* creationName,
                                      bool        useNLSPath, 
                                      const char* defaultExtension)
        : fCreationName(creationName),
          fDefaultExtension(defaultExtension),
          fModuleHandle(0),
          fUseCount(0),
          fUseNLSPath(useNLSPath)
{
  IFUNCTRACE_ALL();
}

/*------------------------------------------------------------------------------
| IRefCountedDLLImp::IRefCountedDLLImp                                         |
------------------------------------------------------------------------------*/
IRefCountedDLLImp::IRefCountedDLLImp( const IModuleHandle& handle )
        : fCreationName(""),
          fModuleHandle(handle),
          fUseCount(1),
          fUseNLSPath(false)
{
  IFUNCTRACE_ALL();
}

/*------------------------------------------------------------------------------
| IRefCountedDLLImp::~IRefCountedDLLImp                                        |
------------------------------------------------------------------------------*/
IRefCountedDLLImp::~IRefCountedDLLImp()
{
  IFUNCTRACE_ALL();
}

/*------------------------------------------------------------------------------
| IRefCountedDLLImp::operator ==                                               |
|   This function returns equality if the handles match.                       |
------------------------------------------------------------------------------*/
bool IRefCountedDLLImp::operator ==(const IRefCountedDLLImp& dllImp) const
{
  IFUNCTRACE_ALL();

  return (dllImp.handle() == handle());
}


/*------------------------------------------------------------------------------
| IRefCountedDLLImp::initialCreationName                                       |
------------------------------------------------------------------------------*/
IString IRefCountedDLLImp::initialCreationName ( ) const
{
  IFUNCTRACE_ALL();
  return fCreationName;
}


/*------------------------------------------------------------------------------
| IRefCountedDLLImp::addRef                                                    |
|   Add a reference to a DLL Module. Load the dll if count is zero.            |
------------------------------------------------------------------------------*/
unsigned long IRefCountedDLLImp::addRef()
{
  IFUNCTRACE_ALL();

  if (useCount() == 0)
    openDLL();
  fUseCount++;
  return( fUseCount );
}


/*------------------------------------------------------------------------------
| IRefCountedDLLImp::removeRef                                                 |
|  Remove a reference. Delete if the count is zero.                            |
------------------------------------------------------------------------------*/
unsigned long IRefCountedDLLImp::removeRef()
{
  IFUNCTRACE_ALL();

  fUseCount--;
  if (fUseCount == 0)
    closeDLL();

  return( fUseCount );
}

/*------------------------------------------------------------------------------
| IRefCountedDLLImp::nlsPathName                                               |
|   Note:  The code here works for OS/2 and Windows but must be replaced by    |
|          derived classes on other platforms.                                 |
------------------------------------------------------------------------------*/
IString IRefCountedDLLImp::nlsPathName( ) const
{
  IFUNCTRACE_ALL();

#ifdef IC_PMWIN
  // Note: When aix supports _searchenv we should be able to remove the ifdefs
  //       and change the code in IMotifDLLImp::nlsPathName to call this.
  IString langEnvironment(getenv("LANG"));
  IString nlsPathEnvironment(getenv("NLSPATH"));
  nlsPathEnvironment.change("%l", langEnvironment);
  nlsPathEnvironment.change("%L", langEnvironment);
  nlsPathEnvironment.change("\\%N", "");
  nlsPathEnvironment.change("\\%N", "");
  nlsPathEnvironment.change("/%N", "");
  nlsPathEnvironment.change("/%N", "");
  putenv(IString("IREFCOUNTEDDLLIMP=")+nlsPathEnvironment);
  IString fileAndExtension(initialCreationName());
  if(fileAndExtension.indexOf(".") == 0)
     fileAndExtension += fDefaultExtension;
  char fullNLSPathName[_MAX_PATH];
  _searchenv(fileAndExtension, "IREFCOUNTEDDLLIMP", fullNLSPathName);
  putenv("IREFCOUNTEDDLLIMP");
  return IString(fullNLSPathName);
#endif 

#ifdef IC_MOTIF
  return IString("");
#endif
}


/*------------------------------------------------------------------------------
| IRefCountedDLLImp::asDebugInfo                                               |
------------------------------------------------------------------------------*/
IString IRefCountedDLLImp::asDebugInfo() const
{
  IString str="IRefCountedDLLImp(initialCreationName(";
  str += fCreationName;
  str += "),Handle(";
  str += IString(fModuleHandle.asString());
  str += "),UseCount(";
  str += IString(fUseCount);
  str += "))";
  return( str );
}

///////////////////////////////////////////////////////////////////////////////
///////   IWindowsDLLImp Functions
///////////////////////////////////////////////////////////////////////////////


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IWindowsDLLImp::IWindowsDLLImp                                               |
|    Constructor from potentially unqualified name.                            |
------------------------------------------------------------------------------*/
IWindowsDLLImp::IWindowsDLLImp ( const char* creationName,
                                 bool        searchNLSPath)
                             : IRefCountedDLLImp ( creationName, searchNLSPath, ".dll" )
{
  IFUNCTRACE_DEVELOP();
}

/*------------------------------------------------------------------------------
| IWindowsDLLImp::IWindowsDLLImp                                               |
|    Constructor from module handle.                                           |
------------------------------------------------------------------------------*/
IWindowsDLLImp::IWindowsDLLImp ( const IModuleHandle& moduleHandle )
                            : IRefCountedDLLImp ( moduleHandle )
{
  IFUNCTRACE_DEVELOP();
  char name[512];
  unsigned long
    success = GetModuleFileName( moduleHandle, (char*)name, sizeof(name) );
  if (success == 0)
  {
    ITHROWGUIERROR2( "GetModuleFileName",
                    IBaseErrorInfo::accessError,
                    IException::recoverable );
  }
  setCreationName(name);
}

/*------------------------------------------------------------------------------
| IWindowsDLLImp::~IWindowsDLLImp                                              |
|    Destructor.                                                               |
------------------------------------------------------------------------------*/
IWindowsDLLImp::~IWindowsDLLImp ( )
{
  IFUNCTRACE_DEVELOP();
}

/*------------------------------------------------------------------------------
| IWindowsDLLImp::openDLL                                                      |
|    Open the DLL and set the module handle.                                   |
------------------------------------------------------------------------------*/
void IWindowsDLLImp::openDLL ( )
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("initialCreationName()=")+initialCreationName());

  HMODULE moduleHandle = 0;

  unsigned long ulErrorCode = 0;
  // Suppress the error message box that appears when the dll load fails
  unsigned int uiErrorMode = SetErrorMode( SEM_FAILCRITICALERRORS );

  if(useNLSPath() == true)
    moduleHandle = LoadLibrary(nlsPathName());
  else
    moduleHandle = LoadLibrary( initialCreationName() );
  if (moduleHandle == 0)
  {
    ulErrorCode = GetLastError();
    if ( ulErrorCode != 0 )
    {
      SetErrorMode( uiErrorMode );
      ITHROWGUIERROR2( IString("LoadLibrary : ") + initialCreationName(),
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
    }
  }


  SetErrorMode( uiErrorMode );
  setHandle(moduleHandle);

  ITRACE_DEVELOP(IString("Handle=")+IString(handle().asUnsigned()).d2x());
  ITRACE_DEVELOP(IString("loadedPathName()=")+loadedPathName());
}

/*------------------------------------------------------------------------------
| IWindowsDLLImp::closeDLL                                                     |
|    Close the DLL and reset the module handle.                                |
------------------------------------------------------------------------------*/
void IWindowsDLLImp::closeDLL ( )
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("Handle=")+IString(handle().asUnsigned()).d2x());
  unsigned long
    ulErrorCode = (! FreeLibrary( handle() ) );
  if (ulErrorCode != 0)
  {
    // Defect icl2_51888. To keep from having an exception occur during
    // a destructor call, we are logging an error instead of throwing
    // an exception.
    IGUIErrorInfo errorString("FreeLibrary");
    ITRACE_RUNTIME(IString(errorString) + initialCreationName());
  }

  setHandle(0);
}

/*------------------------------------------------------------------------------
| IWindowsDLLImp::procAddress                                                  |
|    Return the address of the passed procedure.                               |
------------------------------------------------------------------------------*/
void* IWindowsDLLImp::procAddress ( const char* procName) const
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("ProcName=")+procName);

  FARPROC pfnTemp;
  pfnTemp = GetProcAddress( handle(), procName );
  if ( !pfnTemp )
  {
    ITHROWGUIERROR2( IString("GetProcAddress: ") + procName,
                     IBaseErrorInfo::accessError,
                     IException::recoverable );

  }
  return( pfnTemp );
}

/*------------------------------------------------------------------------------
| IWindowsDLLImp::procAddress                                                  |
|    Return the address of the passed ordinal.                                 |
------------------------------------------------------------------------------*/
void* IWindowsDLLImp::procAddress ( unsigned long procOrdinal) const
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("ProcOrdinal=")+IString(procOrdinal));

  FARPROC pfnTemp;
  long ordinal1 =  MAKELONG(LOWORD(procOrdinal),0);
  char *ordinal = (char *)ordinal1;
  pfnTemp = GetProcAddress( handle(), ordinal );
  if ( !pfnTemp )
  {
    ITHROWGUIERROR2( IString("GetProcAddress #: ") +
                              IString(procOrdinal),
                     IBaseErrorInfo::accessError,
                     IException::recoverable );

  }
  return( pfnTemp );
}

/*------------------------------------------------------------------------------
| IWindowsDLLImp::procType                                                     |
|    Return the type of the passed procedure name.                             |
------------------------------------------------------------------------------*/
IRefCountedDLLImp::ProcType IWindowsDLLImp::procType ( const char* ) const
{
   IFUNCTRACE_ALL();
   return (k32Bit);  // Only 32 bit currently supported on Windows
}


/*------------------------------------------------------------------------------
| IWindowsDLLImp::procType                                                     |
|    Return the type of the passed procedure ordinal.                          |
------------------------------------------------------------------------------*/
IRefCountedDLLImp::ProcType IWindowsDLLImp::procType ( unsigned long ) const
{
   IFUNCTRACE_ALL();
   return (k32Bit);  // Only 32 bit currently supported on Windows
}


/*------------------------------------------------------------------------------
| IWindowsDLLImp::loadedPathName                                               |
|    Return the fully qualified name of the loaded dll.                        |
------------------------------------------------------------------------------*/
IString IWindowsDLLImp::loadedPathName ( ) const
{
  IFUNCTRACE_ALL();
  IASSERTSTATE(handle() != 0);

  if(fLoadedPathName == "")
  {
    if(useNLSPath() == true)
       ((IWindowsDLLImp*)this)->fLoadedPathName = nlsPathName();
    else
    {
      char pathName[512];
      unsigned long ulErrorCode = GetModuleFileName(
                                     handle(),
                                     (char*)pathName,
                                     sizeof(pathName) );
      if (ulErrorCode == 0)
      {
        ITHROWGUIERROR2( "GetModuleFileName",
                         IBaseErrorInfo::accessError,
                         IException::recoverable );
      }
      ((IWindowsDLLImp*)this)->fLoadedPathName = pathName;
    }
  }

  ITRACE_ALL(fLoadedPathName);
  return fLoadedPathName;
}

/*------------------------------------------------------------------------------
| IWindowsDLLImp::nlsPathName                                                  |
------------------------------------------------------------------------------*/
IString IWindowsDLLImp::nlsPathName( ) const
{
   IFUNCTRACE_ALL();
   return Inherited::nlsPathName();
}

#endif // IC_WIN


///////////////////////////////////////////////////////////////////////////////
///////   IOS2DLLImp Functions
///////////////////////////////////////////////////////////////////////////////

#ifdef IC_PM
/*------------------------------------------------------------------------------
| IOS2DLLImp::IOS2DLLImp                                                       |
|    Constructor from potentially unqualified name.                            |
------------------------------------------------------------------------------*/
IOS2DLLImp::IOS2DLLImp ( const char* creationName,
                         bool        searchNLSPath)
                    : IRefCountedDLLImp ( creationName, searchNLSPath, ".dll" )
{
  IFUNCTRACE_DEVELOP();
}

/*------------------------------------------------------------------------------
| IOS2DLLImp::IOS2DLLImp                                                       |
|    Constructor from module handle.                                           |
------------------------------------------------------------------------------*/
IOS2DLLImp::IOS2DLLImp ( const IModuleHandle& moduleHandle )
                    : IRefCountedDLLImp ( moduleHandle )
{
  IFUNCTRACE_DEVELOP();
  char dllName[512];
  unsigned long
     ulErrorCode = DosQueryModuleName( moduleHandle,
                                       sizeof(dllName),
                                       dllName );
  if (ulErrorCode != 0)
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       "DosQueryModuleName",
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
  setCreationName(dllName);
}

/*------------------------------------------------------------------------------
| IOS2DLLImp::~IOS2DLLImp                                                      |
|    Destructor.                                                               |
------------------------------------------------------------------------------*/
IOS2DLLImp::~IOS2DLLImp ( )
{
  IFUNCTRACE_ALL();
}

/*------------------------------------------------------------------------------
| IOS2DLLImp::openDLL                                                          |
|    Open the DLL and set the module handle.                                   |
------------------------------------------------------------------------------*/
void IOS2DLLImp::openDLL ( )
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("initialCreationName()=")+initialCreationName());

  HMODULE moduleHandle = 0;
  char objectNameBuffer[100];

  // Initialize the dll name to the name passed in on the constructor
  IString loadName(initialCreationName());

  // If using the NLSPath, update the name accordingly.
  if(useNLSPath() == true)
     loadName = nlsPathName();

  // If the name doesn't contain a back slash, strip trailing .dll
  if(loadName.indexOf("\\") == 0)
  {
     unsigned long dotDLLPosition = loadName.indexOf(".DLL");
     if(dotDLLPosition !=0)
        loadName = loadName.subString(1, dotDLLPosition-1);
     else
     {
       unsigned long dotdllPosition = loadName.indexOf(".dll");
       if(dotdllPosition !=0)
          loadName = loadName.subString(1, dotdllPosition-1);
     }
  }


  unsigned long ulErrorCode = DosLoadModule( objectNameBuffer,
                                             sizeof(objectNameBuffer),
                                             loadName,
                                             &moduleHandle );
  if ( (ulErrorCode != 0) || (moduleHandle == 0) )
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       IString("DosLoadModule : ") + initialCreationName(),
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
  setHandle(moduleHandle);

  ITRACE_DEVELOP(IString("Handle=")+IString(handle().asUnsigned()).d2x());
  ITRACE_DEVELOP(IString("loadedPathName()=")+loadedPathName());
}

/*------------------------------------------------------------------------------
| IOS2DLLImp::closeDLL                                                         |
|    Close the DLL and reset the module handle.                                |
------------------------------------------------------------------------------*/
void IOS2DLLImp::closeDLL ( )
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("Handle=")+IString(handle().asUnsigned()).d2x());

  unsigned long ulErrorCode = DosFreeModule( handle() );
  if (ulErrorCode != 0)
  {

     // Defect icl2_51888. To keep from having an exception occur during
     // a destructor call, we are logging an error instead of throwing
     // an exception.
     ISystemErrorInfo errorString(ulErrorCode, "DosFreeModule");
     ITRACE_RUNTIME(IString(errorString) + initialCreationName());
  }
  setHandle(0);
}

/*------------------------------------------------------------------------------
| IOS2DLLImp::procAddress                                                      |
|    Return the address of the passed procedure.                               |
------------------------------------------------------------------------------*/
void* IOS2DLLImp::procAddress ( const char* procName) const
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("procName=")+procName);

  unsigned long ulErrorCode = 0;
  int (_System *pfnTemp)() = 0;
  ulErrorCode = DosQueryProcAddr( handle(), 0, (char*)procName, &pfnTemp );
  if (ulErrorCode != 0)
  {
    // If call failed, upper case string and try again
    IString tryUpper( procName );
    tryUpper.upperCase();
    ulErrorCode = DosQueryProcAddr( handle(), 0, (char*)tryUpper, &pfnTemp );
  }
  if (ulErrorCode != 0)
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       IString("DosQueryProcAddr: ") + procName,
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
  return( pfnTemp );
}

/*------------------------------------------------------------------------------
| IOS2DLLImp::procAddress                                                      |
|    Return the address of the passed ordinal.                                 |
------------------------------------------------------------------------------*/
void* IOS2DLLImp::procAddress ( unsigned long procOrdinal) const
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("procOrdinal=")+IString(procOrdinal));

  unsigned long ulErrorCode = 0;
  int (_System *pfnTemp)() = 0;
  ulErrorCode = DosQueryProcAddr( handle(), procOrdinal, (char*)0, &pfnTemp );
  if (ulErrorCode != 0)
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       IString("DosQueryProcAddr #: ") +
                                IString(procOrdinal),
                       IBaseErrorInfo::accessError,
                       IException::recoverable );
  }
  return( pfnTemp );
}

/*------------------------------------------------------------------------------
| IOS2DLLImp::procType                                                         |
|    Return the type of the passed procedure name.                             |
------------------------------------------------------------------------------*/
IRefCountedDLLImp::ProcType IOS2DLLImp::procType ( const char* procedureName) const
{
  IFUNCTRACE_ALL();
  unsigned long
    ulResult = 1;
  unsigned long
    ulErrorCode = DosQueryProcType( this->handle(),
                                  0,
                                  (PSZ)procedureName,
                                  &ulResult );

  if ( ulErrorCode )
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       IString("DosQueryProcType: ") +
                                procedureName,
                       IBaseErrorInfo::accessError,
                       IException::recoverable);
  }
  if(ulResult == 1)
    return (k32Bit);
  else
    return (k16Bit);
}


/*------------------------------------------------------------------------------
| IOS2DLLImp::procType                                                         |
|    Return the type of the passed procedure ordinal.                          |
------------------------------------------------------------------------------*/
IRefCountedDLLImp::ProcType IOS2DLLImp::procType ( unsigned long procedureOrdinal) const
{
   IFUNCTRACE_ALL();
   unsigned long ulResult = 1;
   unsigned long ulErrorCode = DosQueryProcType( this->handle(),
                                  procedureOrdinal,
                                  0,
                                  &ulResult );

  if ( ulErrorCode )
  {
    ITHROWSYSTEMERROR( ulErrorCode,
                       IString("DosQueryProcType #: ") +
                                IString(procedureOrdinal),
                       IBaseErrorInfo::accessError,
                       IException::recoverable);
  }

  if(ulResult == 1)
    return (k32Bit);
  else
    return (k16Bit);
}


/*------------------------------------------------------------------------------
| IOS2DLLImp::loadedPathName                                                   |
|    Return the fully qualified name of the loaded dll.                        |
------------------------------------------------------------------------------*/
IString IOS2DLLImp::loadedPathName ( ) const
{
  IFUNCTRACE_ALL();
  IASSERTSTATE(handle() != 0);

  if(fLoadedPathName == "")
  {
    if(useNLSPath() == true)
       ((IOS2DLLImp*)this)->fLoadedPathName = nlsPathName();
    else
    {
      char pathName[512];
      unsigned long ulErrorCode = DosQueryModuleName(
                                        handle(),
                                        sizeof(pathName),
                                        (char*)pathName);
      if (ulErrorCode != 0)
      {
        ITHROWSYSTEMERROR( ulErrorCode,
                           "DosQueryModuleName",
                           IBaseErrorInfo::accessError,
                           IException::recoverable );
      }
      ((IOS2DLLImp*)this)->fLoadedPathName = pathName;
    }
  }
  ITRACE_ALL(fLoadedPathName);
  return IString( fLoadedPathName );
}

/*------------------------------------------------------------------------------
| IOS2DLLImp::nlsPathName                                                      |
------------------------------------------------------------------------------*/
IString IOS2DLLImp::nlsPathName( ) const
{
   IFUNCTRACE_ALL();
   return Inherited::nlsPathName();
}


#endif // IC_PM


///////////////////////////////////////////////////////////////////////////////
///////   IMotifDLLImp Functions
///////////////////////////////////////////////////////////////////////////////

#ifdef IC_MOTIF

/*------------------------------------------------------------------------------
| IMotifDLLImp::IMotifDLLImp                                                   |
|    Constructor from potentially unqualified name.                            |
------------------------------------------------------------------------------*/
IMotifDLLImp::IMotifDLLImp ( const char* creationName,
                             bool        searchNLSPath)
                    : IRefCountedDLLImp(creationName, searchNLSPath, ".o" ),
                      fLoadedPathName(""),
                      fEntryPoint(0),
                      fLoaderHeader(0),
                      fLoaderSymbols(0),
                      fLoaderInfo(0),
                      fResources(0),
                      fLoaderBias(0)
{
   IFUNCTRACE_DEVELOP();
}

#if 0  // Not currently supported on Motif
/*------------------------------------------------------------------------------
| IMotifDLLImp::IMotifDLLImp                                                   |
|    Constructor from module handle.                                           |
------------------------------------------------------------------------------*/
IMotifDLLImp::IMotifDLLImp ( const IModuleHandle& moduleHandle )
                    : IRefCountedDLLImp ( moduleHandle ),
                      fLoadedPathName("")
{
   IFUNCTRACE_DEVELOP();
}
#endif

/*------------------------------------------------------------------------------
| IMotifDLLImp::~IMotifDLLImp                                                  |
|    Destructor.                                                               |
------------------------------------------------------------------------------*/
IMotifDLLImp::~IMotifDLLImp ( )
{
  IFUNCTRACE_ALL();
}

/*------------------------------------------------------------------------------
| IMotifDLLImp::openDLL                                                        |
|    Open the DLL and set the module handle.                                   |
------------------------------------------------------------------------------*/
void IMotifDLLImp::openDLL()
{
    IFUNCTRACE_DEVELOP();
    ITRACE_DEVELOP(IString("initialCreationName()=")+initialCreationName());
    
    IString loadName(initialCreationName());
    IModuleHandle moduleHandle;
    
    // If library is not found, append DefaultExtension (.o for motif) and try again.
    if(useNLSPath() == true) {
        IString nlsPath = makeNLSPath();

        /* don't append LIBPATH behind NLSPATH, 
           we may want to re-think about this issue latter.

        char *libPath = getenv("LIBPATH");
        IString searchPath("");

        if (libPath == NULL) {
            searchPath = nlsPath;
        } else {
            if (nlsPath == "") {
                searchPath = IString(libPath);
            } else {
                searchPath = nlsPath + IString(":") + IString(libPath);
            }
        }
        */

        ITRACE_DEVELOP(IString("library searchPath=") + nlsPath);
        if ((moduleHandle = (IModuleHandle::Value)loadAndInit(loadName, 1, nlsPath)) == 0) {
            loadName += getDefaultExtension();
            moduleHandle = (IModuleHandle::Value)loadAndInit(loadName, 1, nlsPath);
        }
    } else {
        // leave the search path to the operating system.
        // search the header section of the executable first before searching LIBPATH
        // all these are done thru the call to loadAndInit()
        if ((moduleHandle = (IModuleHandle::Value)loadAndInit(loadName, L_LIBPATH_EXEC, NULL)) == 0) {
            loadName += getDefaultExtension();
            moduleHandle = (IModuleHandle::Value)loadAndInit(loadName, L_LIBPATH_EXEC, NULL);
        }
    }
    
    if (moduleHandle == 0) {
        // error loading library
        IString errorString = IString("LoadLibrary: ") + initialCreationName() + "(errno="
            + IString(errno) + ") loader errors(";
        
        char* buffer[1024];
        loadquery(L_GETMESSAGES, &buffer[0], sizeof(buffer));
        for(int i=0; buffer[i] != 0; i++)
            errorString = errorString + buffer[i] + "\n";
        errorString += ")";
        
        ITHROWGUIERROR(errorString);
    }
    setHandle(moduleHandle);
    
    // Save info to load procedure addresses.
    saveLoaderInformation();
    
    ITRACE_DEVELOP(IString("Handle=")+IString(moduleHandle.asUnsigned()).d2x());
    ITRACE_DEVELOP(IString("loadedPathName()=")+loadedPathName());
}

/*------------------------------------------------------------------------------
| IMotifDLLImp::closeDLL                                                       |
|    Close the DLL and reset the module handle.                                |
------------------------------------------------------------------------------*/
void IMotifDLLImp::closeDLL ( )
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("Handle=")+IString(handle().asUnsigned()).d2x());

   // unload the library object
  terminateAndUnload ((int (*)())(handle().asUnsigned()));
  setHandle(0);
}


struct hdr1 {
    struct filehdr f;
    struct aouthdr a;
    struct scnhdr s[1];
};

/*------------------------------------------------------------------------------
| IMotifDLLImp::procAddress                                                    |
|    Return the address of the passed procedure.                               |
------------------------------------------------------------------------------*/
void* IMotifDLLImp::procAddress(const char* procName) const
{
    IFUNCTRACE_DEVELOP();
    ITRACE_DEVELOP(IString("procName=")+procName);
    
    // Special code to handle returning the address of DLL resources
    if (strstr(procName,"__rsrc") != 0)
        return fResources;
    
    void *pfnTemp;
    // Name *name;                            //leave commented for documentation
    char *rest;
    ulong virtFuncAddr;
    int i;
    
    for (virtFuncAddr=0, i=0; i < fLoaderHeader->l_nsyms; i++) {
        //
        // we've got another symbol ..
        //
        if ((fLoaderSymbols[i].l_smtype & L_EXPORT) != 0) {
            bool found = false;
            //
            // .. and its exported .. stuff its name in str ..
            //
            if (fLoaderSymbols[i].l_zeroes) {
                // the symbol may or may NOT be null terminated
                char n[SYMNMLEN+1];
                strncpy(n, fLoaderSymbols[i]._l._l_name, SYMNMLEN);
                n[SYMNMLEN] = 0;
                if (strcmp(n, procName) == 0)
                    found = true;
            } else {
                // symbol is null termianted
                if (fLoaderSymbols[i].l_offset) {
                    // if there is a string in the string table (l_offset within
                    // string table)
                    char *str = ((char *)fLoaderHeader + fLoaderHeader->l_stoff +
                                 fLoaderSymbols[i].l_offset);
                    if (strcmp(str, procName) == 0)
                        found = true;
                }
            }
            
            if (found) {
                virtFuncAddr = fLoaderSymbols[i].l_value;
                break;
            }
        }
    }
    
    pfnTemp = (virtFuncAddr) ? (void *)(virtFuncAddr + fLoaderBias) : 0;
    
    if (pfnTemp == 0) {
        ITHROWGUIERROR(IString("IDLLModule::procAddress .. can't find ") +
                       IString(procName));
    }
    return( pfnTemp );
}

/*------------------------------------------------------------------------------
| IMotifDLLImp::procAddress                                                    |
|    Return the address of the passed ordinal.                                 |
------------------------------------------------------------------------------*/
void* IMotifDLLImp::procAddress ( unsigned long procOrdinal) const
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("procOrdinal=")+IString(procOrdinal));
   // Not supported on Motif yet (or ever?).
   return 0;
}

/*------------------------------------------------------------------------------
| IMotifDLLImp::procType                                                       |
|    Return the type of the passed procedure name.                             |
------------------------------------------------------------------------------*/
IRefCountedDLLImp::ProcType IMotifDLLImp::procType ( const char* procedureName) const
{
   IFUNCTRACE_ALL();
   // Not supported (not called) on Motif
   return (k32Bit);
}


/*------------------------------------------------------------------------------
| IMotifDLLImp::procType                                                       |
|    Return the type of the passed procedure ordinal.                          |
------------------------------------------------------------------------------*/
IRefCountedDLLImp::ProcType IMotifDLLImp::procType ( unsigned long procedureOrdinal) const
{
   IFUNCTRACE_ALL();
   // Not supported (not called) on Motif
   return (k32Bit);
}


/*------------------------------------------------------------------------------
| IMotifDLLImp::loadedPathName                                                 |
|    Return the fully qualified name of the loaded dll.                        |
------------------------------------------------------------------------------*/
IString IMotifDLLImp::loadedPathName ( ) const
{
    // fLoadedPathName is set in saveLoaderInformation()
    return fLoadedPathName;
}


/*------------------------------------------------------------------------------
| IMotifDLLImp::nlsPathName                                                    |
------------------------------------------------------------------------------*/
IString IMotifDLLImp::nlsPathName( ) const
{
    return fLoadedPathName;
}


/*------------------------------------------------------------------------------
| Our goal here is to replace %L with the language setting                     |
| and remove %N (and what follows) from the NLSPATH string so that it looks    |
| like a LIBPATH                                                               |
------------------------------------------------------------------------------*/
IString IMotifDLLImp::makeNLSPath()
{
    IFUNCTRACE_DEVELOP();

    // We have to regenerate NLSPATH everytime because it can be
    // changed at runtime.

    // Get the current locale first
    char *locale = setlocale(LC_MESSAGES, NULL);

    ITRACE_DEVELOP(IString("locale=") + IString(locale));

    // According to the standard, C and POSIX means don't do localization
    if (strcmp(locale, "C") == 0 || strcmp(locale, "POSIX") == 0) {
        return "";
    }

    // Now, get the NLSPATH
    char *nlsEnv = getenv("NLSPATH");
    if (nlsEnv == NULL)
        return "";
    
    // Must save whatever is in nlsEnv because getenv does not guarantee
    // to save the content across multiple calls. 
    IString nlsPathEnvironment(nlsEnv);
    IString result("");
    
    // change %L to the languange setting (%l is not a valid NLS substitution)
    nlsPathEnvironment.change("%L", locale);

    unsigned int colonPos = nlsPathEnvironment.indexOf(":");
    if (colonPos != 0) {
        unsigned int namePos;
        // at least two path components
        do {
            IString path = nlsPathEnvironment.subString(1, colonPos - 1);
            namePos = path.indexOf("%N");
            if (namePos != 0) {
                // remove %N and whatever follows in the path component
                // and add that section to result
                result += path.subString(1, namePos - 1);
            } else {
                // add whatever is in path to result
                result += path;
            }
            // always add colon because there is always one remaining
            // path component after the do ... while loop
            result += IString(":");

            // removes whatever from %N to colon
            nlsPathEnvironment.remove(1, colonPos);
            // finds the next colon position
            colonPos = nlsPathEnvironment.indexOf(":");
        } while (colonPos != 0);

        // for the very last path component
        namePos = nlsPathEnvironment.indexOf("%N");
        if (namePos != 0) {
            // remove %N and whatever follows in the path component
            // and add that section to result
            result += nlsPathEnvironment.subString(1, namePos - 1);
        } else {
            // add whatever is in path to result
            result += nlsPathEnvironment;
        }

    } else {
        result = nlsPathEnvironment;
    }

    return result;
}


/*------------------------------------------------------------------------------
| IMotifDLLImp__findResources                                                  |
|    Return the address of the resources in this shared library if found.  The |
|    resources are stored with the symbol __rscs.
------------------------------------------------------------------------------*/
static void* 
IMotifDLLImp__findResources(struct ld_info *info) 
{
    int i;
    
    struct hdr1 *hdr = (struct hdr1 *) info->ldinfo_textorg;
    char *data = (char *) info->ldinfo_dataorg;
    ulong datareloc = (ulong)data - hdr->s[hdr->a.o_sndata-1].s_vaddr;
    struct ldhdr *ldhdr;
    struct ldsym *ldsym;
    unsigned long *textOffset;
    
    /* find the loader symbol table */
    ldhdr = (struct ldhdr *)((char *)hdr + hdr->s[hdr->a.o_snloader-1].s_scnptr);
    ldsym = (struct ldsym *)(ldhdr + 1);
    
    /* walk through the symbol table */
    for (i = 0; i < ldhdr->l_nsyms; ldsym++, i++) {
        /* only look at exports which are NOT imports - that
         * is are defined by this module
         */
        if (L_EXPORT!=(ldsym->l_smtype & (L_IMPORT | L_EXPORT)))
            continue;
        
        /* check that it is a valid symbol class */
        if (ldsym->l_smclas != XMC_RW && ldsym->l_smclas != XMC_RO)
            continue;
        
        if (strncmp(ldsym->l_name, "___rsrc", 7) == 0
            || strncmp(ldsym->l_name, "__rsrc", 6) == 0) {
            /* check that it is in .data */
            if (ldsym->l_scnum != hdr->a.o_sndata)
                continue;
            textOffset = (unsigned long *)(ldsym->l_value + datareloc);
            return (void *)(*textOffset);
        }
    }
    return 0;
}


/*------------------------------------------------------------------------------
  | IMotifDLLImp::saveLoaderInformation                                          |
  |    Store info necessary to load procedure addresses.                         |
  ------------------------------------------------------------------------------*/
void 
IMotifDLLImp::saveLoaderInformation()
{
    IFUNCTRACE_DEVELOP();
    //
    // for an understanding of the following code, refer to the
    // AIX RISC System/6000 Files Reference (xcoff and/or a.out sections).
    // Note: changes to the definition, content or use of xcoff structures in
    //       future releases of AIX may require those changes to be reflected
    //       in the following code.
    
    struct ld_info *ld_info_ptr, *ldInfoPtr, *result_ld_info_ptr = 0;
    struct filehdr *filehdr_ptr;
    struct aouthdr *aouthdr_ptr;
    struct scnhdr *scnhdr_ptr;
    int i, bufrSz, rc, objectHasEntryPoint;
    
    //
    // the module handle is either the entry point or ptr to data sect.
    // .. save it.
    //
    fEntryPoint = (int(*)())handle().asUnsigned();
    
    //
    // we'll walk the loaded object's xcoff structures for info to
    // support procAddress() .. first call loadquery() (with
    // successively larger buffers if necessary).
    //
    i = 4;
    do {
        bufrSz = i * (sizeof(struct ld_info) + 254);
        if ((ld_info_ptr = (struct ld_info *)malloc(bufrSz)) == 0) {
            ITHROWGUIERROR("IDLLModule: insufficient memory");
        }
        if ((rc = loadquery(L_GETINFO, ld_info_ptr, bufrSz)) == -1) {
            free(ld_info_ptr);
            if (errno == EINVAL) {
                //
                // can't access buffer pointed to by ld_info_ptr
                //
                ITHROWGUIERROR("IDLLModule: memory access error");
            }
        }
        i *= 2;
    } while(rc == -1);
    
    //
    // Find object we loaded above in the loadquery info.
    // The entry point returned by load() actually points to the entry in the
    // data segment.
    //
    for (ldInfoPtr = ld_info_ptr;
         ;
         ldInfoPtr = (struct ld_info *)((char *)ldInfoPtr + ldInfoPtr->ldinfo_next)) {
        if ((unsigned long)fEntryPoint >= (unsigned long)ldInfoPtr->ldinfo_dataorg &&
            (unsigned long)fEntryPoint < ((unsigned long)ldInfoPtr->ldinfo_dataorg +
                                          ldInfoPtr->ldinfo_datasize)) {
            // save the loaded library path name
            char *filename = ldInfoPtr->ldinfo_filename;
            fLoadedPathName = IString(filename);

            // The codes below are for AIX 4.2+ and should be used when we support
            // loading members from an archive. (Charles, May/27/1998)

            /* 
            char *membername = filename + strlen(filename) + 1;
            if (membername[0] != '\0') {
                fLoadedPathName += IString("(") + IString(membername) + IString(")");
            } 
            */

            filehdr_ptr = (struct filehdr *)ldInfoPtr->ldinfo_textorg;
            result_ld_info_ptr = ldInfoPtr;
            break;
        }
        if (ldInfoPtr->ldinfo_next == 0) {
            //
            // couldn't find object in loadquery info .. something's wrong !!
            //
            free(ld_info_ptr);
            ITHROWGUIERROR("loadquery");
        }
    }
    
    if (filehdr_ptr->f_opthdr != 0) {
        aouthdr_ptr = (struct aouthdr *)((char *)filehdr_ptr + FILHSZ);
        objectHasEntryPoint = (aouthdr_ptr->entry != -1);
        //
        // if the object we loaded has an entry point, we compute the
        // loaderBias now .. otherwise we compute it later when we
        // track down the data section.
        //
        if (objectHasEntryPoint) {
            fLoaderBias = (ulong)(fEntryPoint) - aouthdr_ptr->entry;
        }
    } else {
        //
        // object doesn't have optional xcoff hdr .. can't calculate loaderBias
        //
        ITHROWGUIERROR("missing optional xcoff hdr");
    }
    
    scnhdr_ptr = (struct scnhdr *)((char *)filehdr_ptr + FILHSZ + filehdr_ptr->f_opthdr);
    
    //
    // traverse all section headers ..
    //  - if the dll didn't have an entry point, compute loader bias when we
    //    hit the data section
    //  - also, save the pointer to the loader header when we hit the loader section
    //
    for (i=0; i < filehdr_ptr->f_nscns; i++) {
        if ((!objectHasEntryPoint) && ((scnhdr_ptr[i].s_flags & STYP_DATA) != 0)) {
            fLoaderBias = (ulong)(fEntryPoint) - scnhdr_ptr[i].s_vaddr;
        }
        if ((scnhdr_ptr[i].s_flags & STYP_LOADER) != 0) {
            fLoaderHeader = (struct ldhdr *)((char *)filehdr_ptr + scnhdr_ptr[i].s_scnptr);
        }
    }
    
    if (!objectHasEntryPoint && fLoaderBias == 0) {
        //
        // couldn't find data section header to compute loader bias
        //
        ITHROWGUIERROR("IDLLModule: data sect hdr missing");
    }
    
    if (fLoaderHeader == 0) {
        //
        // couldn't find loader section header .. something's wrong !!
        //
        ITHROWGUIERROR("IDLLModule: no loader sect");
    }
    
    //
    // save ptr to symbol table
    //
    fLoaderSymbols = (struct ldsym *)((char *)fLoaderHeader + LDHDRSZ);
    
    //
    // save ptr to resources (if present)
    //
    
    fResources = IMotifDLLImp__findResources(result_ld_info_ptr);
    // if we want to keep result_ld_info_ptr into fLoaderInfo, we have to
    // save a copy ourselves because we want to free up the potentially large
    // ld_info structure with the free() call below.
    
    free(ld_info_ptr);

    ITRACE_DEVELOP(IString("Resource address is ")
                   +IString((unsigned long)fResources).d2x());
}


#endif  // IC_MOTIF



