/*******************************************************************************
* FILE NAME: idllmod.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of the classes declared in           *
*   idllmod.hpp.                                                               *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Revision: 68  1.6.1.2  source/core/base/idllmod.cpp, appsupport, ioc.v400, 000302  

#include <ibase.hpp>
#include <idllmod.hpp>
#include <idllstat.hpp>
#include <idllimp.hpp>
#include <istring.hpp>
#include <irect.hpp>
#include <itrace.hpp>
#include <iexcept.hpp>
#include <ireslock.hpp>

#ifdef IC_PAGETUNE
    #define _IDLLMOD_CPP_
    #include <ipagetun.h>
#endif

/***************************************************************/
/*                                                             */
/*  Reference Counting Strategy.                               */
/*                                                             */
/*  We need to reference count the dlls because the OS (on OS2 */
/*  and Windows) allows you to open the same dll multiple      */
/*  times but kills the dll on the first attempt to close it.  */
/*  Because we can be opening the same DLL from different      */
/*  places, we need to ensure that the DLL doesn't get closed  */
/*  while it is in use elsewhere.                              */
/*                                                             */
/*  Each time we create an IDLLModule via the primary ctor,    */
/*  the ctor constructs an appropriate subclass of             */
/*  IRefCountedDLLImp and uses this object to open the DLL.    */
/*  The ctor then determines if its internal collection of     */
/*  of IRefCountedDLLImp objects (accessed via a call to       */
/*  IDLLStaticPtr::dllSet()) contains an IRefCountedDLLImp     */
/*  object for the same DLL. If it does, the ctor increments   */
/*  the reference count of the existing one and deletes the    */
/*  the newly created one.                                     */
/*                                                             */
/*  The same DLL can be loaded with several different names:   */
/*  1) fully qualified, 2) unqualified via PATH/LIBPATH, and   */
/*  3) unqualified via the NLSPATH.  No matter how the DLL was */
/*  loaded, we need to ensure that we are referencing it via   */
/*  a single copy of IRefCountedDLLImp so that we don't close  */
/*  the DLL while it is still in use.  Because we are chosing  */
/*  to reference count the first instance of a DLL, we loose   */
/*  both the creation name and the NLSPATH flag used to create */
/*  subsequent IDLLModules if they are different than the      */
/*  first one.  This is not seen as a problem because we do    */
/*  not externalize the creation name.  IDLLModule::name       */
/*  always returns the fully qualified name.                   */
/*                                                             */
/***************************************************************/

/*------------------------------------------------------------------------------
| IDLLModule::IDLLModule                                                       |
------------------------------------------------------------------------------*/
IDLLModule::IDLLModule (const char* resourceFileName,
                        ESearchLocation searchLocation )
   : fDLLImp(0),
     fPrivateData(0)    // Not currently used.
{
  IFUNCTRACE_DEVELOP();
  IASSERTPARM(resourceFileName != 0);
  ITRACE_ALL(IString("DLL: ")+resourceFileName);


  //  Determine if we need to search the NLSPath for the DLL.
  bool searchNLSPath = false;
  if(searchLocation == kNLSPathSearch)
    searchNLSPath = true;

  // Construct a platform specific copy of IRefCountedDLLImp to load
  // the dll and to determine if an existing IRefCountedDLLImp exists
  // for this DLL.
  #ifdef IC_WIN
    fDLLImp = new IWindowsDLLImp( resourceFileName,
                                   searchNLSPath );
  #endif

  #ifdef IC_PM
    fDLLImp = new IOS2DLLImp( resourceFileName,
                               searchNLSPath );
  #endif

  #ifdef IC_MOTIF
    fDLLImp = new IMotifDLLImp( resourceFileName,
                                 searchNLSPath );
  #endif

  // We always open the dll.
  open();


  // Lock Resource Table to determine if we are already referencing this
  // DLL in our collection.
  {
    IResourceLock reslibLock( IDLLStaticPtr::dllKey() );

    // See if the dll is already in our collection
    // If in our collection already, use it and update
    // the reference count.
    IRefCountedDLLImp* currentDLLImp = IDLLStaticPtr::dllSet().find(fDLLImp);
    if(currentDLLImp)
    {
       IRefCountedDLLImp* oldDLLImp = fDLLImp;
       fDLLImp = currentDLLImp;
       fDLLImp->addRef();
       delete oldDLLImp;
    }
    else
    // Otherwise, add it to the collection
    {
       IDLLStaticPtr::dllSet().add( fDLLImp );
    }
  }  // Lock resource table
}


/*------------------------------------------------------------------------------
| IDLLModule::IDLLModule                                                       |
|   Copy Constructor                                                           |
------------------------------------------------------------------------------*/
IDLLModule::IDLLModule(const IDLLModule& reslib)
              : fDLLImp(reslib.fDLLImp),
                fPrivateData(0)    // Not currently used.
{
  IFUNCTRACE_DEVELOP();

  if (fDLLImp != 0)
  {
    /* Lock Resource Table */
    IResourceLock reslibLock( IDLLStaticPtr::dllKey() );
    fDLLImp->addRef();
  }
}

#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| IDLLModule::IDLLModule                                                       |
------------------------------------------------------------------------------*/
IDLLModule::IDLLModule   (const IModuleHandle& moduleHandle)
              : fDLLImp(0),
                fPrivateData(0)    // Not currently used.

{
  IFUNCTRACE_DEVELOP();

  /* Lock Resource Table */
  IResourceLock reslibLock( IDLLStaticPtr::dllKey() );

  // See if the dll is already in our collection
  // If in our collection already, use it and update
  // the reference count.
  IRefCountedDLLImp* currentDLLImp = IDLLStaticPtr::dllSet().find(moduleHandle);
  if(currentDLLImp)
  {
    fDLLImp = currentDLLImp;
    fDLLImp->addRef();
  }
  else
  {
      #ifdef IC_WIN
      fDLLImp = new IWindowsDLLImp( moduleHandle );
    #endif
    #ifdef IC_PM
      fDLLImp = new IOS2DLLImp( moduleHandle );
    #endif
    #ifdef IC_MOTIF
      fDLLImp = new IMotifDLLImp( moduleHandle );
    #endif
    IDLLStaticPtr::dllSet().add( fDLLImp );
    fDLLImp->addRef();
  }
}
#endif  // IC_PMWIN

/*------------------------------------------------------------------------------
| IDLLModule::operator=                                                        |
------------------------------------------------------------------------------*/
IDLLModule& IDLLModule::operator= (const IDLLModule& reslib)
{
  IFUNCTRACE_DEVELOP();

  if (this != &reslib)
  {
    IResourceLock reslibLock( IDLLStaticPtr::dllKey() );
    if (reslib.fDLLImp != 0)
      reslib.fDLLImp->addRef();

    if (fDLLImp != 0)
    {
      fDLLImp->removeRef();
      if (fDLLImp->useCount() == 0)
      {
        IDLLStaticPtr::dllSet().remove( fDLLImp );
        delete fDLLImp;
        fDLLImp = 0;
      }
    }
    fDLLImp = reslib.fDLLImp;
  }

  return( *this);
}



/*------------------------------------------------------------------------------
| IDLLModule::~IDLLModule                                                      |
------------------------------------------------------------------------------*/
IDLLModule::~IDLLModule()
{
  IFUNCTRACE_DEVELOP();

  if (handle())
    close();

  if ( (fDLLImp != 0) && (fDLLImp->useCount() == 0) )
  {
    if ( IDLLStaticPtr::dllSet().find ( fDLLImp ) )
    {
      IDLLStaticPtr::dllSet().remove( fDLLImp );
      delete fDLLImp;
    }
    fDLLImp = 0;
  }
}

/*------------------------------------------------------------------------------
| IDLLModule::open                                                             |
------------------------------------------------------------------------------*/
IDLLModule& IDLLModule::open()
{
  IFUNCTRACE_DEVELOP();

  if (fDLLImp != 0)
    fDLLImp->addRef();

  return( *this );
}

/*------------------------------------------------------------------------------
| IDLLModule::close                                                            |
------------------------------------------------------------------------------*/
IDLLModule& IDLLModule::close()
{
  IFUNCTRACE_DEVELOP();

  /* free the reference to the DLL */
  if (fDLLImp)
    fDLLImp->removeRef();

  return( *this );
}


/*------------------------------------------------------------------------------
| IDLLModule::fileName                                                         |
------------------------------------------------------------------------------*/
IString IDLLModule::fileName() const
{
  return fDLLImp->loadedPathName();
}



/*------------------------------------------------------------------------------
| IDLLModule::isOpen                                                           |
------------------------------------------------------------------------------*/
bool IDLLModule::isOpen() const
{
  if ( (fDLLImp != 0) && (fDLLImp->handle() != 0) )
    return true;
  else
    return false;
}


/*------------------------------------------------------------------------------
| IDLLModule::addRef                                                           |
------------------------------------------------------------------------------*/
unsigned long IDLLModule::addRef(const IModuleHandle& moduleHandle)
{
  IFUNCTRACE_DEVELOP();

  IResourceLock reslibLock( IDLLStaticPtr::dllKey() );

  unsigned long useCount = 0;
  IRefCountedDLLImp* currentDLLImp = IDLLStaticPtr::dllSet().find(moduleHandle);
  if(currentDLLImp)
  {
    useCount = currentDLLImp->addRef();
  }
  return( useCount );
}


/*------------------------------------------------------------------------------
| IDLLModule::removeRef                                                        |
------------------------------------------------------------------------------*/
unsigned long IDLLModule::removeRef(const IModuleHandle& moduleHandle)
{
  IFUNCTRACE_DEVELOP();

  IResourceLock reslibLock( IDLLStaticPtr::dllKey() );

  unsigned long useCount =0;
  IRefCountedDLLImp* currentDLLImp = IDLLStaticPtr::dllSet().find(moduleHandle);
  if(currentDLLImp)
  {
    useCount = currentDLLImp->removeRef();
  }
  return( useCount );
}


/*------------------------------------------------------------------------------
| IDLLModule::procAddress                                                      |
|   Get proceudure address by name                                             |
------------------------------------------------------------------------------*/
void* IDLLModule::procAddress(const char* procedureName) const
{
  IFUNCTRACE_DEVELOP();
  IASSERT(handle() != 0);

  return fDLLImp->procAddress(procedureName);
}

#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| IDLLModule::procAddress                                                      |
|  Get procedure address from Ordinal                                          |
------------------------------------------------------------------------------*/
void* IDLLModule::procAddress(unsigned long procedureOrdinal) const
{
  IFUNCTRACE_DEVELOP();
  IASSERT(handle() != 0);

  return fDLLImp->procAddress(procedureOrdinal);
}
#endif  // IC_PMWIN

#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| IDLLModule::isEntryPoint32Bit                                                |
------------------------------------------------------------------------------*/
bool IDLLModule::isEntryPoint32Bit
                                 ( const char *procedureName ) const
{
  IFUNCTRACE_DEVELOP();

  IRefCountedDLLImp::ProcType procType = fDLLImp->procType(procedureName);
  if(procType == IRefCountedDLLImp::k32Bit)
     return true;
  else
     return false;
}
#endif // IC_PMWIN

#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| IDLLModule::isEntryPoint32Bit                                                |
------------------------------------------------------------------------------*/
bool IDLLModule::isEntryPoint32Bit( unsigned long procedureOrdinal ) const
{
  IFUNCTRACE_DEVELOP();

  IRefCountedDLLImp::ProcType procType = fDLLImp->procType(procedureOrdinal);
  if(procType == IRefCountedDLLImp::k32Bit)
     return true;
  else
     return false;
}
#endif // IC_PMWIN

/*------------------------------------------------------------------------------
| IDLLModule::handle                                                           |
|  Return the Handle used to load Resources                                    |
------------------------------------------------------------------------------*/
IModuleHandle IDLLModule::handle() const
{
  if (fDLLImp != 0)
    return fDLLImp->handle();
  else
    return IModuleHandle(0);
}

