//
//  FILE NAME: iextproc.cpp
//
//  DESCRIPTION:
//
//  This module implements the IExternalProcess class. This class is actually
//  implemented in terms of an underlying implementation class that
//  hides the many differences between the invocation of processes and
//  the tracking of them on the different platforms. This implementation
//  is accessed via the pure virtual IProcImplBase class.
//
//
//  COPYRIGHT:
//    IBM Open Class Library
//    Licensed Materials - Property of IBM
//
//    5645-001
//    (C) Copyright IBM Corporation 1992, 1997
//
//    US Government Users Restricted Rights - Use, duplication, or
//    disclosure restricted by GSA ADP Schedule Contract with IBM
//    Corp.
// ----------------------------------------------------------------------------


// ----------------------------------------------------------------------------
//  Includes
// ----------------------------------------------------------------------------
#include    <icconsta.h>
#include    <iexcept.hpp>
#include    <iprcenv.hpp>
#include    <iextproc.hpp>

// Include the implementation derivative for the current platform
#ifdef      IC_WIN
#include    <iwextprc.hpp>
#endif

#ifdef      IC_PM
#include    <OS2.H>
#include    <ioextprc.hpp>
#endif

#ifdef		IC_POSIX
#include	<ipextprc.hpp>
#endif


#ifdef IC_PAGETUNE
    #define _IEXTPROC_CPP_
    #include <ipagetun.h>
#endif

// ----------------------------------------------------------------------------
//  CLASS: IExtProcImplBase
// ----------------------------------------------------------------------------
IExtProcImplBase::IExtProcImplBase() :

    fId(0)
{
}

IExtProcImplBase::~IExtProcImplBase()
{
}

void IExtProcImplBase::setPath(const IPathName& newPath)
{
    // Don't allow if already started
    throwIfRunning(IC_EXTP_SETPATH);

    fPath = newPath;
}

void IExtProcImplBase::setEnvironment(const IEnvironment& newEnv)
{
    // Don't allow if already started
    throwIfRunning(IC_EXTP_SETENV);

    // Assign the new environment
    fEnviron = newEnv;
}


void IExtProcImplBase::throwIfRunning(const unsigned long reason) const
{
    // If its already running, then don't allow another
    if (isRunning())
    {
        IMessageText errText(IC_EXTP_ALREADYSTARTED,
                             kMsgFileName,
                             IPathName::defaultParser().formatName(getPath()),
                             IMessageText(reason, kMsgFileName));

        IInvalidRequest toThrow(errText,
                                IC_EXTP_ALREADYSTARTED,
                                IException::recoverable);
        ITHROW(toThrow);
    }
}

// ----------------------------------------------------------------------------
//  CLASS: IExternalProcess
// ----------------------------------------------------------------------------
IExternalProcess::IExternalProcess() :

    fImpl(0)
{
    // Create an implementation object of the correct sort
    fImpl = makeImplObj();
}

IExternalProcess::~IExternalProcess()
{
    delete fImpl;
}

void IExternalProcess::start(const IPathName& path)
{
    // Don't allow if already started
    fImpl->throwIfRunning(IC_EXTP_START);

    // Set the path on the implementation object first
    fImpl->setPath(path);
    fImpl->start(0, 0);
}

void IExternalProcess::start()
{
    // Don't allow if already started
    fImpl->throwIfRunning(IC_EXTP_START);

    fImpl->start(0, 0);
}

void IExternalProcess::start( const   IPathName&      path
                            , const IText*          argArray
                            , const unsigned long   argCount)
{
    // Don't allow if already started
    fImpl->throwIfRunning(IC_EXTP_START);

    // Set the passed path first
    fImpl->setPath(path);

    // And now call the implementation object, passing on the arguments
    fImpl->start(argArray, argCount);
}

void IExternalProcess::start(const IText* argArray, const unsigned long argCount)
{
    // Don't allow if already started
    fImpl->throwIfRunning(IC_EXTP_START);

    fImpl->start(argArray, argCount);
}

IExtProcImplBase* IExternalProcess::makeImplObj()
{
    //
    //  According to the current platform, create an implementation
    //  derivative of the correct sort.
    //
    #ifdef IC_WIN
    return new IWinExtProcImpl;
    #endif

    #ifdef IC_PM
    return new IOS2ExtProcImpl;
    #endif

    #ifdef IC_POSIX
    return new IPosixExtProcImpl;
    #endif
}
