/*----------------------------------------------------------------------------*/
/* FILE NAME: inlsutil.cpp                                                    */
/*                                                                            */
/* DESCRIPTION:                                                               */
/*   This file contains the implementation of the NLS utility classes and     */
/*   functions, declared in inlsutil.hpp.                                     */
/*                                                                            */
/*                                                                            */
/* COPYRIGHT:                                                                 */
/*   IBM Open Class Library                                                   */
/*   Licensed Materials - Property of IBM                                     */
/*                                                                            */
/*   5645-001                                                                 */
/*   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.           */
/*                                                                            */
/*   US Government Users Restricted Rights - Use, duplication, or             */
/*   disclosure restricted by GSA ADP Schedule Contract with IBM Corp.        */
/*                                                                            */
/*----------------------------------------------------------------------------*/
#ifdef __MVS__
    #pragma csect(code,"INLSUTIL")
    #pragma csect(static,"SNLSUTIL")
    #pragma comment (copyright, \
    "Licensed Materials - Property of IBM  5645-001 (C) \
    Copyright IBM Corp. 1992, 1997  All Rights Reserved. \
    US Government Users Restricted Rights - Use, duplication, or \
    disclosure restricted by GSA ADP Schedule Contract with IBM Corp.")
#endif
#pragma priority( -2147482024 )


#include "inlsutil.hpp"

#include <iexcept.hpp>
#include <icconsta.h>

#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef IC_PAGETUNE
    #define _INLSUTIL_CPP_
    #include <ipagetun.h>
#endif

#ifdef IC_HAS_RESTARTABLE
  #define ASC_MBLEN(str, n, ps)            ::mbrlen(str, n, ps)
  #define ASC_MBTOWC(pwc, str, n, ps)      ::mbrtowc(pwc, str, n, ps)
  #define ASC_WCTOMB(str, wc, ps)          ::wcrtomb(str, wc, ps)
  #define ASC_CVTMB2WC(str, n, pwc, ps)    cvtMB2WC(str, n, pwc, ps)
#else
  #define ASC_MBLEN(str, n, ps)            ::mblen(str, n)
  #define ASC_MBTOWC(pwc, str, n, ps)      ::mbtowc(pwc, str, n)
  #define ASC_WCTOMB(str, wc, ps)          ::wctomb(str, wc)
  #define ASC_CVTMB2WC(str, n, pwc, ps)    cvtMB2WC(str, n, pwc)
#endif

#ifdef IC_MVS400
  static const char SO = '\x0e';         // shift-out character
  static const char SI = '\x0f';         // shift-in  character
#endif

#ifdef IC_NLS
// IWBuffer's constructor ------------------------------------------
//
// - 'str' is a single null terminated multibyte string
// - the converted wide character string is also null terminated
// - charnum does not include the null terminated character
//
IWBuffer::IWBuffer(const char *str) {

  size_t len = ::strlen(str) + 1;
  wptr = new wchar_t[len];

  // charnum does not include the null terminated character
  charnum = cvtMBS2WCS(str, len, wptr) - 1;
}

// IWBuffer's constructor ------------------------------------------
//
// - 'buf' can be multibyte character stream with zero or
//   more null character embedded, no null terminated character
//   is assumed, conversion is done up to 'blen' bytes
// - if 'blen' is zero, a null wide character is created and
//   charnum has value zero.
//
IWBuffer::IWBuffer(const char *buf, unsigned blen) {

  if (blen == 0) {
    wptr = new wchar_t[1];
    *wptr = L'\0';
    charnum = 0;
  } else {
    wptr = new wchar_t[blen];
    charnum = cvtMBS2WCS(buf, blen, wptr);
  }
}

// IWBuffer's destructor -------------------------------------------
//
IWBuffer::~IWBuffer(void) {
  delete[] wptr;
}

// IWBuffer::positionOf --------------------------------------------
//
// - return the character position where the 'wBuf' is matched.
// - return 0 if the search failed or 'startNum' is out of range.
//
unsigned IWBuffer::positionOf(IWBuffer &wBuf,
                              unsigned startNum) const {

  if ((startNum == 0) || (startNum > charnum)) return 0;

  size_t searchCharNum = wBuf.charNum();

  // Do we have any thing to search for?
  if (searchCharNum == 0) return 0;

  // Does search buffer have more characters than the remaining
  // of the original buffer starting from 'startNum'?
  if (searchCharNum > (charnum-startNum+1)) return 0;

  wchar_t *wp, *endwp;

  for (wp = wptr + startNum - 1,
       endwp = wptr + charnum - searchCharNum;
       wp <= endwp;
       ++wp, ++startNum) {
    if (!::wcsncmp(wp, wBuf, searchCharNum))
      return startNum;
  }
  return 0;
}

// IWBuffer::positionOfAnyOf ---------------------------------------
//
// - return the character position where any one of the characters
//   in 'wBuf' is matched.
// - return 0 if the search failed or 'startNum' is out of range.
//
unsigned IWBuffer::positionOfAnyOf(IWBuffer &wBuf,
                                   unsigned startNum) const {

  if ((startNum == 0) ||
      (startNum > charnum) ||
      (wBuf.charNum() == 0))
   return 0;

  wchar_t *wp, *endwp;
  for (wp = wptr + startNum - 1,
       endwp = wptr + charnum - 1;
       wp <= endwp;
       ++wp, ++startNum) {
    if (wBuf.hasWChar(*wp))
      return startNum;
  }
  return 0;
}

// IWBuffer::positionOfAnyBut --------------------------------------
//
// - return the character position where any one of the characters
//   in 'wBuf' is not matched.
// - return 0 if the search failed or 'startNum' is out of range.
//
unsigned IWBuffer::positionOfAnyBut(IWBuffer &wBuf,
                                    unsigned startNum) const {

  if ((startNum == 0) || (startNum > charnum)) return 0;

  wchar_t *wp, *endwp;
  for (wp = wptr + startNum - 1,
       endwp = wptr + charnum - 1;
       wp <= endwp;
       ++wp, ++startNum) {
    if (!wBuf.hasWChar(*wp))
      return startNum;
  }
  return 0;
}

// IWBuffer::lastPositionOf ----------------------------------------
//
// - return the character position where the 'wBuf' is matched
//   by searching backward.
// - return 0 if the search failed or 'startNum' is 0.
//
unsigned IWBuffer::lastPositionOf(IWBuffer &wBuf,
                                  unsigned startNum) const {
  if (startNum == 0) return 0;

  size_t searchCharNum = wBuf.charNum();

  if ((searchCharNum == 0) ||
      (searchCharNum > charnum))
    return 0;

  if (startNum > (charnum-searchCharNum+1))
    startNum = charnum - searchCharNum + 1;

  for (wchar_t *wp = wptr + startNum - 1;
       wptr <= wp;
       --wp, --startNum) {
    if (!::wcsncmp(wp, wBuf, searchCharNum))
      return startNum;
  }
  return 0;
}

// IWBuffer::lastPositionOfAnyOf -----------------------------------
//
// - return the character position where any one of the characters
//   in 'wBuf' is matched by searching backward.
// - return 0 if the search failed or 'startNum' is 0.
//
unsigned IWBuffer::lastPositionOfAnyOf(IWBuffer &wBuf,
                                       unsigned startNum) const {
  if ((startNum == 0) ||
      (wBuf.charNum() == 0))
    return 0;

  if (startNum > charnum)
    startNum = charnum;

  for (wchar_t *wp = wptr + startNum - 1;
       wptr <= wp;
       --wp, --startNum) {
    if (wBuf.hasWChar(*wp))
      return startNum;
  }
  return 0;
}

// IWBuffer::lastPositionOfAnyBut ----------------------------------
//
// - return the character position where any one of the characters
//   in 'wBuf' is not matched by searching backward.
// - return 0 if the search failed or 'startNum' is 0.
//
unsigned IWBuffer::lastPositionOfAnyBut(IWBuffer &wBuf,
                                        unsigned startNum) const {
  if (startNum == 0) return 0;

  if (startNum > charnum)
    startNum = charnum;

  for (wchar_t *wp = wptr + startNum - 1;
       wptr <= wp;
       --wp, --startNum) {
    if (!wBuf.hasWChar(*wp))
      return startNum;
  }
  return 0;
}

// IWBuffer::reverse -----------------------------------------------
//
// - reverse the buffer contents.
//
void IWBuffer::reverse() const {
  wchar_t *wpStart = wptr;
  wchar_t *wpEnd   = wptr + charnum - 1;
  wchar_t wTemp;

  while (wpStart < wpEnd) {
    wTemp      = *wpEnd;
    *wpEnd--   = *wpStart;
    *wpStart++ = wTemp;
  }
}

// IWBuffer::hasWChar ----------------------------------------------
//
// - return 1 if IWBuffer has 'wCh'; 0 otherwise
//
int IWBuffer::hasWChar(wchar_t wCh) const {
  wchar_t *wp = wptr;
  for (int i = 1; i <= charnum; ++i) {
    if (*wp++ == wCh)
      return 1;
  }
  return 0;
}

// findCharNum -----------------------------------------------------
//
// - given a buffer, its length and the byte position,
//   the corresponding character position is returned.
// - return 0 if the given byte position is 0, greater
//   than the length of the string or invalid multibyte
//   character found in the buffer.
//
unsigned findCharNum(const char *buf, unsigned blen, unsigned pos) {

  if ((buf == NULL) || (pos == 0) || (pos > blen))
    return 0;

  const char *p = buf;
  unsigned idx = 0;
  unsigned numCnt = 0;
  int charlen;

#ifdef IC_HAS_RESTARTABLE
  mbstate_t ss = 0;
#endif

  while (idx < pos) {
    charlen = ASC_MBLEN(p, MB_CUR_MAX, &ss);
    switch (charlen) {
      case 0:
        ++idx; ++p; ++numCnt;
        break;
      case -1:
        return 0;
      case -2:
        // redundant SO/SI found, don't have a full character yet
        idx += MB_CUR_MAX;
        p   += MB_CUR_MAX;
        break;
      default:
        idx += charlen;
        p   += charlen;
        ++numCnt;
    }
  }

  // If 'pos' is pointing to a location that is right after
  // redundant SO/SI are found, we need to increment the 'numCnt'
  // in order to have the correct character count.
  if (charlen == -2)
    return numCnt+1;
  else
    return numCnt;
}

// findNextCharNum -----------------------------------------------------
//
// - given a buffer, its length and the byte position,
//   if the given byte position is pointing at the
//   beginning of the character, the corresponding character
//   position is returned; if it points to the middle
//   of the multibyte character, next character position
//   is returned.
// - return 0 if the given byte position is 0, greater
//   than the length of the string, invalid multibyte
//   character found in the buffer or 'pos' is pointing
//   at the middle of the last multibyte character.
//
unsigned findNextCharNum(const char *buf, unsigned blen, unsigned pos) {

  if ((buf == NULL) || (pos == 0) || (pos > blen))
    return 0;

  const char *p = buf;
  unsigned previdx = 0, idx = 0;
  unsigned numCnt = 0;
  int charlen;
  int redundantSeen = 0;

#ifdef IC_HAS_RESTARTABLE
  mbstate_t ss = 0;
#endif

  while (idx < pos) {
    charlen = ASC_MBLEN(p, MB_CUR_MAX, &ss);
    switch (charlen) {
      case 0:
        // If redundant SO/SI is seen, reset 'redundantSeen' to 0 (not seen);
        // otherwise keep track the current index in 'previdx'.
        if (redundantSeen) {
          redundantSeen = 0;
        } else {
          previdx = idx;
        }
        ++idx; ++p; ++numCnt;
        break;
      case -1:
        // return 0 if the given string has invalid character in it
        return 0;
     case -2:
        redundantSeen = 1;
        idx += MB_CUR_MAX;
        p   += MB_CUR_MAX;

#ifdef IC_MVS400
        // Have we read SOSISOSI... situation ?
        // (i.e. if the last byte read is SI, then we are)
        if (*(p-1) == SI) {
          previdx = idx;
        }
#endif
        break;
     default:
        // If redundant SO/SI is seen, reset 'redundantSeen' to 0 (not seen);
        // otherwise keep track the current index in 'previdx'.
        if (redundantSeen) {
          redundantSeen = 0;
        } else {
          previdx = idx;
        }
        idx += charlen;
        p   += charlen;
        ++numCnt;
    }
  }
  // If 'pos' is pointing to a location that is right after
  // redundant SO/SI are found, we need to increment the 'numCnt'
  // in order to have the correct character count.
  if (charlen == -2)
    ++numCnt;

  // If 'pos' is pointing to the beginning of the character,
  // 'numCnt' is the return value.
  if (pos == (previdx+1)) {
    return numCnt;
  } else {
    // 'pos' is pointing to the middle of the character ...

#ifdef IC_MVS400
    // If 'pos' is pointing a shift-out or shift-in character and
    // we are in a middle of a series of redundant SO/SI, 'numCnt'
    // is the correct character position to return because it is
    // considered as "pointing to the beginning of the character".
    // Otherwise, we need to increment 'numCnt' and return the
    // next character position. We need to call findBytePos() in
    // order to figure out whether 'pos' is indeed pointing to
    // the middle of the last character. In this case, 0 is returned.
    if ((charlen == -2) && ((buf[pos-1] == SO) || (buf[pos-1] == SI))) {
      return numCnt;
    } else {
      ++numCnt;
      if (findBytePos(buf, blen, numCnt)) {
        return numCnt;
      } else { // out of range
        return 0;
      }
    }
#else
    // Increment 'numCnt' in order to get the character position
    // of the next character. Also, we need to call findBytePos() in
    // order to figure out whether 'pos' is indeed pointing to
    // the middle of the last character. In this case, 0 is returned.
    ++numCnt;
    if (findBytePos(buf, blen, numCnt)) {
      return numCnt;
    } else { // out of range
      return 0;
    }
#endif
  }
}

// findBytePos -----------------------------------------------------
//
// - given a buffer, its length and the character position,
//   the beginning of the corresponding byte position is returned.
// - return 0 if the given character position is 0, greater
//   than the number of characters or invalid multibyte
//   character found in the buffer.
//
unsigned findBytePos(const char *buf, unsigned blen, unsigned num) {

  if ((buf == NULL) || (blen == 0) || (num == 0))
    return 0;

  const char *p = buf;
  unsigned idx = 1;
  unsigned numCnt = 1;
  int charlen;

#ifdef IC_HAS_RESTARTABLE
  mbstate_t ss = 0;
#endif

  while ((idx < blen) && (numCnt < num)) {
    charlen = ASC_MBLEN(p, MB_CUR_MAX, &ss);
    switch (charlen) {
      case 0:
        ++idx; ++p; ++numCnt;
        break;
      case -1:
        return 0;
      case -2:
        idx += MB_CUR_MAX;
        p   += MB_CUR_MAX;
        break;
      default:
        idx += charlen;
        p   += charlen;
        ++numCnt;
    }
  }

  // If 'idx' does not exceed the given buffer length and 'numCnt'
  // is the same as 'num', 'idx' is the beginning of the corresponding
  // byte position.
  if ((idx <= blen) && (numCnt == num)) {
#ifdef IC_MVS400
    // call redundantSOSI() to skip all the redundant SO/SI at the front
    // so that the returned index is always pointing to a character
    // rather than redundant SO/SI (Note: it is still possible pointing
    // to the shift-out character).
    return redundantSOSI(buf+idx-1) + idx;
#else
    return idx;
#endif
  } else {
    return 0;
  }
}

// findCurPos ------------------------------------------------------
//
// - given a buffer, its length and the byte position (can be
//   pointed to any part of the character), the beginning of the
//   corresponding byte position for this character is returned.
// - return 0 if the given byte position is 0, greater
//   than the length of the string or invalid multibyte
//   character found in the buffer.
//
unsigned findCurPos(const char *buf, unsigned blen, unsigned pos) {

  if ((buf == NULL) || (pos == 0) || (pos > blen))
    return 0;

  const char *p = buf;
  unsigned curidx = 1, idx = 1;
  int charlen;
  int redundantSeen = 0;

#ifdef IC_HAS_RESTARTABLE
  mbstate_t ss = 0;
#endif

  do {
    charlen = ASC_MBLEN(p, MB_CUR_MAX, &ss);
    switch (charlen) {
      case 0:
        // If redundant SO/SI is seen, reset 'redundantSeen' to 0 (not seen);
        // otherwise keep track the current index in 'curidx'.
        if (redundantSeen) {
          redundantSeen = 0;
        } else {
          curidx = idx;
        }
        ++idx; ++p;
        break;
      case -1:
        // return 0 if the given string has invalid character in it
        return 0;
      case -2:
        // If redundant SO/SI has not been seen, keep track the current
        // index in 'curidx' and then set 'redundantSeen' to 1 to flag
        // that redundant SO/SI is seen now so that if we continue seeing
        // redundant SO/SI, we will not change the value of 'curidx'.
        if (!redundantSeen) {
          curidx = idx;
          redundantSeen = 1;
        }
        idx += MB_CUR_MAX;
        p   += MB_CUR_MAX;
        break;
      default:
        // If redundant SO/SI is seen, reset 'redundantSeen' to 0 (not seen);
        // otherwise keep track the current index in 'curidx'.
        if (redundantSeen) {
          redundantSeen = 0;
        } else {
          curidx = idx;
        }
        idx += charlen;
        p   += charlen;
    }
  } while (idx <= pos);

#ifdef IC_MVS400
  // call redundantSOSI() to skip all the redundant SO/SI at the front
  // so that the returned index is always pointing to a character
  // rather than redundant SO/SI (Note: it is still possible pointing
  // to the shift-out character).
  return redundantSOSI(buf+curidx-1) + curidx;
#else
  return curidx;
#endif
}

// findNextPos -----------------------------------------------------
//
// - given a buffer, its length and the byte position (can be
//   pointed to any part of the character), the byte position
//   after the current character is returned.
// - return 0 if the given byte position is 0, greater
//   than the length of the string, invalid multibyte
//   character found in the buffer or 'pos' is pointing
//   at the last character (i.e. no next character).
//
unsigned findNextPos(const char *buf, unsigned blen, unsigned pos) {

  if ((buf == NULL) || (pos == 0) || (pos > blen))
    return 0;

  const char *p = buf;
  unsigned idx = 1;
  int charlen;

#ifdef IC_HAS_RESTARTABLE
  mbstate_t ss = 0;
#endif

  do {
    charlen = ASC_MBLEN(p, MB_CUR_MAX, &ss);
    switch (charlen) {
      case 0:
        ++idx; ++p;
        break;
      case -1:
        return 0;
      case -2:
        idx += MB_CUR_MAX;
        p   += MB_CUR_MAX;
        break;
      default:
        idx += charlen;
        p   += charlen;
    }
  } while ((idx <= pos) || (charlen == -2));

  // If 'idx' has gone pass the buffer length, 0 is returned;
  // otherwise, 'idx' is the position of the next character.
  if (idx > blen) {
    return 0;
  } else {
#ifdef IC_MVS400
  // call redundantSOSI() to skip all the redundant SO/SI at the front
  // so that the returned index is always pointing to a character
  // rather than redundant SO/SI (Note: it is still possible pointing
  // to the shift-out character).
    return redundantSOSI(buf+idx-1) + idx;
#else
    return idx;
#endif
  }
}

#ifdef IC_MVS400
// redundantSOSI ---------------------------------------------------
//
// - given a buffer, it returns the number of redundant SO/SI
//   characters can be found at the beginning (SO and SI are
//   treated as two bytes).
//
unsigned redundantSOSI(const char *buf) {
  unsigned count = 0;
  while ((*buf == SO) && (*(buf+1) == SI)) {
    count += 2;
    buf += 2;
  }
  return count;
}
#endif

// cvtMB2WC --------------------------------------------------------
//
// - convert the given multibyte character buffer 'pMB' from at most
//   'maxLen' number of bytes to a wide character and store it to
//   the location pointed to by 'pWC'
// - 'ps' contains the shift state for restartable
// - this function assumes that it is always possible to convert
//   a valid character to wide character from 'maxLen' of bytes
// - return 1 if null character is converted
//   (for EBCDIC, return 1 + the length of redundant SO/SI, if any)
// - return 0 if all characters in 'pMB' are redundant SO/SI
// - exception will be thrown if invalid or incomplete multibyte is detected
// - otherwise, the number of bytes in 'pMB' required for
//   conversion is returned
//
#ifdef IC_HAS_RESTARTABLE
  size_t cvtMB2WC(const char *pMB, unsigned maxLen, wchar_t *pWC,
                  mbstate_t *ps)
#else
  size_t cvtMB2WC(const char *pMB, unsigned maxLen, wchar_t *pWC)
#endif
{
  int len;

  len = ASC_MBTOWC(pWC, pMB, maxLen, ps);
  switch (len) {
    case 0:
#ifdef IC_MVS400
      // need to call strlen in case of redundant SOSIs before null
      return(::strlen(pMB)+1);
#else
      return(1);
#endif

    case -1:
#ifdef IC_DEVELOP
      {
        const char *p = pMB;
        for (int i = 0; i < maxLen; ++i) {
          printf("%02x", *p++);
        }
        printf("\n");
        fflush(stdout);
      }
#endif

      // Invalid multibyte is found, throw an exception so that
      // user can aware of it and take action.
#ifdef IC_MOTIF
  {
  IInvalidRequest exception(IMessageText(IC_INVALID_MBCS_STRING),
                           IC_INVALID_MBCS_STRING, IException::unrecoverable);
  ITHROW( exception );
  }
#else
      ITHROWLIBRARYERROR(IC_INVALID_MBCS_STRING,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable);
#endif

    case -2:
#ifdef IC_MVS400
      if (pMB[maxLen-1] == SI) {
        // If the last byte is SI, that means we have redundant SO/SI.
        // In this case, we return zero. Otherwise, incomplete character
        // is found at the end, exception is thrown.
        return 0;
      } else {
#endif

#ifdef IC_DEVELOP
      {
        const char *p = pMB;
        for (int i = 0; i < maxLen; ++i) {
          printf("%02x", *p++);
        }
        printf("\n");
        fflush(stdout);
      }
#endif

        // Incomplete multibyte character is found, throw an exception so that
        // user can aware of it and take action.
#ifdef IC_MOTIF
  {
  IInvalidRequest exception(IMessageText(IC_INVALID_MBCS_STRING),
                           IC_INVALID_MBCS_STRING, IException::unrecoverable);
  ITHROW( exception );
  }
#else
        ITHROWLIBRARYERROR(IC_INVALID_MBCS_STRING,
                           IBaseErrorInfo::invalidRequest,
                           IException::recoverable);
#endif
#ifdef IC_MVS400
      }
#endif

    default:
      // The length that is required to convert this
      // character is returned.
      return len;
  }
}

// cvtMBS2WCS ------------------------------------------------------
//
// - convert the given multibyte character buffer 'pMBS' from
//   'mblen' number of bytes to a stream of wide characters
//   and store them to the buffer pointed to by 'pWCS'
// - if conversion successful, the number of characters being
//   converted is returned; otherwise, exception will be thrown
//
size_t cvtMBS2WCS(const char *pMBS, unsigned mblen, wchar_t *pWCS) {

  if ((pMBS == NULL) || (mblen == 0) || (pWCS == NULL))
    return 0;

  const char *pmbs = pMBS;
  wchar_t *pwcs = pWCS;
  size_t len, length = 0;

#ifdef IC_HAS_RESTARTABLE
  mbstate_t ss = 0;
#endif

  while (length < mblen) {
    len = ASC_CVTMB2WC(pmbs, mblen-length, pwcs, &ss);
    if (len) {
      length += len;
      pmbs = pMBS + length;
      ++pwcs;
    } else {
      // redundant SO/SI is found at the end
      break;
    }
  }
  return(pwcs-pWCS);
}

// cvtWCS2MBS ------------------------------------------------------
//
// - convert 'numWChar' wide characters from 'pWCS' to a stream
//   of multibyte characters and store them to the buffer pointed
//   to by 'pMBS'.
// - it first attempts to convert 'numWChar' number of wide characters
//   to multibyte characters, if the converted length has exceeded
//   'mblen', then if 'alwaysPad' is 0, the converted length 'mbcnt'
//   or 'mbcnt'+1 is returned, depend on whether inserting extra SI
//   is required; otherwise, the remaining spaces are padded with
//   'padChar' and 'mblen' is returned.
// - if 'mblen' is greater than the converted length after all
//   'numWChar' are converted, then if 'alwaysPad' is 0, the converted
//   length 'mbcnt' is returned; otherwise, the remaining spaces
//   are padded with 'padChar' and 'mblen' is returned.
//
// Note of usage:-
//
// - convert all the given wide characters, use
//
//     cvtWCS2MB(pWCS, numWChar, pMBS);
//
// - convert the given wide characters up to the last full character
//   that does not exceed the given length without padding, use
//
//     cvtWCS2MB(pWCS, numWChar, pMBS, mblen);
//
// - convert the given wide characters up to the last full character
//   that does not exceed the given length with padding blanks, use
//
//     cvtWCS2MB(pWCS, numWChar, pMBS, mblen, 1);
//
// - convert the given wide characters up to the last full character
//   that does not exceed the given length with padding characters, use
//
//     cvtWCS2MB(pWCS, numWChar, pMBS, mblen, 1, padChar);
//
size_t cvtWCS2MBS(const wchar_t *pWCS, unsigned numWChar, char *pMBS,
                  unsigned mblen, int alwaysPad, char padChar) {

  if ((pWCS == NULL) || (pMBS == NULL) || (mblen == 0))
    return 0;

  char mbchar[MB_LEN_MAX];
  const wchar_t *pwcs = pWCS;
  char *pmbs = pMBS;
  int len, mbcnt, wccnt;

#ifdef IC_HAS_RESTARTABLE
  mbstate_t ss = 0;
#endif

  for (len = 0, mbcnt = 0, wccnt = 0;
       wccnt < numWChar;
       ++wccnt, ++pwcs) {
    len = ASC_WCTOMB(mbchar, *pwcs, &ss);
    if (len == -1) {
#ifdef IC_MOTIF
  IInvalidRequest exception("cvtWCS2MBS", 0, IException::recoverable);
  ITHROW( exception );
#else
      ITHROWCLIBERROR("cvtWCS2MBS",
                      IBaseErrorInfo::invalidRequest,
                      IException::recoverable);
#endif
    }

    // Can we put the current character within 'mblen' ?
    if ((mbcnt+len) < mblen) {
      // Yes, store it to the multibyte buffer and continue ...

      ::memcpy(pmbs, mbchar, len);
      mbcnt += len;
      pmbs  += len;
    } else {
      // No, fill up the last couple of byte(s)

#ifdef IC_MVS400
      if (len == 1) {
        *pmbs = *mbchar;
      } else {
        if (alwaysPad) {
          if (((mbcnt+len) == mblen) &&
              (*mbchar == SI)) {
            ::memcpy(pmbs, mbchar, len);
          } else {
            ::memset(pmbs, padChar, mblen-mbcnt);

            if (*mbchar != SO) {
              *pmbs = SI;
            }
          }
        } else {
          if (*mbchar == SO) {
            return mbcnt;
          } else {
            if (((mbcnt+len) == mblen) &&
                (*mbchar == SI)) {
              ::memcpy(pmbs, mbchar, len);
            } else {
              *pmbs = SI;
              return mbcnt+1;
            }
          }
        }
      }

#else
      // Does the last converted character fit into 'mblen' ?
      if ((mbcnt+len) == mblen) {
        // Yes, just fill the character
        ::memcpy(pmbs, mbchar, len);
      } else {

        if (alwaysPad) {
          ::memset(pmbs, padChar, mblen-mbcnt);
        } else {
          return mbcnt;
        }
      }
#endif

      // If full charcter fit into 'mblen' or 'alwaysPad' is on,
      // 'mblen' is returned.
      return mblen;

    }
  }

#ifdef IC_MVS400
  // All the required wide characters have been converted,
  // check to see whether we need to put a SI character or not
  // if we are in EBCDIC world
  if ((len > 1) && (*mbchar != SI)) {
    *pmbs++ = SI;
    ++mbcnt;
  }
#endif

  if (alwaysPad) {
    ::memset(pmbs, padChar, mblen-mbcnt);
    return mblen;
  } else {
    return mbcnt;
  }
}

size_t cvtWCS2MBS(const IWBuffer &wBuf, char *pMBS,
                  unsigned mblen, int alwaysPad, char padChar) {

  return cvtWCS2MBS(wBuf, wBuf.charNum(), pMBS, mblen, alwaysPad, padChar);
}

// insWCS2WCS ------------------------------------------------------
//
size_t insWCS2WCS(wchar_t *pwcSrc, int  wcSrcLen,
               wchar_t *pwcIns, int  wcInsLen,
               int wcIndex,     wchar_t *pwcDest)
{
  int  i =0, wcDestLen = 0;

  for (i=0; i < wcIndex ; i++, wcDestLen++) {
      *pwcDest++ = *pwcSrc++;
  } /* endfor */

  for (i=0; i < wcInsLen; i++, wcDestLen++) {
      *pwcDest++ = *pwcIns++;
  } /* endfor */

  for (i=wcIndex; i < wcSrcLen ; i++, wcDestLen++) {
      *pwcDest++ = *pwcSrc++;
  } /* endfor */

  return wcDestLen;
}

// isAllDigits -----------------------------------------------------
//
// - search the given string 'pStr' for length 'len', if all
//   the characters are digits, return 1; otherwise, return 0.
//
int isAllDigits(const char *pStr, unsigned len) {
  for (int i = 0; i < len; ++i) {
    if (!isdigit(*pStr++)) {
      return 0;
    }
  }
  return 1;
}

// isAllHexDigits --------------------------------------------------
//
// - search the given string 'pStr' for length 'len', if all
//   the characters are hex digits, return 1; otherwise, return 0.
//
int isAllHexDigits(const char *pStr, unsigned len) {
  for (int i = 0; i < len; ++i) {
    if (!isxdigit(*pStr++)) {
      return 0;
    }
  }
  return 1;
}

// indexThatIsNot --------------------------------------------------
//
// - search the given string 'pStr' for length 'len', return
//   the character position that does not contain 'ch' (1-based).
//   If all 'len' characters are 'ch', 0 is returned.
//
unsigned indexThatIsNot(const char *pStr, unsigned len, char ch) {
  for (int i = 0; i < len; ++i) {
    if (ch != *pStr++) {
      return(i+1);
    }
  }
  return 0;
}

// getSubString ----------------------------------------------------
//
// - extract 'tLen' bytes from 'sBuf' starting from 'startPos'.
// - if 'startPos' is pointing to the middle of multibyte character
//   it is filled with 'padCharacter' until the next valid character.
// - the resulting string will contain as many characters as possible
//   from the 'tLen' bytes of the original string, 'sBuf'.
// - if at the end, any partial character will be filled with 'padCharacter'
// - if there are more space in 'tLen' than characters can fill,
//   'padCharacter' is used to fill the rest of the bytes
//
// parameters:
// - sBuf         - the source buffer that contains the original string
// - sLen         - the length of the orginal string (can be null embedded)
// - startPos     - the starting position of the extraction (1-based)
// - padCharacter - used as pad character
// - tBuf         - the target buffer which will contain the resulting
//                  string (it is the user responsibility to allocate
//                  enough space for the resulting string)
// - tLen         - the required length of the resulting string
//
char *getSubString(const char *sBuf, unsigned sLen,
                   unsigned startPos, char padCharacter,
                   char *tBuf, unsigned tLen) {

  // Get nothing!!
  if (tLen == 0) return tBuf;

  if (startPos == 0) startPos = 1;

  if (startPos > sLen) {
    // If 'startPos' is beyond 'sLen', just fill 'tBuf' with 'tLen'
    // of 'padCharacter'
    memset(tBuf, padCharacter, tLen);
  } else {
    // Find the next valid character starting from 'startPos'
    unsigned numPos = findNextCharNum(sBuf, sLen, startPos);

    if (numPos) {
      // Use the character position to find back the beginning byte
      // position of this character. From this, we can find out how
      // many 'padCharacter' to be filled between 'startPos' and
      // the 'numPos'th character.
      unsigned curPos = findBytePos(sBuf, sLen, numPos);
      IWBuffer wBuf(sBuf, sLen);
      unsigned padNum = curPos - startPos;
      if (tLen > padNum) {
        memset(tBuf, padCharacter, padNum);
#ifdef IC_MVS400
        unsigned endPos = startPos + tLen - 1;
        if (endPos >= sLen) {
          // If the required 'tLen' bytes goes beyond 'sLen', convert the
          // remaining characters and fill with 'padCharacter' until we
          // have 'tLen-padNum' bytes.
          cvtWCS2MBS(wBuf+numPos-1, wBuf.charNum()-numPos+1, tBuf+padNum,
                     tLen-padNum, 1, padCharacter);
        } else {
          unsigned endNumPos  = findCharNum(sBuf, sLen, endPos);
          unsigned nextNumPos = findCharNum(sBuf, sLen, endPos+1);
          unsigned numChar;
          // Is endPos pointing to the last byte of the multibytes character?
          // If yes, include it in the conversion.
          if ((nextNumPos-endNumPos) == 1) {
            numChar = endNumPos - numPos + 1;
          } else {
            numChar = endNumPos - numPos;
          }
          cvtWCS2MBS(wBuf+numPos-1, numChar, tBuf+padNum,
                     tLen-padNum, 1, padCharacter);
        }
#else
        cvtWCS2MBS(wBuf+numPos-1, wBuf.charNum()-numPos+1, tBuf+padNum,
                   tLen-padNum, 1, padCharacter);
#endif
      } else {
        // There are more 'padNum' than the required 'tLen',
        // fill 'tBuf' with 'tLen' of 'padCharacter'
        memset(tBuf, padCharacter, tLen);
      }
    } else {
      // No more next character, fill 'tBuf' with 'tLen'
      // of 'padCharacter'
      memset(tBuf, padCharacter, tLen);
    }
  }

  return tBuf;
}

// getRemainString -------------------------------------------------
//
// - starting from 'startPos', extract all characters until the
//   end of 'sBuf' and store into 'tBuf'
// - the number of bytes written to 'tBuf' is returned
// - if 'startPos' is pointing to the middle of multibyte character,
//   it is filled with 'padCharacter' until the next valid character.
// - if 'startPos' is bigger than 'sLen', 'tBuf' will be untouched
//   and 0 is returned
//
// parameters:
// - same description as getSubString()
//
unsigned getRemainString(const char *sBuf, unsigned sLen,
                         unsigned startPos, char padCharacter,
                         char *tBuf) {

  if (startPos > sLen) return 0;

  if (startPos == 0) startPos = 1;

  // Find the next valid character starting from 'startPos'
  unsigned numPos = findNextCharNum(sBuf, sLen, startPos);

  if (numPos) {
    // if 'startPos' is pointing to the middle of multibyte character,
    // fill with 'padCharacter' before extracting the remaining character.
    unsigned curPos = findBytePos(sBuf, sLen, numPos);
    IWBuffer wBuf(sBuf, sLen);
    unsigned numChar = wBuf.charNum()-numPos+1;
    unsigned padNum = curPos - startPos;
    memset(tBuf, padCharacter, padNum);
    unsigned mbLen = cvtWCS2MBS(wBuf+numPos-1, numChar, tBuf+padNum);

    return mbLen+padNum;
  } else {
    // startPos is pointing at the middle of the last
    // multibyte character
    unsigned padNum = sLen-startPos+1;
    memset(tBuf, padCharacter, padNum);

    return padNum;
  }
}

#endif // IC_NLS
