// ----------------------------------------------------------------------------
//  FILE NAME: iobservr.cpp
//
//  DESCRIPTION:
//   This file contains the implementation of classes/functions declared
//   in iobservr.hpp.
//
//  COPYRIGHT:
//   IBM Open Class Library
//   (C) Copyright International Business Machines Corporation 1992, 1997
//   Licensed Material - Program-Property of IBM - All Rights Reserved.
//
//   US Government Users Restricted Rights - Use, duplication, or disclosure
//   restricted by GSA ADP Schedule Contract with IBM Corp.
//
// Revision: 45 1.8.1.3 source/core/base/iobservr.cpp, notification, ioc.v400, 000302 
// ----------------------------------------------------------------------------

#include <iobservr.hpp>
#include <inotify.hpp>
#include <inotadr.hpp> 
#include <iintrst.hpp> 
#include <iasynty.hpp> 
#include <iprimlck.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
    #define _IOBSERVR_CPP_
    #include <ipagetun.h>
#endif


//
//  Make the creation of the static resource used to synchronize stuff thread
//  safe.
//
static IPrivateResource& threadsKey()
{
    static IPrivateResource* pthreadsKey = 0;

    if (!pthreadsKey)
    {
        IPrimalLock lockInit;

        if (!pthreadsKey)
            pthreadsKey = new IPrivateResource;
    }
    return *pthreadsKey;
}


class IRequestProcessorDict : public IVPtrKeySet<INotificationProcessor*, IThreadId>
{
  public:
     IRequestProcessorDict(unsigned long numberOfElements)
       : IVPtrKeySet<INotificationProcessor*, IThreadId>(numberOfElements)
    {
    }

    ~IRequestProcessorDict()
    {
    }
};

// ctor/dtor for page tuning
IObserver::~IObserver()
{
    //
    // TBD IObserver should keep a list of notifiers it has registered with and
    //  stopHandlingNotifications for each notifier
    //
    if (fRequestProcessor) 
    {
	    if (fRequestProcessor->removeRef() == 0)
        {
		    fRequestProcessorDict->removeElementWithKey(fThreadId);
		    delete fRequestProcessor; 
        }
    }
}


IObserver::IObserver() :

    fThreadId (INonGUIThread::currentId()),
    fRequestProcessor(NULL)
{
}


// ----------------------------------------------------------------------------
//  IObserver::handleNotificationsFor
//
//  Attach the handler to the specified notifier
// ----------------------------------------------------------------------------
IObserver& IObserver::handleNotificationsFor(   INotifier&          aNotifier,
                                                const IEventData&   userData)
{
    // Have to create a requestProcessor in this thread

    //
    //  If we have cross thread notification, then we create a 
    //  request processor in order to communicate btw notifier & observer.
    //
    if (aNotifier.threadId() != fThreadId)
	    setRequestProcessor();

    aNotifier.addObserver(*this, userData);
    return *this;
}

// ----------------------------------------------------------------------------
// IObserver::stopHandlingNotificationsFor|
//
// Detach the handler from the specified window.
// ----------------------------------------------------------------------------
IObserver& IObserver::stopHandlingNotificationsFor( INotifier& aNotifier)
{
    // This removes for allChangesId interest
    aNotifier.removeObserver(*this);
    return *this;
}


IObserver& IObserver::handleNotificationsFor(IInterest& theInterest) 
{
    INotifierAddress aNotifierAddress = theInterest.notifierAddress();
    INotifier&       aNotifier = aNotifierAddress.notifier();

    //
    // Have to create a requestProcessor in this thread. If we have cross
    //  thread notification, then we create a request processor in order to
    //  communicate btw notifier & observer.
    //
    if (aNotifier.threadId() != fThreadId)
        setRequestProcessor();

    aNotifier.addObserver(*this, theInterest);
    return *this;
}


IObserver&
IObserver::handleNotificationsFor(INotifier& theNotifier,const INotificationId& id)
{
    theNotifier.addObserver(*this, IInterest(theNotifier, id));
    return *this;
}


IObserver& IObserver::stopHandlingNotificationsFor(const IInterest& theInterest)
{
    INotifierAddress aNotifier = theInterest.notifierAddress();
    aNotifier.notifier().removeObserver(*this, theInterest);
    return *this;
}


IObserver&
IObserver::stopHandlingNotificationsFor(INotifier& theNotifier, const INotificationId& id)
{
    theNotifier.removeObserver(*this, IInterest(theNotifier, id));
    return *this;
}

//
//  Asynchronous Observer stuff. Initialize static data
//
IRequestProcessorDict* IObserver::fRequestProcessorDict = 0; 


//Key for INotificationProcessor
const IThreadId& key(const INotificationProcessor* object)
{
    return (object->threadId());
}

const IThreadId& IObserver::threadId() const 
{ 
    return fThreadId;
}

INotificationProcessor* IObserver::requestProcessor()
{
    if (! fRequestProcessor)
        setRequestProcessor();
    return fRequestProcessor;
}


//Only set requestProcessor from the observer thread
IRequestProcessor& IObserver::setRequestProcessor()
{
    if (! fRequestProcessor) 
    {
        //This observer is not referencing a requestProcessor yet
        IResourceLock threadsLock(threadsKey());

        //first time creating dictionary
        if (! fRequestProcessorDict)
        	fRequestProcessorDict = new IRequestProcessorDict(20);

        //Use a requestProcessor from the same thread if available
        if ( fRequestProcessorDict->containsElementWithKey(fThreadId))
        {
            fRequestProcessor = fRequestProcessorDict->elementWithKey(fThreadId);

            // Increment reference for processor
        	fRequestProcessor->addRef();
        }
        else 
        {
        	//If the observer is in a thread doing processmsgs, then use this one
        	fRequestProcessor = new INotificationProcessor(fThreadId); 
        	fRequestProcessorDict->add (fRequestProcessor);

            // Increment reference for processor
        	fRequestProcessor->addRef();
        }
    }
    return *fRequestProcessor;
}



//Class ICrossThreadNotificationLoop
//-----------------------------------------------------
ICrossThreadNotificationLoop::ICrossThreadNotificationLoop (IObserver& theObserver):
  fObserver(theObserver),
  fRequestHandler(0),
  fRequestProcessor(0),
  fIsRunning(false)
{
    IASSERTSTATE ( theObserver.threadId() == INonGUIThread::currentId());

    // Check for a request processor
    INotificationProcessor* theObserverProcessor = theObserver.requestProcessor();
  
    if (theObserverProcessor != 0) 
    {
	    fRequestProcessor = theObserverProcessor;

        // Alias the handler for the request processor
	    fRequestHandler = fRequestProcessor->handlerAlias();
    }

    //  TBD else throw (no request processor for observer!)
}

void ICrossThreadNotificationLoop::cleanup()
{
    // Observer dtor should take care of this
    //  if (!observersInCurrentThread())
    //      delete fRequestProcessor;
}

void ICrossThreadNotificationLoop::run()
{
    //While observers are using the request processor, process notifications
    fIsRunning = true;
    while (observersInCurrentThread() && fIsRunning)
    {
        // processNotification();
	    waitAndDispatchNotification();
    }
    fIsRunning = false;
    cleanup();
}

void ICrossThreadNotificationLoop::stop()
{
    fIsRunning = false;
}


void ICrossThreadNotificationLoop::dispatchNotification()
{
    fRequestHandler->DispatchNextPendingRequest();
}

void ICrossThreadNotificationLoop::waitAndDispatchNotification()
{
    //TBD Use event semaphore here 
    fRequestHandler->WaitAndDispatchNextRequest();
}

bool ICrossThreadNotificationLoop::observersInCurrentThread()
{
    return (fRequestProcessor->refCount() != 0);
}

ICrossThreadNotificationLoop::~ICrossThreadNotificationLoop()
{
    //delete fRequestHandler;
}



//IObserverForConnectionToImp  handles old style notification event data
IObserverForConnectionToImp::IObserverForConnectionToImp()
{
}

IObserverForConnectionToImp::~IObserverForConnectionToImp()
{
}

IObserverForConnectionToImp::IObserverForConnectionToImp(const IObserverForConnectionToImp& copy)
{
}

IObserverForConnectionToImp&     
IObserverForConnectionToImp::operator= (const IObserverForConnectionToImp& rhs)
{
    return *this;
}

void
IObserverForConnectionToImp::handleOldNotificationEvent (const INotificationEvent& event)
{
  // This conditional is here so that the BOSS (a.k.a. CBS) project can use
  // the portapak source base with the old cset compiler. At this time, BOSS
  // is not using our UI classes, therefore it should be safe to always
  // throw the exception in here (we expect it to never be called). 
  // ** THIS CODE SHOULD BE UNCONDITIONALIZED AFTER BOSS GOES TO 96A COMPILERS!
  //#if defined(AIX) //&& !defined(__IBMCPP__))
#if (defined(_AIX) && !defined(__IBMCPP__))
    if (false)
#else
    if (typeid(event) == typeid (INotificationEvent))
#endif
    {
        // Use this method to dispatch an INotificationEventFor
        dispatchOldNotificationEvent(event, (void*)event.eventData().asUnsignedLong());
    }
    else 
    {
        IException unexpectedExc("Received unexpected event type in IObserverForConnectionTo");
        ITHROW(unexpectedExc);
    }
}
