//
//  FILE NAME: iprcenv.cpp
//
//  DESCRIPTION:
//
//  This module implements the IEnvironment class, a specialized dictionary
//  class that stores name/value pairs of text strings in the way that
//  an OS environment does. Its primary use is for the IExternalProcess
//  class, where it is used to pass in a new environment.
//
//
//  COPYRIGHT:
//    IBM Open Class Library
//    Licensed Materials - Property of IBM
//
//    5645-001
//    (C) Copyright IBM Corporation 1992, 1997
//
//    US Government Users Restricted Rights - Use, duplication, or
//    disclosure restricted by GSA ADP Schedule Contract with IBM
//    Corp.
// ----------------------------------------------------------------------------


// ----------------------------------------------------------------------------
//  Includes
// ----------------------------------------------------------------------------
#include    <stdlib.h>
#include    <iprimtyp.hpp>
#include    <itext.hpp>
#include    <iugnrl.hpp>
#include    <iprcenv.hpp>

#ifdef IC_PAGETUNE
    #define _IPRCENV_CPP_
    #include <ipagetun.h>
#endif

#ifdef IC_POSIX
	extern char **environ;
	#define ENVIRON environ
#else
	#define ENVIRON _environ
#endif

// ----------------------------------------------------------------------------
//  IEnvData
//
//  This stuff is just the very simple implementation of the utility
//  class we use to track our list of key/values internally.
// ----------------------------------------------------------------------------
class IEnvData
{
    public :
        IEnvData
        (
            const       IText&      keyText
            , const     IText&      valueText
            ,           IEnvData*   nextNode
        );

        const IText&    keyText();
        IEnvData*       next();
        const IText&    valueText();

        IEnvData*       setNext(IEnvData* nextNode);
        void            setValue(const IText& newText);

    private :
        // Don't need assignment, or default or copy construction
        void            operator=(const IEnvData&);
                        IEnvData(const IEnvData&);
                        IEnvData();

        // Holds the key, value, and link to next node
        IText           fKeyText;
        IText           fValueText;
        IEnvData*       fNextNode;
};

IEnvData::IEnvData( const   IText&      keyText
                    , const IText&      valueText
                    ,       IEnvData*   nextNode) :

    fKeyText(keyText)
    , fValueText(valueText)
    , fNextNode(nextNode)
{
}

inline const IText& IEnvData::keyText()
{
    return fKeyText;
}

inline IEnvData* IEnvData::next()
{
    return fNextNode;
}

inline const IText& IEnvData::valueText()
{
    return fValueText;
}

inline IEnvData* IEnvData::setNext(IEnvData* nextNode)
{
    IEnvData* tmpNode = fNextNode;
    fNextNode = nextNode;
    return tmpNode;
}

inline void IEnvData::setValue(const IText& newText)
{
    fValueText = newText;
}



// ----------------------------------------------------------------------------
//  IEnvironment: Constructors and destructors
//
//  Basically we just have to zero out the head link and copy any current
//  contents if needed.
//
IEnvironment::IEnvironment() :

    fHeadNode(0),
    fIterNode(0),
    fCount(0)
{
}

IEnvironment::IEnvironment(const bool copyCurrentEnv) :

    fHeadNode(0),
    fIterNode(0),
    fCount(0)
{
    // If we were asked to copy the current environment do it
    if (copyCurrentEnv)
    {
        unsigned long   index = 0;
        IText           keyText;
        IText           valueText;
        while (ENVIRON[index])
        {
            //
            //  Copy the current string into both IText objects and
            //  then cut out the parts not needed in each. This gets us
            //  into a UNICode world and lets us use simpler IText stuff
            //  to manipulate them.
            //
            keyText = IText(ENVIRON[index]);
            valueText = IText(ENVIRON[index]);

            // Find the = sign in this entry
            IText::offset_type equalOfs = keyText.find(UASCII::kEqualsSign);

            // Don't freak out in psycho scenario of no equal sign
            if (equalOfs != IText::offset_type(-1))
            {
                //
                //  Terminate at the equal sign for the key and cut up to it
                //  for the value.
                //
                keyText.resize(equalOfs);
                valueText.erase(0, equalOfs+1);

                add(keyText, valueText);
            }
            index++;
        }
    }
}

IEnvironment::IEnvironment(const IEnvironment& toCopy) :

    fHeadNode(0),
    fIterNode(0)
{
    IEnvData* curNode = toCopy.fHeadNode;
    while (curNode)
    {
        add(curNode->keyText(), curNode->valueText());
        curNode = curNode->next();
    }
}

IEnvironment::~IEnvironment()
{
    clear();
}

void IEnvironment::operator=(const IEnvironment& toAssign)
{
    // Clear any current content
    clear();

    // And now iterate the passed environment and make copies of its nodes
    IEnvData* curNode = toAssign.fHeadNode;

    while (curNode)
    {
        add(curNode->keyText(), curNode->valueText());
        curNode = curNode->next();
    }
}

void IEnvironment::add(const IText& keyText, const IText& valueText)
{
    //
    //  Create a new link with the passed value and key, make it the head
    //  of the list, and make the old head its next link.
    //
    fHeadNode = new IEnvData(keyText, valueText, fHeadNode);
    fCount++;
}

void IEnvironment::clear()
{
    //
    //  Iterate the list and delete each element, saving the next
    //  one first each time so we don't lose the trail.
    //
    IEnvData* curNode = fHeadNode;
    while (curNode)
    {
        IEnvData* nextNode = curNode->next();
        delete curNode;
        curNode = nextNode;
    }

    // And now zero out the head node, and zero the node count
    fHeadNode = 0;
    fCount = 0;
}

bool IEnvironment::find(const IText& keyText, IText& valueText) const
{
    // Look for this key. If not, found we are done
    IEnvData* prevNode;
    IEnvData* foundNode = findKey(keyText, prevNode);
    if (!foundNode)
        return false;

    // Pull out the value of the found node
    valueText = foundNode->valueText();
    return true;
}

IText IEnvironment::find(const IText& keyText) const
{
    // Look for this key. If not, found we are done
    IEnvData* prevNode;
    IEnvData* foundNode = findKey(keyText, prevNode);
    if (!foundNode)
        return IText("");

    // Return the value of the found node
    return foundNode->valueText();
}

bool IEnvironment::first(IText& keyText, IText& valueText)
{
    fIterNode = fHeadNode;

    // Optimize for empty list
    if (!fHeadNode)
        return false;

    // Pull out the values of the iterator node
    keyText = fIterNode->keyText();
    valueText = fIterNode->valueText();

    return true;
}

bool IEnvironment::next(IText& keyText, IText& valueText)
{
    // Optimize for empty list
    if (!fIterNode)
        return false;

    // Move to the next node
    fIterNode = fIterNode->next();
    if (!fIterNode)
        return false;

    // Pull out the values of the iterator node
    keyText = fIterNode->keyText();
    valueText = fIterNode->valueText();

    return true;
}

bool IEnvironment::remove(const IText& keyText)
{
    // Look for this key. If not, found we are done
    IEnvData* prevNode;
    IEnvData* removeNode = findKey(keyText, prevNode);
    if (!removeNode)
        return false;

    //
    //  Point the remove node's previous node at the remove node's next
    //  node. This removes the remove node from the list. Handle special
    //  case of it being the first node. If its also the last (i.e. only
    //  one left, that's ok because the head will be set to its next
    //  which is 0.
    //
    if (prevNode)
        prevNode->setNext(removeNode->next());
    else
        fHeadNode = removeNode->next();

    // Bump the counter down, sanity checking first so we don't wrap
    IASSERT(fCount != 0)
    fCount--;

    //
    //  If the iterator node is the remove node, then move the iterator
    //  node forward. Note that we don't have to check for null here
    //  because the remove node is known to be non-zero.
    //
    if (fIterNode == removeNode)
        fIterNode = fIterNode->next();

    // And now delete the remove node to clean it up
    delete removeNode;

    return true;
}

bool IEnvironment::update(const IText& keyText, const IText& newValue)
{
    // Look for this key. If not, found we are done
    IEnvData* prevNode;
    IEnvData* updateNode = findKey(keyText, prevNode);
    if (!updateNode)
        return false;

    updateNode->setValue(newValue);
    return true;
}

IEnvData* IEnvironment::findKey(const IText& keyText, IEnvData*& prevNode) const
{
    IEnvData* curNode = fHeadNode;
    prevNode = 0;
    while (curNode)
    {
        if (!curNode->keyText().compare(keyText))
            break;

        // Save the current node as the previous and move up to next
        prevNode = curNode;
        curNode = curNode->next();
    }
    return curNode;
}
