// Revision: 61 1.3.2.2 source/core/base/iptarray.cpp, appsupport, ioc.v400, 000302  
/*******************************************************************************
* FILE NAME: iptarray.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iptarray.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <ipoint.hpp>
#include <iptarray.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IPTARRAY_CPP_
  #include <ipagetun.h>
#endif

/*------------------------------------------------------------------------------
| IPointArray::IPointArray                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointArray::IPointArray( unsigned long dimension, const IPoint *array )
  : fDim( dimension ), fPtArray(0)
{
  if ( fDim )
  {
    fPtArray = new IPoint[fDim];
    if (array)
    {
      for (int i=0; i < fDim; i++)
      {
        fPtArray[i] = array[i];
      } /* endfor */
    }
  }
}

/*------------------------------------------------------------------------------
| IPointArray::IPointArray                                                     |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointArray::IPointArray ( const IPointArray& pointArray )
  : fDim(pointArray.size()), fPtArray(0)
{
  if ( fDim )
  {
    fPtArray = new IPoint[fDim];
    for (int i=0; i < fDim; i++)
    {
      fPtArray[i] = pointArray[i];
    } /* endfor */
  } /* endif */
}

/*------------------------------------------------------------------------------
| IPointArray::~IPointArray                                                    |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointArray::~IPointArray()
{
  if (fDim)
  {
    delete [] fPtArray;
  } /* endif */
}

/*------------------------------------------------------------------------------
| IPointArray::operator[]                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPoint& IPointArray::operator[] ( unsigned long index )
{
  IASSERT( index < fDim );
  return fPtArray[index];
}

/*------------------------------------------------------------------------------
| IPointArray::operator[]                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
const IPoint& IPointArray::operator[] ( unsigned long index ) const
{
  IASSERT( index < fDim );
  return fPtArray[index];
}

/*------------------------------------------------------------------------------
| IPointArray::size                                                            |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
unsigned long IPointArray::size() const
{
  return fDim;
}

/*------------------------------------------------------------------------------
| IPointArray::resize                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointArray& IPointArray::resize( unsigned long newsize )
{
  if ( newsize )
  {
    if ( newsize != fDim )
    {
      if ( fPtArray )
      {
        IPoint * tempPtArray = new IPoint[newsize];
        for (int i=0; i < ((fDim < newsize) ? fDim : newsize); i++)
        {
          tempPtArray[i] = fPtArray[i];
        } /* endfor */
        delete [] fPtArray;
        fPtArray = tempPtArray;
      }
      else
      {
        fPtArray = new IPoint[newsize];
      }
      fDim = newsize;
    } /* endif */
  }
  else
  {
    if ( fDim )
    {
      fDim = newsize;
      delete [] fPtArray;
      fPtArray = 0;
    }
  } /* endif */
  return *this;
}

/*------------------------------------------------------------------------------
| IPointArray::reverse                                                         |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointArray& IPointArray::reverse()
{
  IPoint tempPt;

  for (int i=0; i < fDim/2; i++)
  {
    tempPt = fPtArray[fDim-1-i];
    fPtArray[fDim-1-i] = fPtArray[i];
    fPtArray[i] = tempPt;
  }
  return *this;
}

/*------------------------------------------------------------------------------
| IPointArray::reversed                                                        |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointArray IPointArray::reversed( ) const
{
  return IPointArray(*this).reverse();
}

/*------------------------------------------------------------------------------
| IPointArray::insert                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointArray& IPointArray::insert( unsigned long index, const IPoint& point)
{
  IASSERT( index < fDim );
  fDim++;
  IPoint * newPtArray = new IPoint[fDim];

  int i;
  for (i=0; i < index; i++)
    newPtArray[i] = fPtArray[i];

  newPtArray[index] = point;

  for (i=index+1; i < fDim; i++)
    newPtArray[i] = fPtArray[i-1];

  delete [] fPtArray;
  fPtArray = newPtArray;

  return *this;
}

/*------------------------------------------------------------------------------
| IPointArray::add                                                             |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointArray& IPointArray::add( const IPoint& point )
{
  resize( fDim+1 );
  fPtArray[fDim-1] = point;
  return *this;
}

/*------------------------------------------------------------------------------
| IPointArray::remove                                                          |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointArray& IPointArray::remove( unsigned long index )
{
  IASSERT( index < fDim );
  fDim--;
  IPoint * newPtArray = new IPoint[fDim];

  int i;
  for (i=0; i < index; i++)
    newPtArray[i] = fPtArray[i];

  for (i=index; i < fDim; i++)
    newPtArray[i] = fPtArray[i+1];

  delete [] fPtArray;
  fPtArray = newPtArray;

  return *this;
}

/*------------------------------------------------------------------------------
| IPointArray::operator=                                                       |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
IPointArray& IPointArray::operator= (const IPointArray& pointArray)
{
  if ( this == &pointArray )
    return *this;

  resize( pointArray.fDim );

  for (int i=0; i < fDim; i++)
  {
    fPtArray[i] = pointArray[i];
  } /* endfor */
  return *this;
}

/*------------------------------------------------------------------------------
| IPointArray::operator==                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
bool IPointArray::operator== (const IPointArray& pointArray) const
{
  bool fAllMatch = True;

  if ( fDim != pointArray.fDim )
    return False;

  for ( int i=0; i < fDim; i++)
  {
    if (fPtArray[i] != pointArray[i])
    {
      fAllMatch = False;
      break;
    }
  }
  return fAllMatch;
}

/*------------------------------------------------------------------------------
| IPointArray::operator!=                                                      |
|                                                                              |
|                                                                              |
------------------------------------------------------------------------------*/
bool IPointArray::operator!= (const IPointArray& pointArray) const
{
  return !(*this == pointArray);
}
