/* NOSHIP */
// ----------------------------------------------------------------------------
// FILE NAME: istpdata.hpp
//
// DESCRIPTION:
//   Declaration of the classes:
//     IStringParserData
//
//  Objects of this class are used to store and manage the data for parsing
//  strings.
//
// COPYRIGHT:
//   IBM Open Class Library
//   Licensed Materials - Property of IBM
//
//   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.
//
//   US Government Users Restricted Rights - Use, duplication, or
//   disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
//
// Revision: 78 1.3.1.3 source/core/base/istpdata.hpp, appsupport, ioc.v400, 000302 
// ----------------------------------------------------------------------------

#ifndef _ISTPDATA_
#define _ISTPDATA_

#include <ibase.hpp>
#include <istring.hpp>

// ----------------------------------------------------------------------------
// Define token pointer type
// ----------------------------------------------------------------------------
#if __IBMCPP__ >= 400
#pragma namemangling(compat)
#endif

typedef IString* ITokenPointer;

#pragma pack(push,4)
#pragma enum(4)

class IC_EXPORTB IStringParserData
{
public:
    // -------------------- Constructor and Destructor ------------------------
    //  The constructor accepts the text string to be parsed and initialized
    //  the parse history information.  The parse history information is used
    //  to track what parsing has been performed (but may not be the final
    //  results). The destructor is used to delete possible parse history
    //  information.
    // ------------------------------------------------------------------------
                        IStringParserData(const IString &text);
    virtual             ~IStringParserData();

    // --------------------------------- Parsing ------------------------------
    //  These methods are used to perform the actual string parsing and to
    //  store a record of the actions performed.  These methods are intended
    //  to be called only by IStringParse objects.
    // ------------------------------------------------------------------------
    IStringParserData&  processToken(IString &token, bool reparse = false);
    IStringParserData&  processPattern(const IString &pattern);
    IStringParserData&  processPattern(const char *pattern);
    IStringParserData&  processPattern(char pattern);
    IStringParserData&  processSkip(bool reparse = false);
    IStringParserData&  processTest(const IStringTest &test);
    IStringParserData&  setPosition(unsigned long position);
    IStringParserData&  changePosition(int delta);
    IStringParserData&  changePosition(unsigned long delta);

    // ------------------------------ Implementation --------------------------
    //  The following methods are used to implement common actions performed
    //  during string parsing.
    // ------------------------------------------------------------------------
    IStringParserData&  reparseLastToken();
    IStringParserData&  reparseTokens(unsigned long startPos, unsigned long stopPos);
    IStringParserData&  saveToken(IString *token);
    IStringParserData&  clearSavedTokens();

    // ---------------------------- Reference Counting ------------------------
    //  The following functions manage the buffer reference count:
    //  addRef    - Increments the usage count.
    //  removeRef - Decrements the usage count and deletes the buffer when the
    //              usage count goes to 0.
    // ------------------------------------------------------------------------
    void                addRef();
    void                removeRef();

    // --------------------------------- Private ------------------------------
private:
    IString         parseText;
    unsigned long   currentPosition;
    unsigned long   patternPosition;
    unsigned long   patternLength;
    unsigned long   tokenCount;
    unsigned long   tokenArraySize;
    IString*        lastToken;
    bool            lastSkip;
    ITokenPointer*  usedTokens;
    unsigned long   refs;
};

#pragma enum(pop)
#pragma pack(pop)

#if __IBMCPP__ >= 400
#pragma namemangling()
#endif

#endif
