// ----------------------------------------------------------------------------
// FILE NAME: istring0.cpp
//
// DESCRIPTION:
//   This file contains the implementation of the constructors for the
//   class IString (and the implementation of related functions).
//
//
// COPYRIGHT:
//   IBM Open Class Library
//   Licensed Materials - Property of IBM
//
//   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.
//
//   US Government Users Restricted Rights - Use, duplication, or
//   disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
//
// Revision: 83 1.7.1.9 source/core/base/istring0.cpp, appsupport, ioc.v400, 000302 
// ----------------------------------------------------------------------------

#ifdef __MVS__
// Need _OPEN_THREADS for thread-safe version of USL
#include <ibhandle.hpp>
#endif // __MVS__

extern "C"
{
    #include <stdio.h>
    #include <float.h>
    #include <ctype.h>
}

#include <istring.hpp>
#include <ibuffer.hpp>
#include <iexcbase.hpp>
#include <idatstrm.hpp>
#include <iprimlck.hpp>
#include <itext.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
    #define _ISTRING0_CPP_
    #include <ipagetun.h>
#endif

#define QUOTE(x) #x
#define STRING(x) QUOTE(x)


#ifdef IC_NLS
#pragma enum(4)
typedef enum
{
    i18nOff = 0,
    i18nOn
}   I18NStates;
#pragma enum(pop)
static I18NStates   gI18NState = i18nOff;
static bool         gI18N_InitDone = false;

/*--------------------------------------------------------------------*/
/* I18N support                                                       */
/*--------------------------------------------------------------------*/
#define  ENVAR_I18N    "ICLUI_I18N"

// ----------------------------------------------------------------------------
// I18N_Init()
//
// Call getenv() to see if the environment variable ICLUI_I18N= is
// defined.  If it is not found, default fgI18NState to OFF (i.e. SWORD
// semantics - just binary compare).  fgI18NState is set to i18nOn
// (I18N semantics - locale sensitive) iff ICLUI_I18N=ON.
// ----------------------------------------------------------------------------
static void I18N_Init()
{
    // Get environment variable  ICLUI_I18N=
    char cEnvStr[3] = "\0";
    const char* pEnv = getenv(ENVAR_I18N);

    // If ICLUI_I18N=  exist
    if (pEnv)
    {
        // 2 characters value found
        if (2 == strlen(pEnv))
        {
            int i;

            strcpy(cEnvStr, pEnv);
            for (i = 0; i < 2 ; i++)
                *(cEnvStr + i) = (char)toupper(*(cEnvStr + i));
            *(cEnvStr + 2) = '\0';
        }
        else
        {
            // Ignore the invalid values
            pEnv = NULL;
        }
    }

    // Default is I18N off, i.e. locale insensitive
    if ( pEnv && !strcmp(cEnvStr, "ON"))
        gI18NState = i18nOn;
    else
        gI18NState = i18nOff;
}
#endif


char* IString::defaultBuffer( )
{
    return (nullBuffer) ?
            nullBuffer
            : (nullBuffer = IBuffer::defaultBuffer()->contents());
}

void IString::setDefaultBuffer(char* newDefaultBuffer)
{
    nullBuffer = newDefaultBuffer;
}

void IBuffer::setDefaultBuffer(IBuffer *newDefaultBuffer)
{
#ifdef IC_DEVELOP
    IASSERT((newDefaultBuffer->len == 0) && (newDefaultBuffer->data[0] == 0))
#endif

    if (IBuffer::nullBuffer)
        IBuffer::nullBuffer -> removeRef();

    // Play it safe and don't assume the buffer is 32 bytes aligned
    // if someone call setDefaultBuffer
    IBuffer::is32bytesAligned = false;
    IBuffer::nullBuffer = newDefaultBuffer;
    IString::setDefaultBuffer(newDefaultBuffer->data);
}

#ifdef IC_NLS
bool IString::isInternationalized()
{
    if (!gI18N_InitDone)
    {
        IPrimalLock lockInit;

        if (!gI18N_InitDone)
        {
            I18N_Init();
            gI18N_InitDone = true;
        }
    }
    return (gI18NState == i18nOn);
}

void IString::enableInternationalization(bool enable)
{
    // Insure that isInternationalized() won't do init and overwrite this
    if (!gI18N_InitDone)
    {
        IPrimalLock lockInit;
        if (!gI18N_InitDone)
            gI18N_InitDone = true;
    }

    if (enable)
        gI18NState = i18nOn;
    else
        gI18NState = i18nOff;
}

void IString::disableInternationalization()
{
    // Insure that isInternationalized() won't do init and overwrite this
    if (!gI18N_InitDone)
    {
        IPrimalLock lockInit;
        if (!gI18N_InitDone)
            gI18N_InitDone = true;
    }
    gI18NState = i18nOff;
}

#endif

/*----------------------------------------------------------------------------*/
const char * IString::nullBuf()
{
   return nullBuffer;
}


/*------------------------------------------------------------------------------
|Member Name: IString :: null                                                  |
|             IString :: zero                                                  |
|             IString :: maxLong                                               |
|             IString :: hexDigits                                             |
|                                                                              |
| These are static IString and char* objects used by the ISring                |
| implementation:                                                              |
|                                                                              |
|   null      - Null (zero length) IString                                     |
|   zero      - IString with value == "0"                                      |
|   maxLong   - IString with ascii representation of LONG_MAX                  |
|   hexDigits - pointer to array of hex digits, in order                       |
------------------------------------------------------------------------------*/

const char
 * const IString::null = "",
 * const IString::zero = "0",
 * const IString::maxLong = STRING(LONG_MAX);

char
 *IString::nullBuffer = 0;


/*------------------------------------------------------------------------------
| IString::IString                                                             |
|                                                                              |
| Initialize the string with as accurrate a representation of the input        |
| argument as possible.                                                        |
|                                                                              |
| All constructors (except as noted below) *must* initialize the pBuffer data  |
| member to point to an appropriate null IBuffer object.  This will cause the  |
| resulting IString to be DBCS-enabled if need be.  This is accomplished via   |
| the static member function IString::defaultBuffer.  It will return a pointer |
| to a null Buffer of the appropriate type.                                    |
|                                                                              |
| The IString is actually "constructed" (that is, storage buffer allocated and |
| initialized) via application of the initBuffer function to the just          |
| initialized IBuffer.                                                         |
|                                                                              |
| The copy constructor simply sets the pBuffer data member to point to the     |
| buffer of the argument IString and bumps that buffer's use count.            |
|                                                                              |
| The special protected constructor that accepts an IBuffer pointer is another |
| special case.  This constructor is used to pass control of an IBuffer to     |
| another IString.  It is typically used within non-const IString member       |
| functions to delay disposal of an IString's IBuffer (in cases where that     |
| IBuffer may be needed during construction of the new IString contents).      |
|                                                                              |
| Notes:                                                                       |
|   1. The use count for the null buffer returned by                           |
|      IString::defaultBuffer is *not* incremented.  This is because           |
|      that buffer (in almost all instances) will be "discarded" (via          |
|      the call to IString::initBuffer).  The exception is the default         |
|      constructor, which will retain its reference to the null buffer         |
|      and therefore must increment the buffer's use count.                    |
------------------------------------------------------------------------------*/
// See istring.inl



/*------------------------------------------------------------------------------
| IString::initBuffer                                                          |
|                                                                              |
| The "base" version of this function simply invokes the corresponding         |
| IBuffer::newBuffer function against the pBuffer member.                      |
|                                                                              |
| The long/unsigned long/double functions convert their arguments to strings   |
| and invoke the "base" version.                                               |
------------------------------------------------------------------------------*/
IString &IString :: initBuffer ( const void *p1,
                                 unsigned    len1,
                                 const void *p2,
                                 unsigned    len2,
                                 const void *p3,
                                 unsigned    len3,
                                 char        padChar )
  {
  return this->setBuffer( this->buffer() -> newBuffer( p1, len1,
                                                     p2, len2,
                                                     p3, len3,
                                                     padChar ) );
  }

/*------------------------------------------------------------------------------
| IString::initBuffer                                                          |
------------------------------------------------------------------------------*/
IString &IString :: initBuffer ( long  aLong )
  {
  char buffer[ 32 ];
  sprintf( buffer, "%ld", aLong );
  return this->initBuffer( buffer,
                           lengthOf( buffer ),
                           0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::initBuffer                                                          |
------------------------------------------------------------------------------*/
IString &IString :: initBuffer ( unsigned long  anUnsignedLong )
  {
  char buffer[ 32 ];
  sprintf( buffer, "%lu", anUnsignedLong );
  return this->initBuffer( buffer,
                           lengthOf( buffer ),
                           0, 0, 0, 0, 0 );
  }
#ifdef _LONG_LONG
/*------------------------------------------------------------------------------
| IString::initBuffer                                                          |
------------------------------------------------------------------------------*/
IString &IString :: initBuffer ( long long  aLongLong )
  {
  char buffer[ 32 ];
  sprintf( buffer, "%lld", aLongLong );
  return this->initBuffer( buffer,
                           lengthOf( buffer ),
                           0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::initBuffer                                                          |
------------------------------------------------------------------------------*/
IString &IString :: initBuffer ( unsigned long long  anUnsignedLongLong )
  {
  char buffer[ 32 ];
  sprintf( buffer, "%llu", anUnsignedLongLong );
  return this->initBuffer( buffer,
                           lengthOf( buffer ),
                           0, 0, 0, 0, 0 );
  }
#endif
/*------------------------------------------------------------------------------
| IString::initBuffer                                                          |
------------------------------------------------------------------------------*/
IString &IString :: initBuffer ( double  aDouble )
  {
  char buffer[ 32 ];
  sprintf( buffer, "%.*g", DBL_DIG, aDouble );
  return this->initBuffer( buffer,
                           lengthOf( buffer ),
                           0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::defaultBuffer                                                       |
|                                                                              |
| If DBCS environment is active, return &IDBCSBuffer::nullBuffer.              |
| Otherwise, return &IBuffer::nullBuffer.                                      |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::data                                                                |
|                                                                              |
| Simply return a pointer to the start of the associated IBuffer's buffer.     |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::lengthOf                                                            |
|                                                                              |
| If the argument char* pointer is zero, return 0;  Otherwise, return the      |
| length as calculated by strlen.                                              |
------------------------------------------------------------------------------*/
// See istring.inl


/*------------------------------------------------------------------------------
| IString::prepareToChange                                                     |
|                                                                              |
| If the reference count for the referenced IBuffer is greater than one        |
| (i.e., if another IString is pointing to it), then make a copy.              |
------------------------------------------------------------------------------*/
IString &IString :: prepareToChange ( )
  {
  if ( this->buffer()->useCount() > 1 )
    {
    IString old( this->buffer() );
    this->initBuffer( this->data(), this->length(),
                      0, 0,
                      0, 0,
                      0 );
    }
  return *this;
  }

/*------------------------------------------------------------------------------
| IString::rightJustify                                                        |
|                                                                              |
| If the new length differs from the current length, delegate the call to the  |
| IBuffer contents.  The receiver's contents will be replaced with the new     |
| buffer returned.                                                             |
|                                                                              |
| Notes:                                                                       |
|   1. This function is implemented here rather than in istring7.cpp (with     |
|      the other "editing" functions) because it is invoked from istring9.cpp  |
|      and we don't want to cause all the editing code to be linked in if      |
|      only conversion functions are required.                                 |
------------------------------------------------------------------------------*/
IString & IString :: rightJustify ( unsigned newLength,
                                    char     padCharacter )
  {
  if ( this->length() != newLength )
    {
    IString old( this->buffer() );

    old.buffer()->addRef();    // dummy add ref
    this->setBuffer( this->buffer() -> rightJustify( newLength,
                                                     padCharacter ));
    old.buffer()->removeRef();  // remove dummy add ref
    }
  return *this;
  }

/*------------------------------------------------------------------------------
| IString::remove                                                              |
|                                                                              |
| If the number of characters to be removed is non-zero, then delegate this    |
| function via the IBuffer contents.  The receiver's contents are replaced     |
| with the IBuffer returned from that call.                                    |
|                                                                              |
| Notes:                                                                       |
|   1. This function is implemented here rather than in istring7.cpp (with     |
|      the other "editing" functions) because it is invoked from istring9.cpp  |
|      and we don't want to cause all the editing code to be linked in if      |
|      only conversion functions are required.                                 |
------------------------------------------------------------------------------*/
IString & IString :: remove ( unsigned startPos )
  {
  return this->remove( startPos, UINT_MAX );
  }

/*------------------------------------------------------------------------------
| IString::remove                                                              |
------------------------------------------------------------------------------*/
IString & IString :: remove ( unsigned startPos,
                              unsigned numChars )
{
    if (numChars && (startPos <= this->length()) && this->length())
    {
        IString old(this->buffer());

        // Dummy add ref
        old.buffer()->addRef();
        this->setBuffer(this->buffer()->remove(startPos,numChars));

        // Remove dummy add ref
        old.buffer()->removeRef();
    }
    return *this;
}

/*------------------------------------------------------------------------------
| IString::remove                                                              |
------------------------------------------------------------------------------*/
IString IString :: remove ( const IString &aString,
                            unsigned       startPos,
                            unsigned       numChars )
  {
  return IString(aString).remove( startPos, numChars );
  }

/*------------------------------------------------------------------------------
| IString :: setBuffer                                                         |
------------------------------------------------------------------------------*/
// See istring.inl

/*------------------------------------------------------------------------------
| IString::asDebugInfo                                                         |
------------------------------------------------------------------------------*/
IString IString :: asDebugInfo ( ) const
  {
  IString result( "IString(@" );
  result += IString( (unsigned long) this );
  result += ",pBuffer->";
  result += this->buffer()->asDebugInfo();
  result += ")";
  return result;
  }

/*------------------------------------------------------------------------------
| IString::operator =                                                          |
------------------------------------------------------------------------------*/
IString &IString :: operator = ( const IString &aString )
  {
  aString.buffer()->addRef();
  this->buffer()->removeRef();
  this->setBuffer( aString.buffer() );
  return *this;
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const void *pBuffer,
                     unsigned    len,
                     char        padCharacter )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( pBuffer,
                    len,
                    0, 0, 0, 0,
                    padCharacter );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const void *pBuffer1,
                     unsigned    len1,
                     const void *pBuffer2,
                     unsigned    len2,
                     char        padCharacter )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( pBuffer1,
                    len1,
                    pBuffer2,
                    len2,
                    0, 0,
                    padCharacter );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const void *pBuffer1,
                     unsigned    len1,
                     const void *pBuffer2,
                     unsigned    len2,
                     const void *pBuffer3,
                     unsigned    len3,
                     char        padCharacter )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( pBuffer1,
                    len1,
                    pBuffer2,
                    len2,
                    pBuffer3,
                    len3,
                    padCharacter );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( int  anInt )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (long) anInt );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( long  aLong )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( aLong );
  }
#ifdef _LONG_LONG
/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( long long aLongLong )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( aLongLong );
  }
#endif
/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( short  aShort )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (long) aShort );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( unsigned  anUnsigned )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (unsigned long) anUnsigned );
  }
/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( unsigned long  anUnsignedLong )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( anUnsignedLong );
  }

#ifdef _LONG_LONG
/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( unsigned long long  anUnsignedLongLong )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( anUnsignedLongLong );
  }
#endif
/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( unsigned short  anUnsignedShort )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (unsigned long) anUnsignedShort );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( double aDouble )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( aDouble );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( char aChar )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void *) &aChar,
                    1,
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( unsigned char aChar )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void *) &aChar,
                    1,
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( signed char aChar )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void *) &aChar,
                    1,
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const char *pChar )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void*) pChar,
                    lengthOf( pChar ),
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const unsigned char *pUnsignedChar )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void*) pUnsignedChar,
                    lengthOf( (const char *) pUnsignedChar ),
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const signed char *psignedChar )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( (const void*) psignedChar,
                    lengthOf( (const char *) psignedChar ),
                    0, 0, 0, 0, 0 );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const IString &aString )
  {
  aString.buffer() -> addRef();
  this->setBuffer( aString.buffer() );
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( )
  : pBuffer( IString::defaultBuffer() )
  {
  this->buffer() -> addRef();
  }

/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const IString *pIString )
  : pBuffer( IString::defaultBuffer() )
  {
  this->initBuffer( pIString->data(),
                    pIString->length(),
                    0, 0, 0, 0, 0 );
  }
/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( IBuffer * oldBuffer )
  {
  this->setBuffer( oldBuffer );
  }
/*------------------------------------------------------------------------------
| IString::IString                                                             |
------------------------------------------------------------------------------*/
IString :: IString ( const IText &anIText )
  : pBuffer( IString::defaultBuffer() )  
  {
  const char *pChar = anIText;
  this->initBuffer( (const void*) pChar,
                    lengthOf( pChar ),
                    0, 0, 0, 0, 0 );
  }
/*------------------------------------------------------------------------------
| IString::~IString                                                            |
------------------------------------------------------------------------------*/
IString :: ~IString ( )
  {
  this->buffer()->removeRef();
  this->pBuffer = 0;
  }

/*------------------------------------------------------------------------------
| IString::writeToStream                                                       |
------------------------------------------------------------------------------*/
void IString :: writeToStream ( IDataStream& toWhere ) const
  {
	long	numChars = this->length();
	
	numChars >>= toWhere;
	toWhere.writeBytes(pBuffer, numChars);
  }

/*------------------------------------------------------------------------------
| IString::readFromStream                                                      |
------------------------------------------------------------------------------*/
void IString :: readFromStream ( IDataStream& fromWhere )
  {
	IString	old(this->buffer());
	long numChars;

	old.buffer()->addRef();
	numChars <<= fromWhere;
	this->initBuffer(0, (unsigned int)numChars, 0, 0, 0, 0, ' ');
	fromWhere.readBytes(pBuffer, numChars);
	old.buffer()->removeRef();
  }
