// ----------------------------------------------------------------------------
// FILE NAME: itmstamp.cpp
//
// DESCRIPTION:
//   This file contains the implementation of classes/functions declared
//   in itmstamp.hpp.
//
// COPYRIGHT:
//   IBM Open Class Library
//   Licensed Materials - Property of IBM
//   (C) Copyright IBM Corporation 1992, 1997  All Rights Reserved.
//
//   US Government Users Restricted Rights - Use, duplication, or
//   disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
//
// Revision: 02  1.7.1.5  source/core/base/itmstamp.cpp, appsupport, ioc.v400, 000302  
// ----------------------------------------------------------------------------

#include <itmstamp.hpp>

extern "C"
{
    #include <time.h>
}

#ifdef _MSC_VER
// MSVC does not like C headers included with extern "C"
#include <math.h>
#else
extern "C"
{
    #include <math.h>
}
#endif

extern "C"
{
    #ifdef IC_WIN
    #include <windows.h>
    #endif
}

#include <istring.hpp>
#include <idate.hpp>
#include <itime.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
    #define _ITMSTAMP_CPP_
    #include <ipagetun.h>
#endif

const double _EXPORT ITimeStamp::secondsInDay  = 86400.0;

// 'baseCorrectionDays' is the number of days off
// ITimeStamp's reference date (i.e.01/01/2000 A.D)
// from julian 0 (i.e. 01/01/4713 B.C).
static const double baseCorrectionDays = -2451545.0;

/*------------------------------------------------------------------------------
| ITimeStamp::ITimeStamp                                                       |
|                                                                              |
| ITimeStamp is internally represented by the number of seconds off the        |
| reference date (i.e. 01/01/2000 A.D) and it is stored in a floating          |
| point type 'fTime'. If the given date is before the reference date, 'fTime'  |
| will have a negative value.                                                  |
|                                                                              |
| In order to get the 'fTime' value, we need to find out the number of days    |
| off the reference date by adding the julian number from 'aDate' and the      |
| 'baseCorrectionDays' (See above for description). The number of seconds can  |
| be found by multiplying the resulting days to 'secondsInDay'. If 'aTime' is  |
| also provided, the number of seconds in 'aTime' are also added in order to   |
| find the value for 'fTime'.                                                  |
------------------------------------------------------------------------------*/
ITimeStamp :: ITimeStamp ( const IDate &aDate,
                           const ITime &aTime )
{
  fTime = ((aDate.julianDate() + baseCorrectionDays) * secondsInDay) +
          aTime.asSeconds();
}

ITimeStamp :: ITimeStamp ( const IDate &aDate )
{
  fTime = (aDate.julianDate() + baseCorrectionDays) * secondsInDay;
}

/*------------------------------------------------------------------------------
| ITimeStamp::operator IDate                                                   |
|                                                                              |
| First, we need to divide 'fTime' by 'secondsInDay' to find out the number    |
| of days off the reference date. Then use the resulting days to subtract      |
| 'baseCorrectionDays' to find back the julian number. Finally, using the      |
| julian number to construct an IDate object.                                  |
------------------------------------------------------------------------------*/
ITimeStamp :: operator IDate ( ) const
{
  unsigned long julianDayNumber = (fTime / secondsInDay) - baseCorrectionDays;
  return IDate(julianDayNumber);
}

/*------------------------------------------------------------------------------
| ITimeStamp::operator ITime                                                   |
|                                                                              |
| First, we need to divide 'fTime' by 'secondsInDay' to find out the number    |
| of days off the reference date. Then call modf() to extract the fractional   |
| part and multiply by 'secondsInDay' to find out the number of seconds that   |
| are left in the fraction of the day. If the resulting seconds is a negative  |
| value, add 'secondsInDay' to get the positive value. Finally, using the      |
| 'seconds' value plus 'floatingErrorCorrection' to construct an ITime object. |
------------------------------------------------------------------------------*/
ITimeStamp :: operator ITime ( ) const
{
  static const double floatingErrorCorrection = 0.0001;
  double integralPart;
  double days = floor(fTime) / secondsInDay;

  // get the seconds that are left in the fraction of the day
  double seconds = modf(days, &integralPart) * secondsInDay;

  if (seconds < 0) {
    seconds += secondsInDay;
  }

  // ITime accepts long as the constructor parameter.
  // In case 'seconds' has value like x.9999x
  // (where x is any number of digits),
  // adding 'floatingErrorCorrection' is required; otherwise
  // the constructed value will be off by one when truncation
  // occur during double to long conversion!!
  return ITime(seconds + floatingErrorCorrection);
}

/*------------------------------------------------------------------------------
| ITimeStamp::currentTimeStamp                                                 |
|                                                                              |
| Get the current local time via system call. Use the provided data to         |
| construct an ITimeStamp accordingly.                                         |
------------------------------------------------------------------------------*/
ITimeStamp ITimeStamp :: currentTimeStamp ( )
{
#ifdef IC_WIN
  SYSTEMTIME now;
  GetLocalTime(&now);

  IDate aDate((IDate::Month) now.wMonth, now.wDay, now.wYear);
  double dateSeconds = (now.wHour * 3600.0) + (now.wMinute * 60.0) +
                       now.wSecond;
  dateSeconds += (aDate.julianDate() + baseCorrectionDays) * secondsInDay;

  return ITimeStamp(dateSeconds);

#endif
#ifdef IC_PM
  time_t t   = time(0);
  tm     now = *localtime(&t);
  return ITimeStamp(IDate(now.tm_year + 1900, now.tm_yday + 1),
                    ITime(now.tm_hour, now.tm_min, now.tm_sec));
#endif
#ifdef IC_POSIX
  time_t t   = time(0);
  tm     now;
  localtime_r(&t, &now);
  return ITimeStamp(IDate(now.tm_year + 1900, now.tm_yday + 1),
                    ITime(now.tm_hour, now.tm_min, now.tm_sec));
#endif
}

/*------------------------------------------------------------------------------
| ITimeStamp::asString                                                         |
|                                                                              |
| Construct an IString that has both the date and time.                        |
------------------------------------------------------------------------------*/
IString ITimeStamp :: asString ( ) const
{
#ifdef IC_WIN
  if (IString::isInternationalized())
  {
#endif

    return ((IDate)(*this)).asString(IDate::yyyy) + IString(" ") +
           ((ITime)(*this)).asString();

#ifdef IC_WIN
  } else {
    IDate aDate = *this;   // construct an IDate from ITimeStamp
    ITime aTime = *this;   // construct an ITime from ITimeStamp

    // Use the IDate and ITime values to initialize the SYSTEMTIME structure
    // and ready for use in GetDateFormat() and GetTimeFormat()
    SYSTEMTIME aSysTime;
    aSysTime.wYear         = aDate.year();
    aSysTime.wMonth        = (WORD) aDate.monthOfYear();
    aSysTime.wDayOfWeek    = ((int) aDate.dayOfWeek() + 1) % 7;
    aSysTime.wDay          = aDate.dayOfMonth();
    aSysTime.wHour         = aTime.hours();
    aSysTime.wMinute       = aTime.minutes();
    aSysTime.wSecond       = aTime.seconds();
    aSysTime.wMilliseconds = 0;

    // Get the formatted date length
    int dateLen = GetDateFormat(GetUserDefaultLCID(),
                                DATE_SHORTDATE,
                                &aSysTime,
                                NULL,
                                NULL,
                                0);

    char *dateBuf = new char[dateLen];

    GetDateFormat(GetUserDefaultLCID(),
                  DATE_SHORTDATE,
                  &aSysTime,
                  NULL,
                  dateBuf,
                  dateLen);

    // Get the formatted time length
    int timeLen = GetTimeFormat(GetUserDefaultLCID(),
                                0,
                                &aSysTime,
                                NULL,
                                NULL,
                                0);

    char *timeBuf = new char[timeLen];

    GetTimeFormat(GetUserDefaultLCID(),
                  0,
                  &aSysTime,
                  NULL,
                  timeBuf,
                  timeLen);

    IString dateTime = IString(dateBuf) + IString(" ") + IString(timeBuf);

    delete[] dateBuf;
    delete[] timeBuf;

    return dateTime;

  }
#endif
}

void ITimeStamp :: writeToStream( IDataStream &toWhere ) const
{
    fTime >>= toWhere;
}

void ITimeStamp :: readFromStream( IDataStream &fromWhere )
{
    fTime <<= fromWhere;
}
