// Revision: 50 1.4.1.5 source/core/base/win32/iwextprc.cpp, oss, ioc.v400, 000302 
//
//  FILE NAME: iwextprc.cpp
//
//  DESCRIPTION:
//
//  This module implements the IWinExtProcImpl which is the Win32
//  implementation derivative of IExtProcImplBase, from iwextproc.hpp.
//  This class provides the platform specific implementation code for
//  managing processes.
//
//
//  COPYRIGHT:
//    IBM Open Class Library
//    Licensed Materials - Property of IBM
//
//    5645-001
//    (C) Copyright IBM Corporation 1992, 1997
//
//    US Government Users Restricted Rights - Use, duplication, or
//    disclosure restricted by GSA ADP Schedule Contract with IBM
//    Corp.
//


// ----------------------------------------------------------------------------
//  Includes
// ----------------------------------------------------------------------------
#include    <windows.h>
#include    <icconsta.h>
#include    <iexcept.hpp>
#include    <iprcenv.hpp>
#include    <iextproc.hpp>
#include    <iwextprc.hpp>
#include    <itext.hpp>
#include    <istring.hpp>
#include    <iugnrl.hpp>
#include    <iautoptr.hpp>
#include    <itrancod.hpp>


#ifdef IC_PAGETUNE
    #define _IWEXTPRC_CPP_
    #include <ipagetun.h>
#endif

bool IWinExtProcImpl::fUseUnicode = true;

// ----------------------------------------------------------------------------
//  CLASS: IWinExtProcImpl
// ----------------------------------------------------------------------------
IWinExtProcImpl::IWinExtProcImpl() :
    fHandle(0)
{
}

IWinExtProcImpl::~IWinExtProcImpl()
{
    if (fHandle)
    {
        CloseHandle(fHandle);
        fHandle = 0;
    }
}

INonGUIApplication::PriorityClass IWinExtProcImpl::getPriority() const
{
    unsigned long prioClass = GetPriorityClass(fHandle);

    if (!prioClass)
    {
        unsigned long error = GetLastError();
        IMessageText errText(IC_EXTP_CANT_GET_PRIO
                            , kMsgFileName
                            , IPathName::defaultParser().formatName(getPath())
                            , IString(error));
        ITHROW(IAccessError(errText, IC_EXTP_CANT_GET_PRIO, IException::recoverable));
    }

    switch (prioClass)
    {
        case HIGH_PRIORITY_CLASS :
            return INonGUIApplication::timeCritical;

        case IDLE_PRIORITY_CLASS :
            return INonGUIApplication::idleTime;

        case REALTIME_PRIORITY_CLASS :
            return INonGUIApplication::foregroundServer;

        case NORMAL_PRIORITY_CLASS :
            return INonGUIApplication::regular;

        default :
            break;
    }

    // This should never happen
    IMessageText errText(   IC_EXTP_UNKNOWN_PRIO
                            , kMsgFileName
                            , IPathName::defaultParser().formatName(getPath()));

    ITHROW(IInvalidParameter(   errText,
                                IC_EXTP_UNKNOWN_PRIO,
                                IException::recoverable));
}

bool IWinExtProcImpl::exitStatus(unsigned long& status) const
{
    if (!GetExitCodeProcess(fHandle, &status))
        return false;

    if (status == STILL_ACTIVE)
        return false;
    return true;
}

bool IWinExtProcImpl::isRunning() const
{
    unsigned long resCode;

    if (!GetExitCodeProcess(fHandle, &resCode))
        return false;

    if (resCode == STILL_ACTIVE)
        return true;
    return false;
}

bool IWinExtProcImpl::setPriority(INonGUIApplication::PriorityClass priority)
{
    return (bool)SetPriorityClass(fHandle, priority);
}

void
IWinExtProcImpl::start(const IText* argArray, const unsigned long argCount)
{
    // Convert the path to text for convenience
    IText pathText(IPathName::defaultParser().formatName(getPath()));

    // Make sure its a legal length for an NT path
    if (pathText.empty() || (pathText.length() > MAX_PATH))
    {
        IMessageText errText(IC_EXTP_BAD_PATH,
                             kMsgFileName,
                             IString((const char*)pathText));

        ITHROW(IInvalidParameter(   errText,
                                    IC_EXTP_BAD_PATH,
                                    IException::recoverable));
    }

    //
    //  Build up the command line string. It will be the paramters that
    //  get passed to the program. Its a null terminated string that holds
    //  all of the parms, white space delimited.
    //
    //  By putting the process name as the first command line arg, we
    //  get the maximum bang for the buck out of CreateProcess(). Otherwise
    //  we would have to do the path searching.
    //
    IText   cmdLine(pathText);
    cmdLine.append(UASCII::kSpace);
    for (unsigned long index = 0; index < argCount; index++)
    {
        cmdLine.append(argArray[index]);
        cmdLine.append(UASCII::kSpace);
    }

    //
    //  Build up the environment string. This one is harder because we
    //  have to have a single string buffer that really contains multiple
    //  null terminated strings, with an extra one on the end. So we
    //  won't try to do that via IText directly. We'll use IText to build
    //  up the key=value strings, but manually copy them onto a raw buffer.
    //
    const unsigned long bufSize = 64 * 1024;

    // Allocate the buffer to build the environment in
    UniChar* envBuffer = new UniChar[bufSize];
    memset(envBuffer, 0, sizeof(UniChar)*bufSize);

    // Put an auto heap pointer on it so it gets cleaned up
    IAutoHeap autoCleanup(envBuffer);

    unsigned long bufPos = 0;

    //
    //  If there are any elements in the environment object, then
    //  lets use them to fill in the environment buffer. Otherwise we
    //  fill it with the needed 2 nul terminators.
    //
    if (fEnviron.getCount())
    {
        IText keyText;
        IText valueText;

        //
        //  Make sure that there is something there. The count was
        //  non-zero so we do it as an assert.
        //
        IASSERTSTATE(fEnviron.first(keyText, valueText));

        do
        {
            // Create the combined string
            IText combinedText = keyText;
            combinedText.append(UASCII::kEqualsSign);
            combinedText.append(valueText);

            // Get the length of the combined string
            IText::length_type combinedLen = combinedText.length();

            // Leave space for the final null, so sub 2 from the buffer size!
            if ((bufPos + combinedLen + 1) >= (bufSize - 2))
            {
                IMessageText errText(IC_EXTP_ENV_BUFFER_FULL, kMsgFileName);
                ITHROW(IResourceExhausted(  errText,
                                            IC_EXTP_ENV_BUFFER_FULL,
                                            IException::recoverable));
            }

            //
            //  Copy this one on the end end of the buffer, skipping over
            //  the null so it goes past it and starts a new string, not
            //  a concatenation.
            //
            combinedText.copy(&envBuffer[bufPos], combinedLen+1);

            // Move up past the text and the nul
            bufPos += combinedLen+1;

        }   while (fEnviron.next(keyText, valueText));

        // And now add a last terminating null at the end
        envBuffer[bufPos] = UASCII::kNull;
    }
     else
    {
        // Set up an empty environment
        envBuffer[0] = 0;
        envBuffer[1] = 0;
    }

    PROCESS_INFORMATION procInfo = {0};
    BOOL result = 0;

    if (fUseUnicode) {
        // Fill out a startup info structure with defaults
        STARTUPINFOW startupInfo = {0};
        startupInfo.cb = sizeof(STARTUPINFOW);
        
        //
        //  Whew... Ok, now we can actually run the stupid program. Note that
        //  we are using the unicode enabled version since all of this incoming
        //  data is Unicode enabled stuff.
        //
        result = CreateProcessW(0, (LPWSTR)cmdLine.c_str(), 0, 0, FALSE,
                                CREATE_UNICODE_ENVIRONMENT | 
                                NORMAL_PRIORITY_CLASS,
                                envBuffer, 0, &startupInfo, &procInfo);
        if (result == 0 && GetLastError() == ERROR_CALL_NOT_IMPLEMENTED)
            fUseUnicode = false;
    }
    
    if (! fUseUnicode) {
        
        STARTUPINFO startupInfo = {0};
        startupInfo.cb = sizeof(STARTUPINFO);

        // Here, we use all ASCII interface, since Win95 (or even Win98)
        // does not support CreateProcessW and Unicode environment string.

        ITranscoder *myTranscoder = ITranscoder::createTranscoder();
        IString mb_cmdLine;
        
        ITranscoder::length_type charBufSize = 
            myTranscoder->byteBufferSize(&envBuffer[0], &envBuffer[bufPos + 1]);
        char *mb_envBuffer = new char[charBufSize + 1];

        // Use the default transcoder to transcode our command line string
        // and the environment buffer.

        const UniChar *from_next = 0;
        char *to_next = 0;

        if (myTranscoder->fromUnicode(cmdLine, mb_cmdLine) == codecvt_base::ok &&
            myTranscoder->fromUnicode(&envBuffer[0], &envBuffer[bufPos + 1], 
                                      from_next, 
                                      &mb_envBuffer[0], &mb_envBuffer[charBufSize],
                                      to_next) == codecvt_base::ok) {
            // Start execution only if we can successfully transcode both the
            // command line and the environment variables.
            result = CreateProcess(0, mb_cmdLine, 0, 0, FALSE,
                                   NORMAL_PRIORITY_CLASS,
                                   (LPVOID)mb_envBuffer, 0,
                                   &startupInfo, &procInfo);
        }
        delete myTranscoder;
        delete mb_envBuffer;
    }

    if (result == 0) {
        IMessageText errText(IC_EXTP_CANT_START,
                             kMsgFileName,
                             IString((const char*)pathText),
                             IString(GetLastError()));
        ITHROW(IAccessError(errText,
                            IC_EXTP_CANT_START,
                            IException::recoverable));
    }

    // It worked so store away the information
    setId(IProcessId((void*)procInfo.dwProcessId));

    // For our internal purposes we store the handle also
    fHandle = procInfo.hProcess;
}

void IWinExtProcImpl::kill()
{
    if (!TerminateProcess(fHandle, 0))
    {
        IMessageText errText(   IC_EXTP_CANT_KILL,
                                kMsgFileName,
                                IPathName::defaultParser().formatName(getPath()),
                                IString(GetLastError()));

        ITHROW(IAccessError(errText,
                            IC_EXTP_CANT_KILL,
                            IException::recoverable));
    }
}

bool IWinExtProcImpl::wait(const long timeout) const
{
    // Wait for the handle for the given timeout
    unsigned long resValue = WaitForSingleObject(fHandle, timeout);

    if (resValue == WAIT_FAILED)
    {
        IMessageText errText(IC_EXTP_ERROR_WAITING_ON,
                             kMsgFileName,
                             IString(GetLastError()),
                             IPathName::defaultParser().formatName(getPath()));

        ITHROW(IAccessError(errText,
                            IC_EXTP_ERROR_WAITING_ON,
                            IException::recoverable));
    }

    // If it didn't fail then it either timed out or its ok
    if (resValue == WAIT_TIMEOUT)
        return false;

    return true;
}
