// Revision: 55 1.14.1.5 source/core/filesys/directry.cpp, filesystem, ioc.v400, 000302 
// directry.cpp
//
// Contains: Implementation of IDirectory
//
// Author:
//      Laura Werner
//
// COPYRIGHT:
//      IBM Open Class Library
//      (C) Copyright International Business Machines Corporation 1992, 1997
//      Licensed Material - Program-Property of IBM - All Rights Reserved.
//      US Government Users Restricted Rights - Use, duplication, or disclosure
//      restricted by GSA ADP Schedule Contract with IBM Corp.
//
#include <igenexc.hpp>
#include <iexcept.hpp>

#include <ifilesys.hpp>
#include <iexcfile.hpp>
#include <ipathnam.hpp>
#include "fsdirrec.hpp"
#include "filesys.hpp"

#ifdef IC_PAGETUNE
    #define _DIRECTRY_CPP_
    #include <ipagetun.h>
#endif

extern IStreamModule gFileSystemModule;

StreamableDefinitionsMacro(IDirectory, gFileSystemModule);

#pragma info(nodcl)
const IFileKind _EXPORT IDirectory::kKind = "IDirectory";
#pragma info(restore)

//------------------------------------------------------------------------------
// Constructors, destructors, etc....
//

IDirectory::IDirectory  ()
{
}

IDirectory::IDirectory (const IHostPathName &path, ECreateFlags flags,
                                                const ICompoundNameParser& parser)
{
        typeSafeAssign( IFSFileSystem::instance()->lookUp(path, parser, flags, kKind),
                                        kKind );
}

IDirectory::IDirectory (const IPathName &path, ECreateFlags flags )
{
        typeSafeAssign( IFSFileSystem::instance()->lookUp(path, flags, kKind),
                                        kKind );
}

IDirectory::IDirectory  (IFSDirectoryRecord* record)
        : IFileSystemEntity(record)
{
}

IDirectory::IDirectory  (const IDirectory& copy)
        : IFileSystemEntity(copy)
{
}

IDirectory::~IDirectory ()
{
}


//------------------------------------------------------------------------------
//      Downcast "Special" Functions
//  These functions will throw a kTypeMismatchNotADirectory exception if types are different.
//

IDirectory::IDirectory  (const IFileSystemEntity& copy)
{
        operator = (copy);
}

IDirectory&
IDirectory::operator = (const IFileSystemEntity& right)
{
        typeSafeAssign (right, IDirectory::kKind);
        return *this;
}


//------------------------------------------------------------------------------
// Creation and deletion of entities in the directory
//

IFile
IDirectory::createFile(const IFileName& name, IFileSize initialSize,
                                                const ICompoundNameParser &parser)
{
        if ( !connected() ) {
                IEntityInvalid e;
                ITHROW(e);
        }

        IFile result;

        try {
                // Try it the simple way first
                result = getRecord()->createFile(name, initialSize);
        }
        catch (const IObjectNotFound&) {
                // An intermediate directory didn't exist.  Try to create it
                result = IFSFileSystem::instance()->createPath(*this,
                                                                                                                IPathName(name, parser),
                                                                                                                IFile::kKind);
        }
        return result;
}

IDirectory
IDirectory::createDirectory     (const IFileName& name,
                                                         const ICompoundNameParser& parser)
{
        if ( !connected() ) {
                IEntityInvalid e;
                ITHROW(e);
        }

        IDirectory result;

        try {
                // Try it the simple way first
                result = getRecord()->createDirectory(name);
        }
        catch (const IObjectNotFound&) {
                // An intermediate directory didn't exist.  Try to create it
                result = IFSFileSystem::instance()->createPath(*this,
                                                                                                        IPathName(name, parser),
                                                                                                        IDirectory::kKind);
        }
        return result;
}

void
IDirectory::deleteAllContents()
{
        IFSDirectoryRecord* Record = (IFSDirectoryRecord*) getRecord();
        Record->deleteAllContents();
}


//------------------------------------------------------------------------------
// Lookup of entities in the directory
//

IFileSystemEntity
IDirectory::lookUp (const IPathName& partialPath) const
{
        IASSERTPARM( !partialPath.isFullPath() )

        IHostPathName aHostPath = defaultParser().formatName(partialPath);

        return getRecord()->lookUpByName(aHostPath);
}

IFileSystemEntity
IDirectory::lookUp      (const IHostPathName& pathOrName,
                                                                                        const ICompoundNameParser &parser) const
{
        IFileSystemEntity result;

        if ( parser == defaultParser())
        {
                // Optimization: Just look it up
                result = getRecord()->lookUpByName(pathOrName);
        }
        else
        {
                // Convert the path to a IPathName, then look that up
                IPathName pathName = parser.parseName(pathOrName);

                result = lookUp(pathName);
        }

        return result;
}


//------------------------------------------------------------------------------
//      Current directory support
//

IDirectory
IDirectory::current()
{
        return IFSFileSystem::instance()->getCurrentDirectory();
}

void
IDirectory::setCurrent(const IDirectory & dir)
{
        dir.getRecord()->setCurrent();
}


//------------------------------------------------------------------------------
//      Temporary directories
//

IDirectory
IDirectory::createTemporary(const IFileName& name )
{
        IDirectory result = IFSFileSystem::instance()->createTempDirectory(name);
        result.getRecord()->setTemporary(true);
        return result;
}


