// Revision: 56 1.18.1.3 source/core/filesys/entity.cpp, filesystem, ioc.v400, 000302 
// entity.cpp
//
// Contains: Implementation of IFileSystemEntity
//
// Author:
//      Laura Werner
//
// COPYRIGHT:
//      IBM Open Class Library
//      (C) Copyright International Business Machines Corporation 1992, 1997
//      Licensed Material - Program-Property of IBM - All Rights Reserved.
//      US Government Users Restricted Rights - Use, duplication, or disclosure
//      restricted by GSA ADP Schedule Contract with IBM Corp.
//
#include <ifilesys.hpp>
#include <iexcfile.hpp>
#include <ipathnam.hpp>
#include <iprimlck.hpp>
#include <istrmmod.hpp>

#include "fsrecord.hpp"
#include "filesys.hpp"

#include <itmstamp.hpp>

#ifdef IC_PAGETUNE
    #define _ENTITY_CPP_
    #include <ipagetun.h>
#endif

//------------------------------------------------------------------------------
// IFileSystemEntity Constants/Globals
//
IStreamModule gFileSystemModule("FileSystem");
StreamableDefinitionsMacro(IFileSystemEntity, gFileSystemModule);


//------------------------------------------------------------------------------
// IFileSystemEntity Lazy evaluator access methods
//
const IFileName& IFileSystemEntity::uniqueName()
{
    static const IFileName* gUniqueName = 0;

    if (!gUniqueName)
    {
        IPrimalLock lockInit;

        if (!gUniqueName)
            gUniqueName = new IFileName("");
    }
    return *gUniqueName;
}


//------------------------------------------------------------------------------
// Constructors, destructors, etc...
//

IFileSystemEntity::IFileSystemEntity ()
        : fRecord(0)
{
}

IFileSystemEntity::~IFileSystemEntity ()
{
        if (fRecord) fRecord->removeRef();
}

IFileSystemEntity::IFileSystemEntity (IFSRecord* record)
        : fRecord(record)
{
        if (fRecord) fRecord->addRef();
}

IFileSystemEntity::IFileSystemEntity (const IFileSystemEntity& copy)
        : fRecord(copy.fRecord)
{
        if (fRecord) fRecord->addRef();
}

IFileSystemEntity::IFileSystemEntity (const IHostPathName& path,
                                                                                const ICompoundNameParser& parser)
        : fRecord(0)
{
        operator=( IFSFileSystem::instance()->lookUp(path, parser, kMustExist, 0) );
}

IFileSystemEntity::IFileSystemEntity (const IPathName &path)
        : fRecord(0)
{
        operator=( IFSFileSystem::instance()->lookUp(path, kMustExist, 0) );
}

IFileSystemEntity&
IFileSystemEntity::operator = (const IFileSystemEntity& right)
{
        if (fRecord != right.fRecord) { // Do nothing if same Record object.
                if (fRecord) fRecord->removeRef();
                fRecord = right.fRecord;
                if (fRecord) fRecord->addRef();
        }
        return (*this);
}


//------------------------------------------------------------------------------
// typeSafeAssign
//      Utility function that does an assignment only if the entity types match
//
void IFileSystemEntity::typeSafeAssign  (const IFileSystemEntity& right,
                                                                                         const IFileKind& kind)
{
        if (fRecord != right.fRecord)
        {
                // If incoming handle has a NIL Record pointer then we do not
                // need to do the type check.  Otherwise do it...

                if ( right.fRecord && !right.isA(kind) ) {
                  IEntityTypeMismatch e(kind, right);
                  ITHROW(e);
                }
                // If we get here, it's safe to do the assignment...
                if (fRecord) fRecord->removeRef();
                fRecord = right.fRecord;
                if (fRecord) fRecord->addRef();
        }
}

bool IFileSystemEntity::isA (const IFileKind& kind) const
{
         return fRecord ? fRecord->isA(kind) : false;
}


//------------------------------------------------------------------------------
// Comparison
//
bool
IFileSystemEntity::operator== (const IFileSystemEntity& other) const
{
        bool result;

        if (fRecord == other.fRecord) {
                // Records pointers are identical, so return true.  Note that NIL pointers
                // are OK here -- this allows kEndOfIteration entities to work correctly...
                result = true;
        } else {
                if (!fRecord || !other.fRecord) {
                        // Record pointers are not identical, and one of them is NIL --
                        // return false...
                        result = false;
                } else {
                        // Record pointers are not identical, and neither is NIL -- let
                        // the Record decide equality...
                        result = fRecord->kind() == other.fRecord->kind() // temporary
                                        && fRecord->isEqual(other.fRecord);
                }
        }
        return (result);
}


//------------------------------------------------------------------------------
// Streaming methods...
//
void
IFileSystemEntity::readFromStream(IDataStream& stream)
{
        // Stream in an entity and assign to myself.  This will catch type mismatches.
        operator=( IFSRecord::streamIn(stream) );
}

void
IFileSystemEntity::writeToStream(IDataStream& stream) const
{
        getRecord()->streamOut(stream);
}


//------------------------------------------------------------------------------
// Entity validity...
//
void IFileSystemEntity::invalidRecord () const
{
        IEntityInvalid e;
        ITHROW(e);
}

void IFileSystemEntity::setInvalid ()
{
        if (fRecord) fRecord->removeRef();
        fRecord = 0;
}

bool IFileSystemEntity::connected       () const
{
        return fRecord && fRecord->testConnection();
}


//------------------------------------------------------------------------------
// IFileSystemEntity attributes...
//
const ICompoundNameParser&
IFileSystemEntity::defaultParser()
{
        return IPathName::defaultParser();
}

IHostPathName
IFileSystemEntity::hostPath(const ICompoundNameParser& parser) const
{
        IHostPathName aHostPath;

        if (parser == IPathName::defaultParser()) {
                // They want a path in this hosts's format.  Easy!
                aHostPath = getRecord()->hostPath();
        } else {
                // We must convert to another host's format.  Ok.
                aHostPath = parser.formatName( getRecord()->pathName() );
        }
        return aHostPath;
}

IFileName IFileSystemEntity::name () const
{
        return getRecord()->name();
}

void IFileSystemEntity::setName (const IFileName& newName) const
{
        getRecord()->setName(newName);
}


ITimeStamp IFileSystemEntity::modificationTime () const
{
        return getRecord()->modificationTime();
}

bool IFileSystemEntity::temporary       () const
{
        return getRecord()->temporary();
}

IPathName                       IFileSystemEntity::path () const
{
        return getRecord()->pathName();
}


//------------------------------------------------------------------------------
// Navigation
//
IVolume IFileSystemEntity::volume () const
{
        return getRecord()->volume();
}

IDirectory IFileSystemEntity::parent    () const
{
        return getRecord()->parent();
}

IDirectory IFileSystemEntity::topmostParent() const
{
        return getRecord()->topmostParent();
}

//------------------------------------------------------------------------------
//      Entity manipulation functions...
//

void
IFileSystemEntity::deleteSelf   (bool deleteChildren)
{
        getRecord()->deleteSelf(deleteChildren);
        setInvalid();
}

void
IFileSystemEntity::moveTo (IDirectory& destination,
                                                        const IFileName& newName,
                                                        EFailureAction action)
{
        IFileSystemMover mover( (IFileSystemMover::EFailureAction)action);
        mover.move( *this, destination, newName );
}

IFileSystemEntity
IFileSystemEntity::copyTo (IDirectory& destination,
                                                        const IFileName& newName,
                                                        EFailureAction action) const
{
        IFileSystemCopier copier((IFileSystemCopier::EFailureAction)action);
        return copier.copy(*this, destination, newName);
}



