/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/

// Revision: 40 1.16.1.6 source/core/filesys/filebuf.cpp, filesystem, ioc.v400, 000302 
// filebuf.cpp
// Created by laura on Fri, Feb 08, 1996 @ 11:58 AM.

#pragma enum(4)
#include <iostream.h>
#pragma enum(pop)

#include <iexcept.hpp>

#include "filebuf.hpp"
#include "fsfilrec.hpp"
#include "fsstream.hpp"

#ifdef IC_PAGETUNE
    #define _FILEBUF_CPP_
    #include <ipagetun.h>
#endif

//-------------------------------------------------------------------------------
// Helper functions for converting between the ios enum values and the ones
// used in the file system.  It would be simpler if the filesystem just used
// the ios enums internally, but it's not allowed to have any dependencies
// on the C++ streams library
//
static IFSFileStream::seek_dir tseek(ios::seek_dir a)
{
        // It's possible for "dir" to have garbage in the high bytes, because of a
        // compiler switch we're using.  Mask it out.
        a = (ios::seek_dir) (a & (ios::beg | ios::cur | ios::end));

    IFSFileStream::seek_dir result = IFSFileStream::beg;

    if (a == ios::cur) {
        result = IFSFileStream::cur;
    }
    else if (a == ios::end) {
        result = IFSFileStream::end;
    }
    return result;
}

//-------------------------------------------------------------------------------
// Class:
//      IFSFileBuf
//
// Purpose:
//      streambuf subclass that works with the Pluto filesystem API
//

// ------------ Constructors, etc. ----------------------

IFSFileBuf::IFSFileBuf(const IFile &file, int modeForMe, int modeForOthers)
        : fFile( )
        , fMode( 0 )
        , fBufferSize( kDefaultBuffer )
        , fBuffer( 0 )
        , fFilePosition( 0u )
        , fByteStream( 0 )
        , fPbackEnd( 0 )
{
        open( file, modeForMe, modeForOthers );
}

IFSFileBuf::IFSFileBuf(const IFSFileBuf &copy)
{
        // Disallowed
}

IFSFileBuf &
IFSFileBuf::operator = (const IFSFileBuf &copy)
{
        // Disallowed
        return *this;
}


IFSFileBuf::~IFSFileBuf()
{
        if (fByteStream) {
                syncout();                              // Make sure everything is flushed
                fByteStream->close();   // Close the file
                delete fByteStream;
        }

        delete[] fBuffer;
}


// ------------ open and close ----------------------

void
IFSFileBuf::open(const IFile &file, int myMode, int otherMode)
{
        IASSERTSTATE( !fByteStream )

        if (!fBuffer) {
                fBuffer = new char[fBufferSize + kPutBackSize];
        }

        fByteStream = file.getRecord()->createStream(myMode, otherMode);
        fMode = myMode;
        // The ANSI standard sort of implies that ios::app implies ios::out,
        // and this was reported as bug #803, so I might as well do it....
        if (fMode & ios::app) {
                fMode |= ios::out;
        }
        fFilePosition = 0u;

        try {
                if (fMode & (ios::app | ios::ate)) {
            // if it's append mode, set file position to the end of file
                        // another possible call is to seek to the end of the file like this:
            // fFilePosition = fByteStream->doSeek(0, IFSFileStream::end);
            fFilePosition = fByteStream->endOfFile();
                }
        }
        catch (...) {
                delete fByteStream;
                fByteStream = 0;
                fMode = 0;
#ifndef SKIPOMPARSE
                throw;
#endif
        }
}


void
IFSFileBuf::close()
{
        IASSERTSTATE( fByteStream )

        if (fByteStream) {
                syncout();                              // Make sure everything is flushed
                fByteStream->close();   // Close the file
                delete fByteStream;
                fByteStream = 0;
        }
        fFile = IFile();
        fMode = 0;
        fFilePosition = 0u;
}



// ------------ Reading and Writing ----------------------

int
IFSFileBuf::overflow(int c)
{
    // Make sure this is legal for our mode. If not, then return EOF.
    if (!(fMode & ios::out ))
        return EOF;

        // Make sure the buffers are synchronized with the file
        sync();

        // Set up a new output buffer
        setp(fBuffer, fBuffer + fBufferSize);

        // If "c" is not EOF, we have to write it.  But we can cheat
        // and just stick it at the start of the write buffer
        if (c != EOF) {
                *pptr() = (char)c;
                pbump(1);
        }
        return 0;       // Anything other than EOF signals success
}

int
IFSFileBuf::underflow()
{
    // Make sure this is legal for our mode. If not, then return EOF.
        if (!(fMode & ios::in))
        return EOF;

        int result = EOF;

    if ( gptr() && gptr() < egptr() ) {
        // There is already data in the buffer.  Just return the current character
        result = *gptr();
    }
    else {
        // We have to read some more data from the file
        // First make sure the buffers are synchronized with the file
        sync();

        // Leave room for putting a few characters back
        char *readBuf = fBuffer + kPutBackSize;

        size_t bytesRead = fByteStream->doRead(readBuf, fBufferSize);
        if (bytesRead) {
                fFilePosition += bytesRead;
                setg(fBuffer, readBuf, readBuf + bytesRead);
                fPbackEnd = gptr();
                result = (unsigned char) *gptr();       // avoid sign extension
        }
    }
        return result;
}

int
IFSFileBuf::pbackfail(int c)
{
    if (!(fMode & ios::in))
        return EOF;

        int success = 1;        // Anything other than EOF

        if (c == EOF) {
                // We're really just seeking one character backwards.
                seekoff(-1, ios::cur, ios::in);
        } else if (!gptr()) {
                // There's no input buffer at all right now.
                // Create one and stick the character onto it
                setg(fBuffer, fBuffer + fBufferSize - 1, fBuffer + fBufferSize);
                fPbackEnd = egptr();
                fBuffer[fBufferSize-1] = (char)c;
        } else {
                // We're putting a back into an existing
                // input buffer.  Slide the buffer over a ways to make room.

                const char * const bufEnd = fBuffer + fBufferSize + kPutBackSize;
                size_t delta = fBufferSize / 4;
                if (delta > bufEnd - fPbackEnd) {
                        delta = bufEnd - fPbackEnd;
                }
                if (delta) {
                        memmove(fBuffer + delta, fBuffer, fBufferSize + kPutBackSize - delta);
                        gbump(delta - 1);
                        fPbackEnd += delta;

                        *gptr() = (unsigned char) c;
                } else {
                        success = 0;    // No room left to put anything back
                }
        }
        return success;
}

int
IFSFileBuf::sync()
{
        syncout();              // Flush the output buffer
        syncin();               // Sync up the input buffer
        return 0;               // Success!
}


// ------------ Stream position ----------------------

streampos
IFSFileBuf::seekoff(streamoff offset, ios::seek_dir dir, int )
{
        sync();                                                                                         // Flush the buffers
        fFilePosition = fByteStream->doSeek(offset, tseek(dir));        // Do the seek
        return fFilePosition;
}


streampos
IFSFileBuf::seekpos(streampos offset, int )
{
        syncout();                      // Flush the output buffer
        setg(0,0,0);            // Discard the input buffer
        fPbackEnd = 0;

        fFilePosition = fByteStream->doSeek(offset, IFSFileStream::beg);
        return fFilePosition;
}

IFileSize
IFSFileBuf::position() const
{
        IFileSize result = fFilePosition;

        //  In the latest draft standard, pbase() etc. are "const"
        //      so this cast won't be necessary.
        IFSFileBuf *This = (IFSFileBuf *) this;

        // Account for any characters that are still in the buffer
        if ( This->pbase() ) {
                result += This->pptr() - This->pbase();
        }
        else if ( This->gptr() ) {
                result -= This->egptr() - This->gptr();
        }

        return result;
}


// ------------ File size ----------------------

IFileSize
IFSFileBuf::endOfFile() const
{
        return fByteStream->endOfFile();
}


void
IFSFileBuf::setEndOfFile(IFileSize size)
{
        fByteStream->setEndOfFile(size);
}

void            IFSFileBuf::syncin()
{
        // If we're in read mode, forget the remaining characters in the
        // input buffer and back up the file pointer to match
        if ( gptr() && gptr() < egptr() ) {
                ptrdiff_t count = egptr() - gptr();
                if (count) {
                        fFilePosition = fByteStream->doSeek(-count, IFSFileStream::cur);
                }
        }
        setg(0,0,0);
        fPbackEnd = 0;
}


void                    IFSFileBuf::syncout()
{
        // If we're in write mode, flush the output bufer
        if ( pbase() ) {
                size_t count = pptr() - pbase();
                if (count) {
                        fByteStream->doWrite(pbase(), count);
                        fFilePosition += count;
                }
                setp(0,0);
        }
}
