// Revision: 52 1.16.1.3 source/core/filesys/filesys.cpp, filesystem, ioc.v400, 000302 
// directry.cpp
//
// Contains:
//    IFSFileSystem, the internal implementation class that provides an abstract
//    interface to much of the filesystem's host-specific innards.
//
// Author:
//    Laura Werner
//
// COPYRIGHT:
//    IBM Open Class Library
//    (C) Copyright International Business Machines Corporation 1992, 1997
//    Licensed Material - Program-Property of IBM - All Rights Reserved.
//    US Government Users Restricted Rights - Use, duplication, or disclosure
//    restricted by GSA ADP Schedule Contract with IBM Corp.
//
#include <ilanglvl.hpp>
#include <ifilesys.hpp>
#include <iexcfile.hpp>
#include <ipathnam.hpp>
#include <igenexc.hpp>
#include <iexcept.hpp>
#include <iprimlck.hpp>

#include "filesys.hpp"
#include "fsalias.hpp"



#ifdef IC_PAGETUNE
    #define _FILESYS_CPP_
    #include <ipagetun.h>
#endif

//-----------------------------------------------------------------------------------
// IFSFileSystem globals and constants
//

// Provide a variable containing the name of the file system's message file.
// The kMsgFileName constant is provided on the compiler's command line
// by the build system.
const char* IFSFileSystem::kMessageFile = kMsgFileName;


//-----------------------------------------------------------------------------------
//  Gen up the file system instance object upon request
//-----------------------------------------------------------------------------------
static IFSFileSystem* gFSInstance = 0;
IFSFileSystem* IFSFileSystem::instance()
{
    if (!gFSInstance)
    {
        IPrimalLock lockInit;

        if (!gFSInstance)
            gFSInstance = IFSFileSystem::initialize();
    }
    return gFSInstance;
}


//-----------------------------------------------------------------------------------
// Constructors, etc...
//
IFSFileSystem::IFSFileSystem()
{
    gFSInstance = this;
}

IFSFileSystem::IFSFileSystem (const IFSFileSystem& )
{
    // Disallowed.
}

IFSFileSystem& IFSFileSystem::operator = (const IFSFileSystem& copy)
{
    // Disallowed.
    return (*this); // Suppress warning.
}

IFSFileSystem::~IFSFileSystem ()
{
}


//-----------------------------------------------------------------------------------
// lookup(IPathName)
//
//    Since most platforms will want a string-based path rather than a
//    componentized version, convert the IPathName to a host-specific path name
//    and look that up.
//
IFileSystemEntity
IFSFileSystem::lookUp(
    const IPathName&                path,
    IFileSystemEntity::ECreateFlags    flags,
    IFileKind                        createKind ) const
{
    IHostPathName aHostPath = getDefaultParser().formatName(path);
    IFileSystemEntity result = lookUp(aHostPath, flags, 0);

    if (!result.valid()) {
        // Couldn't find the entity, so try to auto-create it.
        result = createPath(path, createKind);
    }
    return result;
}

//-----------------------------------------------------------------------------------
// lookup(IHostPathName, parser)
//
//    If the path name happens to be in this host's format, we can just look
//    it up directly.  If not, we have to convert it to a portable IPathName
//    and look that up instead.
//
IFileSystemEntity
IFSFileSystem::lookUp(
    const IHostPathName&            path,
    const ICompoundNameParser&      parser,
    IFileSystemEntity::ECreateFlags flags,
    IFileKind                       createKind ) const
{
    IFileSystemEntity result;

    if ( parser == getDefaultParser() ) {
        result = lookUp(path, flags, createKind);
    }
    else {
        result = lookUp( IPathName(path, parser), flags, createKind );
    }
    return result;
}

//-----------------------------------------------------------------------------------
// lookup(IHostPathName)
//
//    Look up a path name string in this host's format.
//
IFileSystemEntity
IFSFileSystem::lookUp(
    const IHostPathName&            path,
    IFileSystemEntity::ECreateFlags flags,
    IFileKind                       createKind ) const
{
    IFileSystemEntity result;

    if (flags == IFileSystemEntity::kMustExist) {
        // The default behavior is to find an existing file or directory
        // Just look it up, and pass exceptions through.
        result = doLookUp(path);
    }
    else {
        try {
            result = doLookUp(path);

            if (flags == IFileSystemEntity::kReplaceExisting
                                            && result.isA(createKind) ) {
                // Clobber the existing entity because we're in replace mode.
                // But only replace entities of the same type.
                // Mismatched types will be caught later when someone
                // tries to do a type-safe assign.
                result.deleteSelf(true);
            }
            else if ( flags == IFileSystemEntity::kCreateOnly ) {
                // It already exists, but "create only" was specified.  Oops.
                IAlreadyExists e;
                ITHROW(e);
            }
        }
        catch (const IObjectNotFound& ) {
            // It doesn't exist.  That's ok: createPath will create it
        }
        if ( createKind && !result.valid() ) {
            result = createPath( path, createKind );
        }
    }
    return result;
}

//-----------------------------------------------------------------------------------
// createPath(IPathName)
//
//    Create the file system entity specified by the given IPathName,
//    creating intermediate directories if necessary.
//
IFileSystemEntity
IFSFileSystem::createPath(  const IPathName& path,
                            IFileKind createKind ) const
{
    // First we must find or create the parent directory of the entity
    // entity we're supposed to create.
    // As an optimization, we'll strip off the final component
    // and try to find a directory using the resulting path.  This should
    // be the common case, because usually the final component will
    // be the only one that doesn't exist.

    IDirectory parent;
    try {
        IPathName parentPath = path;
        parentPath.truncate(1);
        if (parentPath.numberOfComponents()) {
            IHostPathName hostParent =  getDefaultParser().formatName(parentPath);
            parent = doLookUp(hostParent);
        }
    }
    catch ( const IObjectNotFound & ) {
    }

    // If that didn't work, iterate through the path one component
    // at a time and make sure each one exists
    // The first step is to figure out what directory to start in.

    if ( !parent.valid() ) {
        unsigned long index = 1;
        unsigned long count = path.numberOfComponents();

        if (path.isFullPath()) {
            // A full path is tricky; we must let the appropriate parser
            // convert its first component into a host-specific path,
            // because the component by itself might not be enough

            IPathName tempPath(path);
            tempPath.truncate(tempPath.numberOfComponents() - 1);

            IHostPathName hostPath = getDefaultParser().formatName(tempPath);

            // WJW - I added this to handle the case of creating
            // a new directory in '/' on AIX.
            // Unfortunately, I can't test it because I don't have
            // sufficient privileges on '/'.
            try
              {
                parent = doLookUp(hostPath);
              }
            catch(const IObjectNotFound &)
              {
                // Let the parser handle the creation of a string pointing to
                // the root filesystem.
                IPathName rootName;
                rootName.setFullPath(true);
                hostPath = getDefaultParser().formatName(rootName);
                parent = doLookUp(hostPath);
                parent = parent.createDirectory(tempPath.componentAt(index));
              }

            index++;
        } else {
            // Relative path; just start from the current directory
            parent = IDirectory::current();
        }

        while (index < count) {
            try {
                parent = parent.lookUp( path.componentAt(index) );
            }
            catch ( const IObjectNotFound & ) {
                parent = parent.createDirectory( path.componentAt(index) );
            }
            index++;
        }
    }

    // Now we've found or created the entity's parent directory.
    // Create the entity.
    IFileSystemEntity result;

    if (createKind == IDirectory::kKind) {
        result = parent.createDirectory( path.lastComponent() );
    } else {
        result = parent.createFile( path.lastComponent() );
    }
    return result;
}

//-----------------------------------------------------------------------------------
// createPath(IDirectory, IPathName)
//
//    Create the file system entity specified by the given partial IPathName,
//    relative to the given directory.
//    Creates intermediate directories if necessary.
//
//  This could be replaced by a call to
//            createPath(IPathName(parentArg, path), createKind)
//        saving a little bit of code but losing a bit of speed.
//
IFileSystemEntity
IFSFileSystem::createPath(  const IDirectory &parentArg,
                            const IPathName& path,
                            IFileKind createKind ) const
{
    IFileSystemEntity result;

    IASSERTPARM( !path.isFullPath() )
    IDirectory parent = parentArg;

    // Try to find or create each directory along the path to the desired entity
    unsigned long count = path.numberOfComponents();
    for (unsigned long i = 1; i < count; i++) {
        try {
            parent = parent.lookUp( path.componentAt(i) );
        }
        catch ( const IObjectNotFound & ) {
            parent = parent.createDirectory( path.componentAt(i) );
        }
    }

    // Now that we've found or created the entity's parent directory,
    // create the entity itself.
    if (createKind == IDirectory::kKind) {
        result = parent.createDirectory(path.lastComponent());
    } else {
        result = parent.createFile( path.lastComponent() );
    }
    return result;
}

//-----------------------------------------------------------------------------------
// resolveAlias
//    Try to find a file system entity corresponding to the given IFSAlias
//
IFileSystemEntity
IFSFileSystem::resolveAlias( const IFSAlias& alias ) const
{
    IFileSystemEntity result;

    // I can resolve a "default" alias containing a host-specific path
    // and the name of the IFSFileSystem subclass that created it.
    // In the future this mechanism will be more extensible.

#if IC_RTTI_ENABLED
    if (alias.type() == IFileName(typeid(*this).name()))
#else
    if (alias.type() == this->typeOf())
#endif // IC_RTTI_ENABLED
    {
        result = doLookUp( alias.path() );
    }

    return result;
}
