// Revision: 42 1.6.1.3 source/core/filesys/fsdirrec.cpp, filesystem, ioc.v400, 000302 
// fsdirrec.cpp
//
// Contains:    IFSDirectoryRecord
//                              Internal filesystem class that encapsulates directory operations.
//
// Author:
//      Laura Werner
//
// COPYRIGHT:
//      IBM Open Class Library
//      (C) Copyright International Business Machines Corporation 1997
//      Licensed Material - Program-Property of IBM - All Rights Reserved.
//      US Government Users Restricted Rights - Use, duplication, or disclosure
//      restricted by GSA ADP Schedule Contract with IBM Corp.
//
#include <ifilesys.hpp>
#include <iautoptr.hpp>

#include "fsdirrec.hpp"
#include "fsiterat.hpp"

#ifdef IC_PAGETUNE
    #define _FSDIRREC_CPP_
    #include <ipagetun.h>
#endif

//------------------------------------------------------------------------------
// Constructors, etc...
//
IFSDirectoryRecord::IFSDirectoryRecord()
{
}

IFSDirectoryRecord::IFSDirectoryRecord  (const IFSDirectoryRecord& )
{
        // Disallowed.
}

IFSDirectoryRecord& IFSDirectoryRecord::operator = (const IFSDirectoryRecord& )
{
        // Disallowed.
        return (*this);
}

IFSDirectoryRecord::~IFSDirectoryRecord ()
{

}


//------------------------------------------------------------------------------
// Type safety methods...
//
bool
IFSDirectoryRecord::isA (const IFileKind& kind) const
{
        return kind == IDirectory::kKind || IFSRecord::isA(kind);
}

IFileKind
IFSDirectoryRecord::kind        () const
{
        return IDirectory::kKind;
}


//------------------------------------------------------------------------------
// Deletion...
//
void
IFSDirectoryRecord::deleteSelf(bool deleteChildren)
{
        if (deleteChildren) {
                deleteAllContents();
        }
        IFSRecord::deleteSelf(deleteChildren);
}

void
IFSDirectoryRecord::deleteAllContents   ()
{
        unsigned long           failures = 0u;
        IFileSystemEntity       entity;

        // We're going to delete the directory's contents using the brute
        // force approach: iterate through the contents and delete each one.
        // Set up the iterator
        auto_ptr<IFSIterator> iter = createIterator();

        // First just try to delete everything in one fell swoop.
        try {
                // Always ask for the "first" entity.  The idea here is to avoid requiring
                // that all file system implementations correctly implement deletion
                // during iteration.
                for (entity = iter->first(); entity.valid(); entity= iter->first())
                {
                        entity.deleteSelf(true);
                }
        }
        catch (...) {
                // There was an error deleting one of the entities in this directory.
                // Continue deleting as much as we can, then re-throw the exception.
                // First, increment the failure count and go to the next entity
                failures++;
                entity = iter->next();

                while (entity.valid()) {
                        try {
                                entity.deleteSelf(true);

                                // Now go on to the 'next' entry.
                                // Logically, we want to just keep calling first()
                                // until we've deleted everything -- unfortunately, this doesn't work
                                // unless all entries delete successfully.  We simulate this here by
                                // skipping forward from the first() past the entries that did not
                                // delete...
                                entity = iter->first();
                                for (unsigned long i = 0; i < failures && entity.valid(); i++) {
                                        entity = iter->next();
                                }
                        }
                        catch (...) {
                                // No need for hop-skip-jump games here since the entity
                                // didn't get deleted.  Just go to the next one.
                                failures++;
                                entity = iter->next();
                        }
                }
#ifndef SKIPOMPARSE
                throw;
#endif
        }
}


//------------------------------------------------------------------------------
// Manipulation...
//
IFileSystemEntity
IFSDirectoryRecord::copyTo(IDirectory &to, const IFileName &newName) const
{
        // Return an invalid entity, signaling that it's not possible to
        // copy a whole directory tree in one host filesystem operation.
        // Subclasses can override this if it actually *is* possible.
        return IFileSystemEntity();
}


IFileSystemEntity
IFSDirectoryRecord::moveTo(IDirectory &to, const IFileName &newName)
{
        // Return an invalid entity, signaling that it's not possible to
        // move a whole directory tree in one host filesystem operation.
        // Subclasses can override this if it actually *is* possible.
        return IFileSystemEntity();
}


//------------------------------------------------------------------------------
// Iteration...
//
IFSIterator *
IFSDirectoryRecord::createRecursiveIterator() const
{
        // Return a default recursive iterator that just maintains a stack
        // of single-directory iterators.  On systems where there's a more
        // efficient way to do this, you can override this method.
        return new IFSRecursiveIterator( this );
}

