// Revision: 44 1.8.1.4 source/core/filesys/fsiterat.cpp, filesystem, ioc.v400, 000302 
// fsiterat.cpp
//
// Contains:
//      IFSIterator                     -       Abstract base for single-directory iteration
//      IFSRecursiveIterator-   Concrete class for recursive directory iteration.
//
// Author:
//      Laura Werner
//
// COPYRIGHT:
//      IBM Open Class Library
//      (C) Copyright International Business Machines Corporation 1997
//      Licensed Material - Program-Property of IBM - All Rights Reserved.
//      US Government Users Restricted Rights - Use, duplication, or disclosure
//      restricted by GSA ADP Schedule Contract with IBM Corp.
//
#include <ifilesys.hpp>
#include <ifileitr.hpp>

#include "fsiterat.hpp"
#include "fsdirrec.hpp"


#ifdef IC_PAGETUNE
    #define _FSITERAT_CPP_
    #include <ipagetun.h>
#endif

//===============================================================================
// IFSIterator implementation
//      There's not much here, because most functionality must be implemented
//      by subclasses.
//
IFSIterator::IFSIterator ()
{
}

IFSIterator::IFSIterator (const IFSIterator& copy)
{
        // Disallowed
}

IFSIterator&
IFSIterator::operator = (const IFSIterator& right)
{
        // Disallowed
        return *this;
}

IFSIterator::~IFSIterator ()
{
}


//-------------------------------------------------------------------------------
//===============================================================================
// IFSRecursiveIterator constructor.
//      Create a normal iterator for the top-level directory, plus a stack for
//      doing the recursion.
//
IFSRecursiveIterator::IFSRecursiveIterator(const IFSDirectoryRecord *dir)
        : fIterator( dir->createIterator() )
        , fStack( new IStack< ICountedPointerTo<IFSIterator> > )
        , fFirst( true )
{
}


IFSRecursiveIterator::~IFSRecursiveIterator()
{
        popAll();
        delete fStack;
}


//-------------------------------------------------------------------------------
// Iteration methods
//      The real work happens in getOne; first and next are just wrappers
//      that set up the initial conditions properly
//
IFileSystemEntity
IFSRecursiveIterator::first ()
{
        // If there are any nested iterations on my stack, pop them
        popAll();

        // fFirst indicates that we haven't yet called fIterator->first()
        fFirst = true;
        return getOne( );
}

IFileSystemEntity
IFSRecursiveIterator::next ()
{
        return getOne( );
}


IFileSystemEntity
IFSRecursiveIterator::getOne( )
{
        IFileSystemEntity result;

        do {
                // Get the next entity in the directory over which we're
                // currently iterating.
                result = fFirst ? fIterator->first() : fIterator->next();
                fFirst = false;

                if ( result.isA( IDirectory::kKind) )
                {
                        // When we encounter a sub directory, we have to recursively
                        // iterate over all of its contents.  To do this, push the
                        // current iterator and create a new one for the subdirectory.

                        fStack->push(fIterator);
                        fIterator = IDirectory(result).getRecord()->createIterator();
                        fFirst = true;
                }
                else if (!result.valid())
                {
                        // If there's nothing left in the directory over which we're iterating,
                        // see if we can pop back up a level and continue iterating over
                        // its parent.

                        if ( fStack->numberOfElements() )
                        {
                                fIterator = fStack->top();
                                fStack->pop();
                        }
                        else
                        {
                                // No more levels to pop; we're done!
                                break;
                        }
                }
                // Keep going until we find a valid entity (or until the break statement
                // above is executed.)
        } while (!result.valid());

        return result;
}




void                                            IFSRecursiveIterator::popAll( )
{
        while ( fStack->numberOfElements() ) {
                fIterator = fStack->top();
                fStack->pop();
        }
}



IFSRecursiveIterator::IFSRecursiveIterator(const IFSRecursiveIterator &copy)
{
        // Disallowed
}


IFSRecursiveIterator & IFSRecursiveIterator::operator = (const IFSRecursiveIterator &copy)
{
        // Disallowed
        return *this;
}



