// Revision: 38 1.18.1.3 source/core/filesys/fsrecord.cpp, filesystem, ioc.v400, 000302 
//NOSHIP
// fsrecord.cpp
//
// Contains:    IFSRecord
//                              Internal filesystem class that encapsulates generic
//                              file system entity operations.
//
// Author:
//      Laura Werner
//
// COPYRIGHT:
//      IBM Open Class Library
//      (C) Copyright International Business Machines Corporation 1997
//      Licensed Material - Program-Property of IBM - All Rights Reserved.
//      US Government Users Restricted Rights - Use, duplication, or disclosure
//      restricted by GSA ADP Schedule Contract with IBM Corp.
//
#include <ilanglvl.hpp>
#include <ifilesys.hpp>
#include <ipathnam.hpp>
#include <iexcfile.hpp>

#include <iseq2.h>
#include <igenexc.hpp>
#include <iautoptr.hpp>
#include <idatstrm.hpp>

#include "filesys.hpp"
#include "fsrecord.hpp"
#include "fsalias.hpp"

#ifdef IC_PAGETUNE
    #define _FSRECORD_CPP_
    #include <ipagetun.h>
#endif

//------------------------------------------------------------------------------
// IFSRecord constructors, etc....
//
IFSRecord::IFSRecord()
        : fIsTemporary(false)
{
}

IFSRecord::IFSRecord (const IFSRecord& copy)
        : fIsTemporary(copy.fIsTemporary)
{
}

IFSRecord::~IFSRecord ()
{
}

IFSRecord&
IFSRecord::operator = (const IFSRecord& )
{
      // Disallowed.
      return (*this);
}


//------------------------------------------------------------------------------
// Type safety methods...
//
//      Subclasses must override isA and kind to deal with their own type.  This
//      default implementation of isA is provided just so that all subclasses
//      can call the inherited version of isA without worring about whether their
//      base class actually implements it.
//
bool
IFSRecord::isA (const IFileKind& kind) const
{
        return (false);
}


//------------------------------------------------------------------------------
// Property access...
//
bool
IFSRecord::temporary () const
{
        return fIsTemporary;
}

void
IFSRecord::setTemporary(bool isTemporary)
{
        fIsTemporary = isTemporary;
}

IPathName
IFSRecord::pathName() const
{
        return IPathName(hostPath());
}


//------------------------------------------------------------------------------
// Entity manipulation methods...
//
void
IFSRecord::deleteSelf(bool /*deleteChildren*/ )
{
        IFSRecordLock lock(this);
        doDeleteSelf();
}


//------------------------------------------------------------------------------
// Streaming methods...
//
void
IFSRecord::streamOut(IDataStream & stream)
{
        // To stream out a file system entity, we construct an "alias" for it
        // and then stream that out monomorphically.  In the future, we'll stream
        // several aliases, each representing a different way in which the physical
        // entity could be accessed (locally, via NFS, via LAN Manager, etc.)

        IStreamOutFrame frame(stream);
        IFSRecordLock lock(this);

        (unsigned long)(1) >>= stream;          // Count
        getLocalAlias() >>= stream;
}

IFileSystemEntity
IFSRecord::streamIn(IDataStream & stream)
{
        IStreamInFrame frame(stream);
        IFileSystemEntity entity;

        // Get the number of aliases to read
        unsigned long count;
        count <<= stream;

        // Read them in and try to resolve each one until we get one that works
        // Then read in the remaining ones in order to keep the stream in sync

        for (INumber i = 1; i <= count; i++) {
                IFSAlias alias;
                alias <<= stream;

                if (!entity.valid()) {
                        entity = IFSFileSystem::instance()->resolveAlias(alias);
                }
        }

        if (!entity.valid()) {
                IObjectNotFound e;
                ITHROW(e);
        }
        return entity;
}

IFSAlias
IFSRecord::getLocalAlias( ) const
{
        // The default local alias is just the filesystem type and host path name

#if IC_RTTI_ENABLED
        return IFSAlias(typeid(*IFSFileSystem::instance()).name(), "", hostPath());
#else // IC_RTTI_ENABLED
        return IFSAlias((*IFSFileSystem::instance()).typeOf(), "", hostPath());
#endif // IC_RTTI_ENABLED
}


//--------------------------------------------------------------------------------
// IFSRecordLock implementation
//      Constructing one of these objects will lock the specified record.
//      When the lock object goes out of scope, the record will be unlocked.
//
IFSRecord::
IFSRecordLock::IFSRecordLock(const IFSRecord* record)
      : fLock((IPrivateResource&)record->fLock)
{
}

IFSRecord::
IFSRecordLock::~IFSRecordLock()
{
}

IFSRecord::
IFSRecordLock::IFSRecordLock(const IFSRecordLock&)
      : fLock( *((IResource*)0) ) // Disgusting, but will never be called
{
      // Disallowed
}

void
IFSRecord::
IFSRecordLock::operator=(const IFSRecordLock&)
{
      // Disallowed
}


