/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 46 1.15.1.3 source/core/filesys/fstester.cpp, filesystem, ioc.v400, 000302 
// fstester.cpp

#include <stdarg.h>
#include <string.h>
#include <stdio.h>

#include <idatstrm.hpp>
#include <iautoptr.hpp>
#include <igenexc.hpp>

#include "ifilesys.hpp"
#include "iexcfile.hpp"
#include "ipathnam.hpp"
#include "ifilestr.hpp"
#include "ifileitr.hpp"

#ifdef IC_PAGETUNE
    #define _FSTESTER_CPP_
    #include <ipagetun.h>
#endif

#define TESTASSERT(x) do { if (!(x)) { printf("\nERROR: '%s' failed\n", #x ); gErrors++; } } while(0)

int gErrors = 0;

void currentDir()
{
        printf("\n---------- Current Directory Tests\n");
        //------------------------ Current Directories -------------------------
        IDirectory currentDir = IDirectory::current();
        printf("Current directory is %s\n", (const char*)currentDir.name() );

        IDirectory::setCurrent(currentDir);
        TESTASSERT(   currentDir == IDirectory::current() );
        TESTASSERT( !(currentDir != IDirectory::current()) );
}

void volumeTests()
{
        printf("\n---------- Volume Tests\n");
        //------------------------ Volume Info -------------------------
        IVolume currentVol = IDirectory::current().volume();
        printf("\ncurrent volume is %s\n", (const char*)currentVol.name() );
        printf("  total space on volume is %10llu\n", currentVol.totalSpace() );
        printf("   free space on volume is %10llu\n", currentVol.freeSpace() );

        IVolume sameVol( currentVol.name() );

        printf("sameVol=%s, currentVol=%s\n", (const char*)sameVol.hostPath(), (const char*)currentVol.hostPath());

        TESTASSERT(   sameVol == currentVol );
        TESTASSERT( !(sameVol != currentVol) );

        //------------------------ Volume Iteration -------------------------
        printf("\nIterating over local volumes (STL syntax):\n");
        for (IVolumeIterator iter; iter != IVolumeIterator::end(); iter++) {
                IVolume vol = *iter;
                if (!vol.remote()) {
                        printf("    %-10s %10llu %10llu\n", (const char*)vol.name(),
                                                                vol.totalSpace(), vol.freeSpace());
                }
        }
        //------------------------ Root Iteration -------------------------
        printf("\nIterating over root directories (simpler syntax):\n");
        for (IRootDirectoryIterator dir; dir; dir++) {
                printf("    %-0s\n", (const char*) dir->name() );
        }
}

void iterateTests(IDirectory parent)
{
        // Create some stuff to iterate over
        IDirectory directory = parent.createDirectory("Iterate");
        directory.createFile("File1");
        directory.createDirectory("SubDir1").createFile("File2");
        directory.createFile("Wooooo");

        printf("\n---------- Directory Iteration Tests\n");

        //------------------------ Directory Iteration -------------------------
        {
                printf("Iterating over current directory...\n");
                for (IDirectoryIterator dirIter = directory; dirIter; dirIter++) {
                        printf("    %-10s\n", (const char*)(*dirIter).hostPath());
                }
        }
        //------------------------ Recursive Iteration -------------------------
        {
                printf("\nRecursing over current directory...\n");
                for (IDirectoryIterator anEntity(directory, true); anEntity; anEntity++) {
                        printf("    %-10s\n", (const char*)anEntity->hostPath());
                }
        }

        directory.deleteSelf(true);
}

void fileCreateDelete(IDirectory& parent)
{
        printf("\n---------- File Creation / Deletion Tests\n");

        // Create a file and make sure I can look it up

        const char *kTestFile = "TestFile";
        const IFileSize kFileSize = 200u;

        IFile testFile = parent.createFile(kTestFile, kFileSize);
        IFile file1 = parent.lookUp(kTestFile);
        IFile file2( testFile.hostPath() );

        TESTASSERT( testFile == file1 );
        TESTASSERT( file2 == file2 );
        TESTASSERT( file1.size() == kFileSize );

        printf("Deleting file....\n");
        testFile.deleteSelf();
        try {
                printf("Trying lookup to find deleted file... ");
                file1 = parent.lookUp(kTestFile);
                printf("ERROR: no exception!\n", kTestFile);
                gErrors++;
        }
        catch (const IObjectNotFound &) {
                printf("  Caught expected exception\n");
        }

        printf("Validity tests...\n");
        TESTASSERT( !testFile.valid() );
        TESTASSERT( !file1.connected() );
}

void dirCreateDelete(IDirectory& parent)
{
        printf("\n---------- Directory Creation / Deletion Tests\n");

        // Create a directory and make sure I can look it up
        const char *kTestDir = "TestDir";
        IDirectory testDir = parent.createDirectory(kTestDir);
        IDirectory dir1 = parent.lookUp(kTestDir);
        IDirectory dir2(dir1.hostPath());

        TESTASSERT( testDir == dir1 );
        TESTASSERT( dir1 == dir2 );

        // Create a file in the directory
        const char *kTestFile = "TestFile";
        IFile testFile = testDir.createFile(kTestFile);

        // Try to delete the directory.  Should fail.
        try {
                printf("Delete of non-empty directory... ");
                testDir.deleteSelf();
                printf("\nERROR: no exception!\n");
                gErrors++;
        }
        catch ( const IMustBeEmpty & ) {
                printf("  Caught expected exception\n");
        }

        // Try again, recursively this time.  Should work.
        testDir.deleteSelf(true);

        // The file and directory shouldn't be there now
        // But the file object will still be valid, since it doesn't know
        // that it's been deleted
        printf("Testing file validity...\n");
        TESTASSERT( testFile.valid() );
        TESTASSERT( !testFile.connected() );

        printf("Testing directory validity...\n");
        TESTASSERT( !testDir.valid() );
        TESTASSERT( !testDir.connected() );
}

void tempTests()
{
        const IFileSize kSize = 345u;
        const IFileName kFileName = "FooBar";
        const IFileName kDirName = "BarFoo";

        printf("\n---------- Temp File Tests\n");

        IFile temp1 = IFile::createTemporary();
        IFile temp2 = IFile::createTemporary(IFile::uniqueName(), kSize);
        IFile temp3 = IFile::createTemporary(kFileName);

        printf("Created files '%s', '%s', '%s'...\n", (const char*)temp1.name(),
                                (const char*)temp2.name(), (const char*)temp3.name());

        TESTASSERT( temp1.parent() == temp2.parent() );
        TESTASSERT( temp2.size() == kSize );
        TESTASSERT( temp3.name() == kFileName );

        IDirectory dir1 = IDirectory::createTemporary();
        IDirectory dir2 = IDirectory::createTemporary(kDirName);

        printf("Created directories '%s', '%s'...\n", (const char*)dir1.name(), (const char*)dir2.name());

        TESTASSERT( dir1.parent() == dir2.parent() );
        TESTASSERT( dir1.parent() == temp1.parent() );
        TESTASSERT( dir2.name() == kDirName );

        printf("Trying to create duplicate temp file... ");
        try {
                IFile duplicate = IFile::createTemporary(kFileName);
                printf("\n  ERROR: created duplicate temp file %s.\n", kFileName);
                gErrors++;
        }
        catch (const IException &) {
                printf("  Caught expected exception\n");
        }

        printf("  Deleting temp files and directories...\n");
        temp1.deleteSelf();
        temp2.deleteSelf();
        temp3.deleteSelf();
        dir1.deleteSelf();
        dir2.deleteSelf();
}

IVolume otherVolume(const IDirectory &testDir)
{
        static IVolume otherVol;
#ifndef _AIX
        if (!otherVol.valid()) {
                for (IVolumeIterator iter; iter != IVolumeIterator::end(); iter++) {
                        if ( testDir.volume() != *iter ) {
                                otherVol = *iter;
                                break;
                        }
                }
        }
#else
        if(!otherVol.valid())
          {
            otherVol = IVolume(IHostPathName("/work"));
          }
#endif
        return otherVol;
}

void moveTest(IFileSystemEntity testFile, IDirectory testDir)
{
        IFileName origName = testFile.name();

        const IFileName kMoveTwo = "MoveTwo";
        const IFileName kSubDir = "SubDir";
        const IFileName kOtherDir = "MoveTestDir";

        // Try moving a file
        printf("Testing move of %s...\n", (const char*)testFile.hostPath());

        testFile.moveTo(testDir, kMoveTwo);
        printf("  After simple rename: %s\n", (const char*)testFile.hostPath());

        TESTASSERT( testFile.name() == kMoveTwo );
        TESTASSERT( testFile.parent() == testDir );
        TESTASSERT( testFile == testDir.lookUp(kMoveTwo) );

        IDirectory subDir = testDir.createDirectory(kSubDir);
        testFile.moveTo(subDir);
        printf("  After move into child: %s\n", (const char*)testFile.hostPath());

        TESTASSERT( testFile.name() == kMoveTwo );
        TESTASSERT( testFile.parent() == subDir );
        TESTASSERT( testFile == subDir.lookUp(kMoveTwo) );

        // Find another volume so I can try a cross-volume move
        IVolume otherVol = otherVolume(testDir);
        IDirectory otherDir = otherVol.createDirectory(kOtherDir);

        testFile.moveTo(otherDir, origName);
        printf("  Moved to other volume: %s\n", (const char*)testFile.hostPath());

        TESTASSERT( testFile.name() == origName );
        TESTASSERT( testFile.parent() == otherDir );
        TESTASSERT( testFile == otherDir.lookUp(origName) );
        TESTASSERT( testFile.volume() == otherVol );

        testFile.deleteSelf(true);
        subDir.deleteSelf();
        otherDir.deleteSelf();
}

void copyTest(IFileSystemEntity testFile, IDirectory testDir)
{
        IFileName origName = testFile.name();
        IFileSystemEntity result;

        const IFileName kCopyTwo = "CopyTwo";
        const IFileName kSubDir = "SubDir";
        const IFileName kOtherDir = "CopyTestDir";

        // Try copying it to another name in the same directory
        printf("Testing copy of %s...\n", (const char*)testFile.hostPath());

        result = testFile.copyTo(testDir, kCopyTwo);
        printf("  After simple copy: %s\n", (const char*)result.hostPath());

        TESTASSERT( result != testFile );
        TESTASSERT( result.name() == kCopyTwo );
        TESTASSERT( result.parent() == testDir );
        TESTASSERT( result == testDir.lookUp(kCopyTwo) );
        result.deleteSelf(true);


        IDirectory subDir = testDir.createDirectory(kSubDir);
        result = testFile.copyTo(subDir);
        printf("  After copy into child: %s\n", (const char*)result.hostPath());

        TESTASSERT( result.name() == origName );
        TESTASSERT( result.parent() == subDir );
        TESTASSERT( result == subDir.lookUp(origName) );
        TESTASSERT( result != testFile );
        result.deleteSelf(true);

        // Find another volume so I can try a cross-volume copy
        IVolume otherVol = otherVolume(testDir);
        IDirectory otherDir = otherVol.createDirectory(kOtherDir);

        result = testFile.copyTo(otherDir, origName);
        printf("  Copied to other volume: %s\n", (const char*)result.hostPath());

        TESTASSERT( result.name() == origName );
        TESTASSERT( result.parent() == otherDir );
        TESTASSERT( result == otherDir.lookUp(origName) );
        TESTASSERT( result.volume() == otherVol );
        TESTASSERT( result != testFile );
        result.deleteSelf(true);

        testFile.deleteSelf(true);
        subDir.deleteSelf();
        otherDir.deleteSelf();
}

void moveCopyTest(IDirectory testDir)
{
        {
                printf("\n---------- File/Directory Move Tests\n");

                const char *kMoveFile = "MoveFile";
                const char *kMoveDir = "MoveDir";

                // Move of simple file and directory
                moveTest( testDir.createFile(kMoveFile),                testDir );
                moveTest( testDir.createDirectory(kMoveDir),    testDir );

                // Move of directory with children
                IDirectory withChild = testDir.createDirectory(kMoveDir);
                withChild.createDirectory(kMoveDir).createFile(kMoveFile);
                moveTest(withChild, testDir);
        }
        {
                printf("\n---------- File/Directory Copy Tests\n");

                const char *kCopyFile = "CopyFile";
                const char *kCopyDir = "CopyDir";

                // Copy of simple file and directory
                copyTest( testDir.createFile(kCopyFile),                testDir );
                copyTest( testDir.createDirectory(kCopyDir),    testDir );

                // Copy of directory with children
                IDirectory withChild = testDir.createDirectory(kCopyDir);
                withChild.createDirectory(kCopyDir).createFile(kCopyFile);
                copyTest(withChild, testDir);
        }
}

void stdStreamTest(IDirectory dir)
{
        const char *kStreamFile = "StreamFile";
        const int kBigNum = 8192;                               // Bigger than default buffer size

        IFile testFile = dir.createFile(kStreamFile);

        IFileIOStream testStream(testFile);

        testStream << "Hello there" << endl;
        testStream << "Another line of the file" << endl;

        for (int i = 0; i < kBigNum; i++) {
                testStream << '@';
        }

        testStream.close();

        testFile.deleteSelf();
}

void doAutoCreateTest(IDirectory testDir)
{
        // Starting with a relative path for the thing we want to create,
#ifndef _AIX
        IWin32PathNameParser kParser;
        const IPathName kPartialPath("autotest\\foo\\bar\\baz", kParser);
#else
        IUnixPathNameParser kParser;
        const IPathName kPartialPath("autotest/foo/bar/baz", kParser);
#endif
        IHostPathName partialHost = kParser.formatName(kPartialPath);

        IPathName kFullPath;
        IHostPathName fullHost;

        if (testDir.valid()) {
                kFullPath = IPathName(testDir, kPartialPath);
        } else {
                testDir = IDirectory::current();
                kFullPath = kPartialPath;
        }
        fullHost = kParser.formatName(kFullPath);

#if 0
        for (int i = 1; i <= kFullPath.numberOfComponents(); i++) {
                printf("kFullPath[%d] = %s\n", i, (const char*)kFullPath.componentAt(i));
        }
        printf("testDir = %s\n", (const char*)testDir.hostPath());
#endif

        IFile testFile;
        IDirectory subDir;

        // Make sure kMustExist doesn't auto-create
        printf("Trying kMustExist on nonexistent file...");
        try {
                testFile = IFile( fullHost, IFile::kMustExist, kParser);
                printf("\n  ERROR: kMustExist worked on nonexistent file.\n");
                gErrors++;
        }
        catch ( const IObjectNotFound &e ) {
                printf("  Caught expected %s\n", e.name());
        }

        // Test kCreateIfNeeded on a single-component partial path
        if ( testDir == IDirectory::current() ) {
                printf("Trying kCreateIfNeeded on single component '%s'...\n",
                                                                        (const char*)kPartialPath.componentAt(1));
                subDir = IDirectory( kPartialPath.componentAt(1), IDirectory::kCreateIfNeeded, kParser);
                TESTASSERT( subDir == testDir.lookUp(kPartialPath.componentAt(1), kParser) );
        }

        // Test kCreateIfNeeded on a multi-component path
        printf("Trying kCreateIfNeeded on nonexistent file %s...\n", (const char*)fullHost);
        testFile = IFile(fullHost, IFile::kCreateIfNeeded, kParser);
        printf("testFile.hostPath() = %s\n", (const char*)testFile.hostPath());

        printf("Checking against partial lookup of %s from %s...\n",
                        (const char*)partialHost, (const char*)testDir.hostPath());
        TESTASSERT( testFile == testDir.lookUp(partialHost, kParser) );

        printf("Trying kCreateIfNeeded on existing file...\n");
        testFile = IFile(fullHost, IFile::kCreateIfNeeded, kParser);
        TESTASSERT( testFile == testDir.lookUp(partialHost, kParser) );

        // Look it up again using kCreateOnly: should fail
        printf("Trying kCreateOnly on existing file...");
        try {
                IFile foo( fullHost, IFile::kCreateOnly, kParser);
                printf("\n  ERROR: kCreateOnly worked on existing file.\n");
                gErrors++;
        }
        catch ( const IAlreadyExists &e ) {
                printf("  Caught expected %s\n", e.name());
        }

        // Now delete the file and try kCreateOnly. Should work
        printf("Trying kCreateOnly on nonexistent file...\n");
        testFile.deleteSelf();
        testFile = IFile(fullHost, IFile::kCreateOnly, kParser);
        TESTASSERT( testFile == testDir.lookUp(partialHost, kParser) );

        // Now try kReplaceExisting....
        const IFileSize kSize = 200u;
        printf("Trying kReplaceExisting...\n");
        testFile.setSize(kSize);
        TESTASSERT( testFile.size() == kSize );

        testFile = IFile(fullHost, IFile::kReplaceExisting, kParser);
        TESTASSERT( testFile.size() == 0u );

        // Now try using IDirectory::createFile and createDirectory
        printf("Testing IDirectory::createFile and createDirectory...\n");
#ifndef _AIX
        const IPathName path1("autotest\\idir\\bar\\file", kParser);
        const IPathName path2("autotest\\idir\\file", kParser);
#else
        const IPathName path1("autotest/idir/bar/file", kParser);
        const IPathName path2("autotest/idir/file", kParser);
#endif

        IHostPathName host1 = kParser.formatName(path1);
        IHostPathName host2 = kParser.formatName(path2);

        IFile file1 = testDir.createFile(host1);
        IDirectory dir2 = testDir.createDirectory(host2);

        printf("Cleaning up after auto-create test...\n");
        IDirectory deleteMe = testDir.lookUp( kPartialPath.componentAt(1), kParser );
        deleteMe.deleteSelf(true);
}

void autoCreateTest(IDirectory testDir)
{
        printf("\n---------- Testing auto-creation of full paths....\n");
        doAutoCreateTest(testDir);

        printf("\n---------- Testing auto-creation with partial paths....\n");
        IDirectory saveDir = IDirectory::current();
        IDirectory::setCurrent(testDir);
        doAutoCreateTest( IDirectory() );
        IDirectory::setCurrent(saveDir);
}

void entityStreamTest(IDirectory testDir)
{
        printf("\n---------- Testing IFile streaming ....\n");
        const char *kStreamFile = "StreamFile";

        // Try to stream an IFile out and then back in.
        IFile streamFile(kStreamFile, IFile::kCreateIfNeeded);
        auto_ptr<IDataStream> fileStream = streamFile.createStream();

        streamFile >>= *fileStream;

        printf("Trying simple stream-in...\n");
        IFile sameFile;
        fileStream->reset();
        sameFile <<= *fileStream;
        TESTASSERT( streamFile == sameFile );

        // Try streaming into an IDirectory.  Should fail
        printf("Trying mismatched types...");
        try {
                IDirectory foo;
                fileStream->reset();
                foo <<= *fileStream;
                printf("  ERROR: streamed IFile into IDirectory.\n");
                gErrors++;
        }
        catch ( const IEntityTypeMismatch &e ) {
                printf("  Caught expected %s\n", e.name());
        }


        printf("\n---------- Testing IVolume streaming ....\n");
        IVolume testVol = IDirectory::current().volume();
        fileStream->reset();
        testVol >>= *fileStream;

        printf("Trying simple stream-in...\n");
        IVolume sameVol;
        fileStream->reset();
        sameVol <<= *fileStream;
        TESTASSERT( sameVol == testVol );

        printf("Trying stream-in as IDirectory...\n");
        IDirectory sameDir;
        fileStream->reset();
        sameDir <<= *fileStream;
        TESTASSERT( sameDir == testVol );


        printf("Trying stream-in as IFile...");
        try {
                IFile foo;
                fileStream->reset();
                foo <<= *fileStream;
                printf("  ERROR: streamed IVolume into IFile.\n");
                gErrors++;
        }
        catch ( const IEntityTypeMismatch &e ) {
                printf("  Caught expected %s\n", e.name());
        }

        delete fileStream.release();
        streamFile.deleteSelf();
}


static void printPath(const IPathName &path)
{
        unsigned long count = path.numberOfComponents();

        for (unsigned long i = 1; i <= count; i++) {
                if (i > 1)
                        printf(", ");
                printf("%s", (const char*)path.componentAt(i));
        }
        printf("\n");
}

void pathStreamTest(IDirectory testDir)
{
        printf("\n---------- Testing IPathName streaming ....\n");

        IPathName testPath = testDir.path();

        IFile streamFile("strmfile", IFile::kCreateIfNeeded);

        auto_ptr<IDataStream> stream = IDataStream::createMemoryStream();

        testPath >>= *stream;

        printf("Testing simple stream-in...\n");
        IPathName samePath;
        stream->reset();
        samePath <<= *stream;
        TESTASSERT( samePath == testPath );

        printf("Get entity from streamed-in path...\n");
        IDirectory sameDir = IDirectory(samePath);
        TESTASSERT( sameDir == testDir );

        printf("Construct another path from entity...\n");
        IPathName otherPath = sameDir.path();
        TESTASSERT( otherPath == testPath );
}

void pathNameTests()
{
#ifndef _AIX
        IWin32PathNameParser parser;
#else
        IUnixPathNameParser parser;
#endif
        const int kTestCount = 10;
        printf("Begin PathName Tests ....\n");
        {
                IFileName hostPath[kTestCount];
                IFileName expected[kTestCount];
                bool    isHostPathAbsolute[kTestCount];

#ifndef _AIX
                hostPath[0] = "C:\\Foo\\Bar\\Test1.cpp";
                expected[0] = "C:\\Foo\\Bar\\Test1.cpp";
#else
                hostPath[0] = "/Foo/Bar/Test1.cpp";
                expected[0] = "/Foo/Bar/Test1.cpp";
#endif
                isHostPathAbsolute[0] = true;

#ifndef _AIX
                hostPath[1] = "Data\\Fonts";
                expected[1] = "Data\\Fonts";
#else
                hostPath[1] = "Data/Fonts";
                expected[1] = "Data/Fonts";
#endif
                isHostPathAbsolute[1] = false;

#ifndef _AIX
                hostPath[2] = "\\Data\\Fonts";
                expected[2] = "\\Data\\Fonts";
#else
                hostPath[2] = "/Data/Fonts";
                expected[2] = "/Data/Fonts";
#endif
                isHostPathAbsolute[2] = true;

#ifndef _AIX
                hostPath[3] = ".\\Data\\Fonts";
                expected[3] = "Data\\Fonts";
#else
                hostPath[3] = "./Data/Fonts";
                expected[3] = "Data/Fonts";
#endif
                isHostPathAbsolute[3] = false;

#ifndef _AIX
                hostPath[4] = "C:\\Data\\Fonts\\";
                expected[4] = "C:\\Data\\Fonts";
#else
                hostPath[4] = "/Data/Fonts/";
                expected[4] = "/Data/Fonts";
#endif
                isHostPathAbsolute[4] = true;

#ifndef _AIX
                hostPath[5] = "..\\Data\\Fonts";
                expected[5] = "..\\Data\\Fonts";
#else
                hostPath[5] = "../Data/Fonts";
                expected[5] = "../Data/Fonts";
#endif
                isHostPathAbsolute[5] = false;

#ifndef _AIX
                hostPath[6] = "\\\\twinturbo\\usefulstuff\\games";
                expected[6] = "\\\\twinturbo\\usefulstuff\\games";
#else
                hostPath[6] = "/nfsmounts/Builds";
                expected[6] = "/nfsmounts/Builds";
#endif
                isHostPathAbsolute[6] = true;

#ifndef _AIX
                hostPath[7] = "\\\\twinturbo\\TmpRepository";
                expected[7] = "\\\\twinturbo\\TmpRepository";
#else
                hostPath[7] = "/nfsmounts/Releases";
                expected[7] = "/nfsmounts/Releases";
#endif
                isHostPathAbsolute[7] = true;

#ifndef _AIX
                hostPath[8] = ".C";
                expected[8] = ".C";
#else
                hostPath[8] = ".C";
                expected[8] = ".C";
#endif
                isHostPathAbsolute[8] = false;

#ifndef _AIX
                hostPath[9] = "C";
                expected[9] = "C";
#else
                hostPath[9] = "C";
                expected[9] = "C";
#endif
                isHostPathAbsolute[9] = false;

                for (int count = 0; count < kTestCount; count++)
                {
                        IPathName pathName(hostPath[count], parser);
                        TESTASSERT( (pathName.isFullPath() == isHostPathAbsolute[count]));
                        IFileName fromPathName = parser.formatName(pathName);
                        TESTASSERT (expected[count] == fromPathName);
//                      printf("pathName = %s, expected = %s fromPathName = %s \n",hostPath[count],
//                                                                                              expected[count],fromPathName);

                }

        }

#ifndef _AIX
        IUnixPathNameParser OtherParser;
#else
        IWin32PathNameParser OtherParser;
#endif
        {
#ifndef _AIX
                IFileName hostPath("\\Data\\Fonts");
                IFileName expectedPathName("/Data/Fonts");
#else
                IFileName hostPath("/Data/Fonts");
                IFileName expectedPathName("\\Data\\Fonts");
#endif

                IPathName pathName(hostPath, parser);
                IFileName fromPathName = OtherParser.formatName(pathName);
                TESTASSERT(fromPathName == expectedPathName);

        }
        printf("End PathName Tests.....\n");
}


main()
{
        try {
                printf("\n---------- Creating a temp directory for tests....\n");
                IDirectory testDir = IDirectory::createTemporary();
                printf("created '%s'\n", (const char*) testDir.hostPath());

                tempTests();
                currentDir();
                volumeTests();
                iterateTests(testDir);
                fileCreateDelete(testDir);
                dirCreateDelete(testDir);
                moveCopyTest(testDir);
                stdStreamTest(testDir);
                autoCreateTest(testDir);
                pathNameTests();
                entityStreamTest(testDir);
                pathStreamTest(testDir);

                printf("\n---------- Cleaning up ....\n");

                testDir.deleteSelf();
        }
        catch (const IException &e) {
                printf("\nERROR: Caught unexpected %s(%lu).\n     %s\n",
                                                e.name(), e.errorId(), e.text() );

                for (int i = 0; i < e.locationCount(); i++) {
                        const IExceptionLocation *loc = e.locationAtIndex(i);
                        printf("%-50s; %s, %lu\n", loc->functionName(),
                                                loc->fileName(), loc->lineNumber());
                }

                gErrors++;
        }
        catch (...) {
                printf("\nERROR: Caught unexpected unknown exception.\n");
                gErrors++;
        }

        printf("\n\n------------------------------------------------\n");
        printf("\nTests %s; %d errors\n", gErrors ? "FAILED" : "PASSED", gErrors);
        return gErrors;
}
