/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 83 1.9.1.3 source/core/filesys/icmpdnam.cpp, filesystem, ioc.v400, 000302 

#include <ibase.hpp>
#include <ilanglvl.hpp>
#if IC_RTTI_ENABLED
#include <typeinfo.h>
#endif // IC_RTTI_ENABLED

#include <iseq2.h>
#include <iexcept.hpp>
#include <imrefcnt.hpp>
#include <idatstrm.hpp>

#include "icmpdnam.hpp"

#ifdef IC_PAGETUNE
    #define _ICMPDNAM_CPP_
    #include <ipagetun.h>
#endif

//--------------------------------------------------------------------------
// ICompoundNameImp
//      This is basically just a ref-counted sequence of file names,
//      plus an fIsFull data member.
//

template <class T>
long ComparisonFunction( T const& a, T const& b )
{
        return (a == b) ? 0 : 1;
}

class ICompoundNameImp : public ISequence<IText>, public IMRefCounted {
public:
                        ICompoundNameImp(INumber reserve) : ISequence<IText>(reserve),
                                                                                        fIsFull(false) {}

        bool    operator==(const ICompoundNameImp& other) const
                                { return fIsFull == other.fIsFull &&
                                        this->compare(other, ComparisonFunction) == 0; }

        bool    fIsFull;
};


//--------------------------------------------------------------------------
// ICompoundName constructors, etc.
//
#include <istrmmod.hpp> // fix
IStreamModule gCompoundNameModule("ICompoundName");
StreamableDefinitionsMacro(ICompoundName, gCompoundNameModule);

ICompoundName::ICompoundName ()
        : fPath( new ICompoundNameImp(kInitialSize) )
{
        fPath->addRef();
}

ICompoundName::ICompoundName(const IText&               fullPath,
                                                        const ICompoundNameParser&      aParser)
        : fPath( 0 )
{
        operator=( aParser.parseName(fullPath) );
}

ICompoundName::ICompoundName (const ICompoundName& copy)
        : fPath( copy.fPath )
{
        fPath->addRef();
}

ICompoundName&
ICompoundName::operator= (const ICompoundName& copyMe)
{
        // This is the ONLY case where fPath is allowed to be 0, in order
        // to let my constructors call operator= when they need to
        if (fPath) fPath->removeRef();

        fPath = copyMe.fPath;
        fPath->addRef();
        return *this;
}

ICompoundName::~ICompoundName ()
{
        fPath->removeRef();
}

//--------------------------------------------------------------------------
// Other ICompoundName boilerplate...
//

void
ICompoundName::makeWritable()
{
        if (fPath->count() > 1) {
                ICompoundNameImp* newImp = new ICompoundNameImp(*fPath);
                fPath->removeRef();
                fPath = newImp;
                fPath->addRef();
        }
}

long
ICompoundName::hash () const
{
        long hashVal = 0;

        long count = fPath->count();
        for (long i = 1; i <= count; i++) {
                hashVal += fPath->elementAtPosition(i).hash();
        }
        return hashVal;
}

bool
ICompoundName::operator== (const ICompoundName& other) const
{
        return fPath == other.fPath || *(fPath) == *(other.fPath);
}

bool
ICompoundName::operator!= (const ICompoundName& other) const
{
        return fPath != other.fPath && !( *(fPath) == *(other.fPath) );
}

void
ICompoundName::readFromStream(IDataStream& stream)
{
        makeWritable();

        IStreamInFrame frame(stream);
        short isFull;
        isFull <<= stream;
        fPath->fIsFull = isFull;

        INumber count;
        count <<= stream;

        fPath->removeAll();
        IText element;
        for (INumber i = 1; i <= count; i++) {
                element <<= stream;
                fPath->addAsLast(element);
        }
}


void
ICompoundName::writeToStream(IDataStream& stream) const
{
        IStreamOutFrame frame(stream);

        (short)fPath->fIsFull >>= stream;

        INumber count = fPath->numberOfElements();
        count >>= stream;

        for (INumber i = 1; i <= count; i++) {
                fPath->elementAtPosition(i) >>= stream;
        }
}

//--------------------------------------------------------------------------
// The real meat of ICompoundName....
//

unsigned long
ICompoundName::numberOfComponents () const
{
        return fPath->numberOfElements();
}


IText
ICompoundName::componentAt (unsigned long index) const
{
        IASSERTPARM( index != 0 )
        IASSERTPARM( index <= fPath->numberOfElements() )

        return fPath->elementAtPosition(index);
}

IText                   ICompoundName::lastComponent() const
{
        return fPath->lastElement();
}


void
ICompoundName::factorCommonRoot (const ICompoundName&   anotherPath,
                                                                   ICompoundName&       commonRoot,
                                                                   ICompoundName&       rootToLeaf) const
{
        INumber                                 i;
        INumber                                 commonCount     = 0;

        INumber                                 thisLength              = fPath->numberOfElements();
        ICompoundNameImp*                       otherPath               = anotherPath.fPath;
        INumber                                 thatLength              = otherPath->numberOfElements();

        for (i = 1; i <= thisLength && i <= thatLength; i++) {
                if (otherPath->elementAtPosition(i) == fPath->elementAtPosition(i) )
                        commonCount++;
                else
                        break;
        }

        INumber newCount = thisLength - commonCount;

        rootToLeaf.fPath->removeAll();
        rootToLeaf.setFullPath(false);
        for (i = 1; i <= newCount; i++)
        {
                rootToLeaf.appendComponent( fPath->elementAtPosition(i + commonCount) );
        }

        commonRoot.fPath->removeAll();
        commonRoot.setFullPath(true);
        for (i = 1; i <= commonCount; i++)
        {
                commonRoot.appendComponent(fPath->elementAtPosition(i));
        }
}


void
ICompoundName::append (const ICompoundName& partialPath)
{
        makeWritable();
        IASSERTPARM( ! partialPath.isFullPath() )

        for (unsigned int i = 1; i <= partialPath.numberOfComponents(); i++) {
                fPath->addAsLast( partialPath.componentAt(i) );
        }
}


void
ICompoundName::appendComponent(const IText& componentPath)
{
        makeWritable();
        fPath->addAsLast(componentPath);
}

void
ICompoundName::setFullPath(bool isFull)
{
        makeWritable();
        fPath->fIsFull = isFull;
}

bool
ICompoundName::isFullPath() const
{
        return fPath->fIsFull;
}


void
ICompoundName::truncate (const ICompoundName& path)
{
        makeWritable();
        IASSERTPARM( path.numberOfComponents() != 0 )
        IASSERTPARM( path.numberOfComponents() <= numberOfComponents() )

        INumber chopLength  = path.fPath->numberOfElements();
        INumber delta           = fPath->numberOfElements() - chopLength;

        // Make sure that my last "chopLength" components are really the same as "path"
        for (INumber i = 1; i <= chopLength; i++) {
                // throw a custom exception?
                IASSERTPARM( fPath->elementAtPosition(i+delta)
                                                == path.fPath->elementAtPosition(i) )
        }
        truncate(chopLength);
}


void
ICompoundName::truncate (unsigned long chopThisMuch)
{
        makeWritable();
        IASSERTPARM( chopThisMuch <= fPath->numberOfElements() )

        for (unsigned long i = 0; i < chopThisMuch; i++) {
                fPath->removeLast();
        }
}

//--------------------------------------------------------------------------
// ICompoundNameParser implementation....
//

ICompoundNameParser::~ICompoundNameParser ()
{
}

ICompoundNameParser::ICompoundNameParser ()
{
}

ICompoundNameParser::ICompoundNameParser (const ICompoundNameParser &)
{
}

ICompoundNameParser&
ICompoundNameParser::operator= (const ICompoundNameParser &)
{
        return *this;
}

bool
ICompoundNameParser::operator==(const ICompoundNameParser& other) const
{
#if IC_RTTI_ENABLED
        return typeid(*this) == typeid(other);
#else
        return this->typeOf() == (&other)->typeOf();
#endif // IC_RTTI_ENABLED
}

bool
ICompoundNameParser::operator!=(const ICompoundNameParser& other) const
{
#if IC_RTTI_ENABLED
        return typeid(*this) != typeid(other);
#else
        return this->typeOf() != (&other)->typeOf();
#endif // IC_RTTI_ENABLED
}
