// Revision: 53 1.10.1.3 source/core/filesys/ifileitr.cpp, filesystem, ioc.v400, 000302 
// ifileitr.cpp
//
// Contains: File system iterator classes
//      IDirectoryIterator              - Iterates over the contents of a directory
//      IVolumeIterator                 - Iterates over all mounted volumes on the system
//      IRootDirectoryIterator  - iterator over all root directories on the system
//
// Author:
//      Laura Werner
//
// COPYRIGHT:
//      IBM Open Class Library
//      (C) Copyright International Business Machines Corporation 1992, 1997
//      Licensed Material - Program-Property of IBM - All Rights Reserved.
//      US Government Users Restricted Rights - Use, duplication, or disclosure
//      restricted by GSA ADP Schedule Contract with IBM Corp.
//
#include <ifileitr.hpp>
#include <iprimlck.hpp>

#include "filesys.hpp"
#include "fsiterat.hpp"
#include "fsdirrec.hpp"

#ifdef IC_PAGETUNE
    #define _IFILEITR_CPP_
    #include <ipagetun.h>
#endif

//------------------------------------------------------------------------------------
//      CLASS
//              IDirectoryIterator
//

// Lazily evaluate the end directory iterator
const IDirectoryIterator& IDirectoryIterator::end()
{
    static const IDirectoryIterator* gEndIter = 0;
    if (!gEndIter)
    {
        IPrimalLock lockInit;
        if (!gEndIter)
            gEndIter = new IDirectoryIterator;
    }
    return *gEndIter;
}

IDirectoryIterator::IDirectoryIterator ()
        : fIterator( 0 )
        , fRecursive(false)
{
}

IDirectoryIterator::IDirectoryIterator(const IDirectory &dir, bool recursive)
        : fIterator( recursive ? dir.getRecord()->createRecursiveIterator()
                                                   : dir.getRecord()->createIterator() )
        , fRecursive(recursive)
{
        if (fIterator) {
                fIterator->addRef();
                fEntity = fIterator->first();
        }
}

IDirectoryIterator::IDirectoryIterator(const IDirectoryIterator& copy)
        : fIterator( copy.fIterator )
        , fEntity( copy.fEntity )
        , fRecursive( copy.fRecursive )
{
        if (fIterator) fIterator->addRef();
}

IDirectoryIterator&
IDirectoryIterator::operator=(const IDirectoryIterator& copy)
{
        fEntity = copy.fEntity;

        if (fIterator != copy.fIterator) {
                if (fIterator) fIterator->removeRef();
                fIterator = copy.fIterator;
                if (fIterator) fIterator->addRef();

                fRecursive = copy.fRecursive;
        }
        return *this;
}

IDirectoryIterator::~IDirectoryIterator ()
{
        if (fIterator) fIterator->removeRef();
}

void
IDirectoryIterator::advance()
{
        if (fIterator && fEntity.valid()) {
                fEntity = fIterator->next();
        }
}

IDirectoryIterator
IDirectoryIterator::operator++(int)
{
        // Postfix increment is tricky; we must return the value of the
        // iterator _before_ it was incremented
        IDirectoryIterator tmp = *this;
        advance();
        return tmp;
}

bool
IDirectoryIterator::recursive() const
{
        return fRecursive;
}


//------------------------------------------------------------------------------------
//      CLASS
//              IVolumeIterator
//

// Lazily evaluate the end volume iterator
const IVolumeIterator& IVolumeIterator::end()
{
    static const IVolumeIterator* gEndIter = 0;
    if (!gEndIter)
    {
        IPrimalLock lockInit;
        if (!gEndIter)
            gEndIter = new IVolumeIterator(0);
    }
    return *gEndIter;
}

IVolumeIterator::IVolumeIterator()
        : fIterator( IFSFileSystem::instance()->createVolumesIterator() )
{
        fIterator->addRef();
        fEntity = fIterator->first();
}

// Internal constructor used only for constructing the "end" iterator
IVolumeIterator::IVolumeIterator(IFSIterator* iter)
        : fIterator(iter)
{
        if (fIterator) fIterator->addRef();
}

IVolumeIterator::IVolumeIterator(const IVolumeIterator& copy)
        : fIterator( copy.fIterator )
        , fEntity( copy.fEntity )
{
        if (fIterator) fIterator->addRef();
}

IVolumeIterator&
IVolumeIterator::operator=(const IVolumeIterator& copy)
{
        fEntity = copy.fEntity;

        if (fIterator != copy.fIterator) {
                if (fIterator) fIterator->removeRef();
                fIterator = copy.fIterator;
                if (fIterator) fIterator->addRef();
        }
        return *this;
}

IVolumeIterator::~IVolumeIterator ()
{
        if (fIterator) fIterator->removeRef();
}

void
IVolumeIterator::advance()
{
        if (fIterator && fEntity.valid()) {
                fEntity = fIterator->next();
        }
}

IVolumeIterator
IVolumeIterator::operator++(int)
{
        // Postfix increment is tricky; we must return the value of the
        // iterator _before_ it was incremented
        IVolumeIterator tmp = *this;
        advance();
        return tmp;
}


//------------------------------------------------------------------------------------
//      CLASS
//              IRootDirectoryIterator
//

// Lazily evaluate the end root directory iterator
const IRootDirectoryIterator& IRootDirectoryIterator::end()
{
    static const IRootDirectoryIterator* gEndIter = 0;
    if (!gEndIter)
    {
        IPrimalLock lockInit;
        if (!gEndIter)
            gEndIter = new IRootDirectoryIterator(0);
    }
    return *gEndIter;
}

IRootDirectoryIterator::IRootDirectoryIterator()
        : fIterator( IFSFileSystem::instance()->createRootDirectoryIterator() )
{
        if (fIterator) {
                fIterator->addRef();
                fEntity = fIterator->first();
        }
}

IRootDirectoryIterator::IRootDirectoryIterator(const IRootDirectoryIterator& copy)
        : fIterator( copy.fIterator )
        , fEntity( copy.fEntity )
{
        if (fIterator) fIterator->addRef();
}

// Internal constructor used only for constructing "end" iterator
IRootDirectoryIterator::IRootDirectoryIterator(IFSIterator* iter)
        : fIterator(iter)
{
        if (fIterator) fIterator->addRef();
}

IRootDirectoryIterator&
IRootDirectoryIterator::operator=(const IRootDirectoryIterator& copy)
{
        fEntity = copy.fEntity;

        if (fIterator != copy.fIterator) {
                if (fIterator) fIterator->removeRef();
                fIterator = copy.fIterator;
                if (fIterator) fIterator->addRef();
        }
        return *this;
}

IRootDirectoryIterator::~IRootDirectoryIterator ()
{
        if (fIterator) fIterator->removeRef();
}


void IRootDirectoryIterator::advance()
{
        if (fIterator && fEntity.valid()) {
                fEntity = fIterator->next();
        }
}

IRootDirectoryIterator
IRootDirectoryIterator::operator++(int)
{
        // Postfix increment is tricky; we must return the value of the
        // iterator _before_ it was incremented
        IRootDirectoryIterator tmp = *this;
        advance();
        return tmp;
}

