// Revision: 37 1.12.1.3 source/core/filesys/pathpars.cpp, filesystem, ioc.v400, 000302 
// pathpars.cpp
//
// Contains:
//      IUnixPathNameParser     - Convert Unix-style path strings to/from IPathName
//      IWin32PathNameParser- Convert Win32-style path strings to/from IPathName
//
// Author:
//      Laura Werner
//
// COPYRIGHT:
//      IBM Open Class Library
//      (C) Copyright International Business Machines Corporation 1997
//      Licensed Material - Program-Property of IBM - All Rights Reserved.
//      US Government Users Restricted Rights - Use, duplication, or disclosure
//      restricted by GSA ADP Schedule Contract with IBM Corp.
//
#include <ilanglvl.hpp>
#include <ifilesys.hpp>
#include <ipathnam.hpp>

#include "filesys.hpp"

#ifdef IC_PAGETUNE
    #define _PATHPARS_CPP_
    #include <ipagetun.h>
#endif
//--------------------------------------------------------------------------
// IUnixPathNameParser globals....
//

#if !IC_RTTI_ENABLED
// Type constant for equality comparisons.  Temporary until we get RTTI.
const IUnixPathNameParser::IParserType
                IUnixPathNameParser::kType = "IUnixParser";
#endif // IC_RTTI_ENABLED


//--------------------------------------------------------------------------
// Constructors, etc.....
//
IUnixPathNameParser::~IUnixPathNameParser ()
{
}

IUnixPathNameParser::IUnixPathNameParser ()
{
}

IUnixPathNameParser::IUnixPathNameParser (const IUnixPathNameParser &other)
        : ICompoundNameParser(other)
{
}

IUnixPathNameParser&
IUnixPathNameParser::operator= (const IUnixPathNameParser &other)
{
        ICompoundNameParser::operator=(other);
        return *this;
}


//--------------------------------------------------------------------------
// Access methods.....
//
#if !IC_RTTI_ENABLED
IUnixPathNameParser::IParserType
IUnixPathNameParser::typeOf() const
{
        return kType;
}
#endif // IC_RTTI_ENABLED

UniChar
IUnixPathNameParser::separator() const
{
        return '/';
}


//--------------------------------------------------------------------------
// parseName
//      Convert a unix-style path name string into an IPathName
//
ICompoundName
IUnixPathNameParser::parseName (const IHostPathName& path) const
{
        ICompoundName newPath;
        IText::offset_type length = path.length();
        IText::offset_type pos = 0, next = 0;

        // Is it a full path name?
        if (length && path.at(0) == kUnixPathSeparator) {
                // WJW - Add an empty component to represent the root.
                newPath.appendComponent(IFileName());
                newPath.setFullPath(true);
                pos++;
        }

        // Chop the path into components by searching for '/' separator characters

        while ( (next = path.find('/', pos)) != IText::npos ) {
                // Found one.
                IFileName component = path.substr(pos, next - pos);
                if (component.length() && component != IText(".")) {
                        newPath.appendComponent(component);
                }
                pos = next + 1;
        }
        if (pos < length) {
                // There's a component after the last '/' so append it too
                newPath.appendComponent(path.substr(pos));
        }
        return newPath;
}

//--------------------------------------------------------------------------
// formatName
//      This is easy; just append each component, separated by '/' characters
//
IHostPathName
IUnixPathNameParser::formatName (const ICompoundName& aPath) const
{
        IHostPathName newPath;
        unsigned long componentCount = aPath.numberOfComponents();

        IFileName component;
        // first get/check the volume name
        // WJW - The leading slash will be added appropriately because
        // the first component of a full path is an empty component.
        if (aPath.isFullPath() && aPath.componentAt(1).length() != 0) {
                newPath += '/';
        }
        for (unsigned long count = 1; count <= componentCount; count++)
        {
                newPath += aPath.componentAt(count);
                if (count != componentCount ||
                        (componentCount == 1 && aPath.componentAt(1).length() == 0))
                        newPath += '/';
        }
        return newPath;
}


//--------------------------------------------------------------------------
// IWin32PathNameParser implementation....
//

#if !IC_RTTI_ENABLED
// Type constant for equality comparisons.  Temporary until we get RTTI.
const IWin32PathNameParser::IParserType
                IWin32PathNameParser::kType = "IWin32Parser";
#endif // IC_RTTI_ENABLED


//--------------------------------------------------------------------------
// Constructors, etc.....
//
IWin32PathNameParser::~IWin32PathNameParser ()
{
}

IWin32PathNameParser::IWin32PathNameParser ()
{
}

IWin32PathNameParser::IWin32PathNameParser (const IWin32PathNameParser &other)
        : ICompoundNameParser(other)
{
}

IWin32PathNameParser&
IWin32PathNameParser::operator= (const IWin32PathNameParser &other)
{
        ICompoundNameParser::operator=(other);
        return *this;
}


//--------------------------------------------------------------------------
// Access methods.....
//
#if !IC_RTTI_ENABLED
ICompoundNameParser::IParserType
IWin32PathNameParser::typeOf() const
{
                return kType;
}
#endif // IC_RTTI_ENABLED

UniChar
IWin32PathNameParser::separator() const
{
        return '\\';
}


//--------------------------------------------------------------------------
// parseName
//      Convert a Windows-style path name string into an IPathName.
//      The simple case is easy: just search for \ characters.
//      But we also have to handle UNC names like \\host\share\dir\file.txt
//      and paths containing drive letters, which complicate matters quite a bit
//
ICompoundName
IWin32PathNameParser::parseName(const IHostPathName& path) const
{
        ICompoundName newPath;
        IText::offset_type length = path.length();
        IText::offset_type pos = 0, next = 0;

        if (path.find(":\\") == 1) {                    // Does it start with a drive letter?
                newPath.setFullPath(true);
                newPath.appendComponent( path.substr(0, 2) );
                pos += 3;
        }
        else if (path.find("\\\\") == 0) {      // Is it a UNC-style path?
                newPath.setFullPath(true);
                // Skip the initial two slashes, then search for two more, so we
                // can yank out the \\host\share part of the path
                pos += 1;
                for (int i = 0; i < 2 && pos != IText::npos; i++) {
                        pos = path.find('\\', pos+1);
                }

                if (pos != IText::npos) {
                        // Swallow just the \\host\share part of the path
                        newPath.appendComponent(path.substr(0, pos));
                        pos += 1;
                } else {
                        // Swallow the whole thing
                        newPath.appendComponent(path);
                        pos = length;
                }
        }
        else if (path.find("\\") == 0) {                // "full" path without drive letter
                pos += 1;
                newPath.setFullPath(true);
        }

        // Now, finally, run through the rest of the path and chop it
        // into components by looking for \ characters.

        while ( (next = path.find('\\', pos)) != IText::npos ) {
                IFileName component = path.substr(pos, next - pos);
                if (component.length() && component != IText(".")) {
                        newPath.appendComponent(component);
                }
                pos = next + 1;
        }
        if (pos < length) {
                newPath.appendComponent(path.substr(pos));
        }
        return newPath;
}


//--------------------------------------------------------------------------
// formatName
//      This is easy; just append each component, separated by '\' characters
//
IHostPathName
IWin32PathNameParser::formatName(const ICompoundName& aPath) const
{
        IHostPathName newPath;
        unsigned long componentCount = aPath.numberOfComponents();

        // If this is a full path but the first component is not a drive or host,
        // we need to slap an extra slash on the host path

        int start = 1;
        if (aPath.isFullPath()) {
                // WJW - If this is a path created by a Unix path name
                // parser it will have an empty first component, so
                // we ignore it.
                if(aPath.componentAt(1).length() == 0)
                {
                        start = 2;
                }
                IFileName first = aPath.componentAt(start);
                if ( !(first.at(1) == ':' || first.substr(0, 2) == IText("\\\\")) ) {
                        newPath += '\\';
                }
        }

        for (unsigned long count = start; count <= componentCount; count++) {
                newPath += aPath.componentAt(count);
                if (count != componentCount)
                        newPath += '\\';
        }

        // If it's a naked drive letter, tack on a \ to make it a legal path
        if (newPath.length() == 2 && newPath.at(1) == ':') {
                newPath += "\\";
        }
        return newPath;
}



