/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/

// Revision: 81 1.11.1.3 source/core/filesys/win32/windir.cpp, filesystem, ioc.v400, 000302 
// windir.cpp
// Created by laura on Thu, Jan 24, 1996 @ 12:42 PM.

#include <igenexc.hpp>
#include <iexcfile.hpp>

#include "windir.hpp"
#include "winfile.hpp"
#include "winflsys.hpp"
#include "winvol.hpp"
#include "winhdl.hpp"

#ifdef IC_PAGETUNE
    #define _WINDIR_CPP_
    #include <ipagetun.h>
#endif

INTDirectoryRecord::~INTDirectoryRecord()
{
}

IDirectory                      INTDirectoryRecord::createDirectory(const IFileName& name)
{
        IFSRecordLock lock(this);
        return INTFileSystem::createDirectory( fPath + name );
}

void                            INTDirectoryRecord::setCurrent() const
{
        IFSRecordLock lock(this);
        INTFileSystem::setCurrentDirectory( fPath );
}






IFileSystemEntity INTDirectoryRecord::lookUpByName(const IFileName& theName) const

{
        IFSRecordLock lock(this);
        INTPathName fullPath = fPath + theName;

        return INTFileSystem::lookUpFullPath( fullPath );
}


IFSIterator *   INTDirectoryRecord::createIterator() const
{
        IFSRecordLock lock(this);
        return new INTDirectoryIterator( fPath );
}

// ------------------------------------------------------------------
// copySelf
//      Create a new directory named "newName" in "to" that has the
//      same attributes, permissions, etc. as this directory.  This
//      does NOT copy the contents of the directory.
//
IFileSystemEntity
INTDirectoryRecord::copySelf(IDirectory &to, const IFileName &newName) const
{
        // Create a new directory named "newName" in "to" that has the
        // same attributes, permissions, etc. as this directory.
        INTPathName newPath(to, newName);
        {
                IFSRecordLock lock(this);
                INTFileSystem::createDirectoryEx(fPath, newPath, NULL);
        }
        return new INTDirectoryRecord(newPath);
}



IFile
INTDirectoryRecord::createFile(const IFileName& name,
                                                                IFileSize initialSize)
{
        IFSRecordLock lock(this);
        return INTFileSystem::createFile( fPath + name, initialSize );
}




INTDirectoryRecord::INTDirectoryRecord(const INTPathName &fullPath)
        : fPath(fullPath)
{
}




INTDirectoryIterator::~INTDirectoryIterator()
{
        if (fHandle != INVALID_HANDLE_VALUE) {
                ::FindClose(fHandle);
        }
}

IFileSystemEntity       INTDirectoryIterator::first ()

{
        IFileSystemEntity result;

        if (fHandle != INVALID_HANDLE_VALUE) {
                ::FindClose(fHandle);
                fHandle = INVALID_HANDLE_VALUE;
        }

        WIN32_FIND_DATAW data;

        // If there are no files to find, this will return an invalid handle
        fHandle = INTFileSystem::findFirstFile(fPath + IFileName("*"), &data);

        if (fHandle != INVALID_HANDLE_VALUE) {
                result = makeEntity(data);
                if (!result.valid()) {
                        result = next();
                }
        }
        return result;
}


IFileSystemEntity       INTDirectoryIterator::next ()

{
        IFileSystemEntity result;
        WIN32_FIND_DATAW data;

        while ( !result.valid() && fHandle != INVALID_HANDLE_VALUE ) {
                memset(&data, 0, sizeof(data));
                BOOL found = INTFileSystem::findNextFile(fHandle, &data);
                if (found) {
                        // We found something; try to make an entity from it
                        result = makeEntity(data);
                } else {
                        // We've hit the end of this iteration.
                        ::FindClose(fHandle);
                        fHandle = INVALID_HANDLE_VALUE;
                }
        }
        return result;
}


INTDirectoryIterator::INTDirectoryIterator(const INTPathName path)
        : fPath(path)
        , fHandle(INVALID_HANDLE_VALUE)
{
}


INTDirectoryIterator::INTDirectoryIterator(const INTDirectoryIterator &copy)
        : fPath("")
{
        // Disallowed
}


INTDirectoryIterator & INTDirectoryIterator::operator = (const INTDirectoryIterator &copy)
{
        // Disallowed
        return *this;
}

IFileSystemEntity INTDirectoryIterator::makeEntity(const WIN32_FIND_DATAW &data)
{
        const WCHAR *           name = data.cFileName;
        IFileSystemEntity       result;

        const WCHAR kCurrent[] = L".";
        const WCHAR kParent[]  = L"..";

        // Skip the current directory and parent directory placeholders
        // We can't use lstrcmpW here, because it doesn't work on Win95

        if ( memcmp(name, kCurrent, sizeof(kCurrent)) != 0
                         && memcmp(name, kParent, sizeof(kParent)) != 0 )
        {
                INTPathName fullPath = INTFileSystem::getFullPathName(fPath + name);

                if (data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
                        result = new INTDirectoryRecord( fullPath );
                } else {
                        result = new INTFileRecord( fullPath );
                }
        }
        return result;
}







INTDirectoryRecord::INTDirectoryRecord(const INTDirectoryRecord &)
        : fPath("")

{
        // Disallowed
}


INTDirectoryRecord &            INTDirectoryRecord::operator = (const INTDirectoryRecord &)
{
        // Disallowed
        return *this;
}



void                                            INTDirectoryRecord::doDeleteSelf ()
{
        INTFileSystem::removeDirectory( fPath );
}


IVolume                                 INTDirectoryRecord::volume () const
{
        IFSRecordLock lock(this);
        return new INTVolumeRecord( fPath.getVolumeName() );
}


bool            INTDirectoryRecord::testConnection() const

{
        IFSRecordLock lock(this);
        DWORD attr = INTFileSystem::getFileAttributes( fPath );

        return ( attr != 0xFFFFFFFF ) && ( attr & FILE_ATTRIBUTE_DIRECTORY );
}


IFileName       INTDirectoryRecord::name() const
{
        IFSRecordLock lock(this);
        return fPath.getLeafName();
}


void                                    INTDirectoryRecord::setName(const IFileName& newName)
{
        IFSRecordLock lock(this);

        INTPathName newPath = fPath.getParentName() + newName;
        INTFileSystem::moveFile( fPath, newPath );
        fPath = newPath;
}


ITimeStamp INTDirectoryRecord::modificationTime() const
{
        IFSRecordLock lock(this);

        // GetFileInformation doesn't work for directories on Win95.
        // The workaround is to use FindFirstFile

        WIN32_FIND_DATAW data;
        HANDLE handle = INTFileSystem::findFirstFile(fPath, &data);
        if (handle == INVALID_HANDLE_VALUE) {
                IObjectNotFound e;
                ITHROW(e);
        }
        FindClose(handle);

        return INTFileSystem::fromFileTime(data.ftLastWriteTime);
}





bool                                    INTDirectoryRecord::isEqual ( const IFSRecord *other) const
{
        IFSRecordLock lock1(this);
        IFSRecordLock lock2(other);
        // This method won't be called unless the C++ types are equal....
        return ((INTDirectoryRecord *)other)->fPath == fPath;
}


// ------------------------------------------------------------------
// moveTo
//      Move and/or rename this directory so that its parent is "to"
//      and its name is "newName".  It is OK to return an invalid
//      entity if the move isn't possible (e.g. different volumes).
//      In that case the caller will do a copy followed by a delete.
//
IFileSystemEntity
INTDirectoryRecord::moveTo(IDirectory &to, const IFileName &newName)
{
        IFileSystemEntity result;

        if (volume() == to.volume()) {
                // NT's MoveFile can do directories only on the same volume
                INTPathName newPath(to, newName);

                IFSRecordLock lock(this);
                INTFileSystem::moveFile(fPath, newPath);
                fPath = newPath;

                result = IFileSystemEntity(this);
        }
        return result;
}


IHostPathName           INTDirectoryRecord::hostPath() const
{
        IFSRecordLock lock(this);
        return fPath;
}


IDirectory      INTDirectoryRecord::parent() const
{
        IFSRecordLock lock(this);
        return INTFileSystem::lookUpFullPath( fPath.getParentName() );
}


IDirectory      INTDirectoryRecord::topmostParent() const
{
        return volume();
}

