/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/

// Revision: 78 1.11.1.3 source/core/filesys/win32/winfile.cpp, filesystem, ioc.v400, 000302 
// winfile.cpp
// Created by laura on Sat, Jan 26, 1996 @ 12:40 PM.

#include "winfile.hpp"
#include "winvol.hpp"
#include "winflsys.hpp"
#include "winstrm.hpp"

#define WIN32_LEAN_AND_MEAN 1
#include <windows.h>

#include "winhdl.hpp"

#ifdef IC_PAGETUNE
    #define _WINFILE_CPP_
    #include <ipagetun.h>
#endif

INTFileRecord::~INTFileRecord()
{
}

void                                    INTFileRecord::doDeleteSelf ()
{
        INTFileSystem::deleteFile(fPath);
}


IFileSize               INTFileRecord::size() const
{
        IFSRecordLock lock(this);

        // Use FindFirstFile instead of CreateFile/GetFileInformationByHandle,
        // because the latter fails on Win95 if the file is already open

        WIN32_FIND_DATAW info;
        HANDLE findHdl = INTFileSystem::findFirstFile(fPath, &info);
        if (findHdl == INVALID_HANDLE_VALUE) {
                INTFileSystem::throwLastError(ERROR_ACCESS_DENIED, "FindFirstFile");
        }
        ::FindClose(findHdl);

        IFileSize result = IFileSize(info.nFileSizeHigh) * ULONG_MAX
                                        + IFileSize(info.nFileSizeLow);
        return result;
}

IFileSystemEntity
INTFileRecord::copySelf(IDirectory &to, const IFileName &newName) const
{
        INTPathName newPath(to, newName);
        {
                IFSRecordLock lock(this);
                INTFileSystem::copyFile(fPath, newPath, TRUE);
        }
        return new INTFileRecord(newPath);
}





bool                                    INTFileRecord::isEqual ( const IFSRecord *other) const
{
        IFSRecordLock lock1(this);
        IFSRecordLock lock2(other);
        // This method won't be called unless the C++ types are equal....
        return ((INTFileRecord *)other)->fPath == fPath;
}


IVolume                                 INTFileRecord::volume () const
{
        IFSRecordLock lock(this);
        return new INTVolumeRecord( fPath.getVolumeName() );
}


IFileSystemEntity INTFileRecord::moveTo(IDirectory &to, const IFileName &newName)
{
        IFSRecordLock lock(this);

        INTPathName newPath(to, newName);
        INTFileSystem::moveFile(fPath, newPath);
        fPath = newPath;

        return IFileSystemEntity(this);
}


IDirectory      INTFileRecord::parent() const
{
        IFSRecordLock lock(this);
        return INTFileSystem::lookUpFullPath( fPath.getParentName() );
}


IDirectory      INTFileRecord::topmostParent() const
{
        return volume();
}


bool            INTFileRecord::testConnection() const
{
        IFSRecordLock lock(this);
        DWORD attr = INTFileSystem::getFileAttributes( fPath );

        return ( attr != 0xFFFFFFFF ) && !( attr & FILE_ATTRIBUTE_DIRECTORY );
}





IHostPathName           INTFileRecord::hostPath() const
{
        IFSRecordLock lock(this);
        return fPath;
}

IFSFileStream*  INTFileRecord::createStream(int myMode, int otherMode) const
{
        IFSRecordLock lock(this);
        return new IWinFileStream(fPath, myMode, otherMode);
}

void                            INTFileRecord::setSize (IFileSize newSize)
{
        TWin32Handle handle;
        {
                IFSRecordLock lock(this);
                handle = INTFileSystem::createFile( fPath, GENERIC_WRITE, 0, NULL,
                                                                                        OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL );
        }
        INTFileSystem::setFilePointer( handle, newSize );

        if ( !SetEndOfFile(handle) ) {
                INTFileSystem::throwLastError("SetEndOfFile");
        }
        handle.Close();
}




IFileName       INTFileRecord::name() const
{
        IFSRecordLock lock(this);
        return fPath.getLeafName();
}


void    INTFileRecord::setName(const IFileName& newName)
{
        IFSRecordLock lock(this);
        INTPathName newPath = fPath.getParentName() + newName;
        INTFileSystem::moveFile( fPath, newPath );
        fPath = newPath;
}


ITimeStamp INTFileRecord::modificationTime() const
{
        IFSRecordLock lock(this);
        return INTFileSystem::modificationTime(fPath);
}


void INTFileRecord::setModificationTime(const ITimeStamp& modificationTime)
{
        IFSRecordLock lock(this);
        INTFileSystem::setModificationTime(fPath, modificationTime);
}


INTFileRecord::INTFileRecord(const INTPathName &path)
        : fPath(path)
{
}


INTFileRecord::INTFileRecord(const INTFileRecord &copy)
        : fPath("")
{
        // Disallowed
}


INTFileRecord &         INTFileRecord::operator = (const INTFileRecord &copy)
{
        // Disallowed
        return *this;
}




