/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/

// Revision: 83 1.10.1.4 source/core/filesys/win32/winvol.cpp, filesystem, ioc.v400, 000302 
// winvol.cpp
// Created by laura on Sat, Jan 26, 1996 @ 12:32 PM.

#include <typeinfo.h>
#include <stdio.h>

#include "iexcfile.hpp"
#include "winvol.hpp"
#include "winflsys.hpp"
#include "windir.hpp"
#include "itmstamp.hpp"

#ifdef IC_PAGETUNE
    #define _WINVOL_CPP_
    #include <ipagetun.h>
#endif

INTVolumeIterator::INTVolumeIterator()
        : fDriveStrings( INTFileSystem::copyLogicalDriveStrings() )
        , fCurrentDrive( fDriveStrings )
{
}


INTVolumeIterator::~INTVolumeIterator()
{
        delete[] fDriveStrings;
}

IFileSystemEntity       INTVolumeIterator::first ()

{
        fCurrentDrive = fDriveStrings;
        return next();
}


IFileSystemEntity       INTVolumeIterator::next ()
{
        IFSVolumeRecord *result = 0;

        while ( *fCurrentDrive && !result ) {
                INTPathName path = fCurrentDrive;

                // Skip past the end of this string to the next drive name
                while ( *fCurrentDrive++ ) {
                }

                // Make sure this is a recognizable type of drive
                // and that it's online....
                try {
                        UINT type = INTFileSystem::getDriveType(path);
                        if (type != 0 && type != 1) {
                                DWORD serialNumber, maxComponentLength, fileSystemFlags;
                                INTFileSystem::getVolumeInformation(path, serialNumber,
                                                                                 maxComponentLength, fileSystemFlags);

                                // Create a volume record
                                result = new INTVolumeRecord( INTPathName(path) );
                        }
                }
                catch ( const IVolumeOffline & ) {
                }
                catch ( const IAccessError & ) {
                }
        }

        return IFileSystemEntity(result);
}

INTVolumeIterator::INTVolumeIterator(const INTVolumeIterator &copy)
        : fDriveStrings( new WCHAR[lstrlenW(copy.fDriveStrings)] )
{
        lstrcpyW(fDriveStrings, copy.fDriveStrings);

        // fCurrentDrive should have the same offset in me as in the other
        fCurrentDrive = fDriveStrings + (copy.fCurrentDrive - copy.fDriveStrings);
}


INTVolumeIterator&      INTVolumeIterator::operator = (const INTVolumeIterator &copy)
{
        // Disallowed
        return *this;
}






INTVolumeRecord::INTVolumeRecord(const INTPathName &path)
        : fPath(path)
{
}


INTVolumeRecord::~INTVolumeRecord()
{
}

IFileSize               INTVolumeRecord::freeSpace                              () const
{
        DWORD sectorsPerCluster, bytesPerSector, freeClusters, clusters;
        {
                IFSRecordLock lock(this);
                INTFileSystem::getDiskFreeSpace(fPath, sectorsPerCluster, bytesPerSector,
                                                                                freeClusters, clusters);
        }
        return IFileSize(sectorsPerCluster) * IFileSize(bytesPerSector)
                                * IFileSize(freeClusters);
}


IFileSize               INTVolumeRecord::totalSpace () const
{
        DWORD sectorsPerCluster=0, bytesPerSector=0, freeClusters=0, clusters=0;
        {
                IFSRecordLock lock(this);
                INTFileSystem::getDiskFreeSpace(fPath, sectorsPerCluster, bytesPerSector,
                                                                                freeClusters, clusters);
        }
        return IFileSize(sectorsPerCluster) * IFileSize(bytesPerSector)
                                * IFileSize(clusters);
}


IVolume::EState
INTVolumeRecord::remote () const
{
        IFSRecordLock lock(this);

        UINT type = INTFileSystem::getDriveType(fPath);

        return (type == DRIVE_REMOTE) ? IVolume::kTrue : IVolume::kFalse;
}


IVolume::EState
INTVolumeRecord::removable () const
{
        IFSRecordLock lock(this);
        UINT type = INTFileSystem::getDriveType(fPath);

        return (type == DRIVE_REMOVABLE || type == DRIVE_CDROM) ?
                                IVolume::kTrue : IVolume::kFalse;
}


IFileSystemEntity INTVolumeRecord::lookUpByName(const IFileName& theName) const
{
        IFSRecordLock lock(this);
        INTPathName fullPath = fPath + theName;

        return INTFileSystem::lookUpFullPath( fullPath );
}


IFSIterator *   INTVolumeRecord::createIterator() const
{
        IFSRecordLock lock(this);
        return new INTDirectoryIterator( fPath );
}

// ------------------------------------------------------------------
// copySelf
//      Create a new directory named "newName" in "to" that has the
//      same attributes, permissions, etc. as this volume.  This
//      does NOT copy the contents of the volume.
//
IFileSystemEntity
INTVolumeRecord::copySelf(IDirectory &to, const IFileName &newName) const
{
        INTPathName newPath(to, newName);
        {
                IFSRecordLock lock(this);
                INTFileSystem::createDirectoryEx(fPath, newPath, NULL);
        }
        return new INTDirectoryRecord(newPath);
}



IFile                           INTVolumeRecord::createFile(const IFileName& name,
                                                                IFileSize initialSize)
{
        IFSRecordLock lock(this);
        return INTFileSystem::createFile( fPath + name, initialSize );
}


IDirectory                      INTVolumeRecord::createDirectory(const IFileName& name)
{
        IFSRecordLock lock(this);
        return INTFileSystem::createDirectory( fPath + name );
}

void                                            INTVolumeRecord::setCurrent() const
{
        IFSRecordLock lock(this);
        INTFileSystem::setCurrentDirectory( fPath );
}




bool                                    INTVolumeRecord::isEqual ( const IFSRecord *other) const
{
        IFSRecordLock lock1(this);
        IFSRecordLock lock2(other);

        // This method won't be called unless the C++ types are equal....
        return ((INTVolumeRecord *)other)->fPath == fPath;
}


IVolume                                 INTVolumeRecord::volume () const
{
        return IVolume( (IFSRecord*)this );
}





IDirectory      INTVolumeRecord::parent() const
{
        // NT Volumes don't have parents
        return IDirectory();
}


IDirectory      INTVolumeRecord::topmostParent() const
{
        // NT Volumes don't have parents
        return IDirectory();
}


bool            INTVolumeRecord::testConnection() const
{
        IFSRecordLock lock(this);
        DWORD attr = INTFileSystem::getFileAttributes( fPath );

        return ( attr != 0xFFFFFFFF ) && ( attr & FILE_ATTRIBUTE_DIRECTORY );
}





IHostPathName           INTVolumeRecord::hostPath() const
{
        IFSRecordLock lock(this);
        return fPath;
}


IFileName       INTVolumeRecord::name() const
{
        IFSRecordLock lock(this);
        return fPath.getLeafName();
}





ITimeStamp INTVolumeRecord::modificationTime() const
{
        IFSRecordLock lock(this);

        BY_HANDLE_FILE_INFORMATION info;
        INTFileSystem::getFileInformation(fPath, &info);
        return INTFileSystem::fromFileTime( info.ftLastWriteTime );
}





INTVolumeRecord::INTVolumeRecord(const INTVolumeRecord &copy)
        : fPath("")
{
        // Disallowed
}


INTVolumeRecord &                       INTVolumeRecord::operator = (const INTVolumeRecord &copy)
{
        // Disallowed
        return *this;
}


