// ----------------------------------------------------------------------------
//  FILE: icntstio.cpp
//
//
// COPYRIGHT:
//  IBM Open Class Library
//  (C) Copyright International Business Machines Corporation,  1997
//  Licensed Material - Program-Property of IBM - All Rights Reserved.
//  US Government Users Restricted Rights - Use, duplication, or disclosure
//  restricted by GSA ADP Schedule Contract with IBM Corp.
//
// Revision: 55 1.20.1.4 source/core/streams/icntstio.cpp, oss, ioc.v400, 000302 
// ----------------------------------------------------------------------------

#include "icntstio.hpp"

#include <imstrmbl.hpp>
#include <istrmbf.hpp>
#include <istrmcf.hpp>
#include "istrmctx.hpp"
#include <istrmexc.hpp>
#include <istrmfnd.hpp>
#include <ityperep.hpp>
#include <itrace.hpp>
#include <istring.hpp>
#include <imrefcnt.hpp>
#include "itypenam.hpp"

#include <stdio.h>
#include <stdarg.h>
#include <string.h>


#ifdef IC_PAGETUNE
    #define _ICNTSTIO_CPP_
    #include <ipagetun.h>
#endif

// #define DEBUG_STREAMIO

//
// forward class declarations
//
class ICountedStreamBuffer;

#pragma enum(4)
#pragma pack(push,4)

//
// ICountedStreamFrame
//
class ICountedStreamFrame : public virtual IMRefCounted
{
public:
    ICountedStreamFrame(unsigned long frameDepth, unsigned long length, const char* data = 0);
    ~ICountedStreamFrame();

    ICountedStreamFrame* next()
    {
        return fNext;
    };

    char* buffer()
    {
        return fBuffer;
    };

    unsigned long length()
    {
        return fLength;
    };

    unsigned long depth()
    {
        return fDepth;
    };

    ICountedStreamFrame* nextLink();
    void append(ICountedStreamFrame* frame);
    void print();
    void printAll(unsigned long depth = 0);

private:
    ICountedStreamFrame* fNext;
    unsigned long fDepth;
    char* fBuffer;
    unsigned long fLength;

    ICountedStreamFrame();
    ICountedStreamFrame(const ICountedStreamFrame& other);
};

//
// ICountedStreamableContainer
//
class ICountedStreamableContainer : public IContextContainer
{
public:
    ICountedStreamableContainer(ICountedStreamBuffer* streamBuffer, ICountedStreamFrame* firstFrame);
    ~ICountedStreamableContainer();

    virtual IMStreamable* readAliasedObject(IMRefCounted*& refCounted, const IStreamFinder* finder, const ITypeRepresentation* typeRep);
    virtual const ITypeRepresentation* readObjectType();
    virtual void printAll();

private:
    ICountedStreamBuffer* fBuffer;
    ICountedStreamFrame*  fFrame;

    ICountedStreamableContainer();
    ICountedStreamableContainer(const ICountedStreamableContainer& other);
};

//
// ICountedStreamBuffer
//
class ICountedStreamBuffer
{
public:
    ICountedStreamBuffer(IStreamBuffer* streamBuffer);
    ~ICountedStreamBuffer();

    void    setStream(IDataStream* dataStream);
    void    reset();
    void    flush();

    bool    establishContext();
    void    clearContext();

    void    writeByte(char value);
    void    writeLong(long value);
    void    writeShort(short value);
    void    writeFloat(float value);
    void    writeDouble(double value);

    char    readByte();
    long    readLong();
    short   readShort();
    float   readFloat();
    double  readDouble();

    void    writeBytes(const char* data, unsigned long howMany);
    void    readBytes(char* data, unsigned long howMany);
    void    writeShorts(const short* data, unsigned long howMany);
    void    readShorts(short* data, unsigned long howMany);
    void    writeLongs(const long* data, unsigned long howMany);
    void    readLongs(long* data, unsigned long howMany);
    void    writeFloats(const float* data, unsigned long howMany);
    void    readFloats(float* data, unsigned long howMany);
    void    writeDoubles(const double* data, unsigned long howMany);
    void    readDoubles(double* data, unsigned long howMany);

    void    writeObject(const IMStreamable* object, const ITypeRepresentation* typeRep);
    IMStreamable*    readObject(const IStreamFinder* streamFinder, const ITypeRepresentation* typeRep);
    bool    writeAliasedObject(const IMStreamable* object, const IMRefCounted* refCounted, const ITypeRepresentation* typeRep);
    bool    readAliasedObject(IMStreamable*& object, IMRefCounted*& refCounted, const IStreamFinder* finder, const ITypeRepresentation* typeRep);

    void    writeStartOfFrame();
    void    writeEndOfFrame();
    void    readStartOfFrame();
    void    readEndOfFrame();
    bool    atEndOfFrame();

protected:
    ICountedStreamBuffer();
    ICountedStreamBuffer(const ICountedStreamBuffer& other);

private:
    void            writeCompressedLong(unsigned long length);
    unsigned long   readCompressedLong();

    void            writeContextNumber(unsigned long contextNumber);
    unsigned long   readContextNumber();

    void    writeTypeName(const ITypeName& typeName);
    void    readTypeName(ITypeName& typeName);

    void    invalidFormat();

    IStreamBuffer*          fStreamBuffer;
    IDataStream*            fDataStream;
    IStreamContext*         fStreamContext;
    const IStreamFinder*    fStreamFinder;

    enum
    {
        kStartFrameTag = 0xE0,
        kEndFrameTag = 0xE1,
        kChangeBufferTag = 0xE2,
        kDataBufferTag = 0xE3,
        kNilObjectTag = 0xE4,
        kAliasedObjectTag = 0xE5,
        kTypedAliasedObjectTag = 0xE6,
        kStartAliasedTypeObjectTag = 0xE7,
        kStartTypedObjectTag = 0xE8,
        kStartUntypedObjectTag = 0xE9,
        kEndObjectTag = 0xEA
    };

    friend class ICountedStreamableContainer;
    IMStreamable*    readAliasedObjectFromFrame(ICountedStreamFrame* frame, IMRefCounted*& refCounted, const IStreamFinder* finder, const ITypeRepresentation* typeRep);
    const ITypeRepresentation*    readObjectTypeFromFrame(ICountedStreamFrame* frame);
    IMStreamable*    readAliasedObjectAgain(IMRefCounted*& refCounted);
    const ITypeRepresentation*    readObjectType();

    void    writeStartOfNewFrame(unsigned char tag);
    void    writeEndOfCurrentFrame(bool doFlush);

    void    writeStartOfObject(unsigned char tag);
    void    writeEndOfObject();
    unsigned char    readStartOfObject();
    void    readEndOfObject();

    void    updateAndWriteBuffer();
    unsigned char    readAllFrames(bool isObjectTag);
    void    flushBuffer();
    void    fillBuffer(unsigned long length, bool handleChangeBuffer = true);
    void    skipRemainderOfFrame(unsigned char currentTag);

    ICountedStreamFrame*    fFrame;
    long    fFrameDepth;
    unsigned char*    fInternalBuffer;
    unsigned char*    fBuffer;
    unsigned char*    fPointer;
    unsigned char*    fEndPointer;
    unsigned char*    fFrameLengthPointer;
    bool    fUsingContainer;

    void    swapBytesToBuffer(const unsigned char* data, unsigned long length);
    void    swapBytesFromBuffer(unsigned char* data, unsigned long length);
    void    swapShortsToBuffer(const unsigned char* data, unsigned long length);
    void    swapShortsFromBuffer(unsigned char* data, unsigned long length);
    void    swapLongsToBuffer(const unsigned char* data, unsigned long length);
    void    swapLongsFromBuffer(unsigned char* data, unsigned long length);
    void    swapFloatsToBuffer(const unsigned char* data, unsigned long length);
    void    swapFloatsFromBuffer(unsigned char* data, unsigned long length);
    void    swapDoublesToBuffer(const unsigned char* data, unsigned long length);
    void    swapDoublesFromBuffer(unsigned char* data, unsigned long length);
};

#pragma pack(pop)
#pragma enum(pop)

//
// ICountedStreamFrame
//
ICountedStreamFrame::ICountedStreamFrame(unsigned long frameDepth, unsigned long length, const char* data)
    :    fNext(0)
    ,    fDepth(frameDepth)
    ,    fBuffer(0)
    ,    fLength(length)
{
    fBuffer = new char[fLength];
    if (data != 0)
        memcpy(fBuffer, data, fLength);
}

ICountedStreamFrame::ICountedStreamFrame()
    :    fNext(0)
    ,    fDepth(0)
    ,    fBuffer(0)
    ,    fLength(0)
{
}

ICountedStreamFrame::ICountedStreamFrame(const ICountedStreamFrame& other)
    :    fNext(0)
    ,    fDepth(0)
    ,    fBuffer(0)
    ,    fLength(0)
{
}

ICountedStreamFrame::~ICountedStreamFrame()
{
    delete [] fBuffer;
}

ICountedStreamFrame*
ICountedStreamFrame::nextLink()
{
    if (fNext == 0)
        return 0;

    ITRACE_DEVELOP( IString("nextLink: frame @ ") + IString(long(this)) +
                    IString(": this count ") + IString(count()) +
                    IString(", next count ") + IString(fNext->count()));

    ICountedStreamFrame* result = fNext;
    if (fNext->count() == 1)
        fNext = 0;
    return result;
}

void
ICountedStreamFrame::append(ICountedStreamFrame* frame)
{
    fNext = frame;
    fNext->addRef();
}

void
ICountedStreamFrame::print()
{
    ITRACE_DEVELOP( IString("frame @ ") + IString((long)this) +
                    IString(": length ") + IString(fLength) +
                    IString(", depth ") + IString(fDepth) +
                    IString(", refs ") + IString(count()));

// Not sure how to best do this with the trace system
#if 0
    for (int i = 0; i < fDepth; i++)
        cerr << " ";
    for (int j = 0; j < fLength; j++)
        cerr << "0123456789abcdef"[(unsigned char)fBuffer[j] >> 4] << "0123456789abcdef"[(unsigned char)fBuffer[j] & 0xf];
#endif
}

void ICountedStreamFrame::printAll(unsigned long depth)
{
    for (ICountedStreamFrame* f = this; f != 0; f = f->fNext)
    {
        f->print();
        if (f->fNext && f->fNext->fDepth == depth)
            return;
    }
}

//
// ICountedStreamableContainer
//
ICountedStreamableContainer::ICountedStreamableContainer(ICountedStreamBuffer* streamBuffer, ICountedStreamFrame* firstFrame)
    :    fBuffer(streamBuffer)
    ,    fFrame(firstFrame)
{
    fFrame->addRef();
}

ICountedStreamableContainer::ICountedStreamableContainer()
    :    fBuffer(0)
    ,    fFrame(0)
{
}

ICountedStreamableContainer::ICountedStreamableContainer(const ICountedStreamableContainer& other)
    :    fBuffer(0)
    ,    fFrame(0)
{
}

ICountedStreamableContainer::~ICountedStreamableContainer()
{
    IFUNCTRACE_DEVELOP();
    if (fFrame == 0)
        return;

    ITRACE_DEVELOP(IString("Need code to drop one reference to each frame in container..."));
    printAll();
    ITRACE_DEVELOP(IString("here goes..."));

    ICountedStreamFrame* nextf;
    for (ICountedStreamFrame* f = fFrame; f != 0; f = nextf)
    {
        nextf = f->nextLink();
        f->removeRef();
    }
    ITRACE_DEVELOP(IString("Hope that worked."));
}

IMStreamable*
ICountedStreamableContainer::readAliasedObject(IMRefCounted*& refCounted, const IStreamFinder* finder, const ITypeRepresentation* typeRep)
{
    IFUNCTRACE_DEVELOP();
    fFrame->printAll();

    ICountedStreamFrame* frame = fFrame;
    fFrame = 0;
    return fBuffer->readAliasedObjectFromFrame(frame, refCounted, finder, typeRep);
}

const ITypeRepresentation*
ICountedStreamableContainer::readObjectType()
{
    IFUNCTRACE_DEVELOP();
    fFrame->print();
    return fBuffer->readObjectTypeFromFrame(fFrame);
}

void
ICountedStreamableContainer::printAll()
{
    IFUNCTRACE_DEVELOP();
    if (fFrame)
        fFrame->printAll(fFrame->depth());
}

//
// ICountedStreamBuffer
//
ICountedStreamBuffer::ICountedStreamBuffer(IStreamBuffer* streamBuffer)
    :    fStreamBuffer(streamBuffer)
    ,    fDataStream(0)
    ,    fStreamContext(0)
    ,    fStreamFinder(0)
    ,    fFrame(0)
    ,    fInternalBuffer(0)
    ,    fBuffer(0)
    ,    fPointer(0)
    ,    fEndPointer(0)
    ,    fFrameLengthPointer(0)
    ,    fFrameDepth(0)
    ,    fUsingContainer(false)
{
    fStreamBuffer->addRef();
}

ICountedStreamBuffer::ICountedStreamBuffer()
    :    fStreamBuffer(0)
    ,    fDataStream(0)
    ,    fStreamContext(0)
    ,    fStreamFinder(0)
    ,    fFrame(0)
    ,    fInternalBuffer(0)
    ,    fBuffer(0)
    ,    fPointer(0)
    ,    fEndPointer(0)
    ,    fFrameLengthPointer(0)
    ,    fFrameDepth(0)
    ,    fUsingContainer(false)
{
}

ICountedStreamBuffer::ICountedStreamBuffer(const ICountedStreamBuffer& other)
    :    fStreamBuffer(0)
    ,    fDataStream(0)
    ,    fStreamContext(0)
    ,    fStreamFinder(0)
    ,    fFrame(0)
    ,    fInternalBuffer(0)
    ,    fBuffer(0)
    ,    fPointer(0)
    ,    fEndPointer(0)
    ,    fFrameLengthPointer(0)
    ,    fFrameDepth(0)
    ,    fUsingContainer(false)
{
}

ICountedStreamBuffer::~ICountedStreamBuffer()
{
    IFUNCTRACE_DEVELOP();

    delete [] fInternalBuffer;
    delete fStreamContext;
    if (fFrame)
    {
        ICountedStreamFrame* nextf;
        for (ICountedStreamFrame* f = fFrame; f != 0; f = nextf)
        {
            nextf = f->nextLink();
            ITRACE_DEVELOP(IString(f->count()) + IString(" references"));
            f->removeRef();
        }
    }
    if (fStreamBuffer)
        fStreamBuffer->removeRef();
}

void ICountedStreamBuffer::setStream(IDataStream* dataStream)
{
    fDataStream = dataStream;
}

void ICountedStreamBuffer::reset()
{
    fStreamBuffer->reset();
    if (fStreamContext)
        fStreamContext->reset();
}

void ICountedStreamBuffer::flush()
{
    fStreamBuffer->flush();
}

bool ICountedStreamBuffer::establishContext()
{
    if (!fStreamContext)
        fStreamContext = new IStreamContext();
    return fStreamContext->establish();
}

void ICountedStreamBuffer::clearContext()
{
    fStreamContext->clear();
}

const size_t kCountedBufferSize = 8192;
//const size_t kCountedBufferSize = 100;

#if defined(__WINDOWS__) || defined(__OS2__) || defined(_WIN32)

#define kLittleEndian    1

#elif defined(__MWERKS__) || defined(_AIX)

#define kBigEndian    1

#else

#error "define the byte order for this platform."

#endif

#if defined(kLittleEndian)

inline void
ICountedStreamBuffer::swapBytesToBuffer(const unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();
    const unsigned char* value_ptr = data + length;
    while (length--)
        *fPointer++ = *(--value_ptr);
    if (fFrameDepth == 0)
        flushBuffer();
}

inline void
ICountedStreamBuffer::swapBytesFromBuffer(unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        fillBuffer(length);
    unsigned char* value_ptr = data + length;
    while (length--)
        *(--value_ptr) = *fPointer++;
}

inline void
ICountedStreamBuffer::swapShortsToBuffer(const unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();

    while (length != 0)
    {
        *fPointer++ = data[1];
        *fPointer++ = data[0];
        data += 2;
        length -= 2;
    }
    if (fFrameDepth == 0)
        flushBuffer();
}

inline void
ICountedStreamBuffer::swapShortsFromBuffer(unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        fillBuffer(length);

    while (length != 0)
    {
        *data++ = fPointer[1];
        *data++ = fPointer[0];
        fPointer += 2;
        length -= 2;
    }
}

inline void
ICountedStreamBuffer::swapLongsToBuffer(const unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();

    while (length != 0)
    {
        *fPointer++ = data[3];
        *fPointer++ = data[2];
        *fPointer++ = data[1];
        *fPointer++ = data[0];
        data += 4;
        length -= 4;
    }
    if (fFrameDepth == 0)
        flushBuffer();
}

inline void
ICountedStreamBuffer::swapLongsFromBuffer(unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        fillBuffer(length);

    while (length != 0)
    {
        *data++ = fPointer[3];
        *data++ = fPointer[2];
        *data++ = fPointer[1];
        *data++ = fPointer[0];
        fPointer += 4;
        length -= 4;
    }
}

inline void
ICountedStreamBuffer::swapFloatsToBuffer(const unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();

    while (length != 0)
    {
        *fPointer++ = data[3];
        *fPointer++ = data[2];
        *fPointer++ = data[1];
        *fPointer++ = data[0];
        data += 4;
        length -= 4;
    }
    if (fFrameDepth == 0)
        flushBuffer();
}

inline void
ICountedStreamBuffer::swapFloatsFromBuffer(unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        fillBuffer(length);

    while (length != 0)
    {
        *data++ = fPointer[3];
        *data++ = fPointer[2];
        *data++ = fPointer[1];
        *data++ = fPointer[0];
        fPointer += 4;
        length -= 4;
    }
}

inline void
ICountedStreamBuffer::swapDoublesToBuffer(const unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();

    while (length != 0)
    {
        *fPointer++ = data[7];
        *fPointer++ = data[6];
        *fPointer++ = data[5];
        *fPointer++ = data[4];
        *fPointer++ = data[3];
        *fPointer++ = data[2];
        *fPointer++ = data[1];
        *fPointer++ = data[0];
        data += 8;
        length -= 8;
    }
    if (fFrameDepth == 0)
        flushBuffer();
}

inline void
ICountedStreamBuffer::swapDoublesFromBuffer(unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        fillBuffer(length);

    while (length != 0)
    {
        *data++ = fPointer[7];
        *data++ = fPointer[6];
        *data++ = fPointer[5];
        *data++ = fPointer[4];
        *data++ = fPointer[3];
        *data++ = fPointer[2];
        *data++ = fPointer[1];
        *data++ = fPointer[0];
        fPointer += 8;
        length -= 8;
    }
}

#elif defined(kBigEndian)

inline void
ICountedStreamBuffer::swapBytesToBuffer(const unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();
    while (length--)
        *fPointer++ = *data++;
    if (fFrameDepth == 0)
        flushBuffer();
}

inline void
ICountedStreamBuffer::swapBytesFromBuffer(unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        fillBuffer(length);
    while (length--)
        *data++ = *fPointer++;
}

inline void
ICountedStreamBuffer::swapShortsToBuffer(const unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();
    memcpy(fPointer, data, (size_t)length);
    fPointer += length;
    if (fFrameDepth == 0)
        flushBuffer();
}

inline void
ICountedStreamBuffer::swapShortsFromBuffer(unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        fillBuffer(length);
    memcpy(data, fPointer, (size_t)length);
    fPointer += length;
}

inline void
ICountedStreamBuffer::swapLongsToBuffer(const unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();
    memcpy(fPointer, data, (size_t)length);
    fPointer += length;
    if (fFrameDepth == 0)
        flushBuffer();
}

inline void
ICountedStreamBuffer::swapLongsFromBuffer(unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        fillBuffer(length);
    memcpy(data, fPointer, (size_t)length);
    fPointer += length;
}

inline void
ICountedStreamBuffer::swapFloatsToBuffer(const unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();
    memcpy(fPointer, data, (size_t)length);
    fPointer += length;
    if (fFrameDepth == 0)
        flushBuffer();
}

inline void
ICountedStreamBuffer::swapFloatsFromBuffer(unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        fillBuffer(length);
    memcpy(data, fPointer, (size_t)length);
    fPointer += length;
}

inline void
ICountedStreamBuffer::swapDoublesToBuffer(const unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();
    memcpy(fPointer, data, (size_t)length);
    fPointer += length;
    if (fFrameDepth == 0)
        flushBuffer();
}

inline void
ICountedStreamBuffer::swapDoublesFromBuffer(unsigned char* data, unsigned long length)
{
    if (fPointer + length > fEndPointer)
        fillBuffer(length);
    memcpy(data, fPointer, (size_t)length);
    fPointer += length;
}

#else

#error "write code to swap primitive types"

#endif

void
ICountedStreamBuffer::writeByte(char value)
{
    if (fPointer == fEndPointer)
        updateAndWriteBuffer();
    *fPointer++ = value;
    if (fFrameDepth == 0)
        flushBuffer();
}

void ICountedStreamBuffer::writeShort(short value)
{
    swapBytesToBuffer((unsigned char*)&value, sizeof(value));
}

void ICountedStreamBuffer::writeLong(long value)
{
    swapBytesToBuffer((unsigned char*)&value, sizeof(value));
}

void ICountedStreamBuffer::writeFloat(float value)
{
    swapBytesToBuffer((unsigned char*)&value, sizeof(value));
}

void ICountedStreamBuffer::writeDouble(double value)
{
    swapBytesToBuffer((unsigned char*)&value, sizeof(value));
}

char ICountedStreamBuffer::readByte()
{
    unsigned char value;
    if (fPointer == fEndPointer)
        fillBuffer(1);
    value = *fPointer++;
    return (char) value;
}

short ICountedStreamBuffer::readShort()
{
    unsigned short value;
    swapBytesFromBuffer((unsigned char*)&value, sizeof(value));
    return (short)value;
}

long ICountedStreamBuffer::readLong()
{
    unsigned long value;
    swapBytesFromBuffer((unsigned char*)&value, sizeof(value));
    return (long)value;
}

float ICountedStreamBuffer::readFloat()
{
    float value;
    swapBytesFromBuffer((unsigned char*)&value, sizeof(value));
    return value;
}

double ICountedStreamBuffer::readDouble()
{
    double value;
    swapBytesFromBuffer((unsigned char*)&value, sizeof(value));
    return value;
}

void ICountedStreamBuffer::writeBytes(const char* data, unsigned long howMany)
{
    if (howMany == 0)
        return;

    if (fFrameDepth == 0)
    {
        fStreamBuffer->writeBytes(data, howMany);
    }
    else if (howMany <= kCountedBufferSize - 3)
    {
        if (fPointer + howMany > fEndPointer)
            updateAndWriteBuffer();
        memcpy(fPointer, data, (size_t)howMany);
        fPointer += howMany;
        if (fFrameDepth == 0)
            flushBuffer();
    }
     else
    {
        writeEndOfCurrentFrame(false);
        if (fPointer == fEndPointer)
        {
            flushBuffer();
            fEndPointer += kCountedBufferSize;
        }
        writeByte(kDataBufferTag);
        writeLong(howMany);
        flushBuffer();
        fStreamBuffer->writeBytes(data, howMany);
    }
}

void
ICountedStreamBuffer::readBytes(char* data, unsigned long howMany)
{
    if (howMany == 0)
        return;

    if (fFrameDepth == 0)
    {
        fStreamBuffer->readBytes(data, howMany);
    }
    else if (howMany <= kCountedBufferSize - 3)
    {
        if (fPointer + howMany > fEndPointer)
            fillBuffer(howMany);
        memcpy(data, fPointer, (size_t)howMany);
        fPointer += howMany;
    }
    else
    {
        unsigned char tag = readByte();
        if (tag != kDataBufferTag)
            invalidFormat();
        unsigned long lengthFromStream = fEndPointer - fPointer;
        if (lengthFromStream != howMany)
            invalidFormat();
        memcpy(data, fPointer, (size_t)howMany);
        fPointer += howMany;
    }
}

void
ICountedStreamBuffer::writeShorts(const short* data, unsigned long howMany)
{
    if (howMany == 0)
        return;

    unsigned long length = howMany * sizeof(short);
    if (fFrameDepth == 0)
    {
        while (length > kCountedBufferSize)
        {
            swapShortsToBuffer((const unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
        }
        swapShortsToBuffer((const unsigned char*)data, length);
    }
    else if (length <= kCountedBufferSize - 3)
    {
        if (fPointer + length > fEndPointer)
            updateAndWriteBuffer();
        swapShortsToBuffer((const unsigned char*)data, length);
    }
    else
    {
        writeEndOfCurrentFrame(false);
        if (fPointer == fEndPointer)
        {
            flushBuffer();
            fEndPointer += kCountedBufferSize;
        }
        writeByte(kDataBufferTag);
        writeLong(length);
        flushBuffer();
        fEndPointer += kCountedBufferSize;
        while (length > kCountedBufferSize)
        {
            swapShortsToBuffer((const unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
            flushBuffer();
            fEndPointer += kCountedBufferSize;
        }
        swapShortsToBuffer((const unsigned char*)data, length);
    }
}

void ICountedStreamBuffer::readShorts(short* data, unsigned long howMany)
{
    if (howMany == 0)
        return;

    unsigned long length = howMany * sizeof(short);
    if (fFrameDepth == 0)
    {
        while (length > kCountedBufferSize)
        {
            swapShortsFromBuffer((unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
        }
        swapShortsFromBuffer((unsigned char*)data, length);
    }
    else if (length <= kCountedBufferSize - 3)
    {
        if (fPointer + length > fEndPointer)
            fillBuffer(length);
        swapShortsFromBuffer((unsigned char*)data, length);
    }
    else
    {
        unsigned char tag = readByte();
        if (tag != kDataBufferTag)
            invalidFormat();
        unsigned long lengthFromStream = fEndPointer - fPointer;
        if (lengthFromStream != length)
            invalidFormat();
        swapShortsFromBuffer((unsigned char*)data, length);
    }
}

void ICountedStreamBuffer::writeLongs(const long* data, unsigned long howMany)
{
    if (howMany == 0)
        return;

    unsigned long length = howMany * sizeof(long);
    if (fFrameDepth == 0)
    {
        while (length > kCountedBufferSize)
        {
            swapLongsToBuffer((const unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
        }
        swapLongsToBuffer((const unsigned char*)data, length);
    }
    else if (length <= kCountedBufferSize - 3)
    {
        if (fPointer + length > fEndPointer)
            updateAndWriteBuffer();
        swapLongsToBuffer((const unsigned char*)data, length);
    }
    else
    {
        writeEndOfCurrentFrame(false);
        if (fPointer == fEndPointer)
        {
            flushBuffer();
            fEndPointer += kCountedBufferSize;
        }
        writeByte(kDataBufferTag);
        writeLong(length);
        flushBuffer();
        fEndPointer += kCountedBufferSize;
        while (length > kCountedBufferSize)
        {
            swapLongsToBuffer((const unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
            flushBuffer();
            fEndPointer += kCountedBufferSize;
        }
        swapLongsToBuffer((const unsigned char*)data, length);
    }
}

void ICountedStreamBuffer::readLongs(long* data, unsigned long howMany)
{
    if (howMany == 0)
        return;

    unsigned long length = howMany * sizeof(long);
    if (fFrameDepth == 0)
    {
        while (length > kCountedBufferSize)
        {
            swapLongsFromBuffer((unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
        }
        swapLongsFromBuffer((unsigned char*)data, length);
    }
    else if (length <= kCountedBufferSize - 3)
    {
        if (fPointer + length > fEndPointer)
            fillBuffer(length);
        swapLongsFromBuffer((unsigned char*)data, length);
    }
    else
    {
        unsigned char tag = readByte();
        if (tag != kDataBufferTag)
            invalidFormat();
        unsigned long lengthFromStream = fEndPointer - fPointer;
        if (lengthFromStream != length)
            invalidFormat();
        swapLongsFromBuffer((unsigned char*)data, length);
    }
}

void ICountedStreamBuffer::writeFloats(const float* data, unsigned long howMany)
{
    if (howMany == 0)
        return;

    unsigned long length = howMany * sizeof(float);
    if (fFrameDepth == 0)
    {
        while (length > kCountedBufferSize)
        {
            swapFloatsToBuffer((const unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
        }
        swapFloatsToBuffer((const unsigned char*)data, length);
    }
    else if (length <= kCountedBufferSize - 3)
    {
        if (fPointer + length > fEndPointer)
            updateAndWriteBuffer();
        swapFloatsToBuffer((const unsigned char*)data, length);
    }
    else
    {
        writeEndOfCurrentFrame(false);
        if (fPointer == fEndPointer)
        {
            flushBuffer();
            fEndPointer += kCountedBufferSize;
        }
        writeByte(kDataBufferTag);
        writeLong(length);
        flushBuffer();
        fEndPointer += kCountedBufferSize;
        while (length > kCountedBufferSize)
        {
            swapFloatsToBuffer((const unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
            flushBuffer();
            fEndPointer += kCountedBufferSize;
        }
        swapFloatsToBuffer((const unsigned char*)data, length);
    }
}

void ICountedStreamBuffer::readFloats(float* data, unsigned long howMany)
{
    if (howMany == 0)
        return;

    unsigned long length = howMany * sizeof(float);
    if (fFrameDepth == 0)
    {
        while (length > kCountedBufferSize)
        {
            swapFloatsFromBuffer((unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
        }
        swapFloatsFromBuffer((unsigned char*)data, length);
    }
    else if (length <= kCountedBufferSize - 3)
    {
        if (fPointer + length > fEndPointer)
            fillBuffer(length);
        swapFloatsFromBuffer((unsigned char*)data, length);
    }
    else
    {
        unsigned char tag = readByte();
        if (tag != kDataBufferTag)
            invalidFormat();
        unsigned long lengthFromStream = fEndPointer - fPointer;
        if (lengthFromStream != length)
            invalidFormat();
        swapFloatsFromBuffer((unsigned char*)data, length);
    }
}

void ICountedStreamBuffer::writeDoubles(const double* data, unsigned long howMany)
{
    if (howMany == 0)
        return;

    unsigned long length = howMany * sizeof(double);
    if (fFrameDepth == 0)
    {
        while (length > kCountedBufferSize)
        {
            swapDoublesToBuffer((const unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
        }
        swapDoublesToBuffer((const unsigned char*)data, length);
    }
     else if (length <= kCountedBufferSize - 3)
    {
        if (fPointer + length > fEndPointer)
            updateAndWriteBuffer();
        swapDoublesToBuffer((const unsigned char*)data, length);
    }
    else
    {
        writeEndOfCurrentFrame(false);
        if (fPointer == fEndPointer)
        {
            flushBuffer();
            fEndPointer += kCountedBufferSize;
        }
        writeByte(kDataBufferTag);
        writeLong(length);
        flushBuffer();
        fEndPointer += kCountedBufferSize;
        while (length > kCountedBufferSize)
        {
            swapDoublesToBuffer((const unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
            flushBuffer();
            fEndPointer += kCountedBufferSize;
        }
        swapDoublesToBuffer((const unsigned char*)data, length);
    }
}

void ICountedStreamBuffer::readDoubles(double* data, unsigned long howMany)
{
    if (howMany == 0)
        return;

    unsigned long length = howMany * sizeof(double);
    if (fFrameDepth == 0)
    {
        while (length > kCountedBufferSize)
        {
            swapDoublesFromBuffer((unsigned char*)data, kCountedBufferSize);
            data += kCountedBufferSize;
            length -= kCountedBufferSize;
        }
        swapDoublesFromBuffer((unsigned char*)data, length);
    }
    else if (length <= kCountedBufferSize - 3)
    {
        if (fPointer + length > fEndPointer)
            fillBuffer(length);
        swapDoublesFromBuffer((unsigned char*)data, length);
    }
    else
    {
        unsigned char tag = readByte();
        if (tag != kDataBufferTag)
            invalidFormat();
        unsigned long lengthFromStream = fEndPointer - fPointer;
        if (lengthFromStream != length)
            invalidFormat();
        swapDoublesFromBuffer((unsigned char*)data, length);
    }
}

void ICountedStreamBuffer::writeCompressedLong(unsigned long value)
{
    if (value < 0x80)
    {
        if (fPointer == fEndPointer)
            updateAndWriteBuffer();
        *fPointer++ = (unsigned char)value;
    }
    else if (value < 0x4000)
    {
        if (fPointer + 2 > fEndPointer)
            updateAndWriteBuffer();
        *fPointer++ = (unsigned char)(0x80 | (value >> 7));
        *fPointer++ = (unsigned char)(value & 0x7f);
    }
    else if (value < 0x40000000)
    {
        if (fPointer + 4 > fEndPointer)
            updateAndWriteBuffer();
        *fPointer++ = (unsigned char)(0x80 | (value >> 23));
        *fPointer++ = (unsigned char)(0x80 | ((value >> 16) & 0x7f));
        *fPointer++ = (unsigned char)((value >> 8) & 0xff);
        *fPointer++ = (unsigned char)(value & 0xff);
    }
    else
    {
        // value out of range.
    }
    if (fFrameDepth == 0)
        flushBuffer();
}

unsigned long ICountedStreamBuffer::readCompressedLong()
{
    unsigned char byte1;
    if (fPointer == fEndPointer)
        fillBuffer(1);

    byte1 = *fPointer++;
    unsigned long result;
    if (byte1 & 0x80)
    {
        byte1 &= 0x7f;
        unsigned char byte2;
        if (fPointer == fEndPointer)
            fillBuffer(1);
        byte2 = *fPointer++;
        if (byte2 & 0x80)
        {
            byte2 &= 0x7f;
            unsigned char byte3, byte4;
            if (fPointer + 2 > fEndPointer)
                fillBuffer(2);
            byte3 = *fPointer++;
            byte4 = *fPointer++;
            result = (byte1 << 23) | (byte2 << 16) | (byte3 << 8) | byte4;
        }
        else
        {
            result = (byte1 << 7) | byte2;
        }
    }
    else
    {
        result = byte1;
    }
    return result;
}

void ICountedStreamBuffer::writeContextNumber(unsigned long value)
{
    writeCompressedLong(value);
}

unsigned long ICountedStreamBuffer::readContextNumber()
{
    return readCompressedLong();
}

void ICountedStreamBuffer::writeTypeName(const ITypeName& typeName)
{
    const char* formattedTypeName = typeName.formattedName();
    unsigned long length = (unsigned long)strlen(formattedTypeName);
    writeCompressedLong(length);
    if (fPointer + length > fEndPointer)
        updateAndWriteBuffer();
    memcpy(fPointer, formattedTypeName, (size_t)length);
    fPointer += length;
    if (fFrameDepth == 0)
        flushBuffer();
}

void ICountedStreamBuffer::readTypeName(ITypeName& typeName)
{
    unsigned long length = readCompressedLong();
    if (fPointer + length > fEndPointer)
        fillBuffer(length);
    char stackBuffer[128];
    char* formattedTypeName;
    if (length < 128)
        formattedTypeName = stackBuffer;
    else
        formattedTypeName = new char[length];
    memcpy(formattedTypeName, fPointer, (size_t)length);
    fPointer += length;
    formattedTypeName[length] = 0;
    typeName = ITypeName(formattedTypeName);
    if (formattedTypeName != stackBuffer)
        delete [] formattedTypeName;
}

void ICountedStreamBuffer::flushBuffer()
{
    fStreamBuffer->writeBytes((char*)fBuffer, fPointer - fBuffer);
    fPointer = fBuffer;
    fEndPointer = fBuffer;
    fFrameLengthPointer = 0;
}

void ICountedStreamBuffer::fillBuffer(unsigned long length, bool handleChangeBuffer)
{
    if (fFrame == 0)
    {
        if (fInternalBuffer == 0)
            fInternalBuffer = new unsigned char[kCountedBufferSize];
        fBuffer = fInternalBuffer;
        fPointer = fBuffer;
        fEndPointer = fPointer + length;
        fStreamBuffer->readBytes((char*)fPointer, fEndPointer - fPointer);
    }
    else
    {
        ICountedStreamFrame* frame = fFrame;
        fFrame = frame->nextLink();
        frame->removeRef();
        fBuffer = (unsigned char*)fFrame->buffer();
        fPointer = fBuffer;
        fEndPointer = fBuffer + fFrame->length();
        if (handleChangeBuffer && *fPointer == kChangeBufferTag)
            fPointer++;
    }
}

void ICountedStreamBuffer::writeStartOfNewFrame(unsigned char tag)
{
    if (fFrameDepth == 0)
    {
        updateAndWriteBuffer();
    }
    else if (fEndPointer == fBuffer)
    {
        fEndPointer += kCountedBufferSize;
    }
    else
    {
        unsigned short length = 1;
        switch (tag)
        {
            case kEndFrameTag:
            case kEndObjectTag:
                if (fFrameDepth > 1)
                    length += 2;
                break;
            case kNilObjectTag:
            case kAliasedObjectTag:
            case kTypedAliasedObjectTag:
            case kStartFrameTag:
            case kStartAliasedTypeObjectTag:
            case kStartTypedObjectTag:
            case kStartUntypedObjectTag:
            case kChangeBufferTag:
                length += 2;
                break;
        }

        if (fPointer + length > fEndPointer)
        {
            writeEndOfCurrentFrame(true);
            fEndPointer = fBuffer + kCountedBufferSize;
        }
    }
    *fPointer++ = tag;
    switch (tag)
    {
        case kNilObjectTag:
        case kAliasedObjectTag:
        case kTypedAliasedObjectTag:
            if (fFrameDepth == 0)
            {
                flushBuffer();
                return;
            }
            break;
        case kEndFrameTag:
        case kEndObjectTag:
            fFrameDepth--;
            if (fFrameDepth == 0)
            {
                flushBuffer();
                return;
            }
            break;
        case kStartFrameTag:
        case kStartAliasedTypeObjectTag:
        case kStartTypedObjectTag:
        case kStartUntypedObjectTag:
            fFrameDepth++;
            break;
    }
    fFrameLengthPointer = fPointer;
    fPointer += sizeof(unsigned short);
}

void ICountedStreamBuffer::writeEndOfCurrentFrame(bool doFlush)
{
    if (fFrameDepth == 0)
        return;
    if (fFrameLengthPointer == 0)
        return;
    unsigned char* savedPointer = fPointer;
    fPointer = fFrameLengthPointer;
    unsigned short length = (unsigned short)(savedPointer - (fPointer + sizeof(short)));
    swapBytesToBuffer((unsigned char*)&length, sizeof(length));
    fPointer = savedPointer;
    if (doFlush)
        flushBuffer();
}

void ICountedStreamBuffer::writeStartOfFrame()
{
    writeEndOfCurrentFrame(false);
    writeStartOfNewFrame(kStartFrameTag);
}

void ICountedStreamBuffer::writeEndOfFrame()
{
    writeEndOfCurrentFrame(false);
    writeStartOfNewFrame(kEndFrameTag);
}

void ICountedStreamBuffer::writeStartOfObject(unsigned char tag)
{
    writeStartOfNewFrame(tag);
}

void ICountedStreamBuffer::writeEndOfObject()
{
    writeStartOfNewFrame(kEndObjectTag);
}

void ICountedStreamBuffer::updateAndWriteBuffer()
{
    if (fFrameDepth == 0)
    {
        if (fInternalBuffer == 0)
            fInternalBuffer = new unsigned char[kCountedBufferSize];
        fBuffer = fInternalBuffer;
        fPointer = fBuffer;
        fEndPointer = fBuffer + kCountedBufferSize;
    }
    else
    {
        if (fEndPointer > fBuffer)
            writeEndOfCurrentFrame(true);
        writeStartOfNewFrame(kChangeBufferTag);
    }
}

unsigned char ICountedStreamBuffer::readAllFrames(bool isObjectTag)
{
    if (fPointer == fEndPointer)
        fillBuffer(1);

    unsigned char tag = *fPointer++;
    if (isObjectTag)
    {
        switch (tag)
        {
            case kNilObjectTag:
            case kAliasedObjectTag:
            case kTypedAliasedObjectTag:
                return tag;
            case kStartAliasedTypeObjectTag:
            case kStartTypedObjectTag:
            case kStartUntypedObjectTag:
                break;
            default:
                invalidFormat();
        }
    }
     else if (tag != kStartFrameTag)
    {
        invalidFormat();
    }
    unsigned long frameDepth = fFrameDepth;
    fFrameDepth++;
    fEndPointer += sizeof(unsigned short);
    fStreamBuffer->readBytes((char*)fPointer, fEndPointer - fPointer);
    unsigned long length = (unsigned short)readShort();
    ICountedStreamFrame* frame = new ICountedStreamFrame(frameDepth, length + 1);
    char* buf = frame->buffer();
    *buf++ = tag;
    fStreamBuffer->readBytes(buf, length);
    ICountedStreamFrame* firstFrame = frame;
    ICountedStreamFrame* currentFrame = frame;
    while (1)
    {
        fillBuffer(1);
        tag = *fPointer++;
        frameDepth = fFrameDepth;
        switch (tag)
        {
            case kStartAliasedTypeObjectTag:
            case kStartTypedObjectTag:
            case kStartUntypedObjectTag:
            case kStartFrameTag:
            {
                fFrameDepth++;
                fEndPointer += sizeof(unsigned short);
                fStreamBuffer->readBytes((char*)fPointer, fEndPointer - fPointer);
                length = (unsigned short)readShort();
                break;
            }
            case kEndObjectTag:
            case kEndFrameTag:
            {
                fFrameDepth--;
                if (fFrameDepth == 0)
                {
                    frame = new ICountedStreamFrame(frameDepth, 1);
                    unsigned char* buf = (unsigned char*)frame->buffer();
                    *buf = (unsigned char)tag;
                    currentFrame->append(frame);
                    currentFrame = frame;
                    fFrame = firstFrame;
                    fFrame->addRef();
                    fBuffer = (unsigned char*)fFrame->buffer();
                    fPointer = fBuffer;
                    tag = *fPointer++;
                    fEndPointer = fBuffer + fFrame->length();
                    return tag;
                }
                fEndPointer += sizeof(unsigned short);
                fStreamBuffer->readBytes((char*)fPointer, fEndPointer - fPointer);
                length = (unsigned short)readShort();
                break;
            }
            case kDataBufferTag:
            {
                fEndPointer += sizeof(unsigned long);
                fStreamBuffer->readBytes((char*)fPointer, fEndPointer - fPointer);
                length = (unsigned long)readLong();
                break;
            }
            case kNilObjectTag:
            case kAliasedObjectTag:
            case kTypedAliasedObjectTag:
            case kChangeBufferTag:
            {
                fEndPointer += sizeof(unsigned short);
                fStreamBuffer->readBytes((char*)fPointer, fEndPointer - fPointer);
                length = (unsigned short)readShort();
                break;
            }
            default:
                invalidFormat();
                break;
        }
        frame = new ICountedStreamFrame(frameDepth, length + 1);
        buf = frame->buffer();
        *buf++ = tag;
        fStreamBuffer->readBytes(buf, length);
        currentFrame->append(frame);
        currentFrame = frame;
    }

    #ifdef _MSC_VER
    return ('\0'); // make MSVC happy   .. ARUN
    #endif
}

unsigned char ICountedStreamBuffer::readStartOfObject()
{
    unsigned char objectTag;
    if (fFrameDepth == 0)
    {
        objectTag = readAllFrames(true);
        fFrame->printAll();
    }
    else
    {
        if (fPointer == fEndPointer)
            fillBuffer(1, false);
        objectTag = *fPointer++;
    }

    switch (objectTag)
    {
        case kNilObjectTag:
        case kAliasedObjectTag:
        case kTypedAliasedObjectTag:
            break;
        case kStartAliasedTypeObjectTag:
        case kStartTypedObjectTag:
        case kStartUntypedObjectTag:
            fFrameDepth++;
            break;
        default:
            invalidFormat();
    }
    return objectTag;
}

void ICountedStreamBuffer::readStartOfFrame()
{
    unsigned char tag;
    if (fFrameDepth == 0)
    {
        tag = readAllFrames(false);
        fFrame->printAll();
    }
    else
    {
        if (fPointer == fEndPointer)
            fillBuffer(1, false);
        tag = *fPointer++;
    }
    if (tag != kStartFrameTag)
        invalidFormat();
    fFrameDepth++;
}

void ICountedStreamBuffer::readEndOfObject()
{
    IFUNCTRACE_DEVELOP();
    if (fPointer == fEndPointer)
        fillBuffer(1);
    unsigned char tag = *fPointer++;

    if (tag != kEndObjectTag)
        invalidFormat();

    fFrameDepth--;
    if (fFrameDepth == 0)
    {
        ICountedStreamFrame* frame = fFrame;
        if (frame == 0)
        {
            ITRACE_DEVELOP(IString("Frame is nil..."));
            invalidFormat();
        }
        fFrame = frame->next();

        ITRACE_DEVELOP(IString("Current count ") + IString(frame->count()));
        frame->removeRef();
        if (fFrame != 0)
        {
            ITRACE_DEVELOP( IString("Frame should be nil...  frame count ") +
                            IString(fFrame->count()));
            invalidFormat();
        }
    }
}

void ICountedStreamBuffer::readEndOfFrame()
{
    IFUNCTRACE_DEVELOP();
    if (fPointer < fEndPointer)
    {
        skipRemainderOfFrame(0);
    }
    else
    {
        if (fPointer == fEndPointer)
            fillBuffer(1, false);
        unsigned char tag = *fPointer++;
        if (tag != kEndFrameTag)
            skipRemainderOfFrame(tag);
         else
            fFrameDepth--;
    }

    if (fFrameDepth == 0)
    {
        ICountedStreamFrame* frame = fFrame;
        if (frame == 0)
        {
            ITRACE_DEVELOP(IString("Frame is nil..."));
            invalidFormat();
        }
        fFrame = frame->next();

        ITRACE_DEVELOP(IString("Current count ") + IString(frame->count()));
        frame->removeRef();
        if (fFrame != 0)
        {
            ITRACE_DEVELOP( IString("Frame should be nil...  frame count ") +
                            IString(fFrame->count()));
            invalidFormat();
        }
    }
}

bool ICountedStreamBuffer::atEndOfFrame()
{
    if (fPointer < fEndPointer)
        return false;
    ICountedStreamFrame* frame = fFrame->next();
    return (*((unsigned char*)frame->buffer()) == kEndFrameTag);
}

void ICountedStreamBuffer::skipRemainderOfFrame(unsigned char currentTag)
{
    long depth = fFrameDepth;
#ifdef DEBUG_STREAMIO
    long frameDepth = 0;
#endif
    long nestingDepth = 0;
    if (currentTag != 0)
    {
        switch (currentTag)
        {
            case kStartAliasedTypeObjectTag:
            case kStartUntypedObjectTag:
            case kStartTypedObjectTag:
            {
#ifdef DEBUG_STREAMIO
                if (nestingDepth == 0)
                    frameDepth = fFrameDepth;
#endif
                nestingDepth++;
                fFrameDepth++;
                ICountedStreamableContainer* container = new ICountedStreamableContainer(this, fFrame);
                fStreamContext->addObjectContainer(container, currentTag == kStartTypedObjectTag);
                break;
            }
            case kStartFrameTag:
            {
                fFrameDepth++;
                break;
            }
            case kNilObjectTag:
            case kAliasedObjectTag:
            case kTypedAliasedObjectTag:
            case kDataBufferTag:
            case kChangeBufferTag:
                break;
            default:
                invalidFormat();
                break;
        }
    }

    while (1)
    {
        ICountedStreamFrame* nextFrame = fFrame->next();
        for (long d = 0; d < nestingDepth; d++)
            nextFrame->addRef();
        fillBuffer(1, false);
        unsigned char tag = *fPointer++;
        switch (tag)
        {
            case kStartAliasedTypeObjectTag:
            case kStartUntypedObjectTag:
            case kStartTypedObjectTag:
            {
#ifdef DEBUG_STREAMIO
                if (nestingDepth == 0)
                    frameDepth = fFrameDepth;
#endif
                nestingDepth++;
                fFrameDepth++;
                ICountedStreamableContainer* container = new ICountedStreamableContainer(this, fFrame);
                fStreamContext->addObjectContainer(container, currentTag == kStartTypedObjectTag);
                break;
            }
            case kEndObjectTag:
            {
                fFrameDepth--;
                if (nestingDepth != 0)
                {
                    nestingDepth--;
                    if (nestingDepth == 0)
                    {
                        ITRACE_DEVELOP(IString("frame depth ") + IString(fFrameDepth));
                    }
                }
                break;
            }
            case kStartFrameTag:
            {
                fFrameDepth++;
                break;
            }
            case kEndFrameTag:
            {
                if (fFrameDepth-- == depth)
                    return;
                break;
            }
            case kNilObjectTag:
            case kAliasedObjectTag:
            case kTypedAliasedObjectTag:
            case kDataBufferTag:
            case kChangeBufferTag:
                break;
            default:
                invalidFormat();
                break;
        }
    }
}

IMStreamable*
ICountedStreamBuffer::readAliasedObjectFromFrame(ICountedStreamFrame* frame, IMRefCounted*& refCounted, const IStreamFinder* finder, const ITypeRepresentation* typeRep)
{
    ICountedStreamFrame* saveFrame = fFrame;
    long saveFrameDepth = fFrameDepth;
    unsigned char* saveBuffer = fBuffer;
    unsigned char* savePointer = fPointer;
    unsigned char* saveEndPointer = fEndPointer;
    bool saveUsingContainer = fUsingContainer;
    fFrame = frame;
    fFrameDepth = frame->depth();
    fBuffer = (unsigned char*)frame->buffer();
    fPointer = fBuffer;
    fEndPointer = fBuffer + frame->length();
    fUsingContainer = true;
    IMStreamable* streamable;
    readAliasedObject(streamable, refCounted, finder, typeRep);
    refCounted->removeRef();
    fFrame->removeRef();
    fFrame = saveFrame;
    fFrameDepth = saveFrameDepth;
    fBuffer = saveBuffer;
    fPointer = savePointer;
    fEndPointer = saveEndPointer;
    fUsingContainer = saveUsingContainer;
    return streamable;
}

const ITypeRepresentation*
ICountedStreamBuffer::readObjectTypeFromFrame(ICountedStreamFrame* frame)
{
    ICountedStreamFrame* saveFrame = fFrame;
    long saveFrameDepth = fFrameDepth;
    unsigned char* saveBuffer = fBuffer;
    unsigned char* savePointer = fPointer;
    unsigned char* saveEndPointer = fEndPointer;
    bool saveUsingContainer = fUsingContainer;
    fFrame = frame;
    fFrameDepth = frame->depth();
    fBuffer = (unsigned char*)frame->buffer();
    fPointer = fBuffer;
    fEndPointer = fBuffer + frame->length();
    fUsingContainer = true;
    const ITypeRepresentation* result = readObjectType();
    fFrame = saveFrame;
    fFrameDepth = saveFrameDepth;
    fBuffer = saveBuffer;
    fPointer = savePointer;
    fEndPointer = saveEndPointer;
    fUsingContainer = saveUsingContainer;
    return result;
}

IMStreamable*
ICountedStreamBuffer::readAliasedObjectAgain(IMRefCounted*& refCounted)
{
    IMStreamable* returnValue = 0;
    if (fUsingContainer && (returnValue = fStreamContext->nextObject(refCounted)) != 0)
    {
        long depth = fFrameDepth;
        while (1)
        {
            fillBuffer(1, false);
            unsigned char tag = *fPointer++;
            switch (tag)
            {
                case kStartAliasedTypeObjectTag:
                case kStartUntypedObjectTag:
                case kStartTypedObjectTag:
                case kStartFrameTag:
                {
                    fFrameDepth++;
                    break;
                }
                case kEndObjectTag:
                {
                    if (fFrameDepth-- == depth)
                        return returnValue;
                    break;
                }
                case kEndFrameTag:
                {
                    fFrameDepth--;
                    break;
                }
                case kNilObjectTag:
                case kAliasedObjectTag:
                case kTypedAliasedObjectTag:
                case kDataBufferTag:
                case kChangeBufferTag:
                    break;
                default:
                    invalidFormat();
                    break;
            }
        }
    }
    return returnValue;
}

const ITypeRepresentation*
ICountedStreamBuffer::readObjectType()
{
    unsigned char objectTag = readStartOfObject();
    if (objectTag != kStartTypedObjectTag)
        invalidFormat();
    ITypeName typeName;
    readTypeName(typeName);
    bool setFinder = false;
    if (fStreamFinder == 0)
    {
        fStreamFinder = &IStreamFinder::defaultFinder();
        setFinder = true;
    }
    const ITypeRepresentation* result = ICountedStreamIO::findTypeRepresentation(fStreamFinder, typeName);
    bool newEntry = true;
    fStreamContext->addObjectType(result, newEntry);
    if (setFinder)
        fStreamFinder = 0;
    return result;
}

void
ICountedStreamBuffer::writeObject(const IMStreamable* object, const ITypeRepresentation* typeRep)
{
    IStreamContextFrame frame(*fDataStream);
    writeEndOfCurrentFrame(false);
    if (object == 0)
    {
        writeStartOfObject(kNilObjectTag);
    }
    else
    {
        const ITypeRepresentation* objectTypeRep = &object->typeRepresentation();
        if (typeRep != objectTypeRep)
        {
            writeStartOfObject(kStartTypedObjectTag);
            writeTypeName(ITypeName::nameOfTypeRepresentation(*objectTypeRep));
        }
        else
        {
            writeStartOfObject(kStartUntypedObjectTag);
        }
        object->externalizeToStream(*fDataStream);
        writeEndOfCurrentFrame(false);
        writeEndOfObject();
    }
}

IMStreamable*
ICountedStreamBuffer::readObject(const IStreamFinder* streamFinder, const ITypeRepresentation* typeRep)
{
    IFUNCTRACE_DEVELOP();
    IStreamContextFrame frame(*fDataStream);
    IMStreamable* result = 0;
    ITypeName typeName;
    const ITypeName* objectTypeName = 0;
    const ITypeRepresentation* objectTypeRep = 0;
    unsigned char objectTag = readStartOfObject();
    switch (objectTag)
    {
        case kNilObjectTag:
            result = 0;
            break;
        case kStartTypedObjectTag:
            readTypeName(typeName);
            objectTypeName = &typeName;
            break;
        case kStartUntypedObjectTag:
            if (typeRep == 0)
            {
                ITRACE_DEVELOP(IString("Needs to know type information"));
                invalidFormat();
            }
            objectTypeRep = typeRep;
            objectTypeName = &(ITypeName::nameOfTypeRepresentation(*objectTypeRep));
            break;
        default:
            invalidFormat();
            break;
    }
    if (objectTypeName != 0)
    {
        bool setFinder = false;
        if (streamFinder == 0)
        {
            if (fStreamFinder == 0)
            {
                fStreamFinder = &IStreamFinder::defaultFinder();
                setFinder = true;
            }
            streamFinder = fStreamFinder;
        }
        else if (fStreamFinder == 0)
        {
            fStreamFinder = streamFinder;
            setFinder = true;
        }
        IMRefCounted* refCounted = 0;
        result = ICountedStreamIO::createUninitialized(streamFinder, *objectTypeName, refCounted);
        result->internalizeFromStream(*fDataStream);
        readEndOfObject();
        if (setFinder)
            fStreamFinder = 0;
    }
    return result;
}

bool
ICountedStreamBuffer::writeAliasedObject(const IMStreamable* object, const IMRefCounted* refCounted, const ITypeRepresentation* typeRep)
{
    IStreamContextFrame frame(*fDataStream);
    bool result = false;
    writeEndOfCurrentFrame(false);
    if (object == 0)
    {
        writeStartOfObject(kNilObjectTag);
    }
    else
    {
        const ITypeRepresentation* objectTypeRep = &object->typeRepresentation();
        bool typeOnStream = (typeRep != objectTypeRep);
        bool newEntry = false;
        long objectNumber = fStreamContext->addObject(object, refCounted, typeOnStream, newEntry);
        if (!newEntry)
        {
            if (typeOnStream)
            {
                writeStartOfObject(kTypedAliasedObjectTag);
                writeTypeName(ITypeName::nameOfTypeRepresentation(*objectTypeRep));
            }
            else
            {
                writeStartOfObject(kAliasedObjectTag);
            }
            writeContextNumber(objectNumber);
            if (fFrameDepth == 0)
                flushBuffer();
        }
        else
        {
            if (typeOnStream)
            {
                newEntry = false;
                long objectTypeNumber = fStreamContext->addObjectType(objectTypeRep, newEntry);
                if (!newEntry)
                {
                    writeStartOfObject(kStartAliasedTypeObjectTag);
                    writeContextNumber(objectTypeNumber);
                }
                else
                {
                    writeStartOfObject(kStartTypedObjectTag);
                    writeTypeName(ITypeName::nameOfTypeRepresentation(*objectTypeRep));
                }
            }
            else
            {
                writeStartOfObject(kStartUntypedObjectTag);
            }
            object->externalizeToStream(*fDataStream);
            writeEndOfCurrentFrame(false);
            writeEndOfObject();
            result = true;
        }
    }
    return result;
}

bool
ICountedStreamBuffer::readAliasedObject(IMStreamable*& object, IMRefCounted*& refCounted, const IStreamFinder* streamFinder, const ITypeRepresentation* typeRep)
{
    IFUNCTRACE_DEVELOP();
    IStreamContextFrame frame(*fDataStream);
    bool result = false;
    ITypeName typeName;
    const ITypeName* objectTypeName = 0;
    const ITypeRepresentation* objectTypeRep = 0;
    long contextNumber = 0;
    unsigned char objectTag = readStartOfObject();
    switch (objectTag)
    {
        case kNilObjectTag:
            object = 0;
            refCounted = 0;
            break;
        case kAliasedObjectTag:
            contextNumber = readContextNumber();
            object = fStreamContext->findObject(refCounted, contextNumber, streamFinder, typeRep);
            break;
        case kTypedAliasedObjectTag:
        {
            readTypeName(typeName);
            contextNumber = readContextNumber();
            bool setFinder = false;
            if (streamFinder == 0)
            {
                if (fStreamFinder == 0)
                {
                    fStreamFinder = &IStreamFinder::defaultFinder();
                    setFinder = true;
                }
                streamFinder = fStreamFinder;
            }
            else if (fStreamFinder == 0)
            {
                fStreamFinder = streamFinder;
                setFinder = true;
            }
            typeRep = ICountedStreamIO::findTypeRepresentation(streamFinder, typeName);
            object = fStreamContext->findObject(refCounted, contextNumber, streamFinder, typeRep);
            if (setFinder)
                fStreamFinder = 0;
            break;
        }
        case kStartAliasedTypeObjectTag:
            if ((object = readAliasedObjectAgain(refCounted)) == 0)
            {
                contextNumber = readContextNumber();
                objectTypeRep = fStreamContext->findObjectType(contextNumber);
                objectTypeName = &(ITypeName::nameOfTypeRepresentation(*objectTypeRep));
            }
            break;
        case kStartTypedObjectTag:
            if ((object = readAliasedObjectAgain(refCounted)) == 0)
            {
                readTypeName(typeName);
                objectTypeName = &typeName;
            }
            break;
        case kStartUntypedObjectTag:
            if ((object = readAliasedObjectAgain(refCounted)) == 0)
            {
                if (typeRep == 0)
                {
                    ITRACE_DEVELOP(IString("Need to know type information"));
                    invalidFormat();
                }
                objectTypeRep = typeRep;
                objectTypeName = &(ITypeName::nameOfTypeRepresentation(*objectTypeRep));
            }
            break;
        default:
            invalidFormat();
            break;
    }

    if (objectTypeName == 0)
    {
        if (refCounted)
            refCounted->addRef();
    }
    else
    {
        bool setFinder = false;
        if (streamFinder == 0)
        {
            if (fStreamFinder == 0)
            {
                fStreamFinder = &IStreamFinder::defaultFinder();
                setFinder = true;
            }
            streamFinder = fStreamFinder;
        }
        else if (fStreamFinder == 0)
        {
            fStreamFinder = streamFinder;
            setFinder = true;
        }
        object = ICountedStreamIO::createUninitialized(streamFinder, *objectTypeName, refCounted);
        bool typeOnStream = (objectTag != kStartUntypedObjectTag);
        bool newEntry = true;
        fStreamContext->addObject(object, refCounted, typeOnStream, newEntry);
        result = true;

        if (objectTag == kStartTypedObjectTag)
        {
            newEntry = true;
            if (objectTypeRep == 0)
                objectTypeRep = &object->typeRepresentation();
            fStreamContext->addObjectType(objectTypeRep, newEntry);
        }
        object->internalizeFromStream(*fDataStream);
        readEndOfObject();
        if (setFinder)
            fStreamFinder = 0;
        refCounted->addRef();
    }
    return result;
}

void ICountedStreamBuffer::invalidFormat()
{
    IInvalidDataOnStream exc;
    ITHROW(exc);
}

//
// ICountedStreamIO
//
ICountedStreamIO::ICountedStreamIO(IStreamBuffer* streamBuffer)
    :    IStreamIO()
    ,    fImp(new ICountedStreamBuffer(streamBuffer))
{
}

ICountedStreamIO::ICountedStreamIO()
    :    IStreamIO()
    ,    fImp(0)
{
}

ICountedStreamIO::ICountedStreamIO(const ICountedStreamIO& other)
    :    IStreamIO(other)
    ,    fImp(0)
{
}

ICountedStreamIO::~ICountedStreamIO()
{
    delete fImp;
}

void ICountedStreamIO::setStream(IDataStream* dataStream)
{
    fImp->setStream(dataStream);
}

void ICountedStreamIO::reset()
{
    fImp->reset();
}

void ICountedStreamIO::flush()
{
    fImp->flush();
}

//
// write primitive data types.
//
void ICountedStreamIO::writeByte(char source)
{
    fImp->writeByte(source);
}

void ICountedStreamIO::writeShort(short source)
{
    fImp->writeShort(source);
}

void ICountedStreamIO::writeLong(long source)
{
    fImp->writeLong(source);
}

void ICountedStreamIO::writeFloat(float source)
{
    fImp->writeFloat(source);
}

void ICountedStreamIO::writeDouble(double source)
{
    fImp->writeDouble(source);
}

//
// read primitive data types.
//
char ICountedStreamIO::readByte()
{
    return fImp->readByte();
}

short ICountedStreamIO::readShort()
{
    return fImp->readShort();
}

long ICountedStreamIO::readLong()
{
    return fImp->readLong();
}

float ICountedStreamIO::readFloat()
{
    return fImp->readFloat();
}

double ICountedStreamIO::readDouble()
{
    return fImp->readDouble();
}

void ICountedStreamIO::writeBytes(const char* data, unsigned long howMany)
{
    fImp->writeBytes(data, howMany);
}

void ICountedStreamIO::readBytes(char* data, unsigned long howMany)
{
    fImp->readBytes(data, howMany);
}

void ICountedStreamIO::writeShorts(const short* data, unsigned long howMany)
{
    fImp->writeShorts(data, howMany);
}

void ICountedStreamIO::readShorts(short* data, unsigned long howMany)
{
    fImp->readShorts(data, howMany);
}

void ICountedStreamIO::writeLongs(const long* data, unsigned long howMany)
{
    fImp->writeLongs(data, howMany);
}

void ICountedStreamIO::readLongs(long* data, unsigned long howMany)
{
    fImp->readLongs(data, howMany);
}

void ICountedStreamIO::writeFloats(const float* data, unsigned long howMany)
{
    fImp->writeFloats(data, howMany);
}

void ICountedStreamIO::readFloats(float* data, unsigned long howMany)
{
    fImp->readFloats(data, howMany);
}

void ICountedStreamIO::writeDoubles(const double* data, unsigned long howMany)
{
    fImp->writeDoubles(data, howMany);
}

void ICountedStreamIO::readDoubles(double* data, unsigned long howMany)
{
    fImp->readDoubles(data, howMany);
}

void ICountedStreamIO::writeObject(const IMStreamable* object, const ITypeRepresentation* typeRep)
{
    fImp->writeObject(object, typeRep);
}

IMStreamable*
ICountedStreamIO::readObject(const IStreamFinder* streamFinder, const ITypeRepresentation* typeRep)
{
    return fImp->readObject(streamFinder, typeRep);
}

bool
ICountedStreamIO::writeAliasedObject(const IMStreamable* object, const IMRefCounted* refCounted, const ITypeRepresentation* typeRep)
{
    return fImp->writeAliasedObject(object, refCounted, typeRep);
}

bool
ICountedStreamIO::readAliasedObject(IMStreamable*& object, IMRefCounted*& refCounted, const IStreamFinder* streamFinder, const ITypeRepresentation* typeRep)
{
    return fImp->readAliasedObject(object, refCounted, streamFinder, typeRep);
}

void ICountedStreamIO::writeBeginFrame()
{
    fImp->writeStartOfFrame();
}

void ICountedStreamIO::writeEndFrame()
{
    fImp->writeEndOfFrame();
}

void ICountedStreamIO::readBeginFrame()
{
    fImp->readStartOfFrame();
}

void ICountedStreamIO::readEndFrame()
{
    fImp->readEndOfFrame();
}

bool ICountedStreamIO::atEndOfFrame()
{
    return fImp->atEndOfFrame();
}

bool ICountedStreamIO::establishContext()
{
    return fImp->establishContext();
}

void ICountedStreamIO::clearContext()
{
    fImp->clearContext();
}

IMStreamable*
ICountedStreamIO::createUninitialized(const IStreamFinder* finder, const ITypeName& typeName, IMRefCounted*& refCounted)
{
    return IStreamIO::createUninitialized(finder, typeName, refCounted);
}

const ITypeRepresentation*
ICountedStreamIO::findTypeRepresentation(const IStreamFinder* finder, const ITypeName& typeName)
{
    return IStreamIO::findTypeRepresentation(finder, typeName);
}
