// IBM Open Class Library
//
// Licensed Materials - Property of IBM
//
// "Restricted Materials of IBM"
//
// (C) Copyright IBM Corporation, 1992, 1997 All Rights Reserved
//
// Revision: 40 1.13.1.2 source/core/testfw/iargdict.cpp, testfw, ioc.v400, 000302 

#include <istrwrap.hpp>
#include <iargdict.hpp>
#include <iexcbase.hpp>
#include <itest.hpp>

#include <stdio.h>
#include <assert.h>

// Segment definitions
#ifdef IC_PAGETUNE
    #define _IARGDICT_CPP_
    #include <ipagetun.h>
#endif

//-----------------------------------------------------------------------------
// IArgumentDictionary: PUBLIC Constants
//-----------------------------------------------------------------------------

static const IString gkEmptyText;
#ifdef _AIX
const IString& IArgumentDictionary::kEmptyText = gkEmptyText;
#else
#pragma info(nodcl)
const IString& _Export IArgumentDictionary::kEmptyText = gkEmptyText;
#pragma info(restore)
#endif

//-----------------------------------------------------------------------------
// IArgumentDictionary: PUBLIC Special member Functions
//-----------------------------------------------------------------------------

IArgumentDictionary::IArgumentDictionary()
{
    // We need to use a local text object because using
	// IArgumentDictionary::kEmptyText will result in a crash when this
	// constructor is used to create a static IArgumentDictionary. 
	// This is due to the fact that the order of initialization
    // of static objects between files is undefined.  So when we are trying
	// to construct a static IArgumentDictionary object,
	// IArgumentDictionary::kEmptyText could be undefined.

    fDictionary = new ISortedMap<IStringWrapper,IString>;
	const IString kLocalEmptyText;
	initialize(kLocalEmptyText);
}

IArgumentDictionary::IArgumentDictionary(ITest& theTest, const IString& nakedOptions)
{
	fDictionary = new ISortedMap<IStringWrapper,IString>;
	initialize(nakedOptions, &theTest.outputTextStream());
	importTextArguments(theTest);
}


IArgumentDictionary::IArgumentDictionary(const IString& argumentString, const IString& nakedOptions)
{
	fDictionary = new ISortedMap<IStringWrapper,IString>;
	initialize(nakedOptions);
	importTextArguments(argumentString);
}

IArgumentDictionary::IArgumentDictionary(const int argc, const char *const *argv, const IString& nakedOptions)
{
	
	fDictionary = new ISortedMap<IStringWrapper,IString>;
	initialize(nakedOptions);
	importTextArguments(argc, argv);
}


IArgumentDictionary::IArgumentDictionary(const IArgumentDictionary& source)
{
	fDictionary = new ISortedMap<IStringWrapper,IString>;
	fLocalOut = false;
	*this = source;
}

IArgumentDictionary& IArgumentDictionary::operator=(const IArgumentDictionary& rhs)
{
	if (this != &rhs) {
		fNakedOptions = rhs.fNakedOptions;
		fRemoveOnLookup = rhs.fRemoveOnLookup;
		fLookupSimilarKeys = rhs.fLookupSimilarKeys;
		
		if (fLocalOut) delete fOut;
		fLocalOut = false;
		fOut = 0;
		if (!rhs.fLocalOut) fOut = rhs.fOut; 
		*fDictionary = *rhs.fDictionary;
	}
	return *this;
}

IArgumentDictionary::~IArgumentDictionary()
{

	if (fLocalOut) delete fOut;
	delete fDictionary;
}


//---------------------------------------------------------------------------
//  IArgumentDictionary: PUBLIC Importing Methods
//-----------------------------------------------------------------------------
void IArgumentDictionary::importTextArguments(ITest& theTest)
{
	ISequence<IString>  textArguments;
	theTest.copyInputs(textArguments);
	importTextArguments(textArguments);
	
}

void IArgumentDictionary::importTextArguments(const IString& argumentString)
{
	ISequence<IString> textArguments;
	tokenize(argumentString, &textArguments);
	importTextArguments(textArguments);
}

void IArgumentDictionary::importTextArguments(const int argc, const char *const *argv)
{

	ISequence<IString> textArguments;
	
	// Construct IStrings for the items in argv. We start with index 1 in order
	// to skip over the application name which is always at index 0.
	for (int i=1; i<argc && argv[i]; ++i) {
		textArguments.addAsLast(IString(argv[i]));
	}
	importTextArguments(textArguments);
}



void IArgumentDictionary::specifyNakedOptions(const IString& nakedOptions)
/*
	specifyNakedOptions takes the given IString and parses it into a list of
	space delimited keys. Each key is taken as a naked option. For instance,
	the	input "-p -cc" will add two naked options, "-p" and "-cc", to the set
	of	naked options.
*/
{
	tokenize(nakedOptions, &fNakedOptions);
}


//-----------------------------------------------------------------------------
// Export methods
//-----------------------------------------------------------------------------
void IArgumentDictionary::exportTextArguments(IString& argumentString) /*semantically const*/
{

	ISortedMap<IStringWrapper, IString>::Cursor cursor(*fDictionary);

	const IString space(" ");
	bool first = true;
	
	for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
	{
		IString key = cursor.element().getkey();
		IString value = cursor.element().value();

		if (first) first = false; else argumentString = argumentString + space;
		
		if (key[1] == '-') {
			argumentString = argumentString + key;
		}
		
		if (value.length() != 0) {
			if (key[1] == '-') {
				argumentString = argumentString + space;
			}
			argumentString = argumentString + value;
		}
	}
}


//-----------------------------------------------------------------------------
// IArgumentDictionary: PUBLIC Lookup Methods
//-----------------------------------------------------------------------------

void IArgumentDictionary::lookupSimilarKeys(bool b) // Call findSimilarKey for all lookups.
{
	fLookupSimilarKeys = b;
}


unsigned long IArgumentDictionary::countOfArguments() const
{
	return fDictionary->numberOfElements();
}

void IArgumentDictionary::removeOnLookup(bool b) 
{
	fRemoveOnLookup = b;
}

bool IArgumentDictionary::textAt(const IString& key, IString& value)
/*
	textAt returns the value associated with the given <key>. If there 
	is no such key, textAt returns false.
*/
{
	return lookupValue(key, value);
	
}


bool IArgumentDictionary::nthValue(const short index, IString& value)
/*
	nthValue attempts to return the index-th option-less value. It returns 
	false if no such value exists. 
*/
{
	IString key(index);
	return textAt(key, value);
}



bool IArgumentDictionary::findSimilarKey(const IString& keyToMatch, IString& newKey)  const
/*
	findSimilarKey looks through the dictionary for a key which matches the
	given <keyToMatch>. It returns the first match found without looking for
	others. It returns 0 if no similar key is found.
	
	If a key is shorter than the keyToMatch, it does not match. If a key is
	longer than the keyToMatch, extra characters are ignored. Characters are
	compared, case-insensitive, and if all characters up to keyToMatch.Length()
	match, the key is considered to match.
	
	FindSimilar key does *not* delete the key/value from this dictionary,
	regardless of the setting of fRemoveOnLookup.
*/
{
	IString keyFragment, key;
    // if there is no exact match
	if (!fDictionary->containsElementWithKey(keyToMatch)) {

	    ISortedMap<IStringWrapper,IString>::Cursor cursor(*fDictionary);

		// get string value for key to match.  convert string to all lower case.
		unsigned significantLength = keyToMatch.length();
        IString lcKeyToMatch = keyToMatch;
		lcKeyToMatch.lowerCase();
		
		// look for possible match for key.
		for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
		{
	        key = cursor.element().getkey();
			if (key.length() >= significantLength) {
				keyFragment = key.subString(1,significantLength);
				keyFragment.lowerCase();
				if (lcKeyToMatch ==  keyFragment){
					newKey = key;
					return true;
				}
			}
		}
		return false;
	}
	else {
      newKey = keyToMatch;
	  return true;  
    }

}



//-----------------------------------------------------------------------------
// IArgumentDictionary: PUBLIC Exception-throwing lookup methods
//-----------------------------------------------------------------------------

bool IArgumentDictionary::numberAt(const IString& key, double& value,
	double minAllowed, double maxAllowed)
/*
	This numberAt returns a true result if the argument exists and it parses
	perfectly and it is within the given range, inclusively. If any condition
	fails, it returns false.
*/
{
	double confidence, tempValue;
	if (rawNumberAt(key, tempValue, confidence)) {
		if (confidence != 1.0 || tempValue < minAllowed || tempValue > maxAllowed) {
			if (confidence != 1.0){
				out() << "Syntax error; " << key << " must be followed by a number\n";
			}
			else{
				out() << "Range error; value " << tempValue << " of key \"" << key
					<< "\" must be in the range " << minAllowed << " to "
					<< maxAllowed << (char)'\n';
			}
			return false;
		}
		value = tempValue;
		return true;
	}
	return false;
}

bool IArgumentDictionary::numberAt(const IString& key, long& value,
	long minAllowed, long maxAllowed)
/*
	This numberAt is identical to the one returning a double <value>, but makes
	an additional check that the value is an integer. If not, it returns false.
*/
{
	double confidence, tempValue;
	if (rawNumberAt(key, tempValue, confidence)) {
 		bool isInt = (tempValue - long(tempValue)) < DBL_EPSILON; 
		if (confidence != 1.0 || !isInt || tempValue < minAllowed || tempValue > maxAllowed) {
			if (confidence != 1.0){
				out() << "Syntax error; " << key << " must be followed by a number\n";
			}
			else if (!isInt){
				out() << "Bad value " << tempValue << "; " << key
					<< " must be followed by an integer\n";
			}	
			else {
				out() << "Range error; value " << tempValue << " of key \"" << key
					<< "\" must be in the range " << minAllowed << " to "
					<< maxAllowed << (char)'\n';
			}
			return false;
		}
 		value = tempValue; 
		return true;
	}
	return false;
}

bool IArgumentDictionary::nthNumber(const short index, double& value,
	double minAllowed, double maxAllowed)
/*
	nthNumber returns the <index>-th option-less value parsed as a number in
	the	parameter <value>. It behaves like the range-checked numberAt methods.
*/
{
	IString key(index);
	return (numberAt(key, value, minAllowed, maxAllowed));
}

bool IArgumentDictionary::nthNumber(const short index, long& value,
	long minAllowed, long maxAllowed)
{
	IString key(index);
	return (numberAt(key, value, minAllowed, maxAllowed));
}




//-----------------------------------------------------------------------------
// IArgumentDictionary: static method
//-----------------------------------------------------------------------------

// Turn this on in OpusBug to debug tokenize
static bool gDebugTokenize = false;


void IArgumentDictionary::tokenize(const IString& inputString, ISequence<IString>* tokenSequence)
/*
	tokenize breaks 'inputString' into tokens, and adds them as
	new IStrings on the heap into 'tokenSequence'. It handles
	double quotes and recognizes the escape character '\'. An unmatched
	double quote eats the rest of the line! Caller owns new objects in
	'tokenSequence'.
*/
{
	#define charTypeName(t) \
		(t==kWhite? "kWhite": (t==kQuote? "kQuote": (t==kEscape? "kEscape": (t==kEnd? "kEnd": "kChar"))))
	#define stateName(s) \
		(s==kScan? "kScan": (s==kIdent? "kIdent": (s==kOpenQuote? "kOpenQuote": (s==kInQuote? "kInQuote": (s==kLiteral? "kLiteral": "kStop")))))



#pragma enum(4)
	enum ECharType { kWhite, kChar, kQuote, kEscape, kEnd };
	enum ELexState { kScan, kIdent, kOpenQuote, kInQuote, kLiteral, kStop } state = kScan;
#pragma enum(pop)
	
	static ELexState gStateTable[5][5] = {
	  //              kWhite      kChar         kQuote        kEscape      kEnd

	  /*kScan*/     { kScan,      kIdent,       kOpenQuote,   kIdent,      kScan },
	  /*kIdent*/    { kStop,      kIdent,       kIdent,       kIdent,      kStop },
	  /*kOpenQuote*/{ kInQuote,   kInQuote,     kStop,        kLiteral,    kScan },
	  /*kInQuote*/  { kInQuote,   kInQuote,     kStop,        kLiteral,    kStop },
	  /*kLiteral*/  { kInQuote,   kInQuote,     kInQuote,     kInQuote,    kStop }
	};
	
	char c = '\0';
	IString token; // This holds each token we find
	int i = 1;
	bool isDone = false;
	int length = inputString.length();

	// We go through this outer loop once for each char
	for (;;) { // We'll see kNull when we run off the end

		ECharType charType = kChar;
		
		if (i > length){
			charType = kEnd;
			isDone = true;
		}
		else{
			c = inputString[i];
		}
					
		if (!isDone){
			if (isspace(c)) charType = kWhite;
			else if (c == '\"') charType = kQuote;
			else if (c == '\\') charType = kEscape;  // ReverseSolidus means backslash
			
			else if (c == '\0') charType = kEnd;
		}
		isDone = false;

		if (gDebugTokenize)
			printf("# tokenize: '%c', type = %s, state = %s -> ", c, charTypeName(charType), stateName(state));

		state = gStateTable[state][charType];

		if (gDebugTokenize)
			printf("%s\n", stateName(state));
		
		switch (state) {
			case kIdent:
			case kInQuote: {
				token = token + IString(c); // Append character
				break;
			}
			case kStop: {
				if (gDebugTokenize) {
					printf("# tokenize: TOKEN = \"");
					printf("\"\n");
				}
				tokenSequence->add(token);
				token = IString("");
				state = kScan;
				break;
			}
			default:
				break; // Do nothing
		}

		if (charType == kEnd) break; // This is the only exit from this loop
		i++;
	}
}

//-----------------------------------------------------------------------------
// IArgumentDictionary: PRIVATE Methods
//-----------------------------------------------------------------------------

void IArgumentDictionary::importTextArguments(ISequence<IString>& textArguments)  // make this private

/*
    importTextArguments parses the items in textArguments, which should be
    IString objects, and places them in this dictionary. Arguments are
    added to any existing	dictionary entries. Options are recognized by a 
    leading '-' character.
	It also tries to recognize negative numbers by trying to parse options as
	numbers, and assuming they aren't numbers if the parse fails.

	The following rules are used (nonopt is a IString with no leading '-'):
	
	(1)	option	nonopt		Pair (option, nonopt) added to dictionary
	(2)	option	option2		Pair (option, "") added to dictionary
	(3)	nonopt	nonopt2		Pair ("n", nonopt2) added, where n is 1,2,
	
	importTextArguments checks options to see if they are in the collection
	fNakedOptions. If and only if so the following rule replaces rule 1:
	
	(1a) option	nonopt		Pair (option, ""), then Pair ("n", nonopt) added
	
	Options not in fNakedOptions are assumed to have 0 or 1 values as in rule
	1.

*/
{
	long plainArgumentCount = 1; // Counter for plain (non-option) arguments
	IString key;
	IString arg;
	
#pragma enum(4)
	enum {kIdle, kNeedValue} state = kIdle;
#pragma enum(pop)
	
#ifdef DEBUG
	out() << "\nBegin Debug IArgumentDictionary -----\n";
#endif
	
	ISequence<IString>::Cursor cursor(textArguments);
	for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
	{
		arg = cursor.element();
		bool isOption = false;
#ifdef DEBUG
		out() << "  Input: \"" << arg << "\"\n";
#endif
		// Determine if this is an option. It is if (a) it has a leading
		// minus, and (b) it doesn't parse perfectly as a number.
		if (arg[1] == '-') {
			double d;
			if (sscanf(arg, "%lf", &d) != 1) {
				isOption = true;
			}
		}

		if (isOption) {
			if (state == kNeedValue) { // last option was empty (no value)
				// Save LAST ARG to dictionary with a blank value
				fDictionary->add(IStringWrapper(key, kEmptyText));
			}
				
			key = arg;
			
			IString temp;	
			if (lookupMember(key, fNakedOptions,temp)) {
				fDictionary->add(IStringWrapper(key, kEmptyText));
				state = kIdle;
			}
			else {
				state = kNeedValue;
			}
		}
		else {
			IString newValue(arg);
			
			switch (state) {
			case kNeedValue: // arg is value for last option
				// Save last arg and this arg as a pair
				fDictionary->add(IStringWrapper(key, newValue));
				state = kIdle;
				break;
			case kIdle: // This arg is unaccompanied by a previous -option
				{
					key = IString(plainArgumentCount++);
					fDictionary->add(IStringWrapper(key, newValue));
					break;
				}
			}
		}
	} // for()

	if (state == kNeedValue) { // Final option was empty (no value)
		// Save LAST ARG to dictionary with a blank value
		fDictionary->add(IStringWrapper(key,kEmptyText));

	}

#ifdef DEBUG
	ISortedMap<IStringWrapper, IString>::Cursor cursor(*fDictionary);

	const IString space(" ");
	
	for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
	{
		IStringWrapper elem = cursor.element();
		out() << "  Entry: \"" << elem.key() << "\", \"" << elem.value() << "\"\n";
	}

	out() << "End Debug IArgumentDictionary -----\n";
#endif
}

bool IArgumentDictionary::lookupValue(const IString& keyArg, IString& Value)
/*
	lookupValue looks for the <keyArg> in the dictionary and puts its
	associated value in the argument <Value>.  It returns false if not found.
	If fLookupSimilarKeys is true, it looks for similar keys. If
	fRemoveOnLookup is true, it removes the key/value pair from the	dictionary 
	if found.  
	
	lookupValue encapsulates the logic of fLookupSimilarKeys and
	fRemoveOnLookup. Other methods should avoid refering to these data
	members directly, and should avoid querying the Dictionary directly.
*/
{
	IString key;
	
	if (fLookupSimilarKeys){
		if (!findSimilarKey(keyArg, key))
			return false;
	}
	else {
		key = keyArg;
	}
	
   	ISortedMap<IStringWrapper, IString>::Cursor cursor(*fDictionary);	
	if (fDictionary->locateElementWithKey(key, cursor))
	{
		Value = fDictionary->elementAt(cursor).value();
		if (fRemoveOnLookup)
			fDictionary->removeAt(cursor);
	    return true;
	}
	else return false;
}

bool IArgumentDictionary::lookupMember(const IString& text, const ISequence<IString>& col, IString& matchingText)
/*
	lookupMember looks for the <text> in the <col> and puts the matching
	member in matchingText.  It returns false if text is not found. If
	fLookupSimilarKeys is true, it looks for similar <text> to be similar to 
	each member of the collection. This means <text> may be longer, and need
	not match case.
	
	This code is very similar to findSimilarKey. It is used by
	importTextArguments	to do flexible matching of naked options.
*/
{
	ISequence<IString>::Cursor cursor(col);
	
	for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext()){
		if (cursor.element() == text) {
			matchingText = text;
			return true;
		}
	}
	
	if (fLookupSimilarKeys) {

		IString textstring = text;
		unsigned significantLength = textstring.length();
		textstring.lowerCase();

		for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
		{
			IString temptext = col.elementAt(cursor);
			if (temptext.length() >= significantLength) {
				IString textFragment = temptext.subString(1, significantLength);
				textFragment.lowerCase();
				if (textstring == textFragment){
					matchingText = temptext;
					return true;
				}
			}
		}
		return false;	
	}
	return false;
}


ITieredTextBuffer& IArgumentDictionary::out()
/*
	out returns a buffer for general output. In general this is just one that
	is	created on the spot, but if the ITest constructor is used, we use
	that one.
*/
{
	if (!fOut) {
		fOut = new ITieredTextBuffer;
		assert(fOut != NULL);
		fLocalOut = true;
	}
	return *fOut;
}


bool IArgumentDictionary::rawNumberAt(const IString& key, double& value,
				double& confidence)
/*
	rawNumberAt returns true if a numeric value exists for the given key and
	returns the value. It returns false if the key does not exist. rawNumberAt
	returns the confidence of the parse from 0.0F (bad) to 1.0F (perfect).
*/
{
		IString valueText;
	
		if (textAt(key, valueText))
		{
			confidence = (double) sscanf(valueText, "%lf", &value);
			return true;
		}
	
		else return false;
}

void IArgumentDictionary::initialize(const IString& nakedOptions, ITieredTextBuffer* out)
{
	fRemoveOnLookup = false;
	fLookupSimilarKeys = false;
	fLocalOut = false; // This will get set to true if needed by out()
	fOut = out;
	specifyNakedOptions(nakedOptions);
}
