// Revision: 25 1.5.1.3 source/core/testfw/iruntest.cpp, testfw, ioc.v400, 000302 
/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/

#include <iruntest.hpp>

#include <stdio.h>
#include <string.h>

// Segment definitions
#ifdef IC_PAGETUNE
    #define _RUNTEST_CPP_
    #include <ipagetun.h>
#endif

static void printCommandLineAndRunTestApplicationUsage(int argc, const char *const *argv);


void parseRunTestApplicationArguments(int argc, const char *const *argv,
            const char *&className, unsigned long &numberOfRuns, int& testArgs,
            ITieredTextBuffer::ETier &echoTier, bool& parseError, bool checkTOption,
            const char *&logFileName)
/*
    parseRunTestApplicationArguments is a global function that is only used
    by the beginTestFrameworkMacro macro to do command line argument parsing.
*/
{
    int i = 1;

    // initialize parameters.
    className = NULL;
    echoTier = ITieredTextBuffer::kNormal;
    numberOfRuns = 1;
    testArgs = 0;

    while (i < argc && !parseError) {

        if ((strcmp(argv[i], "-e") == 0) || (strcmp(argv[i], "-echo") == 0)) {

            if (i+1 == argc) {
                printf("# %s - tier not specified with -e[cho] option.\n", argv[0]);
                parseError = true;
            } else {
                i++;
                switch (argv[i][0]) {
                    case 't': echoTier = ITieredTextBuffer::kTop; break;
                    case 'h': echoTier = ITieredTextBuffer::kHeadline; break;
                    case 'g': echoTier = ITieredTextBuffer::kGeneral; break;
                    case 'n': echoTier = ITieredTextBuffer::kNormal; break;
                    case 'd': echoTier = ITieredTextBuffer::kDetail; break;
                    case 'D': echoTier = ITieredTextBuffer::kDebug; break;
                    case 'b': echoTier = ITieredTextBuffer::kBottom; break;
                    default:
                        printf("# %s - invalid tier specified with -e[cho] option.\n", argv[0]);
                        parseError = true;
                        break;
                }
            }

        } else if (strcmp(argv[i], "-n") == 0) {
            if (i+1 == argc) {
                printf("# %s - number of runs not specified with -n option.\n", argv[0]);
                parseError = true;
            } else {
                i++;
                if (sscanf(argv[i], "%lu", &numberOfRuns) != 1) {
                    printf("# %s - invalid number specified with -n option.\n", argv[0]);
                    parseError = true;
                }
            }

        } else if ((strcmp(argv[i], "-o") == 0) ||
                   (strcmp(argv[i], "-options") == 0)) {
            testArgs =i;
            break;

        } else if ((strcmp(argv[i], "-t") == 0) ||
                   (strcmp(argv[i], "-test") == 0)) {
            if (i+1 == argc) {
                printf("# %s - test class not specified with -t option.\n", argv[0]);
                parseError = true;
            } else {
                i++;
                className = argv[i];
            }
        } else if ((strcmp(argv[i], "-l") == 0) ||
                   (strcmp(argv[i], "-log") == 0)){

            if ((i+1 == argc) || (strncmp(argv[i+1], "-", 1)) == 0)
                logFileName = "TestLog";
            else {
                i++;
                logFileName = argv[i];
            }
        } else {
            printf("# %s - %s option not understood\n", argv[0], argv[i]);
            parseError = true;
        }
        i++;
    }

    // only require -t option if application was created NOT using the
    // the runTestImplementationMacro macro.
    if (className == NULL && checkTOption) {
        printf("# %s - test class name not specified\n", argv[0]);
        parseError = true;
    }
    if (parseError) {
        printCommandLineAndRunTestApplicationUsage(argc, argv);
    }
}





void printCommandLineAndRunTestApplicationUsage(int argc, const char *const *argv)
/*
    printCommandLineAndRunTestApplicationUsage is a static function that
    prints out the command line arguments and the usage information.
*/
{
    const char* kRunTestUsage =
        " -t <class> -e <EchoTier> -n <NumberOfRuns> -o <TestArgs...> \n"
        "  -t[est] <class>             # name of the ITest subclass to be run\n"
        "  -e[cho] h|g|n|d|D           # how much detail to show in diagnostic output\n"
        "                              # (headline, general, normal, detail, Debug)\n"
        "  -n <numberOfRuns>           # run test this many times, default is 1\n"
        "  -l[og] [filename]           # write test output to filename. If filename is not\n"
        "                              # specified, the default file name is TestLog\n"
        "  -o[ptions] <testArgs...>    # pass further arguments to test\n"
        ;

    bool first = true;
    printf("# %s - Commandline: (", argv[0]);
    for (int i=1; i < argc; i++) {
        if (!first) {
            printf(" ");
        }
        printf("%s", argv[i]);
        first = false;
    }
    printf(")\n");
    printf("%s %s\n", argv[0], kRunTestUsage);
}





void runTestInRunTestApplication(ITest& theTest, int argc,
            const char *const *argv,
            const unsigned long numberOfRuns,
            const int testArgs, const ITieredTextBuffer::ETier echoTier,
            bool& isTestRun, const char* logFileName)
/*
    runTestInRunTestApplication is a global function that is only used
    by the runTestMacro macro to set the inputs and echo tier for the test and
    run it.
*/
{
    isTestRun = false;

    if (testArgs && testArgs+1 < argc) {
        theTest.setInputs(argc-testArgs-1, &(argv[testArgs+1]));
    }

    if (logFileName)
        theTest.outputTextStream().openLog(logFileName);
    theTest.outputTextStream().setEchoTier(echoTier);
    for (unsigned long l=0; l<numberOfRuns; l++) {
        theTest.run();
        isTestRun = true;
    }
}
