// Revision: 11 1.18.1.3 source/core/testfw/itest.cpp, testfw, ioc.v400, 000302 
/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/

#include <itest.hpp>
#include <ittxtbuf.hpp>
#include <itiertxt.hpp>
#include <istrwrap.hpp>
#include <iexcbase.hpp>
#include <ism2.h>
#include <ilanglvl.hpp>

#include <stdio.h>
#include <string.h>
#include <assert.h>

// Segment definitions
#ifdef IC_PAGETUNE
    #define _ITEST_CPP_
    #include <ipagetun.h>
#endif

//-----------------------------------------------------------------------------
// ITest: Constructor, Destructor, operator=
//-----------------------------------------------------------------------------

ITest::ITest():
        fOutputTextBuffer()
{
        fIsSuccessful = true;
        fIsSuccessDecided = false;
        fIsReset = true;
        fIsStickySuccess = false;
        fExceptionStage = kNone;
        fRunCount = 0;
        fInfoMap = new ISortedMap<IStringWrapper,IString>;
}


ITest::ITest(const ITest& source):
        fOutputTextBuffer(source.fOutputTextBuffer)
{
        importFrom(source);
        fInfoMap = new ISortedMap<IStringWrapper,IString>(*source.fInfoMap);
}


ITest& ITest::operator=(const ITest& rhs)
{
        if (this != &rhs) {
                fOutputTextBuffer = rhs.fOutputTextBuffer;
                importFrom(rhs);
                *fInfoMap = *rhs.fInfoMap;
        }
        return *this;
}



ITest::~ITest()
{
        delete fInfoMap;
}


//----------------------------------------------------------------------------
//  ITest: private static methods
//----------------------------------------------------------------------------
const char* ITest::stageName(ITest::EStage stage)
{
        static const char* kStageNameTable[] = {
                "None",                         // exception did not occur
                "Run",                          // exception in Run() but not in any of the following:
                "Reset",                        // exception in Reset()
                "Setup",                        // exception in Setup()
                "Test",                         // exception in Test()
                "Cleanup",                      // exception in Cleanup()
                "Unknown",
        };

        const char* name = kStageNameTable[ITest::kCleanup + 1];
        if (stage >= ITest::kNone || stage <= ITest::kCleanup)
                name = kStageNameTable[stage];
        return name;
}


void ITest::printException(ITieredTextBuffer& out, const IException& ie, ITest::EStage stage)
{
        out.pushTier(ITieredTextBuffer::kHeadline);
        out << "Exception occured in stage: " << ITest::stageName(stage) << "\n";
        out << "Type of exception : " << ie.name() << "\n";
        out << "Exception message : " << ie.text() << "\n";
        //out << "Location of exception is: " << ie.locationAtIndex(0)->fileName() << "\n";
        out.popTier();
}


//----------------------------------------------------------------------------
//  ITest: Public exception method
//----------------------------------------------------------------------------

ITest::EStage  ITest::exceptionStage() const
{
        return fExceptionStage;
}


//------------------------------------------------------------------------------
// ITest: PUBLIC setup--target, inputs, flag configuration
//------------------------------------------------------------------------------



void ITest::setInputs(const int argc, const char *const *argv)
/*
        Delete current inputs and create IString objects for
        each item in argv and add the text to the inputs.
*/
{
        fInputTexts.removeAll();

        for (int i=0; i<argc; i++) {
                fInputTexts.add(argv[i]);
        }
}

void ITest::copyInputs(int& argc, char **& argv) const
{
        argc = fInputTexts.numberOfElements();
        if (argc == 0)
                argv = NULL;
        else {
                argv = new char*[argc];
                ISequence<IString>::Cursor cursor(fInputTexts);
                unsigned short i = 0;
                for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
                {
                        unsigned short len = cursor.element().length();
                        char *str = new char[len+1];
                        sprintf(str, "%s", cursor.element());
                        str[len] = '\0';
                        argv[i++] = str;
                }
        }
}

//-----------------------------------------------------------------------------
// ITest: PUBLIC Execution and success result
//-----------------------------------------------------------------------------


void ITest::run()
{

        try{
                fIsSuccessful = true;
                ++fRunCount;
                setStage(kRun);

                if (!isReset()) {       // has the test been run before?
                        setStage(kReset);
                        reset();
                        setStage(kRun);
                }

                if (fIsSuccessful) {            // Continue only if reset passed
                        fIsReset = false;               // Once we enter setup, no longer reset
                        setStage(kSetup);
                        setup();
                        setStage(kRun);
                }

                if (fIsSuccessful) {            // Skip test method if reset or setup failed
                        setStage(kTest);
                        test();
                        setStage(kRun);
                        if (!fIsSuccessDecided) { // if test has not decided, mark it as failed
                                outputTextStream().pushTier(ITieredTextBuffer::kHeadline);
                                outputTextStream() << "Fail; programming error; setSuccess() was not called.\n";
                                outputTextStream().popTier();

                                setSuccess(false);
                        }
                }
                setStage(kNone);
        }
        catch (const IException& ie)
        {
                setSuccess(false);
                printException(outputTextStream(), ie, fExceptionStage );
        }

        catch(...)
        {
                setSuccess(false);
                outputTextStream() << "Unknown exception occured in stage: "
                                       << ITest::stageName(fExceptionStage) << "\n";
        }

        // Don't change the EStage here; that's used to record the stage of
        // exceptions happening in the above try block.
        try{
                cleanup();
 }
        catch(const IException& ie)
        {
                setSuccess(false);
                printException(outputTextStream(), ie, ITest::kCleanup);
        }
        catch(...)
        {
                setSuccess(false);
                outputTextStream() << "Unknown exception occured in stage: "
                                       << ITest::stageName(ITest::kCleanup) << "\n";
        }

try{
                ITieredTextBuffer& out = outputTextStream();
                out << *this;
}

        catch(const IException& ie)
        {
                ITieredTextBuffer& out = outputTextStream();
                out.pushTier(ITieredTextBuffer::kHeadline);
                out << "While printing: ";
                out.popTier();
                printException(outputTextStream(), ie, ITest::kRun);
        }
        catch(...)
        {
                ITieredTextBuffer& out = outputTextStream();
                out.pushTier(ITieredTextBuffer::kHeadline);
                out << "While printing: ";
                out.popTier();
                outputTextStream() << "Unknown exception occured in stage: "
                                       << ITest::stageName(ITest::kRun) << "\n";
        }

}



bool ITest::success() const
{
        return fIsSuccessful;
}

//-----------------------------------------------------------------------------
// ITest: PUBLIC test output methods
//-----------------------------------------------------------------------------

ITieredTextBuffer& ITest::outputTextStream()
{
        return fOutputTextBuffer;
}


//-----------------------------------------------------------------------------
// TTest: PUBLIC Bookkeeping methods
//-----------------------------------------------------------------------------
unsigned long ITest::runCount() const
{
        return fRunCount;
}

//-----------------------------------------------------------------------------
// ITest: PUBLIC Metainformation framework methods
//---------------------------------------------------------------------------


void ITest::copyInfo()
{
        /* subclass should override this method to get information about the test
           printed out.

           example:
                   void ISampleTest::copyInfo()
                   {
                        addInfo(kDescriptionKey, "ISampleTest tests.....");
                        addInfo(kInputSyntax, "sampletest -t ISampleTest -o -i");
                   }
        */

}

void ITest::addInfo(const IString& key, const IString& value)
{
        fInfoMap->add(IStringWrapper(key, value));
}


bool ITest::lookupInfo(const IString& key, IString& info)
/*
        lookupInfo returns metainformation about this class of tests. It relies
        on subclasses' overriding copyInfo.
*/
{
        copyInfo();
        if (!fInfoMap->containsElementWithKey(key))
        {
                return false;
        }
        info = fInfoMap->elementWithKey(key).value();

        return true;
}


const IString gKDescriptionKey("Description");
const IString gKInputSyntaxKey("Input Syntax");
const IString gKTargetClassKey("Target Class");
const IString gKTestNameKey("test Name");
const IString gKTestVersionKey("test Version");
const IString gKTestTypeKey("test Type");
const IString gKComponentKey("Component");

#ifdef _AIX
const IString&  ITest::kDescriptionKey = gKDescriptionKey;
const IString&  ITest::kInputSyntaxKey = gKInputSyntaxKey;
const IString&  ITest::kTargetClassKey = gKTargetClassKey;
const IString&  ITest::kTestNameKey = gKTestNameKey;
const IString&  ITest::kTestVersionKey = gKTestVersionKey;
const IString&  ITest::kTestTypeKey = gKTestTypeKey;
const IString&  ITest::kComponentKey = gKComponentKey;
#else
#pragma info(nodcl)
const IString& _Export ITest::kDescriptionKey = gKDescriptionKey;
const IString& _Export ITest::kInputSyntaxKey = gKInputSyntaxKey;
const IString& _Export ITest::kTargetClassKey = gKTargetClassKey;
const IString& _Export ITest::kTestNameKey = gKTestNameKey;
const IString& _Export ITest::kTestVersionKey = gKTestVersionKey;
const IString& _Export ITest::kTestTypeKey = gKTestTypeKey;
const IString& _Export ITest::kComponentKey = gKComponentKey;
#pragma info(restore)
#endif

//-----------------------------------------------------------------------------
// ITest: PROTECTED Framework methods for execution
//-----------------------------------------------------------------------------

void ITest::reset()
{
        if (!fIsReset) {
                fIsSuccessful = true;
                fIsSuccessDecided = false;
                fIsStickySuccess = false;
            fExceptionStage = kNone;
                fIsReset = true;
        }
}

void ITest::setup()
{
}

void ITest::cleanup()
{
}

//-----------------------------------------------------------------------------
// ITest: PROTECTED Methods for test control
//-----------------------------------------------------------------------------

void ITest::setSuccess(bool success)
{
        if (!(fIsStickySuccess && !fIsSuccessful && fIsSuccessDecided)) {
                fIsSuccessful = success;
        }
        fIsSuccessDecided = true;
}

void ITest::setStickySuccess(bool success)
{
        if (!(fIsStickySuccess && !fIsSuccessful && fIsSuccessDecided)) {
                fIsSuccessful = success;
        }
        fIsSuccessDecided = true;
        fIsStickySuccess = true;
}

bool ITest::isSuccessUndecided() const
{
        return !fIsSuccessDecided;
}

bool ITest::isReset() const
{
        return fIsReset;
}



//-----------------------------------------------------------------------------
// ITest: PRIVATE Methods
//-----------------------------------------------------------------------------
void ITest::setInputs(const ISequence<IString>& TextFromWhere)
{
        fInputTexts.removeAll();

        ISequence<IString>::Cursor cursor(TextFromWhere);
        for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
        {
                fInputTexts.addAsLast(cursor.element());
        }
}


void ITest::copyInputs(ISequence<IString>& textsToWhere) const
{
        textsToWhere.removeAll();
        ISequence<IString>::Cursor cursor(fInputTexts);
        for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
        {
                textsToWhere.addAsLast(cursor.element());
        }
}

void ITest::setStage(ITest::EStage stage)
{
        fExceptionStage = stage;
}


void ITest::importFrom(const ITest& source)
/*
        Copies all data members in source into *this, with the exception of
        fOutputTextBuffer
*/
{
        fInputTexts = source.fInputTexts;

        fIsSuccessful = source.fIsSuccessful;
        fIsSuccessDecided = source.fIsSuccessDecided;
        fIsReset = source.fIsReset;
        fIsStickySuccess = source.fIsStickySuccess;
        fRunCount = source.fRunCount;
        fExceptionStage = source.fExceptionStage;
}


//-----------------------------------------------------------------------------
// ITest: PROTECTED output method
//-----------------------------------------------------------------------------

void ITest::print(ITieredTextBuffer& out)
/*
        print prints out a text representation of this test to <out>. Subclasses
        should override this method. In their overridden implementation, they
        shouldfirst call their inherited implementation. For example:

                class IDerivedTest : public IBaseTest {
                        virtual void print(ITieredTextBuffer&);
                };

                void IDerivedTest::print(ITieredTextBuffer& out)
                {
                        TBaseTest::print(out);
                        out << "My stuff goes here...\n";
                }
*/
{
        bool first;

        out << "run count: " << runCount() << (char)'\n';

        //      info: { "group" "unit" }
        out << "info: {";
        first = true;

        copyInfo();

        ISortedMap<IStringWrapper,IString>::Cursor cursor(*fInfoMap);

        for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
        {
                IStringWrapper elem = fInfoMap->elementAt(cursor);
                if (first) {
                        out.setRelativeIndent(1);
                        out << (char)'\n';
                        first = false;
                }

                out << elem.getkey() << " = " << elem.value() << (char)'\n';
        }
        if (!first) out.setRelativeIndent(-1);
        out << "}\n";


        //      inputs: {"-r", "0.0", "1.0"}
        out << "input: {";

        ISequence<IString>::Cursor inputCursor(fInputTexts);
        first = true;

        for (inputCursor.setToFirst(); inputCursor.isValid(); inputCursor.setToNext())
        {
                out << (first? "\"": ", \"") << inputCursor.element() << "\"";
                first = false;
        }

        out << "}\n";

}

ITieredTextBuffer& operator<<(ITieredTextBuffer& out, ITest& theTest)
{
        ITieredTextBuffer::ETier saveTier;
        unsigned short saveIndent;

        const char* testClassName = NULL;
#if IC_RTTI_ENABLED && __IBMCPP__ >= 360 // Juno bug
        const type_info& thisClass = typeid(theTest);
        if (&thisClass) {
                testClassName = thisClass.name();
        }
#endif
        out.pushTier(ITieredTextBuffer::kHeadline);
        out << "test " << testClassName
                << (theTest.success() ? " ( Pass ) {": " ( FAIL ) {");

        out.setTier(ITieredTextBuffer::kDetail);
        out.setRelativeIndent(1);
        out << (char)'\n';

        saveTier = out.tier();
        saveIndent = out.indent();
        theTest.print(out);
        out.setTier(saveTier);
        out.setIndent(saveIndent);

        out.setRelativeIndent(-1);
        out.setTier(ITieredTextBuffer::kHeadline);
        out << "}\n";
        out.popTier();

        return out;
}
