// Revision: 13 1.8.1.4 source/core/testfw/itestcol.cpp, testfw, ioc.v400, 000302 
/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/

#include <itestcol.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
    #define _ITESTCOL_CPP_
    #include <ipagetun.h>
#endif

//-----------------------------------------------------------------------------
// ITestCollection: PUBLIC Special member functions
//-----------------------------------------------------------------------------
ITestCollection::ITestCollection()
{
        fHaltOnFail = false;
        fSubtestsSetup = false;
}

ITestCollection::ITestCollection(const ITestCollection& source)
{
        fHaltOnFail = source.fHaltOnFail;
        fSubtestsSetup = source.fSubtestsSetup;
        fSubtests = source.fSubtests;
}

ITestCollection& ITestCollection::operator=(const ITestCollection& rhs)
{
        if (this != &rhs) {
                ITest::operator=(rhs);
                fHaltOnFail = rhs.fHaltOnFail;
                fSubtestsSetup = rhs.fSubtestsSetup;
                fSubtests = rhs.fSubtests;
        }
        return *this;
}


ITestCollection::~ITestCollection()
{
    IVPtrSequence<ITest*>::Cursor cursor(fSubtests);

    for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
    {
        delete  fSubtests.elementAt(cursor);
    }
}

//-----------------------------------------------------------------------------
// ITestCollection: PUBLIC TTest overrides
//-----------------------------------------------------------------------------


void ITestCollection::copyInfo()
/*
        copyInfo is an override of a ITest method. It supplies metainformation
        about this test subclass to a dictionary supplied by the caller.
*/
{
        addInfo(kDescriptionKey, "ITestCollection is a collection of ITests.");

}


//-----------------------------------------------------------------------------
// ITestCollection: PUBLIC Subtest methods
//-----------------------------------------------------------------------------
void ITestCollection::adoptTest(ITest* subTest)
{
        fSubtests.add(subTest);
}


//-----------------------------------------------------------------------------
// ITestCollection: PUBLIC Halt on fail methods
//-----------------------------------------------------------------------------
void ITestCollection::setHaltOnFail(bool haltOnFail)
{
        fHaltOnFail = haltOnFail;
}

bool ITestCollection::haltOnFail() const
{
        return fHaltOnFail;
}


//-----------------------------------------------------------------------------
// ITestCollection: PUBLIC Operational methods
//-----------------------------------------------------------------------------
void ITestCollection::randomlyReorder(long initialSeed)
/*
        randomlyReorder rearranges the order of the subtests in this set.
        randomlyReorder does this by creating a vector to hold all the subtests
        temporarily, rearranging the pointers in the vector, then adding them back
        into the original set.
*/
{
        const long kCount = fSubtests.numberOfElements();
        long i;

        if (kCount >= 2) {
                ITest** v; // Initializing this will crash CodeCheck 3.08

         IVPtrSequence<ITest*>::Cursor cursor(fSubtests);

                v = (ITest**)new char[kCount*sizeof(ITest*)];


                if (v != NULL ) {

                        // Load the vector with the pointers
                        for (i=0, cursor.setToFirst(); i < kCount, cursor.isValid();
                                i++, cursor.setToNext()) {
                                v[i] = cursor.element();
                        }

                        // Rearrange the pointers, starting with the count-th pointer and
                        // iterating down to 1 (do not swap pointer 0). For each
                        // pointer i, swap it with a randomly selected pointer 0..i.
                        {
                                long j;
                                ITest* temp;
#ifdef IC_POSIX
                                unsigned int seed = initialSeed;
#endif
                                srand(initialSeed);
                                for (i=kCount-1; i>=0; i--) {
                                    // Set j to a random value from 0 to i. Must cast carefully
                                    // here.'
#ifdef IC_POSIX
                                    j = (i+1)*rand_r(&seed)/(RAND_MAX+1);
#else
                                    j = (i+1)*rand()/(RAND_MAX+1);
#endif
                                    temp = v[j];
                                    v[j] = v[i];
                                    v[i] = temp;
                                }
                        }

                        // Reload the pointers into the collection
                        fSubtests.removeAll();
                        for (i=0; i<kCount; i++) {
                                fSubtests.add(v[i]);
                        }

                        delete v;
                        }
        }
}

//-----------------------------------------------------------------------------
// ITestCollection: PROTECTED ITest overrides
//-----------------------------------------------------------------------------
void ITestCollection::reset()
{

        IVPtrSequence<ITest*>::Cursor cursor(fSubtests);
        for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
        {
                cursor.element()->reset();
        }

        ITest::reset();
}

void ITestCollection::setup()
{
        if (!fSubtestsSetup) {
                setupSubtests();
                fSubtestsSetup = true;
        }
        IVPtrSequence<ITest*>::Cursor cursor(fSubtests);
        ITieredTextBuffer::ETier collectionTier = outputTextStream().echoTier();

        for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
        {
                cursor.element()->outputTextStream().setEchoTier(collectionTier);
        propagateInputs(*(cursor.element()));
        }

        ITest::setup();
}


void ITestCollection::test()
{
        IVPtrSequence<ITest*>::Cursor cursor(fSubtests);
        ITest* test;
        unsigned long i = 0;
        bool successes = success(), singleSuccess;
        outputTextStream().pushTier(ITieredTextBuffer::kDebug);
        outputTextStream()
                << "(ITestCollection: Collection begin)\n";
        outputTextStream().popTier();

        outputTextStream().setRelativeIndent(1);
        unsigned short collectionIndent = outputTextStream().indent();

        for (cursor.setToFirst(); cursor.isValid(); cursor.setToNext())
        {
                        i++;
                        test = cursor.element();
                        test->outputTextStream().setIndent(collectionIndent);
                        test->run();
                        singleSuccess = test->success();
                        outputTextStream() << "---[" << i << "] SubTest";
                        outputTextStream() << (singleSuccess? ": Pass\n\n": ": FAIL\n\n");
                        if (!singleSuccess) {
                                successes = false;
                                if (haltOnFail()) break;
                        }
        }

        outputTextStream().setRelativeIndent(-1);
        outputTextStream().pushTier(ITieredTextBuffer::kDebug);
        outputTextStream()
                << "(ITestCollection: Collection end)\n";
        outputTextStream().popTier();

        setSuccess(successes);
}

// Customized operator<<(ITieredTextBuffer) implementation
void ITestCollection::print(ITieredTextBuffer& out)
{
        ITest::print(out);
        out << "Collection of " << fSubtests.numberOfElements();
        out << " subtests; " << (haltOnFail()? "halts": "continues") << " on subtest failure";
        out.pushTier(ITieredTextBuffer::kHeadline);
        out     << " \n";
        out.popTier();
        out << "}\n";
}

//-----------------------------------------------------------------------------
// ITestCollection: PROTECTED Subtest methods
//-----------------------------------------------------------------------------

void ITestCollection::propagateInputs(ITest& subtest)
{

        ISequence<IString> inputs;
        copyInputs(inputs);
        subtest.setInputs(inputs);
}


void ITestCollection::setupSubtests()
{
        // subclass should override this method
}
