// Revision: 35 1.13.1.3 source/core/testfw/itimetst.cpp, testfw, ioc.v400, 000302 
/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/

#include <itimetst.hpp>
#include <iargdict.hpp>
#include <istatset.hpp>
#include <ismpclck.hpp>

#include <math.h>

// Segment definitions
#ifdef IC_PAGETUNE
    #define _ITIMETST_CPP_
    #include <ipagetun.h>
#endif

const double ITimingTest::fgDouble = 0.0;
const double *ITimingTest::fgDoublePointer = &fgDouble;
const double ITimingTest::kDefaultAutomaticSamplingTolerance = .1;
const double ITimingTest::kDefaultMinimumCorrelation = .4;
const double kUninitializedTime = -DBL_MAX;


ITimingTest::~ITimingTest ()
{
	delete fSamples;
	delete fCalibrationSamples;
	delete fClock;
}


ITimingTest::ITimingTest (const ITimingTest& source):ITest (source)
{
	fCalibrationSamples = new IStatisticalSet;
	fSamples = new IStatisticalSet;
	fClock = new ISimpleClock;
	copyDataMembers (source);
}


ITimingTest& ITimingTest::operator= (const ITimingTest& rhs)
{
	if (&rhs != this)
	{
		ITest::operator= (rhs);
		fSamples -> clear ();
		fCalibrationSamples -> clear ();
		copyDataMembers (rhs);
	}
	return *this;
}


void		ITimingTest::copyInfo ()
{
	addInfo(kInputSyntaxKey, "[-c operationsPerSample] [-s sampleCount] [-w maxAutomaticSampleWindows] [-m maxAllowedMicroseconds] [-at automaticSamplingTolerance]	[-co min correlation] [-cc {continually calibrate}] [-na {disable automatic sampling}] [-t {disable timing}]");

    addInfo(kDescriptionKey, "ITimingTest is an abstract class derived from ITest which is used to created timed performance tests.");
}


void		ITimingTest::setTimingOn (bool timingOn)
{
	fTimingIsOn = timingOn;
}


bool	ITimingTest::timingOn () const
{
	return fTimingIsOn;
}


void 		ITimingTest::setTimingCount (unsigned long timingCount)
{
	fTimingCount = timingCount;
}


unsigned long	ITimingTest::timingCount () const
{
	return fTimingCount;
}


void 		ITimingTest::setSampleCount (unsigned long sampleCount)
{
	fSampleCount = sampleCount;
}


unsigned long	ITimingTest::sampleCount () const
{
	return fSampleCount;
}


unsigned long	ITimingTest::samplesTaken () const
{
	return fSamplesTaken;
}


unsigned long	ITimingTest::samplesDiscarded () const
{
	return fDiscardedSamples;
}


void 		ITimingTest::copySamples(long& count, double*& theArray) const  // caller owns storage.
{
	fSamples -> copySamples (count, theArray);
}


void 		ITimingTest::setContinualCalibration (bool continualCalibration)
{
	fContinuallyCalibrate= continualCalibration;
}


bool	ITimingTest::continualCalibration () const
{
	return fContinuallyCalibrate;
}


void		ITimingTest::setMaximumAllowedMedianMicroseconds (double timeLimit)
{
	fMaxAllowedCalibratedMedianTime = timeLimit;
}


double	ITimingTest::maximumAllowedMedianMicroseconds () const
{
	return fMaxAllowedCalibratedMedianTime;
}


double	ITimingTest::calibratedMedianMicroseconds ()
{
	if (fCalibratedMedianMicroseconds == kUninitializedTime)
		outputTextStream () << "Warning -- Median time is not valid.\n";
	return fCalibratedMedianMicroseconds;
}


double	ITimingTest::calibratedMeanMicroseconds ()
{
	if (fCalibratedMeanMicroseconds == kUninitializedTime)
		outputTextStream () << "Warning -- Mean time is not valid.\n";
	return fCalibratedMeanMicroseconds;
}


double	ITimingTest::relativeErrorMicroseconds () const
{
	ITimingTest * const localThis = (ITimingTest * const) this;
	if (fSamplesTaken <= 1)
	{
		localThis -> outputTextStream () << "Warning -- RelativeErrorMicroseconds not valid";
		return kUninitializedTime;
	}
	else
		return fSamples->relativeError ();
}


double	ITimingTest::correlationMicroseconds () const
{
	ITimingTest * const localThis = (ITimingTest * const) this;
	if (fSamplesTaken <= 1)
	{
		localThis -> outputTextStream () << "Warning -- CorrelationMicroseconds not valid";
		return kUninitializedTime;
	}
	else
		return fSamples->correlation (1);
}


double	ITimingTest::errorMicroseconds () const
{
	ITimingTest * const localThis = (ITimingTest * const) this;
	if (fSamplesTaken <= 1)
	{
		localThis -> outputTextStream () << "Warning -- ErrorMicroseconds not valid";
		return kUninitializedTime;
	}
	else
		return (fSamples->halfInterval () + samplingErrorMicroseconds ());
}



bool	ITimingTest::errorTooLarge () const
{
	return (fAutomaticSamplingTolerance < relativeErrorMicroseconds ());
}


bool	ITimingTest::correlationTooLarge () const
{
	return (fMinimumSampleCorrelation < correlationMicroseconds ());
}


void		ITimingTest::calibrate ()
{
	const unsigned long kClockNowSamples = 20;
	IStatisticalSet  emptySamples;
	double start, end;

    // measure the overhead of a call to timeNow.
	for (int i=0; i<kClockNowSamples; i++)
		fClock->timeNow(i==0 ? start: end);
	fClockNowOverhead = (end - start)/kClockNowSamples;
	
    //measure the overhead of a function call
	for (int j = 0; j < fSampleCount; j++)
		emptySamples.addNext (oneEmptySample ());
	fEmptyMedianTime = emptySamples.median ();

	outputTextStream ().pushTier (ITieredTextBuffer::kDebug);
	outputTextStream () << "timeNow overhead = " << fClockNowOverhead << "\nFunction call overhead = " << fEmptyMedianTime << "\n";
	outputTextStream ().popTier ();

	fIsCalibrated = true;
}


double	ITimingTest::clockNowOverhead () const
{
	return fClockNowOverhead;
}


void 		ITimingTest::setAutomaticSampling(bool doAutomaticSampling)
{
	fDoAutomaticSampling = doAutomaticSampling;
}


bool 	ITimingTest::automaticSampling() const
{
	return fDoAutomaticSampling;
}


void 		ITimingTest::setAutomaticSampleTolerance(double tolerance)
{
	fAutomaticSamplingTolerance = tolerance;
}


double 	ITimingTest::automaticSampleTolerance() const
{
	return fAutomaticSamplingTolerance;
}


void 		ITimingTest::setMinimumSampleCorrelation (double correlation)
{
	fMinimumSampleCorrelation = correlation;
}


double 	ITimingTest::minimumSampleCorrelation () const
{
	return fMinimumSampleCorrelation ;
}


void 		ITimingTest::setMaximumSampleWindows(unsigned long maxWindows)
{
	fMaximumSampleWindows = maxWindows;
}


unsigned long ITimingTest::maximumSampleWindows() const
{
	return fMaximumSampleWindows;
}


ITimingTest::ITimingTest ():
	ITest (),
	fTimingCount (kDefaultTimingCount),
	fSampleCount (kDefaultSampleCount),
	fMaximumSampleWindows (kDefaultMaximumSampleWindows),
	fDiscardedSamples (0),
	fSamplesTaken (0),
	fCurrentTimingCount (0),
	fCalibratedMedianMicroseconds (kUninitializedTime),
	fCalibratedMeanMicroseconds (kUninitializedTime),
	fClockNowOverhead (kUninitializedTime),
	fEmptyMedianTime (kUninitializedTime),
	fAutomaticSamplingTolerance (kDefaultAutomaticSamplingTolerance),
	fMinimumSampleCorrelation (kDefaultMinimumCorrelation),
	fMaxAllowedCalibratedMedianTime (-1),
	fContinuallyCalibrate (false),
	fIsCalibrated (false),
	fDoAutomaticSampling (true),
	fTimingIsOn (true)
{
	fCalibrationSamples = new IStatisticalSet;
	fSamples = new IStatisticalSet;
	fClock = new ISimpleClock;
}


ITimingTest::ITimingTest (unsigned long timingCount, unsigned long sampleCount):
	ITest (),
	fTimingCount (timingCount),
	fSampleCount (sampleCount),
	fMaximumSampleWindows (kDefaultMaximumSampleWindows),
	fDiscardedSamples (0),
	fSamplesTaken (0),
	fCurrentTimingCount (0),
	fCalibratedMedianMicroseconds (kUninitializedTime),
	fCalibratedMeanMicroseconds (kUninitializedTime),
	fClockNowOverhead (kUninitializedTime),
	fEmptyMedianTime (kUninitializedTime),
	fAutomaticSamplingTolerance (kDefaultAutomaticSamplingTolerance),
	fMinimumSampleCorrelation (kDefaultMinimumCorrelation),
	fMaxAllowedCalibratedMedianTime (-1),
	fContinuallyCalibrate (false),
	fIsCalibrated (false),
	fDoAutomaticSampling (true),
	fTimingIsOn (true)
{
	fCalibrationSamples = new IStatisticalSet;
	fSamples = new IStatisticalSet;
	fClock = new ISimpleClock;
}


double	ITimingTest::oneSample ()
{
	double start, end;
	double		theTime;

	timingSetup ();

	fClock->timeNow(start);
	for (fCurrentTimingCount = 0; fCurrentTimingCount < fTimingCount; fCurrentTimingCount++)
		timingTest ();
	fClock->timeNow(end);

	timingCleanup ();

	theTime = (end - start) / fTimingCount;
	return theTime;
}


double	ITimingTest::oneEmptySample ()
{
	double start, end;
	double theTime;

	fClock->timeNow(start);
	for (int i=0; i<fTimingCount; i++)
		emptyTimingTest();
	fClock->timeNow(end);
	
	theTime = (end - start)/fTimingCount;

	return theTime;
}


double	ITimingTest::samplingErrorMicroseconds () const
{
	return (2.0 * fClockNowOverhead / fTimingCount);
}


void		ITimingTest::doUntimedTest ()
{
	timingSetup ();
	timingTest ();
	timingCleanup ();
}


void		ITimingTest::timingSetup ()
{
//To be overridden by subclass
}


void		ITimingTest::timingCleanup ()
{
//To be overridden by subclass
}


void		ITimingTest::print (ITieredTextBuffer& out)
{
	ITest::print (out);

	if (fTimingIsOn)
	{
		IString	testName;
		out.setIndent (0);
		out.pushTier(ITieredTextBuffer::kDetail);
		out 	<< "---------------------------------------------------\n";
		if (lookupInfo (kTestNameKey, testName))
			out << "test - " << testName << "\n\n";
			
		out << "TimingCount: " << timingCount () << "  \tTolerance: " <<
						automaticSampleTolerance () << "\n"
			<< "SampleCount: " << sampleCount () << " \tCorrelation Threshold: " <<
						minimumSampleCorrelation () << "\n"
			<< "SamplesTaken: " << samplesTaken () << "\tSamples Discarded: " <<
						samplesDiscarded () << "\n\n";

		out.pushTier(ITieredTextBuffer::kDebug);

		double	*samplesArray;
		long		count;

		copySamples (count, samplesArray);
		for (int i = 0; i < count; i++)
			out << "\n" << "Sample " << i << ": " << samplesArray [i];

		out.pushTier (ITieredTextBuffer::kNormal);

		double m = (floor(calibratedMedianMicroseconds()*100))/100;
		double e = (floor(errorMicroseconds()*100))/100;

		out << "\ntime " << m << " +/- " << e << " us";
		
		if (errorTooLarge ())
			out << "\tError is too large - "<< (floor(relativeErrorMicroseconds()*100))/100;
		if (correlationTooLarge ())
			out << "\tCorrelation is very high - " << (floor(correlationMicroseconds()*100))/100;

		out << "\n";

		out.popTier ();
		out.popTier ();
		out.popTier ();
	}
	else
	{
		out.pushTier(ITieredTextBuffer::kNormal);
		out << "No times taken.";
		out.popTier ();
	}
}


void		ITimingTest::setup ()		//If overridden subclass must call
{
	if (!parseArguments ())
	{
		setSuccess (false);
		return;
	}
	if (!fTimingIsOn)
	{
		setTimingCount (1);			//timingTest will only be called once.
	}
}

	
void		ITimingTest::test ()
{
	fSamples->clear ();
	fCalibrationSamples->clear ();

try{
	if (fTimingIsOn)
	{
		bool		done = false;
		unsigned long	sampleLimit = fSampleCount * fMaximumSampleWindows;
		
		fSamplesTaken = 0;
		fDiscardedSamples = 0;

		while (!done)
		{
			if (fContinuallyCalibrate)
			{
				calibrate ();
				fCalibrationSamples -> addNext (calibration ());
			}

			//The next sample is taken and added to the statistical set
			fSamples -> addNext (oneSample () - calibration ());
			fSamplesTaken++;
			if (!success ())
				break;

			//If the sample limit is hit, no further samples are taken
			//Note that in this case, the error threshold was never met.
			if (fSamplesTaken >= sampleLimit)
			{
				outputTextStream () << "Note -- Error threshold was never met";
				done = true;
			}
			//If the first window is full, test checks to see if it is finished
			else if (fSamplesTaken >= fSampleCount)
			{
				//If the first window has been filled and the error is still too
				//large, the first sample is tossed and a new one taken
				if (fDoAutomaticSampling && errorTooLarge ())
				{
					fSamples -> removeFirst ();
					if (fContinuallyCalibrate)
						fCalibrationSamples -> removeFirst ();
					++fDiscardedSamples;
				}
				//Otherwise, only fSampleCount number of samples are taken
				else
					done = true;
			}
		}
		setSuccess (true);

		fCalibratedMedianMicroseconds = fSamples -> median ();
		fCalibratedMeanMicroseconds = fSamples -> mean ();

		//test makes sure that median time is not negative
		if ((fCalibratedMedianMicroseconds + samplingErrorMicroseconds ()) < 0.0)
		{
			outputTextStream () << "test Failed -- Measured time is negative.\n";
			setSuccess (false);
		}
		//test makes sure that median time has not exceed allowed time
		if ((fMaxAllowedCalibratedMedianTime >= 0) &&
			((fCalibratedMedianMicroseconds - samplingErrorMicroseconds ()) > 						fMaxAllowedCalibratedMedianTime))
		{
			outputTextStream () << "test Failed -- Measured time exceeded allowed time.\n";
			setSuccess (false);
		}
	}
	else
	{
		doUntimedTest ();
		if (isSuccessUndecided ())
			setSuccess (true);
	}
}
catch (const IException& ie)
{
	setSuccess(false);
	throw(ie);
}

}


void		ITimingTest::cleanup ()
{
//If overridden the subclass must call this method
}


double	ITimingTest::calibration ()
{
	if (!fIsCalibrated)
		calibrate ();
	return fEmptyMedianTime;
}


bool	ITimingTest::parseArguments ()
/*
	ParseArguments parses the input arguments. It understands the following options:
		-c 	count		Sets the operations per sample (inner loop) to count
		-s 	count		Sets the sample count (outer loop) to count
		-m 	time		Sets the maximum allowed median time for a PASS to time
		-w	count		Sets the maximum sample windows
		-at	tol		Set automatic sampling tolerance to tol
		-co	corr		Set minimum correlation between samples
		-cc			Continually calibrate
		-t			Turn timing off
		-na			Turn automatic sampling off
*/
{
	static const 	IString kTimingCountKey ("-c");
	const long		kMinTimingCount = 1;
	const long		kMaxTimingCount = 5000;
		long		tempTimingCount = kDefaultTimingCount;

	static const 	IString kSampleCountKey ("-s");
	const long		kMinSampleCount = 1;
	const long		kMaxSampleCount = 50;
		long		tempSampleCount = kDefaultSampleCount;

	static const 	IString kMaxTimeKey ("-m");
	const double	kMinTime = 0.0;

	static const 	IString kToleranceKey ("-at");
	const double	kMinTolerance = 0.0;
	const double	kMaxTolerance = 10.0;

	static const 	IString kCorrelationKey ("-co");
	const double	kMinimumCorrelation = -1.0;
	const double	kMaxCorrelation = 1.0;

	static const 	IString kMaxWindowsKey ("-w");
	const long		kMinWindows = 1;
	const long		kMaxWindows = 10;
		long		tempMaxSampleWindows = kDefaultMaximumSampleWindows;

	static const IString kContinuallyCalibrateKey ("-cc");
	static const IString kDisableTiming ("-t");
	static const IString kDisableAutomaticSampling ("-na");

	static const IString kNakedOptions("-cc -t -na");

	bool goodParse = true;

	IArgumentDictionary args(*this, kNakedOptions);

	IString value;
	if (args.textAt (kContinuallyCalibrateKey, value))
		setContinualCalibration (true);
	if (args.textAt (kDisableTiming, value))
		setTimingOn (false);
	if (args.textAt (kDisableAutomaticSampling, value))
		setAutomaticSampling (false);

	if  (! ((args.textAt (kTimingCountKey, value) ? args.numberAt (kTimingCountKey,
			tempTimingCount, kMinTimingCount, kMaxTimingCount) : true)
		&& (args.textAt (kSampleCountKey, value) ? args.numberAt (kSampleCountKey,
			tempSampleCount, kMinSampleCount, kMaxSampleCount) : true)
		&& (args.textAt (kMaxTimeKey, value) ? args.numberAt (kMaxTimeKey,
			fMaxAllowedCalibratedMedianTime,kMinTime, DBL_MAX) : true)
		&& (args.textAt (kMaxWindowsKey, value) ? args.numberAt (kMaxWindowsKey,
			tempMaxSampleWindows, kMinWindows, kMaxWindows) : true)
		&& (args.textAt (kCorrelationKey, value) ? args.numberAt(kCorrelationKey,
			fMinimumSampleCorrelation, kMinimumCorrelation, kMaxCorrelation) : true)
		&& (args.textAt (kToleranceKey, value) ? args.numberAt (kToleranceKey,
			fAutomaticSamplingTolerance, kMinTolerance, kMaxTolerance) : true)))
	{
		setSuccess(false);
		goodParse = false;
	}

if (tempTimingCount != kDefaultTimingCount)
   	setTimingCount (tempTimingCount);
if (tempSampleCount != kDefaultSampleCount)
   	setSampleCount (tempSampleCount);
if (tempMaxSampleWindows != kDefaultMaximumSampleWindows)
   	setMaximumSampleWindows (tempMaxSampleWindows);

	return goodParse;
}

double	ITimingTest::emptyTimingTest ()
{
	return *fgDoublePointer;
}


void		ITimingTest::copyDataMembers (const ITimingTest& other)
{
	fTimingCount = other.fTimingCount;
	fSampleCount = other.fSampleCount;
	fMaximumSampleWindows = other.fMaximumSampleWindows;
	fDiscardedSamples = other.fDiscardedSamples;
	fSamplesTaken = other.fSamplesTaken;
	fCurrentTimingCount = other.fCurrentTimingCount;
	fCalibratedMedianMicroseconds = other.fCalibratedMedianMicroseconds;
	fCalibratedMeanMicroseconds = other.fCalibratedMeanMicroseconds;
	fClockNowOverhead = other.fClockNowOverhead;
	fEmptyMedianTime = other.fEmptyMedianTime;
	fAutomaticSamplingTolerance = other.fAutomaticSamplingTolerance;
	fMinimumSampleCorrelation = other.fMinimumSampleCorrelation;
	fMaxAllowedCalibratedMedianTime = other.fMaxAllowedCalibratedMedianTime;
	fContinuallyCalibrate = other.fContinuallyCalibrate;
	fIsCalibrated = other.fIsCalibrated;
	fDoAutomaticSampling = other.fDoAutomaticSampling;
	fTimingIsOn = other.fTimingIsOn;

	//Note the IStatisticalSet override of '='
	*fCalibrationSamples = *other.fCalibrationSamples;
	*fSamples = *other.fSamples;
}
