/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 64 1.11.1.4 source/core/text/intl/compare/icolitr.cpp, text, ioc.v400, 000302 
//===============================================================================
//
// File icolitr.cpp : Collation Iterator
//
//
//
// Created by: Helena Shih
//
// Modification History:
//
//      Date         Name          Description
//
//===============================================================================
//
#ifndef IC_ICOLITR
#include "icolitr.hpp"
#endif

#ifndef IC_ICLITIMP
#include "iclitimp.hpp"
#endif

#include <iexcept.hpp>

#ifdef IC_PAGETUNE
        #define ICOLITR_CPP
        #include <ipagetun.h>
#endif

ICollationIterator::ICollationIterator(ICollationIterator::ECollationType type)
{
        fCollType = type;
        switch(type) {
                case ICollationIterator::kHostCollationsOnly:
                        fHostCollations = new IHostCollationIterator();
                        fCollationScope = ICollationIterator::kHostCollationSet;
                        break;
                case ICollationIterator::kPortableCollationsOnly:
                        fPortableCollations = new IPortableCollationIterator();
                        fCollationScope = ICollationIterator::kPortableCollationSet;
                        break;
                default:
                        fHostCollations = new IHostCollationIterator();
                        fPortableCollations = new IPortableCollationIterator();
                        fCollationScope = ICollationIterator::kHostCollationSet;
                        break;
        }
}

ICollationIterator::ICollationIterator(const ICollationIterator& that)
                :       fCollationScope(that.fCollationScope),
                        fCollType(that.fCollType)
{
        if (fCollType == ICollationIterator::kPortableCollationsOnly) {
                fPortableCollations = new IPortableCollationIterator(*that.fPortableCollations);
        } else if (fCollType == ICollationIterator::kHostCollationsOnly) {
                fHostCollations = new IHostCollationIterator(*that.fHostCollations);
        } else {
                fHostCollations = new IHostCollationIterator(*that.fHostCollations);
                fPortableCollations = new IPortableCollationIterator(*that.fPortableCollations);
        }
}

ICollationIterator::~ICollationIterator()
{
        if (fCollType == ICollationIterator::kPortableCollationsOnly) {
                delete fPortableCollations;
        } else if (fCollType == ICollationIterator::kHostCollationsOnly) {
                delete fHostCollations;
        } else {
                delete fPortableCollations;
                delete fHostCollations;
        }
        fCollationScope = ICollationIterator::kHostCollationSet;
        fCollType = ICollationIterator::kAllCollations;
}

// assignment
ICollationIterator&
ICollationIterator:: operator=(const ICollationIterator& that)
{
        fCollationScope = that.fCollationScope;
        fCollType = that.fCollType;
        if (fCollType == ICollationIterator::kPortableCollationsOnly) {
                *fPortableCollations = *(that.fPortableCollations);
        } else if (fCollType == ICollationIterator::kHostCollationsOnly) {
                *fHostCollations = *(that.fHostCollations);
        }
        return *this;
}

// comparison
bool
ICollationIterator::operator==(const ICollationIterator& that) const
{
        if ((fCollationScope == that.fCollationScope) &&
                (fCollType == that.fCollType))
        {
                if (fCollType == ICollationIterator::kPortableCollationsOnly)
                {
                        return (*fPortableCollations == *(that.fPortableCollations));
                }
                if (fCollType == ICollationIterator::kHostCollationsOnly)
                {
                        return (*fHostCollations == *(that.fHostCollations));
                }
                else
                {
                        return ( (*fPortableCollations == *(that.fPortableCollations)) &&
                                         (*fHostCollations == *(that.fHostCollations)) );
                }
        }
        return false;
}

bool
ICollationIterator::operator!=(const ICollationIterator& that) const
{
        return !(*this == that);
}

// cursor actions
void
ICollationIterator::reset()
{
        if (fCollType != ICollationIterator::kPortableCollationsOnly) {
                fHostCollations->reset();
                fCollationScope = ICollationIterator::kHostCollationSet;
        } else {
                fPortableCollations->reset();
                fCollationScope = ICollationIterator::kPortableCollationSet;
        }
}

ICollationIterator::operator bool() const
{
        if (fCollType == ICollationIterator::kPortableCollationsOnly) {
                return (fPortableCollations->operator bool());
        } else if (fCollType == ICollationIterator::kHostCollationsOnly) {
                return(fHostCollations->operator bool());
        } else {
                if (!fHostCollations->operator bool())
                        return fPortableCollations->operator bool();
        }
        return true;
}

ICollation*
ICollationIterator::create(ICollation::ECollationStrength strength) const
{
        if (fCollationScope == ICollationIterator::kHostCollationSet) {
                return fHostCollations->create(strength);
        } else {
                return fPortableCollations->create(strength);
        }
}

IText
ICollationIterator::localePOSIXID() const
{
        if (fCollationScope == ICollationIterator::kHostCollationSet) {
                return fHostCollations->localePOSIXID();
        } else {
                return fPortableCollations->localePOSIXID();
        }
}

ICollationIterator&
ICollationIterator::operator++()         // prefix
{
        try {
                if (fCollType == ICollationIterator::kHostCollationsOnly) {
                        fHostCollations->operator++();
                } else if (fCollType == ICollationIterator::kPortableCollationsOnly) {
                        fPortableCollations->operator++();
                } else if (fCollationScope == ICollationIterator::kHostCollationSet) {
                        if ((!fHostCollations->operator bool()) &&
                                (fCollationScope == ICollationIterator::kHostCollationSet)) {
                                fCollationScope = ICollationIterator::kPortableCollationSet;
                        } else if (fCollationScope == ICollationIterator::kHostCollationSet) {
                                fHostCollations->operator++();
                        } else {
                                fPortableCollations->operator++();
                        }
                }
        }
        catch (...) {
                ITHROW(IInvalidRequest("Invalid advancing request for collation iterator", IBaseErrorInfo::invalidRequest, IException::recoverable));
        }
        return *this;
}

ICollationIterator
ICollationIterator::operator++(int)     // postfix
{
        ICollationIterator tmp(*this);
        try {
                if (fCollType == ICollationIterator::kHostCollationsOnly) {
                        fHostCollations->operator++();
                } else if (fCollType == ICollationIterator::kPortableCollationsOnly) {
                        fPortableCollations->operator++();
                } else if (fCollationScope == ICollationIterator::kHostCollationSet) {
                        if ((!fHostCollations->operator bool()) &&
                                (fCollationScope == ICollationIterator::kHostCollationSet)) {
                                fCollationScope = ICollationIterator::kPortableCollationSet;
                        } else if (fCollationScope == ICollationIterator::kHostCollationSet) {
                                fHostCollations->operator++();
                        } else {
                                fPortableCollations->operator++();
                        }
                }
        }
        catch (...) {
                ITHROW(IInvalidRequest("Invalid advancing request for collation iterator", IBaseErrorInfo::invalidRequest, IException::recoverable));
        }
        return tmp;
}

// displayable key name
IText
ICollationIterator::displayName(const ILocaleKey& desiredLanguage,
                                                                ILocaleKey::ELookupStrategy strategy) const
{
        if (fCollationScope == ICollationIterator::kHostCollationSet) {
                return (fHostCollations->displayName(desiredLanguage, strategy));
        } else {
                return (fPortableCollations->displayName(desiredLanguage, strategy));
        }
}
