// Revision: 66 1.7.6.5 source/core/text/intl/compare/itxtcoll.cpp, text, ioc.v400, 000302 
/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
//===============================================================================
//
// File itxtcoll.cpp
//
//
//
// Created by: Helena Shih
//
// Modification History:
//
//      Date         Name          Description
//
//===============================================================================

#ifndef IC_IEXCBASE
#include <iexcbase.hpp>
#endif

#ifndef IC_IUGNRL
#include <iugnrl.hpp>
#endif

#ifndef IC_ITXTCOLL
#include "itxtcoll.hpp"
#endif

#ifndef IC_ITXCLIMP
#include "itxclimp.hpp"
#endif

#ifdef IC_PAGETUNE
        #define _ITXTCOLL_CPP
        #include <ipagetun.h>
#endif

ICollation*
ICollation::createCollation(const ILocaleKey& localeId,
                                                        ICollation::ECollationStrength strength)
{
        if (localeId.isCLocale()) {
                return (new IBitwiseCollation());
        }
        try {
#ifdef _WIN32
                return (new IWin32Collation(localeId, strength));
#endif
#ifdef USE_POSIX
                return (new IPOSIXCollation(localeId, strength));
#endif
        } catch (...) {
        return (IPortableCollation::createCollation(localeId, strength));
        }
}

ICollation::ICollation(void)
{
        fLocaleKey = ILocaleKey::defaultLocale();
        fStrength = ICollation::kTertiaryDifference;
}

ICollation::ICollation (const ICollation& order)
        : fLocaleKey(order.fLocaleKey), fStrength(order.fStrength)
{
}

ICollation::~ICollation(void)
{
}

ICollation&
ICollation::operator=(const ICollation& that)
{
        if (this != &that) {
                fLocaleKey = that.fLocaleKey;
                fStrength = that.fStrength;
        }
        return *this;
}

UniChar*
ICollation::createUniCharChunk( const IText& source,
                                                                IText::length_type& chunkLength) const
{
        UniChar *sourceStr = 0;
        IText::length_type sourceLen = source.length();
        sourceStr = new UniChar[sourceLen + 1];
#pragma info(none)
        sourceLen = source.copy(sourceStr, sourceLen, 0);
#pragma info(restore)
        sourceStr[sourceLen] = 0;
        return sourceStr;
}

int
ICollation::compare(const IText& source,
                                        const IText& target) const
{
        int sourceLen = source.length(), targetLen = target.length();
        UniChar *sourceLow = 0, *targetLow = 0;
        IText::length_type sourceChunkLength = 0, targetChunkLength = 0;

        if ((sourceLen == 0) && (targetLen == 0)) {
                return 0;
        } else if ((sourceLen == 0) && (targetLen > 0)) {
                return -1;
        } else if ((sourceLen > 0) && (targetLen == 0)) {
                return 1;
        }
        sourceLow = createUniCharChunk(source, sourceChunkLength);
        targetLow = createUniCharChunk(target, targetChunkLength);
        UniChar *sourceHigh = &(sourceLow[sourceLen-1]),
                        *targetHigh = &(targetLow[targetLen-1]);
        int result = do_compare(sourceLow, sourceHigh, targetLow, targetHigh);
        if (sourceChunkLength < sourceLen)
                delete [] sourceLow;
        if (targetChunkLength < targetLen)
                delete [] targetLow;
        return (result);
}

int
ICollation::compare(ITextIterator sourceBegin,
                                        ITextIterator sourceEnd,
                                        ITextIterator targetBegin,
                                        ITextIterator targetEnd) const
{
        if ((sourceEnd < sourceBegin) || (targetEnd < targetBegin))
                ITHROW(IInvalidParameter("Incorrect text iterators!"));
        return (compare(IText(sourceBegin, sourceEnd), IText(targetBegin, targetEnd)));
}
int
ICollation::compare(const UniChar* sourceLow,
                                        const UniChar* sourceHigh,
                                        const UniChar* targetLow,
                                        const UniChar* targetHigh) const
{
        return (do_compare(sourceLow, sourceHigh, targetLow, targetHigh) );
}

IText
ICollation::transform(const IText& source) const
{
        if (source.length() == 0) {
                return IText();
        } else {
                return (do_transform(source));
        }
}

IText
ICollation::transform(ITextIterator sourceBegin,
                                        ITextIterator sourceEnd) const
{
        if (sourceEnd < sourceBegin)
                ITHROW(IInvalidParameter("Incorrect text iterators!"));
        return (do_transform(IText(sourceBegin, sourceEnd)));
}

long
ICollation::hash(const UniChar* sourceLow,
                                const UniChar* sourceHigh) const
{
        if (sourceHigh < sourceLow) {
                ITHROW(IInvalidParameter("Incorrect text offsets!"));
        }
        return (do_hash(sourceLow, sourceHigh) );
}

long
ICollation::hash(const IText& source) const
{
        return source.hash();
}

long
ICollation::hash(ITextIterator sourceBegin,
                                ITextIterator sourceEnd) const
{
        if (sourceEnd < sourceBegin)
                ITHROW(IInvalidParameter("Incorrect text iterators!"));
        IText tmp(sourceBegin, sourceEnd);
        return (hash(tmp));
}

bool
ICollation::isEqual(const IText& sourceText,
                                        const IText& targetText) const
{
        if (this->compare(sourceText, targetText) == ICollation::kSourceEqual)
                return (true);
        else
                return (false);
}

bool
ICollation::isGreaterThan(      const IText& sourceText,
                                                        const IText& targetText) const
{
        if (this->compare(sourceText, targetText) > ICollation::kSourceEqual)
                return (true);
        else
                return (false);
}


bool
ICollation::isLessThan(const IText& sourceText,
                                                const IText& targetText) const
{

        if (this->compare(sourceText, targetText) < ICollation::kSourceEqual)
                return (true);
        else
                return (false);
}

long
ICollation::do_hash(const UniChar* sourceLow,
                                        const UniChar* sourceHigh) const
{
        IText tmp(sourceLow, sourceHigh-sourceLow);
        return tmp.hash();
}

void
ICollation::setStrength(ICollation::ECollationStrength newStrength)
{
        fStrength = newStrength;
}

ICollation::ECollationStrength
ICollation::strength() const
{
        return fStrength;
}

void
ICollation::setLocaleKey(const ILocaleKey& newLocale)
{
        fLocaleKey = newLocale;
}

ILocaleKey
ICollation::localeKey() const
{
        return fLocaleKey;
}
//===============================================================================
//
// Functions for IBitwiseTextOrder
//
//===============================================================================


IBitwiseCollation::IBitwiseCollation () :
    ICollation()
{
}


IBitwiseCollation::IBitwiseCollation (const IBitwiseCollation& order):
        ICollation(order)

{
}

IBitwiseCollation::~IBitwiseCollation()
{

}

IBitwiseCollation&
IBitwiseCollation::operator=(const IBitwiseCollation& that)
{
        return *this;
}

int
IBitwiseCollation::do_compare(const UniChar* sourceLow,
                                                        const UniChar* sourceHigh,
                                                        const UniChar* targetLow,
                                                        const UniChar* targetHigh) const
{
        bool done = false;
        int result;
        UniChar *source, *target;

        if ((sourceHigh < sourceLow) || (targetHigh < targetLow)) {
                ITHROW(IInvalidParameter("Incorrect text offsets!"));
        }
#pragma info(none)
        for (source = (UniChar*)sourceLow, target = (UniChar*)targetLow;
                 ((target < targetHigh) && (source < sourceHigh));
                 source++, target++) {
                if (*source != *target) {
                        if (*source > *target) {
                                result = ICollation::kSourceGreater;
                        } else if (*source < *target) {
                                result = ICollation::kSourceLess;
                        }
                        done = true;
                        break;
                }
        }
#pragma info(restore)
        if (source >= sourceHigh && (done != true)) {
                if (target >= targetHigh) {
                        result = ICollation::kSourceEqual;
                } else {
                        result = ICollation::kSourceLess;
                }
        } else if (target >= targetHigh && (done != true)) {
                result = ICollation::kSourceGreater;
        }

        return (result);
}

IText
IBitwiseCollation::do_transform(const IText& source) const
{
        return source;
}

#pragma info(nouse)
IText
IBitwiseCollation::displayName( const ILocaleKey& desiredLanguage,
                                                                ILocaleKey::ELookupStrategy strategy) const
{
        return (localeKey().posixID());
}

#pragma info(restore)
//===============================================================================
//
// Functions for IPortableCollation
//
//===============================================================================

IText::length_type IPortableCollation::nlength = 0;

IPortableCollation::IPortableCollation()
        : ICollation()
{
}

IPortableCollation::IPortableCollation(const IText& rules)
{
        fCollationRules = rules;
}

#pragma info(nouse)
IPortableCollation*
IPortableCollation::createCollation(const ILocaleKey& localeKey,
                                                                ICollation::ECollationStrength strength)
{
        return 0;
//      return (new ITableBasedCollation());

}
#pragma info(restore)

IPortableCollation::IPortableCollation(const IPortableCollation& order)
        : ICollation(order)
{
}

IPortableCollation::~IPortableCollation(void)
{
        setLocaleKey(ILocaleKey::defaultLocale());
        this->setStrength(ICollation::kTertiaryDifference);
}

IPortableCollation&
IPortableCollation::operator=(const IPortableCollation& that)
{
        return *this;
}

#pragma info(nouse)
IText
IPortableCollation::displayName(const ILocaleKey& desiredLanguage,
                                                                ILocaleKey::ELookupStrategy strategy) const
{
        return IText();
//      return (((ILocaleKey&)localeKey()).displayName(desiredLanguage, strategy));
}
#pragma info(restore)

int
IPortableCollation::compare(const UniChar *sourceLow,
                                                        const UniChar* sourceHigh,
                                                        const UniChar* targetLow,
                                                        const UniChar* targetHigh,
                                                        IText::length_type& matchedSource,
                                                        IText::length_type& matchedTarget) const
{
        if ((sourceHigh < sourceLow) || (targetHigh < targetLow)) {
                ITHROW(IInvalidParameter("Incorrect text offsets!"));
        }
        return (do_matchCompare(sourceLow, sourceHigh, targetLow, targetHigh, matchedSource, matchedTarget));
}


int
IPortableCollation::compare(const ITextIterator& sourceBegin,
                                                        const ITextIterator& sourceEnd,
                                                        const ITextIterator& targetBegin,
                                                        const ITextIterator& targetEnd,
                                                        IText::length_type& matchedSource,
                                                        IText::length_type& matchedTarget) const
{
        if ((sourceEnd < sourceBegin) || (targetEnd < targetBegin))
                ITHROW(IInvalidParameter("Incorrect text iterators!"));
        return (compare(IText(sourceBegin, sourceEnd), IText(targetBegin, targetEnd), matchedSource, matchedTarget));
}

int
IPortableCollation::compare(const IText& source,
                                                        const IText& target,
                                                        IText::length_type& matchedSource,
                                                        IText::length_type& matchedTarget) const
{

        int sourceLen = source.length(), targetLen = target.length();
        UniChar *sourceLow = 0, *targetLow = 0;
        IText::length_type sourceChunkLength = 0, targetChunkLength = 0;

        if ((sourceLen == 0) && (targetLen == 0)) {
                return 0;
        } else if ((sourceLen == 0) && (targetLen > 0)) {
                return -1;
        } else if ((sourceLen > 0) && (targetLen == 0)) {
                return 1;
        }
        sourceLow = createUniCharChunk(source, sourceChunkLength);
        targetLow = createUniCharChunk(target, targetChunkLength);
        UniChar *sourceHigh = &(sourceLow[sourceLen-1]),
                        *targetHigh = &(targetLow[targetLen-1]);
        int result = compare(sourceLow, sourceHigh, targetLow, targetHigh, matchedSource, matchedTarget);
        if (sourceChunkLength < sourceLen)
                delete [] sourceLow;
        if (targetChunkLength < targetLen)
                delete [] targetLow;
        return (result);
}

int
IPortableCollation::do_compare(const UniChar* sourceLow,
                                                                 const UniChar* sourceHigh,
                                                                 const UniChar* targetLow,
                                                                 const UniChar* targetHigh) const
{
        IText::length_type len1;
        IText::length_type len2;

        if ((sourceHigh < sourceLow) || (targetHigh < targetLow)) {
                ITHROW(IInvalidParameter("Incorrect text offsets!"));
        }
        return (do_matchCompare(sourceLow, sourceHigh, targetLow, targetHigh, len1, len2));
}

