/**
 * ISimpleTextBoundary is a concrete implementation of ITextBoundary that uses a state machine.
 *
 * @package     Text and International
 * @category    Text Scanning
 *
 * @copyright
 *              IBM Open Class Library
 *              (C) Copyright International Business Machines Corporation 1997
 *              Licensed Material - Program-Property of IBM - All Rights Reserved.
 *              US Government Users Restricted Rights - Use, duplication, or disclosure
 *              restricted by GSA ADP Schedule Contract with IBM Corp.
 */

#ifndef _ISIMTXBD_
#define _ISIMTXBD_

#include <iprimtyp.hpp>
#include <itext.hpp>

#include <itxtbdry.hpp>
#include <itxtbdat.hpp>

#if __IBMCPP__ >= 400
#pragma namemangling(compat)
#endif

#pragma pack(push,4)
#pragma enum(4)

/**
 * ISimpleTextBoundary is a concrete implementation of ITextBoundary.
 * ISimpleTextBoundary uses a state machine to compute breaks.
 * Different state machines are available that compute breaks for
 * sentences, words, lines, and characters.  They are accessable
 * through static functions of ITextBoundary.
 * @see ITextBoundary
 */
class ISimpleTextBoundary: public ITextBoundary {
public:
    /** Destructor. */
    virtual ~ISimpleTextBoundary();

    /**
     * Return a polymorphic copy of this ITextBoundary.
     */
    virtual ISimpleTextBoundary* clone() const;

    /** Return the text over which this operates. */
    virtual IText text() const;

    /**
     * Change the text over which this operates. The text boundary is
     * reset to the start.
     */
    virtual void  setText(const IText&);

    /**
     * Return the index of the first character in the text being scanned.
     */
    virtual offset_type first();

    /**
     * Return the index of the last character in the text being scanned.
     */
    virtual offset_type last();

    /**
     * Return the character index of the previous text boundary, or kDone if all
     * boundaries have been returned.
     */
    virtual offset_type previous();

    /**
     * Return the character index of the next text boundary, or kDone if all
     * boundaries have been returned.
     */
    virtual offset_type next();

    /**
     * Return the character index of the text boundary that was most recently
     * returned by next(), previous(), first(), or last().
     */
    virtual offset_type current() const;

    /**
     * Return the first boundary after the specified offset.
     * The value returned is always greater than the offset, or is kDone.
     * @param offset the offset to begin scanning.
     * @return The first boundary after the specified offset.
     */
    virtual offset_type nextAfter(offset_type offset);

    /**
     * Return the nth boundary from the current boundary.
     * @param n the signed number of boundaries to traverse.
     * Negative values move to previous boundaries
     * and positive values move to later boundaries. A value of 0 does nothing.
     * @return The character index of the nth boundary from the current position.
     */
    virtual offset_type nthFromCurrent(int n);

    /**
     * Return true if this ITextBoundary is at the same position in the same text,
     * and is the same class and type (word, line, etc.) of ITextBoundary,
     * as the argument.  Text is considered the same if it contains the same
     * characters, it need not be the same object, and styles are not considered.
     */
    virtual bool operator==(const ITextBoundary&) const;

private:
    /**
     * Construct an ISimpleTextBoundary from the provided state table.  This protected
     * constructor is called from the friend class ITextBoundary.
     */
    ISimpleTextBoundary(const ITextBoundaryData* data);

    ISimpleTextBoundary(const ISimpleTextBoundary&); // only used by clone

    /** The assignment operator is required to satisfy the compiler, but never called. */
    ISimpleTextBoundary operator=(const ISimpleTextBoundary&) { return *this; } // do not call

    /**
     * Internal utility used to locate the previous position from which it is safe
     * to do a forward scan (state is known).
     */
    virtual offset_type previousSafePosition(offset_type offset);

    /**
     * Internal utility to get the next position.
     */
    virtual offset_type nextPosition(offset_type offset);

    const ITextBoundaryData* fData;
        const IWordBreakTable*          fForward;       // fData->forward()
        const IWordBreakTable*          fBackward;      // fData->backward()
        const IUnicodeClassMapping*     fMap;           // fData->map()
    IText fText;
    offset_type fPos;

    friend class ITextBoundary;
};

#pragma enum(pop)
#pragma pack(pop)

#if __IBMCPP__ >= 400
#pragma namemangling()
#endif

#endif // _ISIMTXBD_
