/**
 * IWordBreakTable implements a state transition table.
 *
 * @package     Text and International
 * @category    Text Scanning
 *
 * @copyright
 *              IBM Open Class Library
 *              (C) Copyright International Business Machines Corporation 1997
 *              Licensed Material - Program-Property of IBM - All Rights Reserved.
 *              US Government Users Restricted Rights - Use, duplication, or disclosure
 *              restricted by GSA ADP Schedule Contract with IBM Corp.
 */

#ifndef _IWDBKTBL_
#define _IWDBKTBL_

#include <iprimtyp.hpp>
#include <itxtbdat.hpp>

#if __IBMCPP__ >= 400
#pragma namemangling(compat)
#endif

#pragma pack(push,4)
#pragma enum(4)

/**
 * This class implements a state transition table.
 * After each transition, using the get method, the
 * new state is returned along with information about
 * the state change (ex. was it a "marked" transition").
 * For efficiency, none of the arguments to any of these
 * methods are validated.
 */
class IWordBreakTable {
public:
        // For convenience
        typedef ITextBoundaryData::Node Node;
        typedef ITextBoundaryData::Type Type;

        /**
         * Construct a table from the provided data.  See CharacterBreakData or
         * the other ITextBoundaryData subclasses for examples. Each row represents
         * a state, each column within a row represents a transition.  The values
         * in the table represent the new state and mark information.
         * @param cols number of columns in the table (transitions)
         * @param data an encoded byte array containing state and transition data
         * @param data_length the length of the byte array data
         */
    IWordBreakTable(int cols, const Node data[], int data_length);

    /**
     * Get the resulting state moving from oldState accepting input.
     * @param oldState current state
     * @param input input
     * @return resulting state and transition data
     */
        Node get(Node oldState, Type input) const;

    /**
     * Checks to see if the transition into the specified state was "marked."
     * @param state the state as returned by get, initialState, or endState
     * @return true if transition into state was marked.
     */
        bool isMarkState(Node state) const;

    /**
     * Check to see if the state is the end state
     * @param state the state to check
     * @return true if state is an end state
     */
        bool isEndState(Node state) const;

    /**
     * Get the initial state
     * @return the initial state
     */
        Node initialState() const;

private:
        static const Node  kMark_mask;
        static const Node  kIndex_mask;

        static const Node   kInitial_state;
        static const Node   kEnd_state;

        const ITextBoundaryData::Node* fData;
    int fData_length;
    int fCols;
};

#pragma enum(pop)
#pragma pack(pop)

#if __IBMCPP__ >= 400
#pragma namemangling()
#endif

// -------------------------------------

inline IWordBreakTable::Node
IWordBreakTable::get(IWordBreakTable::Node oldState, IWordBreakTable::Type input) const
{
        return fData[(oldState & kIndex_mask) * fCols + input];
}

inline bool
IWordBreakTable::isMarkState(IWordBreakTable::Node state) const
{
        return (state & kMark_mask) != 0;
}

inline bool
IWordBreakTable::isEndState(IWordBreakTable::Node state) const
{
        return (state & kIndex_mask) == kEnd_state;
}

inline IWordBreakTable::Node
IWordBreakTable::initialState() const
{
        return kInitial_state;
}

#endif // _IWDBKTBL_
