// Revision: 35 1.11.1.5 source/core/text/intl/transcod/itrancod.cpp, text, ioc.v400, 000302 
/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
//
//  File: transcod.cpp
//  Function: Portable Transcoder Framework implementation
//  Author: Chen-Lieh Huang
//  Date created: 03/19/1996
//
#include "iautoptr.hpp" // IAutoArray

#ifndef IC_ITRANCOD
#include "itrancod.hpp"
#endif

#ifndef IC_TRANSCOD
#include "transcod.h"
#endif

#ifdef IC_PAGETUNE
        #define _ITRANCOD_CPP
        #include <ipagetun.h>
#endif

// Transcoder exception handling utilities
const char* kMessageFile = kMsgFileName;        // Exception message file

static const int        kBufSize = 256; // lucky guess!


//
//      ITranscoder class implementation
//
ITranscoder::ITranscoder()
        :       fUnmappedBehavior (kUseSub),
                fCharSubstitute (0x1A),
                fUniCharSubstitute (0xFFFD),
                fCharacterEncoding (IText())
{
}


ITranscoder::~ITranscoder()
{
}


ITranscoder::ITranscoder(const ITranscoder &source)
        : fUnmappedBehavior(source.fUnmappedBehavior),
                fCharSubstitute(source.fCharSubstitute),
                fUniCharSubstitute(source.fUniCharSubstitute),
          fCharacterEncoding(source.fCharacterEncoding)
{
}


ITranscoder &
ITranscoder::operator=(const ITranscoder &right)
{
        if (this != &right)
        {
                fUnmappedBehavior = right.fUnmappedBehavior;
                fCharSubstitute = right.fCharSubstitute;
                fUniCharSubstitute = right.fUniCharSubstitute;
                fCharacterEncoding = right.fCharacterEncoding;
        }
        return *this;
}


ITranscoder::EUnmappedBehavior
ITranscoder::unmappedBehavior()
{
        return fUnmappedBehavior;
}


void
ITranscoder::setUnmappedBehavior(EUnmappedBehavior unmappedBehavior)
{
        fUnmappedBehavior = unmappedBehavior;
}


char
ITranscoder::charSubstitute()
{
        return fCharSubstitute;
}


void
ITranscoder::setCharSubstitute(char substitute)
{
        fCharSubstitute = substitute;
}


UniChar
ITranscoder::uniCharSubstitute()
{
        return fUniCharSubstitute;
}


IText
ITranscoder::characterEncoding() const
{
        return fCharacterEncoding;
}


void
ITranscoder::setCharacterEncoding(const IText& encoding)
{
        fCharacterEncoding = encoding;
}


void
ITranscoder::resetState()
{
        // do nothing
}


ITranscoder::result
ITranscoder::flush(const char* to, const char* to_limit, char*& to_next)
{
        // do nothing in ITranscoder class, and simply checks parameters and
        // returns noconv status
        if (to == NULL || to > to_limit)
        {
                IMessageText text(kFlushBufRangeError, kMessageFile);
                IInvalidParameter exc(text, kFlushBufRangeError, IException::recoverable);
                ITHROW(exc);
        }

        return codecvt_base::noconv;
}


ITranscoder::result
ITranscoder::toUnicode(const IString& from, IText& to)
{
        if (from.length() == 0)
        {
                to = IText();
                return codecvt_base::noconv;
        }

        IString                 copy = from;
        const char* from_begin = copy;
        const char* from_end = from_begin + from.length();
        const char* from_next = NULL;

        length_type     uniBufSize = uniCharBufferSize(from);

        UniChar                 stackBuffer[kBufSize+1];
        UniChar*                to_begin = stackBuffer;
        IAutoArray<UniChar> heapBuf = 0;

        if (uniBufSize > kBufSize) {
                // There wasn't enough room in our fixed-size buffer.
                // Allocate a dynamic one and try again.
                heapBuf.reset( new UniChar[uniBufSize+1] );
                to_begin = heapBuf.get();
        }

        UniChar*        to_limit = to_begin + uniBufSize;
        UniChar*        to_next = NULL;

        result status = doToUnicode(
                from_begin, from_end, from_next, to_begin, to_limit, to_next);
        *to_next = 0x0000;

        // Note: Do not use "to = IText(to_begin);". Use the constructor
        // that takes a UniChar* and a actual length. Otherwise,
        // if the transcded data actually has null characters in it,
        // IText constructor will interpret the first null character
        // as the end of the string, losing data.
        to = IText(to_begin, to_next-to_begin);

        return status;
}


ITranscoder::result
ITranscoder::fromUnicode(const IText& from, IString& to)
{
        if (from.empty())
        {
                to = IString();
                return codecvt_base::noconv;
        }

        IText   itext_from = from;
        length_type     from_length = itext_from.length();

        const UniChar*  from_begin;
        UniChar*                                tmp;
        bool                                            mustDispose = false;
        size_t                                  dummy;
        size_t                                  chunkLength;

        from_begin = itext_from.storage_chunk(0, dummy, chunkLength);
        if (chunkLength < from_length) {        // if from_length > 4k (current default)
                mustDispose = true;
                tmp = new UniChar[from_length];
                itext_from.copy(tmp, IText::npos);
                from_begin = tmp;
        }

        const UniChar* from_end = from_begin + from_length;
        const UniChar* from_next = NULL;

        length_type     charBufSize = byteBufferSize(from);

        char            stackBuffer[kBufSize+1];
        char*           to_begin = stackBuffer;
        IAutoArray<char> heapBuf = 0;

        if (charBufSize > kBufSize) {
                heapBuf.reset( new char[charBufSize+1] );
                to_begin = heapBuf.get();
        }

        char*   to_limit = to_begin + charBufSize;
        char*   to_next = NULL;

        result status = doFromUnicode(
                from_begin, from_end, from_next, to_begin, to_limit, to_next);
        *to_next = '\0';

        to = IString(to_begin);

        if (mustDispose)
                delete [] tmp;

        return status;
}


ITranscoder::result
ITranscoder::toUnicode(const IString& from,
                UniChar* to, UniChar* to_limit, UniChar*& to_next)
{
        if (from.length() == 0)
        {
                to_next = to;
                return codecvt_base::noconv;
        }

        IString                 copy = from;
        const char* from_begin = copy;
        const char* from_end = from_begin + from.length();
        const char* from_next = NULL;

        result status = doToUnicode(
                from_begin, from_end, from_next, to, to_limit, to_next);
        *to_next = 0x0000;

        return status;
}


ITranscoder::result
ITranscoder::fromUnicode(const UniChar* from, const UniChar* from_end,
                const UniChar*& from_next, IString& to)
{
        length_type     charBufSize = byteBufferSize(from, from_end);

        char            stackBuffer[kBufSize+1];
        char*           to_begin = stackBuffer;
        IAutoArray<char> heapBuf = 0;

        if (charBufSize > kBufSize) {
                heapBuf.reset( new char[charBufSize+1] );
                to_begin = heapBuf.get();
        }

        char*   to_limit = to_begin + charBufSize;
        char*   to_next = NULL;

        result status = doFromUnicode(
                from, from_end, from_next, to_begin, to_limit, to_next);
        *to_next = '\0';

        to = IString(to_begin);

        return status;
}


ITranscoder::result
ITranscoder::toUnicode(const char* from, const char* from_end,
                const char*& from_next, IText& to)
{
        length_type     uniBufSize = uniCharBufferSize(from, from_end);

        UniChar                 stackBuffer[kBufSize+1];
        UniChar*                to_begin = stackBuffer;
        IAutoArray<UniChar> heapBuf = 0;

        if (uniBufSize > kBufSize) {
                heapBuf.reset( new UniChar[uniBufSize+1] );
                to_begin = heapBuf.get();
        }

        UniChar*        to_limit = to_begin + uniBufSize;
        UniChar*        to_next = NULL;

        result status = doToUnicode(
                from, from_end, from_next, to_begin, to_limit, to_next);
        *to_next = 0x0000;

        to = IText(to_begin, to_next-to_begin);

        return status;
}


ITranscoder::result
ITranscoder::fromUnicode(const IText& from,
                char* to, char* to_limit, char*& to_next)
{
        if (from.empty())
        {
                to_next = to;
                return codecvt_base::noconv;
        }

        IText   itext_from = from;
        length_type     from_length = itext_from.length();

        const UniChar*  from_begin;
        UniChar*                                tmp;
        bool                                            mustDispose = false;
        size_t                                  dummy;
        size_t                                  chunkLength;

        from_begin = itext_from.storage_chunk(0, dummy, chunkLength);
        if (chunkLength < from_length) {        // if from_length > 4k (current default)
                mustDispose = true;
                tmp = new UniChar[from_length];
                itext_from.copy(tmp, IText::npos);
                from_begin = tmp;
        }

        const UniChar* from_end = from_begin + from_length;
        const UniChar* from_next = NULL;

        result status = doFromUnicode(
                from_begin, from_end, from_next, to, to_limit, to_next);

        if (mustDispose)
                delete [] tmp;

        return status;
}


ITranscoder::result
ITranscoder::toUnicode(
                const char* from, const char* from_end, const char*& from_next,
                UniChar* to, UniChar* to_limit, UniChar*& to_next)
{
        return doToUnicode(from, from_end, from_next, to, to_limit, to_next);
}


ITranscoder::result
ITranscoder::fromUnicode(
                const UniChar* from, const UniChar* from_end, const UniChar*& from_next,
                char* to, char* to_limit, char*& to_next)
{
        return doFromUnicode(from, from_end, from_next, to, to_limit, to_next);
}
