// Revision: 73 1.15.1.4 source/core/text/itext/igtxstor.cpp, text, ioc.v400, 000302 
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//      IGTxStor.hpp    (IGapTextStorage.h)
//
//      Author: Richard Gillam
//
//      Function: Defines a class that isdesigned to give higher performance with longer
//              documents than ISimpleTextStorage.  This class's implementation is based on that
//              of ISimpleTextStorage, but instead of maintaining the object's slop space at
//              the end of the character array, it keeps itat the site of the last change to
//              the text, thus speeding up changes that occur in the same section of the text.
//
/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/

//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifndef IC_IGTXSTOR
#include "igtxstor.hpp"
#endif

#ifndef _IUGNRL_HPP
#include "iugnrl.hpp"
#endif

#ifndef _IEXCEPT_
#include <iexcept.hpp>
#endif

#ifndef _ITRACE_
#include <itrace.hpp>
#endif

#include <string.h>
#ifdef IC_PAGETUNE
        #define _IGTXSTOR_CPP
        #include <ipagetun.h>
#endif


const size_t IGapTextStorage::kInsertionSlopRatio = 4;
const size_t IGapTextStorage::kInsertionSlopConstant = 64;

//========================================================================================
// CLASS IGapTextStorage:  The concrete subclass of ITextStorage that is used to handle
// small strings (<4K characters) and individual nodes of large strings.  It stores the
// characters in a single contiguous array that is stored at the end of the IGapTextStorage
// object itself (we override operator new and operator delete to allocate the extra
// space).
//========================================================================================

extern IStreamModule gITextStreamModule;
StreamableDefinitionsMacro(IGapTextStorage, gITextStreamModule);

// compatibility with the goofy way the IBM compiler does memory debugging
#ifdef __DEBUG_ALLOC__
void*
IGapTextStorage::operator new(  size_t          basicSize,
                                                                const char*     file,
                                                                size_t          line,
                                                                size_t          minCapacity)
#else
void*
IGapTextStorage::operator new(  size_t          basicSize,
                                                                size_t          minCapacity)
#endif
{
        // WARNING:  The same value for minCapacity must be passed to both operator new()
        // and the constructor when creating an IGapTextStorage!  Failure to do this could have
        // disastrous consequences!
        // [The same logic for translating a minCapacity value to an actual allocated length
        // must be used in both operator new() and the constructor.  If you change one, change
        // the other.]
        void*           returnVal;
        size_t          slopSize;

        if (minCapacity != 0)
                slopSize = minCapacity + (minCapacity / kInsertionSlopRatio) +
                                                        kInsertionSlopConstant + 1;             // (1 is for the terminating null)
        else
                slopSize = 1;
        returnVal = malloc(basicSize + (slopSize * sizeof(UniChar)));

        IASSERTSTATE(returnVal != 0)

        return returnVal;
}

// compatibility with the goofy way the IBM compiler does memory debugging
#ifdef __DEBUG_ALLOC__
void
IGapTextStorage::operator delete(       void*           it,
                                                                        const char*     file,
                                                                        size_t          line)
#else
void
IGapTextStorage::operator delete(void*          it)
#endif
{
        free(it);
}

IGapTextStorage::IGapTextStorage(length_type    minCapacity)
: ITextStorage(),
  fGapPos(1),
  fGapLength(0)
{
        // WARNING:  The same value for minCapacity must be passed to both operator new()
        // and the constructor when creating an IGapTextStorage!  Failure to do this could have
        // disastrous consequences!
        // [The same logic for translating a minCapacity value to an actual allocated length
        // must be used in both operator new() and the constructor.  If you change one, change
        // the other.]
        if (minCapacity != 0)
                setCapacity(minCapacity + (minCapacity / kInsertionSlopRatio) + kInsertionSlopConstant);
        else
                setCapacity(0);

        fChars = (UniChar*)((unsigned char*)(this) + sizeof(IGapTextStorage));
                                                                                // the characters are stored at the end of the object
        fChars[0] = UUnicodeSpecial::kNull;

        fGapLength = capacity() - 1;
}

IGapTextStorage::IGapTextStorage(const IGapTextStorage& that)
: ITextStorage(that),
  fGapPos(0),
  fGapLength(0)
{
        fChars = (UniChar*)((unsigned char*)(this) + sizeof(IGapTextStorage));
        fChars[0] = UUnicodeSpecial::kNull;
        // DO NOT COPY-CONSTRUCT!
}

IGapTextStorage::~IGapTextStorage()
{
}

IGapTextStorage&
IGapTextStorage::operator=(const IGapTextStorage&       that)
{
        return *this;
        // do not assign!
}

void
IGapTextStorage::initFrom(      ITextStorage*   that,
                                                        offset_type             pos1,
                                                        length_type             length1,
                                                        offset_type             pos2,
                                                        length_type             length2,
                                                        length_type             gapLength,
                                                        bool                    willDropOriginal)
{
        // this function should only be called on freshly-allocated ITextStorages!
        IASSERTSTATE(length() == 0)

        handleInsertGap(0, length1 + length2 + gapLength);

        // move our gap to after the insertion gap to make room for more insertions at
        // the same position (we can't call inherited because handleInsertGap() fiddles
        // with the gap position as well)
        // [don't call moveGapTo here because we don't want any characters to be copied
        // (we don't have to mess with fGapLength because it's already correct])
        fGapPos = pos1 + length1 + gapLength;

        copyIn(0, that, pos1, length1);
        copyIn(length1 + gapLength, that, pos2, length2);

        if (that->isStyled())
                initStylesFrom(that, willDropOriginal);
}

void
IGapTextStorage::copyOut(       UniChar*                that,
                                                        offset_type             thisPos,
                                                        length_type             numChars) const
{
        if (numChars > length() - thisPos)
                numChars = length() - thisPos;

        // if there are characters before the gap, copy them first
        if (thisPos < fGapPos) {

                // if ALL of the characters are before the gap, copy them and we're done
                if (fGapPos - thisPos >= numChars) {
                        memcpy(that, &(fChars[thisPos]), numChars * sizeof(UniChar));
                        numChars = 0;
                }

                // otherwise, just copy the characters before the gap and update our pointers
                else {
                        memcpy(that, &(fChars[thisPos]), (fGapPos - thisPos) * sizeof(UniChar));
                        numChars -= fGapPos - thisPos;
                        that += fGapPos - thisPos;
                        thisPos = fGapPos;
                }
        }

        if (numChars != 0)
                memcpy(that, &(fChars[thisPos + fGapLength]), numChars * sizeof(UniChar));
}

const UniChar*
IGapTextStorage::c_str() const
{
#pragma info(nocnv)
        ((IGapTextStorage*)this)->moveGapTo(length() + 1);
#pragma info(restore)

        return fChars;
}

UniChar&
IGapTextStorage::at(offset_type pos)
{
        IASSERTPARM(pos <= length())

        if (pos < fGapPos)
                return fChars[pos];
        else
                return fChars[pos + fGapLength];
}

ITextStorage*
IGapTextStorage::reserve(       length_type             numChars)
{
        // if we're already big enough to hold the specified number of characters, do nothing
        // (reserve() is never supposed to make an ITextStorage smaller)
        if (capacity() >= numChars)
                return this;

        // otherwise, create a new ITextStorage of the appropriate size, copy our characters into it,
        // and return it
        else
                return makeCopy(numChars);
}

ITextStorage*
IGapTextStorage::compact()
{
        return makeCopy();
}

const UniChar*
IGapTextStorage::storageChunk(  offset_type             searchPos,
                                                                offset_type&    chunkPos,
                                                                length_type&    chunkLength) const
{
        IASSERTPARM(searchPos <= length())

        if (searchPos < fGapPos) {
                chunkPos = 0;
                if (length() < fGapPos)
                        chunkLength = length(); // (this can happen if everything, including the terminating null,
                                                                        // is before the gap)
                else
                        chunkLength = fGapPos;
                return fChars;
        }
        else {
                chunkPos = fGapPos;
                chunkLength = length() - fGapPos;
                return &(fChars[fGapPos + fGapLength]);
        }
}

IFastTextIterator
IGapTextStorage::beginAt(offset_type            position)
{
        UniChar*        charPos;

        if (position < fGapPos)
                charPos = &(fChars[position]);
        else
                charPos = &(fChars[position + fGapLength]);

        IFastTextIterator       temp;

        ITextStorage::fillInIteratorFields(temp, charPos, &(fChars[fGapPos]), fGapLength);

        return temp;
}

void
IGapTextStorage::printDebugInfo() const
{
        ITextStorage::printDebugInfo();
    ITRACE_RUNTIME( IString("fGapPos = ") + IString(fGapPos) +
                        IString("nfGapLength = ") + IString(fGapLength) +
                    IString("Class: IGapTextStorage"));
}

void
IGapTextStorage::handleInsertGap(       offset_type             pos,
                                                                        length_type             numChars)
{
        moveGapTo(pos);
        setLength(length() + numChars);
        fGapPos += numChars;
        fGapLength -= numChars;
}

void
IGapTextStorage::handleDeleteChars(     offset_type             pos,
                                                                        length_type             numChars)
{
        moveGapTo(pos + numChars);
        setLength(length() - numChars);
        fGapPos -= numChars;
        fGapLength += numChars;
}

void
IGapTextStorage::handleCopyIn(  offset_type             thisPos,
                                                                const UniChar*  that,
                                                                length_type             numChars)
{
        if (thisPos < fGapPos) {
                if (thisPos + numChars < fGapPos)
                        memcpy(&(fChars[thisPos]), that, numChars * sizeof(UniChar));
                else {
                        memcpy(&(fChars[thisPos]), that, (fGapPos - thisPos) * sizeof(UniChar));
                        memcpy(&(fChars[fGapPos + fGapLength]), that + (fGapPos - thisPos),
                                                                (numChars - (fGapPos - thisPos)) * sizeof(UniChar));
                }
        }
        else
                memcpy(&(fChars[thisPos + fGapLength]), that, numChars * sizeof(UniChar));
}

void
IGapTextStorage::handleFill(    offset_type             pos,
                                                                UniChar                 fillChar,
                                                                length_type             numCopies)
{
        length_type             charsToCopy = numCopies;
        if (pos < fGapPos && pos + numCopies >= fGapPos)
                charsToCopy = fGapPos - pos;
        numCopies -= charsToCopy;

        UniChar*        curChar = &(fChars[pos]);
        if (pos >= fGapPos)
                curChar += fGapLength;

        while (charsToCopy-- != 0)
                *curChar++ = fillChar;

        curChar += fGapLength;
        while (numCopies-- != 0)
                *curChar++ = fillChar;
}

void
IGapTextStorage::moveGapTo(     offset_type             newPosition)
{
        if (fGapPos == newPosition)
                return;

        length_type             charsMoving;

        if (newPosition < fGapPos) {
                charsMoving = fGapPos - newPosition;
                memmove(&(fChars[newPosition + fGapLength]), &(fChars[newPosition]), charsMoving
                                                        * sizeof(UniChar));
        }
        else {
                charsMoving = newPosition - fGapPos;
                memmove(&(fChars[fGapPos]), &(fChars[fGapPos + fGapLength]), charsMoving
                                                        * sizeof(UniChar));
        }
        fGapPos = newPosition;
}
