/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/
// Revision: 74 1.12.1.4 source/core/text/itext/igtxstor.hpp, text, ioc.v400, 000302 
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
//
//      IGTxStor.hpp    (IGapTextStorage.h)
//
//      Author: Richard Gillam
//
//      Function: Defines a class that isdesigned to give higher performance with longer
//              documents than ISimpleTextStorage.  This class's implementation is based on that
//              of ISimpleTextStorage, but instead of maintaining the object's slop space at
//              the end of the character array, it keeps itat the site of the last change to
//              the text, thus speeding up changes that occur in the same section of the text.
//
//      (C)
//
//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#ifndef _IGTXSTOR_
#define _IGTXSTOR_

#include <itxtstor.hpp>

#if __IBMCPP__ >= 400
#pragma namemangling(compat)
#endif

#pragma pack(push,4)
#pragma enum(4)

//========================================================================================
// CLASS IGapTextStorage:  A concrete subclass of ITextStorage that is similar to
// ISimpleTextStorage, but which maintains its extra character-storage space at the
// site of the last modification to the text instead of at the end.
//========================================================================================

class IGapTextStorage : public ITextStorage {
                StreamableDeclarationsMacro(IGapTextStorage);

        public:
                typedef size_t offset_type;
                typedef size_t length_type;

                static const size_t kInsertionSlopRatio;
                static const size_t kInsertionSlopConstant;

// compatibility with the goofy way the IBM compiler does memory debuging
#ifdef __DEBUG_ALLOC__
                void*                                   operator new(   size_t          basicSize,
                                                                                                const char*     file,
                                                                                                size_t          line,
                                                                                                size_t          minCapacity = 1);
                void                                    operator delete(        void*           it,
                                                                                                        const char*     file,
                                                                                                        size_t          line);
#else
                void*                                   operator new(   size_t  basicSize,
                                                                                                size_t  minCapacity = 1);
                void                                    operator delete(        void*           it);
#endif

                                                                IGapTextStorage(length_type     minCapacity = 1);
                                                                IGapTextStorage(const IGapTextStorage&  that);  // for streaming only!
                virtual                                 ~IGapTextStorage();

                virtual void                    initFrom(       ITextStorage*   that,
                                                                                        offset_type             pos1,
                                                                                        length_type             length1,
                                                                                        offset_type             pos2,
                                                                                        length_type             length2,
                                                                                        length_type             gapLength,
                                                                                        bool                    willDropOriginal);

                virtual void                    copyOut(UniChar*                that,
                                                                                offset_type             thisPos,
                                                                                length_type             numChars) const;

                virtual const UniChar*  c_str() const;
                virtual UniChar&                at(offset_type  pos);

                virtual ITextStorage*   reserve(        length_type             numChars);
                virtual ITextStorage*   compact();
                virtual const UniChar*  storageChunk(   offset_type             searchPos,
                                                                                                offset_type&    chunkPos,
                                                                                                length_type&    chunkLength) const;

                virtual IFastTextIterator
                                                                beginAt(offset_type             position);

                virtual void                    printDebugInfo() const;

        protected:
                virtual void                    handleInsertGap(        offset_type             pos,
                                                                                                        length_type             numChars);

                virtual void                    handleDeleteChars(      offset_type             pos,
                                                                                                        length_type             numChars);

                virtual void                    handleCopyIn(   offset_type             thisPos,
                                                                                                const UniChar*  that,
                                                                                                length_type             numChars);

                virtual void                    handleFill(     offset_type             pos,
                                                                                        UniChar                 fillChar,
                                                                                        length_type             numCopies);

        private:
                IGapTextStorage&                operator=(const IGapTextStorage&        that);

                void                                    moveGapTo(offset_type   newPosition);

                offset_type                             fGapPos;
                length_type                             fGapLength;
                UniChar*                                fChars;
};

#pragma enum(pop)
#pragma pack(pop)

#if __IBMCPP__ >= 400
#pragma namemangling()
#endif

#endif // _IGTXSTOR_
