//==========================================================================
/*
 * Copyright    IBM Open Class Library
 *              (C) Copyright International Business Machines Corporation 1997
 *              Licensed Material - Program-Property of IBM - All Rights Reserved.
 * @version     1.0 23/10/96
 * @author      Helena Shih
 * Based on IOC international support for java
 */

#ifndef _ICPARRAY_
#define _ICPARRAY_

#include <iprimtyp.hpp>

#if __IBMCPP__ >= 400
#pragma namemangling(compat)
#endif

#pragma pack(push,4)

//====================================
// class ICompactArrayOf<AType>
// Provides a compact way to store information that is indexed by Unicode values,
// such as character properties, types, keyboard values, etc.
// The ATypes are used by value, so should be small, integers or pointers.
//====================================
template <class AType>
class ICompactArrayOf {
 public:

        static  const long kUnicodeCount;
    static  const long kBlockShift;
    static  const long kBlockCount;
    static  const long kIndexShift;
    static  const long kIndexCount;
    static  const unsigned long kBlockMask;

        ICompactArrayOf(AType defaultValue);
        ICompactArrayOf(unsigned short *indexArray, AType *newValues, long count);
        ICompactArrayOf(const ICompactArrayOf<AType>& that);
        ~ICompactArrayOf();

    ICompactArrayOf<AType>& operator=(const ICompactArrayOf<AType>& that);

        // Get is inline for speed
        AType get(UniChar index);

        // Set automatically expands the array if it is compacted.

        void set(UniChar index, AType value);
        void set(UniChar start, UniChar end, AType value);

        // Compact the array.
        // The value of cycle determines how large the overlap can be.
        // A cycle of 1 is the most compacted, but takes the most time to do.
        // If values stored in the array tend to repeat in cycles of, say, 16,
        // then using that will be faster than cycle = 1, and get almost the
        // same compression.

        void compact (unsigned long cycle = 1/* default = 1 */);

        // Expanded takes the array back to a 65536 element array

        void expand();

        // Print UniChar Array  : Debug only

        void printIndex(long start, long count);

        void printPlainArray(long start,long count, const UniChar* tempIndex);

        // # of elements in the indexed array
        long getCount();
        void setCount(long n);
        const AType* getArray();
        const unsigned short* getIndex();

protected:

        long findOverlappingPosition(unsigned long start,
                                    const UniChar *tempIndex,
                                    long tempIndexCount,
                                    unsigned long cycle);
private:

    static  const bool debugShowOverlap;
    static  const unsigned long debugShowOverlapLimit;
    static  const bool debugTrace;
    static  const bool debugSmall;
    static  const unsigned long debugSmallLimit;

        AType* fArray;
        unsigned short* fIndex;
        long fCount;
        bool fCompact;
};

#pragma pack(pop)

#if __IBMCPP__ >= 400
#pragma namemangling()
#endif

template <class AType>
inline AType
ICompactArrayOf<AType>::get(UniChar c)
{
     return (fArray[(fIndex[c >> kBlockShift] & 0xFFFF)
                   + (c & kBlockMask)]);
}

#include <icparray.inl>

#endif // _ICPARRAY_
