/*
*****************************************************************************************
*                                                                                       *
* COPYRIGHT:                                                                            *
*   IBM Open Class Library                                                              *
*   (C) Copyright International Business Machines Corporation,  1997                    *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.                  *
*   US Government Users Restricted Rights - Use, duplication, or disclosure             *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                              *
*                                                                                       *
*****************************************************************************************
*/

// Revision: 76 1.6.1.3 source/core/utils/ilonglng.cpp, utilities, ioc.v400, 000302 
/*
 * DoubleLong.C - generic file, containing C++ code for TDoubleLong methods
 *
 * Modification history:-
 *      Deb Orton       --      25 January 1990  - created
 *      Shinzo Watanabe --      17 December 1990 - modifications
 *      Sanjay Poonen
 *      --      2 October 1991
 *              started adding more portable methods in C, thus maintaining
 *              an assembler and C version of most routines.
 *              Added logical operators.
 *      --      20 Jan, 1992
 *              added shift operators and cleaned up extensively for review
 *      --      3 Mar, 1992
 *              added post review changes, and completed C++ implementation
 *              of the optimized divide algorithm for TDoubleLong.
 *      --      22 Apr, 1992
 *              fixed bug in MULUL algorithm, and cleaned up code to create
 *              this generic file which contains C++ code for all methods in
 *              the TDoubleLong class, which can be used as a reference when
 *              porting to another platform.
 *
 *      --      30 Aug, 1993
 *              AMN fixed left and right shift operators for code portability,
 *              i.e, shifts by more that 31 bits not allowed in Intel and PPC.
 *              Also modified code for constructor that takes a long double.
 *              Previous code not portable to INTEL and PPC due to different bit
 *              ordering. Now used FBitDef.h that takes care of differences.
 *              Only using float division. There is a problem to fix with integer division
 *                              (See example just before the actual code for division.)
 *      --      21 Sep, 1993
 *              AMN included the header NumericUtilities that declares _MULUL.
 *              Only uses float division, so DIVUL is deleted, so is FindFirst
 *              used in the integer division code.
 *      --      29 Oct, 1993
 *              CLM made Taligent_INITIALPORTING changes to implement Taligent_NOSTREAMING
 *              and Taligent_NOEXCEPTIONS.  Removed dependency on Numerics.h and replaced
 *              with fenv.h so that feraiseexcept(FE_INVALID) can be used instead of
 *              0.0*kInfinity.
 *      --      08 Nov, 1993
 *              AMN after Mini-Code Review of Nov. 2.a) Included Numerics.h (Floor, Ceiling, Abs...)
 *              and FPEnvironment (to raise inexact); b) Defined new constants: k2p52
 *              and TDoubleLongMax and TDoubleLongMin.; c) In Divide operator replaced
 *              this ->fLeastSignificant by y.fLeastSignificant; d) Replaced #ifndef _AIX_
 *              by #if defined(__68K__)||defined(__X86__) ; e ) Conversion long double -> TDoubleLong

 *               by constructor,
 *              improved code: test for exp<63 removed, inexact is raised using FPEnvironment,
 *              TDoubleLongMax and TDoubleLongMin are used, replaced *this>>= (62-exp) by actual shift code;
 *              f) Added alternative implementation (similar to that of JPO's RTFPInternals.c)
 *              of Conversion long double -> TDoubleLong by constructor. Nicer and (believed to be)
 *              faster. Need timing. ; g) Conversion double -> TDoubleLong
 *              by constructor improved code: used FPEnvironment to raise inexact, used k2p52, TDoubleLongMax
 *              and Min, replaced *this>>= (52-exp) by actual shift code; h) Added alternative
 *              implementation of Conversion double -> TDoubleLong by constructor.
 *              Nicer and (believed to be) faster. Need timing.; i)  Conversion long double -> TDoubleLong
 *              by constructor Improved code: raise inexact using FPEnvironment, use k2p63, TDoubleLongMax
 *              and Min, Abs; j) DDoubletoDLong uses Floor and Ceiling,
 *              casting of Tail to long (instead of to TDoubleLong), use test of Head and Tail
 *              instead of thier absolute values. Still needs to be tested with double-double.
 *
 *      --      09 Feb, 1994
 *              SJP     Changed __PPC__ to _AIX_ for SDK1, revisit this again later.
 *      --      09 March, 1994
 *              AMN     Added macro NO_Internal to protect PrintObject. PrintObject should be gone altogether in future.
 *      --      02 June, 1994
 *              AMN     Double constructor and double cast operator implementations are also suitable for X86.
 *              (_AIX_ protection has been deleted for these routines). This is to fix bug 12835.
 *      __      20 Sept, 1994
 *              AMN     Fixed double/long double constructors and cast operators. Some constants (such as k2p32,...)
 *              must be double (instead of long double).
 *      __      05 Oct, 1994
 *              AMN     More work on double and long double (X86 and AIX) constructors to fix bugs
 *      __      15 Oct, 1994
 *              AMN     Afer PrimitiveTypes code review: split PrimitiveTypes header into three: CPrimitiveTypes.h
 *                      (includes C typedefs), PrimitiveTypesImp.h (inline implementations) and PrimitiveTpes.h
 *                      just TDoubleLong class declaration and declaration of global operators.  Move some stuff
 *                      from DoubleLong.C to PrimitiveTypesImp.h. Also corrected and improved some inline implementations
 *      __      17 Oct, 1994
 *              AMN     Deleted double cast operator
 *
 *
 */
/*

#ifndef Taligent_PRIMITIVETYPES
        #include <Taligent/PrimitiveTypes.h>            //      TDoubleLong declared there
#endif

#ifndef Taligent_NOSTREAMING
        #ifndef Taligent_GLOBALSTREAMING
        #include <Taligent/GlobalStreaming.h>           //      for streaming.
        #endif
#endif
*/



#include <iprimtyp.hpp>
#include <ilonglng.hpp>

//#include <fenv.h>
//#include <numericutilities.h>                 //      MULUL is there
#include <limits.h>
#ifndef Taligent_FPBITDEF
#include <FPBitDef.h>
#endif

#include <inumutil.hpp>

#ifdef IC_PAGETUNE
    #define _ILONGLNG_CPP_
    #include <ipagetun.h>
#endif


/************************************************************************/
/*      ILongLong       constants                                       */
/************************************************************************/
#ifdef __FPCE__
        static const double k2p32 = 0x1.0p+32;
        static const double k2p52 = 0x1.0p+52;
        static const double k2p63 = 0x1.0p+63;
#else
        static const double k2p32 = 4294967296.0;                       //2^32
        static const double k2p52 = 4503599627370496.0;                 //2^52
        static const double k2p63 = 9223372036854775808.0;              //2^63
#endif

/************************************************************************/
/*      Assignment RIGHT SHIFT operator                                 */
/************************************************************************/

ILongLong& ILongLong::operator>>= (int howMuch)
{
        long tmp = howMuch-32;

        // extract mask that will fall over into leastSig
        if (tmp > 0) {                                                  // shift by more than 32
                long mask = fMostSignificant >> (tmp) ;
                fLeastSignificant = mask;

                if(mask<0)
                        fMostSignificant = -1;
                else
                        fMostSignificant = 0;
        }


        else if (tmp == 0) {                                            //shift by 32
                fLeastSignificant = fMostSignificant;
                if(fMostSignificant < 0)
                        fMostSignificant = -1;
                else
                        fMostSignificant = 0;
        }


        else if (tmp > -32) {                                           // shift by less than 32
                long mask = fMostSignificant << (32 - howMuch) ;
                fLeastSignificant >>= howMuch;
                fLeastSignificant += mask;                      // add in the masked extract
                fMostSignificant >>= howMuch;
        }

        return *this;    //shift by 0

}


/************************************************************************/
/*      Assignment LEFT SHIFT operator                                  */
/************************************************************************/
ILongLong& ILongLong::operator<<= (int howMuch)
{
        long tmp = howMuch-32;

        // extract mask that will move into mostSig
        if (tmp > 0) {                                                  // shift by more than 32
                long mask = fLeastSignificant << (tmp);
                fMostSignificant = mask;
                fLeastSignificant = 0;
        }


        else if (tmp == 0) {                                            //shift by 32
                fMostSignificant = fLeastSignificant;
                fLeastSignificant = 0;
        }


        else if (tmp > -32) {                                           // shift by less than 32
                long mask = fLeastSignificant >> (32 - howMuch) ;

                fLeastSignificant <<= howMuch;
                fMostSignificant <<= howMuch;                   // add in the masked extract
                fMostSignificant += mask;
        }

        return *this;    //shift by 0

}


/************************************************************************/
/*      Assignment MULTIPLY operator                                    */
/************************************************************************/
ILongLong& ILongLong::operator*=(const ILongLong& x)
{
        unsigned long   mostSig;
        unsigned long   leastSig;

        _MULUL(x.GetLeastSignificant(), fLeastSignificant, &mostSig, &leastSig);
        fMostSignificant = mostSig + x.fMostSignificant * fLeastSignificant +
                                        x.fLeastSignificant * fMostSignificant;
        fLeastSignificant = leastSig;

        return(*this);
}


/************************************************************************/
/*      Assignment DIVIDE operator                                      */
/************************************************************************/
ILongLong& ILongLong::operator/=(const ILongLong& y)
{
  //    fenv_t e;

        if (y==0L)
                return(*this = ILongLong(1/y.fLeastSignificant));       // parallel integer divide by zero

        //fegetenv(&e);
                long double f = (long double) *this;
                long double g = (long double) y;
                *this  = ILongLong((long double)(f/g));
        //fesetenv(&e);

        return (*this);

}
/************************************************************************/
/*      Assignment MODULUS operator                                      */
/************************************************************************/
ILongLong& ILongLong::operator%=(const ILongLong& y)
{
        ILongLong n = *this / y;
        *this -= (n * y);
        return *this;
}


/************************************************************************/
/*      global method for binary SHIFT Operators                        */
/************************************************************************/
ILongLong operator>>(const ILongLong& x, int a)
{
        ILongLong result = x;
        result >>= a;
        return result;
}

ILongLong operator<<(const ILongLong& x, int a)
{
        ILongLong result = x;
        result <<= a;
        return result;
}

/************************************************************************/
/*      global method for binary MULTIPLY Operator                      */
/************************************************************************/
ILongLong operator*(const ILongLong& x, const ILongLong& y)
{
        ILongLong result = x;
        result *= y;
        return result;
}

ILongLong  operator*(const ILongLong& x, unsigned long a)
{
        return x * ILongLong(a);
}

ILongLong  operator*(const ILongLong& x, long a)
{
        return x * ILongLong(a);
}

ILongLong  operator*(const ILongLong& x, unsigned int a)
{
        return x * ILongLong(a);
}

ILongLong  operator*(const ILongLong& x, int a)
{
        return x * ILongLong(a);
}

ILongLong  operator*(long a, const ILongLong& x)
{
        return x * ILongLong(a);
}

ILongLong  operator*(unsigned long a, const ILongLong& x)
{
        return x * ILongLong(a);
}

ILongLong  operator*(int a, const ILongLong& x)
{
        return x * ILongLong(a);
}

ILongLong  operator*(unsigned int a, const ILongLong& x)
{
        return x * ILongLong(a);
}


/************************************************************************/
/*      global method for binary DIVIDE Operator                        */
/************************************************************************/
ILongLong operator/(const ILongLong& x, const ILongLong& y)
{
        ILongLong result = x;
        result /= y;
        return result;
}

ILongLong  operator/(const ILongLong& x, long a)
{
        return x/ILongLong(a);
}

ILongLong  operator/(const ILongLong& x, unsigned long a)
{
        return x/ILongLong(a);
}

ILongLong  operator/(const ILongLong& x, int a)
{
        return x/ILongLong(a);
}

ILongLong  operator/(const ILongLong& x, unsigned int a)
{
        return x/ILongLong(a);
}

ILongLong  operator/(long a, const ILongLong& x)
{
        return ILongLong(a)/x;
}

ILongLong  operator/(unsigned long a, const ILongLong& x)
{
        return ILongLong(a)/x;
}

ILongLong  operator/(int a, const ILongLong& x)
{
        return ILongLong(a)/x;
}

ILongLong  operator/(unsigned int a, const ILongLong& x)
{
        return ILongLong(a)/x;
}


/************************************************************************/
/*      method for binary bitwise MODULUS Operator                      */
/************************************************************************/

ILongLong operator% (const ILongLong& x, const ILongLong& y)
{
        ILongLong result = x;
        result /= y;
        result = x - (result * y);
        return result;
}

ILongLong operator%(const ILongLong& x, long a)
{
        return (x % ILongLong(a));
}

ILongLong operator%(const ILongLong& x, unsigned long a)
{
        return (x % ILongLong(a));
}

ILongLong  operator%(const ILongLong& x, int a)
{
        return (x % ILongLong(a));
}

ILongLong  operator%(const ILongLong& x, unsigned int a)
{
        return (x % ILongLong(a));
}

ILongLong operator%(long a, const ILongLong& x)
{
        return (ILongLong(a) % x);
}

ILongLong operator%(unsigned long a, const ILongLong& x)
{
        return (ILongLong(a) % x);
}

ILongLong  operator%(int a, const ILongLong& x)
{
        return (ILongLong(a) % x);
}

ILongLong  operator%(unsigned int a, const ILongLong& x)
{
        return (ILongLong(a) % x);
}


/************************************************************************/
/*      Conversion ILongLong -> long double by type cast                */
/************************************************************************/

#if 0
ILongLong::operator long double () const
{
        if(((unsigned long)fMostSignificant | fLeastSignificant) == 0) return 0.0L;
        long double result = (long double)(k2p32) * fMostSignificant + fLeastSignificant;
        return result;
}

#endif

/************************************************************************/
/*      Conversion double -> ILongLong by constructor           */
/************************************************************************/
#ifdef  __HHW_INTEL__ // intel platform

ILongLong::ILongLong(double theNumber)
{
        *this = ILongLong ((long double) theNumber);
}

#elif  _AIX_

ILongLong::ILongLong(double theNumber)
{

        _DoubleOrTwoPieces xD;
        xD._F = theNumber;
        unsigned long expx = (xD._I._One >> 20) & 0x7fful;              // get biased exponent
        unsigned long signx = (xD._I._One >> 31) & 1ul;                 //get sign

        if (expx < 0x0433) {                    //abs(TheNumber) < 2^52
                if(signx) theNumber = -theNumber;
                xD._F = (theNumber + k2p52) - k2p52;    //this takes care of rounding only PPC
                if (xD._F > theNumber) xD._F -= 1.0;    //only takes care of positive values
                xD._F = xD._F + k2p52;
                fLeastSignificant = xD._I._Two;
                fMostSignificant = xD._I._One & (0x000FFFFF);
                if(signx) *this = -*this;
        }


        else if (expx >= 0x043e)                //abs(TheNumber) >= 63
        {                                       // includes NaN and infinity
                if (theNumber != -k2p63)        //k2p63 = 2^63
                feraiseexcept(FE_INVALID);      // set the invalid flag
                if (signx)      *this = ILongLongMin();
                else            *this = ILongLongMax();
        }

        else {                                  //if (52 <= unbias exp < 63).
                *this = ILongLong(xD._I._One,xD._I._Two);
                fMostSignificant &= 0x000FFFFF;
                fMostSignificant += 0x00100000;         //implicit bit
                if (expx > 0x433) {
                        long mask = fLeastSignificant >> (0x0453 - expx) ;
                        // code below equivalent to *this <<=  (exp - 52)
                        fLeastSignificant <<= (expx - 0x0433);
                        fMostSignificant <<= (expx - 0x0433);
                        fMostSignificant += mask; // add in the mask
                }

        if (signx) *this = -*this;
        }

}

#endif

/************************************************************************/
/*      Conversion long double -> ILongLong by constructor              */
/************************************************************************/


#ifdef __HHW_INTEL__


ILongLong::ILongLong(long double theNumber)
{
        _ExtendedOrThreePieces y;
        y._F = theNumber;
        unsigned short expy = y._I._One  & 0x7fff;              // get biased exponent
        unsigned short signy = (y._I._One >> 15) & 1;           // get sign

        if (expy < 0x403e){                     //abs(theNumber) < k2p63
                if(signy) theNumber = - theNumber;
                y._F = (theNumber + k2p63) - k2p63;  //this takes care of rounding
                if (y._F > theNumber) y._F -= 1.0;
                y._F = y._F + k2p63;
                fLeastSignificant = y._I._Three;
                fMostSignificant = y._I._Two & (0x7FFFFFFF);
                if (signy)  *this = -*this;                     //takes care of overall sign
        }
        else if (signy && (theNumber > -k2p63 - 1.0L)) {  //between -k2p63 - 1.0L and  -k2p63
                if (theNumber != -k2p63)
                  RaiseException(FE_INEXACT); //feraiseexcept(FE_INEXACT); // set the inexact flag
                *this = ILongLongMin();
        }
        else                                                    //if unbias exp >= 63
        {                                                       // includes nan and inf
          RaiseException(FE_INVALID); //feraiseexcept(FE_INVALID);// set the invalid flag
                if (signy)      *this = ILongLongMin();
                else            *this = ILongLongMax();
        }
}



#elif  _AIX_


ILongLong::ILongLong(long double theNumber)
{

        _DoubleDouble DDNumber;

        {
        fenv_t e;
        fegetenv (&e);
                DDNumber._LD = Truncate(theNumber);     //get correctly rounded integer part of theNumber
        fesetenv (&e);                                  //not raise any flag yet
        }

        double DDNumberHead = DDNumber._DD._One;
        double DDNumberTail = DDNumber._DD._Two;
        unsigned long expHead = (DDNumber._I4._One >> 20) & 0x7fful;  // get biased exponent
        unsigned long signHead = (DDNumber._I4._One >> 31) & 1ul;  // get sign

 // large exponents(>63), NaNs and infinities, and also exp = 63 with Abs(Head) > 2^63 are invalid

        if ((expHead > 0x043e) || ((expHead == 0x043e) && ((DDNumber._I4._One & 0xffffful) | DDNumber._I4._Two))) {
                feraiseexcept(FE_INVALID);      // set the invalid flag
                if (signHead)   *this = ILongLongMin();
                else            *this = ILongLongMax();
        }

// exponent <= 63. If exp = 63, then Abs(Head) = 2 ^ 63, must be careful and filter first

        else {
                bool nonIntegral = false;
                if( DDNumber._LD != theNumber) nonIntegral = true;      //may deserve inexact

                if (expHead == 0x043e) {                        // if Abs(Head)> 2^63, careful!

                        if (signHead) {
                                if (DDNumberTail > -1.0)  {
                                        if(nonIntegral) feraiseexcept(FE_INEXACT);      // set the inexact flag
                                        *this = ILongLongMin() + ILongLong(DDNumberHead + k2p63) + ILongLong (DDNumberTail);
                                }
                                else {
                                        feraiseexcept(FE_INVALID);   // set the invalid flag
                                        *this = ILongLongMin();
                                }
                        }
                        else {                          //theNumberHead>0.0
                                if (DDNumberTail  < 0.0){
                                        if(nonIntegral) feraiseexcept(FE_INEXACT);              // set the inexact flag
                                        *this = ILongLongMax() + ILongLong(DDNumberHead - k2p63) + ILongLong (DDNumberTail + 1.0);
                                }
                                else {
                                        feraiseexcept(FE_INVALID);   // set the invalid flag
                                        *this = ILongLongMax();
                                }
                        }
                }       // end of Abs(Head) = 63

// at this point exponent is strictly less than 63

                else {
                        if(nonIntegral) feraiseexcept(FE_INEXACT);              // set the inexact flag
                        *this = ILongLong(DDNumberHead) + ILongLong (DDNumberTail);
                }
        }

}


#endif


/********************************************************************************/
/*      End of File: DoubleLong.C                                               */
/********************************************************************************/
