// Revision: 66 1.5.1.3 source/ui/baseapp/iaccel.cpp, userinput, ioc.v400, 000302  
/*******************************************************************************
* FILE NAME: iaccel.cpp                                                        *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iaccel.hpp.                                                             *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#define INCL_WINACCELERATORS
extern "C" {
  #include <iwindefs.h>
}

#include <iaccel.hpp>
#include <iacceltb.hpp>
#include <iexcept.hpp>
#include <ireslib.hpp>
#include <ithread.hpp>
#include <itrace.hpp>
#include <iwindow.hpp>

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _IACCEL_CPP_
  #include <ipagetun.h>
#endif


#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| IAccelerator::IAccelerator                                                   |
------------------------------------------------------------------------------*/
IAccelerator :: IAccelerator ( const IAccelTblHandle& haccel,
                               IWindow* pwndOwner )
{
  this->create( pwndOwner );
  if ( haccel != 0 )
     this->set( haccel );
}
#endif


/*------------------------------------------------------------------------------
| IAccelerator::IAccelerator                                                   |
------------------------------------------------------------------------------*/
IAccelerator :: IAccelerator ( const IResourceId& accelResId,
                               IWindow* pwndOwner )
{
  this->create( pwndOwner );
  this->set( accelResId );
}


/*------------------------------------------------------------------------------
| IAccelerator::IAccelerator                                                   |
------------------------------------------------------------------------------*/
IAccelerator :: IAccelerator ( unsigned long accelResId,
                               IWindow* pwndOwner )
{
  this->create( pwndOwner );
  this->set( IResourceId( accelResId ));
}


/*------------------------------------------------------------------------------
| IAccelerator::create                                                         |
------------------------------------------------------------------------------*/
void IAccelerator :: create ( IWindow* pwndOwner )
{
  pwndClOwner   = pwndOwner;
  haccelClPrior = 0;
  haccelCl      = 0;
  isSetCl       = false;

  IThread::current().initializeGUI();
  haccelClOriginal = 0;
  if ( pwndOwner )
  {
     haccelClOriginal = pwndOwner->acceleratorHandle();
  }
}


/*------------------------------------------------------------------------------
| IAccelerator::~IAccelerator                                                  |
------------------------------------------------------------------------------*/
IAccelerator :: ~IAccelerator ( )
{
   if ( haccelCl != 0  &&  haccelCl != haccelClOriginal )
      IAcceleratorTable::destroyHandle( haccelCl );

   if ( haccelClPrior != 0  &&  haccelClPrior != haccelClOriginal )
      IAcceleratorTable::destroyHandle( haccelClPrior );

   // If for a window that is still open, reset to original...
   if ( pwndClOwner  &&  pwndClOwner->isValid() )
      pwndClOwner->setAcceleratorHandle( haccelClOriginal );
}


/*------------------------------------------------------------------------------
| IAccelerator::set                                                            |
------------------------------------------------------------------------------*/
IAccelerator& IAccelerator :: set ( const IAccelTblHandle& haccel )
{
  // Calling set with 0 for the haccel is equivalent to
  // removeAccel(), but the handling of haccelClPrior won't be
  // the same.  removeAccel() leaves haccelClPrior unchanged, but
  // calling setAccel(0) updates haccelCl and haccelClPrior.
  // What to do??  Let's make set(0) a no-op...

  if ( haccel != 0 )
  {
     IWindow* accelOwner = this->owner();
     if ( accelOwner )
     {
        accelOwner->setAcceleratorHandle( haccel );
     }

     isSetCl = true;

     if ( haccelCl != 0 )
     {
        /* Don't destroy Prior if it's also one of the other active ones */
        if ( haccelClPrior != 0  &&
             haccelClPrior != haccelClOriginal  &&
             haccelClPrior != haccelCl  &&
             haccelClPrior != haccel )
           IAcceleratorTable::destroyHandle( haccelClPrior );
        haccelClPrior = haccelCl;
     } /* endif */

     haccelCl = haccel;
  }

  return *this;
}


/*------------------------------------------------------------------------------
| IAccelerator::set                                                            |
------------------------------------------------------------------------------*/
IAccelerator& IAccelerator :: set ( const IResourceId& accelResId )
{
  IAccelTblHandle
    haccelTemp( accelResId.resourceLibrary().loadAccelTable( accelResId ) );
  return this->set( haccelTemp );
}


/*------------------------------------------------------------------------------
| IAccelerator::remove                                                         |
------------------------------------------------------------------------------*/
IAccelerator& IAccelerator :: remove ( )
{
  if ( haccelCl != 0  &&
       haccelCl != haccelClPrior  &&
       haccelCl != haccelClOriginal )
  {
     IAcceleratorTable::destroyHandle( haccelCl );
     haccelCl = 0;
  } /* endif */

  IWindow* accelOwner = this->owner();
  if ( accelOwner )
  {
     accelOwner->setAcceleratorHandle( 0 );
  }

  isSetCl = false;

  return *this;
}


/*------------------------------------------------------------------------------
| IAccelerator::handle                                                         |
------------------------------------------------------------------------------*/
IAccelTblHandle IAccelerator :: handle ( ) const
{
  IAccelTblHandle haccel( 0 );
  IWindow* accelOwner = this->owner();
  if ( accelOwner )
  {
     haccel = accelOwner->acceleratorHandle();
  }
  return haccel;
}


/*------------------------------------------------------------------------------
| IAccelerator::ownerHandle                                                    |
------------------------------------------------------------------------------*/
IWindowHandle IAccelerator :: ownerHandle ( ) const
{
  IWindow* ownerWindow = this->owner();
  if ( ownerWindow )
  {
     return ownerWindow->handle();
  }
  else
  {
     return 0;
  }
}


/*------------------------------------------------------------------------------
| IAccelerator::reset                                                          |
------------------------------------------------------------------------------*/
IAccelerator& IAccelerator :: reset ( )
{
  if ( haccelCl != 0 )
  {
     // Don't want to destroy the original haccel.
     // Want it around even if unsetAccel() is called repeatedly.
     if ( haccelCl != haccelClOriginal  &&
          haccelCl != haccelClPrior )
        IAcceleratorTable::destroyHandle( haccelCl );

     haccelCl = haccelClPrior;
     haccelClPrior = 0;
     if ( haccelCl != 0 )
     {
        this->set( haccelCl );
        isSetCl = true;
     }
     else
     {
        this->set( haccelClOriginal );
        isSetCl = false;
     }
  } /* endif */

  return *this;
}
