// Revision: 65 1.2.1.11 source/ui/baseapp/icmaphdr.cpp, frame, ioc.v400, 000302 
/*******************************************************************************
* FILE NAME: icmpahdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in icmaphdr.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#include <icmaphdr.hpp>
#include <iframe.hpp>
#include <ievent.hpp>
#include <iexcept.hpp>
#include <itrace.hpp>
#include <icolmap.hpp>

#define INCL_WINFRAMEMGR
#define INCL_GPI
#define INCL_WINPALETTE
#define INCL_WINWINDOWMGR
#include <iwindefs.h>


#ifdef IC_MOTIF
  #include <ixdc.hpp>
  #include <X11/IntrinsicP.h>
#endif


// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICMAPHDR_CPP_
  #include <ipagetun.h>
#endif


///////////////////// Common Code  ////////////////////

/*------------------------------------------------------------------------------
| IColorMapHandler::IColorMapHandler                                           |
|                                                                              |
| Constructor for IColorMapHandler.                                            |
------------------------------------------------------------------------------*/
IColorMapHandler::IColorMapHandler ( IFrameWindow* frameWindow,
                                     IColorMap* colorMap )
                      : fFrameWindow(frameWindow),
                        fColorMap(colorMap),
                        fPaletteSupported(isPaletteDevice()),
                        fHasFirstRealizationOccurred(false),
                        fExtendedData(0)
#ifdef IC_WIN                        
                        ,lastWindowToChange(0)
#endif  // IC_WIN                        
{
   IFUNCTRACE_DEVELOP();
   handleEventsFor(frameWindow);
}


/*------------------------------------------------------------------------------
| IColorMapHandler::~IColorMapHandler                                          |
|                                                                              |
| Destructor for IColorMapHandler.                                             |
------------------------------------------------------------------------------*/
IColorMapHandler::~IColorMapHandler ( )
{
   IFUNCTRACE_DEVELOP();
   stopHandlingEventsFor(fFrameWindow);
}



/*------------------------------------------------------------------------------
| IColorMapHandler::setColorMap                                                |
|                                                                              |
| Store the new color map.                                                     |
------------------------------------------------------------------------------*/
void IColorMapHandler::setColorMap ( IColorMap* colorMap )

{
   IFUNCTRACE_DEVELOP();
   fColorMap = colorMap;

   // If this is a new colormap and realization of the old color map has
   // already occurred, realize this new color map.  Otherwise
   // we will let the handler do it.
   if(fHasFirstRealizationOccurred == true)
      realizeColorMap(fFrameWindow->handle(), false, false);
}


/*------------------------------------------------------------------------------
| IColorMapHandler::colorMap                                                   |
|                                                                              |
| Return the colormap stored in the handler.                                   |
------------------------------------------------------------------------------*/
IColorMap& IColorMapHandler::colorMap ( ) const

{
   IFUNCTRACE_DEVELOP();

   IASSERTSTATE(fColorMap!=0);
   return *fColorMap;
}

///////////////////// PM and Windows //////////////////

#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| IColorMapHandler::handleRefresh                                              |
|                                                                              |
| PMWIN: Called to refresh the application after a palette change.             |
------------------------------------------------------------------------------*/
void IColorMapHandler::handleRefresh( )
{
   IFUNCTRACE_DEVELOP();
   fFrameWindow->refresh(IWindow::paintAllImmediate);
}

#endif

///////////////////// Windows /////////////////////////

#ifdef IC_WIN
/*------------------------------------------------------------------------------
| IColorMapHandler::dispatchHandlerEvent                                       |
|                                                                              |
| Win: Handle palette management messages.                                     |
------------------------------------------------------------------------------*/
bool IColorMapHandler::dispatchHandlerEvent(IEvent& event)
{
   bool fStopDispatch = false;

   if (event.eventId() == WM_PALETTECHANGED && lastWindowToChange != IWindowHandle(event.parameter1()))
   {
      // Don't realize palette again until our application has been activated
      // again and WM_QUERYNEWPPALETTE has been called, OR unless a different 
      
      IMODTRACE_DEVELOP("WM_PALETTECHANGED");
      ITRACE_DEVELOP(IString("Window causing palette change:")
                    +IString((unsigned long)(event.parameter1())).d2x());

      // If our application did not change the palette, select
      // and realize the palette.  We need to check that event.parameter1()
      // (the window causing the palette change) is not our frame
      // window nor the client of the frame.
      IWindowHandle paletteChangingHandle(event.parameter1());
      IWindowHandle paletteOwnerHandle(fFrameWindow->handle());

      if( paletteChangingHandle != paletteOwnerHandle )
      {

         ITRACE_DEVELOP(IString("paletteOwnerHandle: ")
                        +IString(paletteOwnerHandle.asUnsigned()).d2x());

         // Not our window, so realize the palette
         realizeColorMap(paletteOwnerHandle, false, true);
      }
      fStopDispatch = true;
      lastWindowToChange = IWindowHandle(event.parameter1());
   }
   else if (event.eventId() == WM_QUERYNEWPALETTE)
   {
      IMODTRACE_DEVELOP("WM_QUERYNEWPALETTE");

      IFrameWindow* frame
         = dynamic_cast<IFrameWindow*>(event.dispatchingWindow());

      if(frame)
      {
        IWindowHandle paletteOwnerHandle = frame->handle();

             ITRACE_DEVELOP(IString("paletteOwnerHandle: ")
                           +IString(paletteOwnerHandle.asUnsigned()).d2x());

        // Select and realize the palette and invalidate the window.
        realizeColorMap(paletteOwnerHandle, false, false);
        handleRefresh();
      }
      fStopDispatch = true;
      lastWindowToChange = 0;
   }

   return fStopDispatch;
}

/*------------------------------------------------------------------------------
| IColorMapHandler::handleEventsFor                                            |
|                                                                              |
| WIN: Attach the handler to the specified window and realize the palette.     |
------------------------------------------------------------------------------*/
IColorMapHandler &IColorMapHandler::handleEventsFor( IFrameWindow* frameWindow )
{
  IFUNCTRACE_DEVELOP();
  IASSERTPARM(frameWindow!=0)


  if(fPaletteSupported)
  {
    // Start capturing palette-related events
    IHandler::handleEventsFor(frameWindow);
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IColorMapHandler::stopHandlingEventsFor                                      |
|                                                                              |
| WIN: Detach the handler from the specified window.                           |
------------------------------------------------------------------------------*/
IColorMapHandler &IColorMapHandler::stopHandlingEventsFor( IFrameWindow* frameWindow )
{
  IFUNCTRACE_DEVELOP();
  IASSERTPARM(frameWindow!=0)

  if(fPaletteSupported)
  {
    IHandler::stopHandlingEventsFor(frameWindow);
  }


  return *this;
}

/*------------------------------------------------------------------------------
| IColorMapHandler::realizeColorMap                                            |
|                                                                              |
| WIN: Select and realize the stored palette.                                  |
------------------------------------------------------------------------------*/
void IColorMapHandler::realizeColorMap( const IWindowHandle& paletteWindow,
                                        bool                 forceBackground,
                                        bool                 forceRefresh)
{
   IFUNCTRACE_DEVELOP();

   unsigned long entriesChanged = 0;
   if(fPaletteSupported)
   {
      HDC paletteHDC = GetDC(paletteWindow);
      HPALETTE palette = (HPALETTE)fColorMap->nativeColorMap();
      HPALETTE previousPalette = SelectPalette(paletteHDC,
                                               palette, forceBackground);
      entriesChanged = RealizePalette(paletteHDC);
      fHasFirstRealizationOccurred = true;

          ITRACE_DEVELOP(IString("palette: ")+IString((unsigned long)palette).d2x()
            + IString(", previousPalette: ")+IString((unsigned long)previousPalette).d2x()
            + IString(", entries: ")+IString(entriesChanged));

          ITRACE_DEVELOP(IString("  hdc: ")+IString((unsigned long)paletteHDC).d2x()
            + IString(", forceBackground: ")+IString((unsigned long)forceBackground));

      // Invalidate the window so it will be drawn with the new palette
      if(forceRefresh || entriesChanged!=0)
        UpdateColors( paletteHDC );

      // Realize the prior palette in the background.
      if(previousPalette)
      {
        SelectPalette(paletteHDC, (HPALETTE)previousPalette, true);
        RealizePalette(paletteHDC);

          ITRACE_DEVELOP(IString("  Select old palette, hdc: ")+IString((unsigned long)paletteHDC).d2x()
            + IString(", palette: ")+IString((unsigned long)previousPalette).d2x()
            + IString(", forceBackground: ")+IString("true"));
      }

      ReleaseDC(paletteWindow, paletteHDC);


   }
}

/*------------------------------------------------------------------------------
| IColorMapHandler::isPaletteDevice                                            |
|                                                                              |
| WIN: Return true if the screen is using palettes (vs true color).            |
------------------------------------------------------------------------------*/
bool IColorMapHandler::isPaletteDevice( )
{

   HDC hdc = GetDC(HWND_DESKTOP);
   int rasterCaps = GetDeviceCaps( hdc, RASTERCAPS);
   bool paletteSupported = ((rasterCaps & RC_PALETTE) == RC_PALETTE);
   ReleaseDC(HWND_DESKTOP, hdc);

   return paletteSupported;
}

#endif // IC_WIN


//////////////////////// PM ///////////////////////////

#ifdef IC_PM
/*------------------------------------------------------------------------------
| IColorMapHandler::dispatchHandlerEvent                                       |
|                                                                              |
| PM: Handle palette management messages.                                      |
------------------------------------------------------------------------------*/
bool IColorMapHandler::dispatchHandlerEvent(IEvent& event)
{
   bool fStopDispatch = false;
   if (event.eventId() == WM_REALIZEPALETTE)
   {
      IMODTRACE_DEVELOP("WM_REALIZEPALETTE");

      if (fPaletteSupported)
      {
         IFrameWindow* frame = dynamic_cast<IFrameWindow*>(event.dispatchingWindow());
         if(frame)
         {
           IWindowHandle::Value paletteOwnerHandle = frame->handle();

           // Select and realize the palette
           realizeColorMap(paletteOwnerHandle, false, false);

           fStopDispatch = true;
         }
      }
   } // WM_REALIZEPALETTE

   return fStopDispatch;
}

/*------------------------------------------------------------------------------
| IColorMapHandler::handleEventsFor                                            |
|                                                                              |
| PM: Attach the handler to the specified window.                              |
------------------------------------------------------------------------------*/
IColorMapHandler &IColorMapHandler::handleEventsFor( IFrameWindow* frameWindow )
{
  IFUNCTRACE_DEVELOP();
  IASSERTPARM(frameWindow!=0)


  if(fPaletteSupported)
  {
    IHandler::handleEventsFor(frameWindow);
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IColorMapHandler::stopHandlingEventsFor                                      |
|                                                                              |
| PM: Detach the handler from the specified window.                            |
------------------------------------------------------------------------------*/
IColorMapHandler &IColorMapHandler::stopHandlingEventsFor( IFrameWindow* frameWindow )
{
  IFUNCTRACE_DEVELOP();
  IASSERTPARM(frameWindow!=0)

  if(fPaletteSupported)
  {
     IHandler::stopHandlingEventsFor(frameWindow);
     if (frameWindow->isValid())
     {
        IPresSpaceHandle presSpace = WinGetPS(frameWindow->handle());
        GpiSelectPalette(presSpace, NULLHANDLE);
        WinReleasePS( presSpace);
     }
  }

  return *this;
}

/*------------------------------------------------------------------------------
| IColorMapHandler::realizeColorMap                                            |
|                                                                              |
| PM: Select and realize the stored palette.                                   |
------------------------------------------------------------------------------*/
void IColorMapHandler::realizeColorMap( const IWindowHandle& paletteWindow,
                                        bool,
                                        bool                 forceRefresh )
{
   IFUNCTRACE_DEVELOP();

   IPresSpaceHandle palettePresSpace = WinGetPS(paletteWindow);


   HPAL palette = (HPAL)fColorMap->nativeColorMap();
   HPAL previousPalette = (unsigned long)GpiSelectPalette(palettePresSpace, palette);
   unsigned long entriesChanged;
   WinRealizePalette(paletteWindow, palettePresSpace, &entriesChanged);
   fHasFirstRealizationOccurred = true;

     ITRACE_DEVELOP(IString("palette: ")+IString((unsigned long)palette).d2x()
                  + IString(", previousPalette: ")+IString(previousPalette).d2x()
                  + IString(", entries: ")+IString(entriesChanged));

   // Redraw the window
   if(forceRefresh || entriesChanged!=0)
       handleRefresh();

   // Release the PS
   WinReleasePS( palettePresSpace);
}

/*------------------------------------------------------------------------------
| IColorMapHandler::isPaletteDevice                                            |
|                                                                              |
| PM: Return true if the screen is using palettes (vs true color).             |
------------------------------------------------------------------------------*/
bool IColorMapHandler::isPaletteDevice( )
{

   long graphicsCaps;
   HDC hdc = WinQueryWindowDC(HWND_DESKTOP);
   if (!hdc)
       hdc = WinOpenWindowDC(HWND_DESKTOP);
   DevQueryCaps( hdc, CAPS_ADDITIONAL_GRAPHICS, 1,
                 &graphicsCaps);
   bool paletteSupported = ((graphicsCaps & CAPS_PALETTE_MANAGER) ==
                        CAPS_PALETTE_MANAGER);

   return paletteSupported;
}

#endif // IC_PM

///////////////////// Motif ///////////////////////////

#ifdef IC_MOTIF
/*------------------------------------------------------------------------------
| IColorMapHandler::dispatchHandlerEvent                                       |
|                                                                              |
| MOTIF: This function is not used under Motif.                                |
------------------------------------------------------------------------------*/
bool IColorMapHandler::dispatchHandlerEvent(IEvent& event)
{
   return false;
}

/*------------------------------------------------------------------------------
| IColorMapHandler::handleEventsFor                                            |
|                                                                              |
| MOTIF:  On Motif, we set the colormap property of the window.                |
------------------------------------------------------------------------------*/
IColorMapHandler &IColorMapHandler::handleEventsFor( IFrameWindow* frameWindow )
{
  IFUNCTRACE_DEVELOP();
  IASSERTPARM(frameWindow!=0)

  realizeColorMap(fFrameWindow->handle(), false, false);

  return *this;
}


/*------------------------------------------------------------------------------
| IColorMapHandler::realizeColorMap                                            |
|                                                                              |
| MOTIF: Set the colormap property of the frame and its children.              |
------------------------------------------------------------------------------*/
void IColorMapHandler::realizeColorMap( const IWindowHandle& ,
                                        bool, bool )
{
   IFUNCTRACE_DEVELOP();

   if(fPaletteSupported)
   {
      IWindowHandle::Value windowHandle = fFrameWindow->handle();
      Display *display = IXDisplay::display();
      Colormap colormap
         = fColorMap->nativeColorMap();

      // If a native colormap exists, store it as the colormap for the window
      // (one only exists for a display using a palette).
      if(colormap)
      {
         XtVaSetValues( (Widget)windowHandle, XmNcolormap,
                        (XtPointer)colormap, 0 );
         propagateColorMapToChildren(fFrameWindow->handle(), colormap);
         fHasFirstRealizationOccurred = true;
      }
   }
}


/*------------------------------------------------------------------------------
| IColorMapHandler::propagateColorMapToChildren                                |
|                                                                              |
| MOTIF: Set the colormap property of the children of the window.              |
------------------------------------------------------------------------------*/
void IColorMapHandler::propagateColorMapToChildren ( const IWindowHandle& windowHandle,
                                                     unsigned long colormap)
{

   // if widget can have children then pass the color down.
   if (XtIsComposite((Widget)windowHandle) )
   {
     _WidgetRec
       **children;
     unsigned int
       num_children(0);

     XtVaGetValues( (Widget)windowHandle,
                    XmNchildren    , &children,
                    XmNnumChildren , &num_children,
                    NULL );
     for ( int i = 0; i < num_children; i ++ )
     {
       // if child is not being destroy then set its colormap resource
       if ( !children[i]->core.being_destroyed )
       {

         XtVaSetValues( children[i], XmNcolormap,
                         (XtPointer)colormap, 0 );
         propagateColorMapToChildren( (IWindowHandle)children[i], colormap );
       }
     }
   }
}
/*------------------------------------------------------------------------------
| IColorMapHandler::stopHandlingEventsFor                                      |
|                                                                              |
| MOTIF: Reset the default colormap into the window.                           |
------------------------------------------------------------------------------*/
IColorMapHandler &IColorMapHandler::stopHandlingEventsFor( IFrameWindow* frameWindow )
{
  IFUNCTRACE_DEVELOP();
  IASSERTPARM(frameWindow!=0)

    // IC_NOTYET  Need to restore the prior colormap??

  return *this;
}


/*------------------------------------------------------------------------------
| IColorMapHandler::isPaletteDevice                                            |
|                                                                              |
| MOTIF: Return true if the screen is using palettes (vs true color).          |
------------------------------------------------------------------------------*/
bool IColorMapHandler::isPaletteDevice( )
{

   Display* display = IXDisplay::display();

   Window rootWindow = DefaultRootWindow( display );
   XWindowAttributes winAttrib;

   XGetWindowAttributes( display, rootWindow, &winAttrib);

   if (winAttrib.depth <= 8)
      return true;
   else
      return false;
}

#endif  // IC_MOTIF

/*------------------------------------------------------------------------------
| IColorMapHandler::handleEventsFor                                            |
|                                                                              |
| ALL: Disallow attachment of the handler to a non-frame window.                    |
------------------------------------------------------------------------------*/
IHandler &IColorMapHandler::handleEventsFor( IWindow* )
{
   ITHROWLIBRARYERROR(IC_MEMBER_ACCESS_ERROR,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);

   return *this;
}

/*------------------------------------------------------------------------------
| IColorMapHandler::stopHandlingEventsFor                                      |
|                                                                              |
| ALL: Disallow removal of the handler from a non-frame window.                |
------------------------------------------------------------------------------*/
IHandler &IColorMapHandler::stopHandlingEventsFor( IWindow* )
{
   ITHROWLIBRARYERROR(IC_MEMBER_ACCESS_ERROR,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);

   return *this;
}

#if IC_EXTENDED_RTTI_ENABLED
// private functions that don't need an implementation except to make
// the 96a compiler happy
IColorMapHandler::IColorMapHandler (const IColorMapHandler&)
{}

IColorMapHandler& IColorMapHandler::operator= (const IColorMapHandler& newColorMap)
{
   return (IColorMapHandler&)newColorMap;
}
#endif
