// Revision: 42 1.10 source/ui/baseapp/iframext.cpp, frame, ioc.v400, 000302 
/*******************************************************************************
* FILE NAME: iframext.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iframext.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
extern "C" {
  #define INCL_WINFRAMEMGR
  #define INCL_WINSYS
  #include <iwindefs.h>
  #include <math.h>
  #ifdef IC_MOTIF
    #include <Xm/Separator.h>
  #endif //IC_MOTIF
  }

#include <iframext.hpp>
#include <ibcolor.hpp>
#include <ibidiset.hpp>
#include <ibundles.hpp>
#include <iexgrprt.hpp>
#include <igline2d.hpp>
#include <ihandle.hpp>
#include <iplatfrm.hpp>
#include <ipoint.hpp>
#include <iswp.hpp>
#include <itrace.hpp>
#include <iseq2.h>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IFRAMEXT_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| IFrameExtension::IFrameExtension                                             |
------------------------------------------------------------------------------*/
IFrameExtension :: IFrameExtension ( IWindow                     *window,
                                     IFrameWindow::Location       location,
                                     unsigned long                widthOrHeight,
                                     IFrameWindow::SeparatorType  separator )
  : pData( 0 )
  , extType( IFrameExtension::fixed )
  , extFixedSize( widthOrHeight )
  , extLocation( location )
  , extSeparator( separator )
  , extControl( window )
  {
  }

/*------------------------------------------------------------------------------
| IFrameExtension::IFrameExtension                                             |
------------------------------------------------------------------------------*/
IFrameExtension :: IFrameExtension ( IWindow                     *window,
                                     IFrameWindow::Location       location,
                                     double                       percentage,
                                     IFrameWindow::SeparatorType  separator )
  : pData( 0 )
  , extType( IFrameExtension::relative )
  , extRelativeSize( percentage )
  , extLocation( location )
  , extSeparator( separator )
  , extControl( window )
  {
  }

/*------------------------------------------------------------------------------
| IFrameExtension::IFrameExtension                                             |
------------------------------------------------------------------------------*/
IFrameExtension :: IFrameExtension ( IWindow                     *window,
                                     IFrameWindow::Location       location,
                                     IFrameWindow::SeparatorType  separator )
  : pData( 0 )
  , extType( IFrameExtension::minimumSize )
  , extFixedSize( 0 )
  , extLocation( location )
  , extSeparator( separator )
  , extControl( window )
  {
  }

/*------------------------------------------------------------------------------
| IFrameExtension::~IFrameExtension                                            |
------------------------------------------------------------------------------*/
IFrameExtension :: ~IFrameExtension ( )
  {
  }

/*------------------------------------------------------------------------------
| IFrameExtension::attachedToId                                                |
|                                                                              |
| Return id of attached-to frame, based on location data member value.         |
------------------------------------------------------------------------------*/
unsigned long IFrameExtension :: attachedToId ( ) const
  {
  switch ( this->extLocation )
    {
    case IFrameWindow::rightOfTitleBar:
    case IFrameWindow::leftOfTitleBar:
#ifdef IC_MOTIFWIN
      return 0;          // not supported ... causes extension to be hidden
#endif //IC_MOTIFWIN
#ifdef IC_PM
      return FID_TITLEBAR;
#endif //IC_PM
    case IFrameWindow::rightOfMenuBar:
    case IFrameWindow::leftOfMenuBar:
#ifdef IC_MOTIFWIN
      return 0;          // not supported ... causes extension to be hidden
#endif //IC_MOTIFWIN
#ifdef IC_PM
      return FID_MENU;
#endif //IC_PM
    default:
      return FID_CLIENT;
    }
  }

/*------------------------------------------------------------------------------
| IFrameExtension::attachTo                                                    |
|                                                                              |
| Calculate the size and position of the extension from the argument base      |
| control size/position, according to extension location.  Adjust the          |
| size and position of the attached-to frame control.                          |
------------------------------------------------------------------------------*/
ISWP IFrameExtension :: attachTo ( ISWP &baseSWP )
  {
  IMODTRACE_DEVELOP( "IFrameExtension::attachTo" );

  ITRACE_ALL( IString( "input baseSWP=" )
              +
              baseSWP.position().asString()
              +
              baseSWP.size().asString() );

  ISWP
    extSWP;

  // Set extension size...
  extSWP.setSize( this->sizeTo( baseSWP.size() ) );

  // Remap the location for bidi support, if necessary.
  IFrameWindow::Location
    locationForLayout = this->mappedLocation();

  // Set extension position...
  switch ( locationForLayout )
    {
#ifdef IC_ORIGINUPPERLEFT       // 0,0 upper left
    case IFrameWindow::aboveClient:
#else                           // 0,0 lower left
    case IFrameWindow::belowClient:
#endif
    case IFrameWindow::leftOfClient:
    case IFrameWindow::leftOfTitleBar:
    case IFrameWindow::leftOfMenuBar:
      extSWP.setPosition( baseSWP.position() );
      break;
#ifdef IC_ORIGINUPPERLEFT       // 0,0 upper left
    case IFrameWindow::belowClient:
#else                           // 0,0 lower left
    case IFrameWindow::aboveClient:
#endif
    case IFrameWindow::rightOfClient:
    case IFrameWindow::rightOfTitleBar:
    case IFrameWindow::rightOfMenuBar:
      extSWP.setPosition( baseSWP.position()
                          +
                          baseSWP.size()
                          -
                          extSWP.size() );
      break;
    }

  // Remember size/position for use during painting...
  this->extPosition = extSWP.position();
  this->extSize     = extSWP.size();

  // Adjust base control position...
  switch ( locationForLayout )
    {
    case IFrameWindow::aboveClient:
#ifdef IC_ORIGINUPPERLEFT     //  0,0 upper left
      baseSWP
        .setPosition( baseSWP.position()
                      +
                      IPoint( 0, extSWP.size().height()
                                 +
                                 this->separatorWidth() ) );
#endif
      break;
    case IFrameWindow::rightOfClient:
    case IFrameWindow::rightOfTitleBar:
    case IFrameWindow::rightOfMenuBar:
      break;
    case IFrameWindow::belowClient:
#ifdef  IC_ORIGINLOWERLEFT    //  0,0 lower left
      baseSWP
        .setPosition( baseSWP.position()
                      +
                      IPoint( 0, extSWP.size().height()
                                 +
                                 this->separatorWidth() ) );
#endif
      break;
    case IFrameWindow::leftOfClient:
    case IFrameWindow::leftOfTitleBar:
    case IFrameWindow::leftOfMenuBar:
      baseSWP
        .setPosition( baseSWP.position()
                      +
                      IPoint( extSWP.size().width()
                              +
                              this->separatorWidth(), 0 ) );
      break;
    }

  // Adjust base control size...
  if ( this->location() == IFrameWindow::aboveClient
       ||
       this->location() == IFrameWindow::belowClient )
    baseSWP.setSize( baseSWP.size() - IPoint( 0, extSWP.size().height()
                                                 +
                                                 this->separatorWidth() ) );
  else
    baseSWP.setSize( baseSWP.size() - IPoint( extSWP.size().width()
                                              +
                                              this->separatorWidth(), 0 ) );

  ITRACE_ALL( IString( "output baseSWP=" )
              +
              baseSWP.position().asString()
              +
              baseSWP.size().asString() );

  ITRACE_ALL( IString( "output extSWP=" )
              +
              extSWP.position().asString()
              +
              extSWP.size().asString() );

  return extSWP;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::sizeTo                                                      |
|                                                                              |
| Calculates and returns the size this extension would have when attached to   |
| a control of the argument size.                                              |
| Notes: If top/bottom, then the width is the same, otherwise the height is    |
|          the same.                                                           |
|        If fixed size, then the other dimension is fixed, otherwise it is     |
|          calculated relative to the base dimension.                          |
------------------------------------------------------------------------------*/
ISize IFrameExtension :: sizeTo ( const ISize &baseSize ) const
  {
  IMODTRACE_DEVELOP( "IFrameExtension::sizeTo" );

  ITRACE_ALL( IString( "input=" ) + baseSize.asString() );

  ISize
    result = baseSize;

  // Remap the location for bidi support, if necessary.
  IFrameWindow::Location
    locationForLayout = this->mappedLocation();

  switch ( locationForLayout )
    {
    case IFrameWindow::aboveClient:
    case IFrameWindow::belowClient:
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() < result.height() )
          result.setHeight( this->fixedSize() );
        }
      else if ( this->type() == minimumSize )
        {
        result.setHeight( this->control()->minimumSize().height() );
        }
      else
        {
        result.setHeight( (long)(result.height() * this->relativeSize()) );
        }
      break;

    default:
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() < result.width() )
          result.setWidth( this->fixedSize() );
        }
      else if ( this->type() == minimumSize )
        {
        result.setWidth( this->control()->minimumSize().width() );
        }
      else
        {
        result.setWidth( (long)(result.width() * this->relativeSize()) );
        }
      break;
    }

  ITRACE_ALL( IString( "output=" ) + result.asString() );

  return result;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::totalRectFor                                                |
|                                                                              |
| Calculates and returns a rectangle from a base rectangle plus                |
| the frame extensions size and position.                                      |
------------------------------------------------------------------------------*/
IRectangle IFrameExtension::totalRectFor( const IRectangle& baseRect ) const
  {
  IMODTRACE_DEVELOP( "IFrameExtension::totalRectFor" );

  ITRACE_ALL( IString( "input=" ) + baseRect.asString() );

  IRectangle result = baseRect;

  long delta = 0;

  // Remap the location for bidi support, if necessary.
  IFrameWindow::Location
    locationForLayout = this->mappedLocation();

  switch ( locationForLayout )
    {
#ifdef  IC_ORIGINUPPERLEFT   // 0,0 upper left
    // If the extension is below the client, increase the height of the
    // resulting rectangle by the height of the frame extension.
    case IFrameWindow::belowClient:
#else                       // 0,0 lower left
    // If the extension is above the client, increase the height of the
    // resulting rectangle by the height of the frame extension.
    case IFrameWindow::aboveClient:
#endif
      delta = result.height();
      if ( this->type() == fixed )
        {
        delta += this->fixedSize();
        }
      else if ( this->type() == minimumSize )
        {
        delta += this->control()->minimumSize().height();
        }
      else
        {
        delta = (long)ceil( delta / (1.0 - this->relativeSize()) );
        }

      result.sizeTo( IPair( result.width(),
                            delta + this->separatorWidth() ));
      break;

#ifdef  IC_ORIGINUPPERLEFT      //  0,0 upper left
    // If the extension is above the client, increase the height of the
    // resulting rectangle by the height of the frame extension and move it
    // up by the same amount.
    case IFrameWindow::aboveClient:
#else
    // If the extension is below the client, increase the height of the
    // resulting rectangle by the height of the frame extension and move it
    // down by the same amount.
    case IFrameWindow::belowClient:
#endif
      if ( this->type() == fixed )
        {
        delta = this->fixedSize();
        }
      else if ( this->type() == minimumSize )
        {
        delta = this->control()->minimumSize().height();
        }
      else
        {
        delta = (long)ceil( result.height() * this->relativeSize() /
                            (1.0 - this->relativeSize()) );
        }

      delta += this->separatorWidth();
      result
       .sizeTo( IPair( result.width(), result.height() + delta ))
       .moveTo( IPair( result.left(), result.bottom() - delta ));
      break;

    // If the extension is to the left, increase the width of the resulting
    // rectangle by the width of the frame extension and move it to the left
    // by the same amount.
    case IFrameWindow::leftOfClient:
    case IFrameWindow::leftOfTitleBar:
    case IFrameWindow::leftOfMenuBar:
      if ( this->type() == fixed )
        {
        delta = this->fixedSize();
        }
      else if ( this->type() == minimumSize )
        {
        delta = this->control()->minimumSize().width();
        }
      else
        {
        delta = (long)ceil( result.width() * this->relativeSize() /
                            (1.0 - this->relativeSize()) );
        }

      delta += this->separatorWidth();
      result
       .sizeTo( IPair( result.width() + delta, result.height() ))
       .moveTo( IPair( result.left() - delta, result.bottom() ));
      break;

    // If the extension is to the right, increase the width of the resulting
    // rectangle by the width of the frame extension.
    case IFrameWindow::rightOfClient:
    case IFrameWindow::rightOfTitleBar:
    case IFrameWindow::rightOfMenuBar:
      delta = result.width();
      if ( this->type() == fixed )
        {
        delta += this->fixedSize();
        }
      else if ( this->type() == minimumSize )
        {
        delta += this->control()->minimumSize().width();
        }
      else
        {
        delta = (long)ceil( delta / (1.0 - this->relativeSize()) );
        }

      result.sizeTo( IPair( delta, result.height() ));
      break;
    }

  return result;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::baseRectFor                                                 |
|                                                                              |
| Calculates and returns a rectangle from a total rectangle (i.e. a client     |
| rectangle with a frame extension) minus the frame extensions size and        |
| position.                                                                    |
------------------------------------------------------------------------------*/
IRectangle IFrameExtension::baseRectFor( const IRectangle& totalRect ) const
  {
  IMODTRACE_DEVELOP( "IFrameExtension::baseRectFor" );

  ITRACE_ALL( IString( "input=" ) + totalRect.asString() );

  IRectangle result = totalRect;

  long delta = 0;
  bool adjust = true;

  // Remap the location for bidi support, if necessary.
  IFrameWindow::Location
    locationForLayout = this->mappedLocation();

  switch ( locationForLayout )
    {
#ifdef  IC_ORIGINUPPERLEFT      // 0,0 upper left
    // If the extension is below the client, decrease the height of the
    // resulting rectangle by the height of the frame extension.
    case IFrameWindow::belowClient:
#else                           // 0,0 lower left
    // If the extension is above the client, decrease the height of the
    // resulting rectangle by the height of the frame extension.
    case IFrameWindow::aboveClient:
#endif
      delta = result.height();
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() > result.height() )
          {
          delta = this->separatorWidth();
          }
        else
          {
          delta -= this->fixedSize();
          }
        }
      else if ( this->type() == minimumSize )
        {
        delta -= this->control()->minimumSize().height();
        }
      else
        {
        delta -= floor( delta * this->relativeSize() );
        }

      result.sizeTo( IPair( result.width(),
                            delta - this->separatorWidth() ));
      break;

#ifdef  IC_ORIGINUPPERLEFT      //  0,0 upper left
    // If the extension is above the client, decrease the height of the
    // resulting rectangle by the height of the frame extension and move it
    // down by the same amount.
    case IFrameWindow::aboveClient:
#else                           //  0,0 lower left
    // If the extension is below the client, decrease the height of the
    // resulting rectangle by the height of the frame extension and move it
    // up by the same amount.
    case IFrameWindow::belowClient:
#endif
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() > result.height() )
          {
          result.sizeTo( IPair( result.width(), 0 ));
          adjust = false;
          }
        else
          {
          delta = this->fixedSize();
          }
        }
      else if ( this->type() == minimumSize )
        {
        delta = this->control()->minimumSize().height();
        }
      else
        {
        delta = (long)floor( this->relativeSize() * result.height() );
        }

      if ( adjust )
      {
         delta += this->separatorWidth();
         result
          .sizeTo( IPair( result.width(), result.height() - delta ))
          .moveTo( IPair( result.left(), result.bottom() + delta ));
      }
      break;

    // If the extension is to the left, decrease the width of the resulting
    // rectangle by the width of the frame extension and move it to the right
    // by the same amount.
    case IFrameWindow::leftOfClient:
    case IFrameWindow::leftOfTitleBar:
    case IFrameWindow::leftOfMenuBar:
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() > result.width() )
          {
          result.sizeTo( IPair( 0, result.height() ));
          adjust = false;
          }
        else
          {
          delta = this->fixedSize();
          }
        }
      else if ( this->type() == minimumSize )
        {
        delta = this->control()->minimumSize().width();
        }
      else
        {
        delta = (long)floor( this->relativeSize() * result.width() );
        }

      if ( adjust )
      {
         delta += this->separatorWidth();
         result
          .sizeTo( IPair( result.width() - delta, result.height() ))
          .moveTo( IPair( result.left() + delta, result.bottom() ));
      }
      break;

    // If the extension is to the right, decrease the width of the resulting
    // rectangle by the width of the frame extension.
    case IFrameWindow::rightOfClient:
    case IFrameWindow::rightOfTitleBar:
    case IFrameWindow::rightOfMenuBar:
      delta = result.width();
      if ( this->type() == fixed )
        {
        if ( this->fixedSize() > result.height() )
          {
          delta = this->separatorWidth();
          }
        else
          {
          delta -= this->fixedSize();
          }
        }
      else if ( this->type() == minimumSize )
        {
        delta -= this->control()->minimumSize().width();
        }
      else
        {
        delta -= floor( delta * this->relativeSize() );
        }

      result.sizeTo( IPair( delta - this->separatorWidth(),
                            result.height() ));
      break;
    }

  return result;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::drawSeparator                                               |
|                                                                              |
| Draws the separator line between an extension and the adjacent control       |
| (or extension) to which it is attached.                                      |
------------------------------------------------------------------------------*/
void IFrameExtension :: drawSeparator ( const IPresSpaceHandle &hps )
  {
  IMODTRACE_DEVELOP( "IFrameExtension::drawSeparator" );

  // If no separator, don't even bother...
  if ( this->extSeparator == IFrameWindow::none )
    return;

  // Set up arrays for line start/end points and colors:
  IPoint
    start[3],
    end  [3];

  int
    numLines = (int)this->separatorWidth();

  // Otherwise, set up line start/end points according to location...

  // Remap the location for bidi support, if necessary.
  IFrameWindow::Location
    locationForLayout = this->mappedLocation();

  switch ( locationForLayout )
    {
#ifdef  IC_ORIGINUPPERLEFT      // 0,0 upper left
    case IFrameWindow::belowClient:
#else                           // 0,0 lower left
    case IFrameWindow::aboveClient:
#endif
      start[0] = start[1] = start[2] = this->extPosition
                                       -
                                       IPoint( 0, 1 );
      start[1] -= IPoint( 0, 1 );
      start[2] -= IPoint( 0, 2 );
      end[0] = start[0] + IPoint( this->extSize.width(), 0 );
      end[1] = start[1] + IPoint( this->extSize.width(), 0 );
      end[2] = start[2] + IPoint( this->extSize.width(), 0 );
      break;
#ifdef  IC_ORIGINUPPERLEFT      // 0,0 upper left
    case IFrameWindow::aboveClient:
#else                           // 0,0 lower left
    case IFrameWindow::belowClient:
#endif
      start[0] = start[1] = start[2] = this->extPosition
                                       +
                                       IPoint( 0, this->extSize.height() );
      start[1] += IPoint( 0, 1 );
      start[2] += IPoint( 0, 2 );
      end[0] = start[0] + IPoint( this->extSize.width(), 0 );
      end[1] = start[1] + IPoint( this->extSize.width(), 0 );
      end[2] = start[2] + IPoint( this->extSize.width(), 0 );
      break;
    case IFrameWindow::leftOfClient:
    case IFrameWindow::leftOfMenuBar:
    case IFrameWindow::leftOfTitleBar:
      start[0] = start[1] = start[2] = this->extPosition
                                       +
                                       IPoint( this->extSize.width(), 0 );
      start[1] += IPoint( 1, 0 );
      start[2] += IPoint( 2, 0 );
      end[0] = start[0] + IPoint( 0, this->extSize.height() );
      end[1] = start[1] + IPoint( 0, this->extSize.height() );
      end[2] = start[2] + IPoint( 0, this->extSize.height() );
      break;
    case IFrameWindow::rightOfClient:
    case IFrameWindow::rightOfMenuBar:
    case IFrameWindow::rightOfTitleBar:
      start[0] = start[1] = start[2] = this->extPosition
                                       -
                                       IPoint( 1, 0 );
      start[1] -= IPoint( 1, 0 );
      start[2] -= IPoint( 2, 0 );
      end[0] = start[0] + IPoint( 0, this->extSize.height() );
      end[1] = start[1] + IPoint( 0, this->extSize.height() );
      end[2] = start[2] + IPoint( 0, this->extSize.height() );
      break;
    }

  IExtendedRootGrafPort rootPort(hps, ICoordinateSystem::nativeOrientation());
  IColor evenColor(IColor::kButtonDark);
  IColor oddColor (IColor::kButtonMiddle);
  while ( numLines-- )
    {
    ITRACE_ALL( "Line:"
                +
                start[numLines].asString()
                +
                end[numLines].asString() );

    IFrameBundle    bundle ( ((numLines % 2) == 0) ? evenColor : oddColor);
    ILinkedGrafPort linkedPort(&rootPort, &bundle);
    IGLine2D        line( start[numLines], end[numLines] );
    linkedPort.draw( line );
    }
  }


/*------------------------------------------------------------------------------
| IFrameExtension::separatorWidth                                              |
|                                                                              |
| Returns the width of the separator line.                                     |
------------------------------------------------------------------------------*/
unsigned long IFrameExtension :: separatorWidth ( ) const
  {
  unsigned long thickness = 0;
  switch ( this->extSeparator )
    {
    case IFrameWindow::none:
    default:
      break;            // Use default = 0.
    case IFrameWindow::thinLine:
      thickness = 1;
      break;
    case IFrameWindow::thickLine:
      thickness = 3;
      break;
    }
  return thickness;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::setSize                                                     |
|                                                                              |
| Resets the size of the extension.                                            |
------------------------------------------------------------------------------*/
IFrameExtension &IFrameExtension :: setSize ( unsigned long widthOrHeight )
  {
  this->extType = IFrameExtension::fixed;
  this->extFixedSize = widthOrHeight;

#ifdef IC_MOTIF
  if ((this->location() == IFrameWindow::aboveClient) ||
      (this->location() == IFrameWindow::belowClient))
     XtVaSetValues(this->control()->handle(),
                   XmNheight, widthOrHeight,
                   NULL);
  else
     XtVaSetValues(this->control()->handle(),
                   XmNwidth, widthOrHeight,
                   NULL);
#endif //IC_MOTIF

  return *this;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::setSize                                                     |
|                                                                              |
| Convert width/height arg to unsigned long and use overloaded function.       |
------------------------------------------------------------------------------*/
IFrameExtension &IFrameExtension :: setSize ( int widthOrHeight )
  {
  return this->setSize( (unsigned long)widthOrHeight );
  }

/*------------------------------------------------------------------------------
| IFrameExtension::setSize                                                     |
|                                                                              |
| Resets the size of the extension.                                            |
------------------------------------------------------------------------------*/
IFrameExtension &IFrameExtension :: setSize ( double widthOrHeight )
  {
  this->extType = IFrameExtension::relative;
  this->extRelativeSize = widthOrHeight;

  return *this;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::useMinimumSize                                              |
|                                                                              |
| Specifies that the height or width of the extension should be sized to the   |
| height or width of its window's minimum size.                                |
------------------------------------------------------------------------------*/
IFrameExtension &IFrameExtension :: useMinimumSize ( )
  {
  this->extType = IFrameExtension::minimumSize;
  this->control()->setLayoutDistorted( IWindow::minimumSizeChanged, 0 );
  return *this;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::relativeSize                                                |
|                                                                              |
| Returns the relative width or height as a percentage                         |
| (0.0 if the extension is of type fixed).                                     |
------------------------------------------------------------------------------*/
double IFrameExtension :: relativeSize ( ) const
  {
  return ( this->extType == IFrameExtension::relative )
    ? this->extRelativeSize
    : 0.0;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::fixedSize                                                   |
|                                                                              |
| Returns the fixed width or height in window coordinates                      |
| (0 if the extension is of type relative).                                    |
------------------------------------------------------------------------------*/
unsigned long IFrameExtension :: fixedSize ( ) const
  {
  return ( this->extType == IFrameExtension::fixed )
    ? this->extFixedSize
    : 0;
  }

/*------------------------------------------------------------------------------
| IFrameExtension::mappedLocation                                              |
|                                                                              |
| Maps the extension location to the one that the extension effectively uses   |
| on the display.  Placing bidi support into this routine insulates formatting |
| code in IFrameWindow/IFrameHandler from having to be bidi aware.             |
| Note that this routine does not map the title-bar-relative locations.  This  |
| is because enabling an application for bidi generally does not change the    |
| location of the system menu and menu buttons to the right of the title bar.  |
------------------------------------------------------------------------------*/
IFrameWindow::Location IFrameExtension::mappedLocation ( ) const
{
  IFrameWindow::Location
    layoutLocation = this->location();

  // Remap the location for bidi support if necessary.
  if ( IBidiSettings::isBidiSupported() )
  {
     IWindow
      *frameWindow = this->control()->parent();
     IBidiSettings
       settings( *frameWindow );
     if ( settings.windowLayout() == IBidiSettings::layoutRightToLeft )
     {
        // Frame extensions need to be layed out right to left, rather
        // than left to right.
        switch ( layoutLocation )
        {
           case IFrameWindow::leftOfClient:
              layoutLocation = IFrameWindow::rightOfClient;
              break;
           case IFrameWindow::rightOfClient:
              layoutLocation = IFrameWindow::leftOfClient;
              break;
           case IFrameWindow::leftOfMenuBar:
              layoutLocation = IFrameWindow::rightOfMenuBar;
              break;
           case IFrameWindow::rightOfMenuBar:
              layoutLocation = IFrameWindow::leftOfMenuBar;
              break;
           case IFrameWindow::leftOfTitleBar:
           case IFrameWindow::rightOfTitleBar:
              // Don't modify the title bar area, because the OS doesn't
              // switch the position of the system menu, etc.  Also this
              // causes a problem for vertical floating tool bars.
              break;
           default:
              // Not affected by bidi.
              break;
        }
     }
  }  // end bidi supported

  return layoutLocation;
}


class IFrameExtensionsImp : public IVPtrSequence<IFrameExtension*> {
public:
  IFrameExtensionsImp ( );
 ~IFrameExtensionsImp ( );
}; // IFrameExtensionsImp

IFrameExtensionsImp::IFrameExtensionsImp ( )
{
}


IFrameExtensionsImp::~IFrameExtensionsImp ( )
{
}

IFrameExtensions::Cursor::Cursor ( const IFrameExtensions& extensions )
  : fExtensionsImp( extensions.fExtensionsImp )
  , fExtensionsCursorImp( new IFrameExtensionsImp::Cursor( *fExtensionsImp ) )  
{ }
       

IFrameExtensions::Cursor::~Cursor ( )
{
  delete (IFrameExtensionsImp::Cursor*)fExtensionsCursorImp;
}

bool IFrameExtensions::Cursor::setToFirst ( )
{ 
  return ((IFrameExtensionsImp::Cursor*)fExtensionsCursorImp)->setToFirst();
}


bool IFrameExtensions::Cursor::setToLast ()
{ 
  return ((IFrameExtensionsImp::Cursor*)fExtensionsCursorImp)->setToLast();
}

bool IFrameExtensions::Cursor::setToNext ()
{ 
  return ((IFrameExtensionsImp::Cursor*)fExtensionsCursorImp)->setToNext();
}

bool IFrameExtensions::Cursor::setToPrevious ()
{ 
  return ((IFrameExtensionsImp::Cursor*)fExtensionsCursorImp)->setToPrevious();
}

bool IFrameExtensions::Cursor::isValid() const
{ 
  return ((IFrameExtensionsImp::Cursor*)fExtensionsCursorImp)->isValid();
}

void IFrameExtensions::Cursor::invalidate ( )
{ 
  ((IFrameExtensionsImp::Cursor*)fExtensionsCursorImp)->invalidate();
}

IFrameExtension* const IFrameExtensions::Cursor::element ( ) const
{ 
  return ((const IFrameExtensionsImp::Cursor*)fExtensionsCursorImp)->element();
}     


IFrameExtensions :: IFrameExtensions ( ) 
  : fExtensionsImp( new IFrameExtensionsImp )
{ }


IFrameExtensions :: ~IFrameExtensions ( ) 
{
  delete fExtensionsImp;   
}


bool IFrameExtensions::add ( IFrameExtension* const& extension)
{
  return fExtensionsImp->add( extension );
}

void IFrameExtensions::addAsLast ( IFrameExtension* const& extension)
{
  fExtensionsImp->addAsLast( extension );
}
 
void IFrameExtensions::addAsFirst ( IFrameExtension* const& extension )
{
  fExtensionsImp->addAsFirst( extension );
}

void IFrameExtensions::addAsNext ( IFrameExtension* const& extension, 
                                   Cursor& cursor )
{  
  fExtensionsImp->addAsNext( extension, 
                             *(IFrameExtensionsImp::Cursor*)(cursor.fExtensionsCursorImp) );
}

void IFrameExtensions::addAtPosition ( unsigned long pos, 
                                       IFrameExtension* const& extension )
{
  fExtensionsImp->addAtPosition( pos, extension );
}

void IFrameExtensions::addAsPrevious ( IFrameExtension* const& extension, 
                                        Cursor& cursor )
{
  fExtensionsImp->addAsPrevious( extension, 
                                 *(IFrameExtensionsImp::Cursor*)(cursor.fExtensionsCursorImp) );
}

IFrameExtension* IFrameExtensions::elementAt ( const Cursor& cursor) const
{
  return fExtensionsImp->elementAt( *(IFrameExtensionsImp::Cursor*)(cursor.fExtensionsCursorImp) );
}

IFrameExtension* IFrameExtensions::elementAtPosition ( unsigned long position) const
{
  return fExtensionsImp->elementAtPosition( position );
}

IFrameExtension* IFrameExtensions::firstElement ( ) const
{
  return fExtensionsImp->firstElement();
}

IFrameExtension* IFrameExtensions::lastElement ( ) const
{
  return fExtensionsImp->lastElement();
}

bool IFrameExtensions::isEmpty ( ) const 
{
  return fExtensionsImp->isEmpty();
}

unsigned long IFrameExtensions::position ( Cursor const& cursor ) const
{
  return fExtensionsImp->position( *(const IFrameExtensionsImp::Cursor*)cursor.fExtensionsCursorImp );
}

unsigned long IFrameExtensions::numberOfElements ( ) const 
{
  return fExtensionsImp->numberOfElements();
}

unsigned long IFrameExtensions::removeAll ( )
{
  return fExtensionsImp->removeAll();
}

void IFrameExtensions::removeAt ( Cursor& cursor )
{
  fExtensionsImp->removeAt( *(IFrameExtensionsImp::Cursor*)cursor.fExtensionsCursorImp );
}

void IFrameExtensions::removeAtPosition ( unsigned long position )
{
  fExtensionsImp->removeAtPosition( position );
}

void IFrameExtensions::removeFirst ( )
{
  fExtensionsImp->removeFirst();
}

void IFrameExtensions::removeLast ( )
{
  fExtensionsImp->removeLast();
}

void IFrameExtensions::replaceAt ( const Cursor& cursor, 
                                   IFrameExtension* const& extension )
{
  fExtensionsImp->replaceAt( *(const IFrameExtensionsImp::Cursor*)cursor.fExtensionsCursorImp,
                             extension );
}

void IFrameExtensions::setToPosition ( unsigned long pos, Cursor& cursor) const
{
  fExtensionsImp->setToPosition( pos, 
                                 *(IFrameExtensionsImp::Cursor*)cursor.fExtensionsCursorImp);
}

