// Revision: 26 1.22 source/ui/baseapp/inotehdr.cpp, notebook, ioc.v400, 000302 
/*******************************************************************************
* FILE NAME: inotehdr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     INotebookHandler                                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#pragma priority( -2147481424 )

extern "C"
 {
  #define INCL_WINFRAMEMGR
  #define INCL_WININPUT             // WM_CHAR, WinGetKeyState, WinQueryFocus
  #define INCL_WINSTDBOOK
  #define INCL_WINDIALOGS           // Needed for PDLGTEMPLATE in BOOKPAGEINFO
  #include <iwindefs.h>

  #ifdef IC_WIN
    #ifndef _WIN32
      #define _WIN32
    #endif
    #include <commctrl.h>
  #endif
 }

#include <icmdevt.hpp>
#include <inotehdr.hpp>
#include <iwindow.hpp>
#include <inotebk.hpp>
#include <inotebk0.hpp>
#include <iplatfrm.hpp>
#include <icconst.h>
#include <icanvas.hpp>
#include <iframe.hpp>

#ifdef IC_WIN
  #include <iclnbw.h>
  #include <iwcname.hpp>
#endif


// Segment definitions
#ifdef IC_PAGETUNE
  #define _INOTEHDR_CPP_
  #include <ipagetun.h>
#endif

#ifdef IC_WIN
static unsigned long ulCurrentPageId = 0;
#endif

/*------------------------------------------------------------------------------
| INotebookHandler::INotebookHandler                                           |
------------------------------------------------------------------------------*/
INotebookHandler::INotebookHandler ( )
#ifdef IC_WIN
  : fFocusFromPage( false )
#endif // IC_WIN
{
}

/*------------------------------------------------------------------------------
| INotebookHandler::~INotebookHandler                                          |
------------------------------------------------------------------------------*/
INotebookHandler::~INotebookHandler ( )
{
}

/*------------------------------------------------------------------------------
| INotebookHandler::dispatchHandlerEvent                                       |
|                                                                              |
| Dispatch command functions for this handler.                                 |
------------------------------------------------------------------------------*/
bool INotebookHandler::dispatchHandlerEvent ( IEvent &event )
{
  bool stopProcessingEvent = false;

  switch ( event.eventId() )
  {
#ifdef IC_PM
    case WM_HELP:
      {
        if ( !event.window()->hasFocus() )
        {          // Focus is not on the notebook (and hence, not a tab).
          IWindow* parent = event.window()->parent();
          if (parent)
          {
            /******************************************************************/
            /* The user must have requested contextual help from a            */
            /* page window.  We will avoid the notebook's processing          */
            /* of WM_HELP, which is geared for its tabs.  If the help         */
            /* request gets here, the page window must not be a frame         */
            /* (it is probably a canvas).  By routing the requests to         */
            /* the notebook's parent, these contextual helps can be           */
            /* resolved through the help subtable of the notebook's           */
            /* frame window.                                                  */
            /******************************************************************/
            parent->handle().postEvent( event.eventId(),
                                        event.parameter1(),
                                        event.parameter2() );
            stopProcessingEvent = true;
          }
        }
      }
      break;
#endif // IC_PM

#ifdef IC_WIN
    case WM_HELP:
      {
        /**********************************************************************/
        /* Get a pointer to the notebook object.                              */
        /**********************************************************************/
        INotebook* pNotebk = (INotebook *)event.controlWindow();
        IWindow* parent = event.window()->parent();
        HELPINFO helpInfo=*((LPHELPINFO)(void *)(event.parameter2()));

        /**********************************************************************/
        /* If focus is on a tab control's tab, (native and/or ccl)            */
        /*   then post BKN_HELP (processed by IPageHandler::help().           */
        /**********************************************************************/
        if (( pNotebk ) &&
            ( event.handle()==helpInfo.hItemHandle ) )
        {
          /********************************************************************/
          /* Windows tab control processing.  Get the top page handle.        */
          /********************************************************************/
          IPageHandle topPageHandle( 0 );
          if (!pNotebk->isEmpty())
             topPageHandle = pNotebk->topPage();

          if (parent)
          {
            /******************************************************************/
            /* The user must have requested contextual help on a tab          */
            /* control's tab.  Simulate the tab help notification for the     */
            /* CUA '91 notebook control by posting the page help event        */
            /* to the parent of the notebook control.                         */
            /******************************************************************/
            parent->handle().postEvent(
                               WM_COMMAND,
                               IEventParameter1( (unsigned short)pNotebk->id(),
                                                 BKN_HELP ),
                               IEventParameter2( topPageHandle ) );

            stopProcessingEvent = true;
          }
        }
        else
        {
          if (parent)
          {
            /******************************************************************/
            /* The user must have requested contextual help from a            */
            /* page window.  We will avoid the notebook's processing          */
            /* of WM_HELP, which is geared for its tabs.  If the help         */
            /* request gets here, the page window must not be a frame         */
            /* (it is probably a canvas).  By routing the requests to         */
            /* the notebook's parent, these contextual helps can be           */
            /* resolved through the help subtable of the notebook's           */
            /* frame window.  Please note that we must send the help message  */
            /* rather than post it since parameter2 is a pointer to the       */
            /* HELPINFO structure.                                            */
            /******************************************************************/
            parent->handle().sendEvent( event.eventId(),
                                        event.parameter1(),
                                        event.parameter2() );

            stopProcessingEvent = true;
          }
        }
      } //WM_HELP
      break;
#endif // IC_WIN

    case WM_COMMAND:
    {
      // Route any unprocessed WM_COMMAND messages to the notebook's owner.
      // This allows handling of button events at the client/frame level.
      // However, if the notebook is the client window, we must only route
      // the command event to the frame window once since it will route it right
      // back to us if no handler processes it at the frame window level.

      INotebook *notebook = dynamic_cast<INotebook*>(event.dispatchingWindow());
      IWindow *pwndOwner = (notebook == 0) ? 0 : notebook->owner();
      if (pwndOwner &&
          pwndOwner->handle() != IWindow::desktopWindow()->handle() &&
          pwndOwner->handle() != IWindow::objectWindow()->handle())
      {
         ICommandEvent cmdEvent( event );
         bool isClient = (notebook->id() == FID_CLIENT) &&
                          pwndOwner->isFrameWindow();
         if (!isClient || !cmdEvent.isFromFrame())
         {
#ifdef IC_WIN
            unsigned long parm1 = event.parameter1();
            if (isClient)
            {
               // The notebook is the client, but the command event is not from
               // the frame.  Send it on to the frame window, but make sure we know
               // not to process it again.
               parm1 |= 0x80000000ul;
            }
            event.setResult( pwndOwner->sendEvent( event.eventId(),
                                                   IEventParameter1( parm1 ),
                                                   event.parameter2() ));
#endif // IC_WIN

#ifdef IC_PM
            unsigned long parm2 = event.parameter2();
            if (isClient)
            {
               // The notebook is the client, but the command event is not from
               // the frame.  Send it on to the frame window, but make sure we know
               // not to process it again.
               parm2 |= 0x8000ul;
            }
            event.setResult( pwndOwner->sendEvent( event.eventId(),
                                                   event.parameter1(),
                                                   parm2 ));
#endif // IC_PM
            stopProcessingEvent = true;
         }
      }
      break;
    }

#ifdef IC_WIN
    case WM_SIZE:
    case IC_UM_TCM_SIZE:
      {
        /**********************************************************************/
        /* Only process Windows tab control sizing event.                     */
        /**********************************************************************/
        IWindowHandle hwndControl( event.controlHandle() );
        if ( !(hwndControl &&
             ( IWindowClassName( hwndControl ) == WC_TABCONTROL )) )
          break;

        /**********************************************************************/
        /* Obtain pointer to notebook object.                                 */
        /**********************************************************************/
        INotebook*
          pNotebk = (INotebook*)IWindow::windowWithHandle( hwndControl );

        /**********************************************************************/
        /* If all tabs are to be shown (i.e. multi rows), we have to size the */
        /* tab control prior to calculating the display rectangle, to prevent */
        /* the application page window from clipping the bottom row of tabs.  */
        /**********************************************************************/
        ISize newSize( event.parameter2().lowNumber(),
                       event.parameter2().highNumber() );

        if ( pNotebk->areAllTabsVisible() )
        {
          if ( event.eventId() ==  WM_SIZE )
          {
            SetWindowPos( pNotebk->handle(),
                          NULL, 0, 0,
                          (int)newSize.width(),
                          (int)newSize.height(),
                          SWP_NOMOVE | SWP_NOZORDER );

            event.controlHandle().postEvent( IC_UM_TCM_SIZE,
                                             event.parameter1(),
                                             event.parameter2() );
          }
          else
          {
            pNotebk->tabControlResize( newSize );

            stopProcessingEvent = true;
          }
        }
        else
        {
          pNotebk->tabControlResize( newSize );
        }
      } //WM_SIZE & IC_UM_TCM_SIZE
      break;

    case WM_CHAR:
    {
      // Handle tabbing and mnemonics for the notebook.

      IKeyboardEvent keyEvent( event );
      if (keyEvent.isVirtual())
      {
         // Handle tabbing for the notebook.
         IKeyboardEvent::VirtualKey vkKey = keyEvent.virtualKey();

         if ((vkKey == IKeyboardEvent::tab ||
              vkKey == IKeyboardEvent::backTab) &&
             !keyEvent.isUpTransition())
            stopProcessingEvent = handleTab( keyEvent );
      }
      else if ((keyEvent.isCharacter() ||
                (keyEvent.isAltDown() && MapVirtualKey( keyEvent.parameter1(), 2 ))) &&
               !keyEvent.isCtrlDown())
      {
         // Handle mnemonic.
         IWindowHandle hwndControl( event.controlHandle() );
         if (!hwndControl ||
             (IWindowClassName( hwndControl ) != WC_TABCONTROL))
            break;

         INotebook
            *notebook = dynamic_cast<INotebook*>( event.dispatchingWindow() );
         if (!notebook)
            break;

         // Cursor through each of the tab pages to look for the mnemonic.
         int itemCount = TabCtrl_GetItemCount( hwndControl );
         char textBuffer[256];
         IOC_ITEM tabCtrlItem;
         tabCtrlItem.itemHeader.mask = TCIF_TEXT;
         tabCtrlItem.itemHeader.pszText = textBuffer;
         tabCtrlItem.itemHeader.cchTextMax = 256;
         for (int iPageIndex = 0; iPageIndex < itemCount; iPageIndex++)
         {
            if (!TabCtrl_GetItem( hwndControl, iPageIndex, &tabCtrlItem))
            {
               ITHROWGUIERROR( "TabCtrl_GetItem" );
            }

            // Now that we have the tab text, look for mnemonic.
            IString tabText( textBuffer );
            
            // Find first non-escaped & character.
            unsigned beginPos = tabText.indexOf('&');
            while ((beginPos != 0) &&
                   (tabText.indexOf( '&', beginPos+1 ) == beginPos+1))
               beginPos = tabText.indexOf('&', beginPos + 2);

            if (beginPos)
            {
               IString target = IString( "&" ) + IString((char)keyEvent.character() ) ;

               // See if there is a match
               if ((tabText.indexOf( target.upperCase(), beginPos ) == beginPos) ||
                   (tabText.indexOf( target.lowerCase(), beginPos ) == beginPos))
               {
                  // Found a match.  Turn to the indexed page.
                  notebook->turnToPage( notebook->tabPageHandle( iPageIndex ));
                  stopProcessingEvent = true;
                  break;
               }
            }
         }
      }
      break;
    }

    case WM_SETFOCUS:
    {
      // If the notebook has received focus, check to determine whether
      // it was traversed to via a back tab.  If so, determine whether
      // the focus was previously on a descendant (ie, control in the page
      // window).  If so, then allow the notebook to receive focus.  If not,
      // set the focus on the page window if possible.
      
      unsigned long ulKeyState =
         IGETKEYSTATE( IWindow::desktopWindow()->handle(), VK_TAB ) &
         IGETKEYSTATE( IWindow::desktopWindow()->handle(), VK_SHIFT );
    
      if (ulKeyState & 0x8000)
      {
         // Set focus on the page window if it is tabbable.
         if (!fFocusFromPage)
            stopProcessingEvent = gotFocus( event );
         else
            fFocusFromPage = false;
      }
      break;
    }

    case WM_SYSKEYDOWN:
      {
        /**********************************************************************/
        /* Only process Windows tab control keyboard events of interest.      */
        /**********************************************************************/
        IWindowHandle hwndControl( event.controlHandle() );
        if ( !(hwndControl &&
             ( IWindowClassName( hwndControl ) == WC_TABCONTROL )) )
          break;

        /**********************************************************************/
        /* Process key events for the Windows tab control.                    */
        /**********************************************************************/
        IKeyboardEvent keyevt( event );
        if ( !keyevt.isVirtual() || !keyevt.isAltDown() ||
             keyevt.isShiftDown() || keyevt.isCtrlDown() )
        {
          break;
        }

        /**********************************************************************/
        /* Obtain pointer to notebook object.                                 */
        /**********************************************************************/
        INotebook*
          pNotebk = (INotebook*)IWindow::windowWithHandle( hwndControl );

        /**********************************************************************/
        /* Alt+Down:  set the focus to the application's page window          */
        /**********************************************************************/
        if ( keyevt.virtualKey() == IKeyboardEvent::down )
        {
          /********************************************************************/
          /* Get the top page handle if it exists.                            */
          /********************************************************************/
          IPageHandle topPageHandle = pNotebk->topPage();
          if (topPageHandle)
          {
            /******************************************************************/
            /* Get the application page window if it exists and set the       */
            /* focus to it.                                                   */
            /******************************************************************/
            IWindow* pageWindow = pNotebk->window( topPageHandle );
            if (pageWindow)
            {
              pageWindow->setFocus();
              stopProcessingEvent = true;
            }
          }
        }
        /**********************************************************************/
        /* Alt+PageUp:  bring the previous page to the top                    */
        /**********************************************************************/
        else if ( keyevt.virtualKey() == IKeyboardEvent::pageUp )
        {
          IPageHandle topPageHandle = pNotebk->topPage();
          if (topPageHandle)
          {
            IPageHandle prevPageHandle = pNotebk->previousPage( topPageHandle );
            if (prevPageHandle)
            {
              pNotebk->turnToPage( prevPageHandle );
              stopProcessingEvent = true;
            }
          }
        }
        /**********************************************************************/
        /* Alt+PageDown:  bring the next page to the top                      */
        /**********************************************************************/
        else if ( keyevt.virtualKey() == IKeyboardEvent::pageDown )
        {
          IPageHandle topPageHandle = pNotebk->topPage();
          if (topPageHandle)
          {
            IPageHandle nextPageHandle = pNotebk->nextPage( topPageHandle );
            if (nextPageHandle)
            {
              pNotebk->turnToPage( nextPageHandle );
              stopProcessingEvent = true;
            }
          }
        }
      } //WM_SYSKEYDOWN
      break;

    case WM_NOTIFY:
      {
        /**********************************************************************/
        /* Process the Windows tab ctrl's notifications for selection events  */
        /**********************************************************************/
        LPNMHDR lpnmhdr = (LPNMHDR)event.parameter2().asUnsignedLong();

        IWindowHandle hwndControl( lpnmhdr->hwndFrom );

        /**********************************************************************/
        /* Verify that we are processing a Windows tab control                */
        /**********************************************************************/
        if ( hwndControl  &&
             (IWindowClassName( hwndControl ) == WC_TABCONTROL) )
        {
          if ( (lpnmhdr->code == TCN_SELCHANGE) ||
               (lpnmhdr->code == TCN_SELCHANGING) )
          {
            INotebook*
               pNotebk = (INotebook*)IWindow::windowWithHandle( hwndControl );
   
            /******************************************************************/
            /* Fill in the PAGESELECTNOTIFY structure that we use in Windows  */
            /******************************************************************/
            PAGESELECTNOTIFY pageSelectNotify;
            pageSelectNotify.hwndBook = hwndControl;

            if ( lpnmhdr->code == TCN_SELCHANGE )
            {
              /****************************************************************/
              /* Use value for previously selected page that we stored during */
              /* processing of TCN_SELCHANGING.                               */
              /****************************************************************/
              pageSelectNotify.ulPageIdNew =
                   (DWORD)pNotebk->topPage().asUnsigned();
              pageSelectNotify.ulPageIdCur = (DWORD)ulCurrentPageId;

              /****************************************************************/
              /* The tab selection has already occurred and the new tab is    */
              /* now the top page in the tab control.  Do our required        */
              /* processing and issue the selection notification.             */
              /****************************************************************/
              stopProcessingEvent =
                pNotebk->processTabSelect( &pageSelectNotify, false );
            }
            else
            {
              /****************************************************************/
              /* The tab selection for a new tab is pending.                  */
              /* Determine the newly selected tab.  We have to do hit testing */
              /* with the cursor position.                                    */
              /****************************************************************/
              // Get the cursor position when this message received.
              long msgCursorPos = GetMessagePos();

              // The cursor position is in screen coordinates; translate to client
              // coordinates.
              RECT clientRect;
              if (!GetWindowRect( hwndControl, &clientRect ))
              {
                 ITHROWGUIERROR( "GetWindowRect" );
              }
              
              TC_HITTESTINFO hitTestInfo;
              hitTestInfo.pt.x = LOWORD( msgCursorPos ) - clientRect.left;
              hitTestInfo.pt.y = HIWORD( msgCursorPos ) - clientRect.top;
              long newPage = TabCtrl_HitTest( hwndControl, &hitTestInfo );
              if (newPage == -1)
                 pageSelectNotify.ulPageIdNew = pNotebk->topPage().asUnsigned();
              else
                 pageSelectNotify.ulPageIdNew = pNotebk->tabPageHandle( newPage )
                                                .asUnsigned();
              pageSelectNotify.ulPageIdCur = (DWORD)pNotebk->topPage().asUnsigned();
              ulCurrentPageId = pageSelectNotify.ulPageIdCur;

              /****************************************************************/
              /* The tab selection is now in progress (pending).  Do our      */
              /* required processing and issue the selection pending          */
              /* notification.                                                */
              /****************************************************************/
              if ( pNotebk->processTabSelect( &pageSelectNotify, true ) )
                event.setResult( true );
            }
          }
          // If the owner of this tab control is another tab control, we can't
          // forward any of the notification messages to the owner tab
          // control.  Doing so may result in a trap in the common controls dll.
          // Since we don't need to send these notification messages
          // up the owner chain anyway (we don't look for them in any of our
          // handlers), just return true for all of the WM_NOTIFY messages.
          stopProcessingEvent = true;
        }
        break;
      } //WM_NOTIFY

    case WM_SYSCOMMAND:
      {
        /**********************************************************************/
        /* Process the Windows tab control's WM_SYSCOMMAND/SC_CLOSE event     */
        /* for Windows95 to work around a Windows95 bug.                      */
        /**********************************************************************/
        IWindowHandle hwndControl( event.controlHandle() );
        if ( !(hwndControl &&
             ( IWindowClassName( hwndControl ) == WC_TABCONTROL )  &&
	       IPlatform::isWin9x()) )
        {
          break;
        }

        if ( (((unsigned long)event.parameter1()) & 0xFFF0) == SC_CLOSE )
        {
          INotebook*
            pNotebk = (INotebook*)IWindow::windowWithHandle( hwndControl );

          /********************************************************************/
          /* Return true if a tab control to circumvent Windows95 bug.        */
          /********************************************************************/
          if ( pNotebk && !pNotebk->isPMCompatible() )
          {
            return( true );
          }
        }
      } //WM_SYSCOMMAND
      break;

    case WM_SETFONT:
      {
        /**********************************************************************/
        /* Process the Windows tab control's WM_SETFONT event to resize the   */
        /* clipping window properly and not clip tabs                         */
        /**********************************************************************/
        IWindowHandle hwndControl( event.controlHandle() );
        if ( hwndControl && (IWindowClassName(hwndControl) == WC_TABCONTROL) )
        {
          // Get the notebook object
          INotebook* pnb = (INotebook*)IWindow::windowWithHandle(hwndControl);

          // Allow the native control to process the message
          this->defaultProcedure( event );

          // Event handled so don't route on.
          event.setResult( true );

          // If we found the object
          if ( pnb )
          {
            // Resize the window based on the new page window size
            IWindow* clip = pnb->pageClippingWindow();
            if (clip)
              clip->sizeTo( pnb->pageSize() );
          }
          return( true );
        }
      } //WM_SETFONT
      break;

    case IC_UM_IS_AGGREGATE_CTRL:
      {
        event.setResult(true);
        stopProcessingEvent = true;
      }
      break;
#endif //IC_WIN

    /******************************************************************/
    /* Store the colors that are being set and then let this message  */
    /* pass on to the notebook control.                               */
    /******************************************************************/
    case BKM_SETNOTEBOOKCOLORS:
      ((INotebook*)event.window())
          ->storeNotebookColors( event.parameter1(), event.parameter2() );
      break;

    default:
      break;
  } //end switch

  return stopProcessingEvent;
}


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| INotebookHandler::handleTab                                                  |
|                                                                              |
| Process tabbing for the notebook.                                            |
------------------------------------------------------------------------------*/
bool INotebookHandler::handleTab ( IKeyboardEvent& keyEvent )
{
  bool stopProcessingEvent = false;

  IKeyboardEvent::VirtualKey vkKey = keyEvent.virtualKey();
  IWindowHandle hwndFocus = 
     IQUERYFOCUS(IWindow::desktopWindow()->handle());
  INotebook *notebook = (INotebook*)(keyEvent.dispatchingWindow());
  IWindowHandle hwndNotebook = notebook->handle();
  IPageHandle topPage = notebook->topPage();
  IWindow *pageWindow = (topPage) ? notebook->window( topPage ) : 0;

  // If focus is on the notebook itself, then set focus to the
  // page window if it is tabbable.
  if (hwndFocus == hwndNotebook &&
      pageWindow &&
      vkKey == IKeyboardEvent::tab)
  {
     if (pageWindow->isFrameWindow())
     {
        IWindowHandle
           hwndNextControl = GetNextDlgTabItem( pageWindow->handle(),
                                                NULL,
                                                FALSE );
        if (hwndNextControl)
        {
           pageWindow->postEvent( WM_NEXTDLGCTL,
                                  IEventParameter1( hwndNextControl ),
                                  IEventParameter2( true ));
           stopProcessingEvent = true;
        }
     }
     else if (pageWindow->isTabStop())
     {
        pageWindow->setFocus();
        stopProcessingEvent = true;
     }
  }
  else
  if (hwndFocus != hwndNotebook &&
      vkKey == IKeyboardEvent::backTab)
  {
     // The focus must be on a descendant of the page window.  Since we see
     // the event at this level, we must be tabbing out of the page window.
     // Set the focus to the notebook only if it is a back tab.  Also, set
     // flag to indicate that we are the one that set focus on the notebook.
     {
        fFocusFromPage = true;
        notebook->setFocus();
        stopProcessingEvent = true;
     }
  }
  return stopProcessingEvent;
}


/*------------------------------------------------------------------------------
| INotebookHandler::gotFocus                                                   |
|                                                                              |
| Handle a focus request for the notebook.  We only need to handle the focus   |
| request when the notebook has been traversed to via a back tab.              |
------------------------------------------------------------------------------*/
bool INotebookHandler::gotFocus ( IEvent& event )
{
  bool stopProcessingEvent = false;
  INotebook *notebook = (INotebook*)(event.dispatchingWindow());
  IWindowHandle hwndNotebook = notebook->handle();
  IPageHandle topPage = notebook->topPage();
  IWindow *pageWindow = (topPage) ? notebook->window( topPage ) : 0;

  // If the top page window is either a tabbable control or contains a tabbable
  // control, traverse into the page window.
  if (pageWindow)
  {
     if (pageWindow->isFrameWindow())
     {
        // Note:  There is a problem with the GetNextDlgTabItem returning
        //        NULL when the direction flag is TRUE (search previous)
        //        and the starting handle is NULL.  So we just check for
        //        the existence of any tabbable control in the dialog and
        //        let the frame window determine which control gets focus.
        IWindowHandle
           hwndNextControl = GetNextDlgTabItem( pageWindow->handle(),
                                                NULL,
                                                FALSE );
        if (hwndNextControl)
        {
           pageWindow->setFocus();
           stopProcessingEvent = true;
        }
     }
     else if (pageWindow->isTabStop())
     {
        pageWindow->setFocus();
        stopProcessingEvent = true;
     }
  }
  return stopProcessingEvent;
}
#endif // IC_WIN

