/*******************************************************************************
* FILE NAME: iresprv.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     IAixResourceImp                                                          *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Revision: 89  1.4.1.6  source/ui/baseapp/iresprv.cpp, resource, ioc.v400, 000302  

#ifdef IC_MOTIF
#include <iresprv.hpp>
#include <itrace.hpp>
#include <istring.hpp>
#include <ipoint.hpp>
#include <ixdc.hpp>


/*------------------------------------------------------------------------------
| IAixResourceImp::IAixResourceImp                                             |
------------------------------------------------------------------------------*/
IAixResourceImp::IAixResourceImp(const IResourceLibrary & resourceLibrary):
  fResourceLibrary(resourceLibrary),
  fResourceTypeCount(0),
  fResourceStart((res*)(resourceLibrary.resourceAddress())),
  fVersion(1)  // Note: We update the version in resourceByteAddress
{
   IFUNCTRACE_ALL();
   if (fResourceStart)
   {
     fResourceTypeCount = fResourceStart->resTable.restypecount;
   }
}

/*------------------------------------------------------------------------------
| IAixResourceImp::~IAixResourceImp                                            |
------------------------------------------------------------------------------*/
IAixResourceImp::~IAixResourceImp   ( )
{
}


/*------------------------------------------------------------------------------
| IAixResourceImp::resourceTypeAddress - Returns the address of the resource   |
|                          type table for a given resource type. (i.e.         |
|                          RT_STRING, RT_POINTER, etc...)                      |
------------------------------------------------------------------------------*/
restype* IAixResourceImp::resourceTypeAddress (unsigned short resType)
{
  IFUNCTRACE_ALL();
  bool found = false;
  restype *pResourceType = 0;

  /*------------------------------------------------------------------*/
  /* Get starting address of resource types                           */
  /*------------------------------------------------------------------*/
  if (fResourceStart)
  {
      pResourceType = (restype*) &(fResourceStart->resType);
  }

  if(pResourceType)
  {
    /********************************************************************/
    /* Loop through the resource type headers to find passed resType    */
    /********************************************************************/
    {
      short i;

      for( i=1; (!found) && (i <= fResourceTypeCount); i++)
      {
        if (pResourceType->restype == resType)
          found = true;
        else
          pResourceType = (restype*)((char*) pResourceType + sizeof(restype));
      }
    }
  }

  if (found)
    return pResourceType;
  else
  {
    ITRACE_DEVELOP(IString("Resource type not found:")+IString(resType));
    return 0;
  }
}

/*------------------------------------------------------------------------------
| IAixResourceImp::resourceOrdinalIdAddress - Returns the address of the          |
|                          ordinal id table for a given resource.              |
------------------------------------------------------------------------------*/
ordresid* IAixResourceImp::resourceOrdinalIdAddress (unsigned short resType,
                                              unsigned short resourceId)
{

  IFUNCTRACE_ALL();
  bool found = false;
  ordresid *pOrdResId = 0;
  restype *pResourceType = (restype*) resourceTypeAddress(resType);

  if (fResourceStart && pResourceType)
  {
      pOrdResId = (ordresid*) ((char *)fResourceStart + pResourceType->offset);
  }

  if (pOrdResId)
  {
    unsigned short ordIdCount = pResourceType->count - pResourceType->strcount;
    unsigned long resourceBundleId = resourceId;

    if (resType == RT_STRING || resType == RT_MESSAGE)
    {
      resourceBundleId = resourceId/16 + 1;
    }
    short i;

    for( i=1; (!found) && (i <= ordIdCount); i++)
    {
      if (pOrdResId->resourceBundleId == resourceBundleId)
        found = true;
      else
        pOrdResId = (ordresid*) ((char*)pOrdResId + sizeof(ordresid));
    }
  }

  if (found)
    return pOrdResId;
  else
    return 0;
}

/*------------------------------------------------------------------------------
| IAixResourceImp::resourceByteAddress - Returns the address of the               |
|                          resource data for a given resource id.              |
------------------------------------------------------------------------------*/
char *IAixResourceImp::resourceByteAddress (unsigned short resType,
                                            unsigned short resourceId)
{
  IFUNCTRACE_ALL();
  char *result=0;
  resources *pResource;

  /*------------------------------------------------------------------*/
  /* Get starting address of resource types                           */
  /*------------------------------------------------------------------*/
  ordresid *pOrdResId = resourceOrdinalIdAddress(resType, resourceId);

  if (fResourceStart && pOrdResId)
  {
    pResource =(resources*) ((char *)fResourceStart + pOrdResId->offset);
    
    // The version of irc is stored with each resource type.  
    // If the version number is 0xFF, this is a portapak GA version
    // of the resource compiler and we leave the version at the
    // default value of 1.
    if(pResource->restype != 0xFF)
       fVersion = pResource->restype;

    ITRACE_DEVELOP(IString("irc version is: ")+IString(fVersion));

    result = &(pResource->bytes);
  }

  return result;
}

/*------------------------------------------------------------------------------
| IAixResourceImp::resourceData - RT_STRING/RT_MESSAGE                         |
------------------------------------------------------------------------------*/
char *IAixResourceImp::resourceData(rtstring* pResource,
                                    unsigned short resourceId)
{
  IFUNCTRACE_ALL();

  char *result;
  stringdata *pStringData = (stringdata*) &(pResource->stringData);
  unsigned short resourceIdx = resourceId%16;
  short i;

  for (i=0; i <= resourceIdx; i++)
  {
    /**************************************************************/
    /* The actual address of the resource string in memory is     */
    /* returned. This will need to change as the design of how    */
    /* DBCS is handled is finalized.                              */
    /**************************************************************/
    result =  &(pStringData->data);

    /**************************************************************/
    /* Get next string                                            */
    /**************************************************************/
    pStringData = (stringdata*) ((char*)  pStringData + sizeof(stringdata) +
      IString(&(pStringData->data)).length());
  }
  return result;
}


/*------------------------------------------------------------------------------
| IAixResourceImp::resourceData - RT_BITMAP/RT_ICON                            |
------------------------------------------------------------------------------*/
void IAixResourceImp::resourceData(rtbitmap* pResource, char** xpmData )
{
  if(fVersion < 2)
  {
    xpmbitmaparray *xpmBa = (xpmbitmaparray*) &pResource->xpmba;
    *xpmData = (char*)xpmBa + sizeof(xpmbitmaparray);

    // Note: The following code is here to be backwards compatible with
    // the original icon types. In the old way, both the color image
    // and the XOR/AND mask were stored in the file with the mask first
    // and the color image second.  In the new way, there is just a
    // single image using XPMs mechanism to identify the transparent
    // color.
    // Note: We do not support the use of inverse color on AIX.
    switch(xpmBa->type)
    {
    case BFT_ICON :
    case BFT_COLORICON :
    case BFT_POINTER :
    case BFT_COLORPTR :

       if (xpmBa->offsetNext && xpmData!=0)
       {
         *xpmData = *xpmData + xpmBa->offsetNext;
       }
       break;
    }
  }
  else // Version 2 or later
  {
     xpmbitmaparray2 *xpmBa = (xpmbitmaparray2*) &pResource->xpmba;

     // By default we return the first image.  If there is more
     // than one image:
     //    1) For BFT_BITMAP, find the one that best matches the
     //       display depth.    
     //    2) For icon type, alway try to find a 32 by 32 image, color
     //       if depth > 1, black and white if depth 1
     if(xpmBa->offsetNext != 0)
     {
        xpmbitmaparray2* possibleImage = 0;
        xpmbitmaparray2* chosenImage = 0;

        Display* display = IXDisplay::display();
        unsigned long displayDepth = XDefaultDepth(display, XDefaultScreen(display));

        // More than one image, pick the best.
        while(xpmBa && !chosenImage)
        {
          switch(xpmBa->type)
          {
          case BFT_ICON :
          case BFT_POINTER :
          case BFT_COLORICON :
          case BFT_COLORPTR :
             if(xpmBa->width == 32 && xpmBa->height == 32)
             {

                // Always choose the image if 32x32 and depth matches
                if(xpmBa->bitCount == displayDepth)
                   chosenImage = xpmBa;
                else if(possibleImage)
                {
                   // The prior image was 32x32 but the depth didn't match exactly. 
                   // Is this one any better?
                   if(displayDepth > 1 && possibleImage->bitCount == 1)
                   {
                     // At least its a color image on a color display
                     possibleImage = xpmBa;
                   }
                }
                else
                {
                  // The depth doesn't match but it is a possibility because it is
                  // a 32x32 image.
                  possibleImage =  xpmBa;
                }
             }
             break;
          default: // BFT_BITMAP
             {
                // Always choose the image if the depth matches
                if(xpmBa->bitCount == displayDepth)
                   chosenImage = xpmBa;
                else if(xpmBa->bitCount >1 && displayDepth >1)
                {
                   // At least its color image on a color display.
                   possibleImage =  xpmBa;
                }
             }
             break;
          } // end switch

          ITRACE_DEVELOP(IString("  Type: ")+IString(xpmBa->type).d2x()+
                         IString("  BitPlanes: ")+IString(xpmBa->bitCount)+
                         IString("  Size: ")+IString(ISize(xpmBa->width, xpmBa->height).asString())+
                         IString("  Colors: ")+IString(xpmBa->colors));

          if(xpmBa->offsetNext)
            xpmBa = (xpmbitmaparray2*)((char*)xpmBa + xpmBa->offsetNext);
          else
            xpmBa = 0;
        }

        if(chosenImage)
           xpmBa = chosenImage;
        else if(possibleImage)
           xpmBa = possibleImage;
        else // The first (or only) image
           xpmBa = (xpmbitmaparray2*) &pResource->xpmba;

     }

     *xpmData = (char*)xpmBa + sizeof(xpmbitmaparray2);

     ITRACE_DEVELOP(IString("  Chosen Type: ")+IString(xpmBa->type).d2x()+
                    IString("  BitPlanes: ")+IString(xpmBa->bitCount)+
                    IString("  Size: ")+IString(ISize(xpmBa->width, xpmBa->height).asString())+
                    IString("  Colors: ")+IString(xpmBa->colors));
  }  // end else Version > 2
}

#endif
