// Revision: 05 1.7.1.4 source/ui/baseapp/iwindow3.cpp, basewin, ioc.v400, 000302 
/*******************************************************************************
* FILE NAME: iwindows3.cpp                                                     *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iwindow.hpp that are PM unique.                                         *
*                                                                              *
* COPYRIGHT:                                                                   *
*   Licensed Materials - Property of IBM                                       *
*   (C) Copyright IBM Corporation 1992, 1993                                   *
*   All Rights Reserved                                                        *
*   US Government Users Restricted Rights - Use, duplication, or               *
*   disclosure                                                                 *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/

#define INCL_WINERRORS
#define INCL_WINSYS
#define INCL_WINWINDOWMGR
extern "C" {
   #include <iwindefs.h>
}

#ifdef IC_PM // Ensure clean compile in WIN/AIX

#include <iwindow.hpp>
#include <iwinpriv.hpp>
#include <irect.hpp>
#include <icconst.h>
#include <iexcept.hpp>
#include <istring.hpp>
#include <ipoint.hpp>
#include <itrace.hpp>
#include <icolor.hpp>
#include <iinhratt.hpp>
#include <icoordsy.hpp>

// Segment definitions
#ifdef IC_PAGETUNE
  #define _IWINDOW3_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| IWindow::size                                                                |
|                                                                              |
| Get the window size.                                                         |
------------------------------------------------------------------------------*/
ISize IWindow::size() const
{
   SWP swp;
   WinQueryWindowPos(handle(), &swp);

   // no exception required because handle() check for valid
   // window handle and no bit was set.

   return (ISize(swp.cx,swp.cy));
}

/*------------------------------------------------------------------------------
| IWindow::parentSize                                                          |
------------------------------------------------------------------------------*/
ISize IWindow::parentSize ( ) const
{
   return parentSize( this->handle() );
}

/*------------------------------------------------------------------------------
| IWindow::parentSize                                                          |
------------------------------------------------------------------------------*/
ISize IWindow::parentSize ( const IWindowHandle& windowhandle )
{
   IMODTRACE_ALL("parentSize");
   IWindowHandle  parentHwnd = IPARENTOF( windowhandle );
   if ( parentHwnd == IWindowHandle(0) )
      {
      parentHwnd = IWindow::desktopWindow()->handle();
      }
   // the size needed is the size of the coordinate space which is used
   // to position child windows.
#ifdef   IC_WIN
   RECT rect;
   GetClientRect(parentHwnd, &rect);
   ITRACE_ALL( IString("parentSize=") +
               ISize(rect.right - rect.left, rect.bottom - rect.top ).asString() );
   return  ISize(rect.right - rect.left, rect.bottom - rect.top );
#endif   //IC_WIN
#ifdef   IC_PM
   SWP swp;
   WinQueryWindowPos(parentHwnd, &swp);
   ITRACE_ALL( IString("parentSize=") +
               ISize(swp.cx, swp.cy ).asString() );
   return (ISize(swp.cx,swp.cy));
#endif   //IC_PM
}

/*------------------------------------------------------------------------------
| IWindow::position                                                            |
|                                                                              |
| Gets the object position.                                                    |
------------------------------------------------------------------------------*/
IPoint IWindow::position() const
{
   return rect().minXMinY();
}

/*------------------------------------------------------------------------------
| IWindow::rect                                                                |
|                                                                              |
| Gets the object position and size.                                           |
------------------------------------------------------------------------------*/
IRectangle IWindow::rect() const
{
   if (ICoordinateSystem::isConversionNeeded() )
      {
      return ICoordinateSystem::convertToApplication( nativeRect( ),
                                                      this->parentSize( ) );
      }
   else
      {
      return nativeRect( );
      }
}

/*------------------------------------------------------------------------------
| IWindow::nativeRect                                                          |
------------------------------------------------------------------------------*/
IRectangle IWindow::nativeRect ( ) const
{
   SWP   swp;
   WinQueryWindowPos(handle(), &swp);

   // no exception required because handle() check for valid
   // window handle and no bit was set.

   return (IRectangle(swp.x, swp.y, swp.x + swp.cx, swp.y + swp.cy));
}

/*------------------------------------------------------------------------------
| IWindow::sizeTo                                                              |
|                                                                              |
| Size the window.                                                             |
------------------------------------------------------------------------------*/
IWindow& IWindow::sizeTo(const ISize& siz)
{
   IMODTRACE_ALL("Win::sizeTo");
   ITRACE_ALL(siz.asString());

   // If in non-native coordinates a resize will require adjustment
   // of position as well.
   if ( ICoordinateSystem::isConversionNeeded() )
      {
      moveSizeTo( IRectangle( position(), siz ) );
      }
   else
      {
      WinSetWindowPos(handle(), 0,
                      0,
                      0,
                      siz.width(),
                      siz.height(),
                      SWP_SIZE | SWP_NOADJUST);
      }

   // no exception required because handle() check for valid
   // window handle and SWP_SIZE is defined in toolkit.
   return *this;
}


/*------------------------------------------------------------------------------
| IWindow::moveTo                                                              |
|                                                                              |
| Move the object.                                                             |
------------------------------------------------------------------------------*/
IWindow& IWindow::moveTo(const IPoint& pt)
{
   IMODTRACE_ALL("Win::moveTo");
   ITRACE_ALL(pt.asString());
   if ( ICoordinateSystem::isConversionNeeded() )
      {
      moveSizeTo( IRectangle(pt, size()) );
      }
   else
      {
      WinSetWindowPos(handle(),0,
                      pt.x() ,
                      pt.y(),
                      0,
                      0,
                      SWP_MOVE | SWP_NOADJUST);
      }
   return *this;

   // no exception required because handle() check for valid
   // window handle and SWP_MOVE is defined in toolkit.
}

/*------------------------------------------------------------------------------
| IWindow::moveSizeTo                                                          |
|                                                                              |
| Move and size the object.                                                    |
------------------------------------------------------------------------------*/
IWindow& IWindow::moveSizeTo(const IRectangle& rect)
{

   IMODTRACE_ALL("Win::moveSizeTo");
   ITRACE_ALL(rect.asString());

   // recoordination ... translate to native coordinate system
   IRectangle nativeRectangle = rect;
   if ( ICoordinateSystem::isConversionNeeded() )
      {
      nativeRectangle = ICoordinateSystem::convertToNative( rect,
                                                            parentSize() );
      }

   WinSetWindowPos(handle(), 0,
                   nativeRectangle.minX(),
                   nativeRectangle.minY(),
                   nativeRectangle.width(),
                   nativeRectangle.height(),
                    SWP_MOVE | SWP_SIZE | SWP_NOADJUST);

   return *this;
   // no exception required because handle() check for valid
   // window handle and SWP_MOVE and SWP_SIZE are defined in toolkit.
}


/*------------------------------------------------------------------------------
| IWindow::setColor                                                            |
|                                                                              |
| Change the window attribute color using Red, Green and                       |
| Blue mix.                                                                    |
------------------------------------------------------------------------------*/
IWindow& IWindow::setColor(unsigned long ulColorArea, const IColor& color)
{
    if( ulColorArea == PP_FOREGROUNDCOLOR || ulColorArea == PP_BACKGROUNDCOLOR) 
    {
        // check for colour inheritance
        IWindow::ChildCursor childCursor(*this);

        for( childCursor.setToFirst(); childCursor.isValid(); childCursor.setToNext() )
        {
            IWindow*  child = childWindowAt(childCursor);
            if( child )
            {
                IWindowAttributeKeySet::Cursor cursor(child->pWindowData->fAttributeKeySet);
                for( cursor.setToFirst();cursor.isValid();cursor.setToNext() )
                {
                    IAttribute *attribute = (IAttribute *)cursor.element().attribute();
                    IInheritColorAttribute* inheritColorAtt = 
                            dynamic_cast<IInheritColorAttribute*>(attribute);      
                    if (inheritColorAtt)
                    {
                        // We only have to worry about setting inherited colours.
                        // PM will colour any owned windows when a new colour is set UNTIL the 'owned' 
                        // window has it's colour set explicity.  This is done in 
                        // IWindow::addOrReplaceAttribute so we don't have to worry about enforcing 
                        // non - inherited colours.

                        if( ulColorArea == PP_FOREGROUNDCOLOR && 
                            ((inheritColorAtt->inheritColor() 
                                == IInheritColorAttribute::kForegroundColor) ||
                            (inheritColorAtt->inheritColor() 
                                == IInheritColorAttribute::kForegroundAndBackgroundColor)))
                        {
                            child->setForegroundColor( color );
                        }
                        else if( ulColorArea == PP_BACKGROUNDCOLOR && 
                            ((inheritColorAtt->inheritColor() 
                                == IInheritColorAttribute::kBackgroundColor) ||
                            (inheritColorAtt->inheritColor() 
                                == IInheritColorAttribute::kForegroundAndBackgroundColor)))
                        {
                            child->setBackgroundColor( color );
                        }
 
                    }  // if ( inheritColorAtt )
                
                }  // for( cursor )
                    
            }  // if( child )
                
        }  // for( childCursor )
            
    }  // if( foreground or background )	
            
    // change colours for this window
    unsigned long ulClr=color.asRGBLong();
    WinSetPresParam (handle(), ulColorArea, sizeof(ulClr), &ulClr);
		
    return *this;
   // no exception b'cos handle is validated.
}

/*------------------------------------------------------------------------------
| IWindow::color                                                               |
|                                                                              |
| Return the color of window attribute.                                        |
------------------------------------------------------------------------------*/
IColor IWindow::color(unsigned long ulColorArea) const
{
   unsigned long ulClr=0;
   unsigned long ulRetLen = WinQueryPresParam (handle(), ulColorArea, 0, 0,
                             sizeof(ulClr), &ulClr, 0);

   if (ulRetLen == 0) {
      ITHROWLIBRARYERROR(IC_PRESPARAMNOTFOUND,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable);
   } /* endif */

   unsigned char blue  = (unsigned char)(ulClr);
   unsigned char green = (unsigned char)(ulClr >> 8);
   unsigned char red   = (unsigned char)(ulClr >> 16);
   return IColor(red, green, blue);
}

/*------------------------------------------------------------------------------
| IWindow::color                                                               |
|                                                                              |
| Return the color of window attribute.                                        |
------------------------------------------------------------------------------*/
IColor IWindow::color(unsigned long ulColorArea, const IColor& defaultColor) const
{
   unsigned long ulClr=0;
   unsigned long ulRetLen = WinQueryPresParam (handle(), ulColorArea, 0, 0,
                             sizeof(ulClr), &ulClr, 0);

   // if no presentation parameter has been set for this
   // area return the default color.
   if (ulRetLen == 0) {
      return defaultColor;
   } /* endif */

   unsigned char blue  = (unsigned char)(ulClr);
   unsigned char green = (unsigned char)(ulClr >> 8);
   unsigned char red   = (unsigned char)(ulClr >> 16);
   return IColor(red, green, blue);
}


/*------------------------------------------------------------------------------
| IWindow::positionBehindSiblings                                              |
|                                                                              |
| Put this window in the Z order behind all of its siblings.                   |
------------------------------------------------------------------------------*/
IWindow& IWindow::positionBehindSiblings ( )
{

   IMODTRACE_DEVELOP("Win::positionBehindSiblings");
   WinSetWindowPos(handle(), HWND_BOTTOM, 0,0,0,0, SWP_ZORDER);

   return *this;
}

/*------------------------------------------------------------------------------
| IWindow::positionOnSiblings                                                  |
|                                                                              |
| Put this window in the Z order on top of all of its siblings.                |
------------------------------------------------------------------------------*/
IWindow& IWindow::positionOnSiblings ( )
{

   IMODTRACE_DEVELOP("Win::positionBehindSiblings");
   WinSetWindowPos(handle(), HWND_TOP, 0,0,0,0, SWP_ZORDER);

   return *this;
}

/*------------------------------------------------------------------------------
| IWindow::positionBehind                                                      |
|                                                                              |
| Put this window in the Z order behind the passed window.                     |
------------------------------------------------------------------------------*/
IWindow& IWindow::positionBehindSibling ( const IWindowHandle& siblingWindow)
{

   IMODTRACE_DEVELOP("Win::positionBehindSiblings");
   WinSetWindowPos(handle(), siblingWindow, 0,0,0,0,
                   SWP_ZORDER);

   return *this;
}


/*------------------------------------------------------------------------------
| IWindow::isShowing                                                           |
|                                                                              |
| Is the window showing?  True if the window not being moved                   |
| and some part of the window is currently on the screen.                      |
------------------------------------------------------------------------------*/
bool IWindow::isShowing() const
{
   return WinIsWindowShowing(handle());

   // no exception required because handle() check for valid
   // window handle
}

/*------------------------------------------------------------------------------
| IChildCursorData::IChildCursorData                                           |
|                                                                              |
------------------------------------------------------------------------------*/
IChildCursorData::IChildCursorData( IWindowHandle parentHandle )
    : hwndParent( parentHandle ),
      enumHandle( 0 )
{
}

/*------------------------------------------------------------------------------
| IChildCursorData::~IChildCursorData                                          |
|                                                                              |
------------------------------------------------------------------------------*/
IChildCursorData::~IChildCursorData()
{
  if ( this->enumHandle )
  {
    if ( !WinEndEnumWindows( this->enumHandle ) )
    {
      ITHROWGUIERROR("WinEndEnumWindows");
    }
  }
}

/*------------------------------------------------------------------------------
| IChildCursorData::setToFirst                                                |
|                                                                              |
------------------------------------------------------------------------------*/
bool IChildCursorData::setToFirst( IWindowHandle &hwnd )
{
  if ( enumHandle )
  {
    // Enum in progress, check to see if we're at end...
    if ( !hwnd )
      // At end, simply call setToNext() to loop back to start...
      setToNext(hwnd);
    else
      // In middle somewhere.  end this enum...
      invalidate( hwnd );
  }
  // Start enum...
  enumHandle = WinBeginEnumWindows( hwndParent );
  if ( enumHandle)
    // If ok, setToNext() to get first child...
    return setToNext(hwnd);

  // Error, throw exception...
  ITHROWGUIERROR("WinBeginEnumWindows");

  return false;
}

/*------------------------------------------------------------------------------
| IChildCursorData::setToNext                                                 |
|                                                                              |
------------------------------------------------------------------------------*/
bool IChildCursorData::setToNext(IWindowHandle &hwnd)
{
  bool bSuccess(false);
  // Check that enum was begun...
  if ( this->enumHandle )
  {
    // It was, try to get next window...
    hwnd = WinGetNextWindow( this->enumHandle );
    if ( hwnd)
      // Got it, return true...
      bSuccess=true;
  }
  return bSuccess;
}

/*------------------------------------------------------------------------------
| IChildCursorData::invalidate                                                 |
|                                                                              |
------------------------------------------------------------------------------*/
void IChildCursorData::invalidate(IWindowHandle &hwnd)
{
  // See if enum in progress...
  if ( this->enumHandle )
  {
    // Yes, terminate it...
    if ( WinEndEnumWindows( enumHandle ) )
    {
      enumHandle = 0;
      hwnd = 0;
    }
    else
    {
      // Error, throw exception...
      ITHROWGUIERROR("WinEndEnumWindows");
    }
  }
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::ChildCursor                                            |
|                                                                              |
| The constructor simply saves the parent handle and initializes the cursor    |
| state to "invalid" (enumHandle = windowHandle == 0).                         |
------------------------------------------------------------------------------*/
IWindow::ChildCursor::ChildCursor ( IWindow &parent, bool onlyIWindowChildren )
    : hwnd( 0 )
    , pCursorData( new IChildCursorData( parent.handle() ))
{
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::~ChildCursor                                           |
|                                                                              |
| If the cursor enum handle is valid, issue WinEndEnumWindows.                 |
------------------------------------------------------------------------------*/
IWindow::ChildCursor::~ChildCursor ( )
{
  delete pCursorData; //For when IChildCursorData is defined.
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::setToFirst                                             |
|                                                                              |
| If the enumeration has started and we're not at end, reset to beginning.     |
| If we are at end, then advance (WinGetNextWindow will loop).  Save the       |
| first child window's handle in hwnd data member.                             |
------------------------------------------------------------------------------*/
bool IWindow::ChildCursor::setToFirst ( )
{
   return pCursorData->setToFirst( hwnd );
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::setToNext                                              |
|                                                                              |
| Set the saved window handle to the value returned by WinGetNextWindow.       |
------------------------------------------------------------------------------*/
bool IWindow::ChildCursor::setToNext ( )
{
   return pCursorData->setToNext( hwnd );
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::isValid                                                |
|                                                                              |
| Return true if the saved window handle isn't null.                           |
------------------------------------------------------------------------------*/
bool IWindow::ChildCursor::isValid ( ) const
{
  return hwnd != 0;
}

/*------------------------------------------------------------------------------
| IWindow::ChildCursor::invalidate                                             |
|                                                                              |
| If enum handle is valid, end the enumeration.                                |
------------------------------------------------------------------------------*/
void IWindow::ChildCursor::invalidate ( )
{
  pCursorData->invalidate( hwnd );
}

/*------------------------------------------------------------------------------
| IWindow::foregroundColor                                                     |
|                                                                              |
| Returns the foreground color of the window.                                  |
------------------------------------------------------------------------------*/
IColor IWindow::foregroundColor () const
{
  return (color(PP_FOREGROUNDCOLOR, IGUIColor(IGUIColor::windowText)));
}

/*------------------------------------------------------------------------------
| IWindow::backgroundColor                                                     |
|                                                                              |
| Returns the background color of the window.                                  |
------------------------------------------------------------------------------*/
IColor IWindow::backgroundColor () const
{
  return (color(PP_BACKGROUNDCOLOR, IGUIColor(IGUIColor::windowBgnd)));
}

/*------------------------------------------------------------------------------
| IWindow::borderColor                                                         |
|                                                                              |
| Returns the border color of the window.                                      |
------------------------------------------------------------------------------*/
IColor IWindow::borderColor () const
{
  return (color(PP_BORDERCOLOR, IGUIColor(IGUIColor::frameBorder)));
}

/*------------------------------------------------------------------------------
| IWindow::shadowColor                                                         |
|                                                                              |
| Returns the shadow color of the window.                                      |
------------------------------------------------------------------------------*/
IColor IWindow::shadowColor () const
{
  return (color(PP_SHADOW, IGUIColor(IGUIColor::dialogShadow)));
}

/*------------------------------------------------------------------------------
| IWindow::disabledForegroundColor                                             |
|                                                                              |
| Returns the disabled foreground color of the window.                         |
------------------------------------------------------------------------------*/
IColor IWindow::disabledForegroundColor () const
{
  return (color(PP_DISABLEDFOREGROUNDCOLOR,
                IGUIColor(IGUIColor::defaultControl)));
}

/*------------------------------------------------------------------------------
| IWindow::disabledBackgroundColor                                             |
|                                                                              |
| Returns the disabled background color of the window.                         |
------------------------------------------------------------------------------*/
IColor IWindow::disabledBackgroundColor () const
{
  return (color(PP_DISABLEDBACKGROUNDCOLOR,
                IGUIColor(IGUIColor::windowBgnd)));
}

/*------------------------------------------------------------------------------
| IWindow::hilitedForegroundColor                                              |
|                                                                              |
| Returns the hilite foreground color of the window.                           |
------------------------------------------------------------------------------*/
IColor IWindow::hiliteForegroundColor () const
{
  return (color(PP_HILITEFOREGROUNDCOLOR,
                IGUIColor(IGUIColor::hiliteFgnd)));
}

/*------------------------------------------------------------------------------
| IWindow::hilitedBackgroundColor                                              |
|                                                                              |
| Returns the hilite background color of the window.                           |
------------------------------------------------------------------------------*/
IColor IWindow::hiliteBackgroundColor () const
{
  return (color(PP_HILITEBACKGROUNDCOLOR,
                IGUIColor(IGUIColor::hiliteBgnd)));
}

/*------------------------------------------------------------------------------
| IWindow::activeColor                                                         |
|                                                                              |
| Returns the active color of the window.                                      |
------------------------------------------------------------------------------*/
IColor IWindow::activeColor () const
{
  return (color(PP_ACTIVECOLOR,
                IGUIColor(IGUIColor::activeFrameBorder)));
}

/*------------------------------------------------------------------------------
| IWindow::inactiveColor                                                       |
|                                                                              |
| Returns the inactive color of the window.                                    |
------------------------------------------------------------------------------*/
IColor IWindow::inactiveColor () const
{
  return (color(PP_INACTIVECOLOR,
                IGUIColor(IGUIColor::inactiveFrameBorder)));
}

/*------------------------------------------------------------------------------
| IWindow::setForegroundColor                                                  |
|                                                                              |
| Sets the foreground color of the window.                                     |
------------------------------------------------------------------------------*/
IWindow& IWindow::setForegroundColor (const IColor &color)
{
  return (setColor(PP_FOREGROUNDCOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::setBackgroundColor                                                  |
|                                                                              |
| Sets the background color of the window.                                     |
------------------------------------------------------------------------------*/
IWindow& IWindow::setBackgroundColor (const IColor &color)
{
  return (setColor(PP_BACKGROUNDCOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::setBorderColor                                                      |
|                                                                              |
| Sets the border color of the window.                                         |
------------------------------------------------------------------------------*/
IWindow& IWindow::setBorderColor (const IColor &color)
{
  return (setColor(PP_BORDERCOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::setShadowColor                                                      |
|                                                                              |
| Sets the shadow color of the window.                                         |
------------------------------------------------------------------------------*/
IWindow& IWindow::setShadowColor (const IColor &color)
{
  return (setColor(PP_SHADOW, color));
}

/*------------------------------------------------------------------------------
| IWindow::setDisabledForegroundColor                                          |
|                                                                              |
| Sets the disabled foreground color of the window.                            |
------------------------------------------------------------------------------*/
IWindow& IWindow::setDisabledForegroundColor (const IColor &color)
{
  return (setColor(PP_DISABLEDFOREGROUNDCOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::setDisabledBackgroundColor                                          |
|                                                                              |
| Sets the disabled background color of the window.                            |
------------------------------------------------------------------------------*/
IWindow& IWindow::setDisabledBackgroundColor (const IColor &color)
{
  return (setColor(PP_DISABLEDBACKGROUNDCOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::setHiliteForegroundColor                                            |
|                                                                              |
| Sets the hilite foreground color of the window.                              |
------------------------------------------------------------------------------*/
IWindow& IWindow::setHiliteForegroundColor (const IColor &color)
{
  return (setColor(PP_HILITEFOREGROUNDCOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::setHiliteBackgroundColor                                            |
|                                                                              |
| Sets the hilite background color of the window.                              |
------------------------------------------------------------------------------*/
IWindow& IWindow::setHiliteBackgroundColor (const IColor &color)
{
  return (setColor(PP_HILITEBACKGROUNDCOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::setActiveColor                                                      |
|                                                                              |
| Sets the active color of the window.                                         |
------------------------------------------------------------------------------*/
IWindow& IWindow::setActiveColor (const IColor &color)
{
  return (setColor(PP_ACTIVECOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::setInactiveColor                                                    |
|                                                                              |
| Sets the inactive color of the window.                                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::setInactiveColor (const IColor &color)
{
  return (setColor(PP_INACTIVECOLOR, color));
}

/*------------------------------------------------------------------------------
| IWindow::resetForegroundColor                                                |
|                                                                              |
| Resets the foreground color by undoing a previous set.                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetForegroundColor ()
{
  return (resetColor(PP_FOREGROUNDCOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetBackgroundColor                                                |
|                                                                              |
| Resets the background color by undoing a previous set.                       |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetBackgroundColor ()
{
  return (resetColor(PP_BACKGROUNDCOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetBorderColor                                                    |
|                                                                              |
| Resets the border color by undoing a previous set.                           |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetBorderColor ()
{
  return (resetColor(PP_BORDERCOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetShadowColor                                                    |
|                                                                              |
| Resets the shadow color by undoing a previous set.                           |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetShadowColor ()
{
  return (resetColor(PP_SHADOW));
}

/*------------------------------------------------------------------------------
| IWindow::resetDisabledForegroundColor                                        |
|                                                                              |
| Resets the disabled foreground color by undoing a previous set.              |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetDisabledForegroundColor ()
{
  return (resetColor(PP_DISABLEDFOREGROUNDCOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetDisabledBackgroundColor                                        |
|                                                                              |
| Resets the disabled background color by undoing a previous set.              |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetDisabledBackgroundColor ()
{
  return (resetColor(PP_DISABLEDBACKGROUNDCOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetHiliteForegroundColor                                          |
|                                                                              |
| Resets the hilite foreground color by undoing a previous set.                |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetHiliteForegroundColor ()
{
  return (resetColor(PP_HILITEFOREGROUNDCOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetHiliteBackgroundColor                                          |
|                                                                              |
| Resets the hilite background color by undoing a previous set.                |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetHiliteBackgroundColor ()
{
  return (resetColor(PP_HILITEBACKGROUNDCOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetActiveColor                                                    |
|                                                                              |
| Resets the active color by undoing a previous set.                           |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetActiveColor ()
{
  return (resetColor(PP_ACTIVECOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetInactiveColor                                                  |
|                                                                              |
| Resets the inactive color by undoing a previous set.                         |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetInactiveColor ()
{
  return (resetColor(PP_INACTIVECOLOR));
}

/*------------------------------------------------------------------------------
| IWindow::resetColor                                                          |
|                                                                              |
| Worker function to remove presentation parameters from windows.              |
------------------------------------------------------------------------------*/
IWindow& IWindow::resetColor ( unsigned long colorArea )
{
  IGUIErrorInfo dummy( "" );      // Clear out last GUI error.
  if ( ! WinRemovePresParam( this->handle(), colorArea ))
  {
     IGUIErrorInfo colorError( "WinRemovePresParam" );
     if ( colorError.errorId() )
     {        // Only throw exception if real error.
        colorError.throwError( IEXCEPTION_LOCATION() );
     }
  }
  return *this;
}
#endif // IC_PM
