// Revision: 23 1.3 source/ui/baseapp/win/ibrshmgr.cpp, basewin, ioc.v400, 000302 
/*******************************************************************************
* FILE NAME: ibrshmgr.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in ibrshmgr.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
// Priority INT_MIN (-2147483647 - 1) + 1024 + 512
#pragma priority( -2147482112 )

#include <ibrshmgr.hpp>
#include <icolor.hpp>

/*------------------------------------------------------------------------------
| Brush Element class.                                                         |
------------------------------------------------------------------------------*/

/*------------------------------------------------------------------------------
| IBrushElement::IBrushElement                                                         |
|                                                                              |
| Default constructor.                                                         |
------------------------------------------------------------------------------*/
IBrushElement::IBrushElement( ) 
   : lColor(0)
   , hBrush(0)
   , lUseCount(0)
{ 
   // default object only.
}

/*------------------------------------------------------------------------------
| IBrushElement::IBrushElement                                                         |
|                                                                              |
| Normal constructor. Used by IBrushManager for creating a brush.               |
------------------------------------------------------------------------------*/
IBrushElement::IBrushElement( long lClr ) 
   : lColor(lClr)
   , hBrush(0)
   , lUseCount(0) 
{
  IFUNCTRACE_ALL();

  // NOTE: lClr assumed to be an RGB value, adjusted via NEARESTPALETTECOLOR()
  hBrush = CreateSolidBrush( lColor );

  // The use count is 0 when the IBrushElement is called. Must call useBrush after
  // creating it to set the count to 1.
}

/*------------------------------------------------------------------------------
| IBrushElement::~IBrushElement                                                        |
|                                                                              |
| Destructor.  Cleans up the resource object.                                  |
| Should only be invoked as a result of a freeBrush() method call.             |
------------------------------------------------------------------------------*/
IBrushElement::~IBrushElement( )
{
  IFUNCTRACE_ALL();
  // delete the object because no one is using it anymore.
  if ( hBrush )
  {
    DeleteObject( hBrush );
  }
}

/*------------------------------------------------------------------------------
| IBrushElement::useBrush                                                          |
|                                                                              |
| Increase use count, tracking references to this brush resource.              |
------------------------------------------------------------------------------*/
long IBrushElement::useBrush( )
{
  IFUNCTRACE_ALL();
  lUseCount++;
  return lUseCount;
}

/*------------------------------------------------------------------------------
| IBrushElement::freeBrush                                                         |
|                                                                              |
| Decrease use count, tracking references to this brush resource.              |
------------------------------------------------------------------------------*/
long IBrushElement::freeBrush( )
{
  IFUNCTRACE_ALL();
  lUseCount--;
 
  // if no one is referring to the brush, clean it up.
  if ( lUseCount <= 0 )
  {
    delete this;
  }
  return lUseCount;
}

/*------------------------------------------------------------------------------
| IBrushElement::useCount                                                         |
|                                                                              |
| Returns use count                                                            |
------------------------------------------------------------------------------*/
long IBrushElement::useCount( )
{
  IFUNCTRACE_ALL();
  return lUseCount;
}

/*------------------------------------------------------------------------------
| IBrushElement::operator==                                                        |
|                                                                              |
| Equality between IBrushElements.                                                 |
------------------------------------------------------------------------------*/
bool IBrushElement::operator==(IBrushElement& br)
{
   return( hBrush == br.hBrush );
}

/*------------------------------------------------------------------------------
| IBrushElement::handle                                                           |
|                                                                              |
| Returns the brush elements handle                                            |
------------------------------------------------------------------------------*/
HBRUSH IBrushElement::handle() const 
{ 
  return hBrush; 
}

/*------------------------------------------------------------------------------
| IBrushElement::color                                                            |
|                                                                              |
| Returns the brush elements color                                             |
------------------------------------------------------------------------------*/
long IBrushElement::color() const 
{ 
  return lColor; 
}

const long& key(const IBrushElement* k)
{ 
  return k->lColor; 
}

/*------------------------------------------------------------------------------
| Brush Manager class. (list of Brush Elements)                                |
------------------------------------------------------------------------------*/

// Initialize static obects.
//
IBrushManager::BrushList IBrushManager::brushList;
long IBrushManager::lBrushCount = 0;

/*------------------------------------------------------------------------------
| IBrushManager::findBrush (public)                                                |
|                                                                              |
| Return a brush handle for the specified color                                |
------------------------------------------------------------------------------*/
HBRUSH IBrushManager::findBrush( const IColor& color )
{
  IFUNCTRACE_ALL();

   // convert and store via NEARESTPALETTECOLOR, so all 
   // close colors will be using the same brush handle.
  long lRealColor = NEARESTPALETTECOLOR(color.asRGBLong() );

  IBrushElement* pElem(0);
  pElem = brushElement( IColor(lRealColor) );

  if ( !pElem ) 
  {
    // new entry required...
    // create, add, set useCount to 1, and return handle.
    pElem = new IBrushElement( lRealColor );
    brushList.add( pElem );
    lBrushCount++;
  }


  return( pElem->handle() );
}

/*------------------------------------------------------------------------------
| IBrushManager::useBrush (public)                                             |
|                                                                              |
| Increment the use count of the brush.                                        |
------------------------------------------------------------------------------*/
long IBrushManager::useBrush( HBRUSH hBr )
{
  IFUNCTRACE_ALL();
  IBrushElement* pElem(0);
  pElem = brushElement( hBr );
  long lCount(0);

  if (pElem)
  {
    // call brush elements useBrush to bump the use count.
    lCount = pElem->useBrush();
  }
  return lCount;
}

/*------------------------------------------------------------------------------
| IBrushElement::freeBrush (public)                                                |
|                                                                              |
| Locate the specified brush handle, and release this use of it.               |
------------------------------------------------------------------------------*/
void IBrushManager::freeBrush( HBRUSH hBr )
{
  IFUNCTRACE_ALL();
  long lUseCount(0);
  IBrushElement* pElem(0);

  // don't bother if there are no brushes to free.
  if (lBrushCount > 0 )
  {
    pElem = brushElement(hBr);
    if ( pElem)
    {
      // save cursor location, in case elem must be deleted
      bool bFound(false);
      long myKey = brushList.key(pElem);
      BrushList::Cursor myCursor(brushList);
      bFound = brushList.locateElementWithKey(myKey, myCursor);

      // decrease access count on brush elem (may delete the elem)
      lUseCount = pElem->freeBrush();

      // remove from list, if no longer used
      if ( !lUseCount && bFound ) 
      {
        brushList.removeAt(myCursor);
        lBrushCount--;
      }
    }
  }
}

/*------------------------------------------------------------------------------
| IBrushElement::freeBrush (public)                                                |
|                                                                              |
| Locate the specified brush handle, and release this use of it.               |
------------------------------------------------------------------------------*/
void IBrushManager::freeBrush( const IColor& lClr )
{
  IFUNCTRACE_ALL();

  long lUseCount(0);
  IBrushElement* pElem(0);

  pElem = brushElement(lClr);
  if ( pElem)
  {
    // save cursor location, in case elem must be deleted
    bool bFound;
    long myKey = brushList.key(pElem);
    BrushList::Cursor myCursor(brushList);
    bFound = brushList.locateElementWithKey(myKey, myCursor);

    // decrease access count on brush elem (may delete the elem)
    lUseCount = pElem->freeBrush();

    // remove from list, if no longer used
    if ( !lUseCount && bFound ) 
    {
      brushList.removeAt(myCursor);
      lBrushCount--;
    }
  }
}


/*------------------------------------------------------------------------------
| IBrushElement::brushCount (public)                                              |
|                                                                              |
| Returns the number of brushes currently maintained by the brush manager.     |
------------------------------------------------------------------------------*/
long IBrushManager::brushCount( ) 
{ 
  return lBrushCount;
}

/*------------------------------------------------------------------------------
| IBrushElement::brushElement (protected)                                         |
|                                                                              |
| Locate the specified brush handle by color.                                  |
------------------------------------------------------------------------------*/
IBrushElement* IBrushManager::brushElement( const IColor& lClr )
{
  IFUNCTRACE_ALL();
  long lRealColor = NEARESTPALETTECOLOR(lClr.asRGBLong() );

  IBrushElement* pElem(0);
  BrushList::Cursor myCursor(brushList);
   
  bool bFound;
  bFound = brushList.locateElementWithKey(lRealColor, myCursor);
  if ( bFound ) 
  {
    pElem = brushList.elementAt(myCursor);
    if ( pElem )
      return pElem;
  }
  return 0;
}

/*------------------------------------------------------------------------------
| IBrushElement::brushElement (protected)                                         |
|                                                                              |
| Locate the specified brush handle by handle.                                 |
------------------------------------------------------------------------------*/
IBrushElement* IBrushManager::brushElement( HBRUSH hBr )
{
  IFUNCTRACE_ALL();
  IBrushElement* pElem(0);
  BrushList::Cursor myCursor(brushList);

  for (brushList.setToFirst(myCursor);
       myCursor.isValid();
       myCursor.setToNext() ) 
  {
    pElem = brushList.elementAt(myCursor);
    if ( pElem )
    {
      if (pElem->handle() == hBr )
        return pElem;
    }
  }
  return 0;
}
