// Revision: 68 1.11 source/ui/basectl/itabctl.cpp, notebook, ioc.v400, 000302 
/*******************************************************************************
* FILE NAME: itabctl.cpp                                                       *
*                                                                              *
* DESCRIPTION:                                                                 *
*   Implementation of the class(es):                                           *
*     ITabControl                                                              *
*     ITabControl::Cursor                                                      *
*                                                                              *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   Licensed Material - Property of IBM                                        *
*   (C) Copyright IBM Corp.  1997  All Rights Reserved.                        *
*                                                                              *
*******************************************************************************/
#pragma priority( -2147481524 )

extern "C" {
  #define INCL_WINSYS
  #define INCL_WINWINDOWMGR         // WinQueryWindowULong
  #define INCL_WINBUTTONS           // BS_NOTEBOOKBUTTON
  #define INCL_WINMESSAGEMGR        // For WM_PRESPARAMCHANGED.
  #define INCL_WINSTDBOOK
  #define INCL_WINDIALOGS           // Needed for PDLGTEMPLATE in BOOKPAGEINFO
  #include <iwindefs.h>
#ifdef IC_WIN
  #include <commctrl.h>
#endif // IC_WIN
}

#include <ibase.hpp>
#include <icconst.h>
#include <idefstyl.h>
#include <itabctl.hpp>
#include <itabpriv.hpp>
#include <iexcept.hpp>
#include <iframe.hpp>
#include <ihandler.hpp>
#include <inotifev.hpp>
#include <ipushbut.hpp>
#include <itrace.hpp>

#ifdef IC_WIN
#include <ifont.hpp>
#include <isetcv.hpp>
#include <iwcname.hpp>
#endif // IC_WIN

// Segment definitions.
#ifdef IC_PAGETUNE
  #define _ITABCTL_CPP_
  #include <ipagetun.h>
#endif

// Make private data pointer names less verbose.
#define tcd fTabControlData
#define tpd fTabPageData

#ifdef IC_WIN
extern unsigned long propagationInProgress;
#endif

/*------------------------------------------------------------------------------
| Public tab control styles.                                                   |
------------------------------------------------------------------------------*/
const ITabControl::Style
  ITabControl::tabsTop            ( 1, ITabControl__tabsTop ),
  ITabControl::tabsBottom         ( 1, ITabControl__tabsBottom ),
  ITabControl::buttons            ( 1, ITabControl__buttons ),
  ITabControl::border3D           ( 1, ITabControl__border3D ),
#ifdef IC_WIN
  ITabControl::classDefaultStyle  ( 1, ITabControl__classDefaultStyle );
#elif IC_PM
  ITabControl::classDefaultStyle  ( 2, ITabControl__classDefaultStyle );
#endif // IC_WIN

/*------------------------------------------------------------------------------
| Default style for new tab control objects (initial value).                   |
------------------------------------------------------------------------------*/
ITabControl::Style
#ifdef IC_WIN
  ITabControl::fCurrentDefaultStyle ( 1, ITabControl__classDefaultStyle );
#elif IC_PM
  ITabControl::fCurrentDefaultStyle ( 2, ITabControl__classDefaultStyle );
#endif // IC_WIN

/*------------------------------------------------------------------------------
| ITabControl::ITabControl                                                     |
|                                                                              |
| Construct a tab control from a window id, parent, and owner window handles,  |
| and initial size rectangle, and a style.                                     |
------------------------------------------------------------------------------*/
ITabControl :: ITabControl ( unsigned long     windowId,
                             IWindow*          parent,
                             IWindow*          owner,
                             const IRectangle& initial,
                             const Style&      style )
  : fTabControlData( new ITabControlData( this ))
{
  IMODTRACE_DEVELOP( "ITabControl::ctor1" );

  // assertions on input parms
  IASSERTPARM( parent != 0 )

  IWindowHandle hwndOwner = (owner == 0) ? IWindowHandle(0) : owner->handle();

  char* windowClass;
#ifdef IC_WIN
  InitCommonControls();
  windowClass = WC_TABCONTROL;
#elif IC_PM
  windowClass = WC_NOTEBOOK;
#endif // IC_WIN

  IWindowHandle hwndTabControl =
      this -> create( windowId,
                      0,
                      convertToGUIStyle( style ),
                      windowClass,
                      parent->handleForChildCreation(),
                      hwndOwner,
                      initial,
                      0,
                      0,
                      defaultOrdering(),
                      convertToGUIStyle( style, true ) );

  startHandlingEventsFor( hwndTabControl );

  // Create the tab page sequence collection.
  tcd->fPageSequence = new ITabControlPageSequence( );

#ifdef IC_WIN
  tcd->fPageKeyboardHandler = new ITabPageKeyboardHandler();

  // Create the page clipping window.
  tcd->fClipWindow = new ITabControlClipWindow( this );

  // Create the button area if buttons style is present.
  if (style & buttons)
  {
     tcd->fButtonCanvas = new ITabControlButtonCanvas( this );

     // Determine correct minimum size for the button canvas and cache the
     // minimum height.  We need to do this before it is shown, since this
     // height will affect the clipping window size.
     tcd->fDefButtonMargin = tcd->fButtonCanvas->margin();
     tcd->fButtonCanvasHeight = tcd->buttonCanvasHeight();
  }
#endif // IC_WIN

#ifdef IC_PM
  tcd->fTabPageResizeHandler = new ITabPageResizeHandler();
#endif // IC_PM

  // Add our private handler.
  tcd->fDefaultHandler = new ITabControlDefaultHandler();
  tcd->fDefaultHandler->handleEventsFor( this );
}


/*------------------------------------------------------------------------------
| ITabControl::ITabControl                                                     |
|                                                                              |
| Construct a tab control from a window id and the window handle which serves  |
| as both parent and owner.                                                    |
------------------------------------------------------------------------------*/
ITabControl :: ITabControl( unsigned long windowId,
                            IWindow*      parentAndOwner )
  : fTabControlData( new ITabControlData( this ))
{
  IMODTRACE_DEVELOP( "ITabControl::ctor2" );

  IASSERTPARM(parentAndOwner != 0)

  setAutoDestroyWindow( false );
  reserveUserWindowWord( false );
  startHandlingEventsFor( windowId, parentAndOwner );

  tcd->fDefaultHandler = new ITabControlDefaultHandler();
  tcd->fDefaultHandler->handleEventsFor( this );

  // Create the tab page sequence collection.
  tcd->fPageSequence = new ITabControlPageSequence( );

#ifdef IC_WIN
  // Get the window class of the control window to determine if
  // we're dealing with a tab control.
  IWindowHandle
     hwndControl = IWindow::handleWithParent( windowId,
                                             parentAndOwner->handle() );
  if (IWindowClassName( hwndControl ) == WC_TABCONTROL)
  {
     // Create the page clipping window.
     tcd->fClipWindow = new ITabControlClipWindow( this );

     tcd->fPageKeyboardHandler = new ITabPageKeyboardHandler();
  }
  else
  {
     ITHROWLIBRARYERROR1( IC_WRONG_CONTROL_TYPE,
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable,
                          "WC_TABCONTROL" );
  }
#endif // IC_WIN

#ifdef IC_PM
  tcd->fTabPageResizeHandler = new ITabPageResizeHandler();
#endif // IC_PM
}


/*------------------------------------------------------------------------------
| ITabControl::ITabControl                                                     |
|                                                                              |
| Constructor to instantiate from an existing tab control.                     |
------------------------------------------------------------------------------*/
ITabControl :: ITabControl ( const IWindowHandle& handle )
  : fTabControlData( new ITabControlData( this ))
{
  IMODTRACE_DEVELOP( "ITabControl::ctor3" );

  setAutoDestroyWindow( false );
  reserveUserWindowWord( false );
  startHandlingEventsFor( handle );

  tcd->fDefaultHandler = new ITabControlDefaultHandler();
  tcd->fDefaultHandler->handleEventsFor( this );

  // Create the tab page sequence collection.
  tcd->fPageSequence = new ITabControlPageSequence( );

#ifdef IC_WIN
  // Get the window class of the control window to determine if
  // we're dealing with a tab control.
  if (IWindowClassName( handle ) == WC_TABCONTROL)
  {
     // Create the page clipping window.
     tcd->fClipWindow = new ITabControlClipWindow( this );

     tcd->fPageKeyboardHandler = new ITabPageKeyboardHandler();
  }
  else
  {
     ITHROWLIBRARYERROR1( IC_WRONG_CONTROL_TYPE,
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable,
                          "WC_TABCONTROL" );
  }
#endif // IC_WIN

#ifdef IC_PM
  tcd->fTabPageResizeHandler = new ITabPageResizeHandler();
#endif // IC_PM
}


/*------------------------------------------------------------------------------
| ITabControl::~ITabControl                                                    |
|                                                                              |
| Destructor.                                                                  |
------------------------------------------------------------------------------*/
ITabControl :: ~ITabControl ( )
{
  IMODTRACE_DEVELOP( "ITabControl::dtor" );

  tcd->fDefaultHandler->stopHandlingEventsFor( this );

  if (tcd->fAutoDeletePages)
  {
     // Delete all of the tab pages.  Invalidate the pointer to the tab control
     // prior to deletion.
     while (!tcd->fPageSequence->isEmpty())
     {
        ITabPage
           *tabPage = tcd->fPageSequence->firstElement();
        tcd->fPageSequence->removeFirst();
        tabPage->tpd->fTabControl = 0;
        delete tabPage;
     }
  }

  // Delete the private data object.  Its destructor will delete
  // any other collections/objects.
  delete fTabControlData;
}


/*------------------------------------------------------------------------------
| ITabControl::defaultStyle                                                    |
|                                                                              |
| Return the default style for new tab control objects.                        |
------------------------------------------------------------------------------*/
ITabControl::Style ITabControl :: defaultStyle ( )
{
  IMODTRACE_DEVELOP( "ITabControl::defaultStyle" );

  return fCurrentDefaultStyle;
}


#ifdef IC_WIN
/*------------------------------------------------------------------------------
| ITabControl::setDefaultStyle                                                 |
|                                                                              |
| Sets the default style.                                                      |
|                                                                              |
| Note: This function is inlined on all platforms except Windows due to        |
|       variables it references not being exported.                            |
------------------------------------------------------------------------------*/
void ITabControl::setDefaultStyle ( const ITabControl::Style& style )
{
  IMODTRACE_DEVELOP( "ITabControl::setDefaultStyle" );

  fCurrentDefaultStyle = style;
}
#endif // IC_WIN


/*------------------------------------------------------------------------------
| ITabControl::convertToGUIStyle                                               |
|                                                                              |
| Returns base style for the control by default, or extended style if          |
| extended flag (bExtOnly) is set.                                             |
------------------------------------------------------------------------------*/
unsigned long ITabControl :: convertToGUIStyle ( const IBitFlag& style,
                                                 bool  extendedOnly ) const
{
  IMODTRACE_DEVELOP( "ITabControl::convertToGUIStyle" );

  // Obtain the style from the class (IControl) that we inherit from
  unsigned long ulStyle = Inherited::convertToGUIStyle( style, extendedOnly );

  ITabControl::Style& bitStyle = (ITabControl::Style &)style;
  if (extendedOnly)
  {
#ifdef IC_WIN
     if (bitStyle & ITabControl::border3D)
        ulStyle |= WS_EX_CLIENTEDGE;
#endif // IC_WIN
  }
  else
  {
#ifdef IC_WIN
     // Set the new styles required by the tab control.
     ulStyle |= TCS_TABS | TCS_SINGLELINE | TCS_RAGGEDRIGHT | WS_CHILD;
#endif // IC_WIN

#ifdef IC_PM
     if (bitStyle & ITabControl::buttons)
        ulStyle |= BKS_BUTTONAREA;
     if (bitStyle & ITabControl::tabsTop)
        ulStyle |= BKS_MAJORTABTOP;
     if (bitStyle & ITabControl::tabsBottom)
        ulStyle |= BKS_MAJORTABBOTTOM;

     // Set the style to get the tabbed dialog.
     ulStyle |= BKS_TABBEDDIALOG;
#endif // IC_PM
  }

  return ulStyle;
}


/*------------------------------------------------------------------------------
| ITabControl::setTabPosition                                                  |
|                                                                              |
| Set the tab position style of the tab control.                               |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: setTabPosition ( ETabPosition position )
{
  ITRACE_WIN_NOP();
  IMODTRACE_DEVELOP( "ITabControl::setTabPosition" );

#ifdef IC_PM
  unsigned long ulStyle = style();
  unsigned long ulOldStyle = ulStyle;
  unsigned long ulMask = ~(BKS_MAJORTABTOP | BKS_MAJORTABBOTTOM);
  ulStyle &= ulMask;

  switch (position)
  {
    case kTop:
      ulStyle |= BKS_MAJORTABTOP;
      break;
    case kBottom:
      ulStyle |= BKS_MAJORTABBOTTOM;
      break;
    default:
      ITHROWLIBRARYERROR(IC_INVALIDENUMVALUE,
                         IBaseErrorInfo::invalidParameter,
                         IException::recoverable);
      break;
  }

  if (ulStyle != ulOldStyle)
  {
     setStyle( ulStyle );
     refresh();
  }
#endif

  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::tabPosition                                                     |
------------------------------------------------------------------------------*/
ITabControl::ETabPosition ITabControl :: tabPosition ( ) const
{
  ITRACE_WIN_NOP();
  IMODTRACE_DEVELOP( "ITabControl::tabPosition" );

  ETabPosition tabPosition = kTop;

#ifdef IC_PM
  unsigned long ulStyle = style();
  if (ulStyle & BKS_MAJORTABBOTTOM)
     tabPosition = kBottom;
#endif

  return tabPosition;
}


/*------------------------------------------------------------------------------
| ITabControl::buttonParent                                                    |
------------------------------------------------------------------------------*/
IWindow* ITabControl :: buttonParent ( ) const
{
  IMODTRACE_DEVELOP( "ITabControl::buttonParent" );

  // If this 
#ifdef IC_WIN
  return (IWindow*)tcd->fButtonCanvas;
#elif IC_PM
  return (IWindow*)this;
#endif
}


/*------------------------------------------------------------------------------
| ITabControl::addCommonButton                                                 |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: addCommonButton ( IPushButton* button )
{
  IMODTRACE_DEVELOP( "ITabControl::addCommonButton" );

  IASSERTPARM( button != 0 )

  // Throw an exception if the tab control was not constructed with the buttons
  // style or the buttonParent() is not the button's parent.
#ifdef IC_WIN
  if (!tcd->fButtonCanvas)
#elif IC_PM
  if (~style() & BKS_BUTTONAREA)
#endif // IC_WIN
  {
     ITHROWLIBRARYERROR1( IC_INVALID_STYLE_FOR_FUNCTION,
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable,
                          "ITabControl::buttons" );
  }

  if (button->parent() != buttonParent())
  {
     ITHROWLIBRARYERROR( IC_INVALID_BUTTON_PARENT,
                         IBaseErrorInfo::invalidRequest,
                         IException::recoverable );
  }

  ITabControlButton *tcButton = new ITabControlButton();
  tcButton->fApplicationButton = button;

  tcd->addCommonButton( tcButton, button->isDefault() );

  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::addCommonButton                                                 |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: addCommonButton ( EButton buttonType,
                                              bool    isDefaultButton )
{
  // Throw an exception if the tab control was not constructed with the buttons
  // style.
#ifdef IC_WIN
  if (!tcd->fButtonCanvas)
#elif IC_PM
  if (!(style() & BKS_BUTTONAREA))
#endif // IC_WIN
  {
     ITHROWLIBRARYERROR1( IC_INVALID_STYLE_FOR_FUNCTION,
                          IBaseErrorInfo::invalidRequest,
                          IException::recoverable,
                          "ITabControl::buttons" );
  }

  ITabControlButton *tcButton = new ITabControlButton;
  tcButton->fStandardButton = buttonType;

  // Add the standard button as a common button.
  tcd->addCommonButton( tcButton, isDefaultButton );

  return *this;
}

  
/*------------------------------------------------------------------------------
| ITabControl::removeCommonButton                                              |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: removeCommonButton ( IPushButton* button )
{
  IMODTRACE_DEVELOP( "ITabControl::removeCommonButton" );

  // Find this button in the common button sequence.
  ITabControlButton *tcCommonButton = 0;
  if (tcd->fCommonButtons)
  {
     ITabControlButtonSequence::Cursor cursor( *(tcd->fCommonButtons) );
     forCursor (cursor)
     {
        ITabControlButton
           *tcButton = tcd->fCommonButtons->elementAt( cursor );
        if (tcButton->button( this ) == button)
        {
           tcCommonButton = tcButton;
           cursor.setToLast();
        }
     }
  }

  if (!tcCommonButton)
  {
     // Throw an exception if we didn't find the button.
     ITHROWLIBRARYERROR1( IC_CONTROL_NOT_FOUND,
                          IBaseErrorInfo::accessError,
                          IException::recoverable,
                          "IPushButton*");
  }

  // Remove this button from the tab control.
  tcd->removeCommonButton( tcCommonButton );
  delete tcCommonButton;

  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::removeCommonButton                                              |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: removeCommonButton ( EButton buttonType )
{
  IMODTRACE_DEVELOP( "ITabControl::removeCommonButton" );

  ITabControlButton *tcCommonButton = 0;
  if (tcd->fCommonButtons)
  {
     ITabControlButtonSequence::Cursor cursor( *(tcd->fCommonButtons) );
     forCursor (cursor)
     {
        ITabControlButton 
           *tcButton = tcd->fCommonButtons->elementAt( cursor );
        if (tcButton->fStandardButton == buttonType)
        {
           tcCommonButton = tcButton;
           cursor.setToLast();
        }
     }
  }

  if (!tcCommonButton)
  {
     // Throw an exception if we didn't find the button.
     ITHROWLIBRARYERROR1( IC_CONTROL_NOT_FOUND,
                          IBaseErrorInfo::accessError,
                          IException::recoverable,
                          "IPushButton*");
  }

  // Remove this button from the tab control.
  tcd->removeCommonButton( tcCommonButton );
  delete tcCommonButton;

  return *this;
}

  
/*------------------------------------------------------------------------------
| ITabControl::button                                                          |
------------------------------------------------------------------------------*/
IPushButton* ITabControl :: button ( EButton buttonType ) const
{
  IMODTRACE_DEVELOP( "ITabControl::button" );

  // Tab pages can add standard buttons as page buttons.  They can do so
  // prior to adding the tab page to any tab control.  This means that we
  // won't necessarily create the standard buttons until the page is added
  // to the tab control.  In order to allow access to these standard buttons
  // before pages are added to the tab control, we'll go ahead and create them
  // when the application attempts to access them.
  unsigned long index = tcd->buttonIndex( buttonType );
  IPushButton *commonButton = tcd->fStandardButton[ index ];
  if (!commonButton)
  {
     // Create the button since it doesn't already exist.
     commonButton = tcd->createButton( buttonType );
  }
  
  return commonButton;
}


/*------------------------------------------------------------------------------
| ITabControl::setPageBackgroundColor                                          |
|                                                                              |
| Sets the background page color of the tab control.                           |
------------------------------------------------------------------------------*/
ITabControl& ITabControl::setPageBackgroundColor ( const IColor& color )
{
  IMODTRACE_DEVELOP( "ITabControl::setPageBackgroundColor" );

  if (!tcd->fPageBkgrndColorSet || color != tcd->fPageBkgrndColor)
  {
#ifdef IC_WIN
     tcd->fPageBkgrndColorSet = true;
     tcd->fPageBkgrndColor = color;

     // Force a repaint of the clip window if the top page is not auto sized.
     ITabPage *currentPage = topPage();
     if (currentPage && !currentPage->isAutoPageSize())
     {
        tcd->fClipWindow->refresh();
     }
#endif
          
#ifdef IC_PM
     if (!sendEvent( BKM_SETNOTEBOOKCOLORS,
                     (unsigned long)color.asRGBLong(),
                     BKA_BACKGROUNDPAGECOLOR ).asUnsignedLong())
     {
        ITHROWGUIERROR( "BKM_SETNOTEBOOKCOLORS" );
     }
#endif

     // Send a notification to all observers.
     notifyObservers( INotificationEvent( pageBackgroundColorId,
                                          *this ));
  }

  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::resetPageBackgroundColor                                        |
|                                                                              |
| Resets the background page color of the tab control.                         |
------------------------------------------------------------------------------*/
ITabControl& ITabControl::resetPageBackgroundColor ( )
{
  IMODTRACE_DEVELOP( "ITabControl::resetPageBackgroundColor" );

  if (tcd->fPageBkgrndColorSet)
  {
#ifdef IC_PM
     sendEvent( BKM_SETNOTEBOOKCOLORS,
                (unsigned long)IGUIColor(IGUIColor::notebookPageBgnd).asRGBLong(),
                BKA_BACKGROUNDPAGECOLOR );
#endif
     tcd->fPageBkgrndColorSet = false;
     notifyObservers( INotificationEvent( pageBackgroundColorId,
                                          *this ));
  }
  
  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::pageBackgroundColor                                             |
|                                                                              |
| Returns the background page color of the tab control.                        |
------------------------------------------------------------------------------*/
IColor ITabControl::pageBackgroundColor ( ) const
{
  IMODTRACE_DEVELOP( "ITabControl::pageBackgroundColor" );

  // If this color was set previously, retrieve it from where we saved
  // it and return it.  Otherwise, if a color has not been previously
  // set, return the default.
  if (tcd->fPageBkgrndColorSet)
     return (IColor( tcd->fPageBkgrndColor ));

  return( IGUIColor(IGUIColor::notebookPageBgnd) );
}


/*------------------------------------------------------------------------------
| ITabControl::addFirstPage                                                    |
|                                                                              |
| Add a page to the tab control as the first page.                             |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: addFirstPage ( const ITabPage* newPage )
{
  IMODTRACE_DEVELOP( "ITabControl::addFirstPage" );

  IASSERTPARM( newPage != 0 )

  tcd->addPage( 1, newPage );
  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::addLastPage                                                     |
|                                                                              |
| Add a page to the tab control as the last page.                              |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: addLastPage ( const ITabPage* newPage )
{
  IMODTRACE_DEVELOP( "ITabControl::addLastPage" );

  IASSERTPARM( newPage != 0 )

  tcd->addPage( tcd->fPageSequence->numberOfElements()+1, newPage );
  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::addPageBefore                                                   |
|                                                                              |
| Add a given page to the tab control before the referenced page.              |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: addPageBefore ( const ITabPage* newPage,
                                            const ITabPage* referencePage )
{
  IMODTRACE_DEVELOP( "ITabControl::addPageBefore" );

  IASSERTPARM( newPage != 0 )

  tcd->addPage( tcd->tabPageIndex( referencePage ), newPage );
  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::addPageBefore                                                   |
|                                                                              |
| Add a given page to the tab control before the referenced page.              |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: addPageBefore ( const ITabPage* newPage,
                                            Cursor&         cursor )
{
  IMODTRACE_DEVELOP( "ITabControl::addPageBefore" );

  IASSERTPARM( newPage != 0 )

  tcd->addPage( tcd->tabPageIndex( cursor.current() ), newPage );
  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::addPageAfter                                                    |
|                                                                              |
| Add a given page to the tab control after the referenced page.               |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: addPageAfter ( const ITabPage* newPage,
                                           const ITabPage* referencePage )
{
  IMODTRACE_DEVELOP( "ITabControl::addPageAfter" );

  IASSERTPARM( newPage != 0 )

  tcd->addPage( tcd->tabPageIndex( referencePage )+1, newPage );
  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::addPageAfter                                                    |
|                                                                              |
| Add a given page to the tab control after the referenced page.               |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: addPageAfter ( const ITabPage* newPage,
                                           Cursor&         cursor )
{
  IMODTRACE_DEVELOP( "ITabControl::addPageAfter" );

  IASSERTPARM( newPage != 0 )

  tcd->addPage( tcd->tabPageIndex( cursor.current() )+1, newPage );
  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::removePage                                                      |
|                                                                              |
| Remove a given page from the tab control.                                    |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: removePage ( ITabPage* page )
{
  IMODTRACE_DEVELOP( "ITabControl::removePage" );

  if (!tcd->fPageSequence->contains( page ))
  {
     ITHROWLIBRARYERROR1( IC_CONTROL_NOT_FOUND,
                          IBaseErrorInfo::accessError,
                          IException::recoverable,
                          "ITabPage*");
  }

  tcd->deletePage( page, false );
  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::removePage                                                      |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: removePage ( Cursor& cursor )
{
  IMODTRACE_DEVELOP( "ITabControl::removePage" );

  IASSERTPARM( cursor.isValid() )

  tcd->deletePage( cursor.current(), false );
  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::removeAllPages                                                  |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: removeAllPages ( )
{
  IMODTRACE_DEVELOP( "ITabControl::removeAllPages" );

  tcd->deletePage( 0, true );

  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::setTopPage                                                      |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: setTopPage ( const ITabPage* page )
{
  IMODTRACE_DEVELOP( "ITabControl::setTopPage" );

  // Throw an exception if this tab page hasn't been added to the tab control.
  if (!page || page->tpd->fTabControl != this)
  {
     ITHROWLIBRARYERROR( IC_TABPAGE_NOT_FOUND,
                         IBaseErrorInfo::invalidParameter,
                         IException::recoverable );
  }

#ifdef IC_WIN
  unsigned long currentIndex = tcd->tabPageIndex( topPage() ),
                newIndex = tcd->tabPageIndex( page );
  if (TabCtrl_SetCurSel( handle(), newIndex-1 ) == -1)
  {
     ITHROWGUIERROR("TabCtrl_SetCurSel");
  }

  // The tab selection has already occurred and the new tab is now the
  // top page in the tab control.  Do our required processing and issue
  // the selection notification.

  NMHDR notifyInfo;
  notifyInfo.hwndFrom = handle();
  notifyInfo.idFrom = (UINT)id();
  notifyInfo.code = TCN_SELCHANGE;
  parent()->handle().sendEvent( WM_NOTIFY,
                                id(),
                                &notifyInfo );

  tcd->processTabSelect( currentIndex, newIndex );
#endif // IC_WIN

#ifdef IC_PM
  if (!sendEvent( BKM_TURNTOPAGE,
                  page->tpd->fPageId,
                  0 ).asUnsignedLong())
  {
     ITHROWGUIERROR( "BKM_TURNTOPAGE" );
  }
#endif // IC_PM

  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::setTopPage                                                      |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: setTopPage ( const Cursor& cursor )
{
  IMODTRACE_DEVELOP( "ITabControl::setTopPage" );

  return setTopPage( cursor.current() );
}


/*------------------------------------------------------------------------------
| ITabControl::topPage                                                         |
|                                                                              |
| Returns a pointer to the ITabPage object that corresponds to the top page    |
| of the tab control.                                                          |
------------------------------------------------------------------------------*/
ITabPage* ITabControl :: topPage ( ) const
{
  IMODTRACE_DEVELOP( "ITabControl::topPage" );

  ITabPage *topPage = 0;
  if (totalPages())
  {
#ifdef IC_WIN
     long pageIndex = TabCtrl_GetCurSel( handle() );

     if (pageIndex != -1)
     {
        TC_ITEM tabItem;
        tabItem.mask = TCIF_PARAM;
        if (!TabCtrl_GetItem( handle(),
                              pageIndex,
                              &tabItem ))
        {
           ITHROWGUIERROR( "TabCtrl_GetItem" );
        }
        topPage = (ITabPage*)(void*)tabItem.lParam;
     }
#endif // IC_WIN

#ifdef IC_PM
     topPage = tcd->queryTabPage( BKA_TOP, 0 );
#endif // IC_PM
  }

  return topPage;
}


/*------------------------------------------------------------------------------
| ITabControl::firstPage                                                       |
|                                                                              |
| Returns a pointer to the ITabPage object that corresponds to the first page  |
| of the tab control.                                                          |
------------------------------------------------------------------------------*/
ITabPage* ITabControl :: firstPage ( ) const
{
  IMODTRACE_DEVELOP( "ITabControl::firstPage" );

  ITabPage *firstPage = 0;

  if (totalPages())
  {
#ifdef IC_WIN
     TC_ITEM tabItem;
     tabItem.mask = TCIF_PARAM;
     if (!TabCtrl_GetItem( handle(),
                           0,
                           &tabItem ))
     {
        ITHROWGUIERROR( "TabCtrl_GetItem" );
     }
     firstPage = (ITabPage*)(void*)tabItem.lParam;
#endif // IC_WIN

#ifdef IC_PM
     firstPage = tcd->queryTabPage( BKA_FIRST, 0 );
#endif // IC_PM
  }

  return firstPage;
}


/*------------------------------------------------------------------------------
| ITabControl::lastPage                                                        |
|                                                                              |
| Returns a pointer to the ITabPage object that corresponds to the last page   |
| of the tab control.                                                          |
------------------------------------------------------------------------------*/
ITabPage* ITabControl :: lastPage ( ) const
{
  IMODTRACE_DEVELOP( "ITabControl::lastPage" );

  ITabPage *lastPage = 0;
  unsigned long pageCount = totalPages();

  if (pageCount)
  {
#ifdef IC_WIN
     TC_ITEM tabItem;
     tabItem.mask = TCIF_PARAM;
     if (!TabCtrl_GetItem( handle(),
                           pageCount-1,
                           &tabItem ))
     {
        ITHROWGUIERROR( "TabCtrl_GetItem" );
     }
     lastPage = (ITabPage*)(void*)tabItem.lParam;
#endif // IC_WIN

#ifdef IC_PM
     lastPage = tcd->queryTabPage( BKA_LAST, 0 );     
#endif // IC_PM
  }
  
  return lastPage;
}


/*------------------------------------------------------------------------------
| ITabControl::nextPage                                                        |
|                                                                              |
| Returns a pointer to the ITabPage object that corresponds to the next page   |
| of the tab control.                                                          |
------------------------------------------------------------------------------*/
ITabPage* ITabControl :: nextPage ( const ITabPage* page ) const
{
  IMODTRACE_DEVELOP( "ITabControl::nextPage" );

#ifdef IC_WIN
  ITabPage *nextPage = 0;

  // Verify the reference page.  If it does not exist, throw an exception.
  ITabControlPageSequence::Cursor pageCursor( *(tcd->fPageSequence) );
  ITabPage *referencePage = (ITabPage*)page;
  if (tcd->fPageSequence->locate( referencePage, pageCursor ))
  {
     if (pageCursor.setToNext())
     {
        nextPage = tcd->fPageSequence->elementAt( pageCursor );
     }
  }
  else
  {
     ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                         IBaseErrorInfo::invalidParameter,
                         IException::recoverable);
  }

  return nextPage;
#endif // IC_WIN

#ifdef IC_PM
  return tcd->queryTabPage( BKA_NEXT, page );
#endif // IC_PM
}


/*------------------------------------------------------------------------------
| ITabControl::previousPage                                                    |
|                                                                              |
| Returns a pointer to the ITabPage object that corresponds to the previous    |
| page of the tab control.                                                     |
------------------------------------------------------------------------------*/
ITabPage* ITabControl :: previousPage ( const ITabPage* page ) const
{
  IMODTRACE_DEVELOP( "ITabControl::previousPage" );

#ifdef IC_WIN
  ITabPage *previousPage = 0;

  // Verify the reference page.  If it does not exist, throw an exception.
  ITabControlPageSequence::Cursor pageCursor( *(tcd->fPageSequence) );
  ITabPage *referencePage = (ITabPage*)page;
  if (tcd->fPageSequence->locate( referencePage, pageCursor ))
  {
     if (tcd->fPageSequence->setToPrevious( pageCursor ))
     {
        previousPage = tcd->fPageSequence->elementAt( pageCursor );
     }
  }
  else
  {
     ITHROWLIBRARYERROR( IC_INVALIDHANDLE,
                         IBaseErrorInfo::invalidParameter,
                         IException::recoverable);
  }

  return previousPage;
#endif

#ifdef IC_PM
  return tcd->queryTabPage( BKA_PREV, page );
#endif // IC_PM
}


/*------------------------------------------------------------------------------
| ITabControl::pageSize                                                        |
|                                                                              |
| Return the size of the "viewport" in which the tab control draws the pages   |
| it contains.                                                                 |
------------------------------------------------------------------------------*/
ISize ITabControl :: pageSize ( ) const
{
  IMODTRACE_DEVELOP( "ITabControl::pageSize" );

  RECTL tabControlSize = rect().asRECTL();

#ifdef IC_WIN
  tcd->adjustRect( false, &tabControlSize );
#endif // IC_WIN

#ifdef IC_PM
  if (!sendEvent( BKM_CALCPAGERECT,
                  (unsigned long)&tabControlSize,
                  true ).asUnsignedLong())
  {
     ITHROWGUIERROR( "BKM_CALCPAGERECT" );
  }

#endif // IC_PM

  return ISize( tabControlSize );
}


/*------------------------------------------------------------------------------
| ITabControl::totalPages                                                      |
|                                                                              |
| Return the total number of pages in the tab control.                         |
------------------------------------------------------------------------------*/
unsigned long ITabControl :: totalPages ( ) const
{
  IMODTRACE_DEVELOP( "ITabControl::totalPages" );

  unsigned long pageCount = 0;

#ifdef IC_WIN
  pageCount = (unsigned long)TabCtrl_GetItemCount( handle() );
#endif // IC_WIN

#ifdef IC_PM
  pageCount = sendEvent( BKM_QUERYPAGECOUNT,
                         0,
                         BKA_END );
#endif // IC_PM

  return pageCount;
}


/*------------------------------------------------------------------------------
| ITabControl::isEmpty                                                         |
|                                                                              |
| Queries whether the tab control has any pages.                               |
------------------------------------------------------------------------------*/
bool ITabControl :: isEmpty ( ) const
{
  IMODTRACE_DEVELOP( "ITabControl::isEmpty" );

  if (totalPages())
    return false;
  else
    return true;
}


/*------------------------------------------------------------------------------
| ITabControl::setAutoDeletePages                                              |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: setAutoDeletePages ( bool autoDelete )
{
  IMODTRACE_DEVELOP( "ITabControl::setAutoDeletePages" );
  
  tcd->fAutoDeletePages = autoDelete;

  return *this;
}


/*------------------------------------------------------------------------------
| ITabControl::isAutoDeletePages                                               |
------------------------------------------------------------------------------*/
bool ITabControl :: isAutoDeletePages ( ) const
{
  IMODTRACE_DEVELOP( "ITabControl::isAutoDeletePages" );
  
  return tcd->fAutoDeletePages;
}


#ifdef IC_PMWIN
/*------------------------------------------------------------------------------
| ITabControl::setLayoutDistorted                                              |
|                                                                              |
| Cause a font change to result in a layout of the tab control buttons.        |
| Propagate the font change to the page windows and button canvas.             |
------------------------------------------------------------------------------*/
ITabControl& ITabControl :: setLayoutDistorted ( unsigned long layoutAttributesOn,
                                                 unsigned long layoutAttributesOff )
{
  IMODTRACE_DEVELOP( "ITabControl::setLayoutDistorted" );

#ifdef IC_PM
  if ((layoutAttributesOn & IWindow::childMinimumSizeChanged) &&
      (~layoutAttributesOn & IWindow::childWindowCreated))
  {
     // We want to trigger the OS/2 notebook to re-layout the button area.
     // Note:  This is not working yet.
     if (style() & BKS_BUTTONAREA)
     {
        handle().sendEvent( WM_PRESPARAMCHANGED,
                            IEventParameter1( (unsigned long)PP_USER ),
                            IEventParameter2( 0 ) );
     }
  }
  else if (layoutAttributesOn & IWindow::childWindowDestroyed)
  {
     // One of our children has been destroyed.  We need to remove any deleted
     // button from the button sequences.  In Windows, this setLayoutDistorted
     // is in the ITabControlButtonCanvas class.  We'll post a message so
     // we can process the change in the default handler for both platforms.
     handle().sendEvent( IC_UM_TCN_BTNDELETED,
                         IEventParameter1( 0 ),
                         IEventParameter2( 0 ));
  }
#endif

#ifdef IC_WIN
  if ((layoutAttributesOn & IWindow::fontChanged) &&
      !(layoutAttributesOn & IWindow::windowCreated))
  {
     IFont tcFont = font();
   
     // We need to propagate the tab control font to each of the page windows
     // for which the font has not been explicitly set.
     propagationInProgress++;
     
     if (tcd->fPageSequence && !tcd->fPageSequence->isEmpty())
     {
        ITabControlPageSequence::Cursor pageCursor( *(tcd->fPageSequence) );
        IWindow *pageWindow;
        forCursor( pageCursor )
        {
           pageWindow = tcd->fPageSequence->elementAt( pageCursor )->pageWindow();
           if (pageWindow)
              pageWindow->setFont( tcFont );
        }      
     }
   
     // Also need to set the font of the button canvas and recalculate
     // the button canvas height.
     if (tcd->fButtonCanvas)
     {
        tcd->fButtonCanvas->setFont( tcFont );
        tcd->fButtonCanvasHeight = tcd->buttonCanvasHeight();
     }
     
     propagationInProgress--;
   
     // Force a layout of the clip window and button canvas.
     tcd->layoutChildren( size() );
  }
#endif

  Inherited::setLayoutDistorted( layoutAttributesOn, layoutAttributesOff );

  return *this;
}
#endif // IC_PMWIN


/*------------------------------------------------------------------------------
| ITabControl::calcMinimumSize                                                 |
|                                                                              |
| Calculate the minimum screen size needed by the control.  The height is      |
| based on the minimum size of the page window(s), with the tab control        |
| drawing around them.                                                         |
------------------------------------------------------------------------------*/
ISize ITabControl :: calcMinimumSize ( ) const
{
  IMODTRACE_DEVELOP( "ITabControl::calcMinimumSize" );

  ISize minTabControlSize;
  ISize minPageWindowSize;

  // Determine the minimum area needed for the page windows.
  if (!tcd->fPageSequence->isEmpty())
  {
     ITabControlPageSequence::Cursor cursor( *tcd->fPageSequence );
     forCursor( cursor )
     {
        ITabPage *tabPage = tcd->fPageSequence->elementAt( cursor );
        minPageWindowSize = minPageWindowSize.maximum(
                                  tabPage->minimumPageWindowSize() );
     }
  }
  else
  {
     // If no pages have been added to the tab control, use the Inherited
     // default minimum size.
     minPageWindowSize = Inherited::calcMinimumSize();
  }
  
  // Now using the actual control rectangle, get the page rectangle.  This
  // will give us the correct origin for our minimum size page window.
  RECTL pageRect = rect().asRECTL();

#ifdef IC_WIN
  tcd->adjustRect( false, &pageRect );
  RECTL newPageRect = IRectangle( IPoint( pageRect.left, pageRect.bottom ),
                                  minPageWindowSize ).asRECTL();
  // Now convert back to tab control rectangle.
  tcd->adjustRect( true, &newPageRect );
#endif // IC_WIN

#ifdef IC_PM
  if (!sendEvent( BKM_CALCPAGERECT,
                  &pageRect,
                  true ).asUnsignedLong())
  {
     ITHROWGUIERROR( "BKM_CALCPAGERECT" );
  }
  RECTL newPageRect = IRectangle( IPoint( pageRect.xLeft, pageRect.yBottom ),
                                  minPageWindowSize ).asRECTL();
  // Now convert back to tab control rectangle.
  if (!sendEvent( BKM_CALCPAGERECT,
                  &newPageRect,
                  false ).asUnsignedLong())
  {
     ITHROWGUIERROR( "BKM_CALCPAGERECT" );
  }
#endif // IC_PM
           
  minTabControlSize = ISize( newPageRect );

#ifdef IC_WIN
  // Need to account for the button canvas.  Since the buttons can vary per
  // page, we need to cursor through the pages and set up the button canvas
  // to get the minimum size.
  if (tcd->fButtonCanvas)
  {
     tcd->fButtonCanvas->disableUpdate();
     tcd->fButtonCanvas->setMargin( ISize( tcd->fDefButtonMargin.width(),
                                           tcd->fButtonCanvas->margin().height() ));
     ITabPage *currentPage = topPage();
     ISize     minButtonSize = tcd->fButtonCanvas->minimumSize();

     // Remove any page buttons.
     if (currentPage)
     {
        ITabControlButtonSequence
           *pageButtons = currentPage->tpd->fPageButtons;
        if (pageButtons)
        {
           ITabControlButtonSequence::Cursor cursor( *pageButtons );
           forCursor (cursor)
           {
              ITabControlButton *tcPageButton = pageButtons->elementAt( cursor );
              IPushButton *pageButton = tcPageButton->button( this );
              if (tcd->fButtonCanvas->isInLayout( pageButton ))
                 tcd->fButtonCanvas->remove( pageButton );
           }
        }

        // Add the common buttons to the set canvas and get the button canvas'
        // minimum size.
        ITabControlButtonSequence *commonButtons = tcd->fCommonButtons;
        if (commonButtons)
        {
           ITabControlButtonSequence::Cursor cursor( *commonButtons );
           ITabControlButton *tcCommonButton;
           IPushButton *commonButton;
           forCursor (cursor)
           {
              tcCommonButton = commonButtons->elementAt( cursor );
              commonButton = tcCommonButton->button( this );
              if (!tcd->fButtonCanvas->isInLayout( commonButton ))
                 tcd->fButtonCanvas->add( commonButton );
           }

           ISize minCommonButtonSize = tcd->fButtonCanvas->minimumSize();
           minButtonSize = minButtonSize.maximum( minCommonButtonSize );

           forCursor (cursor)
           {
              tcCommonButton = commonButtons->elementAt( cursor );
              commonButton = tcCommonButton->button( this );
              tcd->fButtonCanvas->remove( commonButton );
           }
        }

        // Now cursor through each page and find the button canvas minimum size
        // for the page buttons.
        if (!isEmpty())
        {
           for (ITabPage *tabPage = firstPage();
                tabPage;
                tabPage = nextPage( tabPage ))
           {
              ITabControlButtonSequence *pageButtons = tabPage->tpd->fPageButtons;
              if (pageButtons)
              {
                 ITabControlButtonSequence::Cursor cursor( *pageButtons );
                 ITabControlButton *tcPageButton;
                 IPushButton *pageButton;
                 forCursor (cursor)
                 {
                    tcPageButton = pageButtons->elementAt( cursor );
                    pageButton = tcPageButton->button( this );
                    tcd->fButtonCanvas->add( pageButton );
                 }

                 ISize minPageButtonSize = tcd->fButtonCanvas->minimumSize();
                 minButtonSize = minButtonSize.maximum( minPageButtonSize );

                 forCursor (cursor)
                 {
                    tcPageButton = pageButtons->elementAt( cursor );
                    pageButton = tcPageButton->button( this );
                    tcd->fButtonCanvas->remove( pageButton );
                 }
              }
           }
        }
        minButtonSize += ISize( 4, 0 );

        // Now consider the button canvas width in the tab control minimum size.
        minTabControlSize +=
              ISize( MAX(0, (long)(minButtonSize.width() - minTabControlSize.width())),
                     tcd->fButtonCanvasHeight );

        // Now restore the correct buttons.
        if (!tcd->fInCreate)
           tcd->showButtons( 0, currentPage );
     }
  }

  // Add a little more if this control has a 3d border.
  unsigned long ulExtStyle = GetWindowLong( handle(), GWL_EXSTYLE );
  if (ulExtStyle & WS_EX_CLIENTEDGE)
  {
     minTabControlSize += ISize( 4, 4 );
  }
#endif

  return minTabControlSize;
}


/*------------------------------------------------------------------------------
| ITabControl::Cursor::Cursor                                                  |
|                                                                              |
| Cursor class constructor                                                     |
------------------------------------------------------------------------------*/
ITabControl::Cursor :: Cursor ( const ITabControl& tabControl )
  : fTabControl ((ITabControl&)tabControl),
    fCurrentTabPage (0)
{ }

/*------------------------------------------------------------------------------
| ITabControl::Cursor::~Cursor                                                 |
|                                                                              |
| Default destructor used for page tuning.                                     |
------------------------------------------------------------------------------*/
ITabControl :: Cursor :: ~Cursor ( )
{ }


/*------------------------------------------------------------------------------
| ITabControl::Cursor::setToFirst                                              |
|                                                                              |
| Sets the cursor to point to the first page of the tab control.               |
------------------------------------------------------------------------------*/
bool ITabControl::Cursor :: setToFirst ( )
{
  if (!fTabControl.isEmpty())
  {
     fCurrentTabPage = fTabControl.firstPage();

     // Make cursor valid.
     fReferences = fTabControl.tcd->fValidate;
     return true;
  }
  else
  {
     invalidate();
     return false;
  }
}


/*------------------------------------------------------------------------------
| ITabControl::Cursor::setToNext                                               |
|                                                                              |
| Sets the cursor to point to the next page of the tab control.                |
------------------------------------------------------------------------------*/
bool ITabControl::Cursor :: setToNext ( )
{
  ITabPage *nextPage = fTabControl.nextPage( fCurrentTabPage );

  if (nextPage)
  {
     fCurrentTabPage = nextPage;
     fReferences = fTabControl.tcd->fValidate;
     return true;
  }
  else
  {
     invalidate();
     return false;
  }
}


/*------------------------------------------------------------------------------
| ITabControl::Cursor::setToPrevious                                           |
|                                                                              |
| Sets the cursor to point to the previous page in the tab control.            |
------------------------------------------------------------------------------*/
bool ITabControl::Cursor :: setToPrevious ( )
{
  ITabPage *previousPage = fTabControl.previousPage( fCurrentTabPage );

  if (previousPage)
  {
     fCurrentTabPage = previousPage;
     fReferences = fTabControl.tcd->fValidate;
     return true;
  }
  else
  {
     invalidate();
     return false;
  }
}


/*------------------------------------------------------------------------------
| ITabControl::Cursor::setToLast                                               |
|                                                                              |
| Sets the cursor to point to the last page in the tab control.                |
------------------------------------------------------------------------------*/
bool ITabControl::Cursor :: setToLast ( )
{
  if (!fTabControl.isEmpty())
  {
     fCurrentTabPage = fTabControl.lastPage();
     fReferences = fTabControl.tcd->fValidate;
     return true;
  }
  else
  {
     invalidate();
     return false;
  }
}


/*------------------------------------------------------------------------------
| ITabControl::Cursor::current                                                 |
|                                                                              |
| Return a pointer to an ITabPage object that references the page currently    |
| pointed to by the cursor.                                                    |
------------------------------------------------------------------------------*/
ITabPage* ITabControl::Cursor :: current ( ) const
{
  ITabPage *currentPage = 0;

  if (isValid())
     currentPage = fCurrentTabPage;

  return currentPage;
}


/*------------------------------------------------------------------------------
| ITabControl::Cursor::setCurrent                                              |
|                                                                              |
| Set the cursor to point to the provided page.                                |
------------------------------------------------------------------------------*/
void ITabControl::Cursor :: setCurrent ( const ITabPage* current )
{
  IASSERTPARM( current != 0 )

  fCurrentTabPage = (ITabPage*)current;
  fReferences = fTabControl.tcd->fValidate;
}


/*------------------------------------------------------------------------------
| ITabControl::Cursor::isValid                                                 |
|                                                                              |
| Queries whether the cursor points to a valid item.                           |
------------------------------------------------------------------------------*/
bool ITabControl::Cursor :: isValid ( ) const
{
  if (fCurrentTabPage && (fReferences == fTabControl.tcd->fValidate))
     return true;
  else
     return false;
}


/*------------------------------------------------------------------------------
| ITabControl::Cursor::invalidate                                              |
|                                                                              |
| Flags this cursor as not valid.                                              |
------------------------------------------------------------------------------*/
void ITabControl::Cursor :: invalidate ( )
{
  fCurrentTabPage = 0;
}
