// Revision: 24 1.7.1.4 source/ui/extapp/iclipbrd.cpp, clipboard, ioc.v400, 000302  
/*******************************************************************************
* FILE NAME: iclipbrd.cpp                                                      *
*                                                                              *
* DESCRIPTION:                                                                 *
*   This file contains the implementation of classes/functions declared        *
*   in iclipbrd.hpp.                                                           *
*                                                                              *
* COPYRIGHT:                                                                   *
*   IBM Open Class Library                                                     *
*   (C) Copyright International Business Machines Corporation 1992, 1997       *
*   Licensed Material - Program-Property of IBM - All Rights Reserved.         *
*   US Government Users Restricted Rights - Use, duplication, or disclosure    *
*   restricted by GSA ADP Schedule Contract with IBM Corp.                     *
*                                                                              *
*******************************************************************************/
#pragma priority( -2147481424 )

#include <string.h>
#include <iclipbrd.hpp>
#include <iclipimp.hpp>
#include <ihandle.hpp>
#include <iwindow.hpp>
#include <iexcept.hpp>
#include <icconst.h>
#include <ithread.hpp>
#include <istring.hpp>
#include <igimage.hpp>
#include <itrace.hpp>


// Segment definitions
#ifdef IC_PAGETUNE
  #define _ICLIPBRD_CPP_
  #include <ipagetun.h>
#endif


/*------------------------------------------------------------------------------
| Clipboard::Cursor Data class                                                 |
|  Note: On Motif, fItemHandle is the index of the format while                |
|  on OS/2 and Windows it is the actual handle of the format.  It              |
|  is up to the caller to use them accordingly.                                |
------------------------------------------------------------------------------*/
#pragma enum(4)
#pragma pack(push,4)

class IClipboardCursorData
{
public:
IClipboardCursorData ( )
  : fItemHandle ( 0 )
  {}

unsigned long
  fItemHandle;
friend class IClipboard;
friend class IClipboard::Cursor;
};

#pragma pack(pop)
#pragma enum(pop)

/*------------------------------------------------------------------------------
| IClipboard::IClipboard                                                       |
|                                                                              |
| Private constructor - used for Enumeration clipboard formats.                |
------------------------------------------------------------------------------*/
IClipboard::IClipboard ( )
#ifdef IC_WIN
       : fClipboardImp(new IWindowsClipboardImp(0))
#endif
#ifdef IC_PM
       : fClipboardImp(new IOS2ClipboardImp(0))
#endif
#ifdef IC_MOTIF
       : fClipboardImp(new IMotifClipboardImp(0))
#endif
{
  IFUNCTRACE_DEVELOP();
}


/*------------------------------------------------------------------------------
| IClipboard::IClipboard                                                       |
|                                                                              |
| Constructor to create a clipboard object with an owner.                      |
------------------------------------------------------------------------------*/
IClipboard::IClipboard ( const IWindowHandle& clipboardWindow)
#ifdef IC_WIN
       : fClipboardImp(new IWindowsClipboardImp(clipboardWindow))
#endif
#ifdef IC_PM
       : fClipboardImp(new IOS2ClipboardImp(clipboardWindow))
#endif
#ifdef IC_MOTIF
       : fClipboardImp(new IMotifClipboardImp(clipboardWindow))
#endif
{
   IFUNCTRACE_DEVELOP();
   IASSERTPARM(clipboardWindow!=0 && clipboardWindow.isValid());
}

/*------------------------------------------------------------------------------
| IClipboard::~IClipboard                                                      |
|                                                                              |
| Delete a clipboard object.                                                   |
------------------------------------------------------------------------------*/
IClipboard::~IClipboard ( )
{
  IFUNCTRACE_DEVELOP();

  close();
  delete fClipboardImp;
}

/*------------------------------------------------------------------------------
| IClipboard::open                                                             |
|                                                                              |
| Open the clipboard object.                                                   |
| A window handle is required for Windows but not for OS/2.  For consistency,  |
| we have moved the requirement for specifying a window handle up to the       |
| constructor.  The handle is used as the owner if an owner is                 |
| necessary(empty).  Its used here to open the clipboard (on Windows).         |
------------------------------------------------------------------------------*/
IClipboard &IClipboard::open( )
{
  IFUNCTRACE_DEVELOP();

  fClipboardImp->open();

  return *this;
}

/*------------------------------------------------------------------------------
| IClipboard::close                                                            |
|                                                                              |
| Close the clipboard object.                                                  |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::close( )
{
  IFUNCTRACE_DEVELOP();

  fClipboardImp->close();

  return *this;
}

/*------------------------------------------------------------------------------
| IClipboard::empty                                                            |
|                                                                              |
| Empty the contents of the clipboard and establish a new clipboard owner. We  |
| use the owner that was supplied on the open call.                            |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::empty( )
{
  IFUNCTRACE_DEVELOP();

  fClipboardImp->emptyClipboard();

  return *this;
}

/*------------------------------------------------------------------------------
| IClipboard::isOpen                                                           |
|                                                                              |
| Return true if the clipboard is open.                                        |
------------------------------------------------------------------------------*/
bool IClipboard::isOpen( ) const
{
  IFUNCTRACE_DEVELOP();

  return fClipboardImp->isOpen();
}


/*------------------------------------------------------------------------------
| IClipboard::setText                                                          |
|                                                                              |
| Make a copy of the passed text and place it on the clipboard by calling      |
| setData.                                                                     |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::setText( const char* text)
{
  IFUNCTRACE_DEVELOP();
  IASSERTPARM(text!=0);

  return setData(IClipboard::textFormat, text, strlen(text));
}

/*------------------------------------------------------------------------------
| IClipboard::setBitmap                                                        |
|                                                                              |
| Make a copy of the passed bitmap and place it on the clipboard by calling    |
| setHandle.                                                                   |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::setBitmap( const IBitmapHandle& bitmap)
{
  IFUNCTRACE_DEVELOP();


  IGImage image(bitmap);
  IBitmapHandle newBitmap = image.copyBitmap();

  return setHandle (IClipboard::bitmapFormat, newBitmap.asUnsigned ());
}


/*------------------------------------------------------------------------------
| IClipboard::hasText                                                          |
|                                                                              |
| Returns true if the clipboard has text in it (IClipboard::textFormat).       |
------------------------------------------------------------------------------*/
bool IClipboard::hasText( ) const
{
  IFUNCTRACE_DEVELOP();

  return hasData(IClipboard::textFormat);
}

/*------------------------------------------------------------------------------
| IClipboard::hasText                                                          |
|                                                                              |
| Returns true if the clipboard has a bitmap in it (IClipboard::bitmapFormat). |
------------------------------------------------------------------------------*/
bool IClipboard::hasBitmap( ) const
{
  IFUNCTRACE_DEVELOP();

  return hasData(IClipboard::bitmapFormat);
}


/*------------------------------------------------------------------------------
| IClipboard::text                                                             |
|                                                                              |
| Returns text from the clipboard by calling the function data with the format |
| IClipboard::textFormat.  The function data will throw an exception if the    |
| clipboard has no text format data.                                           |
| Note: The function data will open the clipboard if it is closed (and leave   |
|       it open).                                                              |
------------------------------------------------------------------------------*/
IString IClipboard::text( )
{
  IFUNCTRACE_DEVELOP();

  // Note: We cannot close the clipboard until we make
  //       a copy of the text.
  IString text = IString((char*)data(textFormat));

  fClipboardImp->autoClose();

  return text;

}

/*------------------------------------------------------------------------------
| IClipboard::bitmap                                                           |
|                                                                              |
| Returns a copy of a bitmap on the clipboard by calling the function data with|
| the format IClipboard::bitmapFormat.  The function data will throw an        |
| exception if the clipboard has no bitmap format data.                        |
| Note: The function data will open the clipboard if it is closed (and leave   |
|       it open).                                                              |
------------------------------------------------------------------------------*/
IBitmapHandle IClipboard::bitmap( )
{
  IFUNCTRACE_DEVELOP();
  // Note: We cannot close the clipboard until we make
  //       a copy of the bitmap.
  void* bitmapData = this->data(IClipboard::bitmapFormat);
  

  IBitmapHandle bitmapHandle((IBitmapHandle::Value)bitmapData);
  IGImage       imageCopy (bitmapHandle);

  fClipboardImp->autoClose();

  return imageCopy.handle();

}


/*------------------------------------------------------------------------------
| IClipboard::data                                                             |
|                                                                              |
| Returns data of all pointer types from the clipboard.  Throws an             |
| IInvalidRequest exception if the clipboard doesn't have the requested        |
| format of data (avoid by calling hasData first).                             |
| Note: The function data will open the clipboard if it is closed (and leave   |
|       it open).                                                              |
------------------------------------------------------------------------------*/
void* IClipboard::data( const char* aFormat )
{
 IFUNCTRACE_DEVELOP();
 ITRACE_DEVELOP(IString("Format=")+aFormat);


 unsigned long bytesReturned = 0;  // Not currently supported.
 void* pData = (void*)fClipboardImp->retrieveData(aFormat, bytesReturned);

 if (!pData)
    ITHROWLIBRARYERROR(IC_NOCLIPBOARDDATA,
                       IBaseErrorInfo::invalidRequest,
                       IException::recoverable);

 return pData;
}

/*------------------------------------------------------------------------------
| IClipboard::primaryFormat                                                    |
|                                                                              |
| Returns the primary Format of data on the clipboard.  The primary format     |
| is the first format of data entered after the clipboard is emptied.          |
------------------------------------------------------------------------------*/
IString IClipboard::primaryFormat( )
{

  IFUNCTRACE_DEVELOP();
  
  IClipboard clipboard;

  /*----------------------------------------------------------
  | We need a cursor. The cursor needs an open clipboard
  | instance for its core Windows API to operate.
  | Clipboard isn't open because we just instantiated it
  | in a static function. It could be open elsewhere. In
  | that case we throw an exception and return a "" string.
  -----------------------------------------------------------*/
  clipboard.open();

  IString result("");
  if (clipboard.isOpen())
  {
   IClipboard::Cursor cursor(clipboard);
   if (cursor.setToFirst())
      result = clipboard.fClipboardImp->formatForItem(
                              cursor.fClipboardCursorData->fItemHandle );
   clipboard.close();
  }
  else
    ITHROWLIBRARYERROR(IC_NOCLIPBOARDDATA,
                       IBaseErrorInfo::invalidRequest,
                       IException::recoverable);

  return result; // Null if clipboard in use or empty.
}

/*------------------------------------------------------------------------------
| IClipboard::format                                                           |
|                                                                              |
| Returns a format at the cursor location.                                     |
------------------------------------------------------------------------------*/
IString IClipboard::format ( const IClipboard::Cursor& cursor)
{
  IFUNCTRACE_DEVELOP();

  IASSERTPARM(cursor.isValid());

  return cursor.fClipboard.fClipboardImp->formatForItem
                      (cursor.fClipboardCursorData->fItemHandle);
}

/*------------------------------------------------------------------------------
| IClipboard::format                                                           |
|                                                                              |
| Returns a format for the passed format handle.                               |
------------------------------------------------------------------------------*/
IString IClipboard::format ( const IClipboard::FormatHandle& handle)
{
  IClipboard clipboard;
  return clipboard.fClipboardImp->formatForAtom(handle);
}

/*------------------------------------------------------------------------------
| IClipboard::registerFormat                                                   |
|                                                                              |
| Registers a new Format of clipboard data.  The handle of the format is       |
| returned.  In OS/2, the format is added to the system atom table.  Windows   |
| has a special clipboard registration function.                               |
------------------------------------------------------------------------------*/
IClipboard::FormatHandle IClipboard::registerFormat ( 
                                   const char*   privateFormat,
                                   EElementSize  elementSize)
{
  IFUNCTRACE_DEVELOP();
  ITRACE_DEVELOP(IString("Format =")+privateFormat);

  IClipboard clipboard;
  unsigned long format =
               clipboard.fClipboardImp->
                  registerFormat( privateFormat, elementSize);

  if (!format)
    ITHROWLIBRARYERROR(IC_REGISTERCLIPBOARDFORMATFAILED,
                       IBaseErrorInfo::accessError,
                       IException::recoverable);

  if (!format)
    ITHROWLIBRARYERROR(IC_REGISTERCLIPBOARDFORMATFAILED,
                       IBaseErrorInfo::accessError,
                       IException::recoverable);
  return (FormatHandle)format;
}

/*------------------------------------------------------------------------------
| IClipboard::formatAsHandle                                                   |
|                                                                              |
| Return the handle of a clipboard format.                                     |
------------------------------------------------------------------------------*/
IClipboard::FormatHandle IClipboard::formatAsHandle ( const char* formatName)
{
  IClipboard clipboard;
  FormatHandle formatHandle =
                 clipboard.fClipboardImp->atomFromName(formatName);

  if(!formatHandle)
     ITHROWLIBRARYERROR(IC_CLIPBOARDFORMATNOTFOUND,
                        IBaseErrorInfo::invalidRequest,
                        IException::recoverable);

  return formatHandle;
}


/*------------------------------------------------------------------------------
| IClipboard::formatCount                                                      |
|                                                                              |
| Returns the number of clipboard formats.                                     |
------------------------------------------------------------------------------*/
unsigned long IClipboard::formatCount ( ) const
{
  IFUNCTRACE_DEVELOP();
  return ((IClipboardImp*)fClipboardImp)->itemCount();
}



/*------------------------------------------------------------------------------
| IClipboard::owner                                                            |
|                                                                              |
| Returns the current clipboard owner.                                         |
|                                                                              |
| Note: The clipboard owner is provided on the open function but ownership     |
|       is not established until the call to empty.  This function returns     |
|       the real clipboard owner (e.g. if you call this function after calling |
|       open but before calling empty, you will receive the prior owner ( or   |
|       perhaps none).                                                         |
------------------------------------------------------------------------------*/
IWindowHandle IClipboard::owner( ) const
{
  IFUNCTRACE_DEVELOP();

  return fClipboardImp->realOwner();

}


/*------------------------------------------------------------------------------
| IClipboard::setData                                                          |
|                                                                              |
| Function to place memory data into the clipboard. If necessary (PMWIN), we   |
| allocate shared memory and copy in the new data.                             |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::setData( const char*        formatName,
                                 const void*        data,
                                 unsigned long      dataLength)
{
  IFUNCTRACE_DEVELOP();
  
  ITRACE_DEVELOP(IString("Format=")+formatName);

  fClipboardImp->copyData(formatName, (const char*)data, dataLength);

  return *this;
}

/*------------------------------------------------------------------------------
| IClipboard::setHandle                                                        |
|                                                                              |
| Function to place a handle into the clipboard. It is the caller's            |
| responsibility to copy the data that the handle references into shared       |
| memory so that the paste process can find it.                                |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::setHandle( const char*     formatName,
                                   unsigned long   data)
{
  IFUNCTRACE_DEVELOP();

  fClipboardImp->copyHandle(formatName, data);

   return *this;
}


/*------------------------------------------------------------------------------
| IClipboard::hasData                                                          |
|                                                                              |
| Function to return true if the requested format of data is on the            |
| clipboard.                                                                   |
------------------------------------------------------------------------------*/
bool IClipboard::hasData( const char* formatName) const
{
  IFUNCTRACE_DEVELOP();

  return fClipboardImp->clipboardHasFormat(formatName);
}


/*------------------------------------------------------------------------------
| IClipboard::Cursor::Cursor                                                   |
|                                                                              |
| The constructor simply saves the clipboard object and initializes the cursor |
| state to an "invalid" state.                                                 |
------------------------------------------------------------------------------*/
IClipboard::Cursor::Cursor ( IClipboard& clipboardIn)
  : fClipboard ( clipboardIn ),
    fClipboardCursorData( new IClipboardCursorData )
{
   IFUNCTRACE_DEVELOP();
}

/*------------------------------------------------------------------------------
| IClipboard::~ChildCursor                                                     |
|                                                                              |
| Currently empty.                                                             |
------------------------------------------------------------------------------*/
IClipboard::Cursor::~Cursor ( )
{
   IFUNCTRACE_DEVELOP();
}

/*------------------------------------------------------------------------------
| IClipboard::Cursor::setToFirst                                               |
|                                                                              |
| Set to the first format in the clipboard.                                    |
------------------------------------------------------------------------------*/
bool IClipboard::Cursor::setToFirst ( )
{
  IFUNCTRACE_DEVELOP();
  fClipboardCursorData->fItemHandle =
        fClipboard.fClipboardImp->nextItemHandle(0);

  return isValid();
}

/*------------------------------------------------------------------------------
| IClipboard::Cursor::setToNext                                                |
|                                                                              |
| Move the cursor to the next format.                                          |
------------------------------------------------------------------------------*/
bool IClipboard::Cursor::setToNext ( )
{
  IFUNCTRACE_DEVELOP();
  fClipboardCursorData->fItemHandle =
        fClipboard.fClipboardImp->nextItemHandle (fClipboardCursorData->fItemHandle);

  return isValid();
}

/*------------------------------------------------------------------------------
| IClipboard::Cursor::isValid                                                  |
|                                                                              |
| Return true if current item handle represents a valid item. Because          |
| IClipboardImp->nextItemHandle always returns 0 if the item is not valid,     |
| and this is stored in fItemHandle, a positive fItemHandle indicates a        |
| valid cursor.                                                                |
------------------------------------------------------------------------------*/
bool IClipboard::Cursor::isValid ( ) const
{
  return (fClipboardCursorData->fItemHandle!=0);
}

/*------------------------------------------------------------------------------
| IClipboard::Cursor::invalidate                                               |
|                                                                              |
| Invalidate the cursor by making fItemHandle 0.                               |
------------------------------------------------------------------------------*/
void IClipboard::Cursor::invalidate ( )
{
  fClipboardCursorData->fItemHandle = 0;;
}



/*------------------------------------------------------------------------------
| OS/2 Only functions                                                          |
------------------------------------------------------------------------------*/
#ifdef IC_PM
/*------------------------------------------------------------------------------
| IClipboard::setOwner                                                         |
|                                                                              |
| Identifies a new owner for the clipboard.                                    |
|                                                                              |
| Note:  OS/2 only function.                                                   |
------------------------------------------------------------------------------*/
IClipboard& IClipboard::setOwner ( )
{
  IFUNCTRACE_DEVELOP();
  ((IOS2ClipboardImp*)fClipboardImp)->setRealOwner();

  return *this;
}

#endif // IC_PM



